/*	common/ffp_negotiate.c - modified to run in user space
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file defines the functions used in the Full Feature phase
 *	by the iSCSI initiator and target for parameter negotiation
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/

#ifdef __KERNEL__

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/sched.h>

#else

#include "user_space.h"

#endif

#include "iscsi_common.h"
#include "debug.h"
#include "my_memory.h"

#include "text_param.h"
#include "ffp_negotiate.h"


__u32 __attribute__ ((no_instrument_function))
    give_text_overflow_message(__u32 newlen, __u32 outmax)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE_ERROR("%s %u text bytes exceeds buffer max %u\n",
		    current->comm, newlen, outmax);
	return REASON_OUT_OF_RESOURCES;
}

/*
 *	Returns 0 if ok, else reason code > 0 for reject
 */
static __u32 __attribute__ ((no_instrument_function))
    append_key_value(const char *key, const char *val_string, char *outbuf,
		 __u32 * outlen, __u32 outmax)
{
	__u32 keylen, vallen, newlen, reason = 0;

	keylen = strlen(key);
	vallen = strlen(val_string);
	newlen = *outlen + keylen + vallen + 2;	/* include the '=' and '\0' */
	if (newlen > outmax) {
		reason = give_text_overflow_message(newlen, outmax);
	} else {
		sprintf(outbuf + *outlen, "%s=%s", key, val_string);
		*outlen = newlen;
	}
	return reason;
}

/*
 *	parses data segment sent in a Text Request or Response in FFP
 *	When called,
 *		"inbuf" holds "inlen" bytes of data to be parsed
 *		"p_param_tbl" holds the table of known keys
 *		"discovery_session" is non-zero if this a Discovery Session
 *		"who_sent_it" is USE_BY_INITIATOR if "inbuf" was sent by Initiator,
 *						 USE_BY_TARGET if "inbuf" was sent by Target
 *		"outbuf" is an empty buffer that can hold up to "outmax" bytes
 *				 (or NULL if no reply is being generated)
 *
 *	Returns 0 if ok, else reason code > 0 for reject
 *	Also on return,
 *		if not NULL, "outbuf" has been filled with a response
 *		if not NULL, "length" has been filled with size of response in "outbuf"
 *		"new_max_send_length" is non-zero if MaxRecvDataSegmentLength was recved
 *		if not NULL, "val_string" points at 1st char of string following 'key='
 *  	if not NULL, "result" holds:
 *			TEXT_EMPTY		if no data segment
 *			TEXT_OTHER		if any key other than SendTargets
 *			TEXT_ST_ALL		if only "SendTargets=ALL" key
 *			TEXT_ST_TN		if only "SendTargets=<target-name>" key
 *			TEXT_ST_NULL	if only "SendTargets=" key
 */
__u32
parse_text_buffer(char *inbuf, __u32 inlen,
		  struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		  __u32 discovery_session,
		  __u32 who_sent_it,
		  char *outbuf, __u32 outmax,
		  enum text_types * result,
		  __u32 * new_max_send_length, __u32 * length,
		  char **val_string)
{
	char *ptr, *ptr2, *equal, *endptr;
	enum text_types text_type = TEXT_EMPTY;
	struct parameter_type *param;
	__u32 value, remaining, len = 0, reason = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if ((ptr = inbuf)) {
		/* inlen bytes of text data were read into inbuf */
		/* once around next loop for each key in inbuf */
		for (remaining = inlen; remaining > 0; ptr = ptr2) {
			equal = NULL;
			/* once around next loop for each character in next key */
			for (ptr2 = ptr;
			     *ptr2 != '\0' && remaining > 0;
			     ptr2++, remaining--) {
				/* skip over key itself */
				if (*ptr2 == '=' && equal == NULL)
					equal = ptr2;
			}

			if (*ptr2 != '\0') {
				*ptr2 = '\0';
				goto bad_key;
			}

			if (equal == NULL) {
				goto bad_equal;
			}

			printk("%s %s %s\n", current->comm,
			       discovery_session ? discovery_word :
			       "Text negotiation", ptr);

			/* terminate the key name string and point at the value */
			*equal++ = '\0';
			if (val_string)
				*val_string = equal;	/* return pointer to value string */
			param = find_parameter(ptr, p_param_tbl);
			if (!param) {
				/* didn't find this (nonstandard) key in our table */
				if (discovery_session) {
					/* RFC 3720 Section 12.21. SessionType
					 * ...
					 * "The only requests a target accepts in this type of
					 * session (i.e., discovery) are a text request with a
					 * SendTargets key and a logout request with readons
					 * 'close the session'."
					 */
					goto bad_disco;
				}
				if (text_type != TEXT_EMPTY
				    && text_type != TEXT_OTHER) {
					goto bad_pair;
				}
				/* else normal FFP text negotiation, key is not understood */
				if (!outbuf) {
					/* not sending reply, give error and reset negotiations */
					goto bad_name;
				}
				if ((reason =
				     append_key_value(ptr,
						      notunderstood_word,
						      outbuf, &len,
						      outmax))) {
					break;
				}
				text_type = TEXT_OTHER;
			} else if (!(param->type & who_sent_it)) {
				goto bad_sender;
			} else if (param->
				   special_key_flag & SENDTARGETS_FLAG) {
				/* RFC 3720 Appendix D. SendTargets Operation
				 * "A SendTargets command consists of a single Text request
				 * PDU.  This PDU contains exactly one text key and value.
				 * The text key MUST be sendTargets."
				 */
				if (text_type != TEXT_EMPTY) {
					if (text_type == TEXT_OTHER)
						goto bad_other;
					goto bad_twice;
				}
				if (strcmp(equal, "All") == 0) {
					/* RFC 3720 Appendix D. Sendtargets Operation
					 * All
					 * ...
					 * This value MUST be supported on a discovery session,
					 * and MUST NOT be supported on an operational session."
					 */
					if (!discovery_session) {
						goto bad_normal;
					}
					text_type = TEXT_ST_ALL;
				} else if (*equal != '\0') {
					text_type = TEXT_ST_TN;
				} else {
					text_type = TEXT_ST_NULL;
				}
			} else if (text_type != TEXT_EMPTY
				   && text_type != TEXT_OTHER) {
				goto bad_pair;
			} else if (param->
				   special_key_flag &
				   MAXRECVDATASEGMENTLENGTH_FLAG) {
				/* the only non-informational key negotiable in FFP */
				if (discovery_session) {
					goto bad_disco;
				}
				if (*new_max_send_length) {
					goto bad_twice;
				}
				value = simple_strtoul(equal, &endptr, 0);
				if (strspn(endptr, WHITE_SPACE) !=
				    strlen(endptr)
				    || value < 512 || value > 16777215) {
					goto bad_value;
				}
				*new_max_send_length = value;
				text_type = TEXT_OTHER;
			} else if (param->
				   type & (INITIAL_ONLY | LEADING_ONLY)
				   || !(param->type & INFORMATIONAL_PARAM)) {
				/* eliminates almost all unacceptable keys */
				goto bad_ffp;
			} else if (discovery_session
				   && (param->special_key_flag
				       & (INITIATORALIAS_FLAG |
					  TARGETALIAS_FLAG))) {
				/* no xxxALIAS key can be sent in discovery session */
				goto bad_disco;
			} else if (who_sent_it & USE_BY_INITIATOR) {
				/* can only be TargetName or InitiatorAlias sent by Initiator
				 * initiator can't send TargetName in FFP
				 */
				if (param->
				    special_key_flag & TARGETNAME_FLAG) {
					goto bad_ffp;
				}
				/* key can only be InitiatorAlias */
				text_type = TEXT_OTHER;
			} else {
				/* can only be TargetName, TargetAlias, TargetAddress
				 * sent by Target
				 */
				text_type = TEXT_OTHER;
			}

			for (; *ptr2 == '\0' && remaining > 0;
			     ptr2++, remaining--) {
				/* skip over nulls between keys */
			}
		}
	}
      out:
	if (result)
		*result = text_type;
	if (length)
		*length = len;
	return reason;

      bad_key:
	TRACE_ERROR("%s Unterminated key %s\n", current->comm, ptr);
      bad_protocol_out:
	reason = REASON_PROTOCOL_ERR;
	goto out;

      bad_name:
	TRACE_ERROR("%s Unknown key %s\n", current->comm, ptr);
	goto bad_protocol_out;

      bad_equal:
	TRACE_ERROR("%s Missing '=' in key %s\n", current->comm, ptr);
	goto bad_protocol_out;

      bad_disco:
	TRACE_ERROR("%s %s not negotiable in %s Session\n",
		    current->comm, ptr, discovery_word);
	goto bad_protocol_out;

      bad_pair:
	TRACE_ERROR("%s %s received with %s\n", current->comm, ptr,
		    sendtargets_word);
	goto bad_protocol_out;

      bad_other:
	TRACE_ERROR("%s %s received with other key\n", current->comm, ptr);
	goto bad_protocol_out;

      bad_twice:
	TRACE_ERROR("%s %s received twice\n", current->comm, ptr);
	goto bad_protocol_out;

      bad_ffp:
	TRACE_ERROR("%s %s cannot be sent by %s in Full Feature Phase\n",
		    current->comm, ptr,
		    param->
		    type & USE_BY_INITIATOR ? "target" : "initiator");
	goto bad_protocol_out;

      bad_sender:
	TRACE_ERROR("%s %s cannot be sent by %s\n", current->comm, ptr,
		    param->
		    type & USE_BY_INITIATOR ? "target" : "initiator");
	goto bad_protocol_out;

      bad_normal:
	TRACE_ERROR("%s %s=%s not allowed in Normal Session\n",
		    current->comm, ptr, equal);
	goto bad_protocol_out;

      bad_value:
	TRACE_ERROR("%s Illegal value in %s=%s\n", current->comm, ptr,
		    equal);
	reason = REASON_NEGOTIATION_RESET;
	goto out;
}
