/*	common/my_memory.c - modified to run in user space
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This defines the generic functions used
 *	by both the iSCSI target and the iSCSI initiator for dynamic memory
 *	allocation.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#ifdef __KERNEL__

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>

/* for definition of in_interrupt() */
#include <linux/interrupt.h>

#else

#include "user_space.h"

#endif

#include "iscsi_common.h"
#include "debug.h"
#include "my_memory.h"

#ifdef ISCSI_CHECK_MEMORY

#define MEMORY_TABLE_SIZE	1048576

static void *my_memory_address_table[MEMORY_TABLE_SIZE];
static char *my_memory_id_table[MEMORY_TABLE_SIZE];
static int my_memory_last = -1;	/* currently the last used slot */
static int my_memory_first = 0;	/* currently 1st possible unused slot */
static int my_memory_max = -1;	/* maximum last used slot so far */
static spinlock_t unh_memory_lock = SPIN_LOCK_UNLOCKED;
#endif

/*	called to allocate memory and keep a record of it */
void *my_kmalloc(__u32 size, char *id)
{
	void *ptr;

	TRACE(TRACE_MY_MEMORY, "Enter my_kmalloc %u bytes for %s\n", size,
	      id);

	if ((ptr = kmalloc(size,
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			   ((in_interrupt() || irqs_disabled())
#else
			   (in_interrupt()
#endif
			    ? GFP_ATOMIC : GFP_KERNEL))) == NULL) {
		TRACE_ERROR("Cannot allocate %u bytes for %s\n", size, id);
		goto out;
	}
#ifdef ISCSI_CHECK_MEMORY
	{
		int i;
		unsigned long flags;

		/* search for an unused slot, starting with first possible empty one */
		spin_lock_irqsave(&unh_memory_lock, flags);
		for (i = my_memory_first; i < MEMORY_TABLE_SIZE; i++) {
			if (my_memory_address_table[i] == NULL) {
				my_memory_address_table[i] = ptr;
				my_memory_id_table[i] = id;
				if (my_memory_last < i) {
					my_memory_last = i;
					if (my_memory_max < i)
						my_memory_max = i;
				}
				/* first possible unused slot */
				my_memory_first = i + 1;
				spin_unlock_irqrestore(&unh_memory_lock,
						       flags);
				goto out;
			}
		}
		/* if loop finishes, table is full so we cannot track our allocations */
		spin_unlock_irqrestore(&unh_memory_lock, flags);
		TRACE_ERROR("my_memory_table full, %d slots\n",
			    MEMORY_TABLE_SIZE);

		/* BUG(); */
	}
#endif

      out:
	TRACE(TRACE_MY_MEMORY, "Leave my_kmalloc %p for %s\n", ptr, id);

	return ptr;
}

#ifdef ISCSI_CHECK_MEMORY
/*	release dynamically allocated memory pointed to by non-NULL ptr */
void _my_kfree(void **ptr, char *id)
{
	{
		int i;
		unsigned long flags;

		TRACE(TRACE_MY_MEMORY,
		      "Enter my_kfree %p, last %d, for %s\n", *ptr,
		      my_memory_last, id);

		/* seach table backwards to find slot containing this pointer */
		spin_lock_irqsave(&unh_memory_lock, flags);
		for (i = my_memory_last; i >= 0; i--) {
			if (my_memory_address_table[i] == *ptr) {
				/* found the slot containing pointer, NULL everything */
				kfree(*ptr);
				*ptr = NULL;
				my_memory_address_table[i] = NULL;
				if (my_memory_first > i) {
					my_memory_first = i;
				}
				while (my_memory_last >= 0 &&
				       my_memory_address_table
				       [my_memory_last]
				       == NULL) {
					if (my_memory_first >
					    my_memory_last) {
						my_memory_first =
						    my_memory_last;
					}
					my_memory_last--;
				}
				spin_unlock_irqrestore(&unh_memory_lock,
						       flags);
				goto out;
			}
		}
		spin_unlock_irqrestore(&unh_memory_lock, flags);
		TRACE_ERROR
		    ("trying to free non-malloced memory %p for %s\n",
		     *ptr, id);
		/* BUG(); */
	      out:

		TRACE(TRACE_MY_MEMORY,
		      "Leave my_kfree, last %d, for %s\n", my_memory_last,
		      id);
	}
}

/*	"prints" statistics on my memory usage into buf,
 *	returns no. of chars printed.
 */
int _print_my_kmemory(char *buf)
{
	if (buf == NULL)
		return
		    printk
		    ("iscsi Maximum of %d memory slots used out of %d\n",
		     my_memory_max + 1, MEMORY_TABLE_SIZE);
	else
		return sprintf(buf,
			       "Maximum of %d memory slots used out of %d\n",
			       my_memory_max + 1, MEMORY_TABLE_SIZE);
}


/*	called only once, at very end when module is unloaded,
 *	to check for any allocated but not freed memory
 */
void _my_kempty(void)
{
	TRACE((TRACE_ENTER_LEAVE | TRACE_MY_MEMORY),
	      "Enter my_kempty, last %d\n", my_memory_last);

	print_my_kmemory(NULL);

	while (my_memory_last >= 0) {
		if (my_memory_address_table[my_memory_last] != NULL) {
			TRACE_ERROR("Unfreed memory[%d] %p for %s\n",
				    my_memory_last,
				    my_memory_address_table
				    [my_memory_last],
				    my_memory_id_table[my_memory_last]);
			/***
			kfree(my_memory_address_table[my_memory_last]);
			***/
			my_memory_address_table[my_memory_last] = NULL;
		}
		my_memory_last--;
	}

	/*      everything should now be reset to initial state */
	my_memory_max = -1;
	my_memory_first = 0;

	TRACE((TRACE_ENTER_LEAVE | TRACE_MY_MEMORY),
	      "Leave my_kempty, last %d\n", my_memory_last);
}

/* !__KERNEL__ implementatation is in user_mem_cache.c */
#if defined(__KERNEL__)

/*****************************************
 * 
 * kmem_cache code
 *
 */

struct my_cache_alloc_node {
	/* the allocated object. */
	void *objp;

	/* list link. */
	struct list_head next;
};

struct my_kmem_cache {
	/* the kernel's memory cache. */
	struct kmem_cache *cachep;

	/* locks the alloc_lst. */
	spinlock_t lock;

	/* list of allocated objects. */
	struct list_head alloc_lst;

	/* the cache's name */
	const char *name;
};

my_kmem_cache_t my_kmem_cache_create(const char *name,
				     size_t size,
				     size_t align,
				     unsigned long flags)
{
	struct my_kmem_cache *p;

	p = my_kmalloc(sizeof(*p), (char*) name);
	if (!p) {
		TRACE_ERROR("%s: my_kmalloc failed\n", __func__);
		return NULL;
	}

	INIT_LIST_HEAD(&p->alloc_lst);
	spin_lock_init(&p->lock);
	p->name = name;
	p->cachep = kmem_cache_create(name, size, align, flags, NULL);

	return p;
}

void *my_kmem_cache_alloc(my_kmem_cache_t cachep, gfp_t flags)
{
	struct my_cache_alloc_node *n;
	long lockflags;

	n = my_kmalloc(sizeof(*n), (char*) cachep->name);
	if (!n) {
		TRACE_ERROR("%s: my_kmalloc failed\n", __func__);
		return NULL;
	}

	n->objp = kmem_cache_alloc(cachep->cachep, flags);
	if (!n->objp) {
		TRACE_ERROR("%s: kmem_cache_alloc failed\n", __func__);
		my_kfree((void**) &n, (char*) cachep->name);
		return NULL;
	}

	spin_lock_irqsave(&cachep->lock, lockflags);
	list_add(&n->next, &cachep->alloc_lst);
	spin_unlock_irqrestore(&cachep->lock, lockflags);

	return n->objp;
}

void my_kmem_cache_free(my_kmem_cache_t cachep, void *objp)
{
	struct my_cache_alloc_node *n;
	struct list_head *node;
	long flags;

	n = NULL;

	spin_lock_irqsave(&cachep->lock, flags);
	list_for_each(node, &cachep->alloc_lst) {
		n = list_entry(node, struct my_cache_alloc_node, next);
		if (n->objp == objp) {
			list_del(&n->next);
			break;
		}

	}
	spin_unlock_irqrestore(&cachep->lock, flags);

	if (!n || n->objp != objp) {
		TRACE_ERROR("%s: object=%p not allocated in cache=%s\n",
			    __func__, objp, cachep->name);
	} else {
		kmem_cache_free(cachep->cachep, objp);
	}
}

void my_kmem_cache_destroy(my_kmem_cache_t cachep)
{
	struct my_cache_alloc_node *n;
	struct list_head *node, *next;
	long flags;

	if (!list_empty(&cachep->alloc_lst)) {
		TRACE_ERROR("%s: called when cache=%s has outstanding "
			    "allocations\n", __func__, cachep->name);
		spin_lock_irqsave(&cachep->lock, flags);
		list_for_each_safe(node, next, &cachep->alloc_lst) {
			n = list_entry(node, struct my_cache_alloc_node, next);
			TRACE_ERROR("%s: outstanding object=%p\n", __func__,
				    n->objp);
			list_del(&n->next);
			my_kfree((void**) &n, (char*) cachep->name);
		}
		spin_unlock_irqrestore(&cachep->lock, flags);
	}

	my_kfree((void**) &cachep, (char*) cachep->name);
}


#endif /* __KERNEL__ */

#endif
