/*	common/my_memory.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This defines the generic functions used
 *	by both the iSCSI target and the iSCSI initiator for dynamic memory
 *	allocation.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/
#ifndef _MY_MEMORY_H
#define _MY_MEMORY_H

#if !defined(__KERNEL__)
/* these values don't matter in userspace, but we need the symbols defined. */
#define GFP_KERNEL 0
#define GFP_ATOMIC 0
#define SLAB_HWCACHE_ALIGN 0
#endif

#ifdef __KERNEL__

#include <linux/version.h>
#include <linux/slab.h>		/* kmem_cache */

/*	as defined in linux/slab.h */
extern void kfree(const void *);

#else

#define kmalloc(x,f)	malloc(x)
#define kfree(x)	free(x)

#endif

/* largest number of bytes kmalloc() will allocate as one block of memory */
#define MAX_KMALLOC_SIZE		131072

/*	called to allocate memory and keep a record of it */
void *my_kmalloc(__u32 size, char *id);


#ifdef ISCSI_CHECK_MEMORY

/*	called only by my_kfree() when *ptr is not NULL */
void _my_kfree(void **ptr, char *id);

#endif

/*	called to release memory allocated dynamically */
static inline void __attribute__ ((no_instrument_function))
    my_kfree(void **ptr, char *id)
{
	if (*ptr != NULL) {
#ifdef ISCSI_CHECK_MEMORY
		_my_kfree(ptr, id);
#else
		kfree(*ptr);
		*ptr = NULL;
#endif
	}
}

#ifdef ISCSI_CHECK_MEMORY
int _print_my_kmemory(char *buf);
#endif

/*	"prints" statistics on my memory usage into buf,
 *	returns no. of chars printed.
 */
static inline int __attribute__ ((no_instrument_function))
    print_my_kmemory(char *buf)
{
#ifdef ISCSI_CHECK_MEMORY
	return _print_my_kmemory(buf);
#else
	return 0;
#endif
}

#ifdef ISCSI_CHECK_MEMORY
void _my_kempty(void);
#endif

/*	called only once, at very end when module is unloaded,
 *	to check for any allocated but not freed memory
 */
static inline void __attribute__ ((no_instrument_function))
    my_kempty(void)
{
#ifdef ISCSI_CHECK_MEMORY
	_my_kempty();
#endif
}

/*****************************************
 *
 * kmem_cache wrappers
 *
 */

#if defined(ISCSI_CHECK_MEMORY) || !defined(__KERNEL__)
/** opaque struct defined in my_memory.c. */
struct my_kmem_cache;

/** type abstraction for kmem cache wrappers. */
typedef struct my_kmem_cache *my_kmem_cache_t;

/** create a new memcache. */
my_kmem_cache_t my_kmem_cache_create(const char *name,
				     size_t size,
				     size_t align,
				     unsigned long flags);

/** allocate an object from the given cache. */
void *my_kmem_cache_alloc(my_kmem_cache_t cachep, gfp_t flags);

/** frees an object back into the cache. */
void my_kmem_cache_free(my_kmem_cache_t cachep, void *objp);

/** destroys a mem_cache */
void my_kmem_cache_destroy(my_kmem_cache_t cachep);

#else /********************* ISCSI_CHECK_MEMORY *********************/

typedef struct kmem_cache *my_kmem_cache_t;

/** create a new memcache. */
static inline my_kmem_cache_t my_kmem_cache_create(const char *name,
						   size_t size,
						   size_t align,
						   unsigned long flags)
{
/* I am not sure if its 2.6.22 that got rid of destructors. */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22) 
	return kmem_cache_create(name, size, align, flags, NULL);
#else
	return kmem_cache_create(name, size, align, flags, NULL, NULL);
#endif
}

/** allocate an object from the given cache. */
static inline void *my_kmem_cache_alloc(my_kmem_cache_t cachep, gfp_t flags)
{
	return kmem_cache_alloc(cachep, flags);
}

/** frees an object back into the cache. */
static inline void my_kmem_cache_free(my_kmem_cache_t cachep, void *objp)
{
	kmem_cache_free(cachep, objp);
}

/** destroys a mem_cache */
static inline void my_kmem_cache_destroy(my_kmem_cache_t cachep)
{
	kmem_cache_destroy(cachep);
}

#endif /* ISCSI_CHECK_MEMORY */

#endif /* _MY_MEMORY_H */
