/**
 * \file queue.h
 * \brief Contains the header information for a generic ring queue that holds
 *        void*s.
 * \author Ethan Burns
 * \date 2007-06-19
 */

#ifndef __KERNEL__
#include "user_space.h"
#else
#include <linux/types.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#endif				/* __KERNEL__ */

#include "my_memory.h"
#include "debug.h"

struct queue {
	struct list_head head;
	spinlock_t lock;
};

/**
 * \brief A queue structure holds the information for a queue.
 */
struct queue_elem {
	void *data;
	struct list_head next;
};

/**
 * Initializes a queue.
 */
static inline void queue_init(struct queue *q)
{
	INIT_LIST_HEAD(&q->head);
	spin_lock_init(&q->lock);
}

/**
 * \brief Returns true if the queue is empty.
 * \param q The ring queue.
 * \return 0 if the queue is not empty, !0 if the queue is empty.
 */
static inline int queue_empty(struct queue *q)
{
	return list_empty(&q->head);
}

/**
 * \brief Adds an item to the tail of the queue.
 * \param q The queue to add to.
 * \param elem The element to add.
 * \return 0 on success, !0 on error.
 */
static inline int queue_enqueue(struct queue *q, void *elem)
{
	struct queue_elem *qelem;
	long flags;

	TRACE(TRACE_DEBUG, "%s: q=%p enqueuing %p\n", __func__, q, elem);

	qelem = my_kmalloc(sizeof(*qelem), "qelem");
	if (!qelem)
		return -1;
	qelem->data = elem;

	spin_lock_irqsave(&q->lock, flags);
	list_add_tail(&qelem->next, &q->head);
	spin_unlock_irqrestore(&q->lock, flags);

	return 0;
}

/**
 * \brief Remove the element on the front of the queue.
 * \param q The queue to remove from.
 * \return 0 on success, !0 on error.
 */
static inline int queue_dequeue(struct queue *q)
{
	struct queue_elem *qelem;
	long flags;


	spin_lock_irqsave(&q->lock, flags);
	qelem = container_of(q->head.next, struct queue_elem, next);
	list_del(q->head.next);
	spin_unlock_irqrestore(&q->lock, flags);

	TRACE(TRACE_DEBUG, "%s: q=%p dequeuing %p\n", __func__, q, qelem);

	my_kfree((void**) &qelem, "qelem");

	return 0;
}

/**
 * \brief Get the front element of the queue.
 * \param q The queue to get the front element from.
 * \param elem Buffer used to return the front element.
 * \return 0 on success, !0 on error.
 */
static inline int queue_front(struct queue *q, void **elem)
{
	struct queue_elem *qelem;
	long flags;

	if (queue_empty(q))
		return -1;

	spin_lock_irqsave(&q->lock, flags);
	qelem = container_of(q->head.next, struct queue_elem, next);
	*elem = qelem->data;
	spin_unlock_irqrestore(&q->lock, flags);

	TRACE(TRACE_DEBUG, "%s: q=%p fronting %p\n", __func__, q, *elem);

	return 0;
}

static inline void queue_free(struct queue *q)
{
	while (!list_empty(&q->head))
		queue_dequeue(q);
}

/* vi: set tabstop=8 textwidth=78: */
