/*	common/target_negotiate.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 * 	This defines the functions used in the Login phase
 * 	by the iSCSI target for parameter negotiation
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/
#ifndef	_TARGET_NEGOTIATE_H
#define	_TARGET_NEGOTIATE_H

#ifdef __KERNEL__

#include <asm/semaphore.h>

#else

#include "../common/user_semaphore.h"

#endif

#include "../common/queue.h"

/* max cmnds/session at target */
#define QUEUE_DEPTH_ALLOWED	4

/* names of bit numbers in the iscsi_conn control byte */
#define SILENCE_BIT		0
#define NEED_NOPIN_BIT		1

/* maximum number of jiffies to wait in ISCSI_SENT state before sending NopIn */
#define NOP_TIME_LIMIT		(HZ)

/* number of jiffies between each timer interrupt */
#define TIMER_PERIOD		(5*HZ)

/* maximum number of unreplied-to NopIns to send to initiator before aborting */
#define MAX_OUTSTANDING_NOPINS	6

/* maximum number of jiffies target gives initiator to reply to a NopIn ping */
#define PING_REPLY_TIME_LIMIT	(HZ/2)

/* maximum number of different IP_address:port listening sockets and threads
 * (used as size of iscsi_portal_groups[])
 */
#define MAX_PORTAL	32

#define CONN_DOWN			0
#define CONN_UP				1
#define CONN_SETUP_FAIL		-1

/* queue item to send a reject pdu back to initiator */
struct reject_item {
	struct list_head reject_link;
	__u8 bad_header[ISCSI_HDR_LEN];
	__u8 reason;
};

/* item to hold buffer for unsolicited data */
struct udata_list {
	__u32 offset;		/* offset of this buffer within FirstBurst */
	__u32 length;		/* number of bytes in this buffer */
	__u8 *buffer;		/* the unsolicited data itself */
#ifdef RDMA_MODE
	struct iser_recv_pdu_t *rx_iser_pdu;
#endif
	struct list_head udata_link;
};

/* global struct to store properties corresponding to an iscsi target */
struct iscsi_global {
	/* conn_id: id assigned to uniquely identify connections within session
	 *            initialized to 0 in init_target() when module is loaded
	 *            running counter that is never restored or configured,
	 *            copied to conn->conn_id in build_conn_sess()
	 */
	__u32 conn_id;

	/* ntsih: id assigned to target sessions
	 *            initialized to 0 in init_target() when module is loaded
	 */
	__u16 ntsih;

	/* targ_snack_flg: bits for Data and Status snack support
	 *            initialized to 0 in init_target() when module is loaded
	 *            bits restored/configured by iscsi_manage a= A= d= s= h=
	 *            copied to session->targ_snack_flg in build_conn_sess()
	 */
	__u8 targ_snack_flg;

	/* phase_collapse: 0 never, 1 always, 2 alternate
	 *            initialized to 0 in init_target() when module is loaded
	 *            restored/configured by iscsi_manage p=
	 *            copied to conn->phase_collapse in build_conn_sess()
	 */
	signed char phase_collapse;

	/* session_list: list of all active sessions known to target
	 *            initialized in init_target() when module is loaded
	 *            never restored or configured or copied
	 *            new sessions added to this list in handle_login()
	 *            old sessions removed from this list in iscsi_release_session()
	 */
	struct list_head session_list;

	/* bad_session_list: list of sessions that failed during startup
	 *            initialized in init_target() when module is loaded
	 *            never restored or configured or copied
	 *            new sessions added to this list in handle_login()
	 *            old sessions removed from this list in iscsi_release_session()
	 */
	struct list_head bad_session_list;

	/* session_rwlock: read-write spin_lock to add/remove sessions from both
	 *            session lists, and to control access to conn_list
	 *            initialized to UNLOCKED in init_target() when module is loaded
	 *            never restored or configured or copied
	 */
	rwlock_t session_rwlock;

	/* server_thr: pointers to task struct for the active server threads
	 *            initialized to NULL in init_target() when module is loaded
	 *            never restored or configured or copied
	 *            new threads are added in iscsi_server_thread() as they come up
	 *            old threads are NULLed in iscsi_server_thread() as they go down
	 */
	struct task_struct *server_thr[MAX_PORTAL];

	/* server_socket: socket on which the corresponding server thread listens
	 *            initialized to NULL in init_target() when module is loaded
	 *            never restored or configured or copied
	 *            new sockets are added in bring_up_portal() as they are created
	 *            old sockets are NULLed in free_sock_data() as they are released
	 */
	struct socket *server_socket[MAX_PORTAL];

	/* server_sem: semaphore to wait on after creating/killing a server thread
	 *            initialized to LOCKED in init_target() when module is loaded
	 *            never restored or configured or copied
	 */
	struct semaphore server_sem;

	/* device: device returned by the STML */
	struct STD *device;

	/* force: manageable bits to initialize connection flags
	 *            initialized to 0 in init_target() when module is loaded
	 *            bits restored/configured by iscsi_manage s o r xok tk1
	 *                                                    iseroff rdmaLogin
	 *            copied to conn->connection_flags in build_conn_sess()
	 */
	__u32 force;

	/* nop_period: timeout period to send periodic NopIns to initiator
	 *            initialized to 0 in init_target() when module is loaded
	 *            bits restored/configured by iscsi_manage n=
	 *            copied to conn->nop_period in build_conn_sess()
	 */
	__u32 nop_period;

	/* r2t_period: r2t retransmit period
	 *            initialized to 0 in init_target() when module is loaded
	 *            bits restored/configured by iscsi_manage r2tp=
	 *            copied to session->r2t_period in build_conn_sess()
	 */
	__u32 r2t_period;

	/* max_queue_depth_allowed: the largest command window target shows to ini
	 *            init to QUEUE_DEPTH_ALLOWED in init_target() when module loaded
	 *            bits restored/configured by iscsi_manage Q=
	 *            copied to session->max_queue_depth_allowed in build_conn_sess()
	 */
	__u32 max_queue_depth_allowed;

	/* param_tbl: text parameters accepted by the target
	 *            allocated and initialized in init_target() when module is loaded
	 *            freed in release_portals() when module is unloaded and in
	 *            handle_login() after logging in another conn to existing session
	 *            restored/items configured by iscsi_manage key=value
	 *            copied to session->session_params in build_conn_sess()
	 */
	struct parameter_type (*param_tbl)[MAX_CONFIG_PARAMS];

	/* auth_parameter: context parameters for chap and srp
	 *            allocated and initialized in init_target() when module is loaded
	 *            restored/configured by iscsi_manage t b px= pn= lx= ln= cl=
	 *                                                    sx= sn=
	 *            freed in release_portals() when module is unloaded
	 *            cloned and freed locally in handle_login()
	 */
	struct auth_parameter_type auth_parameter;

#ifdef RDMA_MODE
	/* Added for iSER - Minti */
	struct iser_api_t iser_api;
	struct iser_ext_api_t iser_ext_api;
	int iser_conn_ctr;	/* Integer that increments for every iSER connection
				 * across sessions
				 */
#endif
};

/* stores everything related to an iscsi connection on the target */
struct iscsi_conn {
	/* conn_link: structure for threading this conn onto lists */
	struct list_head conn_link;

	/* conn_id: the id for this connection */
	int conn_id;

	/* cid: connection id assigned by the Initiator */
	__u16 cid;

	/* portal_group_tag: target_portal_group_tag for the connection */
	__u16 portal_group_tag;

	/* pg_ptr: pointer to entry in iscsi_portal_groups[] table
	 * for target's iscsi_server_thread */
	struct portal_group *pg_ptr;

	/* conn_socket: socket used by this connection */
	struct socket *conn_socket;

	/* session: what session does this connection belong to */
	struct iscsi_session *session;

	/* dev: device on which this connection was received */
	struct iscsi_global *dev;

	/* rx_thread: task struct for rx_thread */
	struct task_struct *rx_thread;

	/* tx_thread: task struct for tx_thread */
	struct task_struct *tx_thread;

	/* tx_sem: semaphore to control operation of tx_thread */
	struct semaphore tx_sem;

	/* semaphores to kill rx and tx threads */
	struct semaphore kill_rx_sem;
	struct semaphore kill_tx_sem;

	/* spinlock to avoid a race in iscsi_release_socket */
	spinlock_t sock_lock;

	/* reject_list: list of rejects to be sent back to initiator */
	struct list_head reject_list;

	/* reject_lock: spinlock to add/remove reject items */
	spinlock_t reject_lock;

	/* bad_hdr: complete header of bad PDU for reject */
	__u8 bad_hdr[ISCSI_HDR_LEN];

	/* hdr_crc: 1 if header digest in use, else 0 */
	int hdr_crc;

	/* data_crc: 1 if data digest in use, else 0 */
	int data_crc;

	/* active: 1 if connection is in use, else 0 */
	__u8 active;

	/* phase_collapse: 0 never, 1 always, 2 this time, -1 next time */
	signed char phase_collapse;

	__u8 dummy1;		/* to keep 4-byte alignment in structure */
	__u8 dummy2;		/* to keep 4-byte alignment in structure */

	/* control: bits to control tx thread wakeups, sending NopIn pings */
	unsigned long control;

	/* nop_period: timeout period to send periodic NopIns to initiator */
	__u32 nop_period;
	struct timer_list nop_timer;

	/* nop_period_remaining: time left until send next periodic NopIn */
	__u32 nop_period_remaining;

	/* time_stamp: jiffies of last send on this connection */
	unsigned long time_stamp;

	/* time_last_nopin_sent: jiffies when last nopin was sent on this conn */
	unsigned long time_last_nopin_sent;

	/* outstanding_nopins: number of NopIns sent but not yet replied to */
	atomic_t outstanding_nopins;

	/* connection_flags: bits to control stuff during login and FFP */
	__u32 connection_flags;
#ifdef ISER
	/* Together indicate when the put_data is complete */
	struct semaphore data_in_sem;
	__u32 data_in_complete;
#endif
	/* text_in_progress: ptr to text command in-progress, else NULL */
	void *text_in_progress;

	/* text_in_progress_lock: spinlock to add/remove text_in_progress */
	spinlock_t text_in_progress_lock;

	/* stat_sn: target's StatSN value.  Incremented only by conn's tx thread
	 * as part of sending a response pdu that uses up a StatSN value.  This
	 * is done without the protection of a lock, but it is safe because
	 * there is only one instance of a tx thread for each connection.
	 */
	__u32 stat_sn;

	__u32 max_send_length;	/*initiator's MaxRecvPDULength */
	__u32 max_recv_length;	/*target's MaxRecvPDULength */

	/* fields to keep track of ip addresses */
	struct sockaddr *ip_address;	/* initiator's ipv4/ipv6 address */
	struct sockaddr *local_ip_address;	/* target's ipv4/ipv6 address */

	struct generic_pdu inputpdu;	/* receives input pdus during login */
	struct generic_pdu outputpdu;	/* builds output pdus during login */

	int conn_status;	/* Replaces checks on conn_socket to see if conn is up
				   because conn_socket may be NULL in iSER mode */

#ifdef RDMA_MODE
	/* ringq of received PDUs */
	struct queue rx_pdu_q;
	/* ringq of received iser_task_h handles */
	struct queue rx_handle_q;

	struct semaphore iser_notify_sem;
	struct semaphore iser_conn_done_sem;
	struct semaphore send_done_sem;
	iser_api_handle_t iser_handle;
	int conn_terminated;	/* Set in conn_terminate_notify */
#endif
};



/* stores everything related to an iscsi session on the target */
struct iscsi_session {
	/* sess_link: structure for threading this session onto lists */
	struct list_head sess_link;

	/* isid: id assigned to a session by the Initiator */
	__u8 isid[6];

	/* tsih: id assigned to a session by the Target */
	__u16 tsih;

	/* portal_group_tag: target_portal_group_tag for the session */
	__u16 portal_group_tag;

	/* nconn: no of active connections */
	__u16 nconn;

	/* cmnd_id: these are the values assigned as target_tags */
	__u32 cmnd_id;

	/* free_command_list: list of free command structures for session */
	struct list_head free_command_list;

	/* active_command_list: list of active command structures for session */
	struct list_head active_command_list;

	/* conn_lock: spinlock to add/remove commands and reject items */
	spinlock_t cmnd_lock;

	/* command_list_size: number of items in free/active_command_lists */
	__u32 command_list_size;

	/* active_seqno: id assigned to each active command, reset when free */
	__u32 active_seqno;

	/* conn_list: list of connections within session
	 *      access to this list is under protection of the session_sem
	 *      it can be read only when session_list is read protected
	 *      it can be written only when session_list is write protected
	 */
	struct list_head conn_list;

	/* unsolicited_data_list: list of free buffers for unsolicited data */
	struct list_head unsolicited_data_list;

	/* unsolicited_data_lock: spinlock to access unsolicited_data_list */
	spinlock_t unsolicited_data_lock;

	/* unsolicited_list_size: number of items in unsolicited_data_list */
	__u32 unsolicited_list_size;

	/* unsolicited_buffer_size: number of bytes in one unsolicited buffer */
	__u32 unsolicited_buffer_size;

	/* devdata: pointer to the device specific data */
	struct iscsi_global *devdata;

	/* Added the r2t retransmit timer - SAI */
	__u32 r2t_period;
	struct timer_list *r2t_timer;

	/* the largest command window target shows to ini */
	__u32 max_queue_depth_allowed;

	/* error recovery ver ref18_04 */
	struct task_struct *retran_thread;
	struct semaphore retran_sem;
	struct semaphore thr_kill_sem;

	/* For parameters */
	struct parameter_type (*session_params)[MAX_CONFIG_PARAMS];
	struct session_operational_parameters *oper_param;

	/* version identifiers */
	__u8 version_min;
	__u8 version_max;
	__u8 version_active;

	/* SNACK flags for Error Recovery - SAI */
	__u8 targ_snack_flg;

	/*
	 * SESSION WIDE COUNTERS: exp_cmd_sn and max_cmd_sn
	 */
	__u32 cmd_sn;
	__u32 exp_cmd_sn;
	__u32 max_cmd_sn;

	/* command ordering variables - SAI */
	struct semaphore cmd_order_sem;
	struct order_cmd *cmd_order_head;
	struct order_cmd *cmd_order_tail;
};


/*************************************************************************
 * This function sends a login response with proper status code when an error
 * occurs in target during login phase
 * Returns 0 if login response sent ok, <0 if hit error during the send
 ************************************************************************/
int
login_reject(struct iscsi_conn *conn, int class, int detail,
	     struct generic_pdu *outputpdu);


void __attribute__ ((no_instrument_function))
    print_isid_tsih_message(struct iscsi_session *session, char *message);


int parameter_negotiate(struct iscsi_conn *conn,
			struct parameter_type
			p_param_tbl[MAX_CONFIG_PARAMS],
			struct generic_pdu *inputpdu,
			struct generic_pdu *outputpdu, __u32 when_called,
			struct auth_parameter_type auth_param);


/* set back the leading only keys if these keys were set to be
 * " KEY_TO_BE_NEGOTIATED" in the leading connection negotation.*/
void reset_parameter_table(struct parameter_type
			   p_param_tbl[MAX_CONFIG_PARAMS]);


/*
 * iscsi_release_session: This function is responsible for closing out
 * a session and removing it from whatever list it is on.
 * INPUT: session to be released
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
int iscsi_release_session(struct iscsi_session *session);

#endif
