/* 	common/text_param.c - modified to run in user space
 *
 * 	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *  	This defines the generic functions used in the Login phase
 *  	by both the iSCSI target and the iSCSI initiator for
 *  	parameter negotiation.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#ifdef __KERNEL__

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/ip.h>
#include <linux/net.h>
#include <asm/uaccess.h>

#else

#include "user_space.h"

#endif

#include "iscsi_common.h"
#include "debug.h"
#include "my_memory.h"

/* chap and srp support -CHONG */
#include "../security/chap/chap.h"
#include "../security/srp/srp.h"

#include "text_param.h"


/* Constant strings that are referenced from several places
 * but are only stored once anywhere in memory.
 */
const char targetname_header_word[] = TARGETNAME_HEADER;
const char targetname_word[] = TARGETNAME;
const char initiatorname_word[] = INITIATORNAME;
const char targetportalgrouptag_word[] = TARGETPORTALGROUPTAG;
const char sendtargets_word[] = SENDTARGETS;
const char targetaddress_word[] = TARGETADDRESS;
const char maxrecvdatasegmentlength_word[] = MAXRECVDATASEGMENTLENGTH;
const char maxburstlength_word[] = MAXBURSTLENGTH;
const char firstburstlength_word[] = FIRSTBURSTLENGTH;
const char none_word[] = NONE;
const char yes_word[] = YES;
const char no_word[] = NO;
const char reject_word[] = REJECT;
const char irrelevant_word[] = IRRELEVANT;
const char normal_word[] = NORMAL;
const char discovery_word[] = DISCOVERY;
const char notunderstood_word[] = NOTUNDERSTOOD;

/* These sizes include the final '\0', so they are 1 bigger than strlen() */
const int targetname_header_size = sizeof(targetname_header_word);
const int targetname_size = sizeof(targetname_word);

/* Store the information about the configurable parameters in this array.
 */
static struct parameter_type config_params[MAX_CONFIG_PARAMS] = {
	{			/* RFC 3720 Section 12.4 TargetName */
	      parameter_name:targetname_word,
	      type:
	 ALL | INFORMATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | STRING | ISCSI_NAME,
	      int_value:0,	/* exceptional use of this field by a string */
	      str_value:NULL,
	      value_list:NULL,
	      special_key_flag:TARGETNAME_FLAG,
	 },
	{			/* RFC 3720 Section 12.5 InitiatorName */
	      parameter_name:initiatorname_word,
	      type:
	 INITIAL_ONLY | INFORMATIONAL_PARAM |
	 USE_BY_INITIATOR | KEY_SINGLE_VALUE | STRING | ISCSI_NAME,
	      str_value:NULL,
	      value_list:NULL,
	      special_key_flag:INITIATORNAME_FLAG,
	 },
	{			/* RFC 3720 Section 12.6 TargetAlias */
	      parameter_name:TARGETALIAS,
	      type:
	 ALL | INFORMATIONAL_PARAM |
	 USE_BY_TARGET | KEY_SINGLE_VALUE | STRING | UTF_8,
	      str_value:NULL,
	      value_list:NULL,
	 },
	{			/* RFC 3720 Section 12.7 InitiatorAlias */
	      parameter_name:INITIATORALIAS,
	      type:
	 ALL | INFORMATIONAL_PARAM |
	 USE_BY_INITIATOR | KEY_SINGLE_VALUE | STRING | UTF_8,
	      str_value:NULL,
	      value_list:NULL,
	 },
	{			/* RFC 3720 Section 12.8 TargetAddress */
	      parameter_name:targetaddress_word,
	      type:
	 ALL | INFORMATIONAL_PARAM |
	 USE_BY_TARGET | KEY_SINGLE_VALUE | STRING | TARGET_ADDRESS_TYPE,
	      str_value:NULL,
	      value_list:NULL,
	      special_key_flag:TARGETADDRESS_FLAG,
	 },
	{			/* RFC 3720 Section 12.9 TargetPortalGroupTag */
	      parameter_name:targetportalgrouptag_word,
	      type:
	 INITIAL_ONLY | INFORMATIONAL_PARAM |
	 USE_BY_TARGET | KEY_SINGLE_VALUE | NUMBER | ZERO_TO_65535,
	      special_key_flag:TARGETPORTALGROUPTAG_FLAG,
	 },
	{			/* RFC 3720 Section 12.21 SessionType */
	      parameter_name:SESSIONTYPE,
	      type:
	 LEADING_ONLY | INFORMATIONAL_PARAM |
	 USE_BY_INITIATOR |
	 KEY_SINGLE_VALUE | ENUMERATED | DISCOVERY_NORMAL,
	      str_value:NORMAL,
	      value_list:NORMAL,
	      special_key_flag:SESSIONTYPE_FLAG,
	 },
#ifdef ISER
	/* Added for iSER - Minti
	 * This key should precede any other login keys which may
	 * be affected by the outcome of the negotiation of the RDMAExtensions
	 * key. Sec 8.3
	 */
	{			/* draft-ietf-ips-iser-05.txt Section 6.3 RDMAExtensions */
	      parameter_name:RDMAEXTENSIONS,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | BOOL_AND,
	      str_value:YES,
	      value_list:YES,
	      special_key_flag:RDMAEXTENSIONS_FLAG,
	 },
#endif
	{			/* RFC 3720 Section 12.1 Headerdigest and DataDigest */
	      parameter_name:HEADERDIGEST,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITH_RDMA |
	 KEY_MULTI_VALUE | ENUMERATED | DIGEST_PARAM,
	      str_value:NONE,
	      value_list:NONE,
	      special_key_flag:HEADERDIGEST_FLAG,
	 },
	{			/* RFC 3720 Section 12.1 Headerdigest and DataDigest */
	      parameter_name:DATADIGEST,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITH_RDMA |
	 KEY_MULTI_VALUE | ENUMERATED | DIGEST_PARAM,
	      str_value:NONE,
	      value_list:NONE,
	      special_key_flag:DATADIGEST_FLAG,
	 },
	{			/* RFC 3720 Section 12.2 MaxConnections */
	      parameter_name:MAXCONNECTIONS,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | NUMBER | ONE_TO_65535 | MIN_NUMBER,
	      int_value:2,
	      special_key_flag:MAXCONNECTIONS_FLAG,
	 },
	{			/* RFC 3720 Section 12.3 SendTargets */
	      parameter_name:sendtargets_word,
	      type:
	 FULL_FEATURE_ONLY |
	 USE_BY_INITIATOR | KEY_NO_VALUE | STRING | UTF_8,
	      str_value:NULL,
	      value_list:NULL,
	      special_key_flag:SENDTARGETS_FLAG,
	 },
	{			/* RFC 3720 Section A.3.1 OFMarker, IFMarker */
	      parameter_name:OFMARKER,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITH_RDMA |
	 KEY_SINGLE_VALUE | BOOL_AND,
	      str_value:NO,
	      value_list:NO,
	      special_key_flag:OFMARKER_FLAG,
	 },
	{			/* RFC 3720 Section A.3.1 OFMarker, IFMarker */
	      parameter_name:IFMARKER,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITH_RDMA |
	 KEY_SINGLE_VALUE | BOOL_AND,
	      str_value:NO,
	      value_list:NO,
	      special_key_flag:IFMARKER_FLAG,
	 },
	{			/* RFC 3720 Section A.3.2 OFMarkInt, IFMarkInt */
	      parameter_name:OFMARKINT,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_TWO_VALUE | NUMBER_RANGE | ONE_TO_65535,
	      int_value:2048,
	      str_value:"2048~2048",
	      value_list:"2048~2048",
	      special_key_flag:OFMARKINT_FLAG,
	 },
	{			/* RFC 3720 Section A.3.2 OFMarkInt, IFMarkInt */
	      parameter_name:IFMARKINT,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_TWO_VALUE | NUMBER_RANGE | ONE_TO_65535,
	      int_value:2048,
	      str_value:"2048~2048",
	      value_list:"2048~2048",
	      special_key_flag:IFMARKINT_FLAG,
	 },
	{			/* RFC 3720 Section 12.10 InitialR2T */
	      parameter_name:INITIALR2T,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | BOOL_OR,
	      str_value:YES,
	      value_list:YES,
	      special_key_flag:INITIALR2T_FLAG,
	 },
	{			/* RFC 3720 Section 12.11 ImmediateData */
	      parameter_name:IMMEDIATEDATA,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | BOOL_AND,
	      str_value:YES,
	      value_list:YES,
	      special_key_flag:IMMEDIATEDATA_FLAG,
	 },
	{			/* RFC 3720 Section 12.12 MaxRecvDataSegmentLength */
	      parameter_name:maxrecvdatasegmentlength_word,
	      type:
	 ALL | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | NUMBER | N512_TO_16777215,
	      int_value:8192,
	      special_key_flag:MAXRECVDATASEGMENTLENGTH_FLAG,
	 },
	{			/* RFC 3720 Section 12.13 MaxBurstLength */
	      parameter_name:maxburstlength_word,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | NUMBER | N512_TO_16777215 | MIN_NUMBER,
	      int_value:INITIAL_MAXBURSTLENGTH,
	      special_key_flag:MAXBURSTLENGTH_FLAG,
	 },
	{			/* RFC 3720 Section 12.14 FirstBurstLength */
	      parameter_name:firstburstlength_word,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | NUMBER | N512_TO_16777215 | MIN_NUMBER,
	      int_value:INITIAL_FIRSTBURSTLENGTH,
	      special_key_flag:FIRSTBURSTLENGTH_FLAG,
	 },
	{			/* RFC 3720 Section 12.15 DefaultTime2Wait */
	      parameter_name:DEFAULTTIME2WAIT,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | NUMBER | ZERO_TO_3600 | MAX_NUMBER,
	      int_value:2,
	      special_key_flag:DEFAULTTIME2WAIT_FLAG,
	 },
	{			/* RFC 3720 Section 12.16 DefaultTime2Retain */
	      parameter_name:DEFAULTTIME2RETAIN,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | NUMBER | ZERO_TO_3600 | MIN_NUMBER,
	      int_value:20,
	      special_key_flag:DEFAULTTIME2RETAIN_FLAG,
	 },
	{			/* RFC 3720 Section 12.17 MaxOutstandingR2T */
	      parameter_name:MAXOUTSTANDINGR2T,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | NUMBER | ONE_TO_65535 | MIN_NUMBER,
	      int_value:1,
	      special_key_flag:MAXOUTSTANDINGR2T_FLAG,
	 },
	{			/* RFC 3720 Section 12.18 DataPDUInOrder */
	      parameter_name:DATAPDUINORDER,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | BOOL_OR,
	      str_value:YES,
	      value_list:YES,
	      special_key_flag:DATAPDUINORDER_FLAG,
	 },
	{			/* RFC 3720 Section 12.19 DataSequenceInOrder */
	      parameter_name:DATASEQUENCEINORDER,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_IN_DISCOVERY |
	 KEY_SINGLE_VALUE | BOOL_OR,
	      str_value:YES,
	      value_list:YES,
	      special_key_flag:DATASEQUENCEINORDER_FLAG,
	 },
	{			/* RFC 3720 Section 11.1 AuthMethod */
	      parameter_name:AUTHMETHOD,
	      type:
	 INITIAL_ONLY | SECURITY_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_MULTI_VALUE | ENUMERATED | AUTH_PARAM,
	      str_value:NONE,
	      value_list:NONE,
	      special_key_flag:AUTHMETHOD_FLAG,
	 },
	{			/* RFC 3720 Section 12.20 ErrorRecoveryLevel */
	      parameter_name:ERRORRECOVERYLEVEL,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | NUMBER | ZERO_TO_2 | MIN_NUMBER,
	      int_value:0,
	      special_key_flag:ERRORRECOVERYLEVEL_FLAG,
	 },
	{			/* non-standard key to text X-extension facilities */
	      parameter_name:X_EXTENSIONKEY,
	      type:
	 ALL | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET |
	 KEY_SINGLE_VALUE | STRING | UTF_8,
	      str_value:NULL,
	      value_list:NULL,
	      special_key_flag:X_EXTENSIONKEY_FLAG,
	 }
#ifdef ISER
	/* Added for iSER Minti */
	, {			/* draft-ietf-ips-iser-05.txt Section 6.4 TargetRecvDataSegmentLength */
	      parameter_name:TARGETRECVDATASEGMENTLENGTH,
	      type:
	   INITIAL_ONLY | OPERATIONAL_PARAM |
	   USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITHOUT_RDMA |
	   KEY_SINGLE_VALUE | NUMBER | N512_TO_16777215 | MIN_NUMBER,
	      int_value:8192,
	      special_key_flag:TARGETRECVDATASEGMENTLENGTH_FLAG,
	   },
	{			/* draft-ietf-ips-iser-05.txt Section 6.5 InitiatorRecvDataSegmentLength */
	      parameter_name:INITIATORRECVDATASEGMENTLENGTH,
	      type:
	 INITIAL_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITHOUT_RDMA |
	 KEY_SINGLE_VALUE | NUMBER | N512_TO_16777215 | MIN_NUMBER,
	      int_value:8192,
	      special_key_flag:INITIATORRECVDATASEGMENTLENGTH_FLAG,
	 },
	{			/* draft-ietf-ips-iser-05.txt Section 6.7 MaxOutstandingUnexpectedPDUs */
	      parameter_name:MAXOUTSTANDINGUNEXPECTEDPDUS,
	      type:
	 LEADING_ONLY | OPERATIONAL_PARAM |
	 USE_BY_INITIATOR | USE_BY_TARGET | IRRELEVANT_WITHOUT_RDMA |
	 KEY_SINGLE_VALUE | NUMBER | TWO_TO_4294967295,
	      int_value:2,
	      special_key_flag:MAXOUTSTANDINGUNEXPECTEDPDUS_FLAG,
	 }
#endif
};

/*
 * deal with security keys
 */
struct security_key_struct {
	char *name;
	__u32 bitmask;
};

static struct security_key_struct security_key[] = {
	{CHAP_A, GOT_CHAP_A},
	{CHAP_I, GOT_CHAP_I},
	{CHAP_C, GOT_CHAP_C},
	{CHAP_N, GOT_CHAP_N},
	{CHAP_R, GOT_CHAP_R},
	{SRP_GROUP, GOT_SRP_GROUP},
	{SRP_S, GOT_SRP_S},
	{SRP_B, GOT_SRP_B},
	{SRP_HM, GOT_SRP_HM},
	{SRP_U, GOT_SRP_U},
	{SRP_TARGETAUTH, GOT_SRP_TARGETAUTH},
	{SRP_A, GOT_SRP_A},
	{SRP_M, GOT_SRP_M},
	{NULL, 0}
};

#define	SECURITY_HASH_BITS	5
#define	SECURITY_HASH_SIZE	(1<<SECURITY_HASH_BITS)
#define	SECURITY_HASH_MASK	(SECURITY_HASH_SIZE - 1)

static struct security_key_struct *security_hash_table[SECURITY_HASH_SIZE];

/*	the "perfect" hash function for all the CHAP and SRP keys */
static __u32 __attribute__ ((no_instrument_function))
    security_hash(const char *name)
{
	__u32 len, hash;

	len = strlen(name);
	hash = name[len - 1] + (len << 1);
	return hash & SECURITY_HASH_MASK;
}

/*	returns 0 on failure, else GOT_XXX value for this key */
__u32 __attribute__ ((no_instrument_function))
    is_securitykey(const char *keyname)
{
	__u32 hash;
	struct security_key_struct *key;

	hash = security_hash(keyname);
	if ((key = security_hash_table[hash]) != NULL &&
	    strcmp(keyname, key->name) == 0) {
		return key->bitmask;
	}
	return 0;
}

/*	returns 0 on failure after printing error message,
 *	else GOT_XXX value for this key
 */
__u32 __attribute__ ((no_instrument_function))
    print_bad_security_key(struct unknown_key * key)
{
	__u32 result;

	if (!(result = is_securitykey(key->keyname))) {
		TRACE_ERROR("unknown key \"%s\" in security phase\n",
			    key->keyname);
	}
	return result;
}

void __attribute__ ((no_instrument_function))
    print_not_allowed_security_key(struct unknown_key *key)
{
	TRACE_ERROR("Security key \"%s\" not allowed in this step\n",
		    key->keyname);
}

/*	called once on module load to setup the security key hash table */
void setup_security_hash_table(void)
{
	__u32 hash, count, collides;
	struct security_key_struct *key, *prev_key;

	for (count = collides = 0, key = security_key; key->name;
	     count++, key++) {
		hash = security_hash(key->name);
		if ((prev_key = security_hash_table[hash]) != NULL
		    && strcmp(key->name, prev_key->name)) {
			printk
			    ("Duplicate hash value %u for \"%s\" and \"%s\"\n",
			     hash, security_hash_table[hash]->name,
			     key->name);
			collides++;
		} else if (prev_key == NULL) {
			TRACE(TRACE_NEGOTIATIONS,
			      "%2u: Add \"%s\" at hash value %u\n",
			      count + 1, key->name, hash);
			security_hash_table[hash] = key;
		}
	}
	TRACE(TRACE_NEGOTIATIONS,
	      "%u hash collisions in %u security keys\n", collides, count);
}

/*	Called to get table entry for key identified by keytext string.
 *	Returns pointer to entry if found, NULL if not found.
 */
struct parameter_type * __attribute__ ((no_instrument_function))
    find_parameter(const char *keytext,
	       struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
	struct parameter_type *p, *result = NULL;
	int i;

	TRACE(TRACE_ENTER_LEAVE, "Enter find_parameter %s\n", keytext);

	for (i = 0, p = p_param_tbl; i < MAX_CONFIG_PARAMS; i++, p++) {
		if (!strcmp(p->parameter_name, keytext)) {
			result = p;
			break;
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave find_parameter, p %p\n", result);

	return result;
}

/* Called to get pointer to table entry for key identified by its special flag*/
struct parameter_type * __attribute__ ((no_instrument_function))
    find_flag_parameter(__u64 key_flag,
		    struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
	struct parameter_type *p, *result = NULL;
	int i;

	for (i = 0, p = p_param_tbl; i < MAX_CONFIG_PARAMS; i++, p++) {
		if (p->special_key_flag & key_flag) {
			result = p;
			break;
		}
	}

	return result;
}

/* Called to get int value of table entry for key identified by special flag */
int get_int_param(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		  __u64 flag)
{
	struct parameter_type *p;

	if ((p = find_flag_parameter(flag, p_param_tbl)))
		return p->int_value;
	return -1;
}

/*	Called to set max_recv_length to value we sent to target in a
 *	MaxRecvDataSegmentLength key
 *  Changed for iSER - Minti
 */
void
set_connection_length(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		      __u32 * max_recv_length, __u32 * max_send_length,
		      int role, int iSERFlag)
{
	struct parameter_type *p;

	TRACE(TRACE_ENTER_LEAVE, "Enter set_connection_length\n");

#ifdef ISER
	if (iSERFlag) {
		struct parameter_type *pInitiator, *pTarget;
		TRACE(TRACE_ISER, "iSER: (old) max_send_length: %u"
		      " max_recv_length: %u\n",
		      *max_send_length, *max_recv_length);

		if ((pInitiator =
		     find_flag_parameter
		     (INITIATORRECVDATASEGMENTLENGTH_FLAG,
		      p_param_tbl)) != NULL) {
			if ((pTarget =
			     find_flag_parameter
			     (TARGETRECVDATASEGMENTLENGTH_FLAG,
			      p_param_tbl)) != NULL) {
				if (role == INITIATOR) {
					*max_send_length =
					    pTarget->int_value;
					*max_recv_length =
					    pInitiator->int_value;
				} else {
					*max_send_length =
					    pInitiator->int_value;
					*max_recv_length =
					    pTarget->int_value;
				}
				TRACE(TRACE_ISER,
				      "iSER: (new)max_send_length: "
				      "%u max_recv_length: %u\n",
				      *max_send_length, *max_recv_length);
			}
		}
	} else
#endif
	{
		if ((p =
		     find_flag_parameter(MAXRECVDATASEGMENTLENGTH_FLAG,
					 p_param_tbl))
		    && IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
			/* we did send this key to the other side */
			*max_recv_length = p->int_value;
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave set_connection_length\n");
}

/* called on transition to FFP in order to turn on digest flags for connection*/
void
set_digestflags(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		__u32 * connection_flags)
{
	struct parameter_type *p;

	if ((p = find_flag_parameter(HEADERDIGEST_FLAG, p_param_tbl))) {
		if (strcmp(p->str_value, CRC32C) == 0) {
			*connection_flags |= USE_HEADERDIGEST;
			TRACE(TRACE_ISCSI, "Enabling Header Digests\n");
		}
	}
	if ((p = find_flag_parameter(DATADIGEST_FLAG, p_param_tbl))) {
		if (strcmp(p->str_value, CRC32C) == 0) {
			*connection_flags |= USE_DATADIGEST;
			TRACE(TRACE_ISCSI, "Enabling Data Digests\n");
		}
	}
}

/*	Check that a number is within legal bounds depending on its type.
 *	Returns 1 if ok, 0 if error
 */
static int __attribute__ ((no_instrument_function))
    check_bounds(struct parameter_type *p, int int_value, int who_called)
{
	int retval = 1;

	if (IS_ONE_TO_65535(p->type)) {
		if ((int_value < 1) || (int_value > 65535)) {
			TRACE_ERROR("illegal number %d - "
				    "should be between 1 and 65535\n",
				    int_value);
			retval = 0;
		}
	} else if (IS_ZERO_TO_65535(p->type)) {
		if ((int_value < 0) || (int_value > 65535)) {
			TRACE_ERROR("illegal number %d - "
				    "should be between 0 and 65535\n",
				    int_value);
			retval = 0;
		}
	} else if (IS_N512_TO_16777215(p->type)) {
		if ((int_value < 512) || (int_value > 16777215)) {
			if (who_called == MANAGEMENT &&
			    (p->special_key_flag & (MAXBURSTLENGTH_FLAG |
						    FIRSTBURSTLENGTH_FLAG)))
			{
				/* Burst limits depend on draft,
				   do it at run time */
			} else {
				TRACE_ERROR("illegal number %d - "
					    "should be between 512 and 16777215\n",
					    int_value);
				retval = 0;
			}
		}
	} else if (IS_ZERO_TO_3600(p->type)) {
		if ((int_value < 0) || (int_value > 3600)) {
			TRACE_ERROR("illegal number %d - "
				    "should be between 0 and 3600\n",
				    int_value);
			retval = 0;
		}
	} else if (IS_TWO_TO_4294967295(p->type)) {
		if (int_value < 2) {
			TRACE_ERROR("illegal number %d - "
				    "should be between 2 and 4294967295\n",
				    int_value);
			retval = 0;
		}
	} else if (IS_ZERO_TO_2(p->type)) {
		if ((int_value < 0) || (int_value > 2)) {
			TRACE_ERROR("illegal number %d - "
				    "should be between 0 and 2\n",
				    int_value);
			retval = 0;
		}
	}

	return retval;
}

/*	This is a range, check both numbers and their order.
 *	Returns first number in range if ok, -1 if error
 */
static int __attribute__ ((no_instrument_function))
    check_range(const char *value_list, int value)
{
	char *endptr;
	int lower, upper;

	lower = simple_strtoul(value_list, &endptr, 0);
	if (*endptr == '~') {
		/* first number followed by ~ must be
		   followed by another number */
		upper = simple_strtoul(endptr + 1, &endptr, 0);
	} else {
		/* only 1 value was specified, which is ok */
		upper = lower;
	}

	if (strspn(endptr, WHITE_SPACE) != strlen(endptr) || lower > upper) {
		TRACE_ERROR("illegal range \"%s\"\n", value_list);
		return -1;
	}
	if (value >= 0 && (value < lower || value > upper)) {
		TRACE_ERROR("value %d out of range \"%s\"\n", value,
			    value_list);
		return -1;
	}

	return lower;
}

/*
 * Checks correctness of the type of parameter with the value passed
 * The main types are defined in "text_param.h"
 * If the value is NULL, no checks made.
 * On return, a flag is set in p->neg_info if a fatal error occured
 * Called only from check_correctness
 */

static void __attribute__ ((no_instrument_function))
    check_type_correctness(struct parameter_type *p,
		       char *value, int who_called, unsigned int *int_value,
		       __u64 * login_flags)
{
	char *dummy;
	char *endptr = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_type_correctness, "
	      "parameter: %s, value: %s\n", p->parameter_name, value);

	/* If the value is inquiry, ignore it */
	if (!strcmp(value, "?"))
		goto out;

	/* look for value = "Reject" or "Irrelevant" or "NotUnderstood" */
	if (!strcmp(value, reject_word)) {
		if (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
			/* Reject always fatal as an offer */
			TRACE_ERROR("illegal offer: %s=%s\n",
				    p->parameter_name, value);
			p->neg_info |= KEY_BAD;
		} else if (p->
			   special_key_flag & (OFMARKINT_FLAG |
					       IFMARKINT_FLAG)) {
			/* Reject is not fatal as a reply to
			 * OFMarkInt/IFMarkInt keys
			 *  RFC 3720 Section A.3.2 OFMarkInt, IFMarkInt
			 *  "When the interval is unacceptable the responder
			 *  answers with "Reject".  Reject is resetting
			 *  the marker function in the
			 *  specified direction (Output or Input) to No."
			 */
			/* so we can do a reset later */
			p->neg_info |= KEY_REJECT;
		} else {
			/* Reject always fatal as a reply to all other keys */
			TRACE_ERROR("got key: %s=%s\n", p->parameter_name,
				    value);
			p->neg_info |= KEY_BAD;
		}
		goto out;
	}

	if (!strcmp(value, irrelevant_word)) {
		if (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
			/* Irrelevant always fatal as an offer */
			TRACE_ERROR("illegal offer: %s=%s\n",
				    p->parameter_name, value);
			p->neg_info |= KEY_BAD;
		} else {
			/* Irrelevant is a legal reply in all negotiations
			 * but print trace info only if not discovery
			 * and not a reply to OFMarkInt or IFMarkInt
			 */
			if (!(((*login_flags & DISCOVERY_FLAG)
			       && IS_IRRELEVANT_IN_DISCOVERY(p->type))
			      || (p->special_key_flag & OFMARKINT_FLAG)
			      || (p->special_key_flag & IFMARKINT_FLAG))) {
				TRACE_WARNING("got key: %s=%s\n",
					      p->parameter_name, value);
			}
			p->neg_info |= KEY_IRRELEVANT;
		}
		goto out;
	}

	/*  RFC 3720 Section 5.2 Text Mode Negotiation
	 *  "All keys in this document, except for the X extension formats,
	 *  MUST be supported by iSCSI initiators and targets when used
	 *  as specified here.  If used as specified, these keys MUST NOT be
	 *  answered with NotUnderstood."
	 */
	if (!strcmp(value, notunderstood_word)) {
		if (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
			/* NotUnderstood always fatal as an offer */
			TRACE_ERROR("illegal offer: %s=%s\n",
				    p->parameter_name, value);
			p->neg_info |= KEY_BAD;
		} else if ((p->special_key_flag & X_EXTENSIONKEY_FLAG)) {
			/* other side not expected to understand this offer
			 * anyway, consider its response irrelevant
			 */
			p->neg_info |= KEY_IRRELEVANT;
		} else {
			/* "NotUnderstood" always fatal reply to standard key */
			TRACE_ERROR
			    ("standard key \"%s\" MUST NOT be answered "
			     "with \"%s\"\n", p->parameter_name,
			     notunderstood_word);
			p->neg_info |= KEY_BAD;
		}
		goto out;
	}

	if (IS_STRING(p->type)) {
		/*  key value is a string, just check its total length */

		/*  RFC 3720 Section 5.1 Text Format
		 *  "If not otherwise specified, the maximum length of a
		 *  simple-value (not its encoded representation) is 255
		 *  bytes, not including the delimiter (comma or zero byte)."
		 *
		 *  RFC 3720 Section 3.2.6.1  iSCSI Name Properties
		 *  "Each iSCSI node, whether an initiator or target,
		 *  MUST have an iSCSI name.
		 *
		 *  Initiators and targets MUST support the receipt of iSCSI
		 *  names of up to the maximum length of 223 bytes."
		 */
		char *tptr;
		int len = strlen(value);
		int t_num;

		if (len > MAX_KEY_VALUE_LENGTH) {
			TRACE_ERROR
			    ("value of key \"%s\" longer than %d characters\n",
			     p->parameter_name,
			     (IS_ISCSI_NAME(p->type) ?
			      MAX_ISCSI_NAME_LENGTH :
			      MAX_KEY_VALUE_LENGTH));
			p->neg_info |= KEY_BAD;
		}

		/* check ISCSI_NAME type */
		if (IS_ISCSI_NAME(p->type)) {
			if (who_called == TARGET &&
			    p->special_key_flag & TARGETNAME_FLAG) {
				/* check TargetName sent by the initiator */
				p->int_value = 0;	/* assume an error */
				if ((strncmp(targetname_header_word, value,
					     strlen
					     (targetname_header_word)) !=
				     0)
				    || (tptr = strchr(value, ':')) == NULL) {
					TRACE_WARNING
					    ("invalid %s \"%s\". "
					     "Use Discovery session"
					     " to get proper %s.\n",
					     p->parameter_name, value,
					     p->parameter_name);
					goto out;
				}
				tptr++;	/* point at char after ':' */
				if (strspn(tptr, WHITE_SPACE) ==
				    strlen(tptr)) {
					TRACE_ERROR
					    ("no target number in %s \"%s\"\n",
					     p->parameter_name, value);
					p->neg_info |= KEY_BAD;
					goto out;
				}

				t_num = simple_strtoul(tptr, &endptr, 10);
				if (t_num < 0 || *endptr != '\0') {
					TRACE_ERROR
					    ("bad target number \"%s\" "
					     "in %s \"%s\"\n",
					     tptr, p->parameter_name,
					     value);
					p->neg_info |= KEY_BAD;
					goto out;
				}
				/* check target number from the target name */
				if (!target_in_use(t_num)) {
					TRACE_ERROR
					    ("target number %d not in use "
					     "in %s \"%s\"\n",
					     t_num, p->parameter_name,
					     value);
					p->neg_info |= KEY_BAD;
					goto out;
				}
				/* target number ok, save it for pickup later */
				p->int_value = t_num;
			}

			/*  check the format of the iSCSI name */
			/*  RFC 3720 3.2.6.3  iSCSI Name Structure
			 *  "The type designator strings currently defined are:
			 *     iqn. - iSCSI Qualified name
			 *     eui. - Remainder of the string is an IEEE EUI-64
			 *            identifier, in ASCII-encoded hexadecimal."
			 */
			if (len == 20 && strncmp(value, "eui.", 4) == 0) {
				/* IEEE format world wide unique name --
				   16 hex digits must follow */
				if (strspn
				    (value + 4, "0123456789abcdefABCDEF")
				    == 16)
					/* ok, have exactly 16 hex
					   digits following the "eui." */
					goto out;
			} else if (len > 12 && len <= MAX_ISCSI_NAME_LENGTH
				   && strncmp(value, "iqn.", 4) == 0) {
				/* iSCSI qualified name -- must get "yyyy-mm."
				   then alphanums, ., - */
				if (strspn(value + 4, "0123456789") == 4
				    && *(value + 8) == '-'
				    && strspn(value + 9, "0123456789") == 2
				    && *(value + 11) == '.'
				    && strspn(value + 12,
					      "0123456789abcdefghijklmnopqrstuvwxyz-.:")
				    == len - 12) {
					/* ok, have iqn.yyyy-mm.xxx with xxx
					 * all digits, lower-case letters,
					 * * '-', '.', ':'
					 */
					goto out;
				}
			}

			/* If reach here have invalid iSCSI name */
			if (TRACE_TEST(TRACE_ISCSI)) {
				TRACE_WARNING
				    ("invalid iSCSI name \"%s\"\n", value);
			}
			goto out;
		}
	}

	/* the following are to be tested for each value in the list    */
	do {
		endptr = NULL;

		/* extract just a single value */
		dummy = strchr(value, ',');
		if (dummy) {
			/* found a comma so value contains list of
			   values -- is this ok? */
			if (!IS_KEY_MULTI_VALUE(p->type)
			    && !(p->special_key_flag & TARGETADDRESS_FLAG)) {
				/* this key cannot accept a list of values */
				TRACE_ERROR
				    ("key \"%s\" does not accept a list of values "
				     "\"%s\"\n", p->parameter_name, value);
				p->neg_info |= KEY_BAD;
				break;
			}
			/* terminate first item in the list so it
			   is a string by itself */
			*dummy = '\0';
		}

		TRACE(TRACE_NEGOTIATIONS, "value: %s\n", value);

		/*  Check for length of the key value. */
		/*  RFC 3720 Section 5.1 Text Format
		 *  "If not otherwise specified, the maximum length of a
		 *  simple-value (not its encoded representation) is 255
		 *  bytes, not including the delimiter (comma or zero byte)."
		 */
		if (strlen(value) > MAX_KEY_VALUE_LENGTH) {
			if (TRACE_TEST(TRACE_ISCSI)) {
				TRACE_WARNING("value of key \"%s\" "
					      "exceeds %d characters\n",
					      p->parameter_name,
					      MAX_KEY_VALUE_LENGTH);
			}
		}

		if (*value == '\0' && !IS_KEY_NO_VALUE(p->type)) {
			TRACE_ERROR("no value after '=' for key \"%s\"\n",
				    p->parameter_name);
			p->neg_info |= KEY_BAD;
			break;
		}

		else if (IS_NUMBER(p->type) ||
			 (IS_NUMBER_RANGE(p->type) &&
			  IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))) {
			/* a number or a reply to a range */

			*int_value = simple_strtoul(value, &endptr, 0);

			if (strspn(endptr, WHITE_SPACE) != strlen(endptr)) {
				TRACE_ERROR("illegal number \"%s\"\n",
					    value);
				p->neg_info |= KEY_BAD;
				break;
			}

			if (!check_bounds(p, *int_value, who_called)) {
				p->neg_info |= KEY_BAD;
				break;
			}
		}

		else if (IS_BOOLEAN(p->type)) {
			if (strcmp(value, yes_word)
			    && strcmp(value, no_word)) {
				TRACE_ERROR
				    ("illegal value \"%s\" - expected "
				     "\"%s\" or \"%s\"\n", value, yes_word,
				     no_word);
				p->neg_info |= KEY_BAD;
				break;
			}
		}

		else if (IS_ENUMERATED(p->type)) {
			if (IS_AUTH_PARAM(p->type)) {
				if (strcmp(value, KRB5)
				    && strcmp(value, SPKM1)
				    && strcmp(value, SPKM2)
				    && strcmp(value, SRP)
				    && strcmp(value, CHAP)
				    && strcmp(value, none_word)) {
					TRACE_ERROR
					    ("illegal value \"%s\" - expected "
					     "\"KRB5\" or "
					     "\"SPKM1\" or \"SPKM2\" or \"SRP\" or "
					     "\"CHAP\" or \"%s\"\n", value,
					     none_word);
					p->neg_info |= KEY_BAD;
					break;
				}
			} else if (IS_DISCOVERY_NORMAL(p->type)) {
				if (strcmp(value, discovery_word)
				    && strcmp(value, normal_word)) {
					TRACE_ERROR
					    ("illegal value \"%s\" - expected "
					     "\"%s\" or \"%s\"\n", value,
					     discovery_word, normal_word);
					p->neg_info |= KEY_BAD;
					break;
				}
			}
		}

		else if (IS_NUMBER_RANGE(p->type)
			 && !IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
			/* is an offer of a numeric range */
			if ((*int_value = check_range(value, -1)) < 0
			    || !check_bounds(p, *int_value, who_called)) {
				p->neg_info |= KEY_BAD;
				break;
			}
		}

		if (dummy) {
			/* restore the comma and goto the next value */
			*dummy++ = ',';
		}

		value = dummy;
	}
	while (value);

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave check_type_correctness\n");

	return;
}

/*
 * Checks the correctness of the passed string, checks whether it
 * conforms to the <key>=<value> format
 * extracts the value from the input string and returns it as an
 * output parameter.
 * Returns pointer to parameter in p_param_tbl if ok, else NULL after error
 * (but on non-NULL p->neg_info can still have flag set on fatal error)		*/
static struct parameter_type * __attribute__ ((no_instrument_function))
    check_correctness(char *keytext,
		  const char **p_value,
		  struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		  __u32 who_called,
		  __u32 when_called,
		  __u32 flags,
		  unsigned int *int_value,
		  __u64 * login_flags,
		  struct unknown_key **unknown_key_list)
{
	char *value;
	struct parameter_type *p = NULL;
	struct unknown_key *uptr;
	__u32 count;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_correctness\n");

	TRACE(TRACE_ISCSI, "Got key: %s\n", keytext);

	if ((value = strchr(keytext, '=')) == NULL) {
		/* could not find '=', don't accept this */
		TRACE_ERROR("key \"%s\" not followed by '='\n", keytext);
		goto out;
	}

	/* terminate the key name string and point at the value string */
	*value++ = '\0';

	/* Check for length of the key name. */

	/*  RFC 3720 Section 5.1 Text Format
	 *  "A standard-label MUST begin with a capital letter and
	 *  must not exceed 63 characters."
	 */
	if (strlen(keytext) > MAX_KEY_NAME_LENGTH) {
		if (TRACE_TEST(TRACE_ISCSI)) {
			TRACE_WARNING
			    ("length of key name \"%s\" exceeds %d\n",
			     keytext, MAX_KEY_NAME_LENGTH);
		}
	}

	/*  RFC 3720 Section 10.12.10 Login Parameters
	 *  "All keys in Chapter 12, except for the X extension formats,
	 *  MUST be supported by iSCSI initiators and targets.
	 *  Keys in Chapter 11 only need to be supported when the function
	 *  to which they refer is mandatory to implement."
	 */
	if (!(p = find_parameter(keytext, p_param_tbl))) {
		/* key not found in our table, look for
		   value = "Reject" or "Irrelevant" or "NotUnderstood" */
		if (!strcmp(value, reject_word)
		    || !strcmp(value, irrelevant_word)
		    || !strcmp(value, notunderstood_word)) {
			/* These values always fatal as a value for offer
			   of unknown key */
			TRACE_ERROR
			    ("unknown key with illegal value: %s=%s\n",
			     keytext, value);
		} else {
			/* value not bad, but have we seen this
			   unknown key before? */
			for (count = 0, uptr = *unknown_key_list;
			     uptr != NULL; count++, uptr = uptr->next) {
				if (!strcmp(keytext, uptr->keyname)) {
					/* yes, this unknown key is
					   being negotiated twice */
					TRACE_ERROR
					    ("unknown key \"%s\" received twice\n",
					     keytext);
					goto out;
				}
			}

			/* never saw this key before, should we save
			   it for later checks? */
			/* chap support = CHONG */
			if (!is_securitykey(keytext)) {
				if (TRACE_TEST(TRACE_ISCSI)) {
					TRACE_WARNING
					    ("unknown key \"%s\"\n",
					     keytext);
				}
			}

			if (count >= MAX_UNKNOWN_KEYS) {
				/* no, we have seen too many unknown keys,
				   bail out */
				TRACE_ERROR("%u unknown keys received, "
					    "too many!\n", count);
				goto out;
			}

			else if ((uptr =
				  my_kmalloc(sizeof(struct unknown_key),
					     "Unknown key")) == NULL) {
				goto out;
			}

			/* chap and srp support - CHONG */
			else if ((uptr->keyname =
				  my_kmalloc(strlen(keytext) + 1,
					     "Key Name"))
				 == NULL) {
				my_kfree((void **) &uptr, "Unknown key");
				goto out;
			}

			else if ((uptr->keyvalue =
				  my_kmalloc(strlen(value) + 1,
					     "Key Value"))
				 == NULL) {
				my_kfree((void **) &uptr->keyname,
					 "Key Name");
				my_kfree((void **) &uptr, "Unknown key");
				goto out;
			} else {
				strcpy(uptr->keyname, keytext);
				strcpy(uptr->keyvalue, value);
				uptr->processed = 0;
				uptr->next = *unknown_key_list;
				*unknown_key_list = uptr;

				if ((flags & USE_REFLECT_XKEYS)
				    && (strncmp(keytext, "X-", 2) == 0
					|| strncmp(keytext, "X#", 2) == 0))
					/* this is an X key and user wants its
					 * value reflected back to sender
					 */
					*p_value = value;
				else
					/* pass back Notunderstood value for
					 * unknown key
					 */
					*p_value = notunderstood_word;
			}
		}
		goto out;
	}

	if ((who_called == INITIATOR && !IS_USE_BY_TARGET(p->type))
	    || (who_called == TARGET && !IS_USE_BY_INITIATOR(p->type))) {
		TRACE_ERROR("key \"%s\" cannot be sent to %s\n", keytext,
			    who_called ==
			    INITIATOR ? "initiator" : "target");
		p->neg_info |= KEY_BAD;
		goto out;
	}

	if (IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info)) {
		/* attempting to negotiate this key twice */
		/*  RFC 3720 Section 5.3 Login Phase
		 *  "Neither the initiator nor the target should attempt to
		 *  declare or negotiate a parameter more than once
		 *  during login except for responses to specific keys
		 *  that explicitly allow repeated key declarations
		 *  (e.e., TargetAddress).  An attempt to renegotiate/
		 *  redeclare parameters not specifically allowed MUST
		 *  be detected by the initiator and target.  If such an
		 *  attempt is detected by the target, the target MUST
		 *  respond with Login reject (initiator error);
		 *  if detected by the initiator, the initiator MUST drop the
		 *  connection."
		 */
		TRACE_ERROR("key \"%s\" received twice\n",
			    p->parameter_name);
		p->neg_info |= KEY_BAD;
		goto out;
	}

	if (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
		/* this is new offer from other side, not reply
		   to one of our offers */
		if (!(p->type & when_called)) {
			/* wrong time to be negotiating this parameter */
			TRACE_ERROR
			    ("key \"%s\" cannot be negotiated now\n",
			     keytext);
			p->neg_info |= KEY_BAD;
			goto out;
		}
	}

	check_type_correctness(p, value, who_called, int_value,
			       login_flags);

	/* Return the value pointer on success */
	*p_value = value;

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave check_correctness, p %p\n", p);

	if (IS_NUMBER(p->type))
		TRACE(TRACE_ISCSI_FULL, "p->int_value=%d\n", p->int_value);

	return p;
}

char * __attribute__ ((no_instrument_function))
    my_strdup(const char *str)
{
	int n = strlen(str) + 1;
	char *s = kmalloc(n, GFP_KERNEL);

	if (!s)
		return NULL;
	return strcpy(s, str);
}

void __attribute__ ((no_instrument_function))
    strreplace(char **str, const char *new_str)
{
	if (*str != NULL)
		kfree(*str);
	*str = my_strdup(new_str);
}

/* Copy src parameter table to dst, duplicating any strings
 * except parameter_name, since that is always a constant
 */
void
param_tbl_cpy(struct parameter_type dst[MAX_CONFIG_PARAMS],
	      struct parameter_type src[MAX_CONFIG_PARAMS])
{
	struct parameter_type *dptr;
	int i;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter param_tbl_cpy, dst %p, src %p, size %ld\n", dst, src,
	      (long) sizeof(struct parameter_type) * MAX_CONFIG_PARAMS);

	/* copy everything in the src table to the dst table "as is" */
	memcpy(dst, src,
	       sizeof(struct parameter_type) * MAX_CONFIG_PARAMS);

	/* now go back and "dup" all the strings */
	for (i = 0, dptr = dst; i < MAX_CONFIG_PARAMS; i++, dptr++) {
		if (dptr->str_value) {
			dptr->str_value = my_strdup(dptr->str_value);
		}
		if (dptr->value_list) {
			dptr->value_list = my_strdup(dptr->value_list);
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave param_tbl_cpy\n");
}

/* Copy initial parameter table to dst, duplicating any strings
 * except parameter_name, since that is always a constant
 */
void param_tbl_init(struct parameter_type dst[MAX_CONFIG_PARAMS])
{
	param_tbl_cpy(dst, config_params);
}

/* Free any strings referenced in dst parameter table
 * except parameter_name, since that is always a constant
 */
void param_tbl_uncpy(struct parameter_type dst[MAX_CONFIG_PARAMS])
{
	struct parameter_type *dptr;
	int i;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter param_tbl_uncpy, dst %p, size %ld\n", dst,
	      (long) sizeof(struct parameter_type) * MAX_CONFIG_PARAMS);

	/* go thru and free all the strings that were "duped" during copy */
	for (i = 0, dptr = dst; i < MAX_CONFIG_PARAMS; i++, dptr++) {
		if (dptr->str_value) {
			kfree(dptr->str_value);
		}
		if (dptr->value_list) {
			kfree(dptr->value_list);
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave param_tbl_uncpy\n");
}

void
configure_parameter(int param_neg_info,
		    char *ptr_to_keytext,
		    struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
	struct parameter_type *param = NULL;
	char *value_list = NULL;
	char *endptr;
	int int_value;

	TRACE(TRACE_ENTER_LEAVE, "Enter configure_parameter\n");

	/* Get the value string */
	value_list = strchr(ptr_to_keytext, '=');
	if (value_list)
		*value_list++ = '\0';	/* terminate the key text string */

	if ((param = find_parameter(ptr_to_keytext, p_param_tbl)) != NULL) {
		if (value_list) {
			if (IS_NUMBER(param->type)) {
				/* parameter is a number, set only the
				 * int_value field, don't set value_list
				 * or str_value fields
				 */

				int_value =
				    simple_strtoul(value_list, &endptr, 0);
				if (strspn(endptr, WHITE_SPACE) !=
				    strlen(endptr)) {
					TRACE_ERROR
					    ("illegal number \"%s\"\n",
					     value_list);
					goto out;
				} else {
					param->int_value = int_value;
					TRACE(TRACE_ISCSI_FULL, "Number %d set "
					      "for %s\n", int_value,
					      param->parameter_name);
				}
			} else {
				/* parameter is a string, enumerated, boolean
				 * or range,
				 * set both the value_list and str_value fields,
				 * do not set int_value field
				 */

				if (IS_NUMBER_RANGE(param->type)) {
					/* this is a range, check both
					   numbers and their order */
					if (check_range(value_list, -1) <
					    0)
						goto out;
				}

				strreplace(&param->value_list, value_list);
				endptr = strchr(value_list, ',');
				if (endptr) {	/* value_list really a list,i
						 * use 1st on list as
						 * current value
						 */
					*endptr = '\0';
					strreplace(&param->str_value,
						   value_list);
					*endptr = ',';	/* repair the damage */
				} else {
					/* value_list only single value, make
					   it current val too */
					strreplace(&param->str_value,
						   value_list);
				}
			}
		}

		/* set the negotiation info only if no error */
		param->neg_info = param_neg_info;
	} else {
		TRACE_WARNING("find_parameter did not find a match for %s\n",
		              ptr_to_keytext);
	}

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave configure_parameter\n");
}

#ifdef ISER
/* Will only be called on *leading login* to check if RDMAExtensions was proposed
 * and if so, is the reply expected.
 */
int is_iSER_proposed(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		     int *replyOptional)
{
	struct parameter_type *p = NULL;
	int negFlag = 0;

	if ((p =
	     find_flag_parameter(RDMAEXTENSIONS_FLAG,
				 p_param_tbl)) != NULL) {
		negFlag = IS_KEY_TO_BE_NEGOTIATED(p->neg_info);

		if (negFlag && (!strcmp(p->str_value, no_word))) {
			*replyOptional = 1;
		}
	}
	return negFlag;
}

int is_iSER_enabled(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		    struct session_operational_parameters *oper_param,
		    __u32 secondary_connection)
{
	int iSERChecksFlag = 0;
	struct parameter_type *p;

	if (!secondary_connection) {
		if ((p =
		     find_flag_parameter(RDMAEXTENSIONS_FLAG, p_param_tbl))
		    != NULL) {
			if (!strcmp(p->str_value, yes_word)) {
				iSERChecksFlag = 1;
			}
			TRACE(TRACE_ISER, "RDMAExtensions=%s\n",
			      p->str_value);
		} else {
			TRACE_ERROR("iSER: Null special flags\n");
		}
	} else {
		if (oper_param->RDMAExtensions)
			iSERChecksFlag = 1;

		TRACE(TRACE_ISER, "oper_param RDMAExtensions=%d\n",
		      oper_param->RDMAExtensions);
	}
	return iSERChecksFlag;
}
#endif
/*	RFC 3720 Section 5.2 Text Mode Negotiation
 *	"Some parameters may be subject to integrity rules (e.g., parameter-x
 *	must not exceed parameter-y or parameter-u not 1 implies parameter-v
 *	be Yes).  Whenever required, integrity rules are specified with the
 *	keys.  Checking for compliance with the integrity rule must only be
 *	performed after all the parameters are available (the existent and the
 *	newly negotiated).  An iSCSI target MUST perform integrity checking
 *	before the new parameters take effect.	An initiator MAY perform
 *	integrity checking."
 */
void
check_integrity_rules(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		      struct session_operational_parameters *oper_param,
		      __u32 secondary_connection,
		      __u32 connection_flags, int iSERChecksFlag)
{
	struct parameter_type *p, *p2;
	int discoveryFlag = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_integrity_rules\n");

	/*  RFC 3720 Section A.3.2 OFMarkInt, IFMarkInt
	 *  "Reject is resetting the marker function in the specified
	 *  direction (Output or Input) to No."
	 */
	if ((p = find_flag_parameter(OFMARKINT_FLAG, p_param_tbl)) != NULL) {
		if (p->neg_info & KEY_REJECT) {
			/* got OFMarkInt=Reject, so set OFMarker=No */
			TRACE(TRACE_NEGOTIATIONS, "Checking %s=%d\n",
			      p->parameter_name, p->int_value);
			if ((p = find_flag_parameter(OFMARKER_FLAG,
						     p_param_tbl)) !=
			    NULL) {
				TRACE(TRACE_NEGOTIATIONS, "Have %s=%s\n",
				      p->parameter_name, p->str_value);
				if (!strcmp(p->str_value, yes_word)) {
					/* OFMarker=Yes, change value to No */
					strreplace(&p->str_value, no_word);
					TRACE(TRACE_ISCSI,
					      "Reset %s to %s\n",
					      p->parameter_name,
					      p->str_value);
				}
			}
		}
	}
	if ((p = find_flag_parameter(IFMARKINT_FLAG, p_param_tbl)) != NULL) {
		if (p->neg_info & KEY_REJECT) {
			/* got IFMarkInt=Reject, so set IFMarker=No */
			TRACE(TRACE_NEGOTIATIONS, "Checking %s=%d\n",
			      p->parameter_name, p->int_value);
			if ((p = find_flag_parameter(IFMARKER_FLAG,
						     p_param_tbl)) !=
			    NULL) {
				TRACE(TRACE_NEGOTIATIONS, "Have %s=%s\n",
				      p->parameter_name, p->str_value);
				if (!strcmp(p->str_value, yes_word)) {
					/* IFMarker=Yes, change value to No */
					strreplace(&p->str_value, no_word);
					TRACE(TRACE_ISCSI,
					      "Reset %s to %s\n",
					      p->parameter_name,
					      p->str_value);
				}
			}
		}
	}

	/* the following checks are done only on 1st connection in a session */
	if (!secondary_connection) {
		/*  RFC 3720 Section 12.14 FirstBurstLength
		 *  "FirstBurstLength MUST NOT exceed MaxBurstLength."
		 */
		if ((p = find_flag_parameter(FIRSTBURSTLENGTH_FLAG,
					     p_param_tbl)) != NULL) {
			if ((p2 = find_flag_parameter(MAXBURSTLENGTH_FLAG,
						      p_param_tbl)) !=
			    NULL) {
				if (p->int_value > p2->int_value) {
					if (TRACE_TEST(TRACE_ISCSI)) {
						TRACE_WARNING
						    ("FirstBurstLength %u"
						     " exceeds MaxBurstLength %u\n",
						     p->int_value,
						     p2->int_value);
					}
					/* fix FirstBurstLength */
					p->int_value = p2->int_value;
				}
			}
		}

		/*  RFC 3720 Section 12.21 SessionType
		 *  "The discovery session implies MaxConnections = 1
		 *  and overrides both the default and an explicit setting."
		 */
		if ((p = find_flag_parameter(SESSIONTYPE_FLAG,
					     p_param_tbl)) != NULL) {
			if (!strcmp(p->str_value, discovery_word)
			    && (p2 =
				find_flag_parameter(MAXCONNECTIONS_FLAG,
						    p_param_tbl)) !=
			    NULL) {
				if (p2->int_value != 1) {
					/* do not have MaxConnections=1,
					   change value to 1 */
					p2->int_value = 1;
					TRACE(TRACE_ISCSI,
					      "Reset %s to %u\n",
					      p2->parameter_name,
					      p2->int_value);
				}
			}
		}

		/*  RFC 3720 Section 12.19 DataSequenceInOrder
		 *  "If DataSequenceInOrder is set to Yes, a target may
		 *  retry at most the last R2T, and an initiator may
		 *  at most request retransmission for the last read
		 *  data sequence.  For this reason, if
		 *  ErrorRecoveryLevel is not 0 and DataSequenceInOrder is
		 *  set to Yes then MaxOutstandingR2T MUST be set to 1."
		 */
		if ((p = find_flag_parameter(DATASEQUENCEINORDER_FLAG,
					     p_param_tbl)) != NULL) {
			if (!strcmp(p->str_value, yes_word)
			    && (p2 =
				find_flag_parameter
				(ERRORRECOVERYLEVEL_FLAG, p_param_tbl))
			    != NULL) {	/* have DataSequenceInOrder=Yes */
				if (p2->int_value > 0
				    && (p = find_flag_parameter
					(MAXOUTSTANDINGR2T_FLAG,
					 p_param_tbl))
				    != NULL) {
					/* have ErrorRecoveryLevel>0 */
					if (p->int_value != 1) {
						/* do not have
						 * MaxOutstandingR2T = 1,
						 * change it to 1
						 */
						if (TRACE_TEST
						    (TRACE_ISCSI)) {
							TRACE_WARNING
							    ("%s=%u reset to 1\n",
							     p->
							     parameter_name,
							     p->int_value);
						}
						p->int_value = 1;
					}
				}
			}
		}
	} else {
		/* Testing - TODO Minti */
		if (oper_param->SessionType)
			discoveryFlag = 1;
	}
#ifdef ISER
	/* iSER Specification 8.1-8.6
	 * Negotiations resulting in RDMAExtensions=Yes imply HeaderDigest=None,
	 * DataDigest=None, OFMarker=No, and IFMarker=No, and override both the
	 * default and explicit settings of these keys.
	 * Irrelevant when: SessionType=Discovery
	 */

	if (iSERChecksFlag && (discoveryFlag == 0)) {
		TRACE(TRACE_ISER, "connection_flags=%x\n",
		      connection_flags);
		if ((connection_flags & ISER_ISCSI_KEYS_OFF) == 0)
			handle_iser_keys(p_param_tbl);
	}
#endif
	TRACE(TRACE_ENTER_LEAVE, "Leave check_integrity_rules\n");
}

/*
 * This function receives 'length' bytes into buffer from the socket
 */
int
iscsi_recv_msg(struct socket *sock, int length, char *buffer, int flags)
{
	struct msghdr msg;
	size_t size;
	int received, retval;
	struct iovec iov;
#ifdef __KERNEL__
	mm_segment_t oldfs;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_recv_msg, length %d bytes\n",
	      length);

	/* the following code will receive PDU's from the other side */
	size = length;
	for (received = 0; received < length;) {
		/* most fields remain 0/NULL */
		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_iovlen = 1;
		msg.msg_iov = &iov;

		iov.iov_len = size;
		iov.iov_base = buffer + received;

#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
		retval = sock_recvmsg(sock, &msg, size, MSG_WAITALL);
		set_fs(oldfs);
#else
		TRACE(TRACE_NET,
		      "%s Call readv, fd %d, size %lu, base %p\n",
		      current->comm, sock->fd, (unsigned long) size, buffer + received);
		retval = readv(sock->fd, &iov, 1);
		TRACE(TRACE_NET, "%s Done readv, fd %d, retval %d\n",
		      current->comm, sock->fd, retval);
#endif

		if (retval <= 0) {
			if (retval == 0) {
				TRACE_ERROR("remote peer disconnected\n");
			} else {
				TRACE_ERROR("sock_recvmsg error %d\n",
					    retval);
			}
			retval = -1;
			goto out;
		}
		received += retval;
		size = length - received;

		TRACE(TRACE_NET, "Received: %d, total_rx: %d, data: %d\n",
		      retval, received, length);
	}

	TRACE_BUFFER(TRACE_BUF, buffer, length,
		     "Got buffer, length: %d\n", length);
	retval = 0;

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_recv_msg, retval %d\n",
	      retval);

	return retval;
}

/*
 * returns 1 if needs a break from the loop
 * else returns a 0
 */
int __attribute__ ((no_instrument_function))
    check_out_length(int out_length, int resp_len)
{

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter check_out_length cur_len: %d, add_len: %d\n",
	      out_length, resp_len);

	/* Increment the out_length */
	out_length += resp_len + 2;

	TRACE(TRACE_ENTER_LEAVE, "Leave check_out_length\n");

	if (out_length > MAX_TEXT_LEN)
		return 1;
	else
		return 0;
}

#ifdef RDMA_MODE
int handle_send_rdma(iser_send_control_func_t send_control,
		     void *conn,
		     struct semaphore *send_done_sem,
		     struct msghdr *msg, int count)
{
	int retval = 1;

	/* For transports like IB that require the login phase to use RDMA too */
	struct iser_send_pdu_t send_pdu;
	struct iovec *iov = msg->msg_iov;	/* iov already setup before */
	iser_status_t status;

	send_pdu.p_bhs = (union iser_pdu_bhs_t *) iov->iov_base;
	/* iovec len doesn't include bhs */
	send_pdu.data.tx.buf.size = count - 1;
	send_pdu.data.tx.buf.type = ISER_BUF_TYPE_IOVEC_VIRT;

	if (count > 1) {
		iov++;
		send_pdu.data.tx.buf.p_buf = iov;
	}

	TRACE(TRACE_DEBUG, "handle_send_rdma send_pdu.p_buf:%p iov:%p "
	      "iov_base:%p\n", send_pdu.data.tx.buf.p_buf, iov,
	      iov->iov_base);

	status = send_control((iser_conn_handle_t) conn, &send_pdu);
	if (status != ISER_SUCCESS)
		retval = -1;
#if 0
	/* Wait for send to actually go out (till send_control finished) - This is
	 * not neeeded as the iovec is copied by iSER - Remove when code is stable*/
	if (down_interruptible(send_done_sem) != 0)
		retval = -1;
#endif

	return retval;
}

#endif

/*
 * Returns 0 if pdu sent ok, <0 if hit error during the send
 */
int
handle_send(struct socket *sock, struct msghdr *msg, int send_length,
	    int count)
{
	int retval, total_tx, tx_loop;
#ifdef __KERNEL__
	mm_segment_t oldfs;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	total_tx = 0;
	do {
		TRACE(TRACE_NET,
		      "Sending %d bytes, %d actual_iovs on sock %p\n",
		      send_length, count, sock);
#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
		/*
		 * Added for iSER - Minti
		 * All conn. will use sock_sendmsg since a transition to iSER
		 * mode occurs only in FPP (see iSER Section 7.1).
		 * Use byte stream mode for ALL login PDUs (see DA 11.3.10)
		 */

		tx_loop = sock_sendmsg(sock, msg, send_length);
		set_fs(oldfs);
#else
		TRACE(TRACE_NET, "%s Call writev, fd %d\n",
		      current->comm, sock->fd);
		tx_loop = writev(sock->fd, msg->msg_iov, msg->msg_iovlen);
		TRACE(TRACE_NET, "%s Done writev, fd %d, tx_loop %d\n",
		      current->comm, sock->fd, tx_loop);
#endif

		if (tx_loop <= 0) {
			TRACE_ERROR("sock_sendmsg error %d, total_tx %d\n",
				    tx_loop, total_tx);
			retval = tx_loop;
			goto out;
		}
		total_tx += tx_loop;

		TRACE(TRACE_NET, "tx_loop = %d, total_tx: %d, data: %d\n",
		      tx_loop, total_tx, send_length);
		if (total_tx != send_length) {
			TRACE_ERROR
			    ("sock_sendmsg total_tx %d, expected %d\n",
			     total_tx, send_length);
		}
	} while (total_tx < send_length);
	retval = 0;
      out:
	return retval;
}

/*
 * sends the PDUs to the other end.
 * Returns 0 if pdu sent ok, <0 if hit error during the send
 */
int iscsi_send_msg(struct socket *sock,
#ifdef RDMA_MODE
		   iser_send_control_func_t send_control,
		   void *conn, struct semaphore *send_done_sem,
#endif
		   struct generic_pdu *outputpdu, int flags)
{
	struct msghdr msg;
	struct iovec *iov, actual_iov[3];

	struct iscsi_targ_login_rsp *targ_login_rsp;
	int data_length, count, send_length;
	int retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_send_msg\n");

	/* Can wait for iSER's connection_establish_notify here for efficiency - Minti */
	switch (outputpdu->hdr->opcode & (ISCSI_OPCODE)) {

	case ISCSI_INIT_LOGIN_CMND:
		/* Send the login Command */
		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_init_login_cmnd((struct iscsi_init_login_cmnd
					       *)
					      outputpdu->hdr);
		break;

	case ISCSI_TARG_LOGIN_RSP:
		/* Send the Login Response */
		targ_login_rsp =
		    (struct iscsi_targ_login_rsp *) outputpdu->hdr;

		TRACE(TRACE_ISCSI,
		      "Send Login Response, CSG %d, NSG %d, T %d\n",
		      (outputpdu->hdr->flags & CSG) >> CSG_SHIFT,
		      outputpdu->hdr->flags & NSG,
		      (outputpdu->hdr->flags & T_BIT) >> 7);

		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_targ_login_rsp(targ_login_rsp);
		break;

	default:
		TRACE_ERROR
		    ("sending bad opcode 0x%02X during Login phase\n",
		     outputpdu->hdr->opcode & ISCSI_OPCODE);
		retval = -1;
		goto out;

	}			/* switch */

	/* all pdus store the DSL in the same place */
	outputpdu->hdr->length = cpu_to_be32(outputpdu->text_length);

	/* set up the i/o message and vectors */
	memset(&msg, 0, sizeof(struct msghdr));
	msg.msg_iov = iov = actual_iov;
	msg.msg_flags = MSG_NOSIGNAL;

	count = 1;

	/* NO packets sent or received during login will have
	 * digests of any kind
	 */
	iov->iov_base = (char *) outputpdu->hdr;
	iov->iov_len = send_length = ISCSI_HDR_LEN;

	if ((data_length = outputpdu->text_length) > 0) {
		/* there is data attached to this pdu */

		/* Add padding to end of any attached data */
		data_length += (-data_length) & 3;

		if (data_length > outputpdu->text_length) {
			TRACE(TRACE_NET,
			      "Length with padding = %d, without = %d\n",
			      data_length, outputpdu->text_length);
			memset(outputpdu->text + outputpdu->text_length, 0,
			       data_length - outputpdu->text_length);
		}

		count++;
		iov++;
		iov->iov_base = outputpdu->text;
		iov->iov_len = data_length;
		send_length += data_length;
	}

	/* i/o message and vector now set up */
	msg.msg_iovlen = count;

	TRACE(TRACE_NET, "Sending %d bytes total, %d actual_iovs iov%p\n",
	      send_length, count, actual_iov[0].iov_base);

#ifdef RDMA_MODE
	if (flags & USE_RDMA_LOGIN) {
		retval =
		    handle_send_rdma(send_control, conn, send_done_sem,
				     &msg, count);
	} else
#endif
		retval = handle_send(sock, &msg, send_length, count);


	if (retval < 0)
		goto out;

	TRACE_BUFFER(TRACE_BUF, (char *) outputpdu->hdr, ISCSI_HDR_LEN,
		     "Sent header, length: %d\n", ISCSI_HDR_LEN);

	if (count > 1) {
		TRACE_BUFFER(TRACE_BUF, (char *) outputpdu->text,
			     data_length, "Sent data, length: %d\n",
			     data_length);
	}
	retval = 0;
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_send_msg retval %d\n",
	      retval);
	return retval;
}

/*
 * checks whether the receiving side has received all the responses
 * for the negotiations it started at any time during the current login stage
 * (including the current outgoing pdu, if any).
 * Return:
 *     -1 if error, didn't receive all the responses it expects
 *	0 if success
 */
int
check_neg_responses(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		    __u32 print_error)
{
	int i;
	int retval = 0;
	struct parameter_type *p;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_neg_responses\n");

	for (i = 0; i < MAX_CONFIG_PARAMS; i++) {
		p = &p_param_tbl[i];

		if (IS_KEY_TO_BE_NEGOTIATED(p->neg_info)
		    && IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)
		    && !IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info)
		    && !IS_KEY_REPLY_OPTIONAL(p->neg_info)) {
			/* parameter was sent, no reply was received yet,
			 * and reply is not optional
			 */
			if (!IS_INFORMATIONAL_PARAM(p->type)
			    && !(p->
				 special_key_flag &
				 MAXRECVDATASEGMENTLENGTH_FLAG)) {
				/* informational parameters expect no
				 * response and neither does
				 * MaxRecvPDULength/MaxRecvDataSegmentLength
				 */

				TRACE(TRACE_NEGOTIATIONS,
				      "response not yet received "
				      "for parameter %s\n",
				      p->parameter_name);
				if (print_error) {
					TRACE_ERROR
					    ("response expected for "
					     "parameter " "%s\n",
					     p->parameter_name);
					/* set "got" bit so we won't see this
					 * message again
					 */
					p->neg_info |=
					    KEY_GOT_FROM_OTHER_SIDE;
				}
				retval = -1;
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave check_neg_responses, retval = %d\n", retval);

	return retval;
}

/*
 * checks whether the receiving side supports one of the parameter values
 * offered by the sender.
 * Return:
 *	If supported, returns a pointer into the "supplied_values_from_sender"
 *	list (and its length in "len") of the first item in p's value_list
 *	that matches any item in "supplied_values_from_sender" list.
 *	(note that this value may be terminated by a comma ',' or null '\0')
 *	else returns NULL and len is 0
 */
static const char * __attribute__ ((no_instrument_function))
    check_for_support(struct parameter_type *p,
		  const char *supplied_values_from_sender, __u32 * len)
{
	char *r_ptr, *s_ptr;
	__u32 r_len, s_len;
	const char *sender_value;
	const char *receiver_value;

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter check_for_support, sender_value %s\n",
	      supplied_values_from_sender);

	*len = 0;
	sender_value = NULL;

	/* No support */
	if ((receiver_value = p->value_list) == NULL)
		goto out;

	TRACE(TRACE_NEGOTIATIONS, "%s's value_list: %s\n",
	      p->parameter_name, receiver_value);

	/* Check whether the parameter's str_value agrees with *
	 * one of the supplied values */

	do {
		/* Get the receiver_value from the parameter's value_list */
		r_ptr = strchr(receiver_value, ',');
		if (r_ptr) {
			r_len = r_ptr - receiver_value;
		} else {
			r_len = strlen(receiver_value);
		}

		TRACE(TRACE_NEGOTIATIONS, "receiver_value: %s\n",
		      receiver_value);

		/* Initialise */
		sender_value = supplied_values_from_sender;

		do {
			/* extract just a single value */
			s_ptr = strchr(sender_value, ',');
			if (s_ptr) {
				s_len = s_ptr - sender_value;
			} else {
				s_len = strlen(sender_value);
			}

			TRACE(TRACE_NEGOTIATIONS, "sender_value: %s\n",
			      sender_value);

			if (r_len == s_len
			    && !strncmp(receiver_value, sender_value,
					r_len)) {
				/* Found a match, return pointer to it
				   in sender_value */
				*len = r_len;
				goto out;
			}

			/* goto the next value supplied by sender */
			if (s_ptr) {
				s_ptr++;	/* skip over comma */
			}
			sender_value = s_ptr;
		} while (sender_value);

		/* skip over the comma and goto the next value in
		   receiver's table list */
		if (r_ptr) {
			r_ptr++;
		}
		receiver_value = r_ptr;
	} while (receiver_value);

      out:

	if (sender_value) {
		TRACE(TRACE_ENTER_LEAVE,
		      "Leave check_for_support, len %u, return value %s\n",
		      *len, sender_value);
	} else {
		TRACE(TRACE_ENTER_LEAVE,
		      "Leave check_for_support, len %u, return value NULL\n",
		      *len);
	}

	return sender_value;
}

static void __attribute__ ((no_instrument_function))
    update_key_value(struct parameter_type *p, int int_value,
		 const char *value)
{
	if (IS_NUMBER(p->type) || IS_NUMBER_RANGE(p->type)) {
		if (p->int_value != int_value) {
			/* have a new numeric value for this key */
			p->int_value = int_value;
			TRACE(TRACE_ISCSI, "Update key %s, new value %d\n",
			      p->parameter_name, p->int_value);
		}
	} else if (p->str_value && !strcmp(value, p->str_value)) {
		/* already have this string value in place */
	} else {
		/* have a new string value for this key */
		strreplace(&p->str_value, value);
		TRACE(TRACE_ISCSI, "Update key %s, new value %s\n",
		      p->parameter_name, p->str_value);
	}
}

static void __attribute__ ((no_instrument_function))
    handle_boolean_param(struct parameter_type *p, const char **value);

static int __attribute__ ((no_instrument_function))
    handle_params_noexch(struct parameter_type *p,
		     struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		     __u32 flags,
		     int int_value,
		     const char *value,
		     char *string,
		     int *resp_len,
		     int *FBLength,
		     struct parameter_type **FBp,
		     int *MBLength,
		     struct parameter_type **MBp, __u64 * login_flags);

static int __attribute__ ((no_instrument_function))
    handle_params_resp(struct parameter_type *p,
		   const char *value, int int_value, int *out_length);

#ifdef ISER
/* Force specified keys to No/None are iSER specifies -
 * !!!! In addition, also make
 * DataSequenceInOrder and DataPDUInOrder Yes so that we don't need to perform
 * the merge_offset_length & check_range_list_complete functions
 */
int handle_iser_keys(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
	struct parameter_type *p;
	int flag = 0;

	/* DataSequenceInOrder */
	if ((p =
	     find_flag_parameter(DATASEQUENCEINORDER_FLAG,
				 p_param_tbl)) != NULL) {
		if (!strcmp(p->str_value, no_word)) {
			/* change value to Yes */
			flag = 1;
			strreplace(&p->str_value, yes_word);
			TRACE_WARNING("%s=No with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}
	/* DataPDUInOrder */
	if ((p =
	     find_flag_parameter(DATAPDUINORDER_FLAG,
				 p_param_tbl)) != NULL) {
		if (!strcmp(p->str_value, no_word)) {
			/* change value to Yes */
			flag = 1;
			strreplace(&p->str_value, yes_word);
			TRACE_WARNING("%s=No with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}

	/* IFMarker */
	if ((p = find_flag_parameter(IFMARKER_FLAG, p_param_tbl)) != NULL) {
		if (!strcmp(p->str_value, yes_word)) {
			/* have IFMarker=Yes, change value to No */
			flag = 1;
			strreplace(&p->str_value, no_word);
			TRACE_WARNING("%s=Yes with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}
	/* OFMarker */
	if ((p = find_flag_parameter(OFMARKER_FLAG, p_param_tbl)) != NULL) {
		if (!strcmp(p->str_value, yes_word)) {
			/* have OFMarker=Yes, change value to No */
			flag = 1;
			strreplace(&p->str_value, no_word);
			TRACE_WARNING("%s=Yes with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}
	/* HeaderDigest */
	if ((p = find_flag_parameter(HEADERDIGEST_FLAG,
				     p_param_tbl)) != NULL) {

		if (strcmp(p->str_value, none_word)) {
			/* have HeaderDigest!=None, change value to None */
			flag = 1;
			strreplace(&p->str_value, none_word);
			/* Fix value_list too so that we respond correctly to sender keys */
			strreplace(&p->value_list, none_word);
			TRACE_WARNING("%s!=None with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}
	/* DataDigest */
	if ((p = find_flag_parameter(DATADIGEST_FLAG,
				     p_param_tbl)) != NULL) {

		if (strcmp(p->str_value, none_word)) {
			/* have DataDigest!=None, change value to None */
			flag = 1;
			strreplace(&p->str_value, none_word);
			strreplace(&p->value_list, none_word);
			TRACE_WARNING("%s!=None with RDMAExtensions=Yes"
				      " Resetting %s=%s\n",
				      p->parameter_name, p->parameter_name,
				      p->str_value);
		}
	}
	return flag;
}
#endif

static void __attribute__ ((no_instrument_function))
    set_discovery_flag(struct parameter_type *p, const char *value,
		   __u64 * login_flags)
{
	if (p->special_key_flag & SESSIONTYPE_FLAG) {
		/* set the DISCOVERY_FLAG when have SessionType=Discovery */
		if (!strcmp(value, discovery_word)) {
			*login_flags |= DISCOVERY_FLAG;
		} else {
			*login_flags &= ~DISCOVERY_FLAG;
		}
	}
}

/*
 * scan through the input and process keys
 * p_param_tbl - pointer to the table of parameters
 * sock - pointer to the socket structure
 * process_these_types - types of parameters to be processed.
 * flags_to_be_set - Which flags to set
 * role - who is calling (INITIATOR/TARGET)
 * Return: returns the length of the string added to the output string
 */
int
scan_input_and_process(struct socket *sock,
		       struct parameter_type
		       p_param_tbl[MAX_CONFIG_PARAMS],
		       int process_these_types, __u64 flags_to_be_set,
		       int role, __u32 * max_send_length,
		       __u32 when_called, char *text, int text_length,
		       struct generic_pdu *outputpdu, __u32 flags,
		       __u64 * login_flags,
		       struct unknown_key **unknown_key_list)
{
	struct parameter_type *p = NULL;
	int resp_len, n;
	char *dummy_string;
	char *input_str, *this_key;
	char *input_string;
	int out_length = 0;
	char *output_string;
	int key_value_len;
	const char *value;
	unsigned int int_value;
	char *last_input;
	int FBLength = -1, MBLength = -1;
	struct parameter_type *FBp = NULL, *MBp = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter scan_input_and_process\n");

	ALLOCATE_MAX_TEXT_LEN(dummy_string);
	ALLOCATE_MAX_TEXT_LEN(input_str);

	output_string = outputpdu->text + outputpdu->text_length;

	/* Copy the input text to a local buffer */
	memcpy(input_str, text, text_length);
	input_string = input_str;	/* point at first char */
	last_input = input_str + text_length;	/* point 1 beyond last char */
	*last_input = '\0';	/* ensure '\0' after last char */

	/* each time around this loop deal with one input key=value string */
	while (input_string < last_input) {
		key_value_len = strlen(input_string);
		TRACE(TRACE_NEGOTIATIONS, "key_value_len = %d\n",
		      key_value_len);
		TRACE(TRACE_NEGOTIATIONS, "FBLength %x MBLength %x\n",
		      FBLength, MBLength);

		/* assume no reply will be sent */
		resp_len = 0;

		value = NULL;
		int_value = -1;
		if ((p =
		     check_correctness(input_string, &value, p_param_tbl,
				       role, when_called, flags,
				       &int_value, login_flags,
				       unknown_key_list))
		    == NULL) {

			/* We don't understand the request. */
			if (value == NULL) {	/* Error is fatal */
				out_length = -1;
				goto out;
			}
			if (!(process_these_types & SECURITY_PARAM)
			    || !is_securitykey(input_string)) {
				resp_len = sprintf(dummy_string,
						   "%s=%s", input_string,
						   value);
			}
		} else if (IS_KEY_BAD(p->neg_info)) {
			out_length = -1;
			goto out;
		} else if (!IS_KEY_IRRELEVANT(p->neg_info)) {
			/* p is the pointer to the table entry for the
			 * received parameter and value points to the
			 * string following the "=" in the input
			 */
			TRACE(TRACE_NEGOTIATIONS,
			      "Process input param: %s, value: %s\n",
			      p->parameter_name, value);

			/* check if this is a Full Feature Only key */
			if (IS_FULL_FEATURE_ONLY(p->type)) {
				/* protocol error */
				TRACE_ERROR
				    ("%p only valid in full feature phase\n",
				     p->parameter_name);
				out_length = -1;
				goto out;
			}

			if (p->special_key_flag & flags_to_be_set) {
				/* caller wants to know that this key
				 * was processed
				 */
				*login_flags |= p->special_key_flag;

				set_discovery_flag(p, value, login_flags);
			}

			if (p->type & process_these_types) {
				/* handle key enquiry by replying with
				   our current value */
				if (!strcmp(value, "?")) {
					/* inquiry no longer accepted in
					   draft 11 and after */
					TRACE_ERROR
					    ("inquiry no longer accepted "
					     "param: %s, value: %s\n",
					     p->parameter_name, value);
					resp_len =
					    sprintf(dummy_string, "%s=%s",
						    p->parameter_name,
						    reject_word);
					/* do not consider this param as
					   "gotten from other side" */
					p = NULL;
				} else if (p->str_value
					   && !strcmp(p->str_value, "?")) {
					/* we sent an inquiry and this is
					 * the response
					 * Note this should never happen
					 * for draft 11 and after
					 */
					printk
					    ("iscsi response to inquiry: "
					     "%s=%s\n", p->parameter_name,
					     value);
				} else {
					/* neither an inquiry or a reply to
					 * an inquiry .
					 * First convert boolean to yes or
					 * no result
					 */
					if (IS_BOOLEAN(p->type)) {
						handle_boolean_param(p,
								     &value);
					}

					if (p->special_key_flag &
					    MAXRECVDATASEGMENTLENGTH_FLAG)
					{
						/* this is
						 * MaxRecvDataSegmentLength
						 * that other side wants to
						 * receive, i.e., the max
						 * size PDU we should send.
						 * No reply is necessary for
						 * this key.
						 */
						if (int_value < 512
						    || int_value >
						    16777215) {
							if (TRACE_TEST
							    (TRACE_ISCSI))
							{
								TRACE_WARNING
								    ("%s %d out of "
								     "bounds "
								     "[512..16777215]\n",
								     p->
								     parameter_name,
								     int_value);
							}
						} else {
							/* value is within
							 * bounds, pass it back
							 * to caller after
							 * forcing it to be a
							 * multiple of 512
							 */
							int_value /= 512;
							*max_send_length =
							    512 *
							    int_value;
						}
					} else
					    if (!IS_KEY_SENT_TO_OTHER_SIDE
						(p->neg_info)) {
						handle_params_noexch(p,
								     p_param_tbl,
								     flags,
								     int_value,
								     value,
								     dummy_string,
								     &resp_len,
								     &FBLength,
								     &FBp,
								     &MBLength,
								     &MBp,
								     login_flags);

					} else
					    if (handle_params_resp
						(p, value, int_value,
						 &out_length) < 0) {
						goto out;
					}
				}
			} else {
				if (IS_SECURITY_PARAM(p->type)) {
					TRACE_ERROR
					    ("not in security phase when "
					     " received input "
					     "param: %s, value: %s\n",
					     p->parameter_name, value);
					out_length = -1;
					goto out;
				} else if (IS_OPERATIONAL_PARAM(p->type)) {
					TRACE_ERROR
					    ("not in operational phase when "
					     " received input "
					     "param: %s, value: %s\n",
					     p->parameter_name, value);
					out_length = -1;
					goto out;
				} else if (IS_INFORMATIONAL_PARAM(p->type)) {
					TRACE_ERROR
					    ("not processing informational "
					     " parameters when "
					     "recv input param: %s,value: %s\n",
					     p->parameter_name, value);
					out_length = -1;
					goto out;
				} else {	/* should never happen */
					TRACE_ERROR
					    ("recv unclassied input param: %s, "
					     "value: %s\n",
					     p->parameter_name, value);
					out_length = -1;
					goto out;
				}
			}
		}

		/* Always check if the response length exceeds MAX_TEXT_LEN */
		/* Then the remaining parameters have to be sent in a     */
		/* separate text command/response */

		TRACE(TRACE_NEGOTIATIONS, "resp_len = %d\n", resp_len);

		if (resp_len) {
			if (check_out_length(out_length, resp_len)) {
				*login_flags |= MORE_TO_SEND_FLAG;
				break;
			} else {
				/* Update resp_len & pointer to output_string */
				sprintf(output_string, "%s", dummy_string);
				output_string += resp_len + 1;

				TRACE(TRACE_ISCSI, "Attach key: %s\n",
				      dummy_string);

				/* update out_length */
				out_length += resp_len + 1;

				if (p) {	/* set sent-to-other-side bit */
					p->neg_info |=
					    KEY_SENT_TO_OTHER_SIDE;
				}

				resp_len = 0;
			}
		}

		if (p) {	/* set got-from-other-side bit */
			p->neg_info |= KEY_GOT_FROM_OTHER_SIDE;
		}

		/* Goto the next key=value pair */
		/* remember where this key started */
		this_key = input_string;
		/* this should be start of next key */
		input_string += key_value_len + 1;

		/*  In Drafts 9 and before, Section 3.10.4 Text
		 *  "Every key=value pair (including the last or only pair) MUST
		 *  be followed by at least one null (0x00) delimiter."
		 *
		 *  This implied there could be more than one null (0x00) at the
		 *  end of a string, so skip over these extra nulls now
		 */
		while (*input_string == '\0' && input_string < last_input) {
			input_string++;
		}

		/*  RFC 3720 5.1 Text Format
		 *  "Every key=value pair, including the last or only pair in
		 *  a LTDS,  MUST be followed by one null (0x00) delimiter."
		 *
		 *  This implies that there can NOT be more than one null (0x00)
		 *  at the end of a string, so if there were extra nulls,
		 *  give a warning now.
		 */
		n = (input_string - this_key) - (key_value_len + 1);
		if (n > 0) {
			TRACE_WARNING("%d extra nulls (0x00) found after"
				      " key \"%s\"\n", n, this_key);
		}
	}

	/* process FirstBurstLength and MaxBurstLength if their
	   reply was delayed */
	TRACE(TRACE_NEGOTIATIONS, "FBLength %x MBLength %x\n", FBLength,
	      MBLength);

	if (FBLength > 0 && MBLength > 0) {
		/* both keys were offered by other side in this pdu */
		if (FBLength <= MBLength) {
			/* ok */
			/* MaxBurstLength is bigger than
			   FirstBurstLength, all ok */
		} else {
			/* FirstBurstLength bigger than MaxBurstLength,
			   fix it */

			FBp->int_value = FBLength = MBLength;
			TRACE(TRACE_NEGOTIATIONS, "Updated FBp %p\n", FBp);
		}
	} else if (FBLength > 0) {
		/* only FirstBurstLength was received in this pdu */
		if ((MBp =
		     find_flag_parameter(MAXBURSTLENGTH_FLAG,
					 p_param_tbl)) != NULL) {
			if (FBLength > MBp->int_value) {
				/* The received FirstBurstLength is
				   bigger than our current
				   MaxBurstLength */
				if (IS_KEY_SENT_TO_OTHER_SIDE
				    (MBp->neg_info)) {
					/* MaxBurstLength already sent,
					   can't be fixed */
					TRACE_ERROR
					    ("FirstBurstLength %d bigger "
					     " than negotiated "
					     "MaxBurstLength %d", FBLength,
					     MBp->int_value);
					out_length = -1;
					goto out;
				}
				/* The unsent MaxBurstLength is smaller
				 * than the received FirstBurstLength,
				 * reduce the FirstBurstLength to the
				 * MaxBurstLength and send both
				 */
				FBLength = FBp->int_value = MBLength =
				    MBp->int_value;
				TRACE(TRACE_NEGOTIATIONS,
				      "Updated FBLength %u\n", FBLength);
			} else {
				/* The received FirstBurstLength not
				 * bigger than our current
				 * MaxBurstLength, send MaxBurstLength only if
				 * negotiation needed.
				 */
				if ((IS_KEY_TO_BE_NEGOTIATED(MBp->neg_info)
				     || IS_KEY_GOT_FROM_OTHER_SIDE(MBp->
								   neg_info))
				    && !IS_KEY_SENT_TO_OTHER_SIDE(MBp->
								  neg_info))
				{
					MBLength = MBp->int_value;
				}
			}
		}
	} else if (MBLength > 0) {
		/* only MaxBurstLength was received in this pdu */
		if ((FBp = find_flag_parameter(FIRSTBURSTLENGTH_FLAG,
					       p_param_tbl)) != NULL) {
			/* must have MaxBurstLength > FirstBurstLength */
			if (MBLength < FBp->int_value) {
				/* MaxBurstLength is smaller than
				   FirstBurstLength */
				if (IS_KEY_SENT_TO_OTHER_SIDE
				    (FBp->neg_info)) {
					/* FirstBurstLength already sent,
					   can't be fixed */
					TRACE_ERROR
					    ("MaxBurstLength %d smaller than "
					     " negotiated "
					     "FirstBurstLength %d",
					     MBLength, FBp->int_value);
					out_length = -1;
					goto out;
				}
				/* FirstBurstLength has not been sent yet,
				 * so force it to be equal to
				 * MaxBurstLength and offer it now
				 */
				FBLength = FBp->int_value = MBLength;
				TRACE(TRACE_NEGOTIATIONS,
				      "Updated FBLength %u\n", FBLength);
			} else {
				/* MaxBurstLength not smaller than
				 * FirstBurstLength, send
				 * FirstBurstLength only if it still needs
				 * to be negotiated
				 */
				if ((IS_KEY_TO_BE_NEGOTIATED(FBp->neg_info)
				     || IS_KEY_GOT_FROM_OTHER_SIDE(FBp->
								   neg_info))
				    && !IS_KEY_SENT_TO_OTHER_SIDE(FBp->
								  neg_info))
				{
					FBLength = FBp->int_value;
				}
			}
		}
	}

	if (FBLength > 0) {
		/* reply to or offer the FirstBurstLength */
		TRACE(TRACE_NEGOTIATIONS, "FBLength %x FBp %p\n", FBLength,
		      FBp);
		FBp->neg_info |= KEY_SENT_TO_OTHER_SIDE;
		if (!IS_KEY_GOT_FROM_OTHER_SIDE(FBp->neg_info))
			outputpdu->hdr->flags &= (~T_BIT);
		resp_len =
		    sprintf(dummy_string, "%s=%d", firstburstlength_word,
			    FBLength);
		sprintf(output_string, "%s", dummy_string);
		output_string += resp_len + 1;

		TRACE(TRACE_ISCSI, "Attach key, %s\n", dummy_string);

		/* update out_length */
		out_length += resp_len + 1;
		resp_len = 0;
		FBLength = -1;
	}

	if (MBLength > 0) {
		/* reply to or offer the MaxBurstLength */
		TRACE(TRACE_NEGOTIATIONS, "MBLength %x MBp %p\n", MBLength,
		      MBp);
		MBp->neg_info |= KEY_SENT_TO_OTHER_SIDE;
		if (!IS_KEY_GOT_FROM_OTHER_SIDE(MBp->neg_info))
			outputpdu->hdr->flags &= (~T_BIT);
		resp_len =
		    sprintf(dummy_string, "%s=%d", maxburstlength_word,
			    MBLength);
		sprintf(output_string, "%s", dummy_string);
		output_string += resp_len + 1;

		TRACE(TRACE_ISCSI, "Attach key, %s\n", dummy_string);

		/* update out_length */
		out_length += resp_len + 1;
		MBLength = -1;
	}

      out:
	FREE_STRING(dummy_string);
	FREE_STRING(input_str);

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave scan_input_and_process, out_length %d\n", out_length);

	/* return the length added to the output_string */
	return out_length;
}

/*
 * Called when a boolean value has been received from the other side,
 * either as a new offer or as a reply to a previous offer from us.
 * When called, "value" is known to point to either "Yes" or "No" as
 * received from the other side, and "p" points to the key in our table
 * (which contains our desired value).
 * This routine applies the correct boolean result function to the
 * received value and the value in our table, and sets "value" to
 * point to the correct result ("Yes" or "No").
 * In addition, if a reply with this result is optional, KEY_REPLY_OPTIONAL
 * is set in p->neg_info.
 * If this function was called because of a reply to a previous offer
 * and the reply is not allowed, KEY_BAD is set in p->neg_info.
 */

static void __attribute__ ((no_instrument_function))
    handle_boolean_param(struct parameter_type *p, const char **value)
{
	int receiver_value;
	int sender_value;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_boolean_param %s=%s\n",
	      p->parameter_name, *value);

	/* convert the received Yes/No string to local 1/0 number */
	if (!strcmp(*value, yes_word))
		sender_value = 1;
	else
		sender_value = 0;

	/* convert the Yes/No string in our table to local 1/0 number */
	if (!strcmp(p->str_value, yes_word))
		receiver_value = 1;
	else
		receiver_value = 0;

	if (IS_BOOL_AND(p->type)) {
		/* result function for this key is Boolean AND */
		if (sender_value & receiver_value) {
			/* both sides say YES, so result is YES */
			*value = yes_word;
		} else {
			/* either (or both) of the sides say NO */
			*value = no_word;
			if (sender_value == 0) {
				/* sender says NO, so reply is optional
				 *
				 *  RFC 3720 Section 5.2.2 Simple-value
				 *  Negotiations
				 *  "Specifically, the two cases in
				 *  which answers are OPTIONAL are:
				 *
				 *  - The boolean function is "AND" and
				 *    the value "No" is received. The
				 *    outcome of the negotiation is "No".
				 *
				 *  - The boolean function is "OR" and
				 *    the value "Yes" is received. The
				 *    outcome of the negotiation is "Yes".
				 *
				 *  Responses are REQUIRED in all other
				 *  cases, and the value chosen and
				 *  sent by the acceptor becomes the
				 *  outcome of the  negotiation."
				 */
				p->neg_info |= KEY_REPLY_OPTIONAL;
			} else if (IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
				/* sender sent YES to our offer of NO -- BAD */
				TRACE_ERROR
				    ("received 'Yes' to our offer of"
				     " 'No' for AND function key \"%s\"\n",
				     p->parameter_name);
				p->neg_info |= KEY_BAD;
			}
		}
	} else {
		/* result function for this key is Boolean OR */
		if (sender_value | receiver_value) {
			/* either (or both) of the sides say YES */
			*value = yes_word;
			if (sender_value != 0) {
				/* sender says YES, so reply is optional
				 *
				 * RFC 3720 Section 5.2.2 Simple-value
				 * Negotiations
				 *  "Specifically, the two cases in
				 *  which answers are OPTIONAL are:
				 *
				 *  - The boolean function is "AND"
				 *    and the value "No" is received.
				 *    The outcome of the negotiation is "No".
				 *
				 *  - The boolean function is "OR"
				 *    and the value "Yes" is received.
				 *    The outcome of the negotiation is "Yes".
				 *
				 *  Responses are REQUIRED in all other
				 *  cases, and the value chosen and
				 *  sent by the acceptor becomes the
				 *  outcome of the negotiation."
				 */
				p->neg_info |= KEY_REPLY_OPTIONAL;
			} else if (IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {
				/* sender sent NO to our offer of YES -- BAD */
				TRACE_ERROR
				    ("received 'No' to our offer of "
				     "'Yes' for OR funct key \"%s\"\n",
				     p->parameter_name);
				p->neg_info |= KEY_BAD;
			}
		} else {
			/* both of the sides say NO so result is NO */
			*value = no_word;
		}
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_boolean_param\n");
}

/*
 * Called to store in "string" our reply to a key received from other side.
 * The resultant length is passed back in resp_len.
 * If the offered key is FirstBurstLength or MaxBurstLength,
 * then no reply is created at this time, but the offered value is passed back
 * in FBLength or MBLength and the pointer to the table entry for this key
 * is passed back in FBp or MBp.  This is done to delay the actual response
 * until both values are known, because must have FirstBurst <= MaxBurst.
 */
static int __attribute__ ((no_instrument_function))
    handle_params_noexch(struct parameter_type *p,
		     struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		     __u32 flags,
		     int int_value,
		     const char *value,
		     char *string,
		     int *resp_len,
		     int *FBLength,
		     struct parameter_type **FBp,
		     int *MBLength,
		     struct parameter_type **MBp, __u64 * login_flags)
{
	struct parameter_type *p1 = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_params_noexch\n");

	if (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info)) {

		/* parameter was not sent previously by us, so this
		 * is a new offer from other side to us, formulate our reply
		 */
		if (IS_INFORMATIONAL_PARAM(p->type)) {
			/* offered parameter is declarative, send no reply */
			update_key_value(p, int_value, value);
		} else if ((*login_flags & DISCOVERY_FLAG)
			   && IS_IRRELEVANT_IN_DISCOVERY(p->type)
			   && !(flags & ALLOW_REPLIES_WHEN_IRRELEVANT)) {
			/* offered parameter is irrelevant in Discovery Session
			 * and we do not allow normal replies when they would be
			 * irrelevant, so tell sender this key is irrelevant
			 */
			*resp_len =
			    sprintf(string, "%s=%s", p->parameter_name,
				    irrelevant_word);
		} else if (IS_NUMBER_RANGE(p->type)) {
			/* offered parameter's value should be a
			 * range <key=value1~value2>
			 */
			int send_irrelevant = 0;

			if (p->special_key_flag & OFMARKINT_FLAG) {
				/* reply with Irrelevant if OFMARKER is no */
				if ((p1 =
				     find_flag_parameter(OFMARKER_FLAG,
							 p_param_tbl)) !=
				    NULL) {
					if (!strcmp
					    (p1->str_value, no_word)) {
						send_irrelevant = 1;
					}
				}
			} else if (p->special_key_flag & IFMARKINT_FLAG) {
				/* reply with Irrelevant if IFMARKER is no */
				if ((p1 = find_flag_parameter
				     (IFMARKER_FLAG, p_param_tbl))
				    != NULL) {
					if (!strcmp
					    (p1->str_value, no_word)) {
						send_irrelevant = 1;
					}
				}
			}

			if (send_irrelevant) {
				*resp_len = sprintf(string, "%s=%s",
						    p->parameter_name,
						    irrelevant_word);
			} else if (int_value >= 0) {
				/* keep the offered lower end of range */
				update_key_value(p, int_value, value);
				/* always need a reply to range parameter */
				*resp_len = sprintf(string, "%s=%d",
						    p->parameter_name,
						    p->int_value);
			}
		} else if (!IS_NUMBER(p->type)) {
			/* offered parameter is not informational,
			 * numeric-range or number, must be boolean or string.
			 * check that we support the offered value
			 */
			const char *param_value = NULL;
			char *copy;
			__u32 len;

			if (IS_BOOLEAN(p->type)) {
				/* offered param is boolean, support assured */
				param_value = value;
				len = 0;
			} else {
				param_value =
				    check_for_support(p, value, &len);
			}

			if (param_value != NULL) {
				/* value received is supported by us, use it */
				if (len && param_value[len]) {
					/* param_value is not null-terminated */
					copy =
					    kmalloc(len + 1, GFP_KERNEL);
					if (copy) {
						strncpy(copy, param_value,
							len);
						copy[len] = '\0';
						param_value = copy;
					}
				} else {
					copy = NULL;
				}
				if (!IS_KEY_REPLY_OPTIONAL(p->neg_info)
				    || (flags & USE_FULL_REPLIES)) {
					/* reply not optional or full reply
					 * required
					 */
					*resp_len =
					    sprintf(string, "%s=%s",
						    p->parameter_name,
						    param_value);
				} else {
					/* optional reply not
					 * sent but marked as
					 * sent
					 */
					p->neg_info |=
					    KEY_SENT_TO_OTHER_SIDE;
				}
				update_key_value(p, int_value,
						 param_value);
				if (copy)
					kfree(copy);
			} else {
				/* value not supported by us, send reject */
				*resp_len = sprintf(string, "%s=%s",
						    p->parameter_name,
						    reject_word);
			}
		} else {
			/* offered parameter is a number, apply its
			 * result function (if it has one)
			 */
			if ((IS_MIN_NUMBER(p->type)
			     && p->int_value <= int_value)
			    || (IS_MAX_NUMBER(p->type)
				&& p->int_value >= int_value)) {
				/* correct min or max already in place */
			} else {
				/* the offered value is the one we keep */
				update_key_value(p, int_value, value);
			}

			/*  Delay reply to FirstBurstLength & MaxBurstLength
			 *  until end of the PDU to ensure
			 *  FirstBurstLength <= MaxBurstLength
			 */
			if (p->special_key_flag & FIRSTBURSTLENGTH_FLAG) {
				*FBLength = p->int_value;
				*FBp = p;
				TRACE(TRACE_NEGOTIATIONS,
				      "Updated FBp %p\n", FBp);
				/* reply to this key at end of this pdu */
			} else if (p->
				   special_key_flag & MAXBURSTLENGTH_FLAG)
			{
				*MBLength = p->int_value;
				*MBp = p;
				TRACE(TRACE_NEGOTIATIONS,
				      "Updated MBp %p\n", MBp);
				/* reply to this key at end of this pdu */
			} else {
				/* always need a reply to numeric parameter */
				*resp_len = sprintf(string, "%s=%d",
						    p->parameter_name,
						    p->int_value);
			}
		}
		TRACE(TRACE_ENTER_LEAVE, "Leave handle_params_noexch\n");
		return 1;
	}
	return -1;
}

/*
 * Called to deal with response from other side to parameter we previously sent.
 * input parameter "value" or "int_value" contains value sent by other side.
 * Returns 0 if ok (out_length unchanged),
 *	   -1 on error after error (and sets out_length to -1)
 */
static int __attribute__ ((no_instrument_function))
    handle_params_resp(struct parameter_type *p,
		   const char *value, int int_value, int *out_length)
{
	int retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_params_resp for %s\n",
	      p->parameter_name);

	if (!IS_NUMBER(p->type)) {
		/* received response key is not a number, must be one of:
		 * string, boolean, numeric-range or enumerated symbol
		 */
		const char *param_value = NULL;
		__u32 len;

		/* received response value should just be a single string */
		if (strchr(value, ',')) {
			TRACE_ERROR
			    ("multiple values in negotiation response "
			     "for parameter: %s\n", p->parameter_name);
			goto out_bad;
		}

		if (IS_BOOLEAN(p->type)) {
			/* key is boolean, result already computed in value */
			param_value = value;
		} else if (!IS_NUMBER_RANGE(p->type)) {
			/* key is string or enumerated symbol */
			param_value = check_for_support(p, value, &len);
		}
		if (param_value != NULL) {
			/* received value is supported by this side, use it */
			update_key_value(p, int_value, value);
		} else if (IS_NUMBER_RANGE(p->type)) {
			/* received value must be in range <key=value1~value2> */
			if (IS_KEY_REJECT(p->neg_info)) {
				/* got Reject reply to OFMarkInt/IFMarkInt */
			} else {
				/* received value must be in offered range */
				if (check_range(p->value_list, int_value) <
				    0) {
					TRACE_ERROR
					    ("check_range failed\n");
					goto out_bad;
				}
				if (p->int_value != int_value) {
					/* assign the negotiated value */
					update_key_value(p, int_value,
							 value);
				}
			}
		} else {
			/* received value not among those we offered */
			TRACE_ERROR("bad response %s for key %s\n",
				    value, p->parameter_name);
			goto out_bad;
		}
	} else {
		/* received reply key is a number, int_value has its value */

		/* See if other side applied min/max correctly to our offer */
		if ((p->int_value < int_value && IS_MIN_NUMBER(p->type))
		    || (p->int_value > int_value
			&& IS_MAX_NUMBER(p->type))) {
			/* Error, reply did not do min/max correctly */
			TRACE_ERROR("got reply %d to offer of %d for %s"
				    " parameter %s\n",
				    int_value, p->int_value,
				    IS_MIN_NUMBER(p->type) ? "min" : "max",
				    p->parameter_name);
			goto out_bad;
		} else {
			/* assign the negotiated value */
			update_key_value(p, int_value, value);
		}
	}
	retval = 0;
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_params_resp retval %d\n",
	      retval);
	return retval;
      out_bad:
	*out_length = retval = -1;
	goto out;
}

/*
 * scan through the table and count the keys of each type
 * that still remain to be negotiated
 * p_param_tbl - pointer to the table of parameters
 * returns counts in next 3 parameters
 * Also set DISCOVERY_FLAG in login_flags if SessionType=Discovery will be
 * negotiated.
 */
void
scan_table_and_count(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		     int *nsecurity, int *ninformational,
		     int *noperational, __u64 * login_flags)
{
	struct parameter_type *p;
	int i;

	TRACE(TRACE_ENTER_LEAVE, "Enter scan_table_and_count\n");

	/* initialize all counts to 0 */
	*nsecurity = *ninformational = *noperational = 0;

	for (i = 0, p = p_param_tbl; i < MAX_CONFIG_PARAMS; i++, p++) {
		if ((IS_KEY_TO_BE_NEGOTIATED(p->neg_info))
		    && (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))) {
			/* only count those params that still
			   remain to be negotiated */
			if (IS_SECURITY_PARAM(p->type))
				*nsecurity += 1;
			else if (IS_INFORMATIONAL_PARAM(p->type))
				*ninformational += 1;
			else if (IS_OPERATIONAL_PARAM(p->type))
				*noperational += 1;
			set_discovery_flag(p, p->str_value, login_flags);
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave scan_table_and_count\n");
}

/*
 * scan through the table and find keys that have not been sent to other side
 * but need to be.  If these keys can be processed in this stage, and there is
 * room for them in the output pdu, append them to the end of the output pdu.
 *	p_param_tbl - pointer to the table of parameters
 *	process_these_types - types of parameters to be processed.
 * Returns total length of all strings appended to the output pdu's text string
 */
int
scan_table_and_process(struct socket *sock,
		       struct parameter_type
		       p_param_tbl[MAX_CONFIG_PARAMS],
		       int process_these_types, __u64 flags_to_be_set,
		       int role, struct generic_pdu *outputpdu,
		       __u32 flags, __u64 * login_flags)
{
	struct parameter_type *p;
	int resp_len = 0;
	char *dummy_string;
	int out_length = 0;
	char *output_string;
	int i;

	TRACE(TRACE_ENTER_LEAVE, "Enter scan_table_and_process\n");

	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	output_string = outputpdu->text + outputpdu->text_length;

	for (i = 0, p = p_param_tbl; i < MAX_CONFIG_PARAMS; i++, p++) {
		if ((IS_KEY_TO_BE_NEGOTIATED(p->neg_info))
		    && (p->type & process_these_types)
		    && (!IS_KEY_SENT_TO_OTHER_SIDE(p->neg_info))
		    && !((*login_flags & DISCOVERY_FLAG)
			 && IS_IRRELEVANT_IN_DISCOVERY(p->type)
			 && !(flags & ALLOW_OFFERS_WHEN_IRRELEVANT))) {
			TRACE(TRACE_NEGOTIATIONS,
			      "Process table parameter: %s\n",
			      p->parameter_name);
			/* Add it to the outgoing data */
			if (IS_NUMBER(p->type)) {
				resp_len = sprintf(dummy_string,
						   "%s=%d",
						   p->parameter_name,
						   p->int_value);
			} else if (p->value_list != NULL) {
				resp_len = sprintf(dummy_string,
						   "%s=%s",
						   p->parameter_name,
						   p->value_list);
			} else if (p->str_value != NULL) {
				resp_len = sprintf(dummy_string,
						   "%s=%s",
						   p->parameter_name,
						   p->str_value);
			}

			if (p->special_key_flag & flags_to_be_set) {
				/* caller wants to know that this key
				 * was processed
				 */
				*login_flags |= p->special_key_flag;
			}

			/* Always check if the Text length
			 * exceeds MAX_TEXT_LEN
			 * Then the remaining parameters have to be sent
			 * in a separate login command/response
			 */
			if (check_out_length
			    (outputpdu->text_length + out_length,
			     resp_len)) {
				/* set the more_to_send flag */
				*login_flags |= MORE_TO_SEND_FLAG;
				break;
			} else {
				/* Update resp_len & pointer to output_string */
				strcpy(output_string, dummy_string);
				output_string += resp_len + 1;

				TRACE(TRACE_ISCSI, "Attach key: %s\n",
				      dummy_string);

				/* update out_length */
				out_length += resp_len + 1;

				/* set sent bit */
				p->neg_info |= KEY_SENT_TO_OTHER_SIDE;

				/* in 2 special cases, set reply optional bit */
				if ((IS_BOOL_AND(p->type)
				     && strcmp(p->str_value, no_word) == 0)
				    || (IS_BOOL_OR(p->type)
					&& strcmp(p->str_value,
						  yes_word) == 0)) {
					p->neg_info |= KEY_REPLY_OPTIONAL;
					TRACE(TRACE_NEGOTIATIONS,
					      "Reply optional for %s\n",
					      dummy_string);
				} else
				    if (!IS_INFORMATIONAL_PARAM(p->type)) {
					TRACE(TRACE_NEGOTIATIONS,
					      "Reply required for %s\n",
					      dummy_string);
				} else {
					TRACE(TRACE_NEGOTIATIONS,
					      "Reply not required for %s\n",
					      dummy_string);
				}
				resp_len = 0;
			}
		}
	}

	FREE_STRING(dummy_string);

	TRACE(TRACE_ENTER_LEAVE,
	      "Leave scan_table_and_process, out_length %d\n", out_length);

	/* return the length added to the output_string */
	return out_length;
}

/*
 * Uses the final values negotiated during a successful login of a new session
 * to set up the session-wide operational values used during FFP
 */
void set_session_parameters(struct session_operational_parameters
			    *oper_param_entry,
			    struct parameter_type
			    login_params[MAX_CONFIG_PARAMS])
{
	struct parameter_type *p;
	__u32 i;

	TRACE(TRACE_ENTER_LEAVE, "Enter set_session_parameters\n");

	for (p = login_params, i = 0; i < MAX_CONFIG_PARAMS; p++, i++) {
		/* For MaxConnections */
		if (p->special_key_flag & MAXCONNECTIONS_FLAG)
			oper_param_entry->MaxConnections = p->int_value;

		/* For InitialR2T */
		else if (p->special_key_flag & INITIALR2T_FLAG)
			oper_param_entry->InitialR2T =
			    !strcmp(p->str_value, yes_word);

		/* For ImmediateData */
		else if (p->special_key_flag & IMMEDIATEDATA_FLAG)
			oper_param_entry->ImmediateData =
			    !strcmp(p->str_value, yes_word);

		/* For MaxBurstLength */
		else if (p->special_key_flag & MAXBURSTLENGTH_FLAG)
			oper_param_entry->MaxBurstLength =
			    (p->int_value / 512) * 512;

		/* For FirstBurstLength */
		else if (p->special_key_flag & FIRSTBURSTLENGTH_FLAG)
			oper_param_entry->FirstBurstLength =
			    (p->int_value / 512) * 512;

		/* For DefaultTime2Wait */
		else if (p->special_key_flag & DEFAULTTIME2WAIT_FLAG)
			oper_param_entry->DefaultTime2Wait = p->int_value;

		/* For DefaultTime2Retain */
		else if (p->special_key_flag & DEFAULTTIME2RETAIN_FLAG)
			oper_param_entry->DefaultTime2Retain =
			    p->int_value;

		/* For MaxOutstandingR2T */
		else if (p->special_key_flag & MAXOUTSTANDINGR2T_FLAG)
			oper_param_entry->MaxOutstandingR2T = p->int_value;

		/* For DataPDUInOrder */
		else if (p->special_key_flag & DATAPDUINORDER_FLAG)
			oper_param_entry->DataPDUInOrder =
			    !strcmp(p->str_value, yes_word);

		/* For DataSequenceInOrder */
		else if (p->special_key_flag & DATASEQUENCEINORDER_FLAG)
			oper_param_entry->DataSequenceInOrder =
			    !strcmp(p->str_value, yes_word);

		/* For ErrorRecoveryLevel */
		else if (p->special_key_flag & ERRORRECOVERYLEVEL_FLAG)
			oper_param_entry->ErrorRecoveryLevel =
			    p->int_value;

		/* For SessionType */
		else if (p->special_key_flag & SESSIONTYPE_FLAG) {
			if ((oper_param_entry->SessionType =
			     !strcmp(p->str_value, discovery_word))) {
				/* discovery session */
				/*  RFC 3720 Section 12.21 SessionType
				 *  "The discovery session implies
				 *  MaxConnections = 1 and overrides both
				 *  the default and an explicit setting."
				 */
				oper_param_entry->MaxConnections = 1;
			}
		}

		/* For TargetName (on target, need the number from this name) */
		else if (p->special_key_flag & TARGETNAME_FLAG) {
			/* save number from end of TargetName for the target */
			oper_param_entry->TargetName = p->int_value;
		}
#ifdef ISER
		/* Added for iSER - Minti */
		/* For MaxOutstandingUnexpectedPDUs */
		else if (p->
			 special_key_flag &
			 MAXOUTSTANDINGUNEXPECTEDPDUS_FLAG) {
			oper_param_entry->MaxOutstandingUnexpectedPDUs =
			    p->int_value;
		} else if (p->special_key_flag & RDMAEXTENSIONS_FLAG) {
			oper_param_entry->RDMAExtensions =
			    !strcmp(p->str_value, yes_word);
			TRACE(TRACE_ISER, "RDMAExtensions=%s\n",
			      p->str_value);
			TRACE(TRACE_ISER, "just set RDMAExtensions=%d\n",
			      oper_param_entry->RDMAExtensions);
		}
#endif
	}
	/* Minti - for testing
	   oper_param_entry->RDMAExtensions = 1; */

	TRACE(TRACE_ENTER_LEAVE, "Leave set_session_parameters\n");
}

/*
 * Uses the final values negotiated during a successful login of a new
 * connection to set up the connection-specific operational values used
 * during FFP
 */
void set_connection_parameters(struct connection_operational_parameters
			       *oper_param_entry,
			       struct parameter_type
			       login_params[MAX_CONFIG_PARAMS])
{

	TRACE(TRACE_ENTER_LEAVE, "Enter set_connection_parameters\n");

	/*  for now, there are no connection-specific values used during
	 *  FFP.  These would be the security values and the marker values.
	 */

	TRACE(TRACE_ENTER_LEAVE, "Leave set_connection_parameters\n");
}

/*
 * Returns -1 on error (duplicate key)
 *	    0 if not a duplicate
 */
int __attribute__ ((no_instrument_function))
    check_step_key(struct unknown_key *key, __u32 * got_keys, __u32 got_value)
{
	if (*got_keys & got_value) {
		TRACE_ERROR("duplicate key %s, value \"%s\"\n",
			    key->keyname, key->keyvalue);
		return -1;
	}
	TRACE(TRACE_ISCSI_FULL, "got %s: %s\n", key->keyname,
	      key->keyvalue);
	*got_keys |= got_value;
	return 0;
}

/*
 * Returns -1 on error (duplicate key or illegal numeric value)
 *	    0 if not a duplicate, numeric_value is set
 */
int __attribute__ ((no_instrument_function))
    check_step_key_number(struct unknown_key *key, __u32 * got_keys,
		      __u32 got_value, __u32 limit, __u32 * numeric_value)
{
	int retval;
	char *tmp_string;

	if ((retval = check_step_key(key, got_keys, got_value)) == 0) {
		*numeric_value =
		    simple_strtoul(key->keyvalue, &tmp_string, 0);
		if (!isdigit(*key->keyvalue)
		    || tmp_string == NULL
		    || *tmp_string != '\0' || *numeric_value > limit) {
			TRACE_ERROR("invalid %s number \"%s\"\n",
				    key->keyname, key->keyvalue);
			retval = -1;
		}
	}

	return retval;
}

static int __attribute__ ((no_instrument_function))
    print_int_param(char *buffer, const char *name, int value)
{
	return sprintf(buffer, "%30s  %d\n", name, value);
}

static int __attribute__ ((no_instrument_function))
    print_boolean_param(char *buffer, const char *name, int value)
{
	return sprintf(buffer, "%30s  %s\n", name,
		       value ? yes_word : no_word);
}

static int __attribute__ ((no_instrument_function))
    print_string_param(char *buffer, const char *name, char *value)
{
	return sprintf(buffer, "%30s  %s\n", name,
		       (value ? value : "<NULL>"));
}

/*
 * prints all the keys in the param_tbl and their values
 */
int __attribute__ ((no_instrument_function))
    print_config_info(struct parameter_type param_tbl[MAX_CONFIG_PARAMS],
		  char *buffer)
{
	struct parameter_type *params;
	int pos;
	int i;

	pos = sprintf(buffer, "    configured parameters\n");

	for (i = 0, params = param_tbl; i < MAX_CONFIG_PARAMS;
	     params++, i++) {
		if (params->type & NUMBER) {
			pos += print_int_param(buffer + pos,
					       params->parameter_name,
					       params->int_value);

		} else {
			pos += print_string_param(buffer + pos,
						  params->parameter_name,
						  params->str_value);
		}
	}

	pos += sprintf(buffer + pos, "\n");

	return pos;
}

/*
 * Prints the session-wide parameter values as negotiated during a
 * successful login of the first connection in a new session
 */
int __attribute__ ((no_instrument_function))
    print_session_params(struct session_operational_parameters *sop,
		     struct parameter_type param_tbl[MAX_CONFIG_PARAMS],
		     char *buffer)
{
	struct parameter_type *p;
	int pos, i;

	pos = sprintf(buffer, "    session-wide parameters\n");

	for (p = param_tbl, i = 0; i < MAX_CONFIG_PARAMS; p++, i++) {
		/* For InitiatorName 12.5 */
		if (p->special_key_flag & INITIATORNAME_FLAG) {
			pos += print_string_param(buffer + pos,
						  p->parameter_name,
						  p->str_value);

			/* For TargetName 12.4 */
		} else if (p->special_key_flag & TARGETNAME_FLAG) {
			pos += print_string_param(buffer + pos,
						  p->parameter_name,
						  p->str_value);

			/* For TargetPortalGroupTag 12.9 */
		} else if (p->special_key_flag & TARGETPORTALGROUPTAG_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       p->int_value);

			/* For MaxConnections 12.2 */
		} else if (p->special_key_flag & MAXCONNECTIONS_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->MaxConnections);

			/* For InitialR2T 12.10 */
		} else if (p->special_key_flag & INITIALR2T_FLAG) {
			pos += print_boolean_param(buffer + pos,
						   p->parameter_name,
						   sop->InitialR2T);

			/* For ImmediateData 12.11 */
		} else if (p->special_key_flag & IMMEDIATEDATA_FLAG) {
			pos += print_boolean_param(buffer + pos,
						   p->parameter_name,
						   sop->ImmediateData);

			/* For MaxBurstLength 12.13 */
		} else if (p->special_key_flag & MAXBURSTLENGTH_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->MaxBurstLength);

			/* For FirstBurstLength 12.14 */
		} else if (p->special_key_flag & FIRSTBURSTLENGTH_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->FirstBurstLength);

			/* For DefaultTime2Wait 12.15 */
		} else if (p->special_key_flag & DEFAULTTIME2WAIT_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->DefaultTime2Wait);

			/* For DefaultTime2Retain 12.16 */
		} else if (p->special_key_flag & DEFAULTTIME2RETAIN_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->DefaultTime2Retain);

			/* For MaxOutstandingR2T 12.17 */
		} else if (p->special_key_flag & MAXOUTSTANDINGR2T_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->MaxOutstandingR2T);

			/* For DataPDUInOrder 12.18 */
		} else if (p->special_key_flag & DATAPDUINORDER_FLAG) {
			pos += print_boolean_param(buffer + pos,
						   p->parameter_name,
						   sop->DataPDUInOrder);

			/* For DataSequenceInOrder 12.19 */
		} else if (p->special_key_flag & DATASEQUENCEINORDER_FLAG) {
			pos += print_boolean_param(buffer + pos,
						   p->parameter_name,
						   sop->
						   DataSequenceInOrder);

			/* For ErrorRecoveryLevel 12.20 */
		} else if (p->special_key_flag & ERRORRECOVERYLEVEL_FLAG) {
			pos += print_int_param(buffer + pos,
					       p->parameter_name,
					       sop->ErrorRecoveryLevel);

			/* For SessionType 12.21 */
		} else if (p->special_key_flag & SESSIONTYPE_FLAG) {
			pos += sprintf(buffer + pos, "%*s  %s\n", 30,
				       p->parameter_name,
				       sop->SessionType
				       ? discovery_word : normal_word);
		}
#ifdef ISER
		/* Added for iSER - Minti */
		/* For RDMAExtensions 8.3 */
		else if (p->special_key_flag & RDMAEXTENSIONS_FLAG) {
			pos += print_boolean_param(buffer + pos,
						   p->parameter_name,
						   sop->RDMAExtensions);
		} else if (p->
			   special_key_flag &
			   MAXOUTSTANDINGUNEXPECTEDPDUS_FLAG) {
			pos +=
			    print_boolean_param(buffer + pos,
						p->parameter_name,
						sop->
						MaxOutstandingUnexpectedPDUs);
		}
#endif
	}
	return pos;
}
