/**
 * Copyright (C) 2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file common/unh_threads.h
 * \brief Provides an abstraction layer for spawning a thread in the kernel or
 *        userspace.
 * \author Ethan Burns
 * \date 2007-05-09
 */
#ifndef _UNH_THREADS_H_
#define _UNH_THREADS_H_

#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif				/* !_POSIX_C_SOURCE */

#include "debug.h"

#if defined(__KERNEL__)

#include <linux/version.h>

typedef pid_t thread_handle;
typedef int thread_ret;

/**
 * \brief Spawns a thread that executes the given function.
 * \param handle The thread handle is returned through this parameter.
 * \param fn The function that the newly spawned thread executes.
 * \param arg The argument passed to the new thread.
 * \param flags Thread creation flags.
 * \return 0 on success.  On error, an error number is returned.
 */
static inline int spawn_thread(thread_handle *handle,
                               thread_ret(*fn) (void *),
                               void *arg,
                               unsigned long flags)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	*handle = kernel_thread(fn, arg, flags);

	TRACE(TRACE_THREADS,
	      "%s spawn_thread created child with thread id %#08lx\n",
	      current->comm, (long) *handle);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/**
 * \brief Kills the currently running thread.
 */
static inline void kill_self(void)
{
	TRACE(TRACE_ENTER_LEAVE,
	      "Entering %s (this function never exits)\n", __func__);
	TRACE(TRACE_THREADS, "Sending kill signal to self, pid:%p\n",
	      (void*) (unsigned long) current->pid);
	kill_proc(current->pid, SIGKILL, 1);
}

/**
 * \brief Kills the specified thread.
 * \param handle The thread_handle for the thread to be killed.
 */
static inline void kill_thread(thread_handle handle)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	kill_proc(handle, SIGKILL, 1);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Waits for a thread to exit.
 * \param handle The thread to wait for.
 */
static inline void wait_thread(thread_handle handle)
{
	/* nop */
}

/**
 * \brief Reparents the current process to init.
 * \note Only used for 2.4 kernels.
 * \note This is a nop for kernels > 2.4.18
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
#define __reparent__() reparent_to_init()
#else
#define __reparent__()
#endif

/**
 * \brief The kernel-space implementation of the daemonize function.
 */
#if defined(K26)
#define unh_daemonize(x...) daemonize(x)
#else
#define unh_daemonize(x...)			\
	do {					\
		daemonize();			\
		sprintf(current->comm, x);	\
		__reparent__();			\
	} while (0);
#endif /* K26 */

#else
#include <pthread.h>
typedef pthread_t thread_handle;
typedef void *thread_ret;

/**
 * \brief Spawns a thread that executes the given function.
 * \param handle The thread handle is returned through this parameter.
 * \param fn The function that the newly spawned thread executes.
 * \param arg The argument passed to the new thread.
 * \param flags Thread creation flags.
 * \return 0 on success.  On error, an error number is returned.
 */
static inline int spawn_thread(thread_handle * handle, thread_ret(*fn) (void *),
		 void *arg, unsigned long flags)
{
	int err;
	struct task_struct *current;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	current = pthread_getspecific(current_key);

	err = pthread_create(handle, NULL, fn, arg);

	TRACE(TRACE_THREADS,
	      "%s spawn_thread created child with thread id %#08lx\n",
	      current->comm, (long) *handle);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return err;
}

/**
 * \brief Kills the currently running thread.
 */
static inline void kill_self(void)
{
	TRACE(TRACE_ENTER_LEAVE,
	      "Entering %s (this function never exits)\n", __func__);
	TRACE(TRACE_THREADS, "Sending kill signal to self, pid:%p\n",
	      pthread_self);
	pthread_cancel(pthread_self());
}

/**
 * \brief Kills the specified thread.
 * \param handle The thread_handle for the thread to be killed.
 */
static inline void kill_thread(thread_handle handle)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	pthread_cancel(handle);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Waits for a thread to exit.
 * \param handle The thread to wait for.
 */
static inline void wait_thread(thread_handle handle)
{
	pthread_join(handle, NULL);
}


/**
 * \brief The user-space implementation of the daemonize function.
 * \note This macro requires a variable `current' in the calling scope.
 */
#define unh_daemonize(x...)					\
	do {							\
		daemonize();					\
		current = pthread_getspecific(current_key);	\
		BUG_ON(!current);				\
	} while (0);

#endif				/* __KERNEL__ */

#endif				/* !_UNH_THREADS_H_ */
/* vi: set tabstop=8 textwidth=78: */
