/*	common/user_locks.h -	user space version of
 *				/usr/src/linux/include/linux/spinlock_types.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This defines the lock functions used
 *	by both the iSCSI target and the iSCSI initiator
 *	to run in user space instead of in the kernel.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#ifndef __USER_LOCKS_H__
#define __USER_LOCKS_H__


/***** spin locks *****/

/* see /usr/src/linux/include/linux/spinlock_types.h */
typedef pthread_mutex_t spinlock_t;

#define SPIN_LOCK_UNLOCKED	PTHREAD_MUTEX_INITIALIZER

#define spin_lock_init(lock)			pthread_mutex_init(lock,NULL)

#define spin_lock(lock)				pthread_mutex_lock(lock)
#define spin_unlock(lock)			pthread_mutex_unlock(lock)

#define spin_lock_irqsave(lock, flags)		do { \
						pthread_mutex_lock(lock); \
						flags = 1; \
						} while (0)

#define spin_unlock_irqrestore(lock, flags)	do { \
						flags = 0; \
						pthread_mutex_unlock(lock); \
						} while (0)


/***** read-write locks *****/

typedef pthread_rwlock_t rwlock_t;

#define RW_LOCK_UNLOCKED			PTHREAD_RWLOCK_INITIALIZER

#define rwlock_init(x)				pthread_rwlock_init(x, NULL)

#define read_lock(lock)							\
do {									\
	TRACE(TRACE_SEM, "%s locking %s for read\n", __func__, #lock);	\
	pthread_rwlock_rdlock(lock);					\
} while (0)
#define read_unlock(lock)							\
do {										\
	TRACE(TRACE_SEM, "%s unlocking %s for read\n", __func__, #lock);	\
	pthread_rwlock_unlock(lock);						\
} while (0)

#define read_lock_irqsave(lock, flags)		do { \
						pthread_rwlock_rdlock(lock); \
						flags = 1; \
						} while (0)

#define read_unlock_irqrestore(lock, flags)	do { \
						flags = 0; \
						pthread_rwlock_unlock(lock); \
						} while (0)

#define write_lock(lock) 						\
do {									\
	TRACE(TRACE_SEM, "%s locking %s for write\n", __func__, #lock);	\
	pthread_rwlock_wrlock(lock);					\
} while (0)
#define write_unlock(lock) 							\
do {										\
	TRACE(TRACE_SEM, "%s unlocking %s for write\n", __func__, #lock);	\
	pthread_rwlock_unlock(lock);						\
} while (0)

#define write_lock_irqsave(lock, flags)		do { \
						pthread_rwlock_wrlock(lock); \
						flags = 1; \
						} while (0)

#define write_unlock_irqrestore(lock, flags)	do { \
						flags = 0; \
						pthread_rwlock_unlock(lock); \
						} while (0)


/***** miscellaneous *****/

#define in_interrupt()				(0)
#define irqs_disabled()				(0)

#define lock_kernel()			do { \
					TRACE(TRACE_THREADS,		\
						"   lock kernel from "	\
						"thread id %#08lx\n",	\
						pthread_self());	\
					} while (0)
#define unlock_kernel()			do { \
					TRACE(TRACE_THREADS,		\
						" unlock kernel from "	\
						"thread id %#08lx\n",	\
						pthread_self());	\
					} while (0)
#endif
