/**
 * Copyright (C) 2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file common/user_mem_cache.c
 * \brief Contains function definitions for the user-space implementation of
 *        Linux's kmem_caches.
 * \author Ethan Burns
 * \date 2007-04-20
 */

#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif				/* !_POSIX_C_SOURCE */

#include "user_space.h"
#include "my_memory.h"

struct my_kmem_cache {
	const char *name;
	size_t obj_size;
	size_t obj_align;

	spinlock_t lock;
	atomic_t num_blocks;
	struct list_head blocks;
};

#define CACHE_LIST_DATA(x) (((void *)(x)) + sizeof(struct list_head))
#define CACHE_DATA_LIST(x) (((void *)(x)) - sizeof(struct list_head))

static inline void __kmem_cache_new_block(my_kmem_cache_t c);
static inline void * __kmem_cache_next(my_kmem_cache_t c);
static inline void __kmem_cache_replace(my_kmem_cache_t c, void *data);

/**
 * Allocates a new cache block (which is preceeded by a list link).
 */
static inline void __kmem_cache_new_block(my_kmem_cache_t c)
{
	struct list_head *list;

	list = my_kmalloc(c->obj_size + sizeof(*list), (char *) c->name);
	TRACE(TRACE_MY_MEMORY, "%s: list=%p, data=%p\n",
	      __func__, list, CACHE_LIST_DATA(list));
	__kmem_cache_replace(c, CACHE_LIST_DATA(list));
}

/**
 * Returns the next block's data pointer.
 */
static inline void * __kmem_cache_next(my_kmem_cache_t c)
{
	struct list_head *list;
	long flags;

	if (atomic_read(&c->num_blocks) <= 0)
		return NULL;


	atomic_dec(&c->num_blocks);

	spin_lock_irqsave(&c->lock, flags);
	list = c->blocks.next;
	list_del(list);
	spin_unlock_irqrestore(&c->lock, flags);

	TRACE(TRACE_MY_MEMORY, "%s: list=%p, data=%p\n",
	      __func__, list, CACHE_LIST_DATA(list));
	return CACHE_LIST_DATA(list);
}

/**
 * Replaces a block in the cache.
 */
static inline void __kmem_cache_replace(my_kmem_cache_t c, void *data)
{
	struct list_head *list;
	long flags;

	list = CACHE_DATA_LIST(data);
	TRACE(TRACE_MY_MEMORY, "%s: list=%p, data=%p\n",
	      __func__, list, CACHE_LIST_DATA(list));

	spin_lock_irqsave(&c->lock, flags);
	list_add(list, &c->blocks);
	spin_unlock_irqrestore(&c->lock, flags);

	atomic_inc(&c->num_blocks);
}

my_kmem_cache_t my_kmem_cache_create(const char *name, size_t size,
					size_t align, unsigned long flags)
{
	my_kmem_cache_t cache;

	cache = my_kmalloc(sizeof(*cache), "kmem_cache");
	if (!cache)
		return NULL;

	cache->name = name;
	cache->obj_size = size;
	cache->obj_align = align;

	spin_lock_init(&cache->lock);
	atomic_set(&cache->num_blocks, 0);
	INIT_LIST_HEAD(&cache->blocks);

	while (atomic_read(&cache->num_blocks) < 400)
		__kmem_cache_new_block(cache);

	return cache;
}

void my_kmem_cache_destroy(my_kmem_cache_t cache)
{
	void *data;
	struct list_head *list;
	long flags;

	if (!cache)
		return;

	while (atomic_read(&cache->num_blocks) != 0) {

		spin_lock_irqsave(&cache->lock, flags);
		data = __kmem_cache_next(cache);
		list = CACHE_DATA_LIST(data);
		list_del(list);
		spin_unlock_irqrestore(&cache->lock, flags);

		my_kfree((void**) &list, (char *) cache->name);
	}

	my_kfree((void **) &cache, (char *) cache->name);
}

void *my_kmem_cache_alloc(my_kmem_cache_t cache, gfp_t flags)
{
	void *obj;

	obj = __kmem_cache_next(cache);

	while (!obj) {
		__kmem_cache_new_block(cache);
		obj = __kmem_cache_next(cache);
	}

	return obj;
}

void my_kmem_cache_free(my_kmem_cache_t cache, void *obj)
{
       	__kmem_cache_replace(cache, obj);
}

/* vi: set tabstop=8 textwidth=78: */
