/*	common/user_sockets.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This defines the socket functions used
 *	by both the iSCSI target and the iSCSI initiator
 *	to run in user space instead of in the kernel.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#include "user_space.h"
#include "my_memory.h"

int sock_bind(struct socket *sock, struct sockaddr *myaddr,
	      int sockaddr_len)
{
	struct task_struct *current;

	current = pthread_getspecific(current_key);

	TRACE(TRACE_NET, "%s sock_bind to fd %d\n", current->comm,
	      sock->fd);
	if (bind(sock->fd, myaddr, sockaddr_len)) {
		TRACE_ERROR("bind: %s\n", strerror(errno));
		return -1;
	}
	TRACE(TRACE_NET, "%s sock_bind to fd %d ok\n", current->comm,
	      sock->fd);
	return 0;
}

int sock_listen(struct socket *sock, int backlog)
{
	struct task_struct *current;

	current = pthread_getspecific(current_key);

	TRACE(TRACE_NET, "%s sock_listen to fd %d\n", current->comm,
	      sock->fd);
	if (listen(sock->fd, backlog)) {
		TRACE_ERROR("listen: %s\n", strerror(errno));
		return -1;
	}
	TRACE(TRACE_NET, "%s sock_listen to fd %d ok\n", current->comm,
	      sock->fd);
	return 0;
}

int sock_accept(struct socket *oldsock, struct socket *newsock, int flags)
{
	int newfd;
	struct task_struct *current;

	current = pthread_getspecific(current_key);

	TRACE(TRACE_NET, "%s sock_accept on fd %d\n",
	      current->comm, oldsock->fd);
	if ((newfd = accept(oldsock->fd, NULL, 0)) < 0) {
		TRACE_ERROR("accept: %s\n", strerror(errno));
		return -1;
	}
	TRACE(TRACE_NET, "%s sock_accept on fd %d gives new fd %d\n",
	      current->comm, oldsock->fd, newfd);
	newsock->fd = newfd;
	return 0;
}

int sock_connect(struct socket *sock, struct sockaddr *vaddr,
		 int sockaddr_len, int flags)
{
	struct task_struct *current;

	current = pthread_getspecific(current_key);

	TRACE(TRACE_NET, "%s sock_connect on fd %d\n", current->comm,
	      sock->fd);
	if (connect(sock->fd, vaddr, sockaddr_len) < 0) {
		TRACE_ERROR("connect: %s\n", strerror(errno));
		return -1;
	}
	TRACE(TRACE_NET, "%s sock_connect on fd %d ok\n",
	      current->comm, sock->fd);
	return 0;
}

int sock_getname(struct socket *sock, struct sockaddr *myaddr,
		 int *sockaddr_len, int which)
{
	struct task_struct *current;

	current = pthread_getspecific(current_key);

	if (which) {
		TRACE(TRACE_NET, "%s getpeername on fd %d\n",
		      current->comm, sock->fd);
		if (getpeername
		    (sock->fd, myaddr, (socklen_t *) sockaddr_len)) {
			TRACE_ERROR("getpeername: %s\n", strerror(errno));
			return -1;
		}
		TRACE(TRACE_NET, "%s getpeername on fd %d ok\n",
		      current->comm, sock->fd);
	} else {
		TRACE(TRACE_NET, "%s getsockname on fd %d\n",
		      current->comm, sock->fd);
		if (getsockname
		    (sock->fd, myaddr, (socklen_t *) sockaddr_len)) {
			TRACE_ERROR("getsockname: %s\n", strerror(errno));
			return -1;
		}
		TRACE(TRACE_NET, "%s getsockname on fd %d ok\n",
		      current->comm, sock->fd);
	}
	return 0;
}

struct proto_ops proto_ops_template = {
      release:NULL,
      bind:sock_bind,
      connect:sock_connect,
      socketpair:NULL,
      accept:sock_accept,
      getname:sock_getname,
      ioctl:NULL,
      listen:sock_listen,
      shutdown:NULL,
      setsockopt:NULL,
      getsockopt:NULL,
      sendmsg:NULL,
      recvmsg:NULL
};

void *sock_alloc(void)
{
	return my_kmalloc(sizeof(struct socket), "newsock");
}


int sock_release(struct socket *sockptr)
{
	my_kfree((void **) &sockptr, "newsock");
	return 0;
}


int sock_create(int family, int type, int protocol,
		struct socket **sockptr)
{
	struct socket *sptr;

	sptr = sock_alloc();
	if (sptr == NULL) {
		TRACE_ERROR("malloc struct socket: %s\n", strerror(errno));
		return -1;
	}
	sptr->fd = socket(family, type, protocol);
	if (sptr->fd < 0) {
		TRACE_ERROR("create socket: %s\n", strerror(errno));
		sock_release(sptr);
		return -1;
	}
	sptr->type = type;
	sptr->ops = &proto_ops_template;

	*sockptr = sptr;
	return 0;
}
