/*	common/user_space.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This defines the generic functions used
 *	by both the iSCSI target and the iSCSI initiator
 *	to run in user space instead of in the kernel.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#ifndef __USER_SPACE_H__
#define __USER_SPACE_H__

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/times.h>
#include <netinet/in.h>
#include <netinet/tcp.h>


/* from /usr/src/linux/include/asm-i386/types.h */
typedef unsigned char __u8;
typedef unsigned short __u16;
typedef unsigned int __u32;
typedef unsigned long long __u64;
typedef __u8 u8;
typedef __u16 u16;
typedef __u32 u32;
typedef __u64 u64;

/* from /usr/src/linux/include/linux/types.h */
typedef unsigned char unchar;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;

typedef unsigned int gfp_t;

/* copy from /usr/src/linux/include/linux/compiler.h */
#define	likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)

#define BUG() abort();
#define BUG_ON(x)				\
	do {					\
		if (x) {			\
			TRACE_ERROR(#x);	\
			BUG();			\
		}				\
	} while (0)

#define schedule() sched_yield()

#include "../common/debug.h"

#define MOD_INC_USE_COUNT	do { } while(0)
#define MOD_DEC_USE_COUNT	do { } while(0)

#define module_init(x)		int module_init_##x(void)	\
				{			\
					return x();	\
				}

#define module_exit(x)		void module_exit_##x(void)	\
				{				\
					x();			\
				}

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]

#define swab64(x) \
({ \
	__u64 __x = (x); \
	((__u64)( \
		(__u64)(((__u64)(__x) & (__u64)0x00000000000000ffULL) << 56) | \
		(__u64)(((__u64)(__x) & (__u64)0x000000000000ff00ULL) << 40) | \
		(__u64)(((__u64)(__x) & (__u64)0x0000000000ff0000ULL) << 24) | \
		(__u64)(((__u64)(__x) & (__u64)0x00000000ff000000ULL) <<  8) | \
		(__u64)(((__u64)(__x) & (__u64)0x000000ff00000000ULL) >>  8) | \
		(__u64)(((__u64)(__x) & (__u64)0x0000ff0000000000ULL) >> 24) | \
		(__u64)(((__u64)(__x) & (__u64)0x00ff000000000000ULL) >> 40) | \
		(__u64)(((__u64)(__x) & (__u64)0xff00000000000000ULL) >> 56) )); \
})

#if defined(__LITTLE_ENDIAN)
#define HIPQUAD(addr) \
	((unsigned char *)&addr)[3], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[0]

#define be64_to_cpu(x) swab64((__u64)(x))
#define cpu_to_be64(x) swab64((__u64)(x))

#elif defined(__BIG_ENDIAN)

#define HIPQUAD NIPQUAD
#define be64_to_cpu(x) ((__u64)(x))
#define cpu_to_be64(x) ((__u64)(x))

#else				/* __BIG_ENDIAN */
#error "Please fix asm/byteorder.h"
#endif				/* __LITTLE_ENDIAN */

extern pthread_key_t current_key;

/* for user-space file handling */
#include "../common/user_files.h"

/* user-space version of /usr/src/linux/include/linux/list.h */
#include "../common/user_list.h"

/* from /usr/src/linux/include/linux/stddef.h */
#ifndef offsetof
#ifdef __compiler_offsetof
#define offsetof(TYPE,MEMBER) __compiler_offsetof(TYPE,MEMBER)
#else
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif
#endif

/* from /usr/src/linux/include/linux/kernel.h */
#define	printk(x...)		fprintf(stderr, x)
#define	simple_strtoul		strtoul
#define	cpu_to_be32		htonl
#define	cpu_to_be16		htons
#define	be32_to_cpu		ntohl
#define	be16_to_cpu		ntohs

#include "../common/user_locks.h"

/* dummy from /usr/src/linux/include/linux/sched.h */
struct task_struct {
	volatile long state;	/* -1 unrunnable, 0 runnable, >0 stopped */
	char comm[32];
	pid_t pid;
	pthread_t threadid;
	sigset_t blocked;
};

/* user space sockets */
#include "../common/user_sockets.h"

/* to define scatter-gather lists */
#include <sys/uio.h>

struct scatterlist {
	void *addr;
	unsigned int offset;
	unsigned int length;
};

#define	scatterlist_address(x)		(x)->addr
#define	sg_virt(x)		(x)->addr
#define	scatterlist_offset(x)		(x)->offset
#define	scatterlist_length(x)		(x)->length
#define sg_init_table(x, y)		memset((x), 0, (y * sizeof(struct scatterlist)))

/*
 * Get the next scatterlist entry.
 */
#define sg_next(x) (((struct scatterlist *) x) + 1)

/*
 * Loop over each sg element, following the pointer to a new list if necessary
 */
#define for_each_sg(sglist, sg, nr, __i)	\
	for (__i = 0, sg = (sglist); __i < (nr); __i++, sg = sg_next(sg))

#define PAGE_SHIFT	12
#define	PAGE_SIZE	(1UL << PAGE_SHIFT)
#define MAX_ORDER	16

#define	__get_free_pages(f,order)	malloc(PAGE_SIZE<<order)
#define	free_pages(x,order)		free((void *)(x))

static inline unsigned long virt_to_phys(volatile void *address)
{
	return (unsigned long) address;
}

static inline void *phys_to_virt(unsigned long address)
{
	return (void *) address;
}

/* user space version of /usr/src/linux/include/linux/signal.h */
#include "../common/user_signal.h"

/* user space version of /usr/src/linux/include/linux/timer.h */
#include "../common/user_timer.h"

/* user space version of /usr/include/asm/bitops.h */
#include "../common/user_bitops.h"

/* user space version of /usr/src/linux/include/asm/atomic.h */
#include "../common/user_atomic.h"

/* user space version of /usr/src/linux/include/asm/semaphore.h */
#include "../common/user_semaphore.h"

/* user space version of /usr/src/linux/include/scsi/scsi.h */
#include "../common/user_scsi.h"

/* user space threads */
#include "../common/user_threads.h"

/* defined in common/text_param.c */
extern char * __attribute__ ((no_instrument_function))
    my_strdup(const char *str);

/* user space wait queue */
#include "../common/user_wait.h"

/* user space kernel-string functions */
#include "../common/user_string.h"

/* scsi_cmnd structure accessor macros. */
#include "../common/scsi_cmnd.h"

#endif
