/**
 * Copyright (C) 2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file common/user_wait.h
 * \brief Contains declarations for a user-space implementation of Linux wait
 *        queues.
 * \author Ethan Burns
 * \date 2007-04-20
 */
#ifndef _USER_WAIT_H_
#define _USER_WAIT_H_

#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif				/* !_POSIX_C_SOURCE */

#include <asm/param.h>		/* HZ */

#include <pthread.h>
#include <time.h>

/**
 * \brief A wait_queue_head_t is implemented using pthread condition
 *        variables.
 */
typedef struct {
	pthread_cond_t cond;
	pthread_mutex_t mutex;
} wait_queue_head_t;

/**
 * \brief Initializes a wait_queue_head_t*
 * \param q The queue to initialize.
 */
static inline void init_waitqueue_head(wait_queue_head_t * q)
{
	pthread_cond_init(&q->cond, NULL);
	pthread_mutex_init(&q->mutex, NULL);
}

/**
 * \brief Suspends a thread untill cond evaluates to true.  Cond is checked
 *        each time the queue is woken up.  A timeout will break out early.
 * \param q The wait queue.
 * \param cond A valid C expression.
 * \param timeout Timeout in jiffies.
 */
#define wait_event_interruptible_timeout(q, condition, timeout)		\
	do {								\
		struct timespec time;					\
									\
		clock_gettime(CLOCK_REALTIME, &time);			\
		time.tv_sec += timeout / HZ;				\
									\
		pthread_mutex_lock(&(q).mutex);				\
		while(!(condition)) {					\
			if (pthread_cond_timedwait(&(q).cond,		\
			                           &(q).mutex,		\
			                           &time)		\
		                                   == ETIMEDOUT)	\
				break;					\
		}							\
		pthread_mutex_unlock(&(q).mutex);			\
	} while (0);


/**
 * \brief Suspends a thread untill cond evaluates to true.  Cond is checked
 *        each time the queue is woken up.
 * \param q The wait queue.
 * \param cond A valid C expression.
 */
#define wait_event_interruptible(q, condition)				\
	do {								\
		pthread_mutex_lock(&(q).mutex);				\
		while(!(condition)) {					\
			pthread_cond_wait(&(q).cond, &(q).mutex);	\
		}							\
		pthread_mutex_unlock(&(q).mutex);			\
	} while (0);

/**
 * \brief Wakes up all the threads waiting on the queue.
 * \param q The queue to wake up.
 */
#define wake_up_interruptible(q)					\
	do {								\
		pthread_cond_broadcast(&(q)->cond);			\
	} while (0);

#endif				/* !_USER_WAIT_H_ */
/* vi: set tabstop=8 textwidth=78: */
