/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#ifndef __ISER_API_H__
#define __ISER_API_H__

#include "iser_types.h"
#include "iser_pdu.h"
#include "iser_ext_api.h"

/* ---------------------------------------------------------------------
   Registered functions exported by iSER datamover layer for iSCSI layer
   ---------------------------------------------------------------------
*/

/* [Enable_Datamover] - prototype
   An iSCSI layer requests its local iSER datamover layer to make an
   iSCSI connection iSER-assisted. This connection is identified
   by the connection handle, for which the iSER-specific resource
   allocation was earlier made.
   The Login Response PDU is relevant only for a target, and contains
   the final Login Response that concludes the iSCSI Login phase. The
   Login Response must be sent to the initiator before iSER datamover
   assistance is enabled for the connection.
*/
typedef iser_status_t
    (*iser_enable_datamover_func_t) (iser_conn_handle_t conn_h,
				     iser_transport_conn_t tport_conn,
				     struct iser_send_pdu_t *
				     p_login_resp_pdu);

/* [Deallocate_Connection_Resources] - prototype
   An iSCSI layer requests its local iSER datamover layer to deallocate
   all the iSER-specific resources that may have been allocated earlier
   for the transport connection identified by the passed connection
   handle. Should be invoked by the iSCSI layer if iSER-specific resources
   become no longer necessary before the connection is iSER-enabled.
   (such as the Login failure of the corresponding iSCSI connection).
*/
typedef iser_status_t
    (*iser_dealloc_conn_resources_func_t) (iser_conn_handle_t conn_h);

/* [Send_Control] - prototype
   An iSCSI layer requests its local iSER datamover layer to transmit
   an iSCSI control-type PDU to the peer iSCSI layer operating in the
   remote iSCSI node. The iSER datamover layer sends the PDU after
   adding its own protocol header.
*/
typedef iser_status_t
    (*iser_send_control_func_t) (iser_conn_handle_t conn_h,
				 struct iser_send_pdu_t * p_ctrl_pdu);

/* [Release_Control] - prototype, extention API function
   iSCSI layer notifies its local iSER datamover layer that
   an iSCSI control-type PDU previously passed through the Control_Notify
   primitive is no longer in use and its memory may be released.
*/
typedef iser_status_t
    (*iser_release_control_func_t) (iser_conn_handle_t conn_h,
				    struct iser_recv_pdu_t * p_ctrl_pdu);

/* [Connection_Terminate] - prototype
   An iSCSI layer requests its local iSER datamover layer to terminate
   the transport connection and deallocate all the connection and task
   resources associated with its connection handle.
   After this function returns all I/O resources previously passed to
   the iSER layer are back into the full ownership by the iSCSI layer.
*/
typedef iser_status_t
    (*iser_connection_terminate_func_t) (iser_conn_handle_t conn_h);

/* [Put_Data] - prototype
   An iSCSI layer requests its local iSER datamover layer to transmit the
   data for the SCSI Data-In PDU (identified by the data descriptor p_data)
   to the peer iSCSI layer on the remote iSCSI node.
   The notify_enable flag, when true, is used to request the local iSER datamover
   layer to generate the eventual local completion notification to the iSCSI layer
   for this put_data() invocation.
*/
typedef iser_status_t
    (*iser_put_data_func_t) (void *iser_task_h,
			     struct iser_send_pdu_t * p_data_in_pdu,
			     int notify_enable);

/* [Get_Data] - prototype
   An iSCSI layer requests its local iSER datamover layer to transmit the
   data for the SCSI R2T PDU (identified by the data descriptor p_data)
   from the peer iSCSI layer on the remote iSCSI node.
   The notify_enable flag, when true, is used to request the local iSER datamover
   layer to generate the eventual local completion notification to the iSCSI layer
   for this get_data() invocation.
*/
typedef iser_status_t
    (*iser_get_data_func_t) (void *iser_task_h,
			     struct iser_send_pdu_t * p_r2t_pdu,
			     int notify_enable);

/* [Notice_Key_Values]  - prototype
   An iSCSI layer requests its local iSER datamover layer to take note
   of the negotiated values of the listed keys.
*/
typedef iser_status_t
    (*iser_notice_key_values_func_t) (iser_conn_handle_t conn_h,
				      char *key, char *value);

/* Whatever connection resources requested by iSCSI to be allocated.
 */
struct iser_conn_resources_t {

	enum iser_transport_conn_type conn_type;	/* IN */
	iser_api_handle_t api_h;	/* IN */
	struct sockaddr_in *dst_addr;	/* IN */
	struct sockaddr_in *src_addr;	/* IN */

	unsigned int max_recv_pdu_sz;	// to be deleted!!!
	/* Maximal PDU size, receiving of which should be anticipated */
	unsigned int first_burst_length;	/* IN */
	unsigned int max_recv_dsl;	/* IN */
	unsigned int max_outstand_cmds;	/* IN */

	/* Array of buffer descriptors to be pre-registered, their memory
	   handles to be cached.
	   If NULL is passed, no buffers are pre-registered.
	   Each buffer should be a single contiguous memory region, i.e. of
	   type ISER_BUF_TYPE_SINGLE.
	   Consequent calls to get_data() or put_data() may pass subregions
	   of these pre-registered regions, the appropriate memory handles
	   and virtual addresses will be used (but only if the buffers are
	   single; scatter-gather lists are not searched in the cache).
	 */
	struct iser_data_buf_t *reg_buf;	/* IN */
	unsigned int num_reg_buf;	/* IN */
};				/* iser_conn_resources_t */

/* [Allocate_Connection_Resources] - prototype
   An iSCSI layer requests its local iSER datamover layer to perform all
   iSER-specific resource allocations required for the full feature phase
   of an iSCSI connection. This connection is identified by the passed
   connection handle. Eventually, this connection should be transitioned
   to be an iSER-assisted iSCSI connection.
   May be called multiple times on the same connection handle, upon
   subsequent invocations returns success, if successful for the first time.
   Some additional implementation-specific resource requirements may be
   specified by the iSCSI layer.
*/
typedef iser_status_t(*iser_allocate_conn_resources_func_t) (iser_conn_handle_t conn_h, struct iser_conn_resources_t * conn_resources, int *max_outstanding_command);	// OUT


struct iser_api_t {
	/* API registration (to be defined by iSER, called by iSCSI */
	iser_send_control_func_t send_control;
	iser_release_control_func_t release_control;
	iser_put_data_func_t put_data;
	iser_get_data_func_t get_data;
	iser_allocate_conn_resources_func_t allocate_connection_resources;
	iser_notice_key_values_func_t notice_key_values;
	iser_enable_datamover_func_t enable_datamover;
	iser_connection_terminate_func_t connection_terminate;
	iser_dealloc_conn_resources_func_t deallocate_connection_resources;
};

/* ---------------------------------------------------------------------
   Registered functions exported by iSCSI layer for iSER datamover layer
   ---------------------------------------------------------------------
*/

typedef void
 (*iser_cmd_notify_func_t) (iser_conn_handle_t conn_h, struct iser_recv_pdu_t * p_ctrl_pdu, void *iser_task_h,	/* in */
			    void **iscsi_task_h);	/* out */

typedef void
 (*iser_dout_notify_func_t) (void *iscsi_task_h,	/* in */
			     struct iser_recv_pdu_t * p_ctrl_pdu);
typedef void
 (*iser_ctrl_notify_func_t) (iser_conn_handle_t conn_h,
			     struct iser_recv_pdu_t * p_ctrl_pdu);

/* [Control_Notify] - prototype
   iSER datamover layer notifies the local iSCSI layer of
   the arrival of an iSCSI control-type PDU from the peer iSER layer
   on the remote iSCSI node. The iSCSI layer processes the control-type
   PDU as defined by iSCSI.
*/
//typedef void
//(*iser_control_notify_func_t)(iser_conn_handle_t            conn_h,
//                              struct iser_recv_pdu_t        *p_ctrl_pdu );

/* [Data_Completion_Notify] - prototype
   iSER datamover layer notifies its local iSCSI layer on
   completing the retrieval of the data or upon sending the data,
   as requested in a prior iSCSI data-type PDU, from/to the peer iSER
   layer on the remote iSCSI node.
   SN is either the DataSN associated with the SCSI Data-In PDU or
   R2TSN associated with the R2T PDU depending on the SCSI operation.
   This function is invoked if and only if Notify_Enable was set
   in either a Put_Data or a Get_Data invocation.
*/
typedef void
 (*iser_data_cmpl_notify_func_t) (void *iscsi_task_h, int itt, int sn);

/* [Data_Ack_Notify] - prototype
   A target iSCSI layer being notified of the arrival of a data
   acknowledgement for a certain SCSI Read data PDU is the only example
   of invoking this function
*/
typedef void
 (*iser_data_ack_notify_func_t) (void *iscsi_task_h, int itt, int data_sn);

/* [Connection_Termination_Notify] - prototype
   iSER datamover layer notifies its local iSCSI layer on an
   unsolicited termination or failure of an iSCSI connection.
   The iSCSI layer considers the connection handle to be invalid
   upon being so notified.
   iSER layer MUST deallocate the connection and task resources
   associated with the terminated connection before notifying
   the iSCSI layer of the termination.
*/
typedef void
 (*iser_conn_term_notify_func_t) (iser_conn_handle_t conn_h);

struct iser_api_cb_t {
	/* Callbacks registration (to be defined by iSCSI, called by iSER */
	iser_cmd_notify_func_t command_notify;
	iser_dout_notify_func_t data_out_notify;
	iser_ctrl_notify_func_t control_notify;
	iser_data_cmpl_notify_func_t data_completion_notify;
	iser_data_ack_notify_func_t data_ack_notify;
	iser_conn_term_notify_func_t conn_terminate_notify;
};

/* ----------------------------------------------------------
   Functions exported by the iSER datamover module
   ----------------------------------------------------------
*/

/*
 * Retrieve the iSER API functions and register iSER API callacks.
 * Functions set to NULL are expected to be ignored by both sides.
 *
 * TODO: fix this functionality to use sockaddrs:
 * "Port number is an IP port at which iSER will listen for
 * incoming connections over IB. This port number may be
 * changed afterwards (see iser_ext_set_port)."
 */
iser_status_t iser_api_register_target(char *provider_name,	/* IN */
				       struct iser_api_t *api,	/* OUT */
				       struct iser_api_cb_t *api_cb,	/* IN */
				       struct iser_ext_api_t *ext_api,	/* OUT */
				       struct iser_ext_api_cb_t *ext_api_cb,	/* IN */
				       struct sockaddr *addr, size_t addr_len, iser_api_handle_t * p_api_h);	/* OUT */

/*
 * Retrieve the iSER API functions and register iSER API callacks.
 * Functions set to NULL are expected to be ignored by both sides.
 */
iser_status_t iser_api_register_initiator(char *provider_name,	/* IN */
					  struct iser_api_t *api,	/* OUT */
					  struct iser_api_cb_t *api_cb,	/* IN */
					  struct iser_ext_api_t *ext_api,	/* OUT */
					  struct iser_ext_api_cb_t *ext_api_cb,	/* IN */
					  iser_api_handle_t * p_api_h);	/* OUT */

/* Unregister API previously registered for this entity
*/
iser_status_t iser_api_unregister(iser_api_handle_t api_h);	/* IN */


#endif				/* __ISER_API_H__ */
