/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#ifndef __ISER_EXT_API_H__
#define __ISER_EXT_API_H__

#if defined(__KERNEL__)
#include <linux/socket.h>
#include <linux/in.h>
#endif				/* __KERNEL__ */

#include "iser_types.h"

/* ----------------------------------------------------------
   Opaque handles 
   ---------------------------------------------------------- 
*/

/* Connection request handle, target only */
typedef void *iser_conn_request_t;

/* ----------------------------------------------------------
   Functions exported by iSER datamover layer for iSCSI layer
   ---------------------------------------------------------- 
*/

/* [Connection_Establish]
   An iSCSI layer requests its local iSER datamover layer to establish 
a physical iSER connection. Relevant for the transports, in which 
RDMA capabilities are supported on the lower levels than TCP, like IB. 
In such cases the TCP connection can not be "upgraded" to have RDMA 
capabilities. This extension primitive allows to open the physical 
connection and later pass its handle to all iSER primitives.
In order to be able to establish a new connection the iSCSI layer
serviced by the iSER module must be registered with it first.
The iSCSI layer supplies the registration api ahandle and its own 
connection handle which will identify the connection throughout 
its entire lifecycle.
*/
typedef iser_status_t(*iser_connection_establish_func_t) (iser_api_handle_t api_h,	/* IN */
							  iser_conn_handle_t iscsi_conn_h,	/* IN */
							  struct sockaddr_in * dst_addr,	/* IN */
							  struct sockaddr_in * src_addr);	/* IN */

/* [Connection_Accept]
   An iSCSI layer at a target side, which has been previously notified
about an incoming iSER connection request (Connection_Request_Notify), 
may accept or reject the connection using this primitive, by setting
the boolean "accept" parameter to true or false, respectively.
The iSCSI layer passes the iSER connection request handle received in 
the previous invocation of connection_request_notify() and and its own 
connection handle which will identify the connection throughout 
its entire lifecycle.
*/
typedef iser_status_t(*iser_connection_accept_func_t) (iser_conn_request_t conn_req_h,	/* IN */
						       iser_conn_handle_t iscsi_conn_h,	/* IN */
						       int accept);	/* IN */

/* ----------------------------------------------------------
   Functions exported by iSCSI datamover layer for iSER layer
   ---------------------------------------------------------- 
*/

/* [Connection_Establish_Notify]
   An iSCSI layer at initiator side, which previously requested to 
establish a physical connection, is notified about successful 
connection establishment through this notification callback.
The connection is identified by the handle, which the iSCSI 
layer passed in previous invocation of connection_establish().
Boolean flag "established" is true if connection establishment
was successful, false if it failed.
*/
typedef iser_status_t(*iser_connection_establish_notify_func_t) (iser_conn_handle_t iscsi_conn_h,	/* IN */
								 int established);	/* IN */

/* [Connection_Request_Notify]
   An iSCSI layer at target side, is notified about an incoming 
connection request through this notification callback. The iSER
layer provides a connection request handle, to be used in the
subsequent call to connection_accept(). This call is possible 
only after the iSCSI layer registered itself as a target entity
and provided a port number to listen on.
*/
typedef iser_status_t(*iser_connection_request_notify_func_t) (iser_conn_request_t conn_req_h,	/* IN */
							       struct sockaddr_in * dst_addr,	/* IN */
							       struct sockaddr_in * src_addr);	/* IN */

/* ----------------------------------------------------------
   Callbacks registration
   ---------------------------------------------------------- 
*/
struct iser_ext_api_t {
	iser_connection_establish_func_t connection_establish;
	iser_connection_accept_func_t connection_accept;
};

struct iser_ext_api_cb_t {
	iser_connection_establish_notify_func_t
	    connection_establish_notify;
	iser_connection_request_notify_func_t connection_request_notify;
};


/* Change the port, at which iSER listens for incoming IB 
   connections.
*/
iser_status_t iser_ext_set_port(iser_api_handle_t ext_api_h,	/* IN */
				unsigned int port);	/* IN */


#endif				/* __ISER_EXT_API_H__ */
