/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#ifndef __ISCSI_PDU_H__
#define __ISCSI_PDU_H__

#if defined(__KERNEL__)
#include <linux/types.h>
#else
#include <stdint.h>
#endif				/* __KERNEL__ */

#include "iser_types.h"

/*! --------------------------------------------------------------------
   [iser_pdu_bhs_t]
 
   Description: iSCSI PDU Basic Header Segment (BHS)
   
    The purpose of defining BHS here is define the fields relevant 
    to iSER operation. iSCSI layer should have its own definitions and 
    may view the BHS passed to iSER just as a buffer. 
    In order to avoid potential name collision, the structure has 
    "iser" prefix, although strictly speaking it must be "iscsi".
 -------------------------------------------------------------------- */

#define ISER_PDU_BHS_LENGTH     48

union iser_pdu_bhs_t {

	unsigned char buf[ISER_PDU_BHS_LENGTH];

	struct {
		uint8_t opcode;	/* optional bits + opcode */
		uint8_t flags;	/* opcode-specific flags */
		uint8_t rsvd[2];	/* usually reserved, used in response */
		uint8_t ahs_length;	/* AHS total length */
		uint8_t dlength[3];	/* Data segment length */
		uint8_t lun[8];	/* LUN */
		uint8_t itt[4];	/* Initiator Task Tag */
		uint8_t other[28];	/* opcode-specific */
	} byte;

	struct {
		uint32_t op_flg_rsvd;	/* opcode, flags, reserved */
		uint32_t length;	/* AHS length, Data segment length */
		uint32_t lun[2];	/* LUN */
		uint32_t itt;	/* Initiator Task Tag */
		uint32_t other[7];	/* opcode-specific */
	} dword;
};				/* iser_pdu_bhs_t */

/*! --------------------------------------------------------------------
   [iser_send_pdu_t]
 
   Description: descriptor of a control PDU to be sent by Send_Control
 -------------------------------------------------------------------- */
struct iser_send_pdu_t {

	union iser_pdu_bhs_t *p_bhs;	/* BHS */

	union {
		/* Command */
		struct {
			/* data-out buffer meant for the entire write/bidir command */
			struct iser_data_buf_t buf_out;

			/* data-in buffer meant for the entire read/bidir command */
			struct iser_data_buf_t buf_in;

			/* although not data segment, ahs defined here */
			struct iser_data_buf_t ahs;

			/* size of immediate unsolicited data for write/bidir command, 
			   will be sent as the data segment of the command PDU  */
			unsigned int immediate_sz;
			/* entire unsolicited data for write/bidir command,
			   will be sent in data segments of command and data-out PDUs */
			unsigned int unsolicited_sz;
		} command;
		/* Response */
		struct {
			/* the sense and response information for the command */
			struct iser_data_buf_t buf_status;

		} response;
		/* Task Mgt Request */
		struct {
			/* data-out for the entire write/bidir command, 
			   valid only if Function=TASK REASSIGN */
			struct iser_data_buf_t buf_in;
			/* data-out for the entire read/bidir command, 
			   valid only if Function=TASK REASSIGN */
			struct iser_data_buf_t buf_out;
		} task_mgt_req;
		/* Data-In - not sent */
		struct {
			/* buffer from which the data-in is to be sent */
			struct iser_data_buf_t src_buf;
		} data_in;
		/* R2T - not sent */
		struct {
			/* buffer to which the data-out is to be received */
			struct iser_data_buf_t dst_buf;
		} r2t;

		/*                                    Other Transmitted PDUs 
		 * ------------------------------------------------------------
		 * Data-Out - disregarded, offset & DSL fields of BHS are used
		 * Asynchronous message - sense and iSCSI Event information 
		 * Text Request - iSCSI Text Request
		 * Text Response - iSCSI Text Response
		 * Login Request - iSCSI Login Request
		 * Login Response - iSCSI Login Response
		 * Reject - reject desriptor
		 * Nop-Out - data accompanying Nop-Out PDU (iSCSI ping)
		 * Nop-In - data accompanying Nop-In PDU (iSCSI return ping)
		 * ------------------------------------------------------------
		 */
		struct {
			/* data segment of a generic tx PDU */
			struct iser_data_buf_t buf;
		} tx;
		/* Any Received PDU */
	} data;			/* PDU data segment descriptor */

};				/* iser_send_pdu_t */

/*! --------------------------------------------------------------------
   [iser_recv_pdu_t]
 
   Description: received control PDU descriptor; passed to Control_Notify 
 -------------------------------------------------------------------- */
struct iser_recv_pdu_t {

	union iser_pdu_bhs_t *p_bhs;	/* BHS */

	struct iser_data_buf_t rx_data;	/* data segment */
};				/* iser_recv_pdu_t */


#endif				/* __ISCSI_PDU_H__ */
