/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#ifndef __ISER_H__
#define __ISER_H__

#if defined(__KERNEL__)
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <rdma/rdma_cm.h>
#include <rdma/ib_verbs.h>
#if defined(CONFIG_DEBUG_FS)
#include <linux/debugfs.h>
#include "../common/debugfs.h"
#endif /* CONFIG_DEBUG_FS */
#else
#include "../common/user_space.h"
#include <stdint.h>
#include <rdma/rdma_cma.h>
#endif				/* __KERNEL__ */

#include <asm/param.h>

/* must be included after "../common/user_space.h", or <linux/types.h> */
#include "../common/my_memory.h"
#include "../common/unh_threads.h"
#include "../common/debug.h"

#include "iser_api.h"
#include "iser_ext_api.h"
#include "iser_header.h"
#include "kernel_dep.h"
#include "iser_platform.h"

#define ISER_DEBUGFS_PERMS 0444

#define CONN_TIMEOUT 20 * HZ	/* 20 seconds */

#if defined(__KERNEL__)

/* kernel-space typedefs */
typedef struct {
	struct iser_conn_t *p_conn;

	void *addr;
	unsigned int length;
	u32 lkey;
	u32 rkey;
	struct ib_mr *ib_mr;
	struct work_struct dereg_work;

	struct ib_device *ib_device;
	enum dma_data_direction dma_dir;
	unsigned int phys_array_len;
	struct ib_phys_buf phys_array[];
} ib_mr_t;
typedef struct {
	/* link to rdma_even_channel queue */
	struct list_head queue;
	/* list of all events */
	struct list_head events_list;
	/* the actual rdma_cm_event structure */
	struct rdma_cm_event *rdma_cm_event;
	/* the event type */
	int event;
	/* the cm_id that this event belongs to */
	struct rdma_cm_id *id;
	/* the channel this event came in on. */
	struct rdma_event_channel *channel;

} iser_cm_event_t;
typedef struct ib_pd ib_pd_t;
typedef struct ib_qp ib_qp_t;
typedef struct ib_cq ib_cq_t;	/* this must refer to a struct ib_cq */
typedef struct ib_wc ib_wc_t;
typedef struct ib_sge ib_sge_t;
typedef struct ib_send_wr ib_send_wr_t;
typedef struct ib_recv_wr ib_recv_wr_t;
typedef struct ib_qp_init_attr ib_qp_init_attr_t;
typedef struct ib_device_attr ib_device_attr_t;
#else			/* __KERNEL__ */
/* user-space typedefs */
typedef struct rdma_cm_event iser_cm_event_t;
typedef struct ibv_pd ib_pd_t;
typedef struct ibv_qp ib_qp_t;
typedef struct ibv_cq ib_cq_t;
typedef struct ibv_mr ib_mr_t;
typedef struct ibv_wc ib_wc_t;
typedef struct ibv_sge ib_sge_t;
typedef struct ibv_send_wr ib_send_wr_t;
typedef struct ibv_recv_wr ib_recv_wr_t;
typedef struct ibv_qp_init_attr ib_qp_init_attr_t;
typedef struct ibv_device_attr ib_device_attr_t;
#endif			/* __KERNEL__ */

/* ---------------------------------------------------------------------
 * CONSTANTS & MACROS
 * ------------------------------------------------------------------ */

/* Memory address alignment */
#define ALIGN_MASK(bit)	    ((1UL << bit) - 1UL)
#define ALIGN_TO_BIT(x,bit) ((x + ALIGN_MASK(bit)) & ~ALIGN_MASK(bit))

#ifndef MIN
#define MIN(a,b)    ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b)    ((a) > (b) ? (a) : (b))
#endif

/* Constant PDU lengths calculations */
#define ISER_TOTAL_HEADERS_LEN		\
    (ISER_HDR_LEN + ISER_PDU_BHS_LENGTH)

/* Hash tables */
#define HASH_TABLE_SIZE			    256

/* Length of an object name string */
#define ISER_OBJECT_NAME_SIZE		    64

/* Various size limits */
#define ISER_LOGIN_PHASE_PDU_DATA_LEN	    (8*1024)	/* 8K */
#define ISER_MIN_RECV_DSL		    (8*1024)	/* 8K */
#define ISER_MAX_CMD_SIZE		    (2*1024*1024)	/* 2M */
#define ISER_MAX_FIRST_BURST		    (128*1024)	/* 128K */

#define ISER_MAX_CTRLS_PER_CMD(first_burst,recv_dsl,imm) \
    ((first_burst / recv_dsl) + (first_burst % recv_dsl > 0 ? 1 : 0) + (imm ? 0 : 1))

#define ISER_MAX_OUTSTAND_CMDS		    16
#define ISER_MAX_IMMEDIATE_CMDS		    2

#define ISER_MAX_NOP_IN			    2
#define ISER_MAX_ASYNC_EVT		    2

#define ISER_MAX_LOGIN_REQ		    1
#define ISER_MAX_TEXT_REQ		    1
#define ISER_MAX_NOP_OUT		    2
#define ISER_MAX_TASK_MGT_REQ		    2
#define ISER_MAX_LOGOUT_REQ		    1

/*! --------------------------------------------------------------------
   [enum iser_conn_state]

   Description: iSER connection state
 -------------------------------------------------------------------- */
enum iser_conn_state {

	ISER_CONN_INIT,		/* descriptor allocated, no connection */
	ISER_CONN_PENDING,	/* about to be established */
	ISER_CONN_UP,		/* up and running */
	ISER_CONN_SYNC_TERM,	/* synchronous termination */
	ISER_CONN_ASYNC_TERM,	/* asynchronous termination */
	ISER_CONN_DOWN,		/* shut down */

	ISER_CONN_STATES_NUM
};				/* iser_conn_state */

/*! --------------------------------------------------------------------
   [enum iser_hello_state]

   Description: iSER hello message states
 -------------------------------------------------------------------- */
enum iser_hello_state {
	ISER_HELLO_BEFORE,
	ISER_HELLO_WAIT,
	ISER_HELLO_SENT,
	ISER_HELLO_REPL_SENT,
	ISER_HELLO_DONE
};

/*! --------------------------------------------------------------------
   [enum iser_regd_buf_type]

   Description: Type of a registered memory buffer. Buffers are registered
    through several pathways: buffer pools, dynamically registered user
    buffers, cached registrations of user buffers.
 -------------------------------------------------------------------- */
enum iser_regd_buf_type {
	ISER_REGD_BUF_UNUSED = 0,	/* This buffer is not yet registered */
	ISER_REGD_BUF_POOL,	/* Pool of internally allocated and registered */
	ISER_REGD_BUF_CACHE,	/* Cached registration */
	ISER_REGD_BUF_USER,	/* Dynamically registered user-supplied buffer */
	ISER_REGD_BUF_RDMA,	/* Local memory registered for a RDMA operation */
	ISER_PRE_REGD_BUFF,	// Allocated with kmalloc.

	ISER_REGD_BUF_TYPES_NUM
};				/* iser_regd_buf_type */

/*! --------------------------------------------------------------------
   [enum iser_dto_type]

   Description: Type of a DTO
 -------------------------------------------------------------------- */
enum iser_dto_type {

	ISER_DTO_RCV = 0,	/* Receive buffer */
	ISER_DTO_SEND,		/* Send buffer */
	ISER_DTO_RDMA_W,	/* Active RDMA-W */
	ISER_DTO_RDMA_R,	/* Active RDMA-R */
	ISER_DTO_PASSIVE,	/* Passive side of a remote RDMA op */

	ISER_DTO_TYPES_NUM
};				/* iser_dto_type */

/*! --------------------------------------------------------------------
   [enum iser_data_dir]

   Description: Direction of data transfer
 -------------------------------------------------------------------- */
enum iser_data_dir {

	ISER_DIR_IN = 0,	/* to initiator */
	ISER_DIR_OUT,		/* from initiator */

	ISER_DIRS_NUM
};				/* iser_data_dir */

/*! --------------------------------------------------------------------
   [enum iser_task_status_t]

   Description: Status of the task
 -------------------------------------------------------------------- */
enum iser_task_status_t {
	ISER_TASK_STATUS_INIT = 0,
	ISER_TASK_STATUS_STARTED,
	ISER_TASK_STATUS_COMPLETED
};				/* iser_task_status_t */

/* ---------------------------------------------------------------------
   Structures
 -------------------------------------------------------------------- */

/*! --------------------------------------------------------------------
   [struct iser_mem_handles_t]

   Description: Memory registration handles
 -------------------------------------------------------------------- */
struct iser_mem_handles_t {
	ib_mr_t *ibv_mr;
	unsigned long bytes;
};				/*  iser_mem_handles_t */

struct iser_adaptor_t;

/*! --------------------------------------------------------------------
   [struct iser_buf_pool_t]

   Description: Pool of pre-registered buffers
 -------------------------------------------------------------------- */
#define ISER_BUF_POOL_NAME_SIZE 3*ISER_OBJECT_NAME_SIZE

struct iser_buf_pool_t {

	struct iser_adaptor_t *p_adaptor;
	spinlock_t lock;	/* synchronize access to pool */
	struct list_head regions_list;	/* pools may be listed */
	unsigned int num_regions;	/* number of regions in pool */
	unsigned int entry_sz;	/* size of one buffer entry */
	unsigned int alloc_entry_sz;	/* size of memory allocated per entry */
	unsigned int gfp_order;	/* log2 of number of desired pages */
	int uncached;		/* memory in this pool is uncached */
	char name[ISER_BUF_POOL_NAME_SIZE];

};				/* iser_buf_pool_t */

/*! --------------------------------------------------------------------
   [struct iser_mem_t]

   Description: Address based memory descriptor
 -------------------------------------------------------------------- */
struct iser_mem_t {
	uint64_t addr;
	uint64_t size;
};				/* iser_mem_t */

/*! --------------------------------------------------------------------
   [struct hash_table_t]

   Description: Hash table
 -------------------------------------------------------------------- */
struct hash_table_t {
	/* Each entry - a bucket list head */
	struct list_head bucket_head[HASH_TABLE_SIZE];
	/* Synchronize access */
	spinlock_t lock;
};				/* hash_table_t */

/*! --------------------------------------------------------------------
   [struct iser_op_params_t]

   Description: Values of session-wide operational parameters
 -------------------------------------------------------------------- */
struct iser_op_params_t {
	uint32_t InitiatorRecvDataSegmentLength;	/* 512..2*24-1], default: 8K, func: MIN */
	uint32_t TargetRecvDataSegmentLength;	/* 512..2*24-1], default: 8K, func: MIN */
	uint32_t FirstBurstLength;	/* [512..2**24-1], default: 64K, func: MIN */
	uint32_t MaxBurstLength;	/* [512..2**24-1], default: 256K, func: MIN */
	uint16_t MaxConnections;	/* [1..65535], default: 1, func: MIN */
	uint16_t DefaultTime2Wait;	/* [0..3600], default: 2, func: MAX */
	uint16_t DefaultTime2Retain;	/* [0..3600], default: 20, func: MIN */
	uint16_t MaxOutstandingR2T;	/* [1..65535], default: 1, func: MIN */
	uint8_t InitialR2T;	/* [0,1] == [No,Yes], default: Yes, func: OR */
	uint8_t ImmediateData;	/* [0,1] == [No,Yes], default: Yes, func: AND */
	uint8_t DataPDUInOrder;	/* [0,1] == [No,Yes], default: Yes, func: OR */
	uint8_t DataSequenceInOrder;	/* [0,1] == [No,Yes] */
	uint8_t ErrorRecoveryLevel;	/* [0..2], default: 0, func: MIN */
	uint8_t SessionType;	/* [0,1] == [Normal,Discovery] */
};				/* iser_op_params_t */

enum iser_op_param_default {
	defaultInitiatorRecvDataSegmentLength = 8 * 1024,
	defaultTargetRecvDataSegmentLength = 8 * 1024,
	defaultFirstBurstLength = 64 * 1024,
	defaultMaxBurstLength = 256 * 1024,
	defaultMaxConnections = 1,
	defaultDefaultTime2Wait = 2,
	defaultDefaultTime2Retain = 20,
	defaultMaxOutstandingR2T = 1,
	defaultInitialR2T = 1,
	defaultImmediateData = 0,
	defaultDataPDUInOrder = 1,
	defaultDataSequenceInOrder = 1,
	defaultErrorRecoveryLevel = 0,
};

/**
 * A cache of pre-registered buffers.
 */
struct iser_regd_buf_cache {
	struct list_head buf_list;
	spinlock_t lock;
	unsigned int fill;
	unsigned int misses;
	unsigned int hits;
};


/*! --------------------------------------------------------------------
   [struct iser_conn_t]

   Description: Connection descriptor
 -------------------------------------------------------------------- */
struct iser_conn_t {

	iser_conn_handle_t iscsi_conn_h;	/* iSCSI-supplied handle */
	struct iser_adaptor_t *p_adaptor;	/* IA context */
	struct iser_entity_t *p_entity;	/* iSCSI entity */
	atomic_t state;		/* Connection state */
	enum iser_hello_state hello_state;

	atomic_t refs;
	spinlock_t conn_lock;		/* Guards the connection and related structures */

	struct list_head task_list;	/* Head of connection's tasks list */
	struct list_head hash_list;	/* Entry in connections hash buckets */
	struct list_head adaptor_list;	/* Entry in the adaptor's list of connections */
	struct list_head entity_list;	/* Entry in the entity list of connections */

	/* header and data registered buffer caches. */
	struct iser_regd_buf_cache regd_cache;
	/* max fill for the cache */
	unsigned int cache_max;

	/* number of registered memory regions for this connection. */
	atomic_t n_regd_mr;
	wait_queue_head_t regd_mr_wait;

	struct rdma_cm_id *cm_id;
	ib_pd_t *pd;
	ib_qp_t *qp;
	ib_cq_t *cq;

	/* max number of write SGEs allowed by the hardware */
	unsigned int max_sge_wr;
	/* max number of read SGEs allowed by the hardware */
	unsigned int max_sge_rd;

	/* max number of send SGEs (allowed by iSER) */
	unsigned int max_sge_send;
	/* max number of recv SGEs (allowed by iSER) */
	unsigned int max_sge_recv;

#if !defined(__KERNEL__)
	struct ibv_comp_channel *comp_chan;
#else
	struct workqueue_struct *cq_work_q;
	struct work_struct cq_work;
#endif /* !__KERNEL__ */

	/* serializes the active connect state machine
	   (and the Hello Exchange) */
	wait_queue_head_t conn_wait;
	atomic_t conn_done;

	/* flow control */
	int iser_read_depth;
	int max_rdma_read_req_in;
	int max_rdma_read_req_out;
	atomic_t outstanding_reads;
	wait_queue_head_t read_depth_wait;

	thread_handle cq_thrd_handle;
	struct semaphore startstop_sem;

	atomic_t post_recv_buf_count;	/* Counter of posted recv buffers */
	atomic_t post_send_buf_count;	/* Counter of posted send buffers */

#if defined(CONFIG_ISCSI_DEBUG)
	unsigned int min_post_recv_bufs;
#endif /* CONFIG_ISCSI_DEBUG */

	struct list_head dto_list;
	wait_queue_head_t dto_list_wait;

	struct iser_op_params_t param;	/* Session operational parameters */
	unsigned int max_outstand_cmds;	/* current max allowed commands */
	unsigned int initial_post_recv_bufs_num;	/* posted at the beginning */
	unsigned int alloc_post_recv_bufs_num;	/* allocated through a buffer pool */

	int disc_evt_flag;	/* Set if EVENT_DISCONNECTED is received from DAPL */
	wait_queue_head_t disconnect_wait_q;	/* queue for waiting for a disconnection completion */

	struct iser_buf_pool_t *spare_post_recv_pool;	/* Pool of post_recv buffers */
	struct iser_buf_pool_t *spare_send_data_pool;	/* Pool of send_data buffers */

	char name[ISER_OBJECT_NAME_SIZE];	/* Conn. name - usually ip addr */

#if defined(CONFIG_DEBUG_FS)
	struct dentry *debugfs_dir;
	struct dentry *debugfs_state;
	struct dentry *debugfs_refs;

	struct dentry *debugfs_cache_max;
	struct dentry *debugfs_cache_fill;
	struct dentry *debugfs_cache_misses;
	struct dentry *debugfs_cache_hits;

	struct dentry *debugfs_max_sge_wr;
	struct dentry *debugfs_max_sge_rd;
	struct dentry *debugfs_max_sge_send;
	struct dentry *debugfs_max_sge_recv;
	struct dentry *debugfs_iser_read_depth;
	struct dentry *debugfs_max_rdma_read_req_in;
	struct dentry *debugfs_max_rdma_read_req_out;
	struct dentry *debugfs_outstanding_reads;
	struct dentry *debugfs_post_recv_buf_count;
	struct dentry *debugfs_post_send_buf_count;
#endif /* CONFIG_DEBUG_FS */

};				/* iser_conn_t */


/*! --------------------------------------------------------------------
   [struct iser_regd_buf_t]

   Description: Descriptor of a registered memory buffer
 -------------------------------------------------------------------- */
struct iser_regd_buf_t {
	// memory registration with DAPL
	struct iser_mem_handles_t mem_reg;
	// Origin of the registered buffer
	enum iser_regd_buf_type type;
	// Single buffer virtual addr descr, NULL if not applicable
	struct iser_data_buf_t virt_buf;
	// To be chained here, if freeing upon completion is signaled
	struct list_head free_upon_comp_list;
	// Reference count, memory freed when decremented to 0
	atomic_t ref_count;

	union {
		struct {	/* ISER_PRE_REGD_BUF */
			/* registered buffer cache link. */
			struct list_head rc_next;
			struct iser_conn_t *p_conn;
			unsigned int cache_size;
		} pre_regd;

		struct {	/* ISER_REGD_BUF_USER */
			struct list_head list;
		} user;

		struct {	/* ISER_REGD_BUF_REMOTE */
			uint32_t rmr_context;
		} rdma;
	} as;

};				/* iser_regd_buf_t */


// #ifdef ISER_INITIATOR
//     #define MAX_REGD_BUF_VECTOR_LEN 28
//     #define MAX_PHYS_VEC_IN_TASK    64
// #else
#define MAX_REGD_BUF_VECTOR_LEN	128
#define SQ_SGE_MAX 4
#define MAX_PHYS_VEC_IN_TASK	32
// #endif

/*! --------------------------------------------------------------------
   [struct iser_dto_t]

   Description: DTO descriptor, serves as a cookie as well
 -------------------------------------------------------------------- */
struct iser_dto_t {

	struct iser_task_t *p_task_;
	unsigned int itt;

	struct iser_conn_t *p_conn;
	enum iser_dto_type type;	/* RCV, SEND, RDMA-W, RDMA-R */
	int notify_enable;

	/* simple reference count */
	atomic_t refs;

	/* count of work requests pending for this DTO */
	atomic_t wrs;
	/* set to !0 if any of the wrs had an error */
	unsigned int wr_err;
	unsigned long xfer_len;

	/* Local data - vector of registered buffers */
	unsigned int regd_vector_len;
	struct iser_regd_buf_t *regd[MAX_REGD_BUF_VECTOR_LEN];
	/* here an offset into the registered buffer may be specified */
	unsigned int offset[MAX_REGD_BUF_VECTOR_LEN];
	/* here a smaller size may be specified, if 0, then full size is used */
	unsigned int used_sz[MAX_REGD_BUF_VECTOR_LEN];

	/* to be linked a per-connection dto list etc. */
	struct list_head dto_list;

	union {
		struct {
			struct iser_recv_pdu_t pdu;	/* received pdu descriptor */
		} recv;

		struct {
			ib_mr_t remote;
			unsigned int SN;	/* Data-SN or R2T-SN */
		} rdma;
	} as;

};				/* iser_dto_t */


/*! --------------------------------------------------------------------
   [iser_dto_completion_func_t]

   Description: Function pointer to the DTO completion handler
 -------------------------------------------------------------------- */
typedef int
 (*iser_dto_completion_func_t) (struct iser_dto_t * p_dto,
				unsigned long dto_xfer_len);

/*! --------------------------------------------------------------------
   [iser_enter_hello_exchg_func_t]

   Description: Function pointer to the enter hello exchg handler
 -------------------------------------------------------------------- */
typedef int
 (*iser_enter_hello_exchg_func_t) (struct iser_conn_t * p_conn);

/*! --------------------------------------------------------------------
   [struct iser_task_t]

   Description: Task descriptor
   -------------------------------------------------------------------- */
struct iser_task_t {
	unsigned int itt;	/* Initiator Task Tag */
	void *iscsi_task_h;
	struct iser_conn_t *p_conn;	/* Task is carried out within this connection */

	/* task reference counter */
	atomic_t refs;

	spinlock_t task_lock;

	atomic_t status;		/* task execution status */

	atomic_t rx_count;		/* used-up post-receive buffers */

	/* List of send buffers to free when the task is completed */
	struct list_head send_buff_list;
	/* List of buffers received for this task, to be released when task is complete */
	struct list_head rcv_buff_list;

	struct list_head conn_list;	/* Tasks list of the connection */
	struct list_head hash_list;	/* Hash table bucket entry */

	int dir[ISER_DIRS_NUM];	/* set if direction used */
	unsigned long data_len[ISER_DIRS_NUM];	/* total data length */

	/* RDMA-R DTO, incl. remode reg info and reference to local buf  */
	struct iser_dto_t rdma_read_dto;	/* initiator */
	/* RDMA-R local reg info, referenced by rdma_read_dto */
	struct iser_regd_buf_t rdma_read_local;	/* initiator */
	/* remote data identification */
	ib_mr_t rdma_read_remote;
	/* RDMA-W local reg info, referenced by rdma_write_dto */
	struct iser_regd_buf_t rdma_write_local;	/* initiator */
	/* remote data identification */
	ib_mr_t rdma_write_remote;

	struct iser_mem_t mem_vec[MAX_PHYS_VEC_IN_TASK];

};				/* iser_task_t */

void iser_task_get(struct iser_task_t *);
void iser_task_put(struct iser_task_t *);

#define dto_get_task(x) ((x)->p_task_)

static inline void dto_set_task(struct iser_dto_t *p_dto,
                                struct iser_task_t *p_task)
{
	if (p_task) {
		TRACE(TRACE_ISER, "DTO %p Getting task %p\n",
		      p_dto, p_task);
		iser_task_get(p_task);
		p_dto->itt = p_task->itt;
	} else {
		p_dto->itt = 0;
	}
	p_dto->p_task_ = p_task;
}


/*! --------------------------------------------------------------------
   [struct iser_entity_t]

   Description: descriptor of an iSCSI layer entity
 -------------------------------------------------------------------- */

#define ISER_MAX_ADAPTORS			1
#define ISER_MAX_TARGET_ENTITIES	1
#define ISER_TARGET_MAX_CONN		12
#define ISER_MAX_ENTITIES               ISER_MAX_TARGET_ENTITIES

struct iser_entity_t {
	int registered;
	/* iSER API - called back by iSER */
	struct iser_api_cb_t api_cb;
	/* Extended iSER API - called back by iSER */
	struct iser_ext_api_cb_t ext_api_cb;
	/* DTO completion handler */
	iser_dto_completion_func_t dto_completion_cb;
	/* Enter hello exchange handler */
	iser_enter_hello_exchg_func_t enter_hello_exchg;

	/* listening address (target only) */
	struct sockaddr addr;
	size_t addr_len;

	/* List of all iSER connections on this adaptor */
	spinlock_t conn_lock;
	struct list_head conn_list;
	/* type of the entiry - Initiator,Target */
	enum iser_entity_type type;
	/* name of the iSCSI layer entity provider */
	char provider_name[ISER_OBJECT_NAME_SIZE];
	/* procfs entry */
	struct proc_dir_entry *pent;
};				/* iser_entity_t */

/*! --------------------------------------------------------------------
   [struct iser_adaptor_t]

   Description: iSER Adaptor, representing HCA, RNIC etc.
 -------------------------------------------------------------------- */
struct iser_adaptor_t {
	struct rdma_cm_id *cm_id;

	struct rdma_event_channel *event_chann;
	int has_first_event;

	/* wait queues */
	wait_queue_head_t dat_events_wait_q;	/* wait for kdapl events */
	wait_queue_head_t connect_wait_q;	/* wait for connection, or failure */

	/* pre-registered memory */
	//struct iser_regd_buf_t  regd_mem;

	/* List of all iSER connections on this adaptor */
	spinlock_t conn_lock;	/* synchronizes access to connections list */
	struct list_head conn_list;
	/* Pool of pre-registered headers */

	struct iser_buf_pool_t *header_pool;

	thread_handle event_thrd_handle;
	struct semaphore startstop_sem;

	char name[ISER_OBJECT_NAME_SIZE];

	atomic_t in_callback;

};				/* iser_adaptor_t */

/*! --------------------------------------------------------------------
   [struct iser_global_t]

   Description: Global structures, currently contains:
   + single adaptor descriptor, extention to multiple adaptors possible
   + extended iSER API callbacks set, currently 2. It may become
     useful, if few different iSCSI initiators, or both initiator and
     target from different providers use this iSER datamover.
 -------------------------------------------------------------------- */
struct iser_global_t {
	/* IA context, Future: array of adaptors, by name */
	unsigned int num_adaptors;
	struct iser_adaptor_t adaptor[ISER_MAX_ADAPTORS];
	/* Ext API callbacks registration */
	unsigned int num_entities;
	struct iser_entity_t entity[ISER_MAX_ENTITIES];
	/* Hash tables */
	struct hash_table_t task_hash;	/* hash table for tasks */
	struct hash_table_t conn_hash;	/* connections */

//------------------------------------------------------------------
// For kmem_cache allocation
//------------------------------------------------------------------
	my_kmem_cache_t dto_mem_cache;
	my_kmem_cache_t task_mem_cache;
	my_kmem_cache_t regd_buf_mem_cache;
//------------------------------------------------------------------

#ifdef KDAPL_LOCK
	spinlock_t kdapl_lock;
#endif

#if defined(CONFIG_DEBUG_FS)
	struct dentry *debugfs_num_adaptors;
	struct dentry *debugfs_num_entities;
#endif /* CONFIG_DEBU_FS */

};				/* iser_global_t */

/* ---------------------------------------------------------------------
 * GLOBAL VARIABLES
 * ------------------------------------------------------------------ */
extern struct iser_global_t iser_global;

extern int iser_buffer_aligment_shift;

#if defined(CONFIG_DEBUG_FS)
extern struct dentry *debugfs_datamover;
#endif /* CONFIG_DEBUG_FS */

/* non-sleepable context: !iser_in_interrupt() && !iser_int_disabled() */
/*
#include <asm/hardirq.h>
#define iser_in_interrupt()  in_interrupt()
#define iser_int_disabled()  (!(mfmsr() & MSR_EE))
*/

#endif				/* __ISER_H__ */
/* vi: set tabstop=8: */
