/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#include "iser.h"
#include "iser_global.h"
#include "iser_conn.h"
#include "iser_task.h"
#include "iser_dto.h"
#include "iser_ofa.h"
#include "iser_utils.h"
#include "iser_memory.h"
#include "iser_trace.h"

#include "../common/kernel_workqueue.h"

static void iser_conn_dto_flush_wait(struct iser_conn_t *p_conn);
static void iser_conn_debugfs_init(struct iser_conn_t *p_iser_conn);
static void iser_conn_debugfs_deinit(struct iser_conn_t *p_iser_conn);


static char *conn_state_name[ISER_CONN_STATES_NUM + 1] = {
	"INITIAL",
	"PENDING",
	"UP",
	"SYNC_TERM",
	"ASYNC_TERM",
	"DOWN",

	"ILLEGAL"
};


/**
 * iser_conn_get_state_name -  Retrieves symbolic name of a connection state
 *
 *  returns connection name string
 */
char *iser_conn_get_state_name(struct iser_conn_t *p_iser_conn)	//!< IN - iSER connection descriptor
{
	enum iser_conn_state state;

	state = atomic_read(&p_iser_conn->state);

	return conn_state_name[state <
			       ISER_CONN_STATES_NUM ? state :
			       ISER_CONN_STATES_NUM];
	//return conn_state_name[p_iser_conn->state < ISER_CONN_STATES_NUM ? p_iser_conn->state : ISER_CONN_STATES_NUM];
}				/* iser_conn_get_state_name */

/**
 * iser_conn_alloc -  Allocates and initializes a connection descriptor
 *
 *  returns iSER connection descriptor, or NULL on failure
 */
struct iser_conn_t *iser_conn_alloc(char *name)
{
	struct iser_conn_t *p_iser_conn;

	// Allocate iSER connection structure
	p_iser_conn = my_kmalloc(sizeof(*p_iser_conn), "p_iser_conn");
	if (p_iser_conn == NULL) {
		IERROR("Failed to allocate iSER connection descriptor.\n");
		return NULL;
	}

	memset(p_iser_conn, 0, sizeof(struct iser_conn_t));

	p_iser_conn->hello_state = ISER_HELLO_BEFORE;
	init_waitqueue_head(&p_iser_conn->conn_wait);

	atomic_set(&p_iser_conn->refs, 0);

	atomic_set(&p_iser_conn->n_regd_mr, 0);
	init_waitqueue_head(&p_iser_conn->regd_mr_wait);

	spin_lock_init(&p_iser_conn->conn_lock);

	INIT_LIST_HEAD(&p_iser_conn->task_list);	/* list head of connection's tasks */
	INIT_LIST_HEAD(&p_iser_conn->hash_list);	/* bucket list entry */
	INIT_LIST_HEAD(&p_iser_conn->adaptor_list);	/* adaptor connections list entry */
	INIT_LIST_HEAD(&p_iser_conn->entity_list);	/* entity connections list entry */

#if defined(__KERNEL__)
	p_iser_conn->cq_work_q = create_workqueue(p_iser_conn->name);
	KERNEL_INIT_WORK(&p_iser_conn->cq_work, iser_cq_work_func);
#endif /* __KERNEL__ */

	atomic_set(&p_iser_conn->post_recv_buf_count, 0);
	atomic_set(&p_iser_conn->post_send_buf_count, 0);

#if defined(CONFIG_ISCSI_DEBUG)
	/* set minimum post recv buffer count to "infinity". */
	p_iser_conn->min_post_recv_bufs = (unsigned int) -1;
#endif /* CONFIG_ISCSI_DEBUG */

	INIT_LIST_HEAD(&p_iser_conn->dto_list);
	init_waitqueue_head(&p_iser_conn->dto_list_wait);

	ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_INIT);
	p_iser_conn->disc_evt_flag = 0;

	atomic_set(&p_iser_conn->outstanding_reads, 0);
	init_waitqueue_head(&p_iser_conn->read_depth_wait);

	p_iser_conn->param.InitiatorRecvDataSegmentLength =
	    defaultInitiatorRecvDataSegmentLength;
	p_iser_conn->param.TargetRecvDataSegmentLength =
	    defaultTargetRecvDataSegmentLength;
	p_iser_conn->param.FirstBurstLength = defaultFirstBurstLength;
	p_iser_conn->param.MaxBurstLength = defaultMaxBurstLength;
	p_iser_conn->param.MaxConnections = defaultMaxConnections;
	p_iser_conn->param.DefaultTime2Wait = defaultDefaultTime2Wait;
	p_iser_conn->param.DefaultTime2Retain = defaultDefaultTime2Retain;
	p_iser_conn->param.MaxOutstandingR2T = defaultMaxOutstandingR2T;
	p_iser_conn->param.InitialR2T = defaultInitialR2T;
	p_iser_conn->param.ImmediateData = defaultImmediateData;
	p_iser_conn->param.DataPDUInOrder = defaultDataPDUInOrder;
	p_iser_conn->param.DataSequenceInOrder =
	    defaultDataSequenceInOrder;
	p_iser_conn->param.ErrorRecoveryLevel = defaultErrorRecoveryLevel;

	init_waitqueue_head(&p_iser_conn->disconnect_wait_q);

	iser_init_regd_buf_caches(p_iser_conn);

	strncpy(p_iser_conn->name, name, ISER_OBJECT_NAME_SIZE);

	iser_conn_debugfs_init(p_iser_conn);

	return p_iser_conn;
}				/* iser_conn_alloc */

/**
 * Frees the connection structure allocated by iser_conn_alloc().
 * \param p_conn The connection to free.
 */
static void iser_conn_free(struct iser_conn_t *p_conn)
{
	my_kfree((void**) &p_conn, "p_conn");
}

/**
 * Inc the reference counter of the connection.
 *
 * \param p_conn The connection.
 */
void iser_conn_get(struct iser_conn_t *p_conn)
{
	atomic_inc(&p_conn->refs);
}

/**
 * Decrements the ref counter for the connection and frees the
 * connection when the ref counter is zero.
 *
 * \param p_cone The connection.
 */
void iser_conn_put(struct iser_conn_t *p_conn)
{
	if (atomic_dec_and_test(&p_conn->refs))
		iser_conn_free(p_conn);
}

/**
 * iser_deallocate_connection_resources - iSER API. Implements Deallocate_Connection_Resources primitive.
 *
 * @description
 * Do nothing here... this stuff is all deallocated automatically when
 * the connection disconnect even happens.
 *
 * @returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
 */
iser_status_t iser_deallocate_connection_resources(iser_conn_handle_t iscsi_conn_h)	//!< IN - iSCSI-supplied connection handle
{
#if 0
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_conn_t *p_iser_conn;

	ITRACE_ENTRY();

	// Find the connection
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		IERROR
		    ("Failed to find iSER connection, for iscsi_h = 0x%p\n",
		     iscsi_conn_h);
		return ISER_INVALID_CONN;
		goto deallocate_connection_resources_exit;
	}
	// Deallocate all resources, free connection descriptor
	iser_conn_destroy(p_iser_conn);

      deallocate_connection_resources_exit:

	ITRACE_EXIT();
	return iser_ret;
#else
	return ISER_SUCCESS;
#endif	/* 0 */
}				/* iser_deallocate_connection_resources */

/**
 * iser_param_string_to_boolean - Converts a string into a boolean value, if complies.
 *
 * returns 1 if string equiv. to Yes, 1 if equiv. to No, -1 if not boolean
 */
static int iser_param_string_to_boolean(char *value)	//!< IN - string supposedly containing a boolean value
{
	if (strnicmp(value, "no", 2) == 0 || strncmp(value, "0", 1) == 0)
		return 0;
	else if (strnicmp(value, "yes", 3) == 0
		 || strncmp(value, "1", 1) == 0)
		return 1;
	else
		return -1;
}				/* iser_param_string_to_boolean */

/**
 * iser_notice_key_values - iSER API. Implements Notice_Key_Value primitive
 *
 * An iSCSI layer requests its local iSER datamover layer
 * to take note of the negotiated values of the listed keys.
 *
 * returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN, ISER_ILLEGAL_PARAM)
 */
iser_status_t iser_notice_key_values(iser_conn_handle_t iscsi_conn_h,	//!< IN - iSCSI-supplied connection handle
				     char *key,	//!< IN - string containing the key name
				     char *value)	//!< IN - string containing the value for the key
{
	struct iser_conn_t *p_iser_conn;
	int num_val;

	// Find the connection
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		IERROR("Connection not found, conn_h: %ld\n",
		       (unsigned long) iscsi_conn_h);
		return ISER_INVALID_CONN;
	}

	if (key == NULL || value == NULL) {
		IERROR("NULL key or value\n");
		return ISER_ILLEGAL_PARAM;
	}

	/* Compare key to the supported key p_iser_connnames and retrieve the values */
	if (strcmp(key, "MaxConnections") == 0) {	/* [1..65535] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val > 65535)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.MaxConnections = num_val;
	} else if (strcmp(key, "InitialR2T") == 0) {	/* [0,1] == [No,Yes] */
		num_val = iser_param_string_to_boolean(value);
		if (num_val < 0)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.InitialR2T = num_val;
	} else if (strcmp(key, "ImmediateData") == 0) {	/* [0,1] == [No,Yes] */
		num_val = iser_param_string_to_boolean(value);
		if (num_val < 0)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.ImmediateData = num_val;
	} else if (strcmp(key, "InitiatorRecvDataSegmentLength") == 0) {	/* [512..2**24-1] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val < 512 || num_val > 0x00FFFFFF)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.InitiatorRecvDataSegmentLength =
		    num_val;
	} else if (strcmp(key, "TargetRecvDataSegmentLength") == 0) {	/* [512..2**24-1] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val < 512 || num_val > 0x00FFFFFF)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.TargetRecvDataSegmentLength = num_val;
	} else if (strcmp(key, "MaxBurstLength") == 0) {	/* [512..2**24-1] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val < 512 || num_val > 0x00FFFFFF)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.MaxBurstLength = num_val;
	} else if (strcmp(key, "FirstBurstLength") == 0) {	/* [512..2**24-1] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val < 512 || num_val > 0x00FFFFFF)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.FirstBurstLength = num_val;
	}
	//else if (strcmp(key,"DefaultTime2Wait") == 0) {  /* [0..3600] */
	/*    num_val = simple_strtoul(value,NULL,10);
	   if (num_val > 3600)
	   goto notice_key_values_invalid_val;
	   p_iser_conn->param.DefaultTime2Wait = num_val;
	   } */
	//else if (strcmp(key,"DefaultTime2Retain") == 0) {  /* [0..3600] */
	/*    num_val = simple_strtoul(value,NULL,10);
	   if (num_val > 3600)
	   goto notice_key_values_invalid_val;
	   p_iser_conn->param.DefaultTime2Retain = num_val;
	   } */
	//else if (strcmp(key,"MaxOutstandingR2T") == 0) { /* [1..65535] */
	/*    num_val = simple_strtoul(value,NULL,10);
	   if (num_val > 65535)
	   goto notice_key_values_invalid_val;
	   p_iser_conn->param.MaxOutstandingR2T = num_val;
	   } */
	else if (strcmp(key, "DataPDUInOrder") == 0) {	/* [0,1] == [No,Yes] */
		num_val = iser_param_string_to_boolean(value);
		if (num_val < 0)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.DataPDUInOrder = num_val;
	}
	//else if (strcmp(key,"DataSequenceInOrder") == 0) { /* [0,1] == [No,Yes] */
	/*    num_val = iser_param_string_to_boolean(value);
	   if (num_val < 0)
	   goto notice_key_values_invalid_val;
	   p_iser_conn->param.DataSequenceInOrder = num_val;
	   } */
	else if (strcmp(key, "ErrorRecoveryLevel") == 0) {	/* [0..2] */
		num_val = simple_strtoul(value, NULL, 10);
		if (num_val > 2)
			goto notice_key_values_invalid_val;
		p_iser_conn->param.ErrorRecoveryLevel = num_val;
	} else {
		IERROR("Unsupported key name: %s\n", key);
		return ISER_INVALID_KEY;
	}

	ITRACE(ISER_TRACE_CONN, "Set %s to %d\n", key, num_val);
	return ISER_SUCCESS;

      notice_key_values_invalid_val:
	IERROR("Invalid value: %s for key %s\n", value, key);
	return ISER_INVALID_VALUE;

}				/* iser_notice_key_values */

/**
 * iser_conn_add_task - Add iSER task to an iSER connection
 */
void iser_conn_add_task(struct iser_conn_t *p_iser_conn,	//!< IN - connection descriptor
			struct iser_task_t *p_iser_task)	//!< IN - task descriptor
{
	unsigned long flags;
	ITRACE_ENTRY();

	TRACE(TRACE_ISER, "%s: locking %p->conn_lock\n", __func__, p_iser_conn);
	spin_lock_irqsave(&p_iser_conn->conn_lock, flags);

	list_add(&p_iser_task->conn_list, &p_iser_conn->task_list);
	p_iser_task->p_conn = p_iser_conn;

	spin_unlock_irqrestore(&p_iser_conn->conn_lock, flags);
	TRACE(TRACE_ISER, "%s: unlocking %p->conn_lock\n", __func__, p_iser_conn);

	ITRACE_EXIT();
}				/* iser_conn_add_task */

/**
 * iser_conn_delete_task - Delete iSER task from an iSER connection
 */
void iser_conn_delete_task(struct iser_conn_t *p_iser_conn,	//!< IN - connection descriptor
			   struct iser_task_t *p_iser_task)	//!< IN - task descriptor
{
	unsigned long flags;
	ITRACE_ENTRY();

	TRACE(TRACE_ISER, "%s: locking %p->conn_lock\n", __func__, p_iser_conn);

	BUG_ON(p_iser_task->p_conn == NULL);
	BUG_ON(p_iser_task->p_conn != p_iser_conn);

	spin_lock_irqsave(&p_iser_conn->conn_lock, flags);

	list_del(&p_iser_task->conn_list);
	p_iser_task->p_conn = NULL;

	spin_unlock_irqrestore(&p_iser_conn->conn_lock, flags);
	TRACE(TRACE_ISER, "%s: unlocking %p->conn_lock\n", __func__, p_iser_conn);

	ITRACE_EXIT();
}				/* iser_conn_delete_task */

/**
 * \brief Adds a DTO to the connection's post_recv_dto_list.
 */
void iser_conn_add_dto(struct iser_conn_t *p_conn,
                            struct iser_dto_t *p_dto)
{
	unsigned long flags;

	ITRACE_ENTRY();

	BUG_ON(p_dto->p_conn);

	iser_conn_get(p_conn);
	iser_dto_get(p_dto);

	TRACE(TRACE_ISER, "%s: locking %p->conn_lock\n", __func__, p_conn);
	spin_lock_irqsave(&p_conn->conn_lock, flags);
	list_add(&p_dto->dto_list, &p_conn->dto_list);
	p_dto->p_conn = p_conn;
	spin_unlock_irqrestore(&p_conn->conn_lock, flags);
	TRACE(TRACE_ISER, "%s: unlocking %p->conn_lock\n", __func__, p_conn);

	ITRACE_EXIT();
}

/**
 * \brief Removes a DTO from the receive list.
 */
void iser_conn_remove_dto(struct iser_conn_t *p_conn,
                          struct iser_dto_t *p_dto)
{
	unsigned long flags;

	ITRACE_ENTRY();

	if (p_dto->p_conn != p_conn) {
		IPANIC("Attempt to delete recv DTO: 0x%p from conn: 0x%p "
		       "while referencing conn: 0x%p\n",
		       p_dto, p_conn, p_dto->p_conn);
	}

	TRACE(TRACE_ISER, "%s: locking %p->conn_lock\n", __func__, p_conn);
	spin_lock_irqsave(&p_conn->conn_lock, flags);
	list_del(&p_dto->dto_list);
	p_dto->p_conn = NULL;
	spin_unlock_irqrestore(&p_conn->conn_lock, flags);
	TRACE(TRACE_ISER, "%s: unlocking %p->conn_lock\n", __func__, p_conn);

	iser_dto_put(p_dto);

	wake_up_interruptible(&p_conn->dto_list_wait);
	iser_conn_put(p_conn);
	ITRACE_EXIT();
}

/**
 * iser_release_control - Releases a control DTO descriptor.
 *
 * returns allocated DTO descriptor
 */
iser_status_t
iser_release_control(iser_conn_handle_t iscsi_conn_h,
		     struct iser_recv_pdu_t *p_ctrl_pdu)
{
	struct iser_conn_t *p_iser_conn;
	struct iser_dto_t *p_recv_dto;
	iser_status_t iser_ret = ISER_SUCCESS;

	ITRACE_ENTRY();

	// Find the connection
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		iser_ret = ISER_INVALID_CONN;
		goto release_control_exit;
	}

	if (p_ctrl_pdu == NULL) {
		IERROR("NULL receive control to release, conn: 0x%p\n",
		       iscsi_conn_h);
		iser_ret = ISER_ILLEGAL_PARAM;
		goto release_control_exit;
	}
	// Get the recv DTO descriptor coupled with the PDU
	p_recv_dto =
	    list_entry(p_ctrl_pdu, struct iser_dto_t, as.recv.pdu);
	ITRACE(ISER_TRACE_CTRL_RELEASE,
	       "Control release, DTO:0x%p, as PDU:0x%p\n", p_recv_dto,
	       p_ctrl_pdu);
	if (p_recv_dto->type != ISER_DTO_RCV) {
		IPANIC("Releasing non-RECV type dto: 0x%p, type: %d\n",
		       p_recv_dto, p_recv_dto->type);
	}

	iser_dto_put(p_recv_dto);

      release_control_exit:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_release_control */

/**
 * iser_conn_wait_for_disconnection - Sleep on queue and wait until the connection is fully disconnected
 **/
void iser_conn_wait_for_disconnection(struct iser_conn_t *p_iser_conn)	//!< IN - connection descriptor
{
	struct iser_adaptor_t *p_iser_adaptor = p_iser_conn->p_adaptor;
	int in_callback = atomic_read(&p_iser_adaptor->in_callback);

	if (in_callback && iser_spawn_event_thread(p_iser_adaptor) != 0)
		IPANIC("Failed to start event kernel thread\n");

	ITRACE(ISER_TRACE_CONN,
	       "Waiting for the disconnect event, p_conn: 0x%p\n",
	       p_iser_conn);
	if (atomic_read(&p_iser_conn->state) != ISER_CONN_DOWN) {
		wait_event_interruptible(p_iser_conn->disconnect_wait_q,
					 (atomic_read(&p_iser_conn->state)
					  == ISER_CONN_DOWN));
	}
	ITRACE(ISER_TRACE_CONN, "Got the disconnect event, p_conn: 0x%p\n",
	       p_iser_conn);

	/* now wait for the DTOs to FLUSH */
	iser_conn_dto_flush_wait(p_iser_conn);

	if (in_callback)
		iser_suicide_thread();
}

/**
 * iser_conn_sync_terminate - Triggers start of the disconnection procedures
 */
int iser_conn_sync_terminate(struct iser_conn_t *p_iser_conn)	//!< IN - connection descriptor
{
	int ret_val = -1;

	ITRACE_ENTRY();

	switch (atomic_read(&p_iser_conn->state)) {
	case ISER_CONN_UP:
	case ISER_CONN_PENDING:
		ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_SYNC_TERM);
		ret_val = iser_disconnect(p_iser_conn);	// Start the physical disconnect
		if (ret_val == 0)
			iser_conn_wait_for_disconnection(p_iser_conn);
		else
			IPANIC
			    ("Failed to disconnect gracefully, conn: 0x%p\n",
			     p_iser_conn);
		break;

	case ISER_CONN_SYNC_TERM:
		iser_conn_wait_for_disconnection(p_iser_conn);
		ret_val = 0;
		break;

	case ISER_CONN_ASYNC_TERM:
		ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_SYNC_TERM);
		iser_conn_wait_for_disconnection(p_iser_conn);
		ret_val = 0;
		break;

	case ISER_CONN_DOWN:
		ret_val = 0;	// this may happen only when the iSCSI is being currently notified
	default:
		IERROR("called when in state %s\n",
		       iser_conn_get_state_name(p_iser_conn));
		break;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_conn_sync_terminate */

/**
 * iser_conn_async_terminate - Triggers start of the disconnection procedures
 */
int iser_conn_async_terminate(struct iser_conn_t *p_iser_conn)	//!< IN - connection descriptor
{
	int ret_val = -1;

	ITRACE_ENTRY();

	if (atomic_read(&p_iser_conn->state) == ISER_CONN_UP) {
		// Signal that the connection is being terminated
		ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_ASYNC_TERM);
		ITRACE(ISER_TRACE_CONN,
		       "Conn. 0x%p is being terminated asynchronously\n",
		       p_iser_conn);
		// Start the disconnect procedures
		ret_val = iser_disconnect(p_iser_conn);
	} else {
		IERROR("called when in state %s\n",
		       iser_conn_get_state_name(p_iser_conn));
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_conn_async_terminate */

/**
 * iser_connection_terminate - iSER API. Implements Connection_Terminate primitive.
 *
 *
 * Starts connection teardown process. Waits until all previously posted
 * buffers get flushed. Deallocates all connection resources.
 *
 * returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
 */
iser_status_t iser_connection_terminate(iser_conn_handle_t iscsi_conn_h)	//!< IN - iSCSI-supplied connection handle
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_conn_t *p_iser_conn;

	ITRACE_ENTRY();

	// Find the connection
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		iser_ret = ISER_INVALID_CONN;
		goto out;
	}

	// We need to terminate the connection syncronously
	TRACE(TRACE_DEBUG, "%s: calling iser_conn_sync_terminate\n",
	      __func__);
	if (iser_conn_sync_terminate(p_iser_conn) != 0) {
		iser_ret = ISER_FAILURE;
	}
	iser_conn_destroy(p_iser_conn);

 out:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_connection_terminate */

/**
 * Destroys the cq and qp for the connection, if they have been created.
 * Removes the connection from all of the datastructures that it could
 * reside in:
 * 	- the adaptor
 * 	- the entity
 * 	- the connection hash
 *
 * \param p_iser_conn The connection.
 * \note This must only be called on a connection in state down.
 */
void iser_conn_destroy(struct iser_conn_t *p_iser_conn)
{
	ITRACE_ENTRY();

	BUG_ON(atomic_read(&p_iser_conn->state) != ISER_CONN_DOWN);

	iser_conn_debugfs_deinit(p_iser_conn);

	/* release cached, registered buffers. */
	iser_release_regd_buf_caches(p_iser_conn);

	/* wait for all registered memory regions to be unregistered. */
#if defined(__KERNEL__)
	TRACE(TRACE_ISER, "%s: [%lu] Waiting for %d MRs to unregister\n",
	      __func__, jiffies, atomic_read(&p_iser_conn->n_regd_mr));
	wait_event_timeout(p_iser_conn->regd_mr_wait,
	                   atomic_read(&p_iser_conn->n_regd_mr) == 0,
	                   180 * HZ);
	if (atomic_read(&p_iser_conn->n_regd_mr) == 0)
		TRACE(TRACE_DEBUG, "%s: [%lu] all MRs unregistered properly\n",
		      __func__, jiffies);
	else
		TRACE(TRACE_DEBUG, "%s: [%lu] MRs unregistration timed "
		      "out %d remaining\n",
		      __func__, jiffies, atomic_read(&p_iser_conn->n_regd_mr));
#endif	/* __KERNEL__ */

	// Destroy the CQ and QP for the connection
	iser_destroy_cq_and_qp(p_iser_conn);

	// Detach from the adaptor connections list
	iser_adaptor_remove_connection(p_iser_conn);

	// Detach from the entity connections list
	iser_entity_remove_connection(p_iser_conn);

	// Detach from connections hash
	hash_delete_iser_conn(p_iser_conn);

	ITRACE_EXIT();
}				/* iser_conn_free */

/**
 * iser_complete_connection_termination - Checks if the connection may be terminated and terminates if possible
 *
 * returns 0 if the connection was terminated successfully ; 1 if the connection cannot be terminated now
 */
int iser_complete_connection_termination(struct iser_conn_t *p_iser_conn)	//!< IN - iSER connection descriptor
{
	int rcv_buf_count;
	int send_buf_count;
	int retval;
	unsigned int cur_conn_state;
	struct iser_task_t *p_iser_task;
	struct list_head *node, *next;

	ITRACE_ENTRY();

	rcv_buf_count = atomic_read(&p_iser_conn->post_recv_buf_count);
	send_buf_count = atomic_read(&p_iser_conn->post_send_buf_count);

	// Check if this connection may be terminated now
	cur_conn_state = atomic_read(&p_iser_conn->state);
	if (cur_conn_state != ISER_CONN_ASYNC_TERM
	    && cur_conn_state != ISER_CONN_SYNC_TERM) {
		IPANIC("Illegal state (%s)",
		       iser_conn_get_state_name(p_iser_conn));
	}
	ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_DOWN);

	ITRACE(ISER_TRACE_CONN,
	       "Connection: 0x%p, may be terminated (%s)!\n", p_iser_conn,
	       (cur_conn_state ==
		ISER_CONN_ASYNC_TERM ? "ASYNC" : "SYNC"));

	iser_conn_dto_flush_wait(p_iser_conn);

	// Free all tasks in this connection
	list_for_each_safe(node, next, &p_iser_conn->task_list) {
		p_iser_task = container_of(node, struct iser_task_t,
		                           conn_list);
		TRACE_WARNING("%s: Removing remaining task %p\n",
		              __func__, p_iser_task);
		list_del(&p_iser_task->conn_list);
		p_iser_task->p_conn = NULL;

		/* We are forcing the task to terminate, so we force the
 		 * ref-count to be zero. */
		atomic_set(&p_iser_task->refs, 0);
		iser_task_free(p_iser_task);
	}

	if (cur_conn_state == ISER_CONN_SYNC_TERM) {
		// If the connection was terminated syncronously, wake up the upper layer (waiting on a sleep queue)
		TRACE(TRACE_ISER,
		      "%s: cur_conn_state == ISER_CONN_SYNC_TERM\n",
		      __func__);
		wake_up_interruptible(&p_iser_conn->disconnect_wait_q);
	}

	// If the connection was terminated asynchronously, notify the upper layer
	if (cur_conn_state == ISER_CONN_ASYNC_TERM
	    && p_iser_conn->hello_state == ISER_HELLO_DONE) {

		atomic_inc(&p_iser_conn->p_adaptor->in_callback);
		p_iser_conn->p_entity->api_cb.
		    conn_terminate_notify(p_iser_conn->iscsi_conn_h);
		atomic_dec(&p_iser_conn->p_adaptor->in_callback);
	}

	retval = 0;

	ITRACE_EXIT();
	return retval;
}				/* iser_complete_connection_termination */

/**
 * iser_post_receive_control - Allocates, initializes and posts receive DTO.
 *
 * returns 0 on success, -1 on failure
 */
int iser_post_receive_control(unsigned int buff_size,	//!< IN - for buffer allocation.
			      struct iser_conn_t *p_iser_conn)	//!< IN - iSER connection descriptor
{
	struct iser_dto_t *p_recv_dto;
	struct iser_regd_buf_t *p_regd_buf;
	int ret_val = 0;
	int i;

	ITRACE_ENTRY();

	/* Create & init send DTO descriptor */
	ITRACE(ISER_TRACE_RECV_DTO,
	       "Create & init post-recv DTO descriptor\n");
	p_recv_dto = iser_alloc_dto(p_iser_conn, ISER_DTO_RCV);
	if (!p_recv_dto) {
		IERROR
		    ("Failed to allocate DTO descriptor for post recv buffer\n");
		ret_val = -1;
		goto post_receive_control_exit;
	}

	/* over-ride the default notify_enable */
	p_recv_dto->notify_enable = 0;
	dto_set_task(p_recv_dto, NULL);	/* can't tell at this stage */

	/* Create an iSER headers buffer */
	ITRACE(ISER_TRACE_RECV_DTO, "Allocate iSER buffer\n");
	p_regd_buf = iser_buffer_alloc(ISER_TOTAL_HEADERS_LEN + buff_size,
				       p_iser_conn, DMA_FROM_DEVICE);
	if (p_regd_buf == NULL) {
		IERROR
		    ("Failed to allocate registered buffer from pool, post recv buffer\n");
		ret_val = -1;
		goto post_receive_control_exit;
	}
	i = iser_dto_add_regd_buff(p_recv_dto, p_regd_buf, USE_NO_OFFSET,
				   USE_SIZE(ISER_TOTAL_HEADERS_LEN));

	ITRACE(ISER_TRACE_RECV_DTO, "Call post recieve with the DTO: %p\n",
	       p_recv_dto);
	if (iser_post_recv(p_recv_dto) != 0) {
		IERROR("Failed to post receive DTO\n");
		ret_val = -1;
		goto post_receive_control_exit;
	}

	/* Increment the post-recv buffers counter */
	iser_conn_inc_post_recv_cnt(p_iser_conn);

      post_receive_control_exit:

	if (ret_val != 0 && p_recv_dto != NULL)
		iser_conn_remove_dto(p_iser_conn, p_recv_dto);
	ITRACE_EXIT();
	return ret_val;
}				/* iser_post_receive_control */

/**
 * \brief Allocates a new connection for the given adaptor.
 */
struct iser_conn_t *iser_adaptor_new_conn(struct iser_adaptor_t *p_adaptor,
					  char *name)
{
	struct iser_conn_t *p_iser_conn;

	ITRACE_ENTRY();

	// Allocate iSER connection structure
	p_iser_conn = iser_conn_alloc(name);
	if (p_iser_conn == NULL) {
		IERROR("Failed to allocate iSER connection.\n");
		ITRACE_EXIT();
		return NULL;
	}

	/* Add the connection to the adaptor's connections list */
	iser_adaptor_add_connection(p_adaptor, p_iser_conn);

	ITRACE_EXIT();
	return p_iser_conn;
}

/**
 * Initialize the debugfs entires for the given connection.
 */
static void iser_conn_debugfs_init(struct iser_conn_t *p_iser_conn)
{

#if defined(CONFIG_DEBUG_FS)
	TRACE(TRACE_ISER, "%s: creating debugfs directory\n", __func__);
	if (debugfs_datamover) {
		p_iser_conn->debugfs_dir =
			debugfs_create_dir(p_iser_conn->name,
					   debugfs_datamover);
		if (IS_ERR(p_iser_conn->debugfs_dir)) {
			TRACE_WARNING("Unable to create %s debugfs entry\n",
				      p_iser_conn->name);
			p_iser_conn->debugfs_dir = NULL;
		}
	}

	if (p_iser_conn->debugfs_dir) {
		TRACE(TRACE_ISER, "%s: creating debugfs state\n", __func__);
		p_iser_conn->debugfs_state =
			debugfs_create_atomic_t("state",
						ISER_DEBUGFS_PERMS,
						p_iser_conn->debugfs_dir,
						&p_iser_conn->state);
		if (IS_ERR(p_iser_conn->debugfs_state)) {
			TRACE_WARNING("Unable to create state "
				      "debugfs entry\n");
			p_iser_conn->debugfs_state = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs refs\n", __func__);
		p_iser_conn->debugfs_refs =
			debugfs_create_atomic_t("refs",
						ISER_DEBUGFS_PERMS,
						p_iser_conn->debugfs_dir,
						&p_iser_conn->refs);
		if (IS_ERR(p_iser_conn->debugfs_refs)) {
			TRACE_WARNING("Unable to create refs "
				      "debugfs entry\n");
			p_iser_conn->debugfs_refs = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs cache_max\n",
		      __func__);
		p_iser_conn->debugfs_cache_max =
			debugfs_create_u32("cache_max",
					   0644, /* writable */
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->cache_max);
		if (IS_ERR(p_iser_conn->debugfs_cache_max)) {
			TRACE_WARNING("Unable to create cache_max "
				      "debugfs entry\n");
			p_iser_conn->debugfs_cache_max = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs cache_misses\n",
		      __func__);
		p_iser_conn->debugfs_cache_misses =
			debugfs_create_u32("cache_misses",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->regd_cache.misses);
		if (IS_ERR(p_iser_conn->debugfs_cache_misses)) {
			TRACE_WARNING("Unable to create cache_misses "
				      "debugfs entry\n");
			p_iser_conn->debugfs_cache_misses = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs cache_hits\n",
		      __func__);
		p_iser_conn->debugfs_cache_hits =
			debugfs_create_u32("cache_hits",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->regd_cache.hits);
		if (IS_ERR(p_iser_conn->debugfs_cache_hits)) {
			TRACE_WARNING("Unable to create cache_hits "
				      "debugfs entry\n");
			p_iser_conn->debugfs_cache_hits = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs cache_fill\n",
		      __func__);
		p_iser_conn->debugfs_cache_fill =
			debugfs_create_u32("cache_fill",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->regd_cache.fill);
		if (IS_ERR(p_iser_conn->debugfs_cache_fill)) {
			TRACE_WARNING("Unable to create cache_fill "
				      "debugfs entry\n");
			p_iser_conn->debugfs_cache_fill = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_sge_wr\n",
		      __func__);
		p_iser_conn->debugfs_max_sge_wr =
			debugfs_create_u32("max_sge_wr",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_sge_wr);
		if (IS_ERR(p_iser_conn->debugfs_max_sge_wr)) {
			TRACE_WARNING("Unable to create max_sge_wr "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_sge_wr = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_sge_rd\n",
		      __func__);
		p_iser_conn->debugfs_max_sge_rd =
			debugfs_create_u32("max_sge_rd",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_sge_rd);
		if (IS_ERR(p_iser_conn->debugfs_max_sge_rd)) {
			TRACE_WARNING("Unable to create max_sge_rd "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_sge_rd = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_sge_send\n",
		      __func__);
		p_iser_conn->debugfs_max_sge_send =
			debugfs_create_u32("max_sge_send",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_sge_send);
		if (IS_ERR(p_iser_conn->debugfs_max_sge_send)) {
			TRACE_WARNING("Unable to create max_sge_send "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_sge_send = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_sge_recv\n",
		      __func__);
		p_iser_conn->debugfs_max_sge_recv =
			debugfs_create_u32("max_sge_recv",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_sge_recv);
		if (IS_ERR(p_iser_conn->debugfs_max_sge_recv)) {
			TRACE_WARNING("Unable to create max_sge_recv "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_sge_recv = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs iser_read_depth\n",
		      __func__);
		p_iser_conn->debugfs_iser_read_depth =
			debugfs_create_u32("iser_read_depth",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->iser_read_depth);
		if (IS_ERR(p_iser_conn->debugfs_iser_read_depth)) {
			TRACE_WARNING("Unable to create iser_read_depth "
				      "debugfs entry\n");
			p_iser_conn->debugfs_iser_read_depth = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_rdma_read_req_in\n",
		      __func__);
		p_iser_conn->debugfs_max_rdma_read_req_in =
			debugfs_create_u32("max_rdma_read_req_in",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_rdma_read_req_in);
		if (IS_ERR(p_iser_conn->debugfs_max_rdma_read_req_in)) {
			TRACE_WARNING("Unable to create max_rdma_read_req_in "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_rdma_read_req_in = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs max_rdma_read_req_out\n",
		      __func__);
		p_iser_conn->debugfs_max_rdma_read_req_out =
			debugfs_create_u32("max_rdma_read_req_out",
					   ISER_DEBUGFS_PERMS,
					   p_iser_conn->debugfs_dir,
					   &p_iser_conn->max_rdma_read_req_out);
		if (IS_ERR(p_iser_conn->debugfs_max_rdma_read_req_out)) {
			TRACE_WARNING("Unable to create max_rdma_read_req_out "
				      "debugfs entry\n");
			p_iser_conn->debugfs_max_rdma_read_req_out = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs outstanding_reads\n",
		      __func__);
		p_iser_conn->debugfs_outstanding_reads =
			debugfs_create_atomic_t("outstanding_reads",
						ISER_DEBUGFS_PERMS,
						p_iser_conn->debugfs_dir,
						&p_iser_conn->outstanding_reads);
		if (IS_ERR(p_iser_conn->debugfs_post_recv_buf_count)) {
			TRACE_WARNING("Unable to create outstanding_reads "
				      "debugfs entry\n");
			p_iser_conn->debugfs_outstanding_reads = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs post_recv_buf_count\n",
		      __func__);
		p_iser_conn->debugfs_post_recv_buf_count =
			debugfs_create_atomic_t("post_recv_buf_count",
						ISER_DEBUGFS_PERMS,
						p_iser_conn->debugfs_dir,
						&p_iser_conn->post_recv_buf_count);
		if (IS_ERR(p_iser_conn->debugfs_post_recv_buf_count)) {
			TRACE_WARNING("Unable to create post_recv_buf_count "
				      "debugfs entry\n");
			p_iser_conn->debugfs_post_recv_buf_count = NULL;
		}

		TRACE(TRACE_ISER, "%s: creating debugfs post_send_buf_count\n",
		      __func__);
		p_iser_conn->debugfs_post_send_buf_count =
			debugfs_create_atomic_t("post_send_buf_count",
						ISER_DEBUGFS_PERMS,
						p_iser_conn->debugfs_dir,
						&p_iser_conn->post_send_buf_count);
		if (IS_ERR(p_iser_conn->debugfs_post_send_buf_count)) {
			TRACE_WARNING("Unable to create post_send_buf_count "
				      "debugfs entry\n");
			p_iser_conn->debugfs_post_send_buf_count = NULL;
		}
	}
#endif /* CONFIG_DEBUG_FS */
}

/**
 * Deinitialize the connection's debugfs files.
 */
static void iser_conn_debugfs_deinit(struct iser_conn_t *p_iser_conn)
{
#if defined(CONFIG_DEBUG_FS)
	if (p_iser_conn->debugfs_post_send_buf_count)
		debugfs_remove(p_iser_conn->debugfs_post_send_buf_count);
	if (p_iser_conn->debugfs_post_recv_buf_count)
		debugfs_remove(p_iser_conn->debugfs_post_recv_buf_count);
	if (p_iser_conn->debugfs_outstanding_reads)
		debugfs_remove(p_iser_conn->debugfs_outstanding_reads);
	if (p_iser_conn->debugfs_max_rdma_read_req_out)
		debugfs_remove(p_iser_conn->debugfs_max_rdma_read_req_out);
	if (p_iser_conn->debugfs_max_rdma_read_req_in)
		debugfs_remove(p_iser_conn->debugfs_max_rdma_read_req_in);
	if (p_iser_conn->debugfs_iser_read_depth)
		debugfs_remove(p_iser_conn->debugfs_iser_read_depth);
	if (p_iser_conn->debugfs_max_sge_recv)
		debugfs_remove(p_iser_conn->debugfs_max_sge_recv);
	if (p_iser_conn->debugfs_max_sge_send)
		debugfs_remove(p_iser_conn->debugfs_max_sge_send);
	if (p_iser_conn->debugfs_max_sge_rd)
		debugfs_remove(p_iser_conn->debugfs_max_sge_rd);
	if (p_iser_conn->debugfs_max_sge_wr)
		debugfs_remove(p_iser_conn->debugfs_max_sge_wr);
	if (p_iser_conn->debugfs_cache_fill)
		debugfs_remove(p_iser_conn->debugfs_cache_fill);
	if (p_iser_conn->debugfs_cache_hits)
		debugfs_remove(p_iser_conn->debugfs_cache_hits);
	if (p_iser_conn->debugfs_cache_misses)
		debugfs_remove(p_iser_conn->debugfs_cache_misses);
	if (p_iser_conn->debugfs_cache_max)
		debugfs_remove(p_iser_conn->debugfs_cache_max);
	if (p_iser_conn->debugfs_refs)
		debugfs_remove(p_iser_conn->debugfs_refs);
	if (p_iser_conn->debugfs_state)
		debugfs_remove(p_iser_conn->debugfs_state);
	if (p_iser_conn->debugfs_dir)
		debugfs_remove(p_iser_conn->debugfs_dir);
#endif /* CONFIG_DEBUG_FS */
}

/**
 * Waits for all of the DTOs on the given connection to FLUSH.
 *
 * \param p_conn The connection.
 */
static void iser_conn_dto_flush_wait(struct iser_conn_t *p_conn)
{
	struct list_head *node, *next;
	struct iser_dto_t *p_dto;

	/*
	 * Wait for all DTOs to be completed.  If there is an error, the
	 * DTOs will be flushed, just wait for that to finish.
	 */
	TRACE(TRACE_ISER, "%s: Waiting for all DTOs to FLUSH\n", __func__);
	wait_event_interruptible_timeout(p_conn->dto_list_wait,
	               list_empty(&p_conn->dto_list), 60 * HZ);

	if (list_empty(&p_conn->dto_list))
		TRACE(TRACE_DEBUG, "%s: all DTOs FLUSHED properly\n",
		      __func__);
	else
		TRACE(TRACE_DEBUG, "%s: DTOs FLUSH timeout\n",
		      __func__);

	/* Free all Posted receive DTOs in this connection not flushed
	 * before the timeout expired. */
	list_for_each_safe(node, next, &p_conn->dto_list) {
		p_dto = container_of(node, struct iser_dto_t, dto_list);
/*
		TRACE_WARNING("%s: Removing DTO %p (dto_list=%p) which never FLUSHED\n",
		              __func__, p_dto, &p_dto->dto_list);
*/
		if (p_dto->p_conn)
			iser_conn_remove_dto(p_conn, p_dto);
	}

}

/* vi: set tabstop=8: */
