/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <asm/io.h>
#endif				/* !__KERNEL__ */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_dto.h"
#include "iser_conn.h"
#include "iser_task.h"
#include "iser_memory.h"
#include "iser_ofa.h"
#include "iser_utils.h"
#include "iser_platform.h"
#include "iser_trace.h"

/**
 * iser_dto_add_regd_buff - Increments the reference count for the registered buffer & adds it to the DTO object
 *
 * returns index of used buffer
 */
int iser_dto_add_regd_buff(struct iser_dto_t *p_dto,	//<! iSER DTO object, which the buffer is added to
			   struct iser_regd_buf_t *p_regd_buf,	//<! iSER regsitered buffer to be added to the dto
			   unsigned long use_offset,
			   unsigned long use_size)
{
	int add_idx;
	ITRACE_ENTRY();

	iser_regd_buff_reference(p_regd_buf);

	add_idx = p_dto->regd_vector_len++;

	if (add_idx >= MAX_REGD_BUF_VECTOR_LEN)
		IPANIC("%s: Too many registered buffers %d\n", __func__, add_idx);

	TRACE(TRACE_ISER, "dto=%p, adding p_regd_buf=%p ibv_mr=(%p, %lu) "
	      "at index=%d\n", p_dto, p_regd_buf,
	      p_regd_buf->mem_reg.ibv_mr->addr,
	      (unsigned long) p_regd_buf->mem_reg.ibv_mr->length, add_idx);

	p_dto->regd[add_idx] = p_regd_buf;
	p_dto->used_sz[add_idx] = use_size;
	p_dto->offset[add_idx] = use_offset;

	ITRACE_EXIT();
	return add_idx;
}				/* iser_regd_buff_use */

void iser_dto_clone_regd_buffs(struct iser_dto_t *p_dst_dto,	//<! iSER DTO object, cloning dst
			       struct iser_dto_t *p_src_dto,	//<! iSER DTO object, cloning src
			       unsigned long offset,	//<! offset to skip
			       unsigned long size)	//<! size to use
{
	unsigned long remaining_offset = offset;
	unsigned long remaining_size = size;
	unsigned long regd_buf_size;
	unsigned long used_size;
	int i;

	ITRACE_ENTRY();

	for (i = 0; i < p_src_dto->regd_vector_len; i++) {
		regd_buf_size = p_src_dto->used_sz[i] > 0 ?
		    p_src_dto->used_sz[i] :
		    p_src_dto->regd[i]->mem_reg.bytes;

		if (remaining_offset < regd_buf_size) {
			used_size =
			    MIN(remaining_size,
				regd_buf_size - remaining_offset);
			iser_dto_add_regd_buff(p_dst_dto,
					       p_src_dto->regd[i],
					       USE_OFFSET(p_src_dto->
							  offset[i] +
							  remaining_offset),
					       USE_SIZE(used_size));
			remaining_size -= used_size;
			if (remaining_size == 0) {
				break;
			} else {
				remaining_offset = 0;
			}
		} else {
			remaining_offset -= regd_buf_size;
		}
	}
	if (remaining_size > 0) {
		IPANIC
		    ("size to clone: %ld exceeds by %ld the total size of src DTO: 0x%p, dst DTO: 0x%p, task: 0x%p\n",
		     size, remaining_size, p_src_dto, p_dst_dto,
		     dto_get_task(p_dst_dto));
	}

	ITRACE_EXIT();
}				/* iser_dto_clone_regd_buffs */


/**
 * iser_dto_add_local_memory - Register local memory, possibly using a number of registration
 * calls; add each registered buffer to the current DTO descriptor
 *
 * returns 0 on success, -1 on failure
 */
int iser_dto_add_local_memory(struct iser_dto_t *p_dto,	//<! IN - iSER DTO object
			      struct iser_data_buf_t *p_mem,	//<! IN - memory to register and add
			      struct iser_mem_t *p_phys_vec,	//<! IN - phys mem descriptor vector
			      unsigned int phys_vec_len,	//<! IN - length of phys mem vector
			      unsigned long *p_total_sz)	//<! OUT - total size of memory
{
	unsigned int used_phys_vec_len = 0;
	int phys_vec_to_free = 0;
	struct iser_regd_buf_t *p_regd_buf;
	int cur_buf;
	int aligned_len;
	unsigned long aligned_data_sz;
	int ret_val = 0;
	enum dma_data_direction dir;

	ITRACE_ENTRY();

	if (p_phys_vec == NULL) {
		IPANIC("NULL phys mem vector\n");
	}
	if (phys_vec_len == 0) {
		IPANIC("Zero phys mem vector length\n");
	}
	if (p_dto->type == ISER_DTO_RCV || p_dto->type == ISER_DTO_RDMA_R)
		dir = DMA_FROM_DEVICE;
	else
		dir = DMA_TO_DEVICE;


	if (p_mem->type == ISER_BUF_TYPE_SINGLE) {
		p_phys_vec->addr = (unsigned long) p_mem->p_buf;
		p_phys_vec->size = p_mem->size;

		if (p_total_sz != NULL)
			*p_total_sz = p_mem->size;

		used_phys_vec_len = 1;

		ITRACE(ISER_TRACE_LOCAL_MEM_REG,
		       "Adding SINGLE buffer virt: 0x%p phys: 0x%08lX, sz: %ld to dto: 0x%p\n",
		       p_mem->p_buf, (unsigned long) p_phys_vec->addr,
		       (unsigned long) p_phys_vec->size, p_dto);
	} else {
		if (p_mem->size > phys_vec_len) {
			ITRACE(ISER_TRACE_LOCAL_MEM_REG,
			       "Allocating phys mem descriptor for %d entries\n",
			       p_mem->size);
			p_phys_vec = (struct iser_mem_t *)
			    my_kmalloc(p_mem->size * sizeof(*p_phys_vec), "p_phys_vec");
			if (p_phys_vec == NULL) {
				IERROR
				    ("Failed to allocate phys mem descriptor for %d entries\n",
				     p_mem->size);
				ret_val = -1;
				goto dto_add_local_memory_exit;
			}
			phys_vec_to_free = 1;
		}

		switch (p_mem->type) {
		case ISER_BUF_TYPE_IOVEC_VIRT:
			ITRACE(ISER_TRACE_LOCAL_MEM_REG,
			       "Translating IOVEC_VIRT sz: %d to dto: 0x%p\n",
			       p_mem->size, p_dto);
			used_phys_vec_len =
			    iser_iovec_virt_to_mem_t(dto_get_task(p_dto),
						     dir,
						     p_mem, p_phys_vec,
						     p_total_sz);
			TRACE(TRACE_ISER, "%s: used_phys_vec_len=%d\n",
			      __func__, used_phys_vec_len);
			break;

		case ISER_BUF_TYPE_SCATTERLIST:
			ITRACE(ISER_TRACE_LOCAL_MEM_REG,
			       "Translating SCATTERLIST sz: %d to dto: 0x%p\n",
			       p_mem->size, p_dto);

			used_phys_vec_len =
			    iser_scatterlist_to_mem_t(dto_get_task(p_dto),
						      dir,
						      p_mem,
						      p_phys_vec,
						      p_total_sz);
			break;

		default:
			IPANIC
			    ("Unsupported data buffer descriptor type: %d in p_data: 0x%p\n",
			     p_mem->type, p_mem);
			break;
		}
	}

	cur_buf = 0;
	do {
		/* Determine an aligned stretch */
		aligned_len =
		    iser_iovec_aligned_length(&p_phys_vec[cur_buf],
					      used_phys_vec_len,
					      &aligned_data_sz);
		ITRACE(ISER_TRACE_LOCAL_MEM_REG,
		       "Registering %d aligned phys.entries out of %d, size: %ld, to add to DTO:0x%p, now with %d regd.bufs\n",
		       aligned_len, used_phys_vec_len, aligned_data_sz,
		       p_dto, p_dto->regd_vector_len);

		if (aligned_len > used_phys_vec_len) {
			IPANIC
			    ("found aligned_len > remaing num. of entries to add\n");
		}


		p_regd_buf = iser_alloc_regd_buf();
		if (p_regd_buf == NULL) {
			IERROR("Failed to allocate registered buffer\n");
			ret_val = -1;
			goto dto_add_local_memory_exit;
		}
		memset(p_regd_buf, 0, sizeof(struct iser_regd_buf_t));
		p_regd_buf->type = ISER_REGD_BUF_USER;
		p_regd_buf->virt_buf.type = ISER_BUF_TYPE_SINGLE;

		if (aligned_len == 1) {
			p_regd_buf->virt_buf.p_buf = (void *) (unsigned long) p_phys_vec[cur_buf].addr;
			p_regd_buf->virt_buf.size = p_phys_vec[cur_buf].size;
			TRACE(TRACE_ISER, "%s: virt_buf=%p, size=%u\n",
			      __func__, p_regd_buf->virt_buf.p_buf,
			      p_regd_buf->virt_buf.size);
		}

		ret_val = iser_register_phys_mem(p_dto->p_conn,
						 &p_phys_vec[cur_buf],
						 aligned_len,
						 aligned_data_sz,
						 DAT_MEM_PRIV_ALL_FLAG,
						 &p_regd_buf->mem_reg,
						 NULL,
		                                 dir);

		if (ret_val != 0) {
			IERROR
			    ("Failed to register %d phys entries starting from %d\n",
			     aligned_len, cur_buf);
			// ToDo: deregister all previously allocated memory
			goto dto_add_local_memory_exit;
		}

		iser_dto_add_regd_buff(p_dto, p_regd_buf, USE_NO_OFFSET,
				       USE_ENTIRE_SIZE);
		cur_buf += aligned_len;
		used_phys_vec_len -= aligned_len;

		ITRACE(ISER_TRACE_LOCAL_MEM_REG,
		       "Added regd.buf:0x%p to DTO:0x%p, now %d regd.bufs, to add %d phys.entries\n",
		       p_regd_buf, p_dto, p_dto->regd_vector_len,
		       used_phys_vec_len);

	} while (used_phys_vec_len > 0);

      dto_add_local_memory_exit:

	if (phys_vec_to_free) {
		ITRACE(ISER_TRACE_LOCAL_MEM_REG,
		       "Freeing phys mem descriptor for %d entries\n",
		       p_mem->size);
		my_kfree((void**) &p_phys_vec, "p_phys_vec");
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_dto_add_local_memory */

/**
 * Frees DTO descriptor and all associated objects
 *
 * Locks can not be held while this function is called!
 *
 * returns 0 on success, -1 on failure
 */
int dbg_iser_dto_free(struct iser_dto_t *p_dto, char *caller)
{
	int i;

	ITRACE_ENTRY();

	ITRACE(ISER_TRACE_REGD_BUFFER, "%s called from %s\n",
	       __func__, caller);

	/* decrement the ref count, if there are still refs held for this DTO,
 	 * then do not delete it */
	if (!atomic_dec_and_test(&p_dto->refs)) {
		TRACE(TRACE_ISER, "%s: p_dto->refs=%d\n",
		      __func__, atomic_read(&p_dto->refs));
		ITRACE_EXIT();
		return 0;
	}

	if (dto_get_task(p_dto)) {
		TRACE(TRACE_ISER, "DTO %p Putting task %p\n",
		      p_dto, dto_get_task(p_dto));
		iser_task_put(dto_get_task(p_dto));
		dto_set_task(p_dto, NULL);
	}

	BUG_ON(dto_get_task(p_dto));
	BUG_ON(p_dto->p_conn);

	// Pass all registered buffers, and free them if needed
	for (i = 0; i < p_dto->regd_vector_len; i++) {
		int release_deferred;

		ITRACE(ISER_TRACE_REGD_BUFFER, "Releasing DTO:0x%p, "
		       "regd.buf:0x%p, #%d\n", p_dto, p_dto->regd[i], i);

		if (iser_regd_buff_release(p_dto->regd[i],
		    &release_deferred) < 0) {
			IPANIC("While releasing DTO:0x%p, failed to release "
			       "regd.buf:0x%p, #%d\n",
			       p_dto, p_dto->regd[i], i);
		}

		if (release_deferred) {
			ITRACE(ISER_TRACE_REGD_BUFFER, "DTO:0x%p[%d], "
			       "regd.buf:0x%p, - release deferred due to ref "
			       "count=%d\n", p_dto, i, p_dto->regd[i],
			       release_deferred);
		} else {
			p_dto->regd[i] = NULL;
		}
	}

	switch (p_dto->type) {
	case ISER_DTO_RCV:
		ITRACE(ISER_TRACE_RECV_DTO,
		       "Releasing RECV dto descriptor: 0x%p\n", p_dto);
		break;

	case ISER_DTO_SEND:
		ITRACE(ISER_TRACE_SEND_DTO,
		       "Releasing SEND dto descriptor: 0x%p\n", p_dto);
		break;

	case ISER_DTO_RDMA_R:
	case ISER_DTO_RDMA_W:
		ITRACE(ISER_TRACE_SEND_DTO,
		       "Releasing RDMA dto descriptor: 0x%p\n", p_dto);
		break;

	default:
		IPANIC("Unexpected type: %d, in dto: 0x%p\n", p_dto->type,
		       p_dto);
		break;
	}

	my_kmem_cache_free(iser_global.dto_mem_cache, p_dto);

	ITRACE_EXIT();
	return 0;
}

/**
 * iser_dto_completion_error - Handles error in DTO completion
 *
 * returns 0 on success, -1 on failure
 */
int iser_dto_completion_error(struct iser_dto_t *p_dto)	//!< IN - completed DTO descriptor
{
	struct iser_conn_t *p_iser_conn;
	int ret_val = 0;

	ITRACE_ENTRY();

	p_iser_conn = p_dto->p_conn;
	if (p_iser_conn == NULL) {
		IPANIC("NULL connection in p_dto: 0x%p\n", p_dto);
	}

	switch (p_dto->type) {
	case ISER_DTO_RCV:
		// Decrement the rcv buffers counter
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		// Release the DTO
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		break;

	case ISER_DTO_SEND:
		// Decrement the sent buffers counter
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		// Release the DTO
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		break;

	case ISER_DTO_RDMA_R:
	case ISER_DTO_RDMA_W:
		// Release the DTO
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		break;

	default:
		IPANIC("Unknown iSER DTO type: %d\n", p_dto->type);
		break;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_dto_completion_error */

/**
 * iser_dto_create_recv_pdu - Creates received PDU descriptor by the received DTO
 */
void iser_dto_create_recv_pdu(struct iser_dto_t *p_dto,	//!< IN,OUT - recv DTO descriptor
			      unsigned long dto_xfer_len)	//!< IN - DTO transfer length
{
	unsigned char *p_recv_buf;

	ITRACE_ENTRY();

	ITRACE(ISER_TRACE_RECV_DTO, "p_dto=0x%p, xfer_len=%lu\n", p_dto,
	       dto_xfer_len);
	TRACE(TRACE_ISER, "ISER_HDR_LEN=%d\n", ISER_HDR_LEN);
	TRACE(TRACE_ISER, "ISER_PDU_BHS_LENGTH=%d\n", ISER_PDU_BHS_LENGTH);

	if (dto_xfer_len < ISER_TOTAL_HEADERS_LEN) {
		IPANIC("Received data size (%ld) less than iSER + "
		       "iSCSI headers (%d)\n", dto_xfer_len,
		       ISER_TOTAL_HEADERS_LEN);
	}
	if (p_dto->regd_vector_len != 1) {
		IPANIC("Received data IOV is (%d) expectd 1\n",
		       p_dto->regd_vector_len);
	}

	ITRACE(ISER_TRACE_RECV_DTO,
	       "Reconsurct pdu header from dto->(buf: 0x%p, size: %d) "
	       "offset: %d, used_size: %d\n",
	       p_dto->regd[0]->virt_buf.p_buf,
	       ISER_TOTAL_HEADERS_LEN,
	       p_dto->offset[0],
	       p_dto->used_sz[0]);

	TRACE(TRACE_ISER, "virt_buf.p_buf=%p, virt_buf.size=%d\n",
	      p_dto->regd[0]->virt_buf.p_buf,
	      p_dto->regd[0]->virt_buf.size);
	TRACE(TRACE_ISER, "ibv_mr->addr=%p, ibv_mr->length=%lu\n",
	      p_dto->regd[0]->mem_reg.ibv_mr->addr,
	      (unsigned long) p_dto->regd[0]->mem_reg.ibv_mr->length);

	/* Get the VA of the headers registered memory */
	p_recv_buf = (unsigned char *) p_dto->regd[0]->virt_buf.p_buf;
	/* Account for the offset within it */
	p_recv_buf += p_dto->offset[0];
	/* Skip the iSER header to get the iSCSI PDU BHS */
	p_dto->as.recv.pdu.p_bhs = (union iser_pdu_bhs_t *) (p_recv_buf + ISER_HDR_LEN);

	// If the BHS, DSL & dto_xfer_len aren't consistent, the iSCSI layer will handle it.
	if (dto_xfer_len > ISER_TOTAL_HEADERS_LEN) {	// we have data

		ITRACE(ISER_TRACE_RECV_DTO,
		       "Reconsurct pdu data from dto->(buf: 0x%p, size: %d) "
		       "offset: %d, used_size: %d\n",
		       p_dto->regd[0]->virt_buf.p_buf,
		       p_dto->regd[0]->virt_buf.size,
		       p_dto->offset[0], p_dto->used_sz[1]);

		p_dto->as.recv.pdu.rx_data.p_buf =
			p_dto->regd[0]->virt_buf.p_buf + ISER_TOTAL_HEADERS_LEN;
		p_dto->as.recv.pdu.rx_data.p_buf += p_dto->offset[0];
		p_dto->as.recv.pdu.rx_data.size =
			dto_xfer_len - ISER_TOTAL_HEADERS_LEN;

		if (p_dto->as.recv.pdu.rx_data.size
		    > ((p_dto->regd[0]->virt_buf.size - ISER_TOTAL_HEADERS_LEN)
		       - p_dto->offset[0])) {
			IPANIC("Virtual buffer, dto: 0x%p, exceeds (%d) the regd buff %p (%d = sz:%d - offset:%d)\n",
			     p_dto, p_dto->as.recv.pdu.rx_data.size,
			     p_dto->regd[0],
			     p_dto->regd[0]->virt_buf.size -
			     p_dto->offset[0],
			     p_dto->regd[0]->virt_buf.size,
			     p_dto->offset[0]);
		}

	} else if (dto_xfer_len == ISER_TOTAL_HEADERS_LEN) {	// there is no data
		p_dto->as.recv.pdu.rx_data.p_buf = NULL;
		p_dto->as.recv.pdu.rx_data.size = 0;
	} else {
		p_dto->as.recv.pdu.rx_data.p_buf = NULL;
		p_dto->as.recv.pdu.rx_data.size = 0;
		TRACE(TRACE_ISER, "Unexpected rx size: %ld\n",
		      dto_xfer_len);
	}
	p_dto->as.recv.pdu.rx_data.type = ISER_BUF_TYPE_SINGLE;

	ITRACE_EXIT();

}				/* iser_dto_create_recv_pdu */

/**
 * iser_dto_send_create - Creates a new send DTO descriptor, adds header regd buffer
 *
 * returns allocated DTO descriptor
 */
struct iser_dto_t *iser_dto_send_create(struct iser_conn_t *p_iser_conn,
					union iser_pdu_bhs_t *p_bhs)
{
	struct iser_regd_buf_t *p_regd_hdr = NULL;
	struct iser_dto_t *p_send_dto = NULL;
	unsigned char *p_iser_header;
	union iser_pdu_bhs_t *p_dbg_bhs;

	ITRACE_ENTRY();

	if (!p_bhs) {
		TRACE_ERROR("%s: called with a NULL p_bhs\n", __func__);
		p_send_dto = NULL;
		goto dto_send_create_exit;
	} else if ((void*) p_bhs < (void*) PAGE_SIZE) { 
		TRACE_ERROR("%s: called with an invalid p_bhs [%p]\n",
		            __func__, p_bhs);
		p_send_dto = NULL;
		goto dto_send_create_exit;
	}

	p_send_dto = iser_alloc_dto(p_iser_conn, ISER_DTO_SEND);
	if (!p_send_dto)
		goto dto_send_create_exit;

	p_regd_hdr =
	    iser_buffer_alloc(ISER_TOTAL_HEADERS_LEN, p_iser_conn,
	                      DMA_TO_DEVICE);

	if (p_regd_hdr == NULL) {
		IERROR("iser_buffer_pool_alloc failed\n");
		iser_conn_remove_dto(p_iser_conn, p_send_dto);
		p_send_dto = NULL;
		goto dto_send_create_exit;
	}

	/* Start initializing iSER Header */
	p_iser_header = (unsigned char *) p_regd_hdr->virt_buf.p_buf;
	if (!p_iser_header) {
		IERROR("p_iser_header is null");
		iser_conn_remove_dto(p_iser_conn, p_send_dto);
		p_send_dto = NULL;
		goto dto_send_create_exit;
	} else if ((void*) p_iser_header < (void*) PAGE_SIZE) {
		IERROR("p_iser_header [%p] is invalid", p_iser_header);
		iser_conn_remove_dto(p_iser_conn, p_send_dto);
		p_send_dto = NULL;
		goto dto_send_create_exit;
	}

	memset(p_iser_header, 0, ISER_HDR_LEN);
	ISER_HDR_SET_VERSION(p_iser_header);
	/* Copy the iSCSI BHS */
	memcpy(p_iser_header + ISER_HDR_LEN, p_bhs, ISER_PDU_BHS_LENGTH);

	p_dbg_bhs =
	    (union iser_pdu_bhs_t *) ((unsigned char *) p_iser_header +
				      ISER_HDR_LEN);
	ITRACE(ISER_TRACE_SEND_DTO, "dbg: dsl=%d %d %d\n",
	       p_dbg_bhs->byte.dlength[0], p_dbg_bhs->byte.dlength[1],
	       p_dbg_bhs->byte.dlength[2]);

	/* we use the headers (iser+iscsi) size from the start of the regd buf's memory */
	iser_dto_add_regd_buff(p_send_dto, p_regd_hdr, USE_NO_OFFSET,
			       USE_SIZE(ISER_TOTAL_HEADERS_LEN));

      dto_send_create_exit:
	ITRACE_EXIT();
	return p_send_dto;

}				/* iser_dto_send_create */

/**
 * \brief Creates an iSER hello message DTO.
 * \param p_conn The connection.
 * \param p_hello The hello message buffer.
 */
struct iser_dto_t *iser_dto_raw_create(struct iser_conn_t *p_conn,
				       unsigned char *p_hello,
				       unsigned int length)
{
	struct iser_dto_t *p_dto;
	struct iser_regd_buf_t *p_regd_hdr;
	unsigned char *p_iser_header;

	ITRACE_ENTRY();

	p_dto = iser_alloc_dto(p_conn, ISER_DTO_SEND);
	if (!p_dto)
		goto error;

	p_regd_hdr = iser_buffer_alloc(length, p_conn, DMA_TO_DEVICE);
	if (!p_regd_hdr) {
		IERROR("iser_buffer_alloc failed\n");
		iser_conn_remove_dto(p_conn, p_dto);
		goto error;
	}

	p_iser_header = (unsigned char *) p_regd_hdr->virt_buf.p_buf;
	memcpy(p_iser_header, p_hello, length);

	iser_dto_add_regd_buff(p_dto, p_regd_hdr, USE_NO_OFFSET,
			       USE_SIZE(length));

	ITRACE_EXIT();
	return p_dto;

      error:
	ITRACE_EXIT();
	return NULL;
}

/**
 * Increment the reference count on this DTO.
 *
 * \param p_dto The dto.
 */
void iser_dto_get(struct iser_dto_t *p_dto)
{
	ITRACE_ENTRY();
	atomic_inc(&p_dto->refs);
	ITRACE_EXIT();
}

/**
 * Decrement the DTO reference count.
 *
 * \param p_dto The DTO.
 */
void iser_dto_put(struct iser_dto_t *p_dto)
{
	/* this does the decrement for us. */
	/* TODO: do the decrement here instead of iser_free_dto */
	iser_dto_free(p_dto);
}

/**
 * Allocates a new data transfer object.
 *
 * \param p_conn The connection that this DTO belongs to
 * \param type The type for this DTO.
 * \return The newly allocated DTO, or NULL if there was an allocation error.
 */
struct iser_dto_t *iser_alloc_dto(struct iser_conn_t *p_conn,
                                  enum iser_dto_type type)
{
	struct iser_dto_t *p_dto;

	ITRACE_ENTRY();

	p_dto = my_kmem_cache_alloc(iser_global.dto_mem_cache,
	                         MEM_ALLOC_FLAG);
	if (!p_dto) {
		IERROR("allocation of hello DTO descriptor failed\n");
		ITRACE_EXIT();
		return NULL;
	}
	TRACE(TRACE_ISER, "%s: allocated DTO=%p\n", __func__, p_dto);

	memset(p_dto, 0, sizeof(*p_dto));
	atomic_set(&p_dto->wrs, 0);
	p_dto->type = type;
	p_dto->notify_enable = 1;
	iser_conn_add_dto(p_conn, p_dto);

	ITRACE_EXIT();
	return p_dto;
}

/**
 * iser_dto_copy_send_data - Allocates a send-data buffer and copies there a PDU's data segment
 *
 * returns 0 if successful
 */
int
iser_dto_copy_send_data(struct iser_dto_t *p_send_dto,
			struct iser_data_buf_t *p_data,
			unsigned int total_size)
{
	struct iser_conn_t *p_iser_conn = p_send_dto->p_conn;
	struct iser_regd_buf_t *p_regd_data;
	unsigned long total_data_sz;
	int ret_val = 0;

	ITRACE_ENTRY();

	/* Create the iSER headers buffer */
	p_regd_data = iser_buffer_alloc(total_size, p_iser_conn, DMA_TO_DEVICE);
	if (p_regd_data == NULL) {
		IERROR("iser_buffer_pool_alloc failed\n");
		ret_val = -1;
		goto dto_send_data_copy_exit;
	}

	/* Copy user data to buffer */
	iser_data_buf_memcpy(&p_regd_data->virt_buf, p_data,
			     &total_data_sz);

	/* we use the headers (iser+iscsi) size from the start of the regd buf's memory */
	iser_dto_add_regd_buff(p_send_dto, p_regd_data, USE_NO_OFFSET,
			       USE_SIZE(total_data_sz));

      dto_send_data_copy_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_dto_copy_send_data */

/**
 * Registers one work completion for a DTO.
 *
 * \param p_dto The DTO.
 */
void iser_dto_one_wc(struct iser_dto_t *p_dto, unsigned int err,
                     unsigned long len)
{
	struct iser_entity_t *p_entity;

	ITRACE_ENTRY();

	if (err)
		p_dto->wr_err = err;

	p_dto->xfer_len += len;

	if (atomic_dec_and_test(&p_dto->wrs)) {
		/* when all work completions are done */
		TRACE(TRACE_ISER, "%s: DTO %p complete\n", __func__, p_dto);

/*
		if (p_dto->type == ISER_DTO_RCV
		    && p_dto->p_conn->hello_state == ISER_HELLO_DONE
		    && p_dto->p_conn->max_sge_recv < 2) {
			iser_expand_recv_sges(p_dto);
		}
*/

		if (p_dto->wr_err) {
			/* if any WC was an error */
			TRACE(TRACE_ISER, "%s: DTO errored\n", __func__);
			err = iser_dto_completion_error(p_dto);
			if (err) {
				TRACE_ERROR("%s: iser_dto_completion_error "
					    "failed", __func__);
			}
		} else {
			TRACE(TRACE_ISER, "%s: DTO successful\n", __func__);
			p_entity = p_dto->p_conn->p_entity;
			/* no WCs were in error */

			TRACE(TRACE_ISER, "Calling=%p\n", p_entity->dto_completion_cb);
			err = p_entity->dto_completion_cb(p_dto, p_dto->xfer_len);
			if (err) {
				TRACE_ERROR("Failed to handle successful DTO "
					    "completion\n");
			}
		}
	} else {
		TRACE(TRACE_ISER, "%s: DTO %p still more to go\n", __func__, p_dto);
	}

	ITRACE_EXIT();
}

/**
 * There is an issue with SEND & RECV DTOs: we can not split up sends
 * over multiple WRs, since that is not how it works.
 *
 * If we run into the situation where we would need to split up a SEND, we
 * will just allocate another buffer and merge it.
 *
 * \param dto The DTO to merge the regd buffers of into a newly allocated
 *            sendbuf.
 *
 * \note This routine allocates a sendbuf for the DTO, which will need to be
 *       free()d later.  (See iser_dto_one_wc() in datamover/iser_dto.c).
 *
 * \return 0 on success, !0 on error.
 */
int iser_merge_sges(struct iser_dto_t *dto)
{
	int i, release_deferred;
	unsigned long length, offs;
	struct iser_regd_buf_t *p_regd_hdr = NULL;

	TRACE(TRACE_ISER, "%s: dto=%p, merging SGEs\n",
	      __func__, dto);

	ITRACE_ENTRY();

	length = 0;
	for (i = 0; i < dto->regd_vector_len; i += 1)
		length += dto->regd[i]->mem_reg.ibv_mr->length;

	TRACE(TRACE_ISER, "%s: merge_total_len=%lu\n", __func__, length);

	p_regd_hdr = iser_buffer_alloc(length, dto->p_conn, DMA_TO_DEVICE);
	if (!p_regd_hdr) {
		IERROR("iser_buffer_pool_alloc failed\n");
		ITRACE_EXIT();
		return -1;
	}

	offs = 0;
	for (i = 0; i < dto->regd_vector_len; i += 1) {
		memcpy(p_regd_hdr->virt_buf.p_buf + offs,
		       dto->regd[i]->mem_reg.ibv_mr->addr,
		       dto->regd[i]->mem_reg.ibv_mr->length);

		offs += dto->regd[i]->mem_reg.ibv_mr->length;
	}

	for (i = 0; i < dto->regd_vector_len; i += 1) {
		if (iser_regd_buff_release(dto->regd[i],
		                           &release_deferred) < 0) {

			TRACE_ERROR("%s: error while releasing regd buf=%p\n",
			            __func__, dto->regd[i]);
			ITRACE_EXIT();
			return -1;
		}

		dto->regd[i] = NULL;
	}
	dto->regd_vector_len = 0;

	iser_dto_add_regd_buff(dto, p_regd_hdr, USE_NO_OFFSET,
	                       USE_SIZE(length));

	ITRACE_EXIT();
	return 0;
}

/* vi: set tabstop=8 textwidth=78: */
