/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_global.h"
#include "iser_conn.h"
#include "iser_initiator.h"
#include "iser_target.h"
#include "iser_utils.h"
#include "iser_ofa.h"
#include "iser_memory.h"
#include "iser_trace.h"

// Globals
struct iser_global_t iser_global;

#if defined(CONFIG_DEBUG_FS)

/* eventually debugfs_unh_iscsi should be made global for the entire
 * project and should be non-null */
#define debugfs_unh_iscsi NULL

struct dentry *debugfs_datamover;
#endif /* CONFIG_DEBUG_FS */

// Functions

/**
 * iser_adaptor_init - Initializes iSER adaptor structure.
 *
 *
 * Creates adaptor-scope objects (Interface Adaptor, Protection Zone,
 * Public Service Points).
 *
 * returns 0 on success, -1 on failure
 */
int iser_adaptor_init(struct iser_adaptor_t *p_iser_adaptor,	//!< IN - adaptor descriptor
		      char *name)	//!< IN - IA name
{
//     mem_cache_t* header_mem_cache_p=NULL;
//     unsigned int cache_entry_size=0;

	if (p_iser_adaptor == NULL)
		return -1;

	memset(p_iser_adaptor, 0, sizeof(struct iser_adaptor_t));

	// initialize in_callback counter.
	atomic_set(&p_iser_adaptor->in_callback, 0);

	// Initialize waiting queue for the event handler thread
	init_waitqueue_head(&p_iser_adaptor->dat_events_wait_q);
	init_waitqueue_head(&p_iser_adaptor->connect_wait_q);

	// Initialize list of connections
	spin_lock_init(&p_iser_adaptor->conn_lock);
	INIT_LIST_HEAD(&p_iser_adaptor->conn_list);

	strncpy(p_iser_adaptor->name, name, 64);

	if (iser_ofa_init_adaptor(p_iser_adaptor) != 0) {
		return -1;
	}

	/* Start the event thread */
	if (iser_spawn_event_thread(p_iser_adaptor) != 0) {
		IERROR("Failed to start event kernel thread\n");
		iser_adaptor_release(p_iser_adaptor);
		return -1;
	}

	return 0;
}				/* iser_adaptor_init */

/**
 * iser_adaptor_release - Releases all adaptor-related resources.
 *
 * returns 0 on success, -1 on failure
 */
int iser_adaptor_release(struct iser_adaptor_t *p_iser_adaptor)	//!< IN - adaptor descriptor
{
	IINFO("releasing adaptor\n");

	/* these should alredy have been removed!!!! -- eaburns */
	if (!list_empty(&p_iser_adaptor->conn_list))
		return -1;
	if (iser_free_psp_by_adaptor(p_iser_adaptor) != 0)
		return -1;
	if (iser_ofa_deinit_adaptor(p_iser_adaptor) != 0)
		return -1;

	iser_kill_event_thread(p_iser_adaptor);

	return 0;
}				/* iser_adaptor_release */

/**
 * iser_adaptor_add_connection - Adds a connection to adaptor
 */
void iser_adaptor_add_connection(struct iser_adaptor_t *p_iser_adaptor,	//!< IN - adaptor descriptor
				 struct iser_conn_t *p_iser_conn)	//!< IN - connection to add
{
	unsigned long flags;

	iser_conn_get(p_iser_conn);
	p_iser_conn->p_adaptor = p_iser_adaptor;
	spin_lock_irqsave(&p_iser_adaptor->conn_lock, flags);
	list_add(&p_iser_conn->adaptor_list, &p_iser_adaptor->conn_list);
	spin_unlock_irqrestore(&p_iser_adaptor->conn_lock, flags);
}				/* iser_adaptor_add_connection */

/**
 * iser_adaptor_find_connection - Adds a connection to adaptor
 *
 * returns 0 on success, -1 on failure
 */
struct iser_conn_t *iser_adaptor_find_connection(struct iser_adaptor_t *p_iser_adaptor,	//!< IN - adaptor descriptor
						 struct rdma_cm_id *cm_id)	//!< IN - EP handle to find
{

	struct iser_conn_t *p_iser_conn = NULL;
	struct list_head *p_list;
	unsigned long flags;

	ITRACE_ENTRY();

	spin_lock_irqsave(&p_iser_adaptor->conn_lock, flags);
	p_list = p_iser_adaptor->conn_list.next;
	while (p_list != &p_iser_adaptor->conn_list) {
		p_iser_conn =
		    list_entry(p_list, struct iser_conn_t, adaptor_list);
		if (p_iser_conn->cm_id == cm_id)
			break;
		p_iser_conn = NULL;
		p_list = p_list->next;
	}
	spin_unlock_irqrestore(&p_iser_adaptor->conn_lock, flags);

	ITRACE_EXIT();
	return p_iser_conn;
}				/* iser_adaptor_find_connection */

/**
 * iser_adaptor_remove_connection - Removes a connection from adaptor
 */
void iser_adaptor_remove_connection(struct iser_conn_t *p_iser_conn)	//!< IN - connection to add
{
	struct iser_adaptor_t *p_iser_adaptor;
	unsigned long flags;

	p_iser_adaptor = p_iser_conn->p_adaptor;
	if (p_iser_adaptor == NULL) {
		IPANIC("NULL adaptor in conn: 0x%p\n",
		       p_iser_conn);
	}

	spin_lock_irqsave(&p_iser_adaptor->conn_lock, flags);
	if (p_iser_conn->adaptor_list.next) {
		list_del(&p_iser_conn->adaptor_list);
	}
	spin_unlock_irqrestore(&p_iser_adaptor->conn_lock, flags);

	iser_conn_put(p_iser_conn);
}				/* iser_adaptor_remove_connection */


/**
 * iser_entity_add_connection - Adds a connection to entity
 */
void iser_entity_add_connection(struct iser_entity_t *p_entity,	//!< IN - entity descriptor
				struct iser_conn_t *p_iser_conn)	//!< IN - connection to add
{
	unsigned long flags;

	iser_conn_get(p_iser_conn);
	p_iser_conn->p_entity = p_entity;
	spin_lock_irqsave(&p_entity->conn_lock, flags);
	list_add(&p_iser_conn->entity_list, &p_entity->conn_list);
	spin_unlock_irqrestore(&p_entity->conn_lock, flags);

}				/* iser_entity_add_connection */

/**
 * iser_entity_remove_connection - Removes a connection from adaptor
 */
void iser_entity_remove_connection(struct iser_conn_t *p_iser_conn)	//!< IN - connection to add
{
	unsigned long flags;
	struct iser_entity_t *p_entity;

	p_entity = p_iser_conn->p_entity;
	if (p_entity == NULL) {
		IPANIC("NULL entity in conn: 0x%p\n", p_iser_conn);
	}

	spin_lock_irqsave(&p_entity->conn_lock, flags);
	if (p_iser_conn->entity_list.next) {
		list_del(&p_iser_conn->entity_list);
	}
	spin_unlock_irqrestore(&p_entity->conn_lock, flags);
	iser_conn_put(p_iser_conn);
}				/* iser_entity_remove_connection */

/**
 * iser_global_init - Initializes the global iSER context structure.
 *
 * returns 0 on success, -1 on failure
 */
int iser_global_init()
{
	ITRACE_ENTRY();

	memset(&iser_global, 0, sizeof(struct iser_global_t));
#if defined(CONFIG_DEBUG_FS)
	debugfs_datamover = debugfs_create_dir("datamover",
					       debugfs_unh_iscsi);
	if (IS_ERR(debugfs_datamover)) {
		TRACE_WARNING("Unable to create datamover debugfs entry\n");
		debugfs_datamover = NULL;
	} else {
		iser_global.debugfs_num_adaptors =
			debugfs_create_u32("num_adaptors",
					   ISER_DEBUGFS_PERMS,
					   debugfs_datamover,
					   &iser_global.num_adaptors);
		if (IS_ERR(iser_global.debugfs_num_adaptors)) {
			TRACE_WARNING("Unable to create "
				      "num_adaptors entry\n");
			iser_global.debugfs_num_adaptors = NULL;
		}
		iser_global.debugfs_num_entities =
			debugfs_create_u32("num_entities",
					   ISER_DEBUGFS_PERMS,
					   debugfs_datamover,
					   &iser_global.num_entities);
		if (IS_ERR(iser_global.debugfs_num_entities)) {
			TRACE_WARNING("Unable to create "
				      "num_entities entry\n");
			iser_global.debugfs_num_entities = NULL;
		}
	}
#endif /* CONFIG_DEBUG_FS */

	iser_global.regd_buf_mem_cache = my_kmem_cache_create(
					"unh_regd_buf_mem_cache",
					sizeof(struct iser_regd_buf_t),
					0,
					SLAB_HWCACHE_ALIGN);

	if (!iser_global.regd_buf_mem_cache)
		IPANIC("iser_global_init: kmem_cache_create failed\n");

	iser_global.task_mem_cache = my_kmem_cache_create(
					"unh_task_mem_cache",
					sizeof(struct iser_task_t),
					0,
					SLAB_HWCACHE_ALIGN);
	if (!iser_global.task_mem_cache)
		IPANIC("iser_global_init: kmem_cache_create failed\n");

	iser_global.dto_mem_cache = my_kmem_cache_create(
					"unh_dto_mem_cache",
					sizeof(struct iser_dto_t),
					0,
					SLAB_HWCACHE_ALIGN);
	if (!iser_global.dto_mem_cache)
		IPANIC("iser_global_init: kmem_cache_create failed\n");

	// Allocate adaptors; currently single adaptor
	iser_global.num_adaptors = 1;
	if (iser_adaptor_init(&iser_global.adaptor[0], "Adaptor0") != 0) {
		IERROR("initializing iser failed!\n");
		iser_adaptor_release(&iser_global.adaptor[0]);
		return -1;
	}
	// Start with no entities defined
	iser_global.num_entities = 0;

	// Initialize task hash table
	hash_init(&iser_global.task_hash);
	// Initialize connections hash table
	hash_init(&iser_global.conn_hash);

#ifdef KDAPL_LOCK
	spin_lock_init(&iser_global.kdapl_lock);
#endif

	ITRACE_EXIT();
	return 0;
}				/* iser_global_init */

/**
 * iser_global_release - Releases all resources through the global iSER context structure.
 *
 * returns 0 on success, -1 on failure
 */
int iser_global_release()
{
	int err = 0;
	ITRACE_ENTRY();

	// Are all entities released?
	if (iser_global.num_entities > 0) {
		IERROR
		    ("Some entities did not unregister. Can't release module.\n");
		return -1;
	}

	// Release all adaptors
	err = iser_adaptor_release(&iser_global.adaptor[0]);
	if (err) {
		TRACE_ERROR("%s: failed to release adaptor 0\n",
			    __func__);
		goto out;
	}
	iser_global.num_adaptors = 0;

	if (iser_global.regd_buf_mem_cache)
		my_kmem_cache_destroy(iser_global.regd_buf_mem_cache);
	if (iser_global.task_mem_cache)
		my_kmem_cache_destroy(iser_global.task_mem_cache);
	if (iser_global.dto_mem_cache)
		my_kmem_cache_destroy(iser_global.dto_mem_cache);

#if defined(CONFIG_DEBUG_FS)
	if (iser_global.debugfs_num_adaptors)
		debugfs_remove(iser_global.debugfs_num_adaptors);
	if (iser_global.debugfs_num_entities)
		debugfs_remove(iser_global.debugfs_num_entities);
	if (debugfs_datamover)
		debugfs_remove(debugfs_datamover);
#endif /* CONFIG_DEBUG_FS */


out:
	ITRACE_EXIT();
	return err;
}				/* iser_global_release */

/**
 * iser_find_regd_entity - Look for an iSCSI entity using its type and provider name as the key.
 *
 * returns entity descriptor if found, NULL - if not found
 */
struct iser_entity_t *iser_find_regd_entity(enum iser_entity_type entity_type,	//!< IN - initiator or target
					    char *provider_name)	//!< IN - provider of the iSCSI layer entity
{
	struct iser_entity_t *p_entity = NULL;
	int i;

	for (i = 0; i < ISER_MAX_ENTITIES; i++) {
		if (iser_global.entity[i].registered &&
		    strncmp(provider_name,
			    iser_global.entity[i].provider_name, 64) == 0
		    && entity_type == iser_global.entity[i].type) {
			p_entity = &iser_global.entity[i];
			break;
		}
	}

	return p_entity;
}				/* iser_find_regd_entity */

/**
 * iser_find_free_entity - Look for a free entity descriptor within the global context struct.
 *
 * returns index into the entities array if found, -1 if not found
 */

static int iser_find_free_entity(void)
{
	int i;

	for (i = 0; i < ISER_MAX_ENTITIES; i++) {
		if (!iser_global.entity[i].registered)
			return i;
	}
	/* no free entities */
	return -1;
}				/* iser_find_free_entity */

/**
 * iser_api_register_target - register a target with iser extended api
 *
 *  Create a PSP for the given listening port, register the extended
 * callbacks, supplied by the caller (iSCSI layer) and retrieve the
 * extended iSER API functions. Functions set to NULL are expected
 * to be ignored by both sides.
 *
 * returns iSER status
 */
iser_status_t iser_api_register_target(char *provider_name,	//!< IN - provider of the iSCSI layer entity
				       struct iser_api_t * api,	//!< OUT - iSER API functions
				       struct iser_api_cb_t * api_cb,	//!< IN - extended iSER callbacks, supplied by iSCSI layer
				       struct iser_ext_api_t * ext_api,	//!< OUT - extended iSER API functions
				       struct iser_ext_api_cb_t * ext_api_cb,	//!< IN - extended iSER callbacks, supplied by iSCSI layer
				       struct sockaddr * addr, size_t addr_len, iser_api_handle_t * p_api_h)	//!< OUT - extended API registration handle
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_entity_t *p_entity;
	int entity_num;

	ITRACE_ENTRY();

	if (p_api_h == NULL) {
		IERROR("NULL *p_api_h \n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	*p_api_h = ISER_INVALID_API_H;

	if (provider_name == NULL) {
		IERROR("NULL *provider_name\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	if (api == NULL) {
		IERROR("NULL *api structure\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	if (api_cb == NULL) {
		IERROR("NULL *api_cb structure\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}

	/* Check if this entity is already registered */
	p_entity =
	    iser_find_regd_entity(ISER_ENTITY_TARGET, provider_name);
	if (p_entity != NULL) {
		IERROR("This entity is already registered: %s-%s\n",
		       provider_name,
		       iser_get_entity_type_name(ISER_ENTITY_TARGET));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}

	/* New entity - use the next free descriptor */
	entity_num = iser_find_free_entity();
	if (entity_num < 0) {
		IERROR("Failed to find a free entity entry for: %s-%s\n",
		       provider_name,
		       iser_get_entity_type_name(ISER_ENTITY_TARGET));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}
	p_entity = &iser_global.entity[entity_num];

	memset(p_entity, 0, sizeof(struct iser_entity_t));

	/* Copy iSER API functions to iSCSI structure */
	if (iser_target_register_entity
	    (p_entity, api, ext_api, ext_api_cb, addr, addr_len) != 0) {
		IERROR
		    ("Failed to set TARGET API functions for entity: %s-%s\n",
		     provider_name,
		     iser_get_entity_type_name(ISER_ENTITY_TARGET));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}

	/* Copy iSCSI API callback functions from iSCSI structure */
	memcpy(&p_entity->api_cb, api_cb, sizeof(struct iser_api_cb_t));
	/*
	   p_entity->api_cb.conn_terminate_notify = api_cb->conn_terminate_notify;
	   p_entity->api_cb.command_notify = api_cb->command_notify;
	   p_entity->api_cb.data_out_notify = api_cb->data_out_notify;
	   p_entity->api_cb.control_notify = api_cb->control_notify;
	   p_entity->api_cb.data_ack_notify = api_cb->data_ack_notify;
	   p_entity->api_cb.data_completion_notify = api_cb->data_completion_notify;
	 */
	p_entity->registered = 1;
	*p_api_h = (iser_api_handle_t) p_entity;
	iser_global.num_entities++;
	strncpy(p_entity->provider_name, provider_name, 64);
	p_entity->type = ISER_ENTITY_TARGET;
	spin_lock_init(&p_entity->conn_lock);
	INIT_LIST_HEAD(&p_entity->conn_list);

      api_register_exit:

	ITRACE_EXIT();

	return iser_ret;
}				/* iser_api_register */

/**
 * iser_api_register_initiator - register a initiator with iser extended api
 *
 *  Create a PSP for the given listening port, register the extended
 * callbacks, supplied by the caller (iSCSI layer) and retrieve the
 * extended iSER API functions. Functions set to NULL are expected
 * to be ignored by both sides.
 *
 * returns iSER status
 */
iser_status_t
iser_api_register_initiator(char *provider_name,
			    struct iser_api_t * api,
			    struct iser_api_cb_t * api_cb,
			    struct iser_ext_api_t * ext_api,
			    struct iser_ext_api_cb_t * ext_api_cb,
			    iser_api_handle_t * p_api_h)
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_entity_t *p_entity;
	int entity_num;

	ITRACE_ENTRY();

	if (p_api_h == NULL) {
		IERROR("NULL *p_api_h \n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	*p_api_h = ISER_INVALID_API_H;

	if (provider_name == NULL) {
		IERROR("NULL *provider_name\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	if (api == NULL) {
		IERROR("NULL *api structure\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}
	if (api_cb == NULL) {
		IERROR("NULL *api_cb structure\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_register_exit;
	}

	/* Check if this entity is already registered */
	p_entity =
	    iser_find_regd_entity(ISER_ENTITY_INITIATOR, provider_name);
	if (p_entity != NULL) {
		IERROR("This entity is already registered: %s-%s\n",
		       provider_name,
		       iser_get_entity_type_name(ISER_ENTITY_INITIATOR));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}

	/* New entity - use the next free descriptor */
	entity_num = iser_find_free_entity();
	if (entity_num < 0) {
		IERROR("Failed to find a free entity entry for: %s-%s\n",
		       provider_name,
		       iser_get_entity_type_name(ISER_ENTITY_INITIATOR));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}
	p_entity = &iser_global.entity[entity_num];

	memset(p_entity, 0, sizeof(struct iser_entity_t));

	/* Copy iSER API functions to iSCSI structure */
	if (iser_initiator_register_entity
	    (p_entity, api, ext_api, ext_api_cb) != 0) {
		IERROR
		    ("Failed to set INITIATOR API functions for entity: %s-%s\n",
		     provider_name,
		     iser_get_entity_type_name(ISER_ENTITY_INITIATOR));
		iser_ret = ISER_FAILURE;
		goto api_register_exit;
	}

	/* Copy iSCSI API callback functions from iSCSI structure */
	memcpy(&p_entity->api_cb, api_cb, sizeof(struct iser_api_cb_t));
	/*
	   p_entity->api_cb.conn_terminate_notify = api_cb->conn_terminate_notify;
	   p_entity->api_cb.command_notify = api_cb->command_notify;
	   p_entity->api_cb.data_out_notify = api_cb->data_out_notify;
	   p_entity->api_cb.control_notify = api_cb->control_notify;
	   p_entity->api_cb.data_ack_notify = api_cb->data_ack_notify;
	   p_entity->api_cb.data_completion_notify = api_cb->data_completion_notify;
	 */
	p_entity->registered = 1;
	*p_api_h = (iser_api_handle_t) p_entity;
	iser_global.num_entities++;
	strncpy(p_entity->provider_name, provider_name, 64);
	p_entity->type = ISER_ENTITY_INITIATOR;
	spin_lock_init(&p_entity->conn_lock);
	INIT_LIST_HEAD(&p_entity->conn_list);

      api_register_exit:

	ITRACE_EXIT();

	return iser_ret;
}				/* iser_api_register */

#if 0
/*
 * This function is currently unused.
 */
/**
 * iser_ext_set_port - Set new listening port number.
 *
 * returns iSER status
 */
iser_status_t iser_ext_set_port(iser_api_handle_t api_h,	//!< IN - extended API registration handle
				unsigned int port)	//!< IN - listening port
{
	struct iser_entity_t *p_entity;
	unsigned long entity_index = (unsigned long) api_h;

	if (entity_index >= iser_global.num_entities) {
		return ISER_ILLEGAL_PARAM;
	}
	p_entity = &iser_global.entity[entity_index];

	/* Create PSP for the requested port */
	if (iser_create_psp(p_entity, port) == 0)
		return ISER_SUCCESS;
	else
		return ISER_FAILURE;
}				/* iser_ext_set_port */
#endif				/* 0 */

/**
 * iser_api_unregister - Unregister API entity
 *
 * returns iSER status
 */
iser_status_t iser_api_unregister(iser_api_handle_t api_h)	//!< IN - extended API registration handle
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_entity_t *p_entity;

	ITRACE_ENTRY();

	if (api_h == NULL || api_h == ISER_INVALID_API_H) {
		IERROR("Invalid API registration handle");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_unregister_exit;
	}
	p_entity = (struct iser_entity_t *) api_h;
	if (!p_entity->registered) {
		IERROR("Trying to deregister unregistered entity\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto api_unregister_exit;
	}
	if(!list_empty(&p_entity->conn_list)) {
		TRACE_ERROR("%s: connections still exist\n", __func__);
		iser_ret = ISER_FAILURE;
		goto api_unregister_exit;
	}

	if (iser_free_psp_by_entity(p_entity) != 0) {
		IERROR("Failed to free PSP for entity\n");
		iser_ret = ISER_FAILURE;
		goto api_unregister_exit;
	}
	p_entity->registered = 0;
	if (iser_global.num_entities > 0) {
		iser_global.num_entities--;
	} else {
		IPANIC
		    ("Zero registered entities when trying to unregister one\n");
	}

      api_unregister_exit:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_api_unregister */

static char *iser_entity_type_name[iSER_ENTITY_TYPES_NUM + 1] = {
	"Initiator",
	"Target",
	"Illegal type"
};

/**
 * iser_get_entity_type_name - Get iSER entity type string name
 *
 * returns entity type string for valid entity types, or "Illegal type"
 */
char *iser_get_entity_type_name(enum iser_entity_type type)
{
	if (type < iSER_ENTITY_TYPES_NUM) {
		return iser_entity_type_name[type];
	} else {
		return iser_entity_type_name[iSER_ENTITY_TYPES_NUM];
	}
}				/* iser_get_entity_type_name */

#if defined(__KERNEL__)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
EXPORT_SYMBOL(iser_api_unregister);
EXPORT_SYMBOL(iser_api_register_target);
EXPORT_SYMBOL(iser_api_register_initiator);
#endif
#endif				/* __KERNEL__ */
