/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#ifndef __ISER_HEADER_H__
#define __ISER_HEADER_H__

/* iSER opcode, first nibble of byte zero. */
#define ISER_OPCODE(p) ((*p & 0xF0) >> 4)

/*
 * iSER header opcodes.
 */
enum iser_opcode {
	ISER_OP_NONE = 0,
	ISER_OP_ISCSI = 1,
	ISER_OP_HELLO = 2,
	ISER_OP_HELLO_REPLY = 3
};

/*
 BYTE |0                   1                   2                   3
      |0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
DWORD |       |W|R|                                                   |
    0 | 0001b |S|S|               Reserved                            |
      |       |V|V|                                                   |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    1 |             Write STag (or N/A) = RKey                        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    2 |             Write TO (or N/A) = Virt.Addr.                    |
    3 |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    4 |             Read  STag (or N/A) = RKey                        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    5 |             Read  TO (or N/A) = Virt.Addr.                    |
    6 |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

#define ISER_HDR_LEN 28

/* Version */
#define ISER_HDR_VER_OFFSET     0	/* in byte 0 */
#define ISER_HDR_VER_SHIFT      4	/* last bit 3 */
#define ISER_HDR_VER_MASK       0x0F	/* 4 bits */

/* WSV - Write STag Valid flag */
#define ISER_HDR_WSV_OFFSET     0	/* in byte 0 */
#define ISER_HDR_WSV_SHIFT      3	/* last bit 4 */
#define ISER_HDR_WSV_MASK       0x01	/* 1 bit */

/* RSV - Read STag Valid flag */
#define ISER_HDR_RSV_OFFSET     0	/* in byte 0 */
#define ISER_HDR_RSV_SHIFT      2	/* last bit 5 */
#define ISER_HDR_RSV_MASK       0x01	/* 1 bit */

/* Retrieve a bit field from a header byte array.
   Possible field names are: VER, WSV, RSV
*/
#define ISER_HDR_GET_BITS(p,field) \
    ((((unsigned char *)p)[ISER_HDR_ ## field ## _OFFSET] >> \
    ISER_HDR_ ## field ## _SHIFT) & \
    ISER_HDR_ ## field ## _MASK)

/* Set the passed value in the bit field
   Possible field names are: VER, WSV, RSV
*/
#define ISER_HDR_SET_BITS(p,field,val) \
   do { \
        ((unsigned char *)p)[ISER_HDR_ ## field ## _OFFSET] &= \
           ~(ISER_HDR_ ## field ## _MASK << ISER_HDR_ ## field ## _SHIFT); \
        ((unsigned char *)p)[ISER_HDR_ ## field ## _OFFSET] |= \
           ((val) & ISER_HDR_ ## field ## _MASK) << ISER_HDR_ ## field ## _SHIFT; \
   } while(0)

/**
 * \brief The current iSER version.
 */
#define ISER_CURRENT_VERSION 0x01

#define ISER_HDR_SET_VERSION(p) ISER_HDR_SET_BITS(p,VER,ISER_CURRENT_VERSION)

/* Access to the fields Read S-Tag, Write S-Tag by 32-bit halves.
   Returns l-value.
*/
#define ISER_HDR_W_RKEY(p) \
	(*(uint32_t *) (((unsigned char *) p) + 4))
#define ISER_HDR_W_VADDR(p) \
	(*(uint64_t *) (((unsigned char *) p) + 8))
#define ISER_HDR_R_RKEY(p) \
	(*(uint32_t *) (((unsigned char *) p) + 16))
#define ISER_HDR_R_VADDR(p) \
	(*(uint64_t *) (((unsigned char *) p) + 20))

/*
 BYTE |0                   1                   2                   3
      |0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
DWORD |
    0 | 0010b | Rsvd  | MaxVer| MinVer|           iSER-IRD            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    1 |                           Reserved                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    2 |                           Reserved                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

#define ISER_HELLO_LEN 12

/* MaxVer */
#define ISER_HDR_HELLO_MAXVER_OFFSET 1	/* in byte 1 */
#define ISER_HDR_HELLO_MAXVER_SHIFT 4	/* last 4 bites */
#define ISER_HDR_HELLO_MAXVER_MASK 0xF	/* 4 bits long */

/* MinVer */
#define ISER_HDR_HELLO_MINVER_OFFSET 1	/* in byte 1 */
#define ISER_HDR_HELLO_MINVER_SHIFT 0	/* last 4 bites */
#define ISER_HDR_HELLO_MINVER_MASK 0xF	/* 4 bits long */

/* Access to iSER-IRD, returns an l-value. */
#define ISER_HDR_HELLO_IRD(p) \
    (*(uint16_t *) (((unsigned char *) p) + 2))

/*
 BYTE |0                   1                   2                   3
      |0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
DWORD |       |     |R|       |       |                               |
    0 | 0011b |Rsvd |E| MaxVer| CurVer|           iSER-ORD            |
      |       |     |J|       |       |                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    1 |                           Reserved                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    2 |                           Reserved                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

#define ISER_HELLO_REPLY_LEN 12

/* REJ */
#define ISER_HDR_HELLO_REPLY_REJ_OFFSET 0	/* in byte 0 */
#define ISER_HDR_HELLO_REPLY_REJ_SHIFT 0	/* last byte */
#define ISER_HDR_HELLO_REPLY_REJ_MASK 1	/* 1 bit long */

/* MaxVer (same as in Hello Message header) */
#define ISER_HDR_HELLO_REPLY_MAXVER_OFFSET ISER_HDR_HELLO_MAXVER_OFFSET
#define ISER_HDR_HELLO_REPLY_MAXVER_SHIFT ISER_HDR_HELLO_MAXVER_SHIFT
#define ISER_HDR_HELLO_REPLY_MAXVER_MASK ISER_HDR_HELLO_MAXVER_MASK

/* MinVer (same as in Hello Message header) */
#define ISER_HDR_HELLO_REPLY_MINVER_OFFSET ISER_HDR_HELLO_MINVER_OFFSET
#define ISER_HDR_HELLO_REPLY_MINVER_SHIFT ISER_HDR_HELLO_MINVER_SHIFT
#define ISER_HDR_HELLO_REPLY_MINVER_MASK ISER_HDR_HELLO_MINVER_MASK

/* Access to iSER-ORD, returns an l-value. */
#define ISER_HDR_HELLO_REPLY_ORD(p) \
    (*(uint16_t *) (((unsigned char *) p) + 2))

#endif				/* __ISER_HEADER_H__ */
