/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/io.h>
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_conn.h"
#include "iser_task.h"
#include "iser_dto.h"
#include "iser_ofa.h"
#include "iser_initiator.h"
#include "iser_memory.h"
#include "iser_utils.h"
#include "iser_platform.h"
#include "iser_trace.h"
#include "iser_global.h"

iser_status_t
iser_initiator_connection_establish(iser_api_handle_t api_h,
				    iser_conn_handle_t iscsi_conn_h,
				    struct sockaddr_in *dst_addr,
				    struct sockaddr_in *src_addr);
iser_status_t
iser_initiator_allocate_connection_resources(iser_conn_handle_t conn_h,
					     struct iser_conn_resources_t
					     *conn_resources, int *);

iser_status_t
iser_initiator_enable_datamover(iser_conn_handle_t iser_conn_h,
				iser_transport_conn_t tport_conn,
				struct iser_send_pdu_t *p_login_resp_pdu);
iser_status_t
iser_initiator_connection_accept(iser_conn_request_t conn_req_h,
				 iser_conn_handle_t iscsi_conn_h,
				 int accept);
iser_status_t
iser_initiator_put_data(iser_conn_handle_t conn_h,
			struct iser_send_pdu_t *p_data_in_pdu,
			int notify_enable);
iser_status_t
iser_initiator_get_data(iser_conn_handle_t conn_h,
			struct iser_send_pdu_t *p_r2t_pdu,
			int notify_enable);
iser_status_t
iser_initiator_send_control(iser_conn_handle_t iscsi_conn_h,
			    struct iser_send_pdu_t *p_ctrl_pdu);

int
iser_initiator_dto_completion(struct iser_dto_t *p_dto,
			      unsigned long dto_xfer_len);

int iser_initiator_alloc_ffp_recv_buffs(struct iser_conn_t *p_conn);

int iser_initiator_enter_hello_exchg(struct iser_conn_t *p_conn);

static int iser_initiator_recv_helloreply(struct iser_conn_t *p_conn,
					  struct iser_dto_t *p_dto);

/**
 * iser_initiator_register_entity - Sets the API functions specific to an initiator entity.
 *
 */
int
iser_initiator_register_entity(struct iser_entity_t *p_entity,
			       struct iser_api_t *api,
			       struct iser_ext_api_t *ext_api,
			       struct iser_ext_api_cb_t *ext_api_cb)
{
	ITRACE_ENTRY();

	api->allocate_connection_resources =
	    iser_initiator_allocate_connection_resources;
	api->notice_key_values = iser_notice_key_values;
	api->release_control = iser_release_control;
	api->enable_datamover = iser_initiator_enable_datamover;
	api->connection_terminate = iser_connection_terminate;
	api->deallocate_connection_resources =
	    iser_deallocate_connection_resources;

	api->send_control = iser_initiator_send_control;
	api->get_data = iser_initiator_get_data;
	api->put_data = iser_initiator_put_data;

	/* copy iSER API functions to iSCSI-supplied structure */
	if (ext_api != NULL) {
		ext_api->connection_establish =
		    iser_initiator_connection_establish;
		ext_api->connection_accept = iser_initiator_connection_accept;	/* stub */
	}

	/* copy extended API iSCSI callback functions to iSER */
	if (ext_api_cb != NULL) {
		p_entity->ext_api_cb.connection_establish_notify =
		    ext_api_cb->connection_establish_notify;
		p_entity->ext_api_cb.connection_request_notify =
		    ext_api_cb->connection_request_notify;
	}

	/* Set the DTO completion handler */
	p_entity->dto_completion_cb = iser_initiator_dto_completion;
	p_entity->enter_hello_exchg = iser_initiator_enter_hello_exchg;

	ITRACE_EXIT();
	return 0;
}				/* iser_initiator_register_entity */

/**
 * iser_connection_establish -	Initiates connection establishment process.
 *
 *  returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_ILLEGAL_PARAM)
 */
iser_status_t iser_initiator_connection_establish(iser_api_handle_t api_h,	//!< IN - extended api handle
						  iser_conn_handle_t iscsi_conn_h,	//!< IN - iSCSI supplied connection handle
						  struct sockaddr_in * dst_addr,	//!< IN - destination (target) IP address + port
						  struct sockaddr_in * src_addr)	//!< IN - source (initiator) IP address + port
{
	struct iser_entity_t *p_entity;
	struct iser_conn_t *p_iser_conn;
	struct iser_adaptor_t *p_adaptor;
	char name[ISER_OBJECT_NAME_SIZE];
#if 0
	int ep_create_ret;
#endif

	ITRACE_ENTRY();

	if (api_h == NULL || api_h == ISER_INVALID_API_H) {
		IERROR("API Handle is illegal: 0x%p\n", api_h);
		return ISER_ILLEGAL_PARAM;
	}
	p_entity = (struct iser_entity_t *) api_h;

	if (dst_addr != NULL) {
		snprintf(name, ISER_OBJECT_NAME_SIZE, "%d.%d.%d.%d",
			NIPQUAD(dst_addr->sin_addr));
	} else {
		snprintf(name, ISER_OBJECT_NAME_SIZE, "Unknown");
	}
	ITRACE(ISER_TRACE_CONN, "Connecting to: %s\n", name);


	/* Allocate and initialize iSER connection structure */
	p_iser_conn = iser_conn_alloc(name);
	if (p_iser_conn == NULL) {
		IERROR("iser_conn_alloc failed\n");
		goto iser_connection_establish_failure;
	}

	/* Save the iSCSI connection handle */
	p_iser_conn->iscsi_conn_h = iscsi_conn_h;
	/* Set the entity */
	p_iser_conn->p_entity = p_entity;
	/* Set the adaptor (now single adaptor assumed) */
	p_adaptor = &iser_global.adaptor[0];

	/* Add the connection to the adaptor connections list */
	iser_adaptor_add_connection(p_adaptor, p_iser_conn);
	/* Add the connection to the entity connections list */
	iser_entity_add_connection(p_entity, p_iser_conn);
	/* Add the connection to the connections hash */
	hash_add_iser_conn(p_iser_conn);

	ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_PENDING);

	p_iser_conn->alloc_post_recv_bufs_num =
	    ISER_INITIATOR_EP_AVG_POST_RECV + 2;
	p_iser_conn->initial_post_recv_bufs_num =
	    ISER_INITIATOR_INITIAL_POST_RECV + 2;

	/* Connect the endpoint */
	if (iser_connect(p_iser_conn, dst_addr, src_addr) != 0) {
		IERROR("iser_connect failed\n");
		goto iser_connection_establish_failure;
	}
	ITRACE_EXIT();
	return ISER_SUCCESS;

      iser_connection_establish_failure:
	if (p_iser_conn) {
		ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_DOWN);
		iser_conn_destroy(p_iser_conn);
	}
	IERROR("%s: failed\n", __FUNCTION__);
	ITRACE_EXIT();
	return ISER_FAILURE;
}				/* iser_initiator_connection_establish */

/**
 * Function:	 iser_initiator_connection_accept
 *
* initiators are not supposed to call connection_accept(), so this is a stub
 * so this is a stub
 */
iser_status_t
iser_initiator_connection_accept(iser_conn_request_t conn_req_h,
				 iser_conn_handle_t iscsi_conn_h,
				 int accept)
{
	return ISER_UNREGISTERED;
}				/* iser_initiator_connection_accept */

/**
 * Function:	 iser_initator_get_data
 *
 * Parameters:
 * conn_h - IN - the pdu will be sent on this connection
 * p_data_in_pdu - IN - the pdu to send.
 * notify_enable - IN -0 - iSCSI does not wish to be notified
 * when this operation finishes successfully.
 * 1 - iSCSI wants to be notified when this operation finishes successfully.
 *
 * initiators are not supposed to call get_data(), so this is a stub
 */
iser_status_t
iser_initiator_get_data(iser_conn_handle_t conn_h,
			struct iser_send_pdu_t * p_r2t_pdu,
			int notify_enable)
{
	return ISER_UNREGISTERED;
}				/* iser_initiator_get_data */

/**
 * Function:	 iser_initator_put_data
 *
 * Parameters:
 * conn_h - IN - the pdu will be sent on this connection
 * p_data_in_pdu - IN - the pdu to send.
 * notify_enable - IN -0 - iSCSI does not wish to be notified
 * when this operation finishes successfully.
 * 1 - iSCSI wants to be no// segment length will be set in put_data()tified when this operation finishes successfully.
 *
 *
 * the initiator is not suppose to call put_data().
 * so this here is a stub.
 */
iser_status_t iser_initiator_put_data(iser_conn_handle_t conn_h,	/* IN */
				      struct iser_send_pdu_t * p_data_in_pdu,	/* IN */
				      int notify_enable)
{				/* IN */
	return ISER_UNREGISTERED;
}				/* iser_initiator_put_data */

/**
 * iser_initiator_register_rdma_memory - Registers memory intended for RDMA, obtaining RMR context
 *
 * returns 0 on success, -1 on failure
 */
int iser_initiator_register_rdma_memory(struct iser_task_t *p_iser_task,	//<! IN - task
					enum iser_data_dir cmd_dir,	//<! IN - data flow direction
					struct iser_data_buf_t *p_mem)	//<! IN - memory to register
{
	unsigned long flags;
	struct iser_regd_buf_t *p_rdma_regd;
	struct list_head *p_task_buff_list;
	unsigned int priv_flags;
	struct iser_mem_t *p_phys_vec;
	unsigned long cmd_data_len;
	unsigned int phys_vec_len = 0;
	int phys_vec_to_free = 0;
	int aligned_len;
	int ret_val = 0;
	enum dma_data_direction dir;

	ITRACE_ENTRY();

	priv_flags = IBV_ACCESS_LOCAL_WRITE;

	if (cmd_dir == ISER_DIR_IN) {
		p_rdma_regd = &p_iser_task->rdma_write_local;
		p_task_buff_list = &p_iser_task->rcv_buff_list;
		priv_flags |= IBV_ACCESS_REMOTE_WRITE;
		dir = DMA_FROM_DEVICE;
	} else if (cmd_dir == ISER_DIR_OUT) {
		struct iser_dto_t *p_dto = &p_iser_task->rdma_read_dto;

		p_dto->type = ISER_DTO_PASSIVE;
		dto_set_task(p_dto, p_iser_task);
		iser_task_put(p_iser_task);
		p_dto->p_conn = p_iser_task->p_conn;
		p_dto->regd_vector_len = 1;

		p_dto->regd[0] = p_rdma_regd =
		    &p_iser_task->rdma_read_local;

		p_task_buff_list = &p_iser_task->send_buff_list;
		priv_flags |= IBV_ACCESS_REMOTE_READ;
		dir = DMA_TO_DEVICE;
	} else {
		IPANIC("Unexpected command direction: %d for task: 0x%p\n",
		       cmd_dir, p_iser_task);
		ret_val = -1;
		goto register_rdma_memory_exit;
	}

	if (p_mem->type == ISER_BUF_TYPE_SINGLE) {

		p_phys_vec = p_iser_task->mem_vec;
		phys_vec_len = 1;

		p_phys_vec->addr = (unsigned long) p_mem->p_buf;
		p_phys_vec->size = cmd_data_len = (unsigned long) p_mem->size;

		ITRACE(ISER_TRACE_RDMA_MEM_REG,
		       "Registering SINGLE RDMA buffer virt: 0x%p phys: 0x%08lX, sz: %ld\n",
		       p_mem->p_buf, (unsigned long) p_phys_vec->addr,
		       cmd_data_len);
	} else {
		if (p_mem->size <= MAX_PHYS_VEC_IN_TASK) {
			p_phys_vec = p_iser_task->mem_vec;
		} else {
			ITRACE(ISER_TRACE_RDMA_MEM_REG,
			       "Allocating phys mem descriptor for %d entries\n",
			       p_mem->size);
			p_phys_vec =
			    (struct iser_mem_t *) my_kmalloc(p_mem->size
                                                             * sizeof(*p_phys_vec),
							     "p_phys_vec");
			if (p_phys_vec == NULL) {
				IERROR
				    ("Failed to allocate phys mem descriptor for %d entries\n",
				     p_mem->size);
				ret_val = -1;
				goto register_rdma_memory_exit;
			}
			phys_vec_to_free = 1;
		}

		switch (p_mem->type) {
		case ISER_BUF_TYPE_IOVEC_VIRT:
			ITRACE(ISER_TRACE_RDMA_MEM_REG,
			       "Translating RDMA IOVEC_VIRT sz: %d\n",
			       p_mem->size);
			phys_vec_len =
			    iser_iovec_virt_to_mem_t(p_iser_task,
						     cmd_dir,
						     p_mem,
						     p_phys_vec,
						     &cmd_data_len);
			break;

		case ISER_BUF_TYPE_SCATTERLIST:
			TRACE(TRACE_ISER_MEM,
			       "%s, Translating RDMA SCATTERLIST sz: %d\n",
			      __func__, p_mem->size);
			phys_vec_len =
			    iser_scatterlist_to_mem_t(p_iser_task,
						      cmd_dir,
						      p_mem,
						      p_phys_vec,
						      &cmd_data_len);
			TRACE(TRACE_ISER_MEM, "%s: phys_vec_len=%u\n",
			      __func__, phys_vec_len);
			break;

		default:
			IPANIC
			    ("Unsupported data buffer descriptor type: %d in p_data: 0x%p\n",
			     p_mem->type, p_mem);
			break;
		}

		/* Determine an aligned stretch and check that the entire vector is aligned correctly */
		aligned_len =
		    iser_iovec_aligned_length(p_phys_vec, phys_vec_len,
					      NULL);
		if (aligned_len != phys_vec_len) {
			int i;
			IERROR
			    ("Can't register for rdma because of alignment violation\n");
			for (i = 0; i < phys_vec_len; i++) {
				IINFO("phys_vec[%d] addr=0x%lX size=%ld\n",
				      i,
				      (unsigned long) p_phys_vec[i].addr,
				      (unsigned long) p_phys_vec[i].size);
			}
			ret_val = -1;
			goto register_rdma_memory_exit;
		}
	}

	// Start initializing the RDMA regsitered memory descriptor
	memset(p_rdma_regd, 0, sizeof(*p_rdma_regd));
	p_rdma_regd->type = ISER_REGD_BUF_RDMA;
	p_rdma_regd->virt_buf = *p_mem;

	// Register the user memory
	TRACE(TRACE_ISER_MEM, "%s: registering RDMA memory\n", __func__);
	ret_val = iser_register_phys_mem(p_iser_task->p_conn, p_phys_vec,	/* physical memory vector */
					 phys_vec_len,	/* vector length */
					 cmd_data_len,	/* total size */
					 priv_flags,	/* DAT_MEM_PRIV_ALL_FLAG */
					 &p_rdma_regd->mem_reg,	/* local memory handles returned */
					 &p_rdma_regd->as.rdma.rmr_context,
	                                 dir);

	if (phys_vec_to_free) {
		ITRACE(ISER_TRACE_RDMA_MEM_REG,
		       "Freeing phys mem descriptor for %d entries\n",
		       p_mem->size);
		my_kfree((void**) &p_phys_vec, "p_phys_vec");
	}
	if (ret_val != 0) {
		IERROR("Failed to register %d phys entries\n",
		       phys_vec_len);
		goto register_rdma_memory_exit;
	}

	spin_lock_irqsave(&p_iser_task->task_lock, flags);

	/* Add a reference to this registered data buffer because it is in use by the iSCSI layer now */
	iser_regd_buff_reference(p_rdma_regd);

	p_iser_task->dir[cmd_dir] = 1;
	p_iser_task->data_len[cmd_dir] = cmd_data_len;

	/* Add the registered mem to the list of buffers which are released when the task completes */
	list_add(&p_rdma_regd->free_upon_comp_list, p_task_buff_list);

	spin_unlock_irqrestore(&p_iser_task->task_lock, flags);

      register_rdma_memory_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_initiator_register_rdma_memory */


/**
 * iser_initiator_register_unsolicited_data - Registers memory intended for sending as unsolicited data
 *
 * returns 0 on success, -1 on failure
 */
int iser_initiator_register_local_unsolicited_data(struct iser_task_t *p_iser_task,	//<! IN - task
						   struct iser_data_buf_t *p_mem)	//<! IN - memory to register
{
	unsigned long flags;
	struct iser_dto_t *p_dto;
	unsigned long cmd_data_len;
	int ret_val = 0;
	int i;

	ITRACE_ENTRY();

	p_dto = &p_iser_task->rdma_read_dto;

	p_dto->type = ISER_DTO_PASSIVE;
	dto_set_task(p_dto, p_iser_task);
	p_dto->p_conn = p_iser_task->p_conn;

	if (iser_dto_add_local_memory(p_dto,
				      p_mem,
				      p_iser_task->mem_vec,
				      MAX_PHYS_VEC_IN_TASK,
				      &cmd_data_len) != 0) {
		IERROR("iser_dto_add_local_memory failed\n");
		ret_val = -1;
		goto initiator_register_unsolicited_data_exit;
	}
	/* add_local_memory() references all registered buffers, but this dto is not used in any IO */
	for (i = 0; i < p_dto->regd_vector_len; i++) {
		iser_regd_buff_dereference(p_dto->regd[i]);
	}

	spin_lock_irqsave(&p_iser_task->task_lock, flags);
	p_iser_task->dir[ISER_DIR_OUT] = 1;
	p_iser_task->data_len[ISER_DIR_OUT] = cmd_data_len;
	spin_unlock_irqrestore(&p_iser_task->task_lock, flags);

      initiator_register_unsolicited_data_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_initiator_register_local_unsolicited_data */


/**
 * Function:	 iser_initiator_send_control
 *
 * Parameters:
 * iser_conn_handle_t - IN -	connection handle to note on which
 * connection to send tremotehe pdu
 * iser_pdu_t - IN -	the pdu to send.
 *
 *
 * An iSCSI layer at the INITIATOR requests its local iSER datamover
 * layer to transmit an iSCSI control-type PDU to the peer iSCSI layer
 * operating in the remote iSCSI node. The iSER datamover layer sends
 * the PDU after adding its own protocol header.
 */
iser_status_t iser_initiator_send_control(iser_conn_handle_t iscsi_conn_h,	/* IN */
					  struct iser_send_pdu_t *
					  p_ctrl_pdu)
{				/* IN */
	iser_status_t iser_ret = ISER_SUCCESS;
	union iser_pdu_bhs_t *p_bhs;
	struct iser_conn_t *p_iser_conn = NULL;
	struct iser_dto_t *p_send_dto = NULL;
	struct iser_task_t *p_iser_task = NULL;
	unsigned long buf_offset;
	unsigned long data_seg_len;
	unsigned int itt;
	unsigned char opcode;

	ITRACE_ENTRY();

	/* Check the parameters correctness */
	if (p_ctrl_pdu == NULL) {
		IERROR("NULL control PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;
	}
	p_bhs = p_ctrl_pdu->p_bhs;
	if (p_bhs == NULL) {
		IERROR("NULL BHS in control PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;
	} else if ((void*) p_bhs < (void*) PAGE_SIZE) {
		IERROR("Invalid BHS [%p] in control PDU descriptor\n", p_bhs);
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;
	}

	/* Look up the connection */
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);

	if (p_iser_conn == NULL) {
		IERROR("Failed to find connection, iscsi_conn_h: %lX\n",
		       (unsigned long) iscsi_conn_h);
		iser_pdu_print((char *) __func__, NULL,
			       (char *) p_bhs->buf, NULL);
		iser_ret = ISER_INVALID_CONN;
		goto send_control_error;
	}
	if (atomic_read(&p_iser_conn->state) != ISER_CONN_UP) {
		IERROR
		    ("Failed to send_control as connection is not up, iscsi_conn_h: %lX, p_conn: 0x%p\n",
		     (unsigned long) iscsi_conn_h, p_iser_conn);
		iser_pdu_print((char *) __func__, NULL,
			       (char *) p_bhs->buf, NULL);
		p_iser_conn = NULL;	/* Inhibits connection shutdown */
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	/* Allocate and fill send DTO descriptor, headers buffer and add it to the DTO */
	p_send_dto = iser_dto_send_create(p_iser_conn, p_bhs);
	if (p_send_dto == NULL) {
		IERROR("Failed to create send DTO, conn: 0x%p\n",
		       p_iser_conn);
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	/* Retrieve some BHS fields */
	itt = ntohl(p_bhs->dword.itt);
	data_seg_len =
	    ALIGN_TO_BIT((ntohl(p_bhs->dword.length) & ISCSI_DSL_MASK), 2);
	opcode = p_bhs->byte.opcode & ISCSI_OPCODE_MASK;

	/* According to the PDU opcode ... */
	switch (opcode) {

	case ISCSI_OP_SCSI_CMD:{
			unsigned long edtl;
			unsigned char *p_iser_header;

			/* Retrieve ExpectedDataTransferLength from the Command BHS */
			edtl =
			    ntohl(p_bhs->dword.
				  other[ISCSI_CMD_FIELD_EDTL]);

			/* print out the Command PDU BHS */
			if (ITRACEABLE(ISER_TRACE_SEND_CONTROL)) {
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf, NULL);
			}

			/* Allocate new task descriptor */
			p_iser_task = iser_task_alloc(p_iser_conn, itt);
			if (p_iser_task == NULL) {
				IERROR
				    ("Failed to allocate iser task, conn: 0x%p, itt: %d\n",
				     p_iser_conn, itt);
				iser_ret = ISER_FAILURE;
				goto send_control_error;
			}
			/* Set the DTO's parent task */
			dto_set_task(p_send_dto, p_iser_task);

			/* Headers buffer was added to the DTO descriptor as the first regd buf to use */
			p_iser_header =
			    p_send_dto->regd[0]->virt_buf.p_buf;

			/* Set-up RDMA and data according to the command direction.
			   Note that it's possible to have both directins */
			if (IS_SET_ISCSI_FLAG_READ_CMD(p_bhs->byte.flags)
			    && edtl > 0) {	/* read command */

				/* Register user buffer memory and initialize passive rdma dto descriptor.
				   Total data size is stored in p_iser_task->data_len[ISER_DIR_IN]. */
				TRACE(TRACE_ISER_MEM, "%s: registering read cmnd data, edtl=%lu\n", __func__, edtl);
				if (iser_initiator_register_rdma_memory
				    (p_iser_task, ISER_DIR_IN,
				     &p_ctrl_pdu->data.command.buf_in) !=
				    0) {
					IERROR
					    ("Failed to set up Data-IN RDMA\n");
					iser_ret = ISER_FAILURE;
					goto send_control_error;
				}
				if (edtl >
				    p_iser_task->data_len[ISER_DIR_IN]) {
					IPANIC
					    ("Total data length: %ld, less than EDTL: %ld, in READ cmd BHS itt: %d, p_conn: 0x%p\n",
					     p_iser_task->
					     data_len[ISER_DIR_IN], edtl,
					     itt, p_iser_conn);
				}

				ISER_HDR_SET_BITS(p_iser_header, RSV, 1);
				ISER_HDR_R_VADDR(p_iser_header) =
				cpu_to_be64((unsigned long) p_iser_task->rdma_write_local.mem_reg.ibv_mr->addr);
				ISER_HDR_R_RKEY(p_iser_header) =
				    htonl(p_iser_task->rdma_write_local.as.
					  rdma.rmr_context);

				ITRACE(ISER_TRACE_SEND_CONTROL,
		   		       "Command itt: %d, sending READ "
				       "tags, RMR: 0x%08X, VA: 0x%08lX\n",
				       itt,
				       p_iser_task->rdma_write_local.as.rdma.rmr_context,
				       (unsigned long)p_iser_task->rdma_write_local.mem_reg.ibv_mr->addr);
			}
			/* read cmd */
			if (IS_SET_ISCSI_FLAG_WRITE_CMD(p_bhs->byte.flags)
			     && edtl > 0) {	/* write command */

				TRACE(TRACE_ISER, "%s: WRITE FLAG, edtl=%lu, "
				      "unsolicited_sz=%lu\n",
				      __func__, (unsigned long ) edtl,
				      (unsigned long) p_ctrl_pdu->data.command.unsolicited_sz);

				if (p_ctrl_pdu->data.command.
				    unsolicited_sz < edtl) {
					/* Register user buffer memory and initialize passive rdma dto descriptor.
					   Total data size is stored in p_iser_task->data_len[ISER_DIR_OUT]. */
					TRACE(TRACE_ISER_MEM, "%s: registering write cmnd data, edtl=%lu\n", __func__, edtl);
					if (iser_initiator_register_rdma_memory(p_iser_task, ISER_DIR_OUT, &p_ctrl_pdu->data.command.buf_out) != 0) {
						IERROR
						    ("Failed to register write cmd RDMA memory\n");
						iser_ret = ISER_FAILURE;
						goto send_control_error;
					}

					ISER_HDR_SET_BITS(p_iser_header,
							  WSV, 1);
					ISER_HDR_W_VADDR(p_iser_header) =
					cpu_to_be64((unsigned long) p_iser_task->rdma_read_local.mem_reg.ibv_mr->addr + p_ctrl_pdu->data.command.unsolicited_sz);
					ISER_HDR_W_RKEY(p_iser_header) =
					    htonl(p_iser_task->
						  rdma_read_local.as.rdma.
						  rmr_context);

					ITRACE(ISER_TRACE_SEND_CONTROL,
					"Command itt: %d, sending WRITE "
					"tags, RMR: %08x, VA: %p "
					"+ unsol.sz.: %d\n",
					itt,
					p_iser_task->rdma_read_local.as.rdma.rmr_context,
					p_iser_task->rdma_read_local.mem_reg.ibv_mr->addr,
					p_ctrl_pdu->data.command.unsolicited_sz);
				} else if (p_ctrl_pdu->data.command.
					   buf_out.type !=
					   ISER_BUF_TYPE_SINGLE
					   && p_ctrl_pdu->data.command.
					   buf_out.size >
					   MAX_REGD_BUF_VECTOR_LEN) {
					/* unsolicited data only, scatter list exceeds the max allowed for local memory,
					   if pre-registered cache is used we may end up with more entities in IOV than
					   declared when the end-point was created, thus we try to do a single registration
					 */
					TRACE(TRACE_ISER_MEM, "%s: registering unsolicited write cmnd data only, edtl=%lu\n", __func__, edtl);
					if (iser_initiator_register_rdma_memory(p_iser_task, ISER_DIR_OUT, &p_ctrl_pdu->data.command.buf_out) != 0) {
						IERROR
						    ("Failed to register write cmd RDMA memory\n");
						iser_ret = ISER_FAILURE;
						goto send_control_error;
					}
				} else {	/* unsolicited data only, does not need remote registration */
					if (iser_initiator_register_local_unsolicited_data(p_iser_task, &p_ctrl_pdu->data.command.buf_out) != 0) {
						IERROR
						    ("Failed to register write cmd local unsolicited memory\n");
						iser_ret = ISER_FAILURE;
						goto send_control_error;
					}
				}

				/* Check that EDTL matches provided data buffer length */
				if (edtl >
				    p_iser_task->data_len[ISER_DIR_OUT]) {
					IPANIC
					    ("Total data length: %ld, less than EDTL: %ld, in WRITE cmd BHS itt: %d, p_conn: 0x%p\n",
					     p_iser_task->
					     data_len[ISER_DIR_OUT], edtl,
					     itt, p_iser_conn);
				}

				/* If there is immediate data, add its register buffer reference to the send dto descriptor */
				if (p_ctrl_pdu->data.command.immediate_sz >
				    0) {
					ITRACE(ISER_TRACE_SEND_CONTROL,
					       "Command itt: %d, WRITE, adding imm.data sz: %d\n",
					       itt,
					       p_ctrl_pdu->data.command.
					       immediate_sz);

					iser_dto_clone_regd_buffs(p_send_dto,	/* dst */
								  &p_iser_task->rdma_read_dto,	/* src */
								  0,	/* offset */
								  p_ctrl_pdu->
								  data.
								  command.
								  immediate_sz);
					/*
					   iser_dto_add_regd_buff(p_send_dto,
					   &p_iser_task->rdma_read_local,
					   USE_NO_OFFSET, USE_SIZE(p_ctrl_pdu->data.command.immediate_sz));
					 */
				}
			}	/* write cmd */
		}
		break;		/* ISCSI_OP_SCSI_CMD */

	case ISCSI_OP_DATA_OUT:{
			if (data_seg_len == 0) {
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf, NULL);
				IPANIC
				    ("Zero DataSegmentLength in Data-OUT PDU\n");
			}

			/* Retrive BufferOffset from Data-OUT BHS */
			buf_offset =
			    ntohl(p_ctrl_pdu->p_bhs->dword.
				  other[ISCSI_DATA_OUT_FIELD_OFFSET]);

			if (ITRACEABLE(ISER_TRACE_SEND_CONTROL)) {
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf, NULL);
			}

			/* Find task in the hash */
			p_iser_task =
			    hash_find_iser_task(p_iser_conn, itt);
			if (p_iser_task == NULL) {
				IERROR("Task not found, itt=%d\n", itt);
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf, NULL);
				iser_ret = ISER_INVALID_ITT;
				goto send_control_error;
			}
			if (!p_iser_task->dir[ISER_DIR_OUT]) {
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf, NULL);
				IPANIC
				    ("Data-OUT PDU is sent for a non-WRITE task\n");
			}

			/* Set the DTO's parent task */
			dto_set_task(p_send_dto, p_iser_task);

			/* Set-up the registered buffer entries for the data segment */
			iser_dto_clone_regd_buffs(p_send_dto,	/* dst */
						  &p_iser_task->rdma_read_dto,	/* src */
						  buf_offset,
						  data_seg_len);
			/*
			   iser_dto_add_regd_buff(p_send_dto,
			   &p_iser_task->rdma_read_local,
			   USE_OFFSET(buf_offset), USE_SIZE(data_seg_len));
			 */

			if (buf_offset + data_seg_len >
			    p_iser_task->data_len[ISER_DIR_OUT]) {
				IERROR
				    ("Offset (%ld) & Data segment length (%ld) in Data-Out PDU is inconsistent with total data len (%ld), itt=%d\n",
				     buf_offset, data_seg_len,
				     p_iser_task->data_len[ISER_DIR_OUT],
				     itt);
				iser_ret = ISER_ILLEGAL_PARAM;
				goto send_control_error;
			}

			ITRACE(ISER_TRACE_SEND_CONTROL,
			       "data-out itt: %d, offset: %ld, sz: %ld\n",
			       itt, buf_offset, data_seg_len);
		}
		break;		/* ISCSI_OP_DATA_OUT */

	case ISCSI_OP_TASK_MGT_REQ:	// ToDo: check type, maybe there are buf_in, buf_out
		// ToDo p_ctrl_pdu->data.task_mgt_req.buf_in
	case ISCSI_OP_NOP_OUT:
		/* find task in the hash */
		p_iser_task = hash_find_iser_task(p_iser_conn, itt);
		/* These PDUs do not always relate to a task, but it's ok */
		dto_set_task(p_send_dto, p_iser_task);
		/* Beware: no break here! */

	case ISCSI_OP_LOGIN_REQ:
	case ISCSI_OP_TEXT_REQ:
	case ISCSI_OP_LOGOUT_REQ:{

			if (ITRACEABLE(ISER_TRACE_SEND_CONTROL)) {	/* print out the PDU */
				iser_pdu_print((char *) __func__,
					       p_iser_conn,
					       (char *) p_bhs->buf,
					       &p_ctrl_pdu->data.tx.buf);
			}

			data_seg_len =
			    ALIGN_TO_BIT((ntohl
					  (p_ctrl_pdu->p_bhs->dword.
					   length) & ISCSI_DSL_MASK), 2);
			if (data_seg_len > 0) {
				// Allocate data regd buffer and copy the user data
				iser_dto_copy_send_data(p_send_dto,
							&p_ctrl_pdu->data.
							tx.buf,
							data_seg_len);

				// We compare the actual buf size (total_buf_sz) to the dsl in the BHS (data_seg_len).
				// We assume that a padding (up to 4 bytes) may be used
				/*
				   if (total_buf_sz != ALIGN_TO_BIT(data_seg_len, 2)) {
				   IERROR("Data segment length (%d) in PDU is inconsistent with total data len (%d)\n",
				   data_seg_len, total_buf_sz);
				   iser_ret = ISER_ILLEGAL_PARAM;
				   // ToDo : free resources
				   // erezz: Who should close the connection??
				   goto send_control_error;
				   }
				 */
			}
		}
		break;

	default:{
			IERROR("Unsupported opcode = %d\n", opcode);
			iser_ret = ISER_ILLEGAL_PARAM;
			goto send_control_error;
		}
		break;

	}			/* switch p_bhs->byte.opcode */

	// Post-Receive buffer for a reply PDU
	// ToDo: account for different types of NOP-OUT (req, resp)
	if (opcode != ISCSI_OP_DATA_OUT) {
		if (iser_post_receive_control
		    (p_iser_conn->param.TargetRecvDataSegmentLength,
		     p_iser_conn) != 0) {
			IERROR("post_rcv_buff failed!\n");
			iser_ret = ISER_FAILURE;
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto send_control_error;
		}
	}

	/* Start the DTO */
	if (p_iser_task != NULL) {
		if (opcode == ISCSI_OP_SCSI_CMD) {
			iser_task_set_status(p_iser_task,
					     ISER_TASK_STATUS_STARTED);
		}
	}
	if (iser_start_dto(p_send_dto) != 0) {
		IERROR("Failed to start send DTO\n");
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	ITRACE_EXIT();
	return ISER_SUCCESS;

      send_control_error:
	if (p_send_dto != NULL) {
		iser_conn_remove_dto(p_iser_conn, p_send_dto);
	}
	if (p_iser_conn != NULL) {
		/* drop the connection, any open tasks will be deleted during shutdown */
		IERROR("%s failed, dropping connection:0x%p\n", __func__,
		       p_iser_conn);
		iser_conn_async_terminate(p_iser_conn);
	}

	ITRACE_EXIT();
	return iser_ret;

}				/* iser_initiator_send_control */

/**
 * iser_initiator_dto_completion - Handle a successful DTO completion
 *                                 at Initiator
 *
 * Note: We do not re-post receive buffers in here.  Instead, we post
 *       a recv buffer for each control operation we perform in order to
 *       receive the reply.
 *
 * returns 0 on success, -1 on failure
 */
int iser_initiator_dto_completion(struct iser_dto_t *p_dto,	//!< IN - completed DTO descriptor
				  unsigned long dto_xfer_len)	//!< IN - DTO transfer length
{
	struct iser_conn_t *p_iser_conn;
	struct iser_task_t *p_iser_task = NULL;
	union iser_pdu_bhs_t *p_bhs;
	int ret_val = 0;
	unsigned int rx_count;
	unsigned int itt;
	unsigned char opcode;

	ITRACE_ENTRY();

	p_iser_conn = p_dto->p_conn;
	if (p_iser_conn == NULL) {
		IPANIC("NULL connection in p_dto: 0x%p\n", p_dto);
	}

	if (p_dto->type == ISER_DTO_RCV
	    && p_iser_conn->hello_state == ISER_HELLO_SENT) {
		TRACE(TRACE_ISER,
		      "%s: Handling receive in hello exchange\n",
		      __func__);
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		ret_val =
		    iser_initiator_recv_helloreply(p_iser_conn, p_dto);
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		return ret_val;
	}

	switch (p_dto->type) {
	case ISER_DTO_RCV:	// handle received messages

		/* Number of posted receive buffers has decreased by one */
		iser_conn_dec_post_recv_cnt(p_iser_conn);

		/* Create descriptor for the received PDU */
		iser_dto_create_recv_pdu(p_dto, dto_xfer_len);
		p_bhs = p_dto->as.recv.pdu.p_bhs;

		/* Print the PDU */
		if (ITRACEABLE(ISER_TRACE_RECV_CONTROL)) {
			iser_pdu_print((char *) __func__, p_iser_conn,
				       (char *) p_bhs->buf,
				       &p_dto->as.recv.pdu.rx_data);
		}

		/* Retrieve ITT, note that it may be 0xFFFFFFFF when irrelevant */
		itt = ntohl(p_bhs->dword.itt);
		/* Retrieve opcode */
		opcode = p_bhs->byte.opcode & ISCSI_OPCODE_MASK;

		/* Find task descriptor if relevant */
		if (itt != ISCSI_INVALID_ITT
		    && (opcode == ISCSI_OP_SCSI_RSP
			|| opcode == ISCSI_OP_NOP_IN
			|| opcode == ISCSI_OP_TASK_MGT_RSP
			|| opcode == ISCSI_OP_LOGIN_RSP
			|| opcode == ISCSI_OP_TEXT_RSP
			|| opcode == ISCSI_OP_LOGOUT_RSP)) {
			p_iser_task =
			    hash_find_iser_task(p_iser_conn, itt);
			if (p_iser_task != NULL) {
				// The rcv buffer belongs now to the task
				dto_set_task(p_dto, p_iser_task);
				// Account for the used up posted recv buffer
				iser_task_recvd_pdu_count_inc(p_iser_task);
				// ToDo: something similiar to task->rx_count for non-task related PDUs, on conn. level

				/* SCSI Response PDU completes the task processing */
				if (opcode == ISCSI_OP_SCSI_RSP) {
					// Declare the task completed
					iser_task_set_status(p_iser_task,
							     ISER_TASK_STATUS_COMPLETED);

					// Deregister memory of all send buffers before notifying iSCSI
					/* This happens when the task is freed anyway, but ok. */
					iser_task_release_send_buffers(p_iser_task);

					/* I am not sure why since we are just freeing the task in a minute, but ok */
					rx_count = iser_task_recvd_pdu_count_reset(p_iser_task);
					ITRACE(ISER_TRACE_SEND_CONTROL,
					       "Task complete, itt: %d, received %d buffers\n",
					       itt, rx_count);

					/* This very likely deletes the task. */
					hash_delete_iser_task(p_iser_task);
					if (hash_find_iser_task(p_iser_conn, itt) != NULL) {
						IPANIC
						    ("Task found after delete, itt=0x%0X\n",
						     itt);
					}
				}
			}
		}
		// Notify iSCSI layer
		iser_dto_get(p_dto);
		iser_conn_remove_dto(p_iser_conn, p_dto);
		ITRACE(ISER_TRACE_CTRL_NOTIFY,
		       "Control notify, DTO:0x%p, as PDU:0x%p\n", p_dto,
		       &p_dto->as.recv.pdu);
		p_iser_conn->p_entity->api_cb.control_notify(p_iser_conn->
							     iscsi_conn_h,
							     &p_dto->as.
							     recv.pdu);
		break;

	case ISER_DTO_SEND:	// handle sent messages

		ITRACE(ISER_TRACE_SEND_DTO,
		       "Initiator, Data sent p_dto=0x%p\n", p_dto);
		// One posted send DTO less
		atomic_dec(&p_iser_conn->post_send_buf_count);

		// Release the DTO descriptor
		iser_conn_remove_dto(p_dto->p_conn, p_dto);

		break;		// p_dto->type == ISER_DTO_SEND

	default:
		IPANIC("Illegal initiator iSER DTO type: %d\n",
		       p_dto->type);
		break;

	}			/* switch p_dto->type */

// initiator_dto_completion_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_initiator_dto_completion */


/**
 * iser_initiator_allocate_connection_resources - iSER API. Implements Allocate_Connection_Resources primitive.
 *
 *  Performs exchange of API functions. Allocates additional receive
 *  buffers. Performs static memory registration if instructed to do so.
 *
 *  returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
*/
iser_status_t iser_initiator_allocate_connection_resources(iser_conn_handle_t iscsi_conn_h,	//!< IN - iSCSI-supplied connection handle
							   struct iser_conn_resources_t * conn_resources,	//!< IN - additional resources descriptor
							   int
							   *max_oustanding_commands)
{
	struct iser_entity_t *p_entity;
	struct iser_conn_t *p_iser_conn;
	iser_status_t iser_ret;

	ITRACE_ENTRY();

	p_entity = (struct iser_entity_t *) conn_resources->api_h;
	if (p_entity == NULL) {
		IERROR("NULL API handle\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto allocate_connection_resources_exit;
	}

#if 0
/*                      --- What is 256 * 1024 ---                      */
	if (conn_resources->max_recv_pdu_sz > 256 * 1024) {
		IERROR("max_recv_pdu_sz too large\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto allocate_connection_resources_exit;
	}
#endif /* No upper bound on max recv pdu size */

	if (conn_resources->conn_type == ISER_TRANSPORT_CONN_IB) {
		// Find the connection
		p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
		if (p_iser_conn == NULL) {
			IERROR
			    ("Could not find iSER connection for iscsi_h = 0x%p\n",
			     iscsi_conn_h);
			iser_ret = ISER_INVALID_CONN;
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto allocate_connection_resources_exit;
		}

		if (p_iser_conn->p_entity != p_entity) {
			IERROR
			    ("Entity in conn_res does not match conn for iscsi_h = 0x%p\n",
			     p_iser_conn->iscsi_conn_h);
			iser_ret = ISER_ILLEGAL_PARAM;
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto allocate_connection_resources_exit;
		}
	} else {
		IERROR
		    ("This connection type unsupported, iscsi_h = 0x%p\n",
		     iscsi_conn_h);
		iser_ret = ISER_ILLEGAL_PARAM;
		goto allocate_connection_resources_exit;
	}

	p_iser_conn->max_outstand_cmds = conn_resources->max_outstand_cmds;

	if (p_iser_conn->param.InitiatorRecvDataSegmentLength ==
	    defaultInitiatorRecvDataSegmentLength) {
		p_iser_conn->param.InitiatorRecvDataSegmentLength =
		    conn_resources->max_recv_pdu_sz;
	} else if (p_iser_conn->param.InitiatorRecvDataSegmentLength !=
		   conn_resources->max_recv_pdu_sz) {
		IERROR
		    ("max_recv_pdu_sz(%d) does not match non-default value of InitiatorRecvDataSegmentLength(%d)\n",
		     conn_resources->max_recv_pdu_sz,
		     p_iser_conn->param.InitiatorRecvDataSegmentLength);
		iser_ret = ISER_ILLEGAL_PARAM;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto allocate_connection_resources_exit;
	}

	/* post the initial login response recv buffer.  */
	if (iser_post_receive_control
	    (ISER_LOGIN_PHASE_PDU_DATA_LEN, p_iser_conn) != 0) {
		TRACE_ERROR("%s: iser_post_receive_control failed\n",
			    __func__);
		ITRACE_EXIT();
		return -1;
	}

	ITRACE(ISER_TRACE_CONN,
	       "Max outst. cmds: %d, Allocate post recv bufs: %d, Initially post: %d\n",
	       p_iser_conn->max_outstand_cmds,
	       p_iser_conn->alloc_post_recv_bufs_num,
	       p_iser_conn->initial_post_recv_bufs_num);

	iser_ret = iser_initiator_alloc_ffp_recv_buffs(p_iser_conn);

      allocate_connection_resources_exit:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_allocate_connection_resources */

/**
 * iser_initiator_enable_datamover - iSER API. Implements Enable_Datamover primitive.
 *
 * In target, sends the Login-Response PDU.
 * Upgrades the connection to RDMA-assisted.
 * Performs the Hello Message exchange.
 *
 * returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
 */
iser_status_t iser_initiator_enable_datamover(iser_conn_handle_t iscsi_conn_h,	//!< IN - iSCSI-supplied connection handle
					      iser_transport_conn_t tport_conn,	//!< IN - transport connection descriptor
					      struct iser_send_pdu_t * p_login_resp_pdu)	//!<IN - LoginResponse PDU
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_conn_t *p_iser_conn;

	ITRACE_ENTRY();

	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		iser_ret = ISER_INVALID_CONN;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto enable_data_mover_exit;
	}

	p_iser_conn->p_entity->enter_hello_exchg(p_iser_conn);
	wait_event_interruptible(p_iser_conn->conn_wait,
				 p_iser_conn->hello_state == ISER_HELLO_DONE);

	ITRACE_EXIT();
enable_data_mover_exit:
	return iser_ret;
}				/* iser_enable_datamover */

/**
 * Allocate the recieve buffers for the FFP.
 */
int iser_initiator_alloc_ffp_recv_buffs(struct iser_conn_t *p_conn)
{
	int i;

	/* receive buffer for the hello reply message */
	if (iser_post_receive_control(ISER_HELLO_REPLY_LEN, p_conn) !=
	    0) {
		TRACE_ERROR("%s: iser_post_receive_control failed\n",
			    __func__);
		ITRACE_EXIT();
		return -1;
	}

	// Initial Post-Receive buffers
	TRACE(TRACE_ISER, "%s: initial_post_recv_bufs_num=%d\n",
	      __func__, p_conn->initial_post_recv_bufs_num);
	for (i = 0; i < p_conn->initial_post_recv_bufs_num; i++) {
		if (iser_post_receive_control
		    (p_conn->param.TargetRecvDataSegmentLength,
		     p_conn) != 0) {
			TRACE_ERROR("Failed to post recv buffers\n");
			ITRACE_EXIT();
			return ISER_FAILURE;
		}
	}
	ITRACE(ISER_TRACE_CONN, "Allocated %d post recv bufs\n", i);

	return ISER_SUCCESS;
}

/**
 * \brief Enter the hello message exchange.
 */
int iser_initiator_enter_hello_exchg(struct iser_conn_t *p_conn)
{
	unsigned char hello[ISER_HELLO_LEN];
	struct iser_dto_t *p_hello_dto;

	ITRACE_ENTRY();

	/*
	 * Build the hello message.
	 */
	memset(&hello, 0, ISER_HELLO_LEN);
	hello[0] = ISER_OP_HELLO << 4;
	ISER_HDR_HELLO_IRD(&hello) = htons(p_conn->max_rdma_read_req_in);
	ISER_HDR_SET_BITS(&hello, HELLO_MAXVER, ISER_CURRENT_VERSION);
	ISER_HDR_SET_BITS(&hello, HELLO_MINVER, ISER_CURRENT_VERSION);

	p_hello_dto = iser_dto_raw_create(p_conn,
					  (unsigned char *) &hello,
					  ISER_HELLO_LEN);
	p_conn->hello_state = ISER_HELLO_SENT;
	iser_start_dto(p_hello_dto);

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Called to receive the hello reply message.
 */
static int iser_initiator_recv_helloreply(struct iser_conn_t *p_conn,
					  struct iser_dto_t *p_dto)
{
	unsigned char *pdu;
	int rej, max_ver, min_ver;

	ITRACE_ENTRY();

	pdu = p_dto->regd[0]->virt_buf.p_buf;

	/*
	 * Check that we have a Hello Reply message.
	 */
	if (ISER_OPCODE(pdu) != ISER_OP_HELLO_REPLY) {
		TRACE(TRACE_ISER, "%s: received message is not a Hello "
		      "Reply message\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	rej = ISER_HDR_GET_BITS(pdu, HELLO_REPLY_REJ);
	if (rej) {
		TRACE(TRACE_ISER,
		      "%s: received Hello Reply has the reject "
		      "bit set\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	p_conn->iser_read_depth = ntohs(ISER_HDR_HELLO_REPLY_ORD(pdu));
	max_ver = ISER_HDR_GET_BITS(pdu, HELLO_REPLY_MAXVER);
	min_ver = ISER_HDR_GET_BITS(pdu, HELLO_REPLY_MINVER);
	TRACE(TRACE_ISER, "%s: Hello Reply  message\n\tMaxVer=%d"
	      "\n\tMinVer=%d\n", __func__, max_ver, min_ver);

	/*
	 * Check if the version values above were ok.  If they were not,
	 * lets set the reject flag before sending the hello reply.
	 */
	if (max_ver != ISER_CURRENT_VERSION ||
	    min_ver != ISER_CURRENT_VERSION) {
		IERROR("%s: %s has incompatible version\n", __func__,
		       max_ver != ISER_CURRENT_VERSION
		       ? "max_ver" : "min_ver");
		ITRACE_EXIT();
		return -1;
	}

	p_conn->hello_state = ISER_HELLO_DONE;
	wake_up_interruptible(&p_conn->conn_wait);

	ITRACE_EXIT();
	return 0;
}
