/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/io.h>		/* struct iovec */
#include <asm/scatterlist.h>	/* struct scatterlist */
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_memory.h"
#include "iser_ofa.h"
#include "iser_global.h"
#include "iser_utils.h"
#include "iser_platform.h"
#include "iser_trace.h"

#define GFP_ORDER_TO_SIZE(gfp_order)	(0x01 << (gfp_order + PAGE_SHIFT))

/* Define this to zero pre-registered buffers when adding them to the cache. */
#define PRE_REGD_ZERO_ON_CACHE

static struct iser_regd_buf_t *
cache_get_pre_regd_buf(struct iser_regd_buf_cache *cache, size_t size);

static void cache_put_pre_regd_buf(struct iser_regd_buf_cache *cache,
				   struct iser_regd_buf_t *buf);
static void cache_init(struct iser_regd_buf_cache *cache);

static void cache_release(struct iser_regd_buf_cache *cache);
static struct iser_regd_buf_t *__iser_buffer_alloc(unsigned long size,
						   struct iser_conn_t *p_conn,
						   enum dma_data_direction dir);
static int __iser_buffer_free(struct iser_regd_buf_t *p_entry);

/**
 * Initialize the pre-registered buffer caches.
 */
void iser_init_regd_buf_caches(struct iser_conn_t *p_conn)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	cache_init(&p_conn->regd_cache);
	p_conn->cache_max = PRE_REGD_CACHE_MAX_DEFAULT;

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * Free the memory for the registered buffer caches.
 * (this also unregisters it).
 */
void iser_release_regd_buf_caches(struct iser_conn_t *p_conn)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	cache_release(&p_conn->regd_cache);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * Get a registered buffer.  If one is available in an appropriately
 * sized cache, use that, if not get a new buffer.
 */
struct iser_regd_buf_t *iser_buffer_alloc(unsigned long size,
					  struct iser_conn_t *p_conn,
					  enum dma_data_direction dir)
{
	struct iser_regd_buf_t *b;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	b = cache_get_pre_regd_buf(&p_conn->regd_cache, size);

	if (unlikely(!b)) {
		b = __iser_buffer_alloc(size, p_conn, dir);
		TRACE(TRACE_ISER_MEM, "%s: allocated new buf=%p, size=%lu\n",
		      __func__, b, size);
	} else {
		TRACE(TRACE_ISER_MEM, "%s: allocated cache=%p, "
		      " buf=%p, size=%lu\n",
		      __func__, &p_conn->regd_cache, b, size);

		/* Set the virtual buffer size to match the allocated
		 * size.  We may have pulled a larger buffer out of
		 * the cache. */
		b->virt_buf.size = size;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return b;
}

/**
 * Free a pre-registered buffer.  If it fits nicely into a cache, just
 * cache it, otherwise get rid of it.
 */
int iser_buffer_free(struct iser_regd_buf_t *buf)
{
	int ret = 0;
	struct iser_conn_t *p_conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	p_conn = buf->as.pre_regd.p_conn;

	if (p_conn->regd_cache.fill < p_conn->cache_max) {
		TRACE(TRACE_ISER_MEM, "%s: returning buf=%p to "
		      "cache=%p\n", __func__, buf,
		      &buf->as.pre_regd.p_conn->regd_cache);

		cache_put_pre_regd_buf(&buf->as.pre_regd.p_conn->regd_cache,
				       buf);

	} else {
		TRACE(TRACE_ISER_MEM, "%s: discarding buf=%p\n",
		      __func__, buf);
		ret = __iser_buffer_free(buf);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}


/*!
 * @brief
 *
 * @returns
 *
 * @notes
 *   allocates buffer descriptor from iser_global.regd_buf_mem_cache
 */
static struct iser_regd_buf_t *__iser_buffer_alloc(unsigned long size,
						   struct iser_conn_t *p_conn,
						   enum dma_data_direction dir)
{
	int err;
	struct iser_mem_t phys_mem;
	struct iser_regd_buf_t *p_entry = NULL;
	void *buff = NULL;

	ITRACE_ENTRY();

	p_entry = iser_alloc_regd_buf();
	if (p_entry == NULL) {
		IERROR("Failed to allocate buffer descriptor\n");
		return NULL;
	}
	TRACE(TRACE_ISER_MEM, "%s: allocated regd_buf_t=%p\n",
	      __func__, p_entry);

	buff = my_kmalloc(size, "iser_buffer_alloc");
	if (buff == NULL) {
		IERROR("Failed to allocate from kmalloc\n");
		return NULL;
	}

	p_entry->as.pre_regd.cache_size = size;
	TRACE(TRACE_ISER_MEM, "%s: Used kmalloc, size: %lu\n",
	      __func__, size);

	phys_mem.addr = (uint64_t) (unsigned long) buff;
	phys_mem.size = size;
	err = iser_register_phys_mem(p_conn, &phys_mem, 1, size,
				     DAT_MEM_PRIV_ALL_FLAG,
				     &p_entry->mem_reg, NULL, dir);
	if (err) {
		TRACE_ERROR("%s: iser_register_phys_mem failed\n",
			    __func__);
		iser_free_regd_buf(buff);
		ITRACE_EXIT();
		return NULL;
	}
	p_entry->virt_buf.p_buf = buff;
	p_entry->virt_buf.size = size;
	p_entry->virt_buf.type = ISER_BUF_TYPE_SINGLE;
	p_entry->type = ISER_PRE_REGD_BUFF;
	p_entry->as.pre_regd.p_conn = p_conn;
	atomic_set(&p_entry->ref_count, 0);


	ITRACE_EXIT();
	return p_entry;
}

static int __iser_buffer_free(struct iser_regd_buf_t *p_entry)
{
	int err;
	ITRACE_ENTRY();


	TRACE(TRACE_ISER_MEM, "%s: unregistering regd_buf_t=%p\n",
	      __func__, p_entry);
	err = iser_unregister_memory(&p_entry->mem_reg);
	if (err) {
		TRACE_ERROR("%s: iser_unregister_memory failed\n",
			    __func__);
		ITRACE_EXIT();
		return -1;
	}

	TRACE(TRACE_ISER_MEM, "%s: iser_buffer_free: %p\n",
	      __func__, p_entry->virt_buf.p_buf);
	my_kfree((void**) &p_entry->virt_buf.p_buf, "iser_buffer_free");

	TRACE(TRACE_ISER_MEM, "%s: freeing regd_buf_t=%p\n",
	      __func__, p_entry);
	iser_free_regd_buf(p_entry);

	ITRACE_EXIT();
	return 0;
}

/**
 * iser_regd_buff_reference - Increments the reference count of a registered buffer
 */
int iser_regd_buff_reference(struct iser_regd_buf_t *p_regd_buf)	//<! iSER registered buffer
{
	int new_val;

	atomic_inc(&p_regd_buf->ref_count);
	new_val = atomic_read(&p_regd_buf->ref_count);
	ITRACE(ISER_TRACE_REGD_BUFFER,
	       "Buffer: 0x%p ref_count: %d (after inc)\n", p_regd_buf,
	       new_val);

	return new_val;
}				/* iser_regd_buff_reference */

/**
 * iser_regd_buff_dereference - Decrements the reference count of a registered buffer
 *
 * returns true if reached zero
 */
int iser_regd_buff_dereference(struct iser_regd_buf_t *p_regd_buf)	//<! iSER registered buffer
{
	int new_val;

	ITRACE_ENTRY();
	atomic_dec(&p_regd_buf->ref_count);
	new_val = atomic_read(&p_regd_buf->ref_count);
	ITRACE(ISER_TRACE_REGD_BUFFER,
	       "Buffer: 0x%p ref_count: %d (after dec)\n", p_regd_buf,
	       new_val);
	if (new_val < 0) {
		IINFO
		    ("Negative value after dereferencing buffer: 0x%p ref_count: %d\n",
		     p_regd_buf, new_val);
		atomic_set(&p_regd_buf->ref_count, 0);
		new_val = 0;
	}

	ITRACE_EXIT();
	return new_val;
}				/* iser_regd_buff_dereference */


int iser_regd_buff_real_release(struct iser_regd_buf_t *p_regd_buf)
{
	int ret_val = 0;

	ITRACE_ENTRY();
	ITRACE(ISER_TRACE_REGD_BUFFER,
	       "Release now, regd.buffer:0x%p, type: %d\n", p_regd_buf,
	       p_regd_buf->type);

	switch (p_regd_buf->type) {
//     case ISER_REGD_BUF_POOL:
//         if (iser_buffer_pool_free(p_regd_buf) != 0) {
//             ITRACE(ISER_TRACE_ERRORS,
//                    "iser_buffer_pool_free failed! buffer: 0x%p, type: BUF_POOL(%d)\n",
//                    p_regd_buf,p_regd_buf->type);
//             ret_val = -1;
//             goto regd_buff_release_exit;
//         }
//         break;

	case ISER_PRE_REGD_BUFF:
		if (iser_buffer_free(p_regd_buf) != 0) {
			ITRACE(ISER_TRACE_ERRORS,
			       "failed to free buffer type: ISER_PRE_REGD_BUFF, buffer: 0x%p\n",
			       p_regd_buf);
			ret_val = -1;
			goto regd_buff_release_exit;
		}
		break;


	case ISER_REGD_BUF_USER:
		if (iser_unregister_memory(&p_regd_buf->mem_reg) != 0) {
			ITRACE(ISER_TRACE_ERRORS,
			       "iser_unregister_memory failed! buffer: 0x%p, type: BUF_USER(%d)\n",
			       p_regd_buf, p_regd_buf->type);
			ret_val = -1;
			goto regd_buff_release_exit;
		}
		// for user buffers iser_regd_buf_t is allocated from kmem_cache
		iser_free_regd_buf(p_regd_buf);
		break;

	case ISER_REGD_BUF_RDMA:
		if (iser_unregister_memory(&p_regd_buf->mem_reg) != 0) {
			ITRACE(ISER_TRACE_ERRORS,
			       "iser_unregister_memory failed! buffer: 0x%p, type: BUF_RDMA(%d)\n",
			       p_regd_buf, p_regd_buf->type);
			ret_val = -1;
			goto regd_buff_release_exit;
		}
		// for rdma buffers a static iser_regd_buf_t field of task descriptor is
		// used (as it is always a single one) - so no release necessary
		break;

	case ISER_REGD_BUF_UNUSED:	// ISER_REGD_BUF_UNUSED obviously should not be released
		break;
	default:
		IPANIC
		    ("Unexpected buffer type to release, buffer: 0x%p, type: %d\n",
		     p_regd_buf, p_regd_buf->type);
		break;
	}

      regd_buff_release_exit:
	ITRACE_EXIT();
	return ret_val;
}

/**
 * Decrements the reference count for the registered buffer &
 * releases it from the DTO
 *
 * returns 0 if success, -1 if error, *release_deferred is true if not released immediately
 */
int iser_regd_buff_release(struct iser_regd_buf_t *p_regd_buf,	//<! IN - registered buffer to release
			   int *release_deferred)	//<! OUT - true is release deferred
{
	int ret;
	ITRACE_ENTRY();

	if (release_deferred != NULL) {
		*release_deferred = 0;
	}
	if (atomic_read(&p_regd_buf->ref_count) == 0) {
		IINFO(" buffer: 0x%p ref_count: 0 on entry\n", p_regd_buf);
		ret = iser_regd_buff_real_release(p_regd_buf);
		ITRACE_EXIT();
		return ret;
	}
	// Decrement the ref_count & decide how to release the buffer according to its type & ref_count
	if (iser_regd_buff_dereference(p_regd_buf) == 0) {
		ret = iser_regd_buff_real_release(p_regd_buf);
		ITRACE_EXIT();
		return ret;
	} else {
		if (release_deferred != NULL) {
			*release_deferred =
			    atomic_read(&p_regd_buf->ref_count);
		}
		ITRACE(ISER_TRACE_REGD_BUFFER,
		       "Release deferred, regd.buffer: 0x%p, type: %d\n",
		       p_regd_buf, p_regd_buf->type);
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * Allocates a registered buffer descriptor
 */
struct iser_regd_buf_t *iser_alloc_regd_buf(void)
{
	struct iser_regd_buf_t *b;

	b = my_kmem_cache_alloc(iser_global.regd_buf_mem_cache,
	                        MEM_ALLOC_FLAG);
	if (!b)
		return NULL;

	/* Does this cache do this for us automatically? */
	memset(b, '\0', sizeof(*b));

	return b;
}

/**
 * Allocates a registered buffer descriptor
 */
void iser_free_regd_buf(struct iser_regd_buf_t *buf)
{
	my_kmem_cache_free(iser_global.regd_buf_mem_cache, buf);
}

/**
 * Gets the next pre-registered buffer from the given cache.
 *
 * \param cache The cache to get the buffer from.
 */
static struct iser_regd_buf_t *
cache_get_pre_regd_buf(struct iser_regd_buf_cache *cache, size_t size)
{
	long flags;
	struct list_head *node;
	struct iser_regd_buf_t *b, *ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	b = ret = NULL;

	spin_lock_irqsave(&cache->lock, flags);

	if (unlikely(cache->fill == 0)) {
		cache->misses += 1;
		goto out;
	}
	BUG_ON(list_empty(&cache->buf_list));

	list_for_each(node, &cache->buf_list) {
		b = list_entry(node, struct iser_regd_buf_t,
			       as.pre_regd.rc_next);

		if (b->as.pre_regd.cache_size >= size) {
			ret = b;
			break;
		}
	}

	if (ret) {
		TRACE(TRACE_ISER_MEM, "%s: retrieving regd_buf=%p, "
		      "size=%u, from cache=%p\n",
		      __func__, b, b->as.pre_regd.cache_size, cache);
		BUG_ON(b->as.pre_regd.cache_size !=
		       b->mem_reg.ibv_mr->length);
		list_del(&b->as.pre_regd.rc_next);
		cache->fill -= 1;
		cache->hits += 1;
	} else
		cache->misses += 1;


out:
	spin_unlock_irqrestore(&cache->lock, flags);

	TRACE(TRACE_ENTER_LEAVE, "Exiting %s\n", __func__);
	return ret;
}

/**
 * Put a buffer into the give cache.
 *
 * \param cache The cache to put the buffer into.
 * \param buf The buffer to add to the cache.
 */
static void cache_put_pre_regd_buf(struct iser_regd_buf_cache *cache,
				   struct iser_regd_buf_t *buf)
{
	long flags;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	TRACE(TRACE_ISER_MEM, "%s: cacheing regd_buf=%p in cache=%p\n",
	      __func__, buf, cache);

	if (buf->virt_buf.type != ISER_BUF_TYPE_SINGLE) {
		TRACE_ERROR("%s: attempting to cache non-single "
			    "buffer=%p type\n", __func__, buf);
		goto out;
	}

	if (buf->type != ISER_PRE_REGD_BUFF) {
		TRACE_ERROR("%s: attempting to cache a non-pre-registered "
			    "buffer=%p\n", __func__, buf);
		goto out;
	}

#if defined(PRE_REGD_ZERO_ON_CACHE)
	memset(buf->virt_buf.p_buf, '\0', buf->virt_buf.size);
#endif

	spin_lock_irqsave(&cache->lock, flags);
	list_add(&buf->as.pre_regd.rc_next, &cache->buf_list);
	cache->fill += 1;
	spin_unlock_irqrestore(&cache->lock, flags);

out:
	TRACE(TRACE_ENTER_LEAVE, "Exiting %s\n", __func__);
}

/**
 * Initialize a cache.
 */
static void cache_init(struct iser_regd_buf_cache *cache)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	cache->fill = 0;
	cache->misses = 0;
	cache->hits = 0;
	spin_lock_init(&cache->lock);
	INIT_LIST_HEAD(&cache->buf_list);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}


/**
 * Release the buffers in a given cache (and de-register them).
 */
static void cache_release(struct iser_regd_buf_cache *cache)
{
	long flags;
	struct iser_regd_buf_t *b;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	spin_lock_irqsave(&cache->lock, flags);

	while (!list_empty(&cache->buf_list)) {
		b = list_entry(cache->buf_list.next,
			       struct iser_regd_buf_t,
			       as.pre_regd.rc_next);
		list_del(&b->as.pre_regd.rc_next);

		spin_unlock_irqrestore(&cache->lock, flags);
		__iser_buffer_free(b);
		spin_lock_irqsave(&cache->lock, flags);
	}
	cache->fill = 0;
	cache->misses = 0;
	cache->hits = 0;

	spin_unlock_irqrestore(&cache->lock, flags);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/* vi: set tabstop=8: */
