/**
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file datamover/iser_ofa.c
 * \brief Contains definitions for the iSER/OFA interface
 * \author Ethan Burns
 * \date 2007-04-25
 */
#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif				/* !_POSIX_C_SOURCE */

#include "iser.h"
#include "iser_conn.h"
#include "iser_dto.h"
#include "iser_global.h"
#include "iser_memory.h"
#include "iser_ofa.h"
#include "iser_target.h"
#include "iser_trace.h"

/**
 * \brief Maximum value for the max CQ entries.
 * \note This is used when a device reports a max_cqe value that is
 *       unreasonably high and that they can not actually use.
 */
#define MAX_MAX_CQE	100

static int iser_handle_one_event(struct iser_adaptor_t *p_adaptor,
				 iser_cm_event_t *event);
static int iser_handle_one_wc(struct iser_conn_t *p_adaptor,
			      ib_wc_t *wc);
static int iser_handle_connect_request(struct iser_adaptor_t *p_adaptor,
				       iser_cm_event_t *event);
static int iser_handle_addr_resolved(struct iser_adaptor_t *p_adaptor,
				     iser_cm_event_t *event);
static int iser_handle_route_resolved(struct iser_adaptor_t *p_adaptor,
				      iser_cm_event_t *event);
static int iser_handle_established(struct iser_adaptor_t *p_adaptor,
				   iser_cm_event_t *event);
static int iser_handle_disconnect(struct iser_adaptor_t *p_adaptor,
				  iser_cm_event_t *event);
static int iser_handle_connection_error(struct iser_adaptor_t *p_adaptor,
                                        iser_cm_event_t *event);
static int convert_dto_to_ibv_sg(struct iser_dto_t *, ib_sge_t **, int *,
                                 unsigned int, unsigned int);
static void print_buffer(void *addr, unsigned long len,
			 unsigned long orig_offset);
static void print_dto_data(struct iser_dto_t *dto, unsigned long len);
static void iser_setup_send_wr(ib_send_wr_t *swr,
			       struct iser_dto_t *dto,
                               unsigned long);
static void trace_ibv_wc_opcode(const char *func, int opcode);
static int do_one_wr(struct iser_dto_t *, struct iser_conn_t *,
                     unsigned long, unsigned int, unsigned int);

#if defined(__KERNEL__)
static void
iser_print_ib_device_info(struct ib_device *dev);
#else
#define iser_print_ib_device_info(x) /* */
#endif /* __KERNEL__ */

/**
 * \brief Sets the work request wr_id field of the given work request
 *
 * \param x A work request structure.
 * \param y A pointer to the DTO structure for this request.
 * \note The wr_id field is used to retrieve the work request from a work
 *       completion.
 */
#define SET_WR_ID(x, y) ((x).wr_id = (unsigned long) (y))

/**
 * \brief Gets the DTO that a work completion is signaling completion of.
 * \param x A pointer to the work completion structure.
 */
#define GET_DTO_FROM_WC(x) ((struct iser_dto_t *) (unsigned long) (x->wr_id))

/**
 * \brief Creates an RDMA CM identifier, RDMA event channel and a protection
 *        domain for the given adaptor.
 * \return 0 on success, -1 on failure.
 */
int iser_ofa_init_adaptor(struct iser_adaptor_t *p_adaptor)
{
	int err;

	ITRACE_ENTRY();

	p_adaptor->event_chann = rdma_create_event_channel();
	if (!p_adaptor->event_chann) {
		TRACE_ERROR("%s: rdma_create_event_channel failed\n",
			    __func__);
		goto error;
	}

	err = iser_rdma_create_id(p_adaptor->event_chann, &p_adaptor->cm_id,
			     p_adaptor /* context */ , RDMA_PS_TCP);
	if (err != 0) {
		TRACE_ERROR("%s: rdma_create_id failed\n", __func__);
		goto error_id;
	}

	ITRACE_EXIT();
	return 0;

      error_id:
	rdma_destroy_event_channel(p_adaptor->event_chann);
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Destroys a RDMA CM Identifier, event channel and protection domain
 *        for the given adaptor.
 * \return 0 on success, -1 on failure.
 */
int iser_ofa_deinit_adaptor(struct iser_adaptor_t *p_adaptor)
{
	int err;
	int ret;

	ITRACE_ENTRY();

	ret = 0;
	err = iser_rdma_destroy_id(p_adaptor->cm_id);
	if (err != 0) {
		TRACE_ERROR("%s: rdma_destroy_id failed\n", __func__);
		ret = -1;
	}
	rdma_destroy_event_channel(p_adaptor->event_chann);

	ITRACE_EXIT();
	return ret;
}

/**
 * \brief Binds an address to the given adaptor and listens for new connection
 *        requests.
 *
 * \note This is equivilant to listening on a socket.
 *
 * \return 0 on success, -1 on failure.
 */
int iser_listen(struct iser_entity_t *p_entity, struct sockaddr *addr,
		size_t addr_len)
{
	int err;
	int i, succeeded;
	ITRACE_ENTRY();

	for (succeeded = i = 0; i < iser_global.num_adaptors; i++) {
		err = rdma_bind_addr(iser_global.adaptor[i].cm_id, addr);
		if (err) {
			TRACE_WARNING("%s: rdma_bind_addr failed\n",
				      __func__);
			continue;
		}

		err = rdma_listen(iser_global.adaptor[i].cm_id, 1);
		if (err) {
			TRACE_WARNING("%s: rdma_listen failed\n",
				      __func__);
			continue;
		}
		succeeded++;
	}
	if (succeeded == 0) {
		ITRACE_EXIT();
		return -1;
	}

	p_entity->addr = *addr;
	p_entity->addr_len = addr_len;
	ITRACE_EXIT();
	return 0;
}

/**
 * \brief frees all public service points for an iscsi entity.
 * \return 0 on success, -1 on failure.
 */
int iser_free_psp_by_entity(struct iser_entity_t *p_entity)
{
	ITRACE_ENTRY();
	TRACE_WARNING("unimplemented function\n");
	ITRACE_EXIT();

	return 0;
}

/**
 * \brief frees a public service point for an adaptor object.
 * 	itrace_entry();
 * 		itrace_exit();
 *
 * \return 0 on success, -1 on failure.
 */
int iser_free_psp_by_adaptor(struct iser_adaptor_t *p_adaptor)
{
	ITRACE_ENTRY();
	TRACE_WARNING("unimplemented function\n");
	ITRACE_EXIT();

	return 0;
}

/**
 * \brief Creates a command queue and a queue pair for the given connection.
 *
 * \return 0 on success, -1 on failure.
 */
int iser_create_cq_and_qp(struct iser_conn_t *p_conn,
			  unsigned int max_rdma_read_req_in,
			  unsigned int max_rdma_read_req_out)
{
	int err;
	unsigned int max_reqs;
	ib_qp_init_attr_t qp_init_attr;
	ib_device_attr_t device_attr;

	ITRACE_ENTRY();

	err = iser_ib_query_device(p_conn->cm_id, &device_attr);
	if (err < 0) {
		TRACE_ERROR("%s: ibv_device_query failed\n", __func__);
		goto error;
	}

	p_conn->max_sge_wr = device_attr.max_sge;
	p_conn->max_sge_rd = device_attr.max_sge_rd;
	p_conn->max_sge_send = SQ_SGE_MAX;
	p_conn->max_sge_recv = SQ_SGE_MAX;


	/* Post Recvs can only have a one element SG, aparantly... */
	p_conn->max_sge_recv = 1;

#if defined(OSC_IWARP)
	/* OSC iWARP only supports a single SGE. */
	p_conn->max_sge_wr = 1;
	p_conn->max_sge_rd = 1;
	p_conn->max_sge_send = 1;
	p_conn->max_sge_recv = 1;
#endif /* OSC_IWARP */

	TRACE(TRACE_ISER, "%s: device_attrmax_sge=%d\n",
	      __func__, device_attr.max_sge);
	TRACE(TRACE_ISER, "%s: p_conn->max_sge_wr=%d, p_conn->max_sge_rd=%d\n",
	      __func__, p_conn->max_sge_wr, p_conn->max_sge_rd);
#if !defined(__KERNEL__) && !defined(OSC_IWARP)
	p_conn->comp_chan = ibv_create_comp_channel(p_conn->cm_id->verbs);
	if (!p_conn->comp_chan) {
		TRACE_ERROR("%s: ibv_create_comp_channel failed\n", __func__);
		goto error;
	}
#endif /* !__KERNEL__ && !OSC_IWARP */

	p_conn->cq = iser_ib_create_cq(p_conn,
				       device_attr.max_cqe > MAX_MAX_CQE
	                               	? MAX_MAX_CQE
	                               	: device_attr.max_cqe);
	if (!p_conn->cq) {
		TRACE_ERROR("ibv_create_cq failed\n");
		goto error_cq;
	}

	max_reqs = max_rdma_read_req_in + max_rdma_read_req_out;
	if (max_reqs > device_attr.max_qp_wr)
		max_reqs = device_attr.max_qp_wr;

	memset(&qp_init_attr, 0, sizeof(qp_init_attr));
	qp_init_attr.send_cq = p_conn->cq;
	qp_init_attr.recv_cq = p_conn->cq;
	qp_init_attr.cap.max_send_wr = max_reqs;
	p_conn->max_rdma_read_req_out = max_reqs;
	qp_init_attr.cap.max_recv_wr = max_reqs;
	p_conn->max_rdma_read_req_in = max_reqs;
	qp_init_attr.cap.max_send_sge = p_conn->max_sge_send;
	qp_init_attr.cap.max_recv_sge = p_conn->max_sge_recv;
	qp_init_attr.qp_type = IBV_QPT_RC;
	/* only generate completion queue entries if requested */
#if defined(__KERNEL__)
	qp_init_attr.sq_sig_type = IB_SIGNAL_ALL_WR;
	qp_init_attr.event_handler = __iser_qp_event_handler;
#endif /* !__KERNEL__ */

	TRACE(TRACE_ISER,
	      "\n%s:\n\tsend_cq=%p"
	      "\n\trecv_cq=%p"
	      "\n\tmax_send_wr=%d"
	      "\n\tmax_recv_wr=%d"
	      "\n\tmax_send_sge=%d"
	      "\n\tmax_recv_sge=%d\n\n",
	      __func__,
	      qp_init_attr.send_cq,
	      qp_init_attr.recv_cq,
	      qp_init_attr.cap.max_send_wr,
	      qp_init_attr.cap.max_recv_wr,
	      qp_init_attr.cap.max_send_sge,
	      qp_init_attr.cap.max_recv_sge);

	err = rdma_create_qp(p_conn->cm_id, p_conn->pd, &qp_init_attr);
	if (err) {
		TRACE_ERROR("%s: ibv_create_qp failed\n", __func__);
		goto error_qp;
	}
	p_conn->qp = p_conn->cm_id->qp;

	ITRACE_EXIT();
	return 0;
      error_qp:
	err = iser_ib_destroy_cq(p_conn->cq);
	if (err)
		TRACE_ERROR("%s: iser_ib_desroy_cq failed\n", __func__);
	p_conn->cq = NULL;
      error_cq:
#if !defined(__KERNEL__) && !defined(OSC_IWARP)
	err = ibv_destroy_comp_channel(p_conn->comp_chan);
	if (err)
		TRACE_ERROR("%s: ibv_destroy_comp_channel failed\n", __func__);
#endif /* !__KERNEL__ && !OSC_IWARP */
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Connects an End-Point.
 * \return 0 on success, -1 on failure.
 */
int iser_connect(struct iser_conn_t *p_conn,
		 struct sockaddr_in *dst_addr,
		 struct sockaddr_in *src_addr)
{
	int ret;
	ITRACE_ENTRY();

	atomic_set(&p_conn->conn_done, 0);

	p_conn->cm_id = p_conn->p_adaptor->cm_id;

	ret =
	    rdma_resolve_addr(p_conn->cm_id, (struct sockaddr *) src_addr,
			      (struct sockaddr *) dst_addr, 0);
	if (ret) {
		TRACE_ERROR("%s: rdma_resolve_addr failed\n", __func__);
		goto error;
	}

	wait_event_interruptible_timeout(p_conn->conn_wait,
	                                 atomic_read(&p_conn->conn_done),
	                                 CONN_TIMEOUT);

	if (!atomic_read(&p_conn->conn_done)) {
		TRACE_ERROR("%s: connection not completed\n", __func__);
		goto error;
	}

	if (!p_conn->cq || !p_conn->qp) {
		TRACE_ERROR("%s: cq or qp not properly created\n", __func__);
		goto error;
	}

	ITRACE_EXIT();
	return 0;
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Disconnects a connection
 * \return 0 on success, -1 on failure.
 */
int iser_disconnect(struct iser_conn_t *p_conn)
{
	int err;
	ITRACE_ENTRY();

	err = rdma_disconnect(p_conn->cm_id);
	if (err) {
		TRACE_ERROR("%s: rdma_disconnect failed\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Releases a command queue and queue pair for a given connection.
 *
 * \return 0 on success, -1 on failure
 */
int iser_destroy_cq_and_qp(struct iser_conn_t *p_conn)
{
	int err;
	ib_qp_t *qp;
	ib_cq_t *cq;

	ITRACE_ENTRY();

	qp = p_conn->qp;
	p_conn->qp = NULL;
	cq = p_conn->cq;
	p_conn->cq = NULL;

#if !defined(__KERNEL__)
	iser_kill_cq_thread(p_conn);
#else
	destroy_workqueue(p_conn->cq_work_q);
#endif /* !__KERNEL__ */

	if (qp) {
		err = iser_ib_destroy_qp(qp);
		if (err)
			TRACE_ERROR("%s: iser_ib_destroy_qp failed\n",
				    __func__);

	}

	if (cq) {
		err = iser_ib_destroy_cq(cq);
		if (err)
			TRACE_ERROR("%s: iser_ib_desroy_cq failed\n",
				    __func__);
	}

#if !defined(__KERNEL__) && !defined(OSC_IWARP)
	err = ibv_destroy_comp_channel(p_conn->comp_chan);
	if (err)
		TRACE_ERROR("%s: ibv_destroy_comp_channel failed\n", __func__);
#endif /* !__KERNEL__ && !OSC_IWARP */

	ITRACE_EXIT();

	return 0;
}

/**
 * \brief Accepts an incoming connection request.
 * \return 0 on success, -1 on failure.
 */
int iser_accept(struct iser_conn_t *p_conn,
		struct iser_data_buf_t *p_priv_data)
{
	int err;

	struct rdma_conn_param conn_param = {
		.responder_resources = 1,
		.initiator_depth = 1,
		.retry_count = 5,
	};

	ITRACE_ENTRY();

	err = rdma_accept(p_conn->cm_id, &conn_param);
	if (err) {
		TRACE_ERROR("%s: rdma_accept failed\n", __func__);
		goto error;
	}

	iser_print_ib_device_info(p_conn->cm_id->device);

	ITRACE_EXIT();
	return 0;
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Rejects an incoming connection request.
 * \return 0 on success, -1 on failure.
 */
int iser_reject(struct iser_conn_t *p_conn)
{
	int err;
	ITRACE_ENTRY();

	err = rdma_reject(p_conn->cm_id, NULL, 0);
	if (err) {
		TRACE_ERROR("%s: rdma_reject failed\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Spawns new instances of the event threads.
 * \return 0 on success, -1 on failure.
 */
int iser_spawn_event_thread(struct iser_adaptor_t *p_adaptor)
{
	int err;

	ITRACE_ENTRY();

	init_MUTEX_LOCKED(&p_adaptor->startstop_sem);

	err = spawn_thread(&p_adaptor->event_thrd_handle,
			   iser_event_handler_thread, p_adaptor, 0);
	if (err) {
		TRACE_ERROR("%s: spawn_thread failed\n", __func__);
		ITRACE_EXIT();
		return -1;
	}
	down(&p_adaptor->startstop_sem);

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Kills the event threads.
 */
void iser_kill_event_thread(struct iser_adaptor_t *p_adaptor)
{
	ITRACE_ENTRY();

	kill_thread(p_adaptor->event_thrd_handle);
	wait_thread(p_adaptor->event_thrd_handle);
	TRACE(TRACE_ISER, "%s: event thread terminated\n", __func__);

	ITRACE_EXIT();
}

/**
 * \brief Send kill signal to the current thread.
 */
void iser_suicide_thread(void)
{
	ITRACE_ENTRY();
	kill_self();
	ITRACE_EXIT();
}

/**
 * \brief Event handling thread.
 * \param void_adaptor A void* that contains the address of the iser_adaptor_t
 *                     that this thread is handling events for.
 * \return 0 on success, -1 on failure.
 */
thread_ret iser_event_handler_thread(void *void_adaptor)
{
	int err;
	struct iser_adaptor_t *p_adaptor = void_adaptor;
	iser_cm_event_t *event;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	ITRACE_ENTRY();

	unh_daemonize(__func__);

	/* allow the creating thread to continue */
	up(&p_adaptor->startstop_sem);

	for (;;) {
		err = iser_rdma_get_cm_event(p_adaptor->event_chann, &event);
		if (err) {
			TRACE_ERROR("%s: rdma_get_cm_event failed\n",
				    __func__);
			goto error;
		}

		err = iser_handle_one_event(p_adaptor, event);
		if (err) {
			TRACE_ERROR("%s: iser_handle_one_event failed\n"
			            "\tevent=%d\n",
				    __func__, event->event);
		}

		if (event->event != RDMA_CM_EVENT_REJECTED)
			iser_rdma_ack_cm_event(event);
	}

	ITRACE_EXIT();
	return (thread_ret) 0;

      error:
	ITRACE_EXIT();
	return (thread_ret) - 1;
}

/**
 * \brief Register a memory buffer.
 * \return 0 on success, -1 on failure.
 */
int iser_register_phys_mem(struct iser_conn_t *p_conn,
			   struct iser_mem_t *p_phys_vec,
			   unsigned int phys_vec_len,
			   unsigned long total_sz,
			   unsigned int priv_flags,
			   struct iser_mem_handles_t *p_mem_reg,
			   uint32_t * p_rmr_ctxt,
                           enum dma_data_direction dir)
{
	ITRACE_ENTRY();
#if defined(OSC_IWARP)
	BUG_ON(phys_vec_len != 1);
#endif				/* OSC_IWARP */
#if !defined(__KERNEL__)
	errno = 0;
#endif /* !__KERNEL__ */

	p_mem_reg->ibv_mr = iser_ib_reg_mr(p_conn,
				           p_phys_vec,
	                                   phys_vec_len,
	                                   total_sz,
	                                   priv_flags,
	                                   dir);
	if (!p_mem_reg->ibv_mr) {
		TRACE_ERROR("%s: iser_ib_reg_mr failed for "
		            "addr=%#lx, len=%lu\n", __func__,
		            (unsigned long) p_phys_vec->addr, total_sz);
#if !defined(__KERNEL__)
		perror("iser_register_phys_mem: iser_ib_reg_mr()");
#endif /* !__KERNEL__ */
		ITRACE_EXIT();
		return -1;
	}
	p_mem_reg->bytes = total_sz;

	TRACE(TRACE_ISER_MEM, "%s: registered new stag=%u\n"
	      "\taddr=%p, len=%lu, ibv_mr=%p\n", __func__,
	      p_mem_reg->ibv_mr->lkey,
	      p_mem_reg->ibv_mr->addr,
	      (unsigned long) p_mem_reg->ibv_mr->length,
	      p_mem_reg->ibv_mr);

	if (p_rmr_ctxt)
		*p_rmr_ctxt = p_mem_reg->ibv_mr->lkey;

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Unregister previously registered memory.
 * \return 0 on success, -1 on failure.
 */
int iser_unregister_memory(struct iser_mem_handles_t *mem_reg)
{
	int err;

	ITRACE_ENTRY();

	TRACE(TRACE_ISER_MEM, "%s: unregistering mem_reg->ibv_mr=%p\n",
	      __func__, mem_reg->ibv_mr);

	err = iser_ib_dereg_mr(mem_reg->ibv_mr);
	if (err) {
		TRACE_ERROR("%s: iser_ib_dereg_mr failed\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Post a receive buffer.
 * \return 0 on success, -1 on failure.
 */
int iser_post_recv(struct iser_dto_t *p_recv_dto)
{
	int err;
	ib_recv_wr_t rwr;
	ib_recv_wr_t *bad_wr;
	struct iser_conn_t *p_conn;

	ITRACE_ENTRY();

	p_conn = p_recv_dto->p_conn;
	if (p_conn == NULL) {
		TRACE_ERROR("%s: NULL p_conn in dto: 0x%p\n",
			    __func__, p_recv_dto);
		ITRACE_EXIT();
		return -1;
	}

	TRACE(TRACE_ISER_MEM, "%s: posting recv for dto=%p\n",
	      __func__, p_recv_dto);

	atomic_set(&p_recv_dto->wrs, 1);

	memset(&rwr, 0, sizeof(rwr));
	SET_WR_ID(rwr, p_recv_dto);

	convert_dto_to_ibv_sg(p_recv_dto, &rwr.sg_list, &rwr.num_sge,
	                      0, p_recv_dto->regd_vector_len);

	err = iser_ib_post_recv(p_conn->qp, &rwr, &bad_wr);
	if (err) {
		TRACE_ERROR("%s: iser_ib_post_recv failed=%d\n",
		            __func__, err);
		goto error;
	}
	my_kfree((void**) &rwr.sg_list, "*sg_list");

	TRACE(TRACE_ISER_MEM, "%s: posted receive for DTO=%p\n",
	      __func__, p_recv_dto);

	ITRACE_EXIT();
	return 0;

      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Initiate a DTO operation (PostSend, RDMA-W,
 *        RDMA_RMAX_REGD_BUF_VECTOR_LEN).
 * \return 0 on success, -1 on failure.
 */
int iser_start_dto(struct iser_dto_t *dto)
{
	struct iser_conn_t *p_conn;
	unsigned int wrs, vec_start, vec_end;
	unsigned long offs;
	unsigned int max_sge;

	ITRACE_ENTRY();

	/* Retrieve connection */
	p_conn = dto->p_conn;
	if (p_conn == NULL) {
		TRACE_ERROR("%s: NULL p_conn in dto: 0x%p\n", __func__,
			    dto);
		ITRACE_EXIT();
		return -1;
	}

	BUG_ON(dto->type != ISER_DTO_SEND
	       && dto->type != ISER_DTO_RDMA_W
	       && dto->type != ISER_DTO_RDMA_R);

	if (dto->type == ISER_DTO_SEND) {
		if (dto->p_conn->max_sge_send < dto->regd_vector_len)
			iser_merge_sges(dto);
		max_sge = dto->p_conn->max_sge_send;
	} else if (dto->type == ISER_DTO_RDMA_W)
		max_sge = dto->p_conn->max_sge_wr;
	else if (dto->type == ISER_DTO_RDMA_R)
		max_sge = dto->p_conn->max_sge_rd;
	else {
		max_sge = 1;
		BUG();
	}

	wrs = dto->regd_vector_len / max_sge;
	if (dto->regd_vector_len % max_sge!= 0)
		wrs += 1;

	atomic_set(&dto->wrs, wrs);

	offs = 0;
	vec_start = 0;
	vec_end = max_sge;
	TRACE(TRACE_ISER, "%s max_sge=%d\n", __func__, max_sge);

	iser_dto_get(dto);
	do {
		if (vec_end > dto->regd_vector_len)
			vec_end = dto->regd_vector_len;

		offs += do_one_wr(dto, dto->p_conn, offs, vec_start, vec_end);
		/*
		 * Notice that we can't use dto->p_conn after here, since
		 * after the final WR is sent, the DTO may be removed from the
		 * connection by this point (since it could actually be done
		 * now).
		 */

		atomic_inc(&p_conn->post_send_buf_count);

		ITRACE(ISER_TRACE_SEND_DTO, "Started SEND DTO, p_dto: 0x%p, "
		       "IOV len: %d, Notify: %d, Posted sends: %d\n",
		       dto, vec_end - vec_start, dto->notify_enable,
		       atomic_read(&p_conn->post_send_buf_count));

		wrs -= 1;
		vec_start = vec_end + 1;
		vec_end += max_sge;
	} while (wrs);

	iser_dto_put(dto);

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Handles a single RDMA event.
 * \param p_adaptor A pointer to the adaptor that the event arrived on.
 * \param event The RDMA event.
 * \return 0 on success, !0 on error (An error message is printed).
 */
static int iser_handle_one_event(struct iser_adaptor_t *p_adaptor,
				 iser_cm_event_t *event)
{
	int ret = -1;

	ITRACE_ENTRY();

	switch (event->event) {
	case RDMA_CM_EVENT_ADDR_RESOLVED:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_ADDR_RESOLVED "
		      "event cm_id=%p\n", event->id);
		ret = iser_handle_addr_resolved(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_ADDR_ERROR:
		TRACE_WARNING("Got RDMA_CM_EVENT_ADDR_ERROR event "
		              "cm_id=%p\n", event->id);
		ret = iser_handle_connection_error(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_ROUTE_RESOLVED:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_ROUTE_RESOLVED "
		      "event cm_id=%p\n", event->id);
		ret = iser_handle_route_resolved(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_ROUTE_ERROR:
		TRACE_WARNING("Got RDMA_CM_EVENT_ROUTE_ERROR "
		              "event cm_id=%p\n", event->id);
		ret = iser_handle_connection_error(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_CONNECT_REQUEST:
		TRACE(TRACE_RDMA_EVENTS,
		      "Got RDMA_CM_EVENT_CONNECT_REQUEST event "
		      "cm_id=%p\n", event->id);
		ret = iser_handle_connect_request(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_CONNECT_RESPONSE:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_CONNECT_RESPONSE "
		      "event cm_id=%p\n", event->id);
		break;
	case RDMA_CM_EVENT_CONNECT_ERROR:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_CONNECT_ERROR "
		      "event cm_id=%p\n", event->id);
		ret = iser_handle_connection_error(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_UNREACHABLE:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_UNREACHABLE event "
		      "cm_id=%p\n", event->id);
		ret = iser_handle_connection_error(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_REJECTED:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_REJECTED event "
		      "cm_id=%p\n", event->id);
		ret = iser_handle_connection_error(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_ESTABLISHED:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_ESTABLISHED event "
		      "cm_id=%p\n", event->id);
		ret = iser_handle_established(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_DISCONNECTED:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_DISCONNECTED event "
		      "cm_id=%p\n", event->id);
		ret = iser_handle_disconnect(p_adaptor, event);
		break;
	case RDMA_CM_EVENT_DEVICE_REMOVAL:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_DEVICE_REMOVAL "
		      "event cm_id=%p\n", event->id);
		break;
#if !defined(__KERNEL__)
	case RDMA_CM_EVENT_MULTICAST_JOIN:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_MULTICAST_JOIN "
		      "event cm_id=%p\n", event->id);
		break;
	case RDMA_CM_EVENT_MULTICAST_ERROR:
		TRACE(TRACE_RDMA_EVENTS, "Got RDMA_CM_EVENT_MULTICAST_ERROR "
		      "event cm_id=%p\n", event->id);
		break;
#endif /* !__KERNEL__ */
	default:
		TRACE_ERROR("%s: Invalid event number=%d cm_id=%p\n",
		            __func__, event->event, event->id);
		break;
	}

	ITRACE_EXIT();
	return ret;
}

/**
 * \brief Handles a single work completion.
 * \param p_adaptor The adaptor that the completion happened on.
 * \param wc The work completion structure.
 * \return 0 on success, !0 on failure.
 */
static int iser_handle_one_wc(struct iser_conn_t *p_conn,
			      ib_wc_t *wc)
{
	struct iser_dto_t *dto;

	ITRACE_ENTRY();

	dto = GET_DTO_FROM_WC(wc);

	TRACE(TRACE_ISER, "%s: One WR for DTO [%p, %s] is complete "
	      "(%d remaining) cm_id=%p\n", __func__, dto,
	      dto_type_str(dto), atomic_read(&dto->wrs) - 1,
	      p_conn->cm_id);

	if (wc->status == IBV_WC_WR_FLUSH_ERR) {
		/* a FLUSH is not an error, we use the error completion, but
 		 * are quiet about it. */
		TRACE(TRACE_ISER, "%s: DTO %p FLUSHED\n", __func__, dto);
		iser_dto_one_wc(dto, 1, 0);
		ITRACE_EXIT();
		return 0;
	}

	if (wc->status != IBV_WC_SUCCESS) {
		TRACE_ERROR("Unsuccessful work completion for DTO %p: %s\n",
			    dto, iser_wc_status_string(wc->status));
		iser_dto_one_wc(dto, 1, 0);
		goto error;
	}

	trace_ibv_wc_opcode(__func__, wc->opcode);

	switch (wc->opcode) {
	case IBV_WC_RDMA_WRITE:
	case IBV_WC_SEND:
		TRACE(TRACE_ISER, "%s: sent %u bytes\n",
		      __func__, wc->byte_len);
		print_dto_data(dto, wc->byte_len);
		iser_dto_one_wc(dto, 0, wc->byte_len);
		break;
	case IBV_WC_RDMA_READ:
	case IBV_WC_RECV:
		TRACE(TRACE_ISER, "%s: received %u bytes\n",
		      __func__, wc->byte_len);
		print_dto_data(dto, wc->byte_len);
		iser_dto_one_wc(dto, 0, wc->byte_len);
		break;
	default:
		TRACE_WARNING("%s: bad opcode=%d\n", __func__, wc->opcode);
		break;
	}

	ITRACE_EXIT();
	return 0;
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Handles a RDMA connection request event.
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_connect_request(struct iser_adaptor_t *p_adaptor,
				       iser_cm_event_t *event)
{
	int err;
	struct sockaddr *local_addr;
	struct sockaddr_in *remote_addr;
	struct iser_conn_t *p_conn;
	char name[ISER_OBJECT_NAME_SIZE];

	ITRACE_ENTRY();

	remote_addr = (struct sockaddr_in *) &event->id->route.addr.dst_addr;
	snprintf(name, ISER_OBJECT_NAME_SIZE, "%d.%d.%d.%d",
		 NIPQUAD(remote_addr->sin_addr));

	p_conn = iser_adaptor_new_conn(p_adaptor, name);
	if (p_conn == NULL)
		goto error;

	p_conn->cm_id = event->id;

	p_conn->pd = iser_ib_alloc_pd(p_conn->cm_id);
	if (!p_conn->pd) {
		TRACE_ERROR("%s: iser_ib_alloc_pd failed\n", __func__);
		goto error_pd;
	}

	local_addr = &event->id->route.addr.src_addr;
	print_buffer(local_addr, sizeof(*local_addr), 0);
	err = iser_target_handle_connection_request(p_conn, local_addr);
	if (err) {
		TRACE_ERROR
		    ("iser_target_handle_connection_request failed\n");
		err = rdma_reject(event->id, NULL, 0);
		if (err)
			TRACE_ERROR("%s: rdma_reject failed\n", __func__);
		goto error_handle;
	}

	ITRACE_EXIT();
	return 0;

      error_handle:
	err = iser_ib_dealloc_pd(p_conn->pd);
	if (err)
		TRACE_ERROR("%s: iser_ib_dealloc_pd failed\n", __func__);
      error_pd:
	iser_conn_destroy(p_conn);
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Handles a RDMA address resolved event.
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_addr_resolved(struct iser_adaptor_t *p_adaptor,
				     iser_cm_event_t *event)
{
	int err;
	struct iser_conn_t *p_conn;

	ITRACE_ENTRY();

	p_conn = iser_adaptor_find_connection(p_adaptor, event->id);
	if (p_conn == NULL) {
		TRACE_ERROR("Connection not found, EP handle= %p\n",
			    event->id);
		goto error;
	}
	TRACE(TRACE_ISER, "Found connection %p\n", p_conn->cm_id);

	iser_print_ib_device_info(p_conn->cm_id->device);

	p_conn->pd = iser_ib_alloc_pd(p_conn->cm_id);
	if (!p_conn->pd) {
		TRACE_ERROR("%s: iser_ib_alloc_pd failed\n", __func__);
		goto error;
	}

	err = iser_create_cq_and_qp(p_conn,
				    /* receive queue depth (+4 for login). */
				    p_conn->initial_post_recv_bufs_num + 4,
				    /* send queue depth */
				    ISER_MAX_OUTSTAND_CMDS);
	if (err) {
		TRACE_ERROR("%s: iser_create_cq_and_qp failed\n",
			    __func__);
		goto error_cq_and_qp;
	}

	err = rdma_resolve_route(event->id, 0);
	if (err) {
		TRACE_ERROR("%s: rdma_resolve_route failed\n", __func__);
		goto error_resolve;
	}

	ITRACE_EXIT();
	return 0;

      error_resolve:
	err = iser_destroy_cq_and_qp(p_conn);
	if (err) {
		TRACE_ERROR("%s: iser_destroy_cq_and_qp failed\n",
			    __func__);
	}
      error_cq_and_qp:
	err = iser_ib_dealloc_pd(p_conn->pd);
	if (err)
		TRACE_ERROR("%s: iser_ib_dealloc_pd failed\n", __func__);
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Handles a RDMA route resolved event.
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_route_resolved(struct iser_adaptor_t *p_adaptor,
				      iser_cm_event_t *event)
{
	int err;
	struct iser_conn_t *p_conn;
	struct rdma_conn_param conn_param = {
		.responder_resources = 1,
		.initiator_depth = 1,
		.retry_count = 5,
	};

	ITRACE_ENTRY();

	p_conn = iser_adaptor_find_connection(p_adaptor, event->id);
	if (p_conn == NULL) {
		TRACE_ERROR("Connection not found, EP handle= %p\n",
			    event->id);
		goto error;
	}
	TRACE(TRACE_ISER, "Found connection %p\n", p_conn->cm_id);

	err = rdma_connect(event->id, &conn_param);
	if (err) {
		TRACE_ERROR("%s: rdma_connect failed\n", __func__);
		goto error;
	}

	atomic_set(&p_conn->conn_done, 1);
	wake_up_interruptible(&p_conn->conn_wait);

	ITRACE_EXIT();
	return 0;

      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Handles a RDMA connection estabished event.
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_established(struct iser_adaptor_t *p_adaptor,
				   iser_cm_event_t *event)
{
	struct iser_conn_t *p_conn;

	ITRACE_ENTRY();

	p_conn = iser_adaptor_find_connection(p_adaptor, event->id);
	if (p_conn == NULL) {
		TRACE_ERROR("Connection not found, EP handle= %p\n",
			    event->id);
		goto error;
	}
	TRACE(TRACE_ISER, "Found connection %p\n", p_conn->cm_id);

#if !defined(__KERNEL__)
	if (iser_spawn_cq_thread(p_conn) != 0) {
		TRACE_ERROR("%s: iser_spawn_cq_thread failed\n", __func__);
		return -1;
	}
#else
	iser_begin_request_notify(p_conn);
#endif /* !__KERNEL__ */

	ISER_CONN_SET_STATE(p_conn, ISER_CONN_UP);

	atomic_inc(&p_conn->p_adaptor->in_callback);
	TRACE(TRACE_ISER, "p_conn %p with iscsi_conn_h %p\n",
	      p_conn, p_conn->iscsi_conn_h);
	p_conn->p_entity->ext_api_cb.connection_establish_notify(p_conn->iscsi_conn_h, 1);
	atomic_dec(&p_conn->p_adaptor->in_callback);


	ITRACE_EXIT();
	return 0;
      error:
	ITRACE_EXIT();
	return -1;
}

/**
 * \brief Handles a RDMA disconnect event.
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_disconnect(struct iser_adaptor_t *p_adaptor,
				  iser_cm_event_t *event)
{
	struct iser_conn_t *p_conn;

	ITRACE_ENTRY();

	p_conn = iser_adaptor_find_connection(p_adaptor, event->id);
	if (p_conn == NULL) {
		TRACE_ERROR("Connection not found, EP handle= %p\n",
			    event->id);
		ITRACE_EXIT();
		return -1;
	}

	TRACE(TRACE_ISER, "Found connection %p\n", p_conn->cm_id);
	if (atomic_read(&p_conn->state) == ISER_CONN_PENDING) {
		ISER_CONN_SET_STATE(p_conn, ISER_CONN_DOWN);
		atomic_inc(&p_adaptor->in_callback);
		p_conn->p_entity->ext_api_cb.
		    connection_establish_notify(p_conn->iscsi_conn_h, 0);
		atomic_dec(&p_adaptor->in_callback);
		iser_conn_destroy(p_conn);
	} else {
		if (atomic_read(&p_conn->state) == ISER_CONN_UP) {
			ISER_CONN_SET_STATE(p_conn, ISER_CONN_ASYNC_TERM);
			TRACE(TRACE_ISER, "Conn. 0x%p is being "
			      "terminated asynchronously\n", p_conn);
		}

		/*
		 * Complete the termination process if possible
		 * (no eventspending)
		 */
		iser_complete_connection_termination(p_conn);
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Handles a RDMA connection error event.
 * \notes This is called from: RDMA_CM_EVENT_ADDR_ERROR,
 *                             RDMA_CM_EVENT_ROUTE_ERROR,
 *                             RDMA_CM_EVENT_CONNECT_ERROR,
 *                             RDMA_CM_EVENT_UNREACHABLE
 *                         and RDMA_CM_EVENT_REJECTED
 * \param p_adaptor The adaptor that caused the event.
 * \param event The event structure for this event.
 * \return 0 on success, !0 on error (an error message is printed).
 */
static int iser_handle_connection_error(struct iser_adaptor_t *p_adaptor,
                                        iser_cm_event_t *event)
{
	struct iser_conn_t *p_conn;

	ITRACE_ENTRY();

	p_conn = iser_adaptor_find_connection(p_adaptor, event->id);
	if (p_conn == NULL) {
		TRACE_ERROR("Connection not found, EP handle= %p\n",
			    event->id);
		ITRACE_EXIT();
		return -1;
	}

	if (atomic_read(&p_conn->state) == ISER_CONN_PENDING) {
		ISER_CONN_SET_STATE(p_conn, ISER_CONN_DOWN);

		atomic_inc(&p_adaptor->in_callback);
		p_conn->p_entity->ext_api_cb.
		  connection_establish_notify(p_conn->iscsi_conn_h, 0);
		atomic_dec(&p_adaptor->in_callback);

		atomic_set(&p_conn->conn_done, 1);
		wake_up_interruptible(&p_conn->conn_wait);

	} else {
		TRACE_ERROR("Event unexpected in current conn state\n");
		ITRACE_EXIT();
		return -1;
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Converts a Data Transfer Object to an infiniband scatter gather list.
 * \param dto The data transfer object.
 * \param sg_list Used to return the scatter gather list.
 * \param num_sge Used to return the length of the scatter gather list.
 * \param vec_offset The vector entry offset.  This field is used to construct
 *                   multiple work requests from the same dto.
 * \return 0 on success, !0 on error.
 *
 * \note The pointer returned in sg_list is allocated with my_kmalloc() and
 *       must be freed by the caller with my_kfree().
 */
static int convert_dto_to_ibv_sg(struct iser_dto_t *dto,
				 ib_sge_t **sg_list, int *num_sge,
                                 unsigned int vec_entry_start,
                                 unsigned int vec_entry_end)
{
	int i, j;
	ITRACE_ENTRY();

	*num_sge = vec_entry_end - vec_entry_start;
	*sg_list = my_kmalloc((*num_sge) * sizeof(**sg_list), "*sg_list");
	BUG_ON(!*sg_list);

	TRACE(TRACE_ISER_MEM, "dto=%p, vec_entry_start=%u, vec_entry_end=%d, "
	      "num_sge=%d\n", dto, vec_entry_start, vec_entry_end, *num_sge);
	j = 0;
	for (i = vec_entry_start; i < vec_entry_end; i++) {
		if(dto->regd[j] == NULL) {
			TRACE_WARNING("dto->regd[%d] == NULL\n", j);
			BUG();
		}
		(*sg_list)[j].addr =
		    (unsigned long) dto->regd[i]->virt_buf.p_buf;
		(*sg_list)[j].length = dto->regd[i]->virt_buf.size;
		(*sg_list)[j].lkey = dto->regd[i]->mem_reg.ibv_mr->lkey;
		TRACE(TRACE_ISER_MEM, "%s: regd[%d]=%p "
		      "sg_list[%d].addr=%p, "
		      "sg_list[%d].length=%u, "
		      "sg_list[%d].lkey=%u\n\n",
		      __func__, i, dto->regd[i],
		      j, (void*) (unsigned long) (*sg_list)[j].addr,
		      j, (*sg_list)[j].length,
		      j,  (*sg_list)[j].lkey);
		j += 1;
	}

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Prints the regd buffers of a DTO object.
 * \param dto The DTO to print.
 * \param len The number of bytes to print.
 */
static void print_dto_data(struct iser_dto_t *dto, unsigned long len)
{
	unsigned long bytes = 0;
	unsigned long size;
	int i;


	if (!(TRACE_TEST(TRACE_BUF) || TRACE_TEST(TRACE_ISER_MEM)))
		return;

	BUG_ON(!dto);

	for (i = 0; i < dto->regd_vector_len; i++) {
		if (bytes >= len)
			break;

		if (!dto->regd[i])
			goto bug;
		else if (!dto->regd[i]->virt_buf.p_buf)
			goto bug;

		size = dto->regd[i]->virt_buf.size;
		size = size + bytes > len ? len - bytes : size;
		TRACE(TRACE_ISER_MEM, "%s: dto->regd[%d]->virt_buf.p_buf=%p\n",
		      __func__, i, dto->regd[i]->virt_buf.p_buf);
		print_buffer(dto->regd[i]->virt_buf.p_buf, size, bytes);
		bytes += dto->regd[i]->virt_buf.size;
	}
	iser_printf("\n%4ld bytes\n\n", bytes);

	return;
 bug:
	TRACE_ERROR("%s: i=%d\n", __func__, i);
	BUG();
}

/**
 * \brief Prints a memory buffer's contents.
 *
 * \param addr The starting address of the memory buffer.
 * \param len The length in bytes to print.
 * \param orig_offset The number for the starting byte (for printing scatter
 *                    lists).
 */
static void print_buffer(void *addr, unsigned long len,
			 unsigned long orig_offset)
{
	int i;

	if (TRACE_TEST(TRACE_BUF) || TRACE_TEST(TRACE_ISER_MEM)) {
		for (i = 0; i < len; i++) {
			if ((i + orig_offset) == 0
			     || (i + orig_offset) % 16 == 0)
				iser_printf("\n%4ld: ", i + orig_offset);
			iser_printf("%02x ", ((uint8_t *) addr)[i]);
		}
	}
}

/**
 * \brief Sets up a send work request given a data transfer object.
 * \param swr A pointer to the send work request to setup.
 * \param dto A pointer to the dto.
 * \param offset The byte-offset of the transfer for this WR.
 */
static void iser_setup_send_wr(ib_send_wr_t *swr,
			       struct iser_dto_t *dto,
                               unsigned long offset)
{
	switch (dto->type) {
	case ISER_DTO_SEND:
		swr->opcode = IBV_WR_SEND;
		break;
	case ISER_DTO_RDMA_W:
		swr->opcode = IBV_WR_RDMA_WRITE;
		swr->wr.rdma.rkey = dto->as.rdma.remote.rkey;
		swr->wr.rdma.remote_addr =
		    (unsigned long) dto->as.rdma.remote.addr
		    + offset;
		TRACE(TRACE_ISER, "%s: rkey=%u, rto=%llu\n", __func__,
		      swr->wr.rdma.rkey,
		      (unsigned long long) swr->wr.rdma.remote_addr);
		break;
	case ISER_DTO_RDMA_R:
		swr->opcode = IBV_WR_RDMA_READ;
		swr->wr.rdma.rkey = dto->as.rdma.remote.rkey;
		swr->wr.rdma.remote_addr =
		    (unsigned long) dto->as.rdma.remote.addr
		    + offset;
		TRACE(TRACE_ISER, "%s: rkey=%u, to=%llu\n", __func__,
		      swr->wr.rdma.rkey, (unsigned long long) swr->wr.rdma.remote_addr);
		break;
	default:
		BUG();
	}

	swr->send_flags = IBV_SEND_SIGNALED;
}

/**
 * \brief Prints and IBV_WC opcode name.
 * \param func The calling function name
 * \param opcode The opcode number to print the name for.
 */
static void trace_ibv_wc_opcode(const char *func, int opcode)
{
	switch (opcode) {
	case IBV_WC_RDMA_WRITE:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_RDMA_WRITE\n", func);
		break;
	case IBV_WC_SEND:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_SEND\n", func);
		break;
	case IBV_WC_RDMA_READ:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_RDMA_READ\n", func);
		break;
	case IBV_WC_RECV:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_RECV\n", func);
		break;
	case IBV_WC_COMP_SWAP:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_COMP_SWAP\n", func);
		break;
	case IBV_WC_FETCH_ADD:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_FETCH_ADD\n", func);
		break;
	case IBV_WC_BIND_MW:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_BIND_MW\n", func);
		break;
	case IBV_WC_RECV_RDMA_WITH_IMM:
		TRACE(TRACE_ISER, "%s: Got IBV_WC_RECV_RDMA_WITH_IMM\n",
		      func);
		break;
	}
}

/**
 * Polls the CQ for the given connection for events.
 *
 * Returns the number of events handled on success, or a negative number
 * on error.
 */
int iser_do_poll_cq(struct iser_conn_t *p_conn)
{
	int err;
	int count;
	ib_wc_t wc;

	count = 0;

	/*
	 * This happens because the cq_handler_thread is started
	 * before we post receive buffers (which happens after the
	 * MPARequest/MPAReply exchange).  OSC crashes if we
	 * ibv_poll_cq() while not receive buffers are posted, just
	 * re-loop untill there are some.
	 */
	if (atomic_read(&p_conn->post_recv_buf_count) <= 0) {
		TRACE_WARNING("%s: No receive buffers posted for conn=%p\n",
		              __func__, p_conn);
		schedule();
		goto exit;
	}

	for ( ; ; ) {
		if (!p_conn->cq) {
			/* the CQ could have been destroyed, don't poll if it
 			 * was. */
			break;
		}

		/* opcode that we don't handle */
		wc.opcode = IBV_WC_FETCH_ADD;

		err = iser_ib_poll_cq(p_conn->cq, 1, &wc);
		if (err < 0) {
			TRACE_ERROR("%s: iser_ib_poll_cq failed\n", __func__);
			goto error;
		} else if (err == 0) {
			TRACE(TRACE_ISER, "%s: empty poll\n", __func__);
			goto exit;
		} else if (err > 1) {
			TRACE_WARNING("%s: iser_ib_poll_cq returned %d\n",
			              __func__, err);
		}
		count += 1;

		err = iser_handle_one_wc(p_conn, &wc);
		if (err) {
			TRACE_ERROR("%s: iser_handle_one_wc failed\n",
				    __func__);
		}
	}
 exit:
	return count;
 error:
	return -1;
}

/**
 * Starts one work request for a DTO.
 * A DTO can be multiple work requests since we may be limited by how many
 * sge's we can put in a work request.
 *
 * \param p_dto The DTO.
 * \param p_conn The connection this DTO belonged to.  Notice how, I say
 *               *belonged* to in past tense, this DTO could be completed and
 *               already removed from the connection before this function
 *               really finishes.
 * \param offset The byte offset of the transfer.
 * \param vec_entry_start The index of the first vector entry to use for this
 *                        WR.
 * \param vec_entry_end The index of the last vector entry to use for this WR.
 * \return 0 on success, !0 on error.
 *
 * \note If this is called with the final WR for the DTO, the DTO may be
 *       removed from the connection by the time this returns.  The caller may
 *       also need to iser_dto_get() and iser_dto_put() the DTO if they plan
 *       on using a reference to it *after* this call returns.
 */
static int do_one_wr(struct iser_dto_t *p_dto,
                     struct iser_conn_t *p_conn,
                     unsigned long offset,
                     unsigned int vec_entry_start,
                     unsigned int vec_entry_end)
{
	int err = 0;
	ib_send_wr_t swr, *bad_wr;

	ITRACE_ENTRY();
	TRACE(TRACE_ISER, "%s: offset=%lu, vec_entry_start=%u "
	      " vec_entry_end=%u\n", __func__, offset,
	      vec_entry_start, vec_entry_end);

	memset(&swr, 0, sizeof(swr));
	SET_WR_ID(swr, p_dto);

	iser_setup_send_wr(&swr, p_dto, offset);
	convert_dto_to_ibv_sg(p_dto, &swr.sg_list, &swr.num_sge,
	                      vec_entry_start, vec_entry_end);

	err = iser_ib_post_send(p_conn->qp, &swr, &bad_wr);
	if (err) {
		TRACE_ERROR("Failed to start SEND DTO, err=%d dto: 0x%p\n",
		            err, p_dto);
		ITRACE_EXIT();
		return -1;
	}
	my_kfree((void**) &swr.sg_list, "*sg_list");

	TRACE(TRACE_ISER, "%s: One WR for DTO [%p, %s] is posted "
	      "(%d posted) cm_id=%p\n", __func__, p_dto,
	      dto_type_str(p_dto), atomic_read(&p_dto->wrs),
	      p_conn->cm_id);

	ITRACE_EXIT();
	return 0;
}


#if defined(__KERNEL__)
/*
 * Print information about the given ib_device structure.
 */
static void
iser_print_ib_device_info(struct ib_device *dev)
{
	if (!dev)
		TRACE_ERROR("%s: device is null\n", __func__);
	else {
		printk(KERN_INFO
		       "%20s = %p\n" /* dev */
		       "%20s = [%s]\n" /* name */
		       "%20s = %p\n" /* dma_device */
		       "%20s = %p\n" /* alloc_pd */
		       "%20s = %p\n" /* dealloc_pd */
		       "%20s = %p\n" /* create_qp */
		       "%20s = %p\n" /* destroy_qp */
		       "%20s = %p\n" /* post_send */
		       "%20s = %p\n" /* post_recv */
		       "%20s = %p\n" /* create_cq */
		       "%20s = %p\n" /* destroy_cq */
		       "%20s = %p\n" /* poll_cq */
		       "%20s = %p\n" /* req_notify_cq */
		       "%20s = %p\n" /* reg_phys_mr */
		       "%20s = %p\n" /* dereg_mr */
		       "%20s = %p\n" /* dma_ops */
		       "%20s = %d (%s)\n" /* reg_state */
		       "%20s = [%s]\n" /* node_desc */,
		       "device", dev,
		       "name", dev->name,
		       "dma_device", dev->dma_device,
		       "alloc_pd", dev->alloc_pd,
		       "dealloc_pd", dev->dealloc_pd,
		       "create_qp", dev->create_qp,
		       "destroy_qp", dev->destroy_qp,
		       "post_send", dev->post_send,
		       "post_recv", dev->post_recv,
		       "create_cq", dev->create_cq,
		       "destroy_cq", dev->destroy_cq,
		       "poll_cq", dev->poll_cq,
		       "req_notify_cq", dev->req_notify_cq,
		       "reg_phys_mr", dev->reg_phys_mr,
		       "dereg_mr", dev->dereg_mr,
		       "dma_ops", dev->dma_ops,
		       "reg_state", dev->reg_state,
		       dev->reg_state == IB_DEV_UNINITIALIZED
		       ? "IB_DEV_UNINITIALIZED"
		       : (dev->reg_state == IB_DEV_REGISTERED
		          ? "IB_DEV_REGISTERED"
		          : (dev->reg_state == IB_DEV_UNREGISTERED
		             ? "IB_DEV_UNREGISTERED"
		             : "<UNKNOWN>")),
		       "node_desc", dev->node_desc);
	}
}
#endif /* __KERNEL__*/

/* vi: set tabstop=8 textwidth=78: */
