/**
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file datamover/iser_ofa_kernel.c
 * \brief Contains kernel-space implementsions of the user-space OFA
 *        API.
 * \author Ethan Burns
 * \date 2007-08-06
 */

#define _POSIX_C_SOURCE 200112L

#include <asm/atomic.h>
#include <linux/list.h>
#include <linux/wait.h>

#include "iser_ofa.h"

#include "../common/kernel_workqueue.h"
#include "../common/debug.h"

/**
 * \brief True if event_channel is initialized, else false.
 */
static atomic_t event_channel_ref = ATOMIC_INIT(0);

/**
 * \brief The RDMA event channel.  There is only one...
 */
static struct rdma_event_channel event_channel;

/**
 * \brief Handles incoming rdma events.
 */
static int rdma_event_handler_func(struct rdma_cm_id *id,
                                   struct rdma_cm_event *event)
{
	iser_cm_event_t *iser_event;
	long flags;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	iser_event = my_kmalloc(sizeof(*iser_event), "iser_event");
	if (!iser_event) {
		TRACE(TRACE_DEBUG, "%s: my_kmalloc failed\n", __func__);
		TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
		return -ENOMEM;
	}

	iser_event->rdma_cm_event = event;
	iser_event->event = event->event;
	iser_event->id = id;

	TRACE(TRACE_DEBUG, "%s: new event=%p\n", __func__, iser_event);

	spin_lock_irqsave(&event_channel.lock, flags);
		list_add(&iser_event->queue, &event_channel.queue);
		atomic_inc(&event_channel.num_events);
		iser_event->channel = &event_channel;
	spin_unlock_irqrestore(&event_channel.lock, flags);

	TRACE(TRACE_DEBUG, "%s: waking up channel->waitq=%p\n",
	      __func__, &event_channel.waitq);
	wake_up_interruptible(&event_channel.waitq);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/**
 * \breif Creates an RDMA event channel.  This is a queue of connection
 *        events.
 */
struct rdma_event_channel *rdma_create_event_channel(void)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	if (atomic_add_return(1, &event_channel_ref) == 1) {
		TRACE(TRACE_ISER, "%s: initializing event_channel\n",
		      __func__);

		spin_lock_init(&event_channel.lock);
		INIT_LIST_HEAD(&event_channel.queue);
		atomic_set(&event_channel.num_events, 0);
		init_waitqueue_head(&event_channel.waitq);
		event_channel.handler = rdma_event_handler_func;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return &event_channel;
}

/**
 * \brief Destroys an rdma_event_channel structure.
 */
void rdma_destroy_event_channel(struct rdma_event_channel *channel)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	atomic_dec(&event_channel_ref);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Gets a connection event from the event channel.
 */
int iser_rdma_get_cm_event(struct rdma_event_channel *channel,
                           iser_cm_event_t **event)
{
	long flags;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	TRACE(TRACE_DEBUG, "%s: waiting on channel->waitq=%p\n",
	      __func__, &channel->waitq);
	wait_event_interruptible(channel->waitq,
	                         atomic_read(&channel->num_events) != 0);
	TRACE(TRACE_DEBUG, "%s: woke up after wait on channel->waitq=%p\n",
	      __func__, &channel->waitq);

	if (atomic_read(&channel->num_events) == 0) {
		*event = NULL;
		TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
		return -EAGAIN;
	}

	spin_lock_irqsave(&channel->lock, flags);
		*event = container_of(channel->queue.next,
		                      iser_cm_event_t,
		                      queue);
	spin_unlock_irqrestore(&channel->lock, flags);

	TRACE(TRACE_DEBUG, "%s giving caller event=%p, num=%d\n", __func__,
	      *event, (*event)->event);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/**
 * \brief Acks the RDMA CM event.
 *
 * This function removes the event from the event queue and frees
 * the event structure.
 */
int iser_rdma_ack_cm_event(iser_cm_event_t *iser_event)
{
	unsigned long flags;
	TRACE(TRACE_ENTER_LEAVE, "Entering %s, iser_event=%p, event=%d\n",
	      __func__, iser_event, iser_event->event);

	spin_lock_irqsave(&iser_event->channel->lock, flags);
		list_del(&iser_event->queue);
		atomic_dec(&iser_event->channel->num_events);
	spin_unlock_irqrestore(&iser_event->channel->lock, flags);

	my_kfree((void**) &iser_event, "iser_event");

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/**
 * The worker function for deregistering memory regions.  The reasoning
 * behind this is that memory deregistration may sleep, we need to make
 * sure that it is freed in a non-atomic, interrupt-enabled context.
 */
static void iser_mr_dereg_work_func(work_func_arg *work)
{
	int i;
	ib_mr_t *mr;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	mr = container_of(GET_WORK_STRUCT_P(work), ib_mr_t, dereg_work);

	TRACE(TRACE_ISER_MEM, "%s: deregistering mr=%p, lkey=%u, "
	      "addr=%p+%lu\n", __func__, mr, mr->lkey,
	      (void*) (unsigned long) mr->addr, (unsigned long) mr->length);

	ib_dereg_mr(mr->ib_mr);

	for (i = 0; i < mr->phys_array_len; i += 1) {
		ib_dma_unmap_single(mr->ib_device,
		                    mr->phys_array[i].addr,
		                    mr->phys_array[i].size,
		                    mr->dma_dir);
	}

	BUG_ON(!mr->p_conn);
	if (atomic_dec_and_test(&mr->p_conn->n_regd_mr))
		wake_up(&mr->p_conn->regd_mr_wait);

	TRACE(TRACE_ISER, "%s: %d remaining MRs\n",
	      __func__, atomic_read(&mr->p_conn->n_regd_mr));

	my_kfree((void**) &mr, "mr");

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Registers a memory region.
 */
ib_mr_t *iser_ib_reg_mr(struct iser_conn_t *p_conn,
                        struct iser_mem_t *vec,
                        unsigned int vec_len,
                        size_t length,
                        int access,
                        enum dma_data_direction dir)
{
	ib_mr_t *mr;
	u64 iova_start;
	void *virt_addr;
	unsigned long old_addr, old_size;
	int i;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	mr = my_kmalloc(sizeof(*mr)
	                + sizeof(struct ib_phys_buf) * vec_len,
	                "mr");
	if (!mr) {
		TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
		return NULL;
	}

	/* these two are for debugging only */
	old_addr = (unsigned long) vec[0].addr;
	old_size = (unsigned long) vec[0].size;

	virt_addr = (void*) (unsigned long) vec[0].addr;
	vec[0].addr &= PAGE_MASK;
	iova_start = vec[0].addr;
	vec[0].size += (unsigned long) virt_addr & ~PAGE_MASK;

	TRACE(TRACE_ISER_MEM, "%s: virt_addr=%#lx\n"
	      "\tORIGINAL vec[0].addr=%#lx, ORIGINAL vec[0].size=%lu\n"
	      "\tALIGNED vec[0].addr=%#lx, ALIGNED vec[0].size=%lu\n",
	      __func__,
	      (unsigned long) virt_addr,
	      old_addr, old_size,
	      (unsigned long) vec[0].addr,
	      (unsigned long) vec[0].size);

	/* DMA map memory */
	mr->phys_array_len = vec_len;
	for (i = 0; i < vec_len; i += 1) {
		mr->phys_array[i].addr =
		        ib_dma_map_single(p_conn->pd->device,
		                          (void*) (unsigned long) vec[i].addr,
		                          vec[i].size,
		                          dir);
		mr->phys_array[i].size = vec[i].size;
		TRACE(TRACE_ISER_MEM, "%s: dma mapped [%i] %p len=%lu -> "
		      "%p len=%lu\n",
		      __func__, i,
		      (void*) (unsigned long) vec[i].addr,
		      (unsigned long) vec[i].size,
		      (void*) (unsigned long) mr->phys_array[i].addr,
		      (unsigned long) mr->phys_array[i].size);
	}
	mr->dma_dir = dir;
	mr->ib_device = p_conn->pd->device;


	/*
	 *
	 * CHECKS FROM THE IW_CXGB3 DRIVER
	 *
	 */
	if ((iova_start & ~PAGE_MASK)
	    != (mr->phys_array[0].addr & ~PAGE_MASK)) {
		TRACE_ERROR("cxgb3 will return EINVAL because we "
		            "\"don't have enough alignment\"\n");
	}
	if (mr->phys_array_len > 1
	    && ((mr->phys_array[0].addr
	         + mr->phys_array[0].size) & ~PAGE_MASK)) {
		TRACE_ERROR("cxgb3 will return EINVAL because the end "
		            "of the first buffer is not page "
		            "aligned\n");
	}
	for (i = 0 ; i < mr->phys_array_len ; i += 1) {
		if (i != 0 && mr->phys_array[i].addr & ~PAGE_MASK) {
			TRACE_ERROR("cxgb3 will return EINVAL from "
				    "build_phys_page_list because the "
			            "[%d] buffer is not page aligend\n", i);
		}
		if (i != 0 && i != mr->phys_array_len - 1
		    && (mr->phys_array[i].size & ~PAGE_MASK)) {
			TRACE_ERROR("cxgb3 will return EINVAL from "
			            "build_phys_page_list because the "
			            "[%d] buffer does not end on a "
			            "page aligned addr\n", i);
		}
	}


	mr->ib_mr = ib_reg_phys_mr(p_conn->pd,
	                           &mr->phys_array[0],
	                           mr->phys_array_len,
	                           access,
	                           &iova_start);
	if (IS_ERR(mr->ib_mr)) {
		TRACE_ERROR("%s: ib_reg_phys_mr failed=%lu\n", __func__,
		            (unsigned long) -PTR_ERR(mr->ib_mr));
		my_kfree((void**) &mr, "mr");
		TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
		return NULL;
	}

	TRACE(TRACE_ISER_MEM, "\tvirt_addr=%p, stag=%#x\n",
	      virt_addr, mr->ib_mr->lkey);

	mr->p_conn = p_conn;
	atomic_inc(&mr->p_conn->n_regd_mr);
	mr->lkey = mr->ib_mr->lkey;
	mr->rkey = mr->ib_mr->rkey;
	mr->addr = virt_addr;
	mr->length = length;
	KERNEL_INIT_WORK(&mr->dereg_work, iser_mr_dereg_work_func);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return mr;
}

/**
 * Deregisters a memory region.  The actual de-registration is done as
 * deferred work since it requires a sleepable context.
 */
int iser_ib_dereg_mr(ib_mr_t *mr)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	schedule_work(&mr->dereg_work);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/**
 * \brief This event handler is called when there is an event on the CQ.
 *        We just scheds the iser_cq_work_func() for the given
 *        connection since events need to be handled with interrupts
 *        enabled.
 */
void __iser_cq_completion_handler(struct ib_cq *cq, void *context)
{
	struct iser_conn_t *p_conn = context;

	if (p_conn->cq)
		queue_work(p_conn->cq_work_q, &p_conn->cq_work);
}

/**
 * \brief This work is schedueled when an event may be ready on the CQ.
 */
void iser_cq_work_func(work_func_arg *work)
{
	int err;
	int reqflags;
	struct iser_conn_t *p_conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	p_conn = container_of(GET_WORK_STRUCT_P(work), struct iser_conn_t, cq_work);
	reqflags = IB_CQ_NEXT_COMP;
#if 0
	/* this is not supported in older kernels */
	reqflags |= IB_CQ_REPORT_MISSED_EVENTS;
#endif

	err = ib_req_notify_cq(p_conn->cq, reqflags);
	if (err < 0)
		TRACE_ERROR("%s: ib_req_notify_cq failed\n", __func__);

	iser_do_poll_cq(p_conn);

#if 0
	/* this is not supported in older kernels */
	else if (err > 0)
		iser_do_poll_cq((struct iser_conn_t *) context);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief A do-nothing function that is called when a CQ event happens.
 */
void __iser_cq_event_handler(struct ib_event *event, void *context)
{

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief A do-nothing function that is called when a QP event happens.
 */
void __iser_qp_event_handler(struct ib_event *event, void *context)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/* vi: set tabstop=8 textwidth=72: */
