/**
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file datamover/iser_ofa_kernel.h
 * \brief Contains declarations and inline definitions for the
 *        kernel-space OFA API.
 * \author Ethan Burns
 * \date 2007-08-06
 */

#define _POSIX_C_SOURCE 200112L

#if !defined(__KERNEL__)
#error This can only be compiled from with in the linux kernel.
#endif /* !__KERNEL__ */

#include <linux/types.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/version.h>

#include "../common/debug.h"

#define iser_printf(x...) printk(x)

/* define user-space access flags in terms of kernel-space access flags */
#define IBV_ACCESS_LOCAL_WRITE IB_ACCESS_LOCAL_WRITE
#define IBV_ACCESS_REMOTE_WRITE IB_ACCESS_REMOTE_WRITE
#define IBV_ACCESS_REMOTE_READ IB_ACCESS_REMOTE_READ
#define IBV_ACCESS_REMOTE_ATOMIC IB_ACCESS_REMOTE_ATOMIC
#define IBV_ACCESS_MW_BIND IB_ACCESS_MW_BIND

#define IBV_QPT_RC IB_QPT_RC

#define IBV_WC_SUCCESS IB_WC_SUCCESS
#define IBV_WC_FAILURE IB_WC_FAILURE
#define IBV_WC_RDMA_WRITE IB_WC_RDMA_WRITE
#define IBV_WC_RDMA_READ IB_WC_RDMA_READ
#define IBV_WC_SEND IB_WC_SEND
#define IBV_WC_RECV IB_WC_RECV
#define IBV_WC_COMP_SWAP IB_WC_COMP_SWAP
#define IBV_WC_FETCH_ADD IB_WC_FETCH_ADD
#define IBV_WC_BIND_MW IB_WC_BIND_MW
#define IBV_WC_RECV_RDMA_WITH_IMM IB_WC_RECV_RDMA_WITH_IMM

#define IBV_WR_RDMA_WRITE IB_WR_RDMA_WRITE
#define IBV_WR_WRITE_WITH_IMM IB_WR_WRITE_WITH_IMM
#define IBV_WR_SEND IB_WR_SEND
#define IBV_WR_SEND_WITH_IMM IB_WR_SEND_WITH_IMM
#define IBV_WR_RDMA_READ IB_WR_RDMA_READ

#define IBV_WC_WR_FLUSH_ERR IB_WC_WR_FLUSH_ERR

#define IBV_SEND_SIGNALED IB_SEND_SIGNALED;

/* RDMA event channel functionality */
struct rdma_event_channel {
	spinlock_t lock;
	struct list_head queue;
	atomic_t num_events;
	wait_queue_head_t waitq;
	rdma_cm_event_handler handler;
};

struct rdma_event_channel *rdma_create_event_channel(void);

ib_mr_t *iser_ib_reg_mr(struct iser_conn_t *p_conn,
                        struct iser_mem_t *vec,
                        unsigned int vec_len,
                        size_t length,
                        int access,
                        enum dma_data_direction dir);

int iser_ib_dereg_mr(ib_mr_t *mr);

void rdma_destroy_event_channel(struct rdma_event_channel *channel);

int iser_rdma_get_cm_event(struct rdma_event_channel *channel,
                           iser_cm_event_t **event);

int iser_rdma_ack_cm_event(iser_cm_event_t *event);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
void iser_cq_work_func(struct work_struct *);
#else
void iser_cq_work_func(void *);
#endif /* >= 2.6.20 */

void __iser_cq_completion_handler(struct ib_cq *cq, void *context);

void __iser_cq_event_handler(struct ib_event *event, void *context);

void __iser_qp_event_handler(struct ib_event *event, void *context);

/** \brief Creates an rdma_cm_id */
static inline int
iser_rdma_create_id(struct rdma_event_channel *channel,
                    struct rdma_cm_id **id, void *context,
                    enum rdma_port_space ps)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	*id = rdma_create_id(channel->handler, context, ps);
	if (IS_ERR(*id)) {
		TRACE_ERROR("%s: rdma_create_id failed=%lu\n",
		            __func__, (unsigned long) -PTR_ERR(*id));
		*id = NULL;
		return -1;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/** \brief Destroys an rdma_cm_id structure. */
static inline int iser_rdma_destroy_id(struct rdma_cm_id *id)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	rdma_destroy_id(id);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return 0;
}

/** \brief Gets device attributes */
static inline int
iser_ib_query_device(struct rdma_cm_id *id,
                     ib_device_attr_t *device_attr)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	ret = ib_query_device(id->device, device_attr);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Creates a command queue */
static inline ib_cq_t *
iser_ib_create_cq(struct iser_conn_t *p_conn, int cqe)
{
	ib_cq_t *ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_create_cq(p_conn->cm_id->device,
	                   __iser_cq_completion_handler,
	                   __iser_cq_event_handler,
	                   p_conn,
	                   cqe, 0);
	/*
	 * Do not request notifications untill later.  There is no
	 * synchronization between WCs and connection events, we don't
	 * want to get a WC untill after we are in established.
	 */

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Destroys a command queue */
static inline int iser_ib_destroy_cq(ib_cq_t *cq)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_destroy_cq(cq);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Polls a command queue.  */
static inline int iser_ib_poll_cq(ib_cq_t *cq, int num_entries,
                                  ib_wc_t *wc)
{
	int ret;

	//TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_poll_cq(cq, num_entries, wc);

	//TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Destroys a queue pair. */
static inline int iser_ib_destroy_qp(ib_qp_t *qp)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_destroy_qp(qp);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Posts a receive buffer. */
static inline int iser_ib_post_recv(ib_qp_t *qp,
                                    ib_recv_wr_t *recv_wr,
                                    ib_recv_wr_t **bad_recv_wr)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	TRACE(TRACE_DEBUG, "%s: qp->device->post_recv=%p\n",
	      __func__, qp->device->post_recv);
	if(!qp->device->post_recv) {
		TRACE_ERROR("%s: post_recv function is NULL\n",
		            __func__);

		/* null terminate just incase the device struct is
		 * filled with jibberish. */
		qp->device->name[IB_DEVICE_NAME_MAX - 1] = '\0';
		TRACE(TRACE_DEBUG, "%s: qp->device->name=[%s]\n",
		      __func__, qp->device->name);

		ret = -1;
	} else {
		ret = ib_post_recv(qp, recv_wr, bad_recv_wr);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Posts a send buffer. */
static inline int iser_ib_post_send(ib_qp_t *qp,
                                    ib_send_wr_t *send_wr,
                                    ib_send_wr_t **bad_send_wr)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_post_send(qp, send_wr, bad_send_wr);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Allocate a protection domain. */
static inline ib_pd_t *iser_ib_alloc_pd(struct rdma_cm_id *id)
{
	ib_pd_t *ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	TRACE(TRACE_ISER, "%s: calling ibv_alloc_pd(%p);\n", __func__,
	      id->device);

	ret = ib_alloc_pd(id->device);
	if (IS_ERR(ret)) {
		TRACE_ERROR("%s: ib_alloc_pd failed=%lu\n",
		            __func__, (unsigned long) -PTR_ERR(ret));
		ret = NULL;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Deallocates a protection domain. */
static inline int iser_ib_dealloc_pd(ib_pd_t *pd)
{
	int ret;

	TRACE(TRACE_ENTER_LEAVE, "Entering  %s\n", __func__);

	ret = ib_dealloc_pd(pd);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return ret;
}

/** \brief Begin requesting work completions on the given connection */
static inline void iser_begin_request_notify(struct iser_conn_t *p_conn)
{
	/*
	 * just call the cq work function, this will poll for any
	 * events we already got, then will request more notifications.
	 */
	iser_cq_work_func(&p_conn->cq_work);
}

/** \brief Returns a printable string for the status enum */
static inline const char *iser_wc_status_string(int status)
{
	switch(status) {
        case IB_WC_SUCCESS:
		return "IB_WC_SUCCESS";
        case IB_WC_LOC_LEN_ERR:
		return "IB_WC_LOC_LEN_ERR";
        case IB_WC_LOC_QP_OP_ERR:
		return "IB_WC_LOC_QP_OP_ERR";
        case IB_WC_LOC_EEC_OP_ERR:
		return "IB_WC_LOC_EEC_OP_ERR";
        case IB_WC_LOC_PROT_ERR:
		return "IB_WC_LOC_PROT_ERR";
        case IB_WC_WR_FLUSH_ERR:
		return "IB_WC_WR_FLUSH_ERR";
        case IB_WC_MW_BIND_ERR:
		return "IB_WC_MW_BIND_ERR";
        case IB_WC_BAD_RESP_ERR:
		return "IB_WC_BAD_RESP_ERR";
        case IB_WC_LOC_ACCESS_ERR:
		return "IB_WC_LOC_ACCESS_ERR";
        case IB_WC_REM_INV_REQ_ERR:
		return "IB_WC_REM_INV_REQ_ERR";
        case IB_WC_REM_ACCESS_ERR:
		return "IB_WC_REM_ACCESS_ERR";
        case IB_WC_REM_OP_ERR:
		return "IB_WC_REM_OP_ERR";
        case IB_WC_RETRY_EXC_ERR:
		return "IB_WC_RETRY_EXC_ERR";
        case IB_WC_RNR_RETRY_EXC_ERR:
		return "IB_WC_RNR_RETRY_EXC_ERR";
        case IB_WC_LOC_RDD_VIOL_ERR:
		return "IB_WC_LOC_RDD_VIOL_ERR";
        case IB_WC_REM_INV_RD_REQ_ERR:
		return "IB_WC_REM_INV_RD_REQ_ERR";
        case IB_WC_REM_ABORT_ERR:
		return "IB_WC_REM_ABORT_ERR";
        case IB_WC_INV_EECN_ERR:
		return "IB_WC_INV_EECN_ERR";
        case IB_WC_INV_EEC_STATE_ERR:
		return "IB_WC_INV_EEC_STATE_ERR";
        case IB_WC_FATAL_ERR:
		return "IB_WC_FATAL_ERR";
        case IB_WC_RESP_TIMEOUT_ERR:
		return "IB_WC_RESP_TIMEOUT_ERR";
        case IB_WC_GENERAL_ERR:
		return "IB_WC_GENERAL_ERR";
	default:
		return "UNKNOWN";
	}
}

/* vi: set tabstop=8 textwidth=72: */
