/**
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file datamover/iser_ofa_user.c
 * \brief Contains definitions for using OFA user-space API functions.
 * \author Ethan Burns
 * \date 2007-09-01
 */
#if !defined(__KERNEL__)

#include "iser_ofa.h"
#include "iser_trace.h"

static thread_ret iser_cq_handler_thread(void *void_conn);

/**
 * \brief Spawns a thread to poll the command queue for a given connection.
 * \param p_conn The given conneciton.
 * \return 0 on success, !0 on failure.
 */
int iser_spawn_cq_thread(struct iser_conn_t *p_conn)
{
	int err;
	ITRACE_ENTRY();

	init_MUTEX_LOCKED(&p_conn->startstop_sem);

	err = spawn_thread(&p_conn->cq_thrd_handle,
			   iser_cq_handler_thread, p_conn, 0);
	if (err) {
		TRACE_ERROR("%s: spawn_thread failed\n", __func__);
		ITRACE_EXIT();
		return -1;
	}
	down(&p_conn->startstop_sem);

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Kills the command queue polling thread for the given connection.
 * \param p_conn The given conneciton.
 */
void iser_kill_cq_thread(struct iser_conn_t *p_conn)
{
	thread_handle thread;

	ITRACE_ENTRY();

	if (p_conn->cq_thrd_handle == 0) {
		ITRACE_EXIT();
		return;
	}

	thread = p_conn->cq_thrd_handle;

	/*
	 * setting this to zero will signal the thread to stop on its next
	 * loop pass.  this should give the thread time to ack all its events.
	 */
	p_conn->cq_thrd_handle = 0;

	kill_thread(thread);
	wait_thread(thread);
	TRACE(TRACE_ISER, "%s: cq thread terminated\n", __func__);

	ITRACE_EXIT();
}

/**
 * \brief Command queue handling thread.
 * \param void_adaptor A void* that contains the address of the iser_adaptor_t
 *                     that this thread is handling events for.
 * \return 0 on success, -1 on failure.
 */
static thread_ret iser_cq_handler_thread(void *void_conn)
{
	int err;
	struct task_struct *current;
	struct iser_conn_t *p_conn = void_conn;

#if !defined(OSC_IWARP)
	struct ibv_cq *cq_p;
	void *context_p;
#endif /* !OSC_IWARP */


	ITRACE_ENTRY();

	unh_daemonize(__func__);

	/* allow the creating thread to continue */
	up(&p_conn->startstop_sem);

	while (p_conn->cq_thrd_handle) {

#if !defined(OSC_IWARP)
		/* request notification of events on the completion channel */
		TRACE(TRACE_ISER, "%s: requesting notify\n", __func__);
		err = ibv_req_notify_cq(p_conn->cq, 0);
		if (err) {
			TRACE_ERROR("%s: ibv_req_notify_cq failed\n",
			            __func__);
			goto error;
		}
#endif /* !OSC_IWARP */


		TRACE(TRACE_ISER, "%s: polling event\n", __func__);
		err = iser_do_poll_cq(p_conn);
		if (err < 0)
			goto error;


#if !defined(OSC_IWARP)
		/*
		 * wait for notification of an event on the completion
		 * channel.
		 *
		 * context_p is unused
		 */ 
		TRACE(TRACE_ISER, "%s: getting event\n", __func__);
		err = ibv_get_cq_event(p_conn->comp_chan, &cq_p, &context_p);
		if (err) {
			TRACE_ERROR("%s: ibv_get_cq_event failed\n", __func__);
			goto error;
		}
		if (cq_p != p_conn->cq) {
			TRACE_WARNING("%s: got event on the wrong CQ %p, "
			              "expected %p\n", __func__, cq_p,
			              p_conn->cq);
		}

		/* we need to ack 1 here reguardless of how many events poll
		 * says it gets. */
		TRACE(TRACE_ISER, "%s: acking event\n", __func__);
		ibv_ack_cq_events(p_conn->cq, 1);
#endif /* !OSC_IWARP */
	}

	/*
	 * Poll one  more time since ibv_get_cq_event() may have notified
	 * us of an event.
	 */
	err = iser_do_poll_cq(p_conn);
	if (err < 0)
		goto error;

	ITRACE_EXIT();
	return (thread_ret) 0;

      error:
	ITRACE_EXIT();
	return (thread_ret) - 1;
}

#endif /* !__KERNEL__ */
