/**
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * \file datamover/iser_ofa_user.h
 * \brief Contains declarations and definitions for the OFA user-space
 *        API.
 * \author Ethan Burns
 * \date 2007-08-06
 */

#define _POSIX_C_SOURCE 200112L

#if defined(__KERNEL__)
#error This can only be compiled in user-space
#endif /* __KERNEL__ */

#define iser_printf(x...) printf(x)

int iser_spawn_cq_thread(struct iser_conn_t *p_conn);
void iser_kill_cq_thread(struct iser_conn_t *p_conn);

/** \brief Creates an rdma_cm_id */
static inline int
iser_rdma_create_id(struct rdma_event_channel *channel,
                    struct rdma_cm_id **id, void *context,
                    enum rdma_port_space ps)
{
	return rdma_create_id(channel, id, context, ps);
}

/** \brief Destroys an rdma_cm_id structure. */
static inline int iser_rdma_destroy_id(struct rdma_cm_id *id)
{
	return rdma_destroy_id(id);
}

/** \brief Gets device attributes */
static inline int
iser_ib_query_device(struct rdma_cm_id *id,
                     ib_device_attr_t *device_attr)
{
	return ibv_query_device(id->verbs, device_attr);
}

/** \brief Creates a command queue */
static inline ib_cq_t *
iser_ib_create_cq(struct iser_conn_t *p_conn, int cqe)
{
	return ibv_create_cq(p_conn->cm_id->verbs, cqe, NULL,
	                     p_conn->comp_chan, 0);
}

/** \brief Destroys a command queue */
static inline int iser_ib_destroy_cq(ib_cq_t *cq)
{
	return ibv_destroy_cq(cq);
}

/** \brief Polls a command queue.  */
static inline int iser_ib_poll_cq(ib_cq_t *cq, int num_entries,
                                  ib_wc_t *wc)
{
	return ibv_poll_cq(cq, num_entries, wc);
}

/** \brief Destroys a queue pair. */
static inline int iser_ib_destroy_qp(ib_qp_t *qp)
{
	return ibv_destroy_qp(qp);
}

/** \brief Register a memory region */
static inline ib_mr_t *
iser_ib_reg_mr(struct iser_conn_t *p_conn,
               struct iser_mem_t *vec,
               unsigned int vec_len,
               size_t length,
               int access,
               enum dma_data_direction dir)
{
	void *addr;

	BUG_ON(vec_len != 1);

	addr = (void*) (unsigned long) vec->addr;

	return ibv_reg_mr(p_conn->pd, addr, length, access);
}

/** \brief Deregisters a memory region.  */
static inline int iser_ib_dereg_mr(ib_mr_t *mr)
{
	return ibv_dereg_mr(mr);
}

/** \brief Posts a receive buffer. */
static inline int iser_ib_post_recv(ib_qp_t *qp,
                                    ib_recv_wr_t *recv_wr,
                                    ib_recv_wr_t **bad_recv_wr)
{
	return ibv_post_recv(qp, recv_wr, bad_recv_wr);
}

/** \brief Posts a send buffer. */
static inline int iser_ib_post_send(ib_qp_t *qp,
                                    ib_send_wr_t *send_wr,
                                    ib_send_wr_t **bad_send_wr)
{
	return ibv_post_send(qp, send_wr, bad_send_wr);
}

/** \brief Gets the next rdma_cm_event on the queue.  */
static inline int
iser_rdma_get_cm_event(struct rdma_event_channel *channel,
                       iser_cm_event_t **event)
{
	return rdma_get_cm_event(channel, event);
}

/** \brief Acks the front rdma_cm_event. */
static inline int iser_rdma_ack_cm_event(iser_cm_event_t *event)
{
	return rdma_ack_cm_event(event);
}

/** \brief Allocate a protection domain. */
static inline ib_pd_t *iser_ib_alloc_pd(struct rdma_cm_id *id)
{
	return ibv_alloc_pd(id->verbs);
}

/** \brief Deallocates a protection domain. */
static inline int iser_ib_dealloc_pd(ib_pd_t *pd)
{
	return ibv_dealloc_pd(pd);
}

/** \brief Returns a printable string for the status enum */
static inline const char *iser_wc_status_string(int status)
{
	switch(status) {
        case IBV_WC_SUCCESS:
		return "IBV_WC_SUCCESS";
        case IBV_WC_LOC_LEN_ERR:
		return "IBV_WC_LOC_LEN_ERR";
        case IBV_WC_LOC_QP_OP_ERR:
		return "IBV_WC_LOC_QP_OP_ERR";
        case IBV_WC_LOC_EEC_OP_ERR:
		return "IBV_WC_LOC_EEC_OP_ERR";
        case IBV_WC_LOC_PROT_ERR:
		return "IBV_WC_LOC_PROT_ERR";
        case IBV_WC_WR_FLUSH_ERR:
		return "IBV_WC_WR_FLUSH_ERR";
        case IBV_WC_MW_BIND_ERR:
		return "IBV_WC_MW_BIND_ERR";
        case IBV_WC_BAD_RESP_ERR:
		return "IBV_WC_BAD_RESP_ERR";
        case IBV_WC_LOC_ACCESS_ERR:
		return "IBV_WC_LOC_ACCESS_ERR";
        case IBV_WC_REM_INV_REQ_ERR:
		return "IBV_WC_REM_INV_REQ_ERR";
        case IBV_WC_REM_ACCESS_ERR:
		return "IBV_WC_REM_ACCESS_ERR";
        case IBV_WC_REM_OP_ERR:
		return "IBV_WC_REM_OP_ERR";
        case IBV_WC_RETRY_EXC_ERR:
		return "IBV_WC_RETRY_EXC_ERR";
        case IBV_WC_RNR_RETRY_EXC_ERR:
		return "IBV_WC_RNR_RETRY_EXC_ERR";
        case IBV_WC_LOC_RDD_VIOL_ERR:
		return "IBV_WC_LOC_RDD_VIOL_ERR";
        case IBV_WC_REM_INV_RD_REQ_ERR:
		return "IBV_WC_REM_INV_RD_REQ_ERR";
        case IBV_WC_REM_ABORT_ERR:
		return "IBV_WC_REM_ABORT_ERR";
        case IBV_WC_INV_EECN_ERR:
		return "IBV_WC_INV_EECN_ERR";
        case IBV_WC_INV_EEC_STATE_ERR:
		return "IBV_WC_INV_EEC_STATE_ERR";
        case IBV_WC_FATAL_ERR:
		return "IBV_WC_FATAL_ERR";
        case IBV_WC_RESP_TIMEOUT_ERR:
		return "IBV_WC_RESP_TIMEOUT_ERR";
        case IBV_WC_GENERAL_ERR:
		return "IBV_WC_GENERAL_ERR";
	default:
		return "UNKNOWN";
	}
}

/* vi: set tabstop=8 textwidth=72: */
