/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/kernel.h>
#include <linux/types.h>
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_global.h"
#include "iser_utils.h"
#include "iser_trace.h"

#include <scsi/scsi.h>

#define PRINT_BUF_LEN                      1024	/* size of the buffer for trace string accumulation */
#define ISER_PDU_PRINT_MAX_DATA_BLOCK_LEN   256	/* max size of data printed in a single block */

static int print_buf_pos = 0;	/* current position witin the print buffer */
static int print_buf_trunc = 0;	/* set if the buffer was truncated to avoid overflow */
static char print_buf[PRINT_BUF_LEN];	/* print buffer */

#define PRINT_BUFF_CAT( fmt , args... ) \
    do { \
        if( !print_buf_trunc && print_buf_pos + 32 < PRINT_BUF_LEN ) \
            print_buf_pos += sprintf(print_buf+print_buf_pos, fmt , ## args); \
        else if( !print_buf_trunc ) { \
            print_buf_pos += sprintf(print_buf+print_buf_pos, " ... (deleted)\n"); \
            print_buf_trunc = 1; \
        } \
    } while(0)

#define PRINT_BUFF_OUT() \
    do { \
        sprintf(print_buf+print_buf_pos, "\n"); \
        print_buf_pos = 0; print_buf_trunc = 0; \
        TRACE(TRACE_RDMA_EVENTS, "%s", print_buf); \
    } while(0)

union u64_view {
	__u64 x;
	__u8 c[8];
};

static char *string_llx(__u64 x)
{
	int i;
	static char str[20];
	static char convert[] = "0123456789abcdef";
	char *ptr;
	__u32 c;
	union u64_view tmp;

	tmp.x = x;
	ptr = str;
	*ptr++ = '0';
	*ptr++ = 'x';
	for (i = 7; i >= 0; i--) {
		c = tmp.c[i];
		*ptr++ = convert[c >> 4];
		*ptr++ = convert[c & 0xf];
	}
	*ptr = '\0';
	return str;
}


static void print_rsvd_u8(int n, __u8 rsvd)
{
	if (rsvd != 0)
		PRINT_BUFF_CAT("rsvd%d: 0x%.2x ", n, rsvd);
}


static void print_rsvd_u16(int n, __u16 rsvd)
{
	if (rsvd != 0)
		PRINT_BUFF_CAT("rsvd%d: 0x%.4x ", n, rsvd);
}


static void print_rsvd_u32(int n, __u32 rsvd)
{
	if (rsvd != 0)
		PRINT_BUFF_CAT("rsvd%d: 0x%.8x ", n, rsvd);
}


static void print_rsvd_u64(int n, __u64 rsvd)
{
	if (rsvd != 0ll)
		PRINT_BUFF_CAT("rsvd%d: %s ", n, string_llx(rsvd));
}


static void print_opcode(__u8 opcode)
{
	PRINT_BUFF_CAT("(0x%.2X%s) ", opcode & 0x3f,
		       ((opcode & 0x40) != 0 ? "-Imm" : ""));
}


static void print_flags(__u8 flags)
{
	PRINT_BUFF_CAT("flags: 0x%.2x ", flags);
}


static void print_version(char *which, __u8 version)
{
	PRINT_BUFF_CAT("Version%s: 0x%.2x ", which, version);
}


static void print_response(__u8 response)
{
	if (response == 0x00) {
		PRINT_BUFF_CAT("Resp: Completed(0) ");
	} else if (response == 0x01) {
		PRINT_BUFF_CAT("Resp: Failure(1)   ");
	} else {
		PRINT_BUFF_CAT("Resp: Vendor(%d)   ", response);
	}
}


static void print_status(__u8 status)
{
	if (status == 0x00) {
		PRINT_BUFF_CAT("Status: GOOD(0) ");
	} else if (status == 0x02) {
		PRINT_BUFF_CAT("Status: CHCK(2) ");
	} else if (status == 0x08) {
		PRINT_BUFF_CAT("Status: BUSY(2) ");
	} else {
		PRINT_BUFF_CAT("Status: 0x%.2x ", status);
	}
}


static void print_lun(__u64 lun)
{
	PRINT_BUFF_CAT("LUN: %s ", string_llx(lun));
}


static void print_isid_tsih(__u8 isid[6], __u16 tsih)
{
	PRINT_BUFF_CAT("ISID: 0x%.2x %.2x %.2x %.2x %.2x %.2x ",
		       isid[0], isid[1], isid[2], isid[3], isid[4],
		       isid[5]);
	PRINT_BUFF_CAT("TSIH: %u ", be16_to_cpu(tsih));
}


static void print_dsl(__u32 length)
{
	__u32 be32_length = be32_to_cpu(length);

	PRINT_BUFF_CAT("DSL: 0x%X(%u) ", be32_length, be32_length);
}

static void print_itt(__u32 init_task_tag)
{
	if (init_task_tag == 0xffffffff)
		PRINT_BUFF_CAT("ITT: 0x%08X ", init_task_tag);
	else {
		__u32 be32_init_task_tag = be32_to_cpu(init_task_tag);
		PRINT_BUFF_CAT("ITT: 0x%X(%u) ",
			       be32_init_task_tag, be32_init_task_tag);
	}
}


static void print_ttt(__u32 target_xfer_tag)
{
	if (target_xfer_tag == 0xffffffff)
		PRINT_BUFF_CAT("TTT: 0x%08x ", target_xfer_tag);
	else
		PRINT_BUFF_CAT("TTT: %u ", be32_to_cpu(target_xfer_tag));
}


static void print_cid(__u16 cid)
{
	PRINT_BUFF_CAT("CID: %u ", be16_to_cpu(cid));
}


static void print_expstatsn(__u32 exp_stat_sn)
{
	if (exp_stat_sn != 0)
		PRINT_BUFF_CAT("ExpStatSN: %u ", be32_to_cpu(exp_stat_sn));
}


static void print_cmdsn_expstatsn(__u32 cmd_sn, __u32 exp_stat_sn)
{
	PRINT_BUFF_CAT("CmdSN: %u ", be32_to_cpu(cmd_sn));
	print_expstatsn(exp_stat_sn);
}


static void
print_statsn_exp_max(__u32 stat_sn, __u32 exp_cmd_sn, __u32 max_cmd_sn)
{
	if (stat_sn != 0)
		PRINT_BUFF_CAT("StatSN: %u ", be32_to_cpu(stat_sn));
	PRINT_BUFF_CAT("ExpCmdSN: %u ", be32_to_cpu(exp_cmd_sn));
	PRINT_BUFF_CAT("MaxCmdSN: %u ", be32_to_cpu(max_cmd_sn));
}


static void print_residual(__u32 resid)
{
	if (resid != 0)
		PRINT_BUFF_CAT("ResidualCount: %u ", be32_to_cpu(resid));
}


static void print_datasn(__u32 data_sn)
{
	PRINT_BUFF_CAT("DataSN: %u ", be32_to_cpu(data_sn));
}

static void print_offset(__u32 offset, __u32 length)
{
	__u32 be32_offset = be32_to_cpu(offset);
	__u32 be32_length = be32_to_cpu(length);

	PRINT_BUFF_CAT("Buf: %u-%u ",
		       be32_offset, be32_offset + be32_length);
}


static void print_rtt(__u32 ref_task_tag)
{
	if (ref_task_tag != 0)
		PRINT_BUFF_CAT("RTT: 0x%.8x ", be32_to_cpu(ref_task_tag));
}


static void print_exp_data_sn(__u32 exp_data_sn)
{
	if (exp_data_sn != 0)
		PRINT_BUFF_CAT("ExpDataSN: %u ", be32_to_cpu(exp_data_sn));
}


static void print_begrun(__u32 begrun)
{
	PRINT_BUFF_CAT("BegRun: %u ", be32_to_cpu(begrun));
}


static void print_runlen(__u32 runlen)
{
	PRINT_BUFF_CAT("RunLength: %u ", be32_to_cpu(runlen));
}

/* command opcodes which don't appear in <scsi/scsi.h> */
#define EXTENDED_COPY                   0x83
#define MOVE_MEDIUM_ATTACHED            0xa7
#define READ_ELEMENT_STATUS_ATTACHED    0xb4
#define RECEIVE_COPY_RESULTS            0x84
#define REPORT_DEVICE_ID                0xa3
#define REPORT_LUNS                     0xa0
#define SET_DEVICE_ID                   0xa4

static void print_scsi_cdb_op(__u8 opcode)
{
	char *result = "UNKNOWN";

	/*  opcode symbols defined in <scsi/scsi.h> */
	/*  this table includes only those symbols we have actually seen
	   being sent to/from a vendor */
	switch (opcode) {
	case TEST_UNIT_READY:	/* 0x00 */
		result = "TEST_UNIT_READY";
		break;
	case REZERO_UNIT:	/* 0x01 */
		result = "REWIND";
		break;
	case REQUEST_SENSE:	/* 0x03 */
		result = "REQUEST_SENSE";
		break;
	case READ_BLOCK_LIMITS:	/* 0x05 */
		result = "READ_BLOCK_LIMITS";
		break;
	case READ_6:		/* 0x08 */
		result = "READ_6";
		break;
	case WRITE_6:		/* 0x0a */
		result = "WRITE_6";
		break;
	case WRITE_FILEMARKS:	/* 0x10 */
		result = "WRITE_FILEMARKS";
		break;
	case INQUIRY:		/* 0x12 */
		result = "INQUIRY";
		break;
	case MODE_SENSE:	/* 0x1a */
		result = "MODE_SENSE";
		break;
	case READ_CAPACITY:	/* 0x25 */
		result = "READ_CAPACITY";
		break;
	case READ_10:		/* 0x28 */
		result = "READ_10";
		break;
	case WRITE_10:		/* 0x2a */
		result = "WRITE_10";
		break;
	case READ_12:		/* 0xa8 */
		result = "READ_12";
		break;
	case WRITE_12:		/* 0xaa */
		result = "WRITE_12";
		break;
	case FORMAT_UNIT:	/* 0x04 */
		result = "FORMAT_UNIT";
		break;
	case REASSIGN_BLOCKS:	/* 0x07 */
		result = "REASSIGN_BLOCKS";
		break;
	case SEEK_6:		/* 0x0b */
		result = "SEEK_6";
		break;
	case READ_REVERSE:	/* 0x0f */
		result = "READ_REVERSE";
		break;
	case SPACE:		/* 0x11 */
		result = "SPACE";
		break;
	case RECOVER_BUFFERED_DATA:	/* 0x14 */
		result = "RECOVER_BUFFERED_DATA";
		break;
	case MODE_SELECT:	/* 0x15 */
		result = "MODE_SELECT";
		break;
	case RESERVE:		/* 0x16 */
		result = "RESERVE";
		break;
	case RELEASE:		/* 0x17 */
		result = "RELEASE";
		break;
	case COPY:		/* 0x18 */
		result = "COPY";
		break;
	case ERASE:		/* 0x19 */
		result = "ERASE";
		break;
	case START_STOP:	/* 0x1b */
		result = "START_STOP";
		break;
	case RECEIVE_DIAGNOSTIC:	/* 0x1c */
		result = "RECEIVE_DIAGNOSTIC";
		break;
	case SEND_DIAGNOSTIC:	/* 0x1d */
		result = "SEND_DIAGNOSTIC";
		break;
	case ALLOW_MEDIUM_REMOVAL:	/* 0x1e */
		result = "ALLOW_MEDIUM_REMOVAL";
		break;
	case SET_WINDOW:	/* 0x24 */
		result = "SET_WINDOW";
		break;
	case SEEK_10:		/* 0x2b */
		result = "SEEK_10";
		break;
	case WRITE_VERIFY:	/* 0x2e */
		result = "WRITE_VERIFY";
		break;
	case VERIFY:		/* 0x2f */
		result = "VERIFY";
		break;
	case SEARCH_HIGH:	/* 0x30 */
		result = "SEARCH_HIGH";
		break;
	case SEARCH_EQUAL:	/* 0x31 */
		result = "SEARCH_EQUAL";
		break;
	case SEARCH_LOW:	/* 0x32 */
		result = "SEARCH_LOW";
		break;
	case SET_LIMITS:	/* 0x33 */
		result = "SET_LIMITS";
		break;
	case PRE_FETCH:	/* 0x34 */
		result = "PRE_FETCH";
		break;
		/* case READ_POSITION:         // 0x34
		   result = "READ_POSITION";
		   break; */
	case SYNCHRONIZE_CACHE:	/* 0x35 */
		result = "SYNCHRONIZE_CACHE";
		break;
	case LOCK_UNLOCK_CACHE:	/* 0x36 */
		result = "LOCK_UNLOCK_CACHE";
		break;
	case READ_DEFECT_DATA:	/* 0x37 */
		result = "READ_DEFECT_DATA";
		break;
	case MEDIUM_SCAN:	/* 0x38 */
		result = "MEDIUM_SCAN";
		break;
	case COMPARE:		/* 0x39 */
		result = "COMPARE";
		break;
	case COPY_VERIFY:	/* 0x3a */
		result = "COPY_VERIFY";
		break;
	case WRITE_BUFFER:	/* 0x3b */
		result = "WRITE_BUFFER";
		break;
	case READ_BUFFER:	/* 0x3c */
		result = "READ_BUFFER";
		break;
	case UPDATE_BLOCK:	/* 0x3d */
		result = "UPDATE_BLOCK";
		break;
	case READ_LONG:	/* 0x3e */
		result = "READ_LONG";
		break;
	case WRITE_LONG:	/* 0x3f */
		result = "WRITE_LONG";
		break;
	case CHANGE_DEFINITION:	/* 0x40 */
		result = "CHANGE_DEFINITION";
		break;
	case WRITE_SAME:	/* 0x41 */
		result = "WRITE_SAME";
		break;
	case READ_TOC:		/* 0x43 */
		result = "READ_TOC";
		break;
	case LOG_SELECT:	/* 0x4c */
		result = "LOG_SELECT";
		break;
	case LOG_SENSE:	/* 0x4d */
		result = "LOG_SENSE";
		break;
	case MODE_SELECT_10:	/* 0x55 */
		result = "MODE_SELECT_10";
		break;
	case RESERVE_10:	/* 0x56 */
		result = "RESERVE_10";
		break;
	case RELEASE_10:	/* 0x57 */
		result = "RELEASE_10";
		break;
	case MODE_SENSE_10:	/* 0x5a */
		result = "MODE_SENSE_10";
		break;
	case PERSISTENT_RESERVE_IN:	/* 0x5e */
		result = "PERSISTENT_RESERVE_IN";
		break;
	case PERSISTENT_RESERVE_OUT:	/* 0x5f */
		result = "PERSISTENT_RESERVE_OUT";
		break;
	case EXTENDED_COPY:	/* 0x83 */
		result = "EXTENDED_COPY";
		break;
	case RECEIVE_COPY_RESULTS:	/* 0x84 */
		result = "RECEIVE_COPY_RESULTS";
		break;
	case REPORT_DEVICE_ID:	/* 0xa3 */
		result = "REPORT_DEVICE_ID";
		break;
	case REPORT_LUNS:	/* 0xa0 */
		result = "REPORT_LUNS";
		break;
	case SET_DEVICE_ID:	/*  0xa4 */
		result = "SET_DEVICE_ID";
		break;
	case MOVE_MEDIUM:	/* 0xa5 */
		result = "MOVE_MEDIUM";
		break;
	case MOVE_MEDIUM_ATTACHED:	/* 0xa7 */
		result = "MOVE_MEDIUM_ATTACHED";
		break;
	case WRITE_VERIFY_12:	/* 0xae */
		result = "WRITE_VERIFY_12";
		break;
	case SEARCH_HIGH_12:	/* 0xb0 */
		result = "SEARCH_HIGH_12";
		break;
	case SEARCH_EQUAL_12:	/* 0xb1 */
		result = "SEARCH_EQUAL_12";
		break;
	case SEARCH_LOW_12:	/* 0xb2 */
		result = "SEARCH_LOW_12";
		break;
	case READ_ELEMENT_STATUS_ATTACHED:	/* 0xb4 */
		result = "READ_ELEMENT_STATUS_ATTACHED";
		break;
	case READ_ELEMENT_STATUS:	/* 0xb8 */
		result = "READ_ELEMENT_STATUS";
		break;
	case SEND_VOLUME_TAG:	/* 0xb6 */
		result = "SEND_VOLUME_TAG";
		break;
	case WRITE_LONG_2:	/* 0xea */
		result = "WRITE_LONG_2";
		break;
	}			/* switch */

	PRINT_BUFF_CAT("CDB-OP: %s ", result);
}

static void print_init_scsi_cmnd(char *buffer)
{
	struct iscsi_init_scsi_cmnd *cmd =
	    (struct iscsi_init_scsi_cmnd *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	PRINT_BUFF_CAT("EDTL: %u ", be32_to_cpu(cmd->xfer_len));
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	print_scsi_cdb_op(cmd->cdb[0]);
	PRINT_BUFF_CAT
	    ("CDB-DATA: 0x%.2x%.2x%.2x%.2x %.2x%.2x%.2x%.2x %.2x%.2x%.2x%.2x %.2x%.2x%.2x%.2x ",
	     cmd->cdb[0], cmd->cdb[1], cmd->cdb[2], cmd->cdb[3],
	     cmd->cdb[4], cmd->cdb[5], cmd->cdb[6], cmd->cdb[7],
	     cmd->cdb[8], cmd->cdb[9], cmd->cdb[10], cmd->cdb[11],
	     cmd->cdb[12], cmd->cdb[13], cmd->cdb[14], cmd->cdb[15]);
}


static void print_targ_scsi_rsp(char *buffer)
{
	struct iscsi_targ_scsi_rsp *cmd =
	    (struct iscsi_targ_scsi_rsp *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_response(cmd->response);
	print_status(cmd->status);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_exp_data_sn(cmd->exp_data_sn);
	if (cmd->bidi_resid != 0)
		PRINT_BUFF_CAT("BidirResidualCount: %u ",
			       be32_to_cpu(cmd->bidi_resid));
	print_residual(cmd->resid);
}


static void print_init_text_cmnd(char *buffer)
{
	struct iscsi_init_text_cmnd *cmd =
	    (struct iscsi_init_text_cmnd *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	print_rsvd_u64(4, cmd->rsvd4);
	print_rsvd_u64(5, cmd->rsvd5);
}


static void print_targ_text_rsp(char *buffer)
{
	struct iscsi_targ_text_rsp *cmd =
	    (struct iscsi_targ_text_rsp *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_rsvd_u32(4, cmd->rsvd4);
	print_rsvd_u64(5, cmd->rsvd5);
}


static void print_init_login_cmnd(char *buffer)
{
	struct iscsi_init_login_cmnd *cmd =
	    (struct iscsi_init_login_cmnd *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_version("Max", cmd->version_max);
	print_version("Min", cmd->version_min);
	print_dsl(cmd->length);
	print_isid_tsih(cmd->isid, cmd->tsih);
	print_itt(cmd->init_task_tag);
	print_cid(cmd->cid);
	print_rsvd_u16(1, cmd->rsvd1);
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	print_rsvd_u64(2, cmd->rsvd2);
	print_rsvd_u64(3, cmd->rsvd3);
}


static void print_targ_login_rsp(char *buffer)
{
	struct iscsi_targ_login_rsp *cmd =
	    (struct iscsi_targ_login_rsp *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_version("Max", cmd->version_max);
	print_version("Active", cmd->version_active);
	print_dsl(cmd->length);
	print_isid_tsih(cmd->isid, cmd->tsih);
	print_itt(cmd->init_task_tag);
	print_rsvd_u32(1, cmd->rsvd1);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	if (cmd->status_class != 0)
		PRINT_BUFF_CAT("StatusClass: 0x%.2x ", cmd->status_class);
	if (cmd->status_detail != 0)
		PRINT_BUFF_CAT("StatusDetail: 0x%.2x ",
			       cmd->status_detail);
	print_rsvd_u16(2, cmd->rsvd2);
	print_rsvd_u64(3, cmd->rsvd3);
}


static void print_init_logout_cmnd(char *buffer)
{
	struct iscsi_init_logout_cmnd *cmd =
	    (struct iscsi_init_logout_cmnd *) buffer;

	print_opcode(cmd->opcode);
	PRINT_BUFF_CAT("reasoncod: 0x%.2x ", cmd->flags);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_cid(cmd->cid);
	print_rsvd_u16(2, cmd->rsvd2);
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	print_rsvd_u64(4, cmd->rsvd4);
	print_rsvd_u64(5, cmd->rsvd5);
}


static void print_targ_logout_rsp(char *buffer)
{
	struct iscsi_targ_logout_rsp *cmd =
	    (struct iscsi_targ_logout_rsp *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_response(cmd->response);
	print_rsvd_u8(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_rsvd_u32(3, cmd->rsvd3);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_rsvd_u32(4, cmd->rsvd4);
	PRINT_BUFF_CAT("Time2Wait: 0x%.8x ", be16_to_cpu(cmd->time2wait));
	PRINT_BUFF_CAT("Tm2Retain: 0x%.8x ",
		       be16_to_cpu(cmd->time2retain));
	print_rsvd_u32(5, cmd->rsvd5);
}


static void print_init_scsi_data_out(char *buffer)
{
	struct iscsi_init_scsi_data_out *cmd =
	    (struct iscsi_init_scsi_data_out *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_rsvd_u32(3, cmd->rsvd3);
	print_expstatsn(cmd->exp_stat_sn);
	print_rsvd_u32(4, cmd->rsvd4);
	print_datasn(cmd->data_sn);
	print_offset(cmd->offset, cmd->length);
	print_rsvd_u32(5, cmd->rsvd5);
}


static void print_targ_scsi_data_in(char *buffer)
{
	struct iscsi_targ_scsi_data_in *cmd =
	    (struct iscsi_targ_scsi_data_in *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u8(1, cmd->rsvd1);
	print_status(cmd->status);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_datasn(cmd->data_sn);
	print_offset(cmd->offset, cmd->length);
	print_residual(cmd->resid);
}


static void print_targ_rjt(char *buffer)
{
	struct iscsi_targ_rjt *cmd = (struct iscsi_targ_rjt *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	if (cmd->reason != 0)
		PRINT_BUFF_CAT("Reason: 0x%.2x ", cmd->reason);
	print_rsvd_u8(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_rsvd_u32(4, cmd->rsvd4);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_datasn(cmd->data_sn);
	print_rsvd_u64(4, cmd->rsvd4);
	print_rsvd_u64(5, cmd->rsvd5);
}


static void print_init_nopout(char *buffer)
{
	struct iscsi_init_nopout *cmd =
	    (struct iscsi_init_nopout *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	print_rsvd_u64(2, cmd->rsvd2);
	print_rsvd_u64(3, cmd->rsvd3);
}


static void print_targ_nopin(char *buffer)
{
	struct iscsi_targ_nopin *cmd = (struct iscsi_targ_nopin *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_rsvd_u32(2, cmd->rsvd2);
	print_rsvd_u64(3, cmd->rsvd3);
}


static void print_targ_r2t(char *buffer)
{
	struct iscsi_targ_r2t *cmd = (struct iscsi_targ_r2t *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	PRINT_BUFF_CAT("R2TSN: %u ", be32_to_cpu(cmd->r2t_sn));
	print_offset(cmd->offset, cmd->xfer_len);
}


static void print_targ_async_msg(char *buffer)
{
	struct iscsi_targ_async_msg *cmd =
	    (struct iscsi_targ_async_msg *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(2, cmd->rsvd2);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_rsvd_u32(3, cmd->rsvd3);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	PRINT_BUFF_CAT("AsyncEvnt: %u ", cmd->async_event);
	PRINT_BUFF_CAT("AsyncVCod: %u ", cmd->async_vcode);
	if (cmd->parameter1 != 0)
		PRINT_BUFF_CAT("   Param1: %u ", cmd->parameter1);
	if (cmd->parameter2 != 0)
		PRINT_BUFF_CAT("   Param2: %u ", cmd->parameter2);
	if (cmd->parameter3 != 0)
		PRINT_BUFF_CAT("   Param3: %u ", cmd->parameter3);
	print_rsvd_u32(5, cmd->rsvd5);
}


static void print_init_task_mgt_command(char *buffer)
{
	struct iscsi_init_task_mgt_command *cmd =
	    (struct iscsi_init_task_mgt_command *) buffer;

	print_opcode(cmd->opcode);
	PRINT_BUFF_CAT("Function: 0x%.2x ", cmd->function);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_rtt(cmd->ref_task_tag);
	print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
	if (cmd->ref_cmd_sn != 0)
		PRINT_BUFF_CAT("RefCmdSN: %u ",
			       be32_to_cpu(cmd->ref_cmd_sn));
	print_exp_data_sn(cmd->exp_data_sn);
	print_rsvd_u64(4, cmd->rsvd4);
}


static void print_targ_task_mgt_response(char *buffer)
{
	struct iscsi_targ_task_mgt_response *cmd =
	    (struct iscsi_targ_task_mgt_response *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_response(cmd->response);
	print_rsvd_u8(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_rsvd_u32(2, cmd->rsvd2);
	print_statsn_exp_max(cmd->stat_sn, cmd->exp_cmd_sn,
			     cmd->max_cmd_sn);
	print_rsvd_u32(4, cmd->rsvd4);
	print_rsvd_u64(5, cmd->rsvd5);
}


static void print_init_snack(char *buffer)
{
	struct iscsi_init_snack *cmd = (struct iscsi_init_snack *) buffer;

	print_opcode(cmd->opcode);
	print_flags(cmd->flags);
	print_rsvd_u16(1, cmd->rsvd1);
	print_dsl(cmd->length);
	print_lun(cmd->lun);
	print_itt(cmd->init_task_tag);
	print_ttt(cmd->target_xfer_tag);
	print_rsvd_u32(2, cmd->rsvd2);
	print_expstatsn(cmd->exp_stat_sn);
	print_rsvd_u64(3, cmd->rsvd3);
	print_begrun(cmd->begrun);
	print_runlen(cmd->runlen);
}


//----------------------------------------------------------------------
/*!
  @brief Prints out an iSCSI PDU - its BHS and, if supplied, the data segment
*/
// ---------------------------------------------------------------------
void iser_pdu_print(char *prefix,	//!< string, to be printed before the PDU
		    void *p_id,	//!< pointer, to be printed before the PDU
		    char *p_bhs,	//!< bhs buffer 
		    struct iser_data_buf_t *p_data)	//!< data segmentdescriptor
{
	__u32 opcode;
	int max_data_print_len = 0x1000;

	PRINT_BUFF_CAT(prefix);
	if ((char *) p_id != NULL)
		PRINT_BUFF_CAT("(0x%p): ", (char *) p_id);

	opcode = p_bhs[0] & ISCSI_OPCODE_MASK;

	switch (opcode) {
	case ISCSI_OP_LOGIN_REQ:
		{
			PRINT_BUFF_CAT("LOGIN REQUEST");
			print_init_login_cmnd(p_bhs);
			break;
		}

	case ISCSI_OP_TEXT_REQ:
		{
			PRINT_BUFF_CAT("TEXT REQUEST");
			print_init_text_cmnd(p_bhs);
			break;
		}

	case ISCSI_OP_SCSI_CMD:
		{
			PRINT_BUFF_CAT("SCSI COMMAND");
			print_init_scsi_cmnd(p_bhs);
			break;
		}

	case ISCSI_OP_DATA_OUT:
		{
			PRINT_BUFF_CAT("DATA_OUT");
			print_init_scsi_data_out(p_bhs);
			break;
		}

	case ISCSI_OP_TASK_MGT_REQ:
		{
			PRINT_BUFF_CAT("TAST MNGT CMD");
			print_init_task_mgt_command(p_bhs);
			break;
		}

	case ISCSI_OP_LOGOUT_REQ:
		{
			PRINT_BUFF_CAT("LOGOUT REQUEST");
			print_init_logout_cmnd(p_bhs);
			break;
		}

	case ISCSI_OP_NOP_OUT:
		{
			PRINT_BUFF_CAT("NOP_OUT");
			print_init_nopout(p_bhs);
			break;
		}

		/* SNACK Handling by Target - SAI */
	case ISCSI_OP_SNACK_REQ:
		{
			PRINT_BUFF_CAT("SNACK Request");
			print_init_snack(p_bhs);
			break;
		}

	case ISCSI_OP_NOP_IN:
		{
			PRINT_BUFF_CAT("NOP_IN");
			print_targ_nopin(p_bhs);
			break;
		}

	case ISCSI_OP_SCSI_RSP:
		{
			PRINT_BUFF_CAT("SCSI RESPONSE");
			print_targ_scsi_rsp(p_bhs);
			break;
		}

	case ISCSI_OP_TASK_MGT_RSP:
		{
			PRINT_BUFF_CAT("MGMT RESPONSE");
			print_targ_task_mgt_response(p_bhs);
			break;
		}

	case ISCSI_OP_LOGIN_RSP:
		{
			PRINT_BUFF_CAT("LOGIN RESPONSE");
			print_targ_login_rsp(p_bhs);
			break;
		}

	case ISCSI_OP_TEXT_RSP:
		{
			PRINT_BUFF_CAT("TEXT RESPONSE");
			print_targ_text_rsp(p_bhs);
			break;
		}

	case ISCSI_OP_DATA_IN:
		{
			PRINT_BUFF_CAT("DATA_IN");
			print_targ_scsi_data_in(p_bhs);
			break;
		}

	case ISCSI_OP_LOGOUT_RSP:
		{
			PRINT_BUFF_CAT("LOGOUT RESPONSE");
			print_targ_logout_rsp(p_bhs);
			break;
		}

	case ISCSI_OP_R2T:
		{
			PRINT_BUFF_CAT("R2T");
			print_targ_r2t(p_bhs);
			break;
		}

	case ISCSI_OP_ASYNC:
		{
			PRINT_BUFF_CAT("ASYNC MSG");
			print_targ_async_msg(p_bhs);
			break;
		}

	case ISCSI_OP_REJECT:
		{
			PRINT_BUFF_CAT("TARGET REJECT");
			print_targ_rjt(p_bhs);
			break;
		}

	default:
		{
			PRINT_BUFF_CAT("UNKNOWN OPCODE");
			break;
		}
	}
	if (p_data != NULL) {
		PRINT_BUFF_CAT("DATA: 0x%p (sz: %d, ptr: 0x%p, type: %s)",
			       p_data, p_data->size, p_data->p_buf,
			       iser_data_buf_get_type_name(p_data));
	}
	PRINT_BUFF_OUT();

	// Print data if available
	if (max_data_print_len > 0 && p_data != NULL && p_data->size > 0) {
		char *p_chunk_buf = NULL;
		unsigned int chunk_len = 0;
		unsigned int len_to_print = 0;
		unsigned int chunk_printed_len = 0;
		unsigned int total_printed_len = 0;
		unsigned int cur_chunk = 0;
		int all_printed = 0;

		do {
			if (p_data->type == ISER_BUF_TYPE_SINGLE) {
				p_chunk_buf =
				    p_data->p_buf + total_printed_len;
				len_to_print =
				    MIN(p_data->size - total_printed_len,
					max_data_print_len -
					total_printed_len);
				len_to_print =
				    MIN(len_to_print,
					ISER_PDU_PRINT_MAX_DATA_BLOCK_LEN);

				PRINT_BUFF_CAT
				    ("DATA(0x%p,SINGLE,%d+%d/%d): ",
				     p_data, total_printed_len,
				     len_to_print, p_data->size);

				/* Update counters */
				total_printed_len += len_to_print;
				if (total_printed_len == max_data_print_len
				    || total_printed_len == p_data->size) {
					all_printed = 1;
				}
			} else {
				switch (p_data->type) {
				case ISER_BUF_TYPE_IOVEC_VIRT:{
						p_chunk_buf =
						    iser_iovec_virt_entry_addr
						    (p_data->p_buf,
						     cur_chunk) +
						    chunk_printed_len;
						chunk_len =
						    iser_iovec_entry_len
						    (p_data->p_buf,
						     cur_chunk);

						len_to_print =
						    MIN(chunk_len -
							chunk_printed_len,
							max_data_print_len
							-
							total_printed_len);
						len_to_print =
						    MIN(len_to_print,
							ISER_PDU_PRINT_MAX_DATA_BLOCK_LEN);

						PRINT_BUFF_CAT
						    ("DATA(0x%p,IOVECVIRT[%d],%d+%d/%d): ",
						     p_data, cur_chunk,
						     chunk_printed_len,
						     len_to_print,
						     chunk_len);
						break;
					}
				case ISER_BUF_TYPE_SCATTERLIST:{
						/* XXX -- This code is broken
 						 * with SG chaining!!!!
 						 * EAB */
						p_chunk_buf =
						    iser_scatterlist_entry_to_virt
						    (p_data->p_buf,
						     cur_chunk) +
						    chunk_printed_len;
						chunk_len =
						    iser_scatterlist_entry_len
						    (p_data->p_buf,
						     cur_chunk);

						len_to_print =
						    MIN(chunk_len -
							chunk_printed_len,
							max_data_print_len
							-
							total_printed_len);
						len_to_print =
						    MIN(len_to_print,
							ISER_PDU_PRINT_MAX_DATA_BLOCK_LEN);

						PRINT_BUFF_CAT
						    ("DATA(0x%p,SCATTERLIST[%d],%d+%d/%d): ",
						     p_data, cur_chunk,
						     chunk_printed_len,
						     len_to_print,
						     chunk_len);
						break;
					}
				default:
					all_printed = 1;
					break;
				}	/* switch */

				/* Update counters */
				total_printed_len += len_to_print;
				if (total_printed_len ==
				    max_data_print_len) {
					all_printed = 1;
				} else {
					chunk_printed_len += len_to_print;
					if (chunk_printed_len == chunk_len) {
						chunk_printed_len = 0;
						cur_chunk++;
						if (cur_chunk ==
						    p_data->size) {
							all_printed = 1;
						}
					}
				}
			}

			if (p_chunk_buf != NULL && len_to_print > 0) {
				int i;

				for (i = 0; i < len_to_print; i++) {
					PRINT_BUFF_CAT("%02X",
						       p_chunk_buf[i]);
					if (i % 4 == 3) {
						PRINT_BUFF_CAT(" ");
					}
				}
				PRINT_BUFF_CAT("\n");
				PRINT_BUFF_OUT();

				if (opcode == ISCSI_OP_LOGIN_REQ
				    || opcode == ISCSI_OP_TEXT_REQ
				    || opcode == ISCSI_OP_LOGIN_RSP
				    || opcode == ISCSI_OP_TEXT_RSP) {
					PRINT_BUFF_CAT("TEXT(0x%p): ",
						       p_data);
					for (i = 0; i < len_to_print; i++) {
						PRINT_BUFF_CAT("%c",
							       (p_chunk_buf
								[i] ==
								'\0' ? ' '
								:
								p_chunk_buf
								[i]));
					}
					PRINT_BUFF_CAT("\n");
					PRINT_BUFF_OUT();
				}
			} else {
				all_printed = 1;
			}
		} while (!all_printed);
	}
}				/* iser_pdu_print */

/*
    Copyright (C) 2001-2003 InterOperability Lab (IOL)
                            University of New Hampshier (UNH)
                            Durham, NH 03824

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
    USA.

    The name of IOL and/or UNH may not be used to endorse or promote products
    derived from this software without specific prior written permission.
*/
