/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

/*
#ifndef MODULE  
    #define MODULE
#endif

#ifndef __KERNEL__
    #define __KERNEL__
#endif
*/

#if defined(__KERNEL__)
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/interrupt.h>
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_global.h"
#include "iser_conn.h"
#include "iser_kdapl.h"
#include "iser_memory.h"
#include "iser_procfs.h"
#include "iser_trace.h"

/* ---------------------------------------------------------------------
 * CONSTANTS
 * ------------------------------------------------------------------ */
#define ISER_DM_PROC_FS_PARENT  "voltaire"

static char iser_procfs_path[64];
static struct proc_dir_entry *iser_root_pe = NULL;
static struct proc_dir_entry *iser_iscsi_pe = NULL;
static struct proc_dir_entry *iser_adaptors_pe = NULL;
static struct proc_dir_entry *iser_bufpools_pe = NULL;

/*****************************************************************************
 * iser_procfs_handle_long_read() - Generic handling of long proc reads
 *
 *****************************************************************************/
int
iser_procfs_handle_long_read(char *page,
			     char **start,
			     off_t offset,
			     int count,
			     int *eof,
			     read_proc_buf_func_t buf_f,
			     read_proc_free_buf_func_t free_buf_f,
			     char **p_buf, int *p_buf_sz)
{
	int copy_sz;

	// IERROR("A. page %p start %p offset %d count %d p_buf %p sz %d\n",
	//       page, *start, (int)offset, count, *p_buf, *p_buf_sz );

	if (*p_buf == NULL) {	/* first time */
		if (offset) {
			return 0;
		}
		*p_buf = buf_f(p_buf_sz);
		if (*p_buf == NULL) {
			// vmallock must have failed
			*eof = 1;
			*p_buf_sz = 0;
			return 0;
		}
	}
	if ((offset + count) < *p_buf_sz) {
		*eof = 0;
		//*start = page;
		copy_sz = count;
	} else if ((offset + count) >= *p_buf_sz) {
		*eof = 1;
		//*start = NULL;
		copy_sz = (*p_buf_sz) - offset;
	}

	*start = page;
	if (*p_buf_sz)
		memcpy(page, (*p_buf) + offset, copy_sz);
	if (*eof) {
		free_buf_f(*p_buf);
		*p_buf = NULL;
		*p_buf_sz = 0;
	}
	// IERROR("B. page %p start %p offset %d count %d eof %d p_buf %p sz %d copysz %d\n",
	//  page, *start, (int)offset, count, *eof, *p_buf, *p_buf_sz, copy_sz );

	return copy_sz;
}				/* iser_procfs_handle_long_read */

/* ---------------------------------------------------------------------
 * PROC FS entity-related, and "iscsi" ENTRY - FUNCTIONS
 * ------------------------------------------------------------------ */

/** --------------------------------------------------------------------
 Function:   iser_control_read_proc
 
 Description:
    proc fs "control" read-callback
 -------------------------------------------------------------------- */
static int
iser_procfs_entity_port_read_proc(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
	struct iser_entity_t *p_entity;
	int n;

	p_entity = (struct iser_entity_t *) data;
	n = sprintf(page, "%d\n", p_entity->port);

	return n;
}				/* iser_control_read_proc */

/**
 * iser_procfs_add_entity - Adds a proc fs entry for a registered iSCSI entity 
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_add_entity(struct iser_entity_t *p_entity)
{
	struct proc_dir_entry *pent;
	char dir_name[256];

	ITRACE_ENTRY();

	sprintf(dir_name, "%s-%s",
		p_entity->provider_name,
		iser_get_entity_type_name(p_entity->type));

	// Create entity dir
	p_entity->pent = proc_mkdir(dir_name, iser_iscsi_pe);
	if (p_entity->pent == NULL) {
		IERROR("Could not create entity proc-fs dir: %s\n",
		       dir_name);
		ITRACE_EXIT();
		return -1;
	}
	// Create "port" sub-entry
	if (p_entity->type == ISER_ENTITY_TARGET) {
		pent =
		    create_proc_entry("port", S_IFREG | S_IRUGO | S_IWUSR,
				      p_entity->pent);
		if (pent != NULL) {
			pent->data = p_entity;
			pent->read_proc =
			    iser_procfs_entity_port_read_proc;
			pent->write_proc = NULL;
		} else {
			IERROR
			    ("Could not create entity's port proc-fs entry: %s/port\n",
			     dir_name);
			remove_proc_entry(dir_name, iser_iscsi_pe);
			p_entity->pent = NULL;
			ITRACE_EXIT();
			return -1;
		}
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_add_entity */

/**
 * iser_procfs_delete_entity - Deletes a proc fs entry for a registered iSCSI entity 
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_delete_entity(struct iser_entity_t *p_entity)
{
	char dir_name[256];

	ITRACE_ENTRY();

	if (p_entity->pent != NULL) {
		if (p_entity->type == ISER_ENTITY_TARGET) {
			remove_proc_entry("port", p_entity->pent);
		}
		sprintf(dir_name, "%s-%s",
			p_entity->provider_name,
			iser_get_entity_type_name(p_entity->type));
		remove_proc_entry(dir_name, iser_iscsi_pe);
		p_entity->pent = NULL;
	} else {
		IERROR("NULL procfs entity for: %s\n", dir_name);
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_delete_entity */

/** --------------------------------------------------------------------
 Function:   iser_control_read_proc
 
 Description:
    proc fs "control" read-callback
 -------------------------------------------------------------------- */
static int
iser_procfs_connection_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	struct iser_conn_t *p_iser_conn;
	int n = 0;
	char buf[256];

	p_iser_conn = (struct iser_conn_t *) data;

	page[0] = '\0';
	n += sprintf(buf,
		     "iSCSI handle: 0x%08lX\nkDAPL EP handle: 0x%08lX\n",
		     (unsigned long) p_iser_conn->iscsi_conn_h,
		     (unsigned long) p_iser_conn->ep_handle);
	strcat(page, buf);

	n += sprintf(buf, "State: %s\n",
		     iser_conn_get_state_name(p_iser_conn));
	strcat(page, buf);

	n += sprintf(buf,
		     "Post recv buffers:\n\tAllocated: %d\n\tInitial posted: %d\n\tCurrently posted: %d\n",
		     p_iser_conn->alloc_post_recv_bufs_num,
		     p_iser_conn->initial_post_recv_bufs_num,
		     atomic_read(&p_iser_conn->post_recv_buf_count));
	strcat(page, buf);

	n += sprintf(buf, "Post send buffers:\n\tCurrently posted: %d\n",
		     atomic_read(&p_iser_conn->post_send_buf_count));
	strcat(page, buf);

	n += sprintf(buf,
		     "Operational parameters:\n"
		     "\tInitiatorRecvDataSegmentLength: %d\n"
		     "\tTargetRecvDataSegmentLength: %d\n",
		     p_iser_conn->param.InitiatorRecvDataSegmentLength,
		     p_iser_conn->param.TargetRecvDataSegmentLength);
	strcat(page, buf);

	n += sprintf(buf,
		     "\tFirstBurstLength: %d\n"
		     "\tMaxBurstLength: %d\n",
		     p_iser_conn->param.FirstBurstLength,
		     p_iser_conn->param.MaxBurstLength);
	strcat(page, buf);

	n += sprintf(buf,
		     "\tImmediateData: %d\n"
		     "\tMaxOutstandingCmds: %d\n"
		     "\n",
		     p_iser_conn->param.ImmediateData,
		     p_iser_conn->max_outstand_cmds);
	strcat(page, buf);

	/*
	   "\tMaxConnections: %d\n\t"
	   "\tDefaultTime2Wait: %d\n\t"
	   "\tDefaultTime2Retain: %d\n\t"
	   "\tMaxOutstandingR2T: %d\n\t"
	   "\tInitialR2T: %d\n\t"
	   "\tDataPDUInOrder: %d\n\t"
	   "\tDataSequenceInOrder: %d\n\t"
	   "\tErrorRecoveryLevel: %d\n\t"
	 */
	/*
	   p_iser_conn->param.MaxConnections,
	   p_iser_conn->param.DefaultTime2Wait,
	   p_iser_conn->param.DefaultTime2Retain,
	   p_iser_conn->param.MaxOutstandingR2T,
	   p_iser_conn->param.InitialR2T,                 
	   p_iser_conn->param.DataPDUInOrder,
	   p_iser_conn->param.DataSequenceInOrder,
	   p_iser_conn->param.ErrorRecoveryLevel);
	 */

	return n;
}				/* iser_procfs_connection_read_proc */

/**
 * iser_procfs_add_connection - Adds a proc fs entry for a new iSER connection 
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_add_connection(struct iser_conn_t *p_iser_conn)
{
	struct proc_dir_entry *pent;

	ITRACE_ENTRY();

	pent = create_proc_entry(p_iser_conn->name,
				 S_IFREG | S_IRUGO | S_IWUSR,
				 p_iser_conn->p_entity->pent);
	if (pent != NULL) {
		pent->data = p_iser_conn;
		pent->read_proc = iser_procfs_connection_read_proc;
		pent->write_proc = NULL;
	} else {
		IERROR("Could not create connection proc-fs entry: %s\n",
		       p_iser_conn->name);
		ITRACE_EXIT();
		return -1;
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_add_connection */

/**
 * iser_procfs_delete_connection - Deletes a proc fs entry for a registered iSCSI entity 
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_delete_connection(struct iser_conn_t *p_iser_conn)
{
	ITRACE_ENTRY();

	if (p_iser_conn->p_entity->pent != NULL) {
		remove_proc_entry(p_iser_conn->name,
				  p_iser_conn->p_entity->pent);
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_delete_connection */

/** --------------------------------------------------------------------
 Function:   iser_version_read_proc
 
 Description:
    version read-callback
 -------------------------------------------------------------------- */
int
iser_version_read_proc(char *page, char **start, off_t off,
		       int count, int *eof, void *data)
{
	int len = 0;
	len +=
	    sprintf(page + len, "%s",
		    "$URL: http://web.voltaire.com/repos/iser-t/trunk/datamover/iser_procfs.c $, $Rev: 50 $\n");

	return len;
}				/* iser_version_read_proc */

/** --------------------------------------------------------------------
 Function:   iser_procfs_adaptor_read_proc
 
 Description:
    adaptor read-callback
 -------------------------------------------------------------------- */
int
iser_procfs_adaptor_read_proc(char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{

	struct iser_adaptor_t *p_iser_adaptor =
	    (struct iser_adaptor_t *) data;
	int len = 0;

	len += sprintf(page + len,
		       "dat_ia_query_results:\nadapter_name=%s\nvendor_name=%s\nhardware_version_major=%u\n",
		       p_iser_adaptor->dat_ia_attr.adapter_name,
		       p_iser_adaptor->dat_ia_attr.vendor_name,
		       p_iser_adaptor->dat_ia_attr.hardware_version_major);
	len += sprintf(page + len,
		       "hardware_version_minor=%u\nfirmware_version_major=%u\nfirmware_version_minor=%u\n",
		       p_iser_adaptor->dat_ia_attr.hardware_version_minor,
		       p_iser_adaptor->dat_ia_attr.firmware_version_major,
		       p_iser_adaptor->dat_ia_attr.firmware_version_minor);
	len += sprintf(page + len,
		       "ia_address_ptr=0x%p\nmax_eps=%u\nmax_dto_per_ep=%u\nmax_rdma_read_per_ep_in=%u\n",
		       p_iser_adaptor->dat_ia_attr.ia_address_ptr,
		       p_iser_adaptor->dat_ia_attr.max_eps,
		       p_iser_adaptor->dat_ia_attr.max_dto_per_ep,
		       p_iser_adaptor->dat_ia_attr.
		       max_rdma_read_per_ep_in);
	len +=
	    sprintf(page + len,
		    "max_rdma_read_per_ep_out=%u\nmax_evds=%u\nmax_evd_qlen=%u\nmax_iov_segments_per_dto=%u\n",
		    p_iser_adaptor->dat_ia_attr.max_rdma_read_per_ep_out,
		    p_iser_adaptor->dat_ia_attr.max_evds,
		    p_iser_adaptor->dat_ia_attr.max_evd_qlen,
		    p_iser_adaptor->dat_ia_attr.max_iov_segments_per_dto);
	len +=
	    sprintf(page + len,
		    "max_lmrs=%u\nmax_lmr_block_size=%lld\nmax_lmr_virtual_address=%llx\nmax_pzs=%u\n",
		    p_iser_adaptor->dat_ia_attr.max_lmrs,
		    p_iser_adaptor->dat_ia_attr.max_lmr_block_size,
		    p_iser_adaptor->dat_ia_attr.max_lmr_virtual_address,
		    p_iser_adaptor->dat_ia_attr.max_pzs);
	len +=
	    sprintf(page + len,
		    "max_mtu_size=%lld\nmax_rdma_size=%lld\nmax_rmrs=%u\n",
		    p_iser_adaptor->dat_ia_attr.max_mtu_size,
		    p_iser_adaptor->dat_ia_attr.max_rdma_size,
		    p_iser_adaptor->dat_ia_attr.max_rmrs);

	return len;
}				/* iser_procfs_adaptor_read_proc */

/**
 * iser_procfs_add_adaptor - Adds a proc fs entry for an Interface Adaptor
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_add_adaptor(struct iser_adaptor_t *p_iser_adaptor)
{
	struct proc_dir_entry *pent;

	ITRACE_ENTRY();

	if (iser_adaptors_pe != NULL && p_iser_adaptor != NULL) {
		pent = create_proc_entry(p_iser_adaptor->name,
					 S_IFREG | S_IRUGO | S_IWUSR,
					 iser_adaptors_pe);
		if (pent != NULL) {
			pent->data = p_iser_adaptor;
			pent->read_proc = iser_procfs_adaptor_read_proc;
			pent->write_proc = NULL;
		} else {
			IERROR
			    ("Could not create adaptor proc-fs entry: %s\n",
			     p_iser_adaptor->name);
			ITRACE_EXIT();
			return -1;
		}
	}
	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_add_adaptor */

/**
 * iser_procfs_delete_adaptor - Deletes a proc fs entry for an Interface Adaptor
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_delete_adaptor(struct iser_adaptor_t *p_iser_adaptor)
{
	ITRACE_ENTRY();

	if (iser_adaptors_pe != NULL && p_iser_adaptor != NULL) {
		remove_proc_entry(p_iser_adaptor->name, iser_adaptors_pe);
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_delete_adaptor */


/* ---------------------------------------------------------------------
 * PROC FS "bufpool" ENTRY - FUNCTIONS
 * ------------------------------------------------------------------ */

// #ifdef BUFF_POOL
// static char *p_bufpool_read_proc_buf = NULL;
// static int  bufpool_read_proc_buf_sz = 0;

/**
 * iser_procfs_bufpool_all_read_proc  - proc fs "bufpool/all" read-callback
 *
 * returns number of characters in output buffer
 */
// static int
// iser_procfs_bufpool_all_read_proc (char *page, char **start, off_t off,
//                                    int count, int *eof, void *data)
// {
//     int n;
//
//     ITRACE_ENTRY();
//
//     n = iser_procfs_handle_long_read(page,
//                                      start,
//                                      off,
//                                      count,
//                                      eof,
//                                      iser_buffer_pool_all_procfs_read_create,
//                                      iser_buffer_pool_all_procfs_read_free,
//                                      &p_bufpool_read_proc_buf,
//                                      &bufpool_read_proc_buf_sz);
//     ITRACE_EXIT();
//     return n;
// } /* iser_procfs_bufpool_all_read_proc */
// #endif

/**
 * iser_procfs_cleanup - Cleans up iSER proc fs entries and associated structures.
 * 
 * returns 0 on success, -1 on failure
 */
int iser_procfs_cleanup()
{
	ITRACE_ENTRY();

	if (iser_root_pe != NULL) {
		if (iser_iscsi_pe != NULL) {
			remove_proc_entry("iscsi", iser_root_pe);
			iser_iscsi_pe = NULL;
		}
		if (iser_adaptors_pe != NULL) {
			remove_proc_entry("adaptors", iser_root_pe);
			iser_adaptors_pe = NULL;
		}
		if (iser_bufpools_pe != NULL) {
			remove_proc_entry("all", iser_bufpools_pe);
			remove_proc_entry("bufpools", iser_root_pe);
			iser_bufpools_pe = NULL;
		}
		remove_proc_entry("trace", iser_root_pe);
		remove_proc_entry("version", iser_root_pe);

		remove_proc_entry(iser_procfs_path, NULL);
		iser_root_pe = NULL;
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_procfs_cleanup */


/**
 * iser_procfs_init - Initializes iSER proc fs entries and associated structures.
 * 
 * returns 0 on success, -1 on failure
 */
#undef IERROR
#if defined(__KERNEL__)
#define IERROR printk
#else
#include <stdio.h>
#define IERROR printf
#endif				/* __KENEL__ */

int iser_procfs_init()
{
	struct proc_dir_entry *pent;
	int ret_val = 0;

	//  ITRACE_ENTRY();

	// make proc directories        
	// ---------------------

	memset(iser_procfs_path, 0, sizeof(iser_procfs_path));
	/* Try to create iser in parent dir, if fails, try to create /proc/iser */
	iser_root_pe = proc_mkdir(ISER_DM_PROC_FS_PARENT "/iser", NULL);
	if (iser_root_pe != NULL) {
		strncpy(iser_procfs_path, ISER_DM_PROC_FS_PARENT "/iser",
			64);
	} else {
		iser_root_pe = proc_mkdir("iser", NULL);
		if (iser_root_pe != NULL) {
			strncpy(iser_procfs_path, "iser", 64);
		} else {
			IERROR
			    ("Could not create iser root procfs directory\n");
			ret_val = -1;
			goto procfs_init_exit;
		}
	}
	iser_root_pe->owner = THIS_MODULE;
	IINFO("Created iSER procfs directory: /proc/%s\n",
	      iser_procfs_path);

	iser_iscsi_pe = proc_mkdir("iscsi", iser_root_pe);
	if (iser_iscsi_pe == NULL) {
		IERROR("Could not create iser/iscsi procfs directory\n");
		ret_val = -1;
		goto procfs_init_exit;
	}
	iser_iscsi_pe->owner = THIS_MODULE;

	iser_adaptors_pe = proc_mkdir("adaptors", iser_root_pe);
	if (iser_adaptors_pe == NULL) {
		IERROR
		    ("Could not create iser/adaptors procfs directory\n");
		ret_val = -1;
		goto procfs_init_exit;
	}
	iser_adaptors_pe->owner = THIS_MODULE;

	iser_bufpools_pe = proc_mkdir("bufpools", iser_root_pe);
	if (iser_bufpools_pe == NULL) {
		IERROR
		    ("Could not create iser/bufpools procfs directory\n");
		ret_val = -1;
		goto procfs_init_exit;
	}
	iser_bufpools_pe->owner = THIS_MODULE;

	// make proc files
	// ---------------

	/* iser/version */
	pent =
	    create_proc_entry("version", S_IFREG | S_IRUGO, iser_root_pe);
	if (pent != NULL) {
		pent->data = NULL;
		pent->read_proc = iser_version_read_proc;
		pent->write_proc = NULL;
		pent->owner = THIS_MODULE;
	} else {
		IERROR("Could not create device proc entry: version\n");
		ret_val = -1;
		goto procfs_init_exit;
	}
	/* iser/trace */
	pent =
	    create_proc_entry("trace", S_IFREG | S_IRUGO | S_IWUSR,
			      iser_root_pe);
	if (pent != NULL) {
		pent->data = NULL;
		pent->read_proc = iser_trace_read_proc;
		pent->write_proc = iser_trace_write_proc;
		pent->owner = THIS_MODULE;
	} else {
		IERROR("Could not create device proc entry: trace\n");
		ret_val = -1;
		goto procfs_init_exit;
	}

// #ifdef BUFF_POOL
//     /* iser/bufpools/all */
//     pent = create_proc_entry("all", S_IFREG | S_IRUGO | S_IWUSR, iser_bufpools_pe);
//     if(pent != NULL) {
//         pent->data = NULL;
//         pent->read_proc  = iser_procfs_bufpool_all_read_proc;
//         pent->write_proc = NULL;
//         pent->owner = THIS_MODULE;
//     }
//     else {
//         IERROR("Could not create device proc entry: control\n");
//         ret_val = -1;
//         goto procfs_init_exit;
//     }
// #endif
	/* iser/control */
	/*pent = create_proc_entry("control", S_IFREG | S_IRUGO | S_IWUSR, iser_root_pe);
	   if(pent != NULL) {
	   pent->data = NULL;
	   pent->read_proc  = iser_control_read_proc; 
	   pent->write_proc = iser_control_write_proc;
	   pent->owner = THIS_MODULE;        
	   }
	   else {
	   IERROR("Could not create device proc entry: control\n");
	   ret_val = -1;
	   goto procfs_init_exit;
	   } */

      procfs_init_exit:
	// ITRACE_EXIT();
	return ret_val;
}				/* iser_procfs_init */
