/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_conn.h"
#include "iser_task.h"
#include "iser_dto.h"
#include "iser_target.h"
#include "iser_ofa.h"
#include "iser_memory.h"
#include "iser_utils.h"
#include "iser_trace.h"
#include "iser_global.h"
#include "iser_header.h"

#if defined(__KERNEL__)
#include <linux/module.h>
#include <linux/kernel.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif				/* __KERNEL__ */

iser_status_t
iser_target_connection_establish(iser_api_handle_t api_h,
				 iser_conn_handle_t iscsi_conn_h,
				 struct sockaddr_in * dst_addr,
				 struct sockaddr_in * src_addr);
iser_status_t
iser_target_connection_accept(iser_conn_request_t conn_req_h,
			      iser_conn_handle_t iscsi_conn_h, int accept);
iser_status_t
iser_target_allocate_connection_resources(iser_conn_handle_t conn_h,
					  struct iser_conn_resources_t
					  *conn_resources, int *);
iser_status_t iser_target_enable_datamover(iser_conn_handle_t iser_conn_h,
					   iser_transport_conn_t
					   tport_conn,
					   struct iser_send_pdu_t
					   *p_login_resp_pdu);
iser_status_t iser_target_put_data(void *iser_task_h,
				   struct iser_send_pdu_t *p_data_in_pdu,
				   int notify_enable);

iser_status_t
iser_target_get_data(void *iser_task_h,
		     struct iser_send_pdu_t *p_r2t_pdu, int notify_enable);

iser_status_t
iser_target_send_control(iser_conn_handle_t iscsi_conn_h,
			 struct iser_send_pdu_t *p_ctrl_pdu);

int
iser_target_dto_completion(struct iser_dto_t *p_dto,
			   unsigned long dto_xfer_len);

static int iser_target_alloc_ffp_recv_buffs(struct iser_conn_t *p_conn);

int iser_target_enter_hello_exchg(struct iser_conn_t *p_conn);

static int iser_target_recv_hello(struct iser_conn_t *p_conn,
				  struct iser_dto_t *p_dto);
static int iser_target_sent_helloreply(struct iser_conn_t *p_conn);

thread_ret iser_target_start_read_dto(void *void_dto);

/**
 * iser_target_register_entity - Sets the API functions specific to an target entity.
 *
 */
int
iser_target_register_entity(struct iser_entity_t *p_entity,
			    struct iser_api_t *api,
			    struct iser_ext_api_t *ext_api,
			    struct iser_ext_api_cb_t *ext_api_cb,
			    struct sockaddr *addr, size_t addr_len)
{
	struct sockaddr_in *in;

	ITRACE_ENTRY();

	in = (struct sockaddr_in *) addr;

	if (iser_listen(p_entity, addr, addr_len) != 0) {
#if !defined(__KERNEL__)
		IERROR("Failed to create PSP, %s:%d\n",
		       inet_ntoa(in->sin_addr), in->sin_port);
#endif /* !__KERNEL__ */
		ITRACE_EXIT();
		return -1;
	}

	api->allocate_connection_resources =
	    iser_target_allocate_connection_resources;
	api->notice_key_values = iser_notice_key_values;
	api->release_control = iser_release_control;
	api->enable_datamover = iser_target_enable_datamover;
	api->connection_terminate = iser_connection_terminate;
	api->deallocate_connection_resources =
	    iser_deallocate_connection_resources;

	api->send_control = iser_target_send_control;
	api->get_data = iser_target_get_data;
	api->put_data = iser_target_put_data;

	/* copy iSER API functions to iSCSI-supplied structure */
	if (ext_api != NULL) {
		ext_api->connection_establish =
		    iser_target_connection_establish;
		ext_api->connection_accept = iser_target_connection_accept;
	}

	/* copy extended API iSCSI callback functions to iSER */
	if (ext_api_cb != NULL) {
		p_entity->ext_api_cb.connection_establish_notify =
		    ext_api_cb->connection_establish_notify;
		p_entity->ext_api_cb.connection_request_notify =
		    ext_api_cb->connection_request_notify;
	}

	/* Set the DTO completion handler */
	p_entity->dto_completion_cb = iser_target_dto_completion;
	p_entity->enter_hello_exchg = iser_target_enter_hello_exchg;

	ITRACE_EXIT();
	return 0;
}				/* iser_target_register_entity */

/**
 * iser_target_connection_establish
 *
 * targets are not supposed to call connection_establish(), so this is a stub
 */
iser_status_t
iser_target_connection_establish(iser_api_handle_t api_h,
				 iser_conn_handle_t iscsi_conn_h,
				 struct sockaddr_in * dst_addr,
				 struct sockaddr_in * src_addr)
{
	return ISER_UNREGISTERED;
}				/* iser_target_connection_establish */

/**
 * iser_target_handle_connection_request - Handles a connection request
 *
 * returns 0 on success, -1 on failure
 */
int
iser_target_handle_connection_request(struct iser_conn_t *p_iser_conn,
				      struct sockaddr *local_addr)
{
	struct sockaddr_in *src_addr = NULL;
	int ret_val = 0;
	iser_status_t iser_ret;
	struct iser_entity_t *p_entity;
	int max_ctrl_per_cmd, moc;

	ITRACE_ENTRY();

	/*
	 * Aparantly the addresses returned by RDMA events are not set properly, so
	 * we really have no way of finding the entity that the connection request
	 * is coming in on... since there is only one entity now anyway, just grab
	 * that one.
	 *
	 * TODO: compare `local_addr' to the address that each
	 *       iser_global.entity[] is listening on to see which
	 *       entity the connection was received on.
	 */
	p_entity = &iser_global.entity[0];

	/* Add the connection to the entity's connections list */
	iser_entity_add_connection(p_entity, p_iser_conn);

	/*
	 * Compute the initial post receive buffer count and max
	 * outstanding commands.
	 */
	max_ctrl_per_cmd =
	    ISER_MAX_CTRLS_PER_CMD(p_iser_conn->param.FirstBurstLength,
				   p_iser_conn->param.
				   TargetRecvDataSegmentLength,
				   p_iser_conn->param.ImmediateData);
	if (p_iser_conn->param.FirstBurstLength == 0) {
		moc = ISER_MAX_OUTSTAND_CMDS;
	} else {
		moc = (ISER_TARGET_EP_MAX_RECV_DTOS
		       - ISER_TARGET_MAX_RX_MISC_PDUS)
		    / max_ctrl_per_cmd;
		moc = MIN(moc, ISER_MAX_OUTSTAND_CMDS);
	}
	p_iser_conn->initial_post_recv_bufs_num =
	    p_iser_conn->alloc_post_recv_bufs_num =
	    ISER_TARGET_MAX_RX_MISC_PDUS + (moc * max_ctrl_per_cmd);
	p_iser_conn->max_outstand_cmds = moc;

	TRACE(TRACE_ISER, "%s: initial_post_recv_bufs_num=%d\n",
	      __func__, p_iser_conn->initial_post_recv_bufs_num);

	/*
	 * Create the command queue and queue pair.
	 */
	ret_val = iser_create_cq_and_qp(p_iser_conn,
					/* receive queue depth
 	                                 * (4 extra for login) */
					p_iser_conn->
					initial_post_recv_bufs_num + 4,

					/* send queue depth */
					//ISER_MAX_OUTSTAND_CMDS,
					p_iser_conn->
					initial_post_recv_bufs_num + 4);

	if (ret_val != 0) {
		iser_conn_destroy(p_iser_conn);
		ret_val = -1;
		goto target_handle_connection_request_exit;
	}

	/*
	 * Call back iSCSI to notify it about a new connection request.
	 */
	atomic_inc(&p_iser_conn->p_adaptor->in_callback);
	iser_ret =
	    p_iser_conn->p_entity->ext_api_cb.
	    connection_request_notify((iser_conn_handle_t *) p_iser_conn,
				      (struct sockaddr_in *) local_addr,
				      src_addr);
	atomic_dec(&p_iser_conn->p_adaptor->in_callback);
	TRACE(TRACE_ISER, "Notifying of conneciton request %p\n",
	      p_iser_conn);
	if (iser_ret != ISER_SUCCESS) {
		IERROR("connection_request_notify() failed, status: %d\n",
		       iser_ret);
		iser_conn_destroy(p_iser_conn);
		ret_val = -1;
	}

      target_handle_connection_request_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_target_handle_connection_request */

/**
 * iser_target_connection_accept - Accept or reject a connection request.
 *
 * Accepts or rejects a connection request received previously
 * through connection_request_notify callback. This function is called
 * asynchronously, i.e. it may be called from within the
 * connection_request_notify callback or from a different thread.
 * Here iSCSI layer uses connection request handle to identify the
 * pending connection, and passes its own connection handle, and from
 * now on all API functions and callbacks for this connection use this
 * iSCSI-supplied handle.
 *
 * returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_ILLEGAL_PARAM)
 */
iser_status_t iser_target_connection_accept(iser_conn_request_t conn_req_h,
					    iser_conn_handle_t iscsi_conn_h,
					    int accept)
{
	struct iser_conn_t *p_iser_conn =
	    (struct iser_conn_t *) conn_req_h;
	iser_status_t ret_val = ISER_SUCCESS;
	char bufpool_name[128];

	ITRACE_ENTRY();

	if (p_iser_conn == NULL) {
		ret_val = ISER_ILLEGAL_PARAM;
		IERROR("bad handle\n");
		goto iser_connection_accept_exit;
	}

	if (accept) {
		/* Allocate post-receive buffers for the login phase */
		sprintf(bufpool_name,
			"%s-%s-login-pdu-data",
			p_iser_conn->p_entity->provider_name,
			iser_get_entity_type_name(p_iser_conn->p_entity->
						  type));

		/* allocate a post receive buffer for login */
		if (iser_post_receive_control
		    (ISER_LOGIN_PHASE_PDU_DATA_LEN, p_iser_conn) != 0) {
			iser_reject(p_iser_conn);
			ret_val = ISER_FAILURE;
			IERROR("Failed to post recv buffer for login, "
			       "disconnecting\n");
			goto iser_connection_accept_exit;
		}

		// Add the connection to the connections hash
		p_iser_conn->iscsi_conn_h = iscsi_conn_h;
		hash_add_iser_conn(p_iser_conn);
		TRACE(TRACE_ISER,
		      "Adding iscsi_conn_h %p to p_iser_conn %p\n",
		      iscsi_conn_h, p_iser_conn);

		// Change the connection status
		ISER_CONN_SET_STATE(p_iser_conn, ISER_CONN_PENDING);

		/*
		 * Accept the connection
		 * It is possible to get an ESTABLISHED event directly after
		 * calling iser_accpet, so everything needs to be ready for an
		 * ESTABLISHED event before calling this function.
		 */
		if (iser_accept(p_iser_conn, NULL) != 0) {
			ret_val = ISER_FAILURE;
			IERROR("iser_cr_accept != 0\n");
			hash_delete_iser_conn(p_iser_conn);
			goto iser_connection_accept_exit;
		}

	} else if (iser_reject(p_iser_conn) != 0) {
		IERROR("iser_cr_regect != 0\n");
		ret_val = ISER_FAILURE;
		goto iser_connection_accept_exit;
	}

      iser_connection_accept_exit:
	if (ret_val != ISER_SUCCESS && p_iser_conn != NULL) {
		iser_connection_terminate(p_iser_conn);
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_target_connection_accept */

/**
 * iser_target_put_data
 *
 * Parameters:
 * conn_h - IN - the pdu will be sent on this connection
 * p_data_in_pdu - IN - the pdu to send.
 * notify_enable - IN -0 - iSCSI does not wish to be notified
 * when this operation finishes successfully.
 * 1 - iSCSI wants to be notified when this operation finishes successfully.
 *
 * An iSCSI layer at the TARGET requests its local iSER datamover layer to
 * transmit the data for the SCSI Data-In PDU (identified by the data descriptor p_data)
 * to the peer iSCSI layer on the remote iSCSI node.
 * The notify_enable flag, when true, is used to request the local iSER datamover
 * layer to generate the eventual local completion notification to the iSCSI layer
 * for this put_data() invocation.
 */
iser_status_t
iser_target_put_data(void *iser_task_h,
		     struct iser_send_pdu_t * p_data_in_pdu,
		     int notify_enable)
{
	iser_status_t iser_ret = ISER_SUCCESS;
	union iser_pdu_bhs_t *p_bhs;
	struct iser_conn_t *p_iser_conn = NULL;
	struct iser_dto_t *p_rdma_dto = NULL;
	struct iser_task_t *p_iser_task = NULL;
	unsigned int itt;
	unsigned long buffer_offset;
	unsigned long desired_xfer_len;
	unsigned long total_data_len;

	ITRACE_ENTRY();

	/* Check the parameters correctness */
	if (p_data_in_pdu == NULL) {
		IERROR("NULL control DATA-IN PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto target_put_data_exit;
	}
	p_bhs = p_data_in_pdu->p_bhs;
	if (p_bhs == NULL) {
		IERROR("NULL BHS in control DATA-IN PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto target_put_data_exit;
	}
	itt = ntohl(p_bhs->dword.itt);
	p_iser_task = iser_task_h;
	TRACE(TRACE_ISER, "%s: setting p_iser_task=%p, iscsi_task_h=%p\n",
	      __func__, p_iser_task, p_iser_task->iscsi_task_h);
	if (p_iser_task == NULL) {
		IPANIC("NULL iser task\n");
		BUG();
	}
	p_iser_conn = p_iser_task->p_conn;

	if (atomic_read(&p_iser_conn->state) != ISER_CONN_UP) {
		IERROR("Connection is not up, p_conn: 0x%p\n",
		       p_iser_conn);
		iser_ret = ISER_FAILURE;
		goto target_put_data_exit;
	}

	if (!p_iser_task->dir[ISER_DIR_IN]) {
		IERROR
		    ("The current task, itt: %d, is not READ, can't put_data, "
		     "p_conn: 0x%p, p_task: 0x%p\n", itt, p_iser_conn,
		     p_iser_task);
		iser_ret = ISER_FAILURE;
		goto target_put_data_exit;
	}

	desired_xfer_len =
	    (unsigned long) ntohl(p_bhs->dword.length) & ISCSI_DSL_MASK;
	buffer_offset =
	    (unsigned long) ntohl(p_bhs->dword.
				  other[ISCSI_DATA_IN_FIELD_OFFSET]);
	if (buffer_offset + desired_xfer_len >
	    p_iser_task->data_len[ISER_DIR_IN]) {
		IPANIC
		    ("Buffer offset: %ld, and Xfer len: %ld exceed EDTL: %ld of READ cmd\n",
		     buffer_offset, desired_xfer_len,
		     p_iser_task->data_len[ISER_DIR_IN]);
	}

	/* Initialize RDMA-W DTO descriptor */
	p_rdma_dto = iser_alloc_dto(p_iser_conn, ISER_DTO_RDMA_W);
	if (p_rdma_dto == NULL) {
		IERROR("allocation of RDMA DTO descriptor failed\n");
		iser_ret = ISER_FAILURE;
		goto target_put_data_exit;
	}
	dto_set_task(p_rdma_dto, p_iser_task);
	/* Request completion notification */
	p_rdma_dto->notify_enable = notify_enable;

	p_rdma_dto->regd_vector_len = 0;
	/* Remote setup */
	p_rdma_dto->as.rdma.remote.rkey = p_iser_task->rdma_write_remote.rkey;
	p_rdma_dto->as.rdma.remote.lkey = p_iser_task->rdma_write_remote.lkey;
	p_rdma_dto->as.rdma.remote.addr = p_iser_task->rdma_write_remote.addr + buffer_offset;
	p_rdma_dto->as.rdma.remote.length = desired_xfer_len;

	TRACE(TRACE_ISER, "p_rdma_dto->as.rdma.remote.rkey=%d\n \
			p_rdma_dto->as.rdma.remote.lkey=%d\n \
			p_rdma_dto->as.rdma.remote.addr=%p\n \
			p_rdma_dto->as.rdma.remote.length=%lu\n",
	      p_rdma_dto->as.rdma.remote.rkey,
	      p_rdma_dto->as.rdma.remote.lkey,
	      p_rdma_dto->as.rdma.remote.addr,
	      (unsigned long) p_rdma_dto->as.rdma.remote.length);

	/* Set Data-SN as the RDMA DTO identifier */
	p_rdma_dto->as.rdma.SN = ntohl(p_bhs->dword.other[ISCSI_DATA_IN_FIELD_DATA_SN]);

	/* Register memory and add registered buffers to the DTO; Get total data size */
	if (iser_dto_add_local_memory(p_rdma_dto, &p_data_in_pdu->data.data_in.src_buf,	/* buffer to register */
				      p_iser_task->mem_vec, MAX_PHYS_VEC_IN_TASK, &total_data_len) != 0) {	/* total size of memory returned */
		IERROR("iser_dto_add_local_memory failed\n");
		iser_ret = ISER_FAILURE;
		goto target_put_data_exit;
	}
	TRACE(TRACE_ISER, "p_rdma_dto->as.rdma.remote.rkey=%d\n \
			p_rdma_dto->as.rdma.remote.lkey=%d\n \
			p_rdma_dto->as.rdma.remote.addr=%p\n \
			p_rdma_dto->as.rdma.remote.length=%lu\n",
	      p_rdma_dto->as.rdma.remote.rkey,
	      p_rdma_dto->as.rdma.remote.lkey,
	      p_rdma_dto->as.rdma.remote.addr,
	      (unsigned long) p_rdma_dto->as.rdma.remote.length);

	if (total_data_len < desired_xfer_len) {
		IPANIC
		    ("Total data length: %ld, less than desired Xfer Len: %ld, in READ cmd itt: %d, p_conn: 0x%p\n",
		     total_data_len, desired_xfer_len, itt, p_iser_conn);
	}

	/* Flush the memory which is going to be read by HCA during RDMA-W */
	TRACE(TRACE_ISER, "p_rdma_dto->as.rdma.remote.rkey=%d\n \
			p_rdma_dto->as.rdma.remote.lkey=%d\n \
			p_rdma_dto->as.rdma.remote.addr=%p\n \
			p_rdma_dto->as.rdma.remote.length=%lu\n",
	      p_rdma_dto->as.rdma.remote.rkey,
	      p_rdma_dto->as.rdma.remote.lkey,
	      p_rdma_dto->as.rdma.remote.addr,
	      (unsigned long) p_rdma_dto->as.rdma.remote.length);

	/* Start the DTO */
	if (iser_start_dto(p_rdma_dto) != 0) {
		IERROR("Failed to start RDMA-W DTO\n");
		iser_ret = ISER_FAILURE;
	}

      target_put_data_exit:
	if (iser_ret != ISER_SUCCESS) {
		if (p_rdma_dto != NULL)
			iser_conn_remove_dto(p_iser_conn, p_rdma_dto);
		if (iser_ret == ISER_ILLEGAL_PARAM) {
			// ISER_ERROR: just return an error code
			IERROR("Illegal parameters sent to %s\n",
			       __func__);
		} else if (p_iser_conn != NULL
			   && atomic_read(&p_iser_conn->state) ==
			                             ISER_CONN_UP) {
			// ISER_ERROR: drop the connection if this function failed because of an iSER layer error
			// (Any open tasks will be deleted during the disconnection)
			IERROR("RDMA-W DTO failed, dropping conn: 0x%p\n",
			       p_iser_conn);
			// We need to terminate the connection asyncronously
			iser_conn_async_terminate(p_iser_conn);
		}
	}
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_target_put_data */

/**
 * iser_target_get_data
 *
 * Parameters:
 * conn_h - IN - the pdu will be sent on this connection.
 * p_r2t_pdu - IN - data will be read to this buffer.
 * notify_enable - IN -0 - iSCSI does not wish to be notified
 * when this operation finishes successfully.
 * 1 - iSCSI wants to be notified when this operation finishes successfully.
 *
 *
 * An iSCSI layer at the TARGET requests its local iSER datamover
 * layer to transmit the data for the SCSI R2T PDU
 * (identified by the data descriptor p_data).
 * from the peer iSCSI layer on the remote iSCSI node.
 * The notify_enable flag, when true, is used to request the local iSER datamover
 * layer to generate the eventual local completion notification to the iSCSI layer
 * for this get_data() invocation.
 */
iser_status_t
iser_target_get_data(void *iser_task_h,
		     struct iser_send_pdu_t * p_r2t_pdu, int notify_enable)
{
	iser_status_t iser_ret = ISER_SUCCESS;
	thread_handle t_handle;
	union iser_pdu_bhs_t *p_bhs;
	struct iser_conn_t *p_iser_conn = NULL;
	struct iser_task_t *p_iser_task = NULL;
	struct iser_dto_t *p_rdma_dto = NULL;
	unsigned int itt;
	unsigned long buffer_offset;
	unsigned long desired_xfer_len;
	unsigned long total_data_len;

	ITRACE_ENTRY();

	/* Check the parameters correctness */
	if (p_r2t_pdu == NULL) {
		IERROR("NULL control R2T PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto target_get_data_exit;
	}
	p_bhs = p_r2t_pdu->p_bhs;
	if (p_bhs == NULL) {
		IERROR("NULL BHS in control R2T PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto target_get_data_exit;
	}
	p_iser_task = iser_task_h;
	if (p_iser_task == NULL) {
		IPANIC("NULL iser task\n");
		BUG();
	}
	itt = ntohl(p_bhs->dword.itt);
	p_iser_conn = p_iser_task->p_conn;

	if (atomic_read(&p_iser_conn->state) != ISER_CONN_UP) {
		IERROR("Connection is not up, p_conn: 0x%p\n",
		       p_iser_conn);
		iser_ret = ISER_FAILURE;
		goto target_get_data_exit;
	}

	if (!p_iser_task->dir[ISER_DIR_OUT]) {
		IERROR
		    ("The current task, itt: %d, is not WRITE, can't get_data, "
		     "p_conn: 0x%p, p_task: 0x%p\n", itt, p_iser_conn,
		     p_iser_task);
		iser_ret = ISER_FAILURE;
		goto target_get_data_exit;
	}

	/* flow control RDMA Read Requests */
	if (p_iser_conn->iser_read_depth == 0) {
		IERROR("%s: The current connection does not allow read "
		       "requests from the Target.  Can't get_data\n",
		       __func__);
		iser_ret = ISER_FAILURE;
		goto target_get_data_exit;
	}

	buffer_offset =
	    (unsigned long) ntohl(p_bhs->dword.
				  other[ISCSI_R2T_FIELD_OFFSET]);
	desired_xfer_len =
	    (unsigned long) ntohl(p_bhs->dword.
				  other[ISCSI_R2T_FIELD_DESIRED_CNT]);
	if (buffer_offset + desired_xfer_len >
	    p_iser_task->data_len[ISER_DIR_OUT]) {
		IERROR("task itt=%d p_iser_task=0x%p , bhs itt=%d\n",
		       p_iser_task->itt, p_iser_task, itt);
		IPANIC
		    ("Buffer offset: %ld, and Xfer len: %ld exceed EDTL: %ld of WRITE cmd\n",
		     buffer_offset, desired_xfer_len,
		     p_iser_task->data_len[ISER_DIR_OUT]);
	}

	/* Initialize RDMA-R DTO descriptor */
	p_rdma_dto = iser_alloc_dto(p_iser_conn, ISER_DTO_RDMA_R);
	if (p_rdma_dto == NULL) {
		IERROR("allocation of RDMA DTO descriptor failed\n");
		iser_ret = ISER_FAILURE;
		goto target_get_data_exit;
	}
	dto_set_task(p_rdma_dto, p_iser_task);
	/* Request completion notification */
	p_rdma_dto->notify_enable = notify_enable;
	p_rdma_dto->regd_vector_len = 0;
	/* Remote setup */
	p_rdma_dto->as.rdma.remote.rkey =
	    p_iser_task->rdma_read_remote.rkey;
	p_rdma_dto->as.rdma.remote.lkey =
	    p_iser_task->rdma_read_remote.lkey;
	p_rdma_dto->as.rdma.remote.addr =
	    p_iser_task->rdma_read_remote.addr + buffer_offset;
	p_rdma_dto->as.rdma.remote.length = desired_xfer_len;

	TRACE(TRACE_ISER, "p_rdma_dto->as.rdma.remote.rkey=%d\n \
			p_rdma_dto->as.rdma.remote.lkey=%d\n \
			p_rdma_dto->as.rdma.remote.addr=%p\n \
			p_rdma_dto->as.rdma.remote.length=%lu\n", p_rdma_dto->as.rdma.remote.rkey, p_rdma_dto->as.rdma.remote.lkey, p_rdma_dto->as.rdma.remote.addr, (unsigned long) p_rdma_dto->as.rdma.remote.length);
	/* Set R2T-SN as the RDMA DTO identifier */
	p_rdma_dto->as.rdma.SN =
	    ntohl(p_bhs->dword.other[ISCSI_R2T_FIELD_R2T_SN]);

	/* Register memory and add registered buffers to the DTO; Get total data size */
	if (iser_dto_add_local_memory(p_rdma_dto, &p_r2t_pdu->data.r2t.dst_buf,	/* buffer to register */
				      p_iser_task->mem_vec, MAX_PHYS_VEC_IN_TASK, &total_data_len) != 0) {	/* total size of memory returned */
		IERROR("iser_dto_add_local_memory failed\n");
		iser_ret = ISER_FAILURE;
		goto target_get_data_exit;
	}
	if (total_data_len < desired_xfer_len) {
		IPANIC
		    ("Total data length: %ld, less than desired Xfer len: %ld, in WRITE cmd itt: %d, p_conn: 0x%p\n",
		     total_data_len, desired_xfer_len, itt, p_iser_conn);
	}

	/* Invalidate the memory which is going to be overwritten by HCA during RDMA-R */

	atomic_inc(&p_iser_conn->outstanding_reads);
	if (atomic_read(&p_iser_conn->outstanding_reads)
	    > p_iser_conn->iser_read_depth) {
		atomic_dec(&p_iser_conn->outstanding_reads);
		spawn_thread(&t_handle, iser_target_start_read_dto,
			     p_rdma_dto, 0);
	} else {
		TRACE(TRACE_ISER, "%s: Starting DTO in same thread\n",
		      __func__);
		if (iser_start_dto(p_rdma_dto) != 0) {
			IERROR("%s: Failed to start RDMA-R DTO\n",
			       __func__);
			iser_ret = ISER_FAILURE;
			goto target_get_data_exit;
		}
	}

      target_get_data_exit:
	if (iser_ret != ISER_SUCCESS) {
		if (p_rdma_dto != NULL)
			iser_conn_remove_dto(p_iser_conn, p_rdma_dto);
		if (iser_ret == ISER_ILLEGAL_PARAM) {
			// ISER_ERROR: just return an error code
			IERROR("Illegal parameters sent to %s\n",
			       __func__);
		} else if (p_iser_conn != NULL
			   && atomic_read(&p_iser_conn->state) ==
			                                ISER_CONN_UP) {
			// ISER_ERROR: drop the connection if this function failed because of an iSER layer error
			// (Any open tasks will be deleted during the disconnection)
			IERROR("RDMA-R DTO failed, dropping conn: 0x%p\n",
			       p_iser_conn);
			// We need to terminate the connection asyncronously
			iser_conn_async_terminate(p_iser_conn);
		}
	}
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_target_get_data */

/**
 * \brief This thread function is used to delay the call to
 *        iser_start_dto() untill the RDMA Read queue is not full.
 */
thread_ret iser_target_start_read_dto(void *void_dto)
{
	struct iser_dto_t *p_rdma_dto;
	struct iser_conn_t *p_conn;

	p_rdma_dto = (struct iser_dto_t *) void_dto;
	p_conn = p_rdma_dto->p_conn;

	wait_event_interruptible(p_conn->read_depth_wait,
				 !iser_conn_read_depth_met(p_conn));
	atomic_inc(&p_conn->outstanding_reads);

	/* Start the DTO */
	TRACE(TRACE_ISER, "%s: Starting DTO in new thread\n", __func__);
	if (iser_start_dto(p_rdma_dto) != 0) {
		IERROR("%s: Failed to start RDMA-R DTO\n", __func__);
		return (thread_ret) ISER_FAILURE;
	}

	return (thread_ret) NULL;
}

/**
 * iser_target_send_control
 *
 * Parameters:
 * conn_h - IN - the pdu will be sent on this connection
 * p_ctrl_pdu - IN - the pdu to send.
 *
 *
 * An iSCSI layer at the TARGET requests its local iSER datamover layer
 * to transmit an iSCSI control-type PDU to the peer iSCSI layer
 * operating in the remote iSCSI node. The iSER datamover layer sends
 * the PDU after adding its own protocol header.
 */
iser_status_t iser_target_send_control(iser_conn_handle_t iscsi_conn_h,	/* IN */
				       struct iser_send_pdu_t * p_ctrl_pdu)
{				/* IN */
	iser_status_t iser_ret = ISER_SUCCESS;
	union iser_pdu_bhs_t *p_bhs;
	struct iser_conn_t *p_iser_conn = NULL;
	struct iser_dto_t *p_send_dto = NULL;
	struct iser_task_t *p_iser_task = NULL;
	unsigned int itt;
	unsigned int data_seg_len;
	unsigned char opcode;

	ITRACE_ENTRY();

	/* Check the parameters correctness */
	if (p_ctrl_pdu == NULL) {
		IERROR("NULL control PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;
	}
	p_bhs = p_ctrl_pdu->p_bhs;
	if (p_bhs == NULL) {
		IERROR("NULL BHS in control PDU descriptor\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;
	}

	/* Look up the connection */
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);

	/* print out the PDU */
	iser_pdu_print((char *) __func__, p_iser_conn,
		       (char *) p_bhs->buf,
		       &p_ctrl_pdu->data.tx.buf);

	if (p_iser_conn == NULL) {
		IERROR("Failed to find connection, iscsi_conn_h: %lX\n",
		       (unsigned long) iscsi_conn_h);
		iser_pdu_print((char *) __func__, NULL,
			       (char *) p_bhs->buf, NULL);
		iser_ret = ISER_INVALID_CONN;
		goto send_control_error;
	}
	if (atomic_read(&p_iser_conn->state) != ISER_CONN_UP) {
		IERROR
		    ("Failed to send_control as connection is not up, iscsi_conn_h=0x%lx, p_conn=%p\n",
		     (unsigned long) iscsi_conn_h, p_iser_conn);
		iser_pdu_print((char *) __func__, NULL,
			       (char *) p_bhs->buf, NULL);
		p_iser_conn = NULL;	/* Inhibits connection shutdown */
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	/* Allocate and fill send DTO descriptor, headers buffer and add it to the DTO */
	p_send_dto = iser_dto_send_create(p_iser_conn, p_bhs);
	if (p_send_dto == NULL) {
		IERROR("Failed to create send DTO, conn: 0x%p\n",
		       p_iser_conn);
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	/* Retrieve some BHS fields */
	itt = ntohl(p_bhs->dword.itt);
	data_seg_len =
	    ALIGN_TO_BIT((ntohl(p_bhs->dword.length) & ISCSI_DSL_MASK), 2);
	opcode = p_bhs->byte.opcode & ISCSI_OPCODE_MASK;
	ITRACE(ISER_TRACE_SEND_CONTROL,
	       "About to send itt: %u dsl: %u opcode: %X\n", itt,
	       data_seg_len, opcode);
	/* According to the PDU opcode ... */
	switch (opcode) {
	case ISCSI_OP_SCSI_RSP:{
			unsigned int rx_count;
			int i;

			/* find task in the hash */
			p_iser_task =
			    hash_find_iser_task(p_iser_conn, itt);
			if (p_iser_task == NULL) {
				IERROR("Task not found, itt:0x%0X\n", itt);
				iser_ret = ISER_INVALID_ITT;
				goto send_control_error;
			}

			/* Declare task completed */
			iser_task_set_status(p_iser_task,
					     ISER_TASK_STATUS_COMPLETED);

			/* Release all received buffers */
			iser_task_release_recv_buffers(p_iser_task);

			/* Re-post buffers that were used-up due to receive */
			rx_count =
			    iser_task_recvd_pdu_count_reset(p_iser_task);
			ITRACE(ISER_TRACE_SEND_CONTROL,
			       "Reposting %d buffers upon sending SCSI Response, itt: %d\n",
			       rx_count, itt);
			for (i = 0; i < rx_count; i++) {
				if (iser_post_receive_control
				    (p_iser_conn->param.
				     TargetRecvDataSegmentLength,
				     p_iser_conn) != 0) {
					IERROR("post_rcv_buff failed!\n");
					iser_ret = ISER_FAILURE;
					goto send_control_error;
				}
			}

			/* Now can start the DTO, its completion will release all task resources */
			dto_set_task(p_send_dto, p_iser_task);

			// If data segment present, Allocate data regd buffer and copy the user data
			if (data_seg_len > 0) {
				iser_dto_copy_send_data(p_send_dto,
							&p_ctrl_pdu->data.
							response.
							buf_status,
							data_seg_len);
			}
			hash_delete_iser_task(p_iser_task);
			if (hash_find_iser_task(p_iser_conn, itt) != NULL) {
				IPANIC
				    ("Task found after delete, itt=0x%0X\n",
				     itt);
			}

		}
		break;		/* ISCSI_OP_SCSI_RSP */

	case ISCSI_OP_LOGIN_RSP:
		// Post recv buffer unless it's a final request
		if (!(IS_SET_ISCSI_FLAG_LOGIN_TRANSIT(p_bhs->byte.flags) &&
		      IS_ISCSI_LOGIN_NSG_FULL_FEATURE(p_bhs->byte.
						      flags))) {
			if (iser_post_receive_control
			    (ISER_LOGIN_PHASE_PDU_DATA_LEN,
			     p_iser_conn) != 0) {
				IERROR
				    ("Failed to post recv buffer for the next LoginRequest PDU\n");
				iser_ret = ISER_FAILURE;
				goto send_control_error;
			}
		}
		// beware - no break here !!!

	case ISCSI_OP_NOP_IN:
	case ISCSI_OP_TASK_MGT_RSP:
	case ISCSI_OP_TEXT_RSP:
	case ISCSI_OP_LOGOUT_RSP:
	case ISCSI_OP_ASYNC:
	case ISCSI_OP_REJECT:
		// If data segment present, Allocate data regd buffer and copy the user data
		if (data_seg_len > 0) {
			iser_dto_copy_send_data(p_send_dto,
						&p_ctrl_pdu->data.tx.buf,
						data_seg_len);
		}
		break;

	default:
		IERROR("Illegal opcode = %d\n", opcode);
		iser_ret = ISER_ILLEGAL_PARAM;
		goto send_control_error;

	}			/* switch p_bhs->byte.opcode */

	if (iser_start_dto(p_send_dto) != 0) {
		IERROR("Failed to start send DTO\n");
		iser_ret = ISER_FAILURE;
		goto send_control_error;
	}

	ITRACE_EXIT();
	return ISER_SUCCESS;

      send_control_error:
	if (p_send_dto != NULL)
		iser_conn_remove_dto(p_send_dto->p_conn, p_send_dto);
	if (p_iser_conn != NULL) {
		/* drop the connection, any open tasks will be deleted during shutdown */
		IERROR("%s failed, dropping connection:0x%p\n", __func__,
		       p_iser_conn);
		iser_conn_async_terminate(p_iser_conn);
	}

	ITRACE_EXIT();
	return iser_ret;

}				/* iser_target_send_control */

/**
 * iser_target_create_recvd_task - Handle a received iSCSI command PDU
 *
 * returns 0 on success, -1 on failure
 */
struct iser_task_t *iser_target_create_recvd_task(struct iser_conn_t *p_iser_conn,	//!< IN -
						  unsigned int itt,	//!< IN -
						  unsigned char *p_iser_hdr,	//!< IN -
						  union iser_pdu_bhs_t *p_bhs)	//!< IN -
{
	struct iser_task_t *p_iser_task;
	unsigned long edtl;

	ITRACE_ENTRY();

	p_iser_task = iser_task_alloc(p_iser_conn, itt);
	if (p_iser_task != NULL) {
		/* Retrieve ExpectedDataTransferLength from the Command BHS */
		edtl =
		    (unsigned long) ntohl(p_bhs->dword.
					  other[ISCSI_CMD_FIELD_EDTL]);

		if (IS_SET_ISCSI_FLAG_WRITE_CMD(p_bhs->byte.flags)) {
			p_iser_task->data_len[ISER_DIR_OUT] = edtl;
			p_iser_task->dir[ISER_DIR_OUT] = 1;
			if (ISER_HDR_GET_BITS(p_iser_hdr, WSV)) {
				p_iser_task->rdma_read_remote.lkey = ntohl(ISER_HDR_W_RKEY(p_iser_hdr));
				p_iser_task->rdma_read_remote.rkey = ntohl(ISER_HDR_W_RKEY(p_iser_hdr));
				p_iser_task->rdma_read_remote.addr = (void*) (unsigned long) be64_to_cpu(ISER_HDR_W_VADDR(p_iser_hdr));
				// segment length will be set in get_data()
				TRACE(TRACE_ISER, "Got remote stag=%u "
				      "TO=%p for WRITE cmnd\n",
				      p_iser_task->rdma_read_remote.rkey,
				      p_iser_task->rdma_read_remote.addr);
			}
			if (IS_SET_ISCSI_FLAG_READ_CMD(p_bhs->byte.flags)) {
				IERROR
				    ("Bidirectional commands unsupported\n");
				return NULL;
			}
		} else if (IS_SET_ISCSI_FLAG_READ_CMD(p_bhs->byte.flags)) {
			p_iser_task->data_len[ISER_DIR_IN] = edtl;
			p_iser_task->dir[ISER_DIR_IN] = 1;
			if (ISER_HDR_GET_BITS(p_iser_hdr, RSV)) {
				p_iser_task->rdma_write_remote.rkey = ntohl(ISER_HDR_R_RKEY(p_iser_hdr));
				p_iser_task->rdma_write_remote.lkey = ntohl(ISER_HDR_R_RKEY(p_iser_hdr));
				p_iser_task->rdma_write_remote.addr = (void*) (unsigned long) be64_to_cpu(ISER_HDR_R_VADDR(p_iser_hdr));
				// segment length will be set in put_data()
				TRACE(TRACE_ISER, "Got remote stag=%u "
				      "TO=%p for READ cmnd\n",
				      p_iser_task->rdma_write_remote.rkey,
				      p_iser_task->rdma_write_remote.addr);
			}
		} else {
			TRACE(TRACE_ISER,
			      "Got command with NO DIRECTION.\n");
		}
		iser_task_set_status(p_iser_task,
				     ISER_TASK_STATUS_STARTED);
	} else
		ITRACE("%s\n", "Failed creating iSER Task");

	ITRACE_EXIT();
	return p_iser_task;
}				/* iser_target_create_recvd_task */

/**
 * iser_target_dto_completion - Handle a successful DTO completion at Target
 *
 * returns 0 on success, -1 on failure
 */
int iser_target_dto_completion(struct iser_dto_t *p_dto,	//!< IN - completed DTO descriptor
			       unsigned long dto_xfer_len)	//!< IN - DTO transfer length
{
	struct iser_conn_t *p_iser_conn;
	struct iser_task_t *p_iser_task = NULL;
	union iser_pdu_bhs_t *p_bhs;
	int should_notify;
	int ret_val = 0;
	unsigned int itt;
	unsigned char opcode;
	unsigned int rdma_sn;

	ITRACE_ENTRY();

	p_iser_conn = p_dto->p_conn;
	if (p_iser_conn == NULL) {
		IPANIC("NULL connection in p_dto: 0x%p\n", p_dto);
	}

	/*
	 * If we are expecting a HELLO message, it is the *only* message
	 * that we should receive; anything else is an error.
	 */
	if (p_dto->type == ISER_DTO_RCV
	    && p_iser_conn->hello_state == ISER_HELLO_WAIT) {
		TRACE(TRACE_ISER,
		      "%s: Handling receive in hello exchange\n",
		      __func__);
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		ret_val = iser_target_recv_hello(p_iser_conn, p_dto);
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		return ret_val;
	}

	if (p_dto->type == ISER_DTO_SEND
	    && p_iser_conn->hello_state == ISER_HELLO_REPL_SENT) {
		TRACE(TRACE_ISER, "%s: Hello reply sent\n", __func__);
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		ret_val = iser_target_sent_helloreply(p_iser_conn);
		iser_conn_remove_dto(p_dto->p_conn, p_dto);
		return ret_val;
	}

	switch (p_dto->type) {
	case ISER_DTO_RCV:	// handle received messages
		ITRACE(ISER_TRACE_RECV_DTO,
		       "Target, Data received p_dto=0x%p\n", p_dto);

		/* Number of posted receive buffers has decreased by one */
		iser_conn_dec_post_recv_cnt(p_iser_conn);
		ITRACE(ISER_TRACE_RECV_DTO,
		       "Decremented posted receive buffer count\n");

		/* Create descriptor for the received PDU */
		iser_dto_create_recv_pdu(p_dto, dto_xfer_len);
		p_bhs = p_dto->as.recv.pdu.p_bhs;

		ITRACE(ISER_TRACE_RECV_DTO, "Created receive PDU\n");
		/* Print the PDU */
		iser_pdu_print((char *) __func__, p_iser_conn,
			       (char *) p_bhs->buf,
			       &p_dto->as.recv.pdu.rx_data);
		TRACE(TRACE_RDMA_EVENTS, "Finished printing PDU\n");

		/* Retrieve ITT, note that it may be 0xFFFFFFFF when irrelevant */
		itt = ntohl(p_bhs->dword.itt);
		/* Retrieve opcode */
		opcode = p_bhs->byte.opcode & ISCSI_OPCODE_MASK;
		if (opcode == ISCSI_OP_SCSI_CMD) {
			unsigned char *p_iser_header;

			/* Headers buffer was added to the DTO descriptor as the first regd buf to use,
			   it contains the iSER header as well as the iSCSI BHS */
			p_iser_header = p_dto->regd[0]->virt_buf.p_buf;
			/* create a task */
			p_iser_task =
			    iser_target_create_recvd_task(p_iser_conn, itt,
							  p_iser_header,
							  p_bhs);
			if (p_iser_task == NULL) {
				// ToDo: This task cannot be handled. We should shut down the connection
				IPANIC("Failed to allocate a new task\n");
			}
			ITRACE(ISER_TRACE_RECV_CONTROL,
			       "Created task 0x%p for itt=%d\n",
			       p_iser_task, itt);
			if (p_bhs->byte.opcode & ISCSI_OP_IMMEDIATE) {
				// ToDo: not just one - for entire cmd; also should track the number
				// of imm. commands and if greater than max then drop(?) and post 1.
				if (iser_post_receive_control
				    (p_iser_conn->param.
				     TargetRecvDataSegmentLength,
				     p_iser_conn) != 0) {
					IERROR
					    ("Failed to repost recv buffer upon recv\n");
					ret_val = -1;
				}
			}
			dto_set_task(p_dto, p_iser_task);
			iser_task_recvd_pdu_count_inc(p_iser_task);
			if (p_dto->as.recv.pdu.rx_data.p_buf != NULL) {
				iser_regd_buff_reference(p_dto->regd[0]);
				list_add(&p_dto->regd[0]->free_upon_comp_list,
					 &p_iser_task->rcv_buff_list);
			}
			// Notify iSCSI layer
			iser_dto_get(p_dto);
			iser_conn_remove_dto(p_iser_conn, p_dto);
			ITRACE(ISER_TRACE_CTRL_NOTIFY,
			       "Control notify, DTO:0x%p, as PDU:0x%p\n",
			       p_dto, &p_dto->as.recv.pdu);

			atomic_inc(&p_iser_conn->p_adaptor->in_callback);
			p_iser_conn->p_entity->api_cb.
			    command_notify(p_iser_conn->iscsi_conn_h,
					   &p_dto->as.recv.pdu,
					   p_iser_task,
					   &p_iser_task->iscsi_task_h);
		} else if (opcode == ISCSI_OP_DATA_OUT) {
			if (itt != ISCSI_INVALID_ITT) {
				p_iser_task =
				    hash_find_iser_task(p_iser_conn, itt);
			} else {
				IPANIC
				    ("invalid ITT=%d need to close connection\n",
				     itt);
				BUG();
			}
			dto_set_task(p_dto, p_iser_task);
			iser_task_recvd_pdu_count_inc(p_iser_task);
			if (p_dto->as.recv.pdu.rx_data.p_buf != NULL) {
				iser_regd_buff_reference(p_dto->regd[0]);
				list_add(&p_dto->regd[0]->free_upon_comp_list,
					 &p_iser_task->rcv_buff_list);
			}
			// Notify iSCSI layer
			iser_dto_get(p_dto);
			iser_conn_remove_dto(p_iser_conn, p_dto);
			ITRACE(ISER_TRACE_CTRL_NOTIFY,
			       "Data out notify, DTO:0x%p, as PDU:0x%p\n",
			       p_dto, &p_dto->as.recv.pdu);

			TRACE(TRACE_ISER, "p_iser_conn=%p "
			      "\np_iser_conn->p_adaptor=%p "
			      "\n&p_iser_conn->p_adaptor->in_callback=%p "
			      "\np_iser_conn->p_entity=%p ",
			      p_iser_conn, p_iser_conn->p_adaptor,
			      &p_iser_conn->p_adaptor->in_callback,
			      p_iser_conn->p_entity);

			atomic_inc(&p_iser_conn->p_adaptor->in_callback);
			p_iser_conn->p_entity->api_cb.
			    data_out_notify(p_iser_task->iscsi_task_h,
					    &p_dto->as.recv.pdu);
		} else {
			// Post-Receive buffer for the next unsolicited PDU
			// ToDo: make the access to TTT field more elegant
			if ((opcode == ISCSI_OP_NOP_OUT &&
			     ntohl(p_bhs->dword.
				   other[ISCSI_DATA_OUT_FIELD_TTT]) ==
			     0xFFFFFFFF) || opcode == ISCSI_OP_TASK_MGT_REQ
			    || opcode == ISCSI_OP_TEXT_REQ
			    || opcode == ISCSI_OP_LOGOUT_REQ) {
				if (iser_post_receive_control
				    (p_iser_conn->param.
				     TargetRecvDataSegmentLength,
				     p_iser_conn) != 0) {
					IERROR
					    ("Failed to repost recv buffer upon recv\n");
					ret_val = -1;
				}
			}
			if (itt != ISCSI_INVALID_ITT) {
				p_iser_task =
				    hash_find_iser_task(p_iser_conn, itt);
			}
			if (p_iser_task != NULL) {
				// The rcv buffer belongs now to the task
				dto_set_task(p_dto, p_iser_task);
				iser_task_recvd_pdu_count_inc(p_iser_task);
			}
			// Notify iSCSI layer
			iser_dto_get(p_dto);
			iser_conn_remove_dto(p_iser_conn, p_dto);
			ITRACE(ISER_TRACE_CTRL_NOTIFY,
			       "Control notify, DTO:0x%p, as PDU:0x%p\n",
			       p_dto, &p_dto->as.recv.pdu);

			atomic_inc(&p_iser_conn->p_adaptor->in_callback);
			p_iser_conn->p_entity->api_cb.
			    control_notify(p_iser_conn->iscsi_conn_h,
					   &p_dto->as.recv.pdu);
		}

		atomic_dec(&p_iser_conn->p_adaptor->in_callback);
		TRACE(TRACE_ISER, "%s: out of callback\n", __func__);
		break;

	case ISER_DTO_SEND:	// handle sent messages
		ITRACE(ISER_TRACE_SEND_DTO,
		       "Target, Data sent p_dto=0x%p\n", p_dto);

		// One posted send DTO less
		atomic_dec(&p_iser_conn->post_send_buf_count);

		// Release the DTO descriptor
		iser_conn_remove_dto(p_dto->p_conn, p_dto);

		break;

	case ISER_DTO_RDMA_W:
		ITRACE(ISER_TRACE_PUT_DATA,
		       "RDMA Write finished, p_dto: 0x%p\n", p_dto);

		atomic_dec(&p_iser_conn->post_send_buf_count);

		p_iser_task = dto_get_task(p_dto);
		if (p_iser_task == NULL
		    || (void*) p_iser_task <= (void*) PAGE_SIZE) {
			IPANIC("NULL task in p_dto: 0x%p\n", p_dto);
		}
		should_notify = p_dto->notify_enable;
		rdma_sn = p_dto->as.rdma.SN;
		itt = p_iser_task->itt;

		// Notify iSCSI if requested by put_data()
		if (should_notify) {
			TRACE(TRACE_ISER, "%s: notifying iser_task=%p iscsi_task_h=%p\n",
			      __func__, p_iser_task, p_iser_task->iscsi_task_h);
			atomic_inc(&p_iser_conn->p_adaptor->in_callback);
			p_iser_conn->p_entity->api_cb.data_completion_notify(p_iser_task->iscsi_task_h, itt, rdma_sn);	/* Data-SN */
			atomic_dec(&p_iser_conn->p_adaptor->in_callback);
		}

		// Free DTO, this is supposed to de-register user memory
		// This also removes the task
		iser_conn_remove_dto(p_dto->p_conn, p_dto);

		break;

	case ISER_DTO_RDMA_R:
		ITRACE(ISER_TRACE_GET_DATA,
		       "RDMA Read finished, p_dto: 0x%p\n", p_dto);

		atomic_dec(&p_iser_conn->post_send_buf_count);

		atomic_dec(&p_iser_conn->outstanding_reads);
		wake_up_interruptible(&p_iser_conn->read_depth_wait);

		p_iser_task = dto_get_task(p_dto);
		if (p_iser_task == NULL) {
			IPANIC("NULL task in p_dto: 0x%p\n", p_dto);
		}
		should_notify = p_dto->notify_enable;
		rdma_sn = p_dto->as.rdma.SN;
		itt = p_iser_task->itt;

		// Notify iSCSI if requested by get_data()
		if (should_notify) {
			atomic_inc(&p_iser_conn->p_adaptor->in_callback);
			p_iser_conn->p_entity->api_cb.data_completion_notify(p_iser_task->iscsi_task_h, itt, rdma_sn);	/* R2T-SN */

			atomic_dec(&p_iser_conn->p_adaptor->in_callback);
		}

		// Free DTO, this is supposed to de-register user memory
		// This also frees the task
		iser_conn_remove_dto(p_dto->p_conn, p_dto);

		break;

	default:
		IPANIC("Unknown iSER DTO type: %d in p_dto: 0x%p\n",
		       p_dto->type, p_dto);
		break;

	}			/* switch p_dto->type */

	// target_dto_completion_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_target_dto_completion */


// ISER_ERROR: 12.1.3.1 (insufficient resources)
#define ASSIGN_OR_BUST(lval,def,rval,s) do {	    \
	if (lval == def) { lval = rval; }		    \
	else if (lval != rval) {			    \
		IERROR("%s(%d) does not match non-default(!=%d) value of parameter(%d)\n", \
			s,lval,def,rval);		    \
			iser_ret = ISER_ILLEGAL_PARAM;		    \
			goto allocate_connection_resources_exit;    \
	}} while(0)

/**
 * iser_target_allocate_connection_resources - iSER API. Implements Allocate_Connection_Resources primitive.
 *
 *  Performs exchange of API functions. Allocates additional receive
 *  buffers. Performs static memory registration if instructed to do so.
 *
 *  returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
*/
iser_status_t
iser_target_allocate_connection_resources(iser_conn_handle_t iscsi_conn_h,
					  struct iser_conn_resources_t
					  * conn_resources,
					  int *max_outstanding_cmds)
{
	struct iser_entity_t *p_entity;
	struct iser_conn_t *p_iser_conn;
	int uncache_buf_pools = 0;
	iser_status_t iser_ret;
	char bufpool_name[128];

	ITRACE_ENTRY();

	p_entity = (struct iser_entity_t *) conn_resources->api_h;
	if (p_entity == NULL) {
		IERROR("NULL API handle\n");
		iser_ret = ISER_ILLEGAL_PARAM;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto allocate_connection_resources_exit;
	}

	if (conn_resources->conn_type == ISER_TRANSPORT_CONN_IB) {
		// Find the connection
		p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
		if (p_iser_conn == NULL) {
			IERROR
			    ("Could not find iSER connection for iscsi_h = 0x%p\n",
			     iscsi_conn_h);
			iser_ret = ISER_INVALID_CONN;
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto allocate_connection_resources_exit;
		}

		if (p_iser_conn->p_entity != p_entity) {
			IERROR
			    ("Entity in conn_res does not match conn for iscsi_h = 0x%p\n",
			     p_iser_conn->iscsi_conn_h);
			iser_ret = ISER_ILLEGAL_PARAM;
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto allocate_connection_resources_exit;
		}
	} else {
		IERROR
		    ("This connection type unsupported, iscsi_h = 0x%p\n",
		     iscsi_conn_h);
		iser_ret = ISER_ILLEGAL_PARAM;
		goto allocate_connection_resources_exit;
	}

	/* Target */
	ASSIGN_OR_BUST(p_iser_conn->param.TargetRecvDataSegmentLength,
		       defaultTargetRecvDataSegmentLength,
		       conn_resources->max_recv_pdu_sz, "max_recv_pdu_sz");

	ASSIGN_OR_BUST(p_iser_conn->param.FirstBurstLength,
		       defaultFirstBurstLength,
		       conn_resources->first_burst_length,
		       "first_burst_length");

	if (p_iser_conn->param.TargetRecvDataSegmentLength == 0) {
		IPANIC("TargetRecvDataSegmentLength must not be zero\n");
	}

	*max_outstanding_cmds = p_iser_conn->max_outstand_cmds;

	uncache_buf_pools = 1;

	ITRACE(ISER_TRACE_CONN,
	       "Max outst. cmds: %d, FirstBurst: %d, MaxRecvPduSz: %d"
	       " (tgt: %d), Allocate PRBs: %d, Initial PRBs: %d\n",
	       p_iser_conn->max_outstand_cmds,
	       p_iser_conn->param.FirstBurstLength,
	       conn_resources->max_recv_pdu_sz,
	       p_iser_conn->param.TargetRecvDataSegmentLength,
	       p_iser_conn->alloc_post_recv_bufs_num,
	       p_iser_conn->initial_post_recv_bufs_num);

	/* Allocate post-receive buffers for the full-featured phase */
	snprintf(bufpool_name, ISER_BUF_POOL_NAME_SIZE,
		 "%s-%s-%s-post-recv", p_entity->provider_name,
		 iser_get_entity_type_name(p_entity->type),
		 p_iser_conn->name);

	iser_ret = iser_target_alloc_ffp_recv_buffs(p_iser_conn);


      allocate_connection_resources_exit:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_allocate_connection_resources */

/**
 * iser_target_enable_datamover - iSER API. Implements Enable_Datamover primitive.
 *
 * Posts receive buffers. In target, sends the Login-Response PDU.
 * Upgrades the connection to RDMA-assisted.
 *
 * returns iSER status (ISER_SUCCESS, ISER_FAILURE, ISER_INVALID_CONN)
 */
iser_status_t iser_target_enable_datamover(iser_conn_handle_t iscsi_conn_h,	//!< IN - iSCSI-supplied connection handle
					   iser_transport_conn_t tport_conn,	//!< IN - transport connection descriptor
					   struct iser_send_pdu_t * p_login_resp_pdu)	//!<IN - LoginResponse PDU
{
	iser_status_t iser_ret = ISER_SUCCESS;
	struct iser_conn_t *p_iser_conn;

	ITRACE_ENTRY();

	// Find the connection
	p_iser_conn = hash_find_iser_conn(iscsi_conn_h);
	if (p_iser_conn == NULL) {
		iser_ret = ISER_INVALID_CONN;
		// ISER_ERROR: 12.1.3.1 (insufficient resources)
		goto enable_data_mover_exit;
	}

	/* enter the hello exchange state *just* *before* sending our
	 * final login reply PDU to get rid of a race condition where
	 * we recv the Hello Message before we have correctly set the
	 * hello_state. -- eaburns */
	p_iser_conn->p_entity->enter_hello_exchg(p_iser_conn);

	if (p_login_resp_pdu != NULL) {
		ITRACE(ISER_TRACE_CONN,
		       "calling send control with login response\n");
		iser_ret =
		    iser_target_send_control(iscsi_conn_h,
					     p_login_resp_pdu);
		if (iser_ret != ISER_SUCCESS) {
			IERROR
			    ("Failed to send the final Login_Response PDU\n");
			// ISER_ERROR: 12.1.3.1 (insufficient resources)
			goto enable_data_mover_exit;
		}
		ITRACE(ISER_TRACE_CONN, "sent login response PDU ok\n");
	}

	wait_event_interruptible(p_iser_conn->conn_wait,
				 p_iser_conn->hello_state == ISER_HELLO_DONE);

      enable_data_mover_exit:
	ITRACE_EXIT();
	return iser_ret;
}				/* iser_enable_datamover */

/**
 * Allocate the recv buffers for the FFP of the connection.
 */
static int iser_target_alloc_ffp_recv_buffs(struct iser_conn_t *p_conn)
{
	int i;

	/* allocate a post receive buffer for the hello exchange */
	if (iser_post_receive_control(ISER_HELLO_LEN, p_conn) != 0) {
		iser_reject(p_conn);
		IERROR("Failed to post recv buffer for hello message, "
		       "disconnecting\n");
		return ISER_FAILURE;
	}

	/* allocate post receive buffers for FFP */
	for (i = 0; i < p_conn->initial_post_recv_bufs_num; i++) {
		if (iser_post_receive_control
		    (p_conn->param.TargetRecvDataSegmentLength,
		     p_conn) != 0) {
			IERROR("Failed to post recv buffer %d\n", i);
			// ToDo: recovery?
			ITRACE_EXIT();
			return ISER_FAILURE;
		}
	}
	ITRACE(ISER_TRACE_CONN, "Allocated %d post recv bufs "
	       "for FFP\n", i);

	return ISER_SUCCESS;
}

/**
 * \brief Enter the hello exchange after a RDMA connection has been
 *        established.
 */
int iser_target_enter_hello_exchg(struct iser_conn_t *p_conn)
{
	ITRACE_ENTRY();

	p_conn->hello_state = ISER_HELLO_WAIT;

	ITRACE_EXIT();
	return 0;
}

/**
 * \brief Handles receiving a hello message.
 * \param p_conn The connection receiving the message.
 * \param p_dto The DTO for the received hello message.
 */
static int iser_target_recv_hello(struct iser_conn_t *p_conn,
				  struct iser_dto_t *p_dto)
{
	int retval = 0;
	int iser_ird, max_ver, min_ver;
	unsigned char *pdu;
	struct iser_dto_t *p_reply_dto;
	unsigned char hello_repl[ISER_HELLO_REPLY_LEN];

	ITRACE_ENTRY();

	pdu = p_dto->regd[0]->virt_buf.p_buf;

	/*
	 * This MUST be a hello message.
	 */
	if (ISER_OPCODE(pdu) != ISER_OP_HELLO) {
		TRACE_ERROR("%s: received messages is not a Hello "
		      "message\n", __func__);
		ITRACE_EXIT();
		return -1;
	}

	/*
	 * Grab the version fields.
	 */
	iser_ird = ntohs(ISER_HDR_HELLO_IRD(pdu));
	max_ver = ISER_HDR_GET_BITS(pdu, HELLO_MAXVER);
	min_ver = ISER_HDR_GET_BITS(pdu, HELLO_MINVER);
	TRACE(TRACE_ISER, "%s: Hello message\n\tMaxVer=%d"
	      "\n\tMinVer=%d\n\tIRD=%d\n", __func__,
	      max_ver, min_ver, iser_ird);

	if (iser_ird < p_conn->max_rdma_read_req_out)
		p_conn->iser_read_depth = iser_ird;
	else
		p_conn->iser_read_depth = p_conn->max_rdma_read_req_out;

	/*
	 * Build the hello reply message.
	 */
	memset(&hello_repl, 0, sizeof(hello_repl));
	hello_repl[0] = ISER_OP_HELLO_REPLY << 4;	/* set the opcode */
	ISER_HDR_HELLO_REPLY_ORD(&hello_repl) =
	    htons(p_conn->iser_read_depth);
	ISER_HDR_SET_BITS(&hello_repl, HELLO_REPLY_REJ, 0x00);
	ISER_HDR_SET_BITS(&hello_repl, HELLO_REPLY_MAXVER,
			  ISER_CURRENT_VERSION);
	ISER_HDR_SET_BITS(&hello_repl, HELLO_REPLY_MINVER,
			  ISER_CURRENT_VERSION);

	/*
	 * Check if the version values above were ok.  If they were not,
	 * lets set the reject flag before sending the hello reply.
	 */
	if (max_ver != ISER_CURRENT_VERSION ||
	    min_ver != ISER_CURRENT_VERSION) {
		TRACE_ERROR("%s: %s has incompatible version\n", __func__,
		            max_ver != ISER_CURRENT_VERSION
		            ? "max_ver" : "min_ver");
		ISER_HDR_SET_BITS(&hello_repl, HELLO_REPLY_REJ, 1);
		retval = -1;
	}

	/*
	 * It is an iSER level error if the initiator proposes an
	 * iSER-IRD value > 0 while the target is proposing an iSER-ORD
	 * value of zero.
	 */
	if (iser_ird != 0 && p_conn->max_rdma_read_req_out == 0) {
		TRACE_ERROR("%s: iSER-IRD > 0, while iSER-ORD == 0\n",
		            __func__);
		ISER_HDR_SET_BITS(&hello_repl, HELLO_REPLY_REJ, 1);
		retval = -1;
	}

	p_reply_dto = iser_dto_raw_create(p_conn,
					  (unsigned char *) &hello_repl,
					  ISER_HELLO_REPLY_LEN);
	p_conn->hello_state = ISER_HELLO_REPL_SENT;
	iser_start_dto(p_reply_dto);

	ITRACE_EXIT();
	return retval;
}

/**
 * \brief Handles a DTO completion when finished sending a hello reply.
 */
static int iser_target_sent_helloreply(struct iser_conn_t *p_conn)
{
	ITRACE_ENTRY();

	p_conn->hello_state = ISER_HELLO_DONE;
	wake_up_interruptible(&p_conn->conn_wait);

	ITRACE_EXIT();
	return 0;
}
