/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_conn.h"
#include "iser_ofa.h"
#include "iser_utils.h"
#include "iser_memory.h"
#include "iser_trace.h"

/**
 * iser_task_alloc - Allocates and initializes a connection descriptor
 */
struct iser_task_t *iser_task_alloc(struct iser_conn_t *p_iser_conn,	//!< IN - iSER connection descriptor
				    unsigned int itt)	//!< IN - ITT
{
	struct iser_task_t *p_iser_task;

	ITRACE_ENTRY();

	// Yoni : remove after debug
	if (p_iser_conn == NULL) {
		ITRACE(ISER_TRACE_TASK,
		       "%s() p_iser_conn is NULL!!! !!!\n", __func__);
	}

	p_iser_task = my_kmem_cache_alloc(iser_global.task_mem_cache,
	                               MEM_ALLOC_FLAG);
	if (p_iser_task == NULL) {
		IERROR("Failed to allocate iser_task\n");
		goto task_alloc_exit;
	}
	// Yoni : remove after debug
	ITRACE(ISER_TRACE_TASK,
	       "%s() about to memset(0) task descriptor\n", __func__);

	memset(p_iser_task, 0, sizeof(struct iser_task_t));

	/* Init the task */
	p_iser_task->itt = itt;

	atomic_set(&p_iser_task->refs, 0);
	spin_lock_init(&p_iser_task->task_lock);
	atomic_set(&p_iser_task->rx_count, 0);

	INIT_LIST_HEAD(&p_iser_task->send_buff_list);
	INIT_LIST_HEAD(&p_iser_task->rcv_buff_list);

	/* Type "remote" makes RMR context field valid */
	p_iser_task->rdma_write_local.type = ISER_REGD_BUF_UNUSED;
	p_iser_task->rdma_read_local.type = ISER_REGD_BUF_UNUSED;

	/* Init the ref count for these buffers */
	atomic_set(&p_iser_task->rdma_write_local.ref_count, 0);
	atomic_set(&p_iser_task->rdma_read_local.ref_count, 0);

	/* Set the task status */
	atomic_set(&p_iser_task->status, ISER_TASK_STATUS_INIT);

	/* Add task to connection */
	iser_conn_add_task(p_iser_conn, p_iser_task);

	/* Add the task to the tasks hash table */
	hash_add_iser_task(p_iser_task);

	ITRACE(ISER_TRACE_TASK, "Allocated iSER task: 0x%p\n",
	       p_iser_task);

      task_alloc_exit:
	ITRACE_EXIT();
	return p_iser_task;
}				/* iser_task_alloc */

/**
 * iser_task_release_send_buffers - Frees all sent buffers of a task (upon completion)
 */
void iser_task_release_send_buffers(struct iser_task_t *p_iser_task)	//!< IN - iSER task descriptor
{
	unsigned long flags;
	struct iser_regd_buf_t *p_regd_buf;
	struct list_head *node;
	int tries = 0;

	ITRACE_ENTRY();

	ITRACE(ISER_TRACE_TASK,
	       "Releasing send buffs for iSER task: 0x%p\n", p_iser_task);

	// Free all sent buffers from the list
	spin_lock_irqsave(&p_iser_task->task_lock, flags);
	while (!list_empty(&p_iser_task->send_buff_list)) {
		int release_deferred;

		node = p_iser_task->send_buff_list.next;

		// Get the next send buffer & remove it from the list
		p_regd_buf = container_of(node, struct iser_regd_buf_t,
			                  free_upon_comp_list);
		list_del(&p_regd_buf->free_upon_comp_list);
		spin_unlock_irqrestore(&p_iser_task->task_lock, flags);

		if (iser_regd_buff_release(p_regd_buf, &release_deferred) <
		    0) {
			IPANIC
			    ("Failed to release send buffer after task complete, task: 0x%p, itt: %d - error occured\n",
			     p_iser_task, p_iser_task->itt);
		}
		if (release_deferred) {
			IPANIC
			    ("Failed to release send buffer after task complete, task: 0x%p, itt: %d - %d references remain\n",
			     p_iser_task, p_iser_task->itt,
			     release_deferred);
			IERROR("calling schedule\n");
			tries++;
			schedule();
		}
		spin_lock_irqsave(&p_iser_task->task_lock, flags);

	}
	spin_unlock_irqrestore(&p_iser_task->task_lock, flags);
	if (tries)
		IERROR("Succeeded to release send buffer after %d tries\n",
		       tries);

	ITRACE_EXIT();
}				/* iser_task_release_send_buffers */

/**
 * iser_task_release_recv_buffers - Frees all receive buffers of a task (upon completion)
 */
void iser_task_release_recv_buffers(struct iser_task_t *p_iser_task)	//!< IN - iSER task descriptor
{
	unsigned long flags;
	struct list_head *node;
	struct iser_regd_buf_t *p_regd_buf;

	ITRACE_ENTRY();

	// Free all rcv buffers from the list
	spin_lock_irqsave(&p_iser_task->task_lock, flags);
	while (!list_empty(&p_iser_task->rcv_buff_list)) {
		int release_deferred;

		node = p_iser_task->rcv_buff_list.next;

		// Get the next rcv buffer & remove it from the list
		p_regd_buf = container_of(node, struct iser_regd_buf_t,
			       free_upon_comp_list);
		list_del(&p_regd_buf->free_upon_comp_list);
		spin_unlock_irqrestore(&p_iser_task->task_lock, flags);

		if (iser_regd_buff_release(p_regd_buf, &release_deferred) <
		    0) {
			IPANIC
			    ("Failed to release recv buffer after task complete, task: 0x%p, itt: %d - error occured\n",
			     p_iser_task, p_iser_task->itt);
		}
		if (release_deferred) {
			IPANIC
			    ("Failed to release recv buffer after task complete, task: 0x%p, itt: %d - %d references remain\n",
			     p_iser_task, p_iser_task->itt,
			     release_deferred);
		}

		spin_lock_irqsave(&p_iser_task->task_lock, flags);
	}
	spin_unlock_irqrestore(&p_iser_task->task_lock, flags);

	ITRACE_EXIT();
}				/* iser_task_release_recv_buffers */


/**
 * iser_task_recvd_pdu_count_inc - Increments counter of received pdus
 */
void iser_task_recvd_pdu_count_inc(struct iser_task_t *p_iser_task)
{
	ITRACE_ENTRY();

	if (p_iser_task == NULL)
		IPANIC("NULL task descriptor\n");

	atomic_inc(&p_iser_task->rx_count);

	ITRACE_EXIT();
}				/* iser_task_recvd_pdu_count_inc */

/**
 * iser_task_recvd_pdu_count_reset - Decrements counter of received pdus
 */
int iser_task_recvd_pdu_count_reset(struct iser_task_t *p_iser_task)	//!< IN - iSER task descriptor
{
	int ret_val = 0;

	ITRACE_ENTRY();

	if (p_iser_task == NULL)
		IPANIC("NULL task descriptor\n");

	ret_val = atomic_read(&p_iser_task->rx_count);
	atomic_set(&p_iser_task->rx_count, 0);

	ITRACE_EXIT();
	return ret_val;
}				/* iser_task_recvd_pdu_count_reset */

/**
 * iser_task_set_status - Sets atask status
 */
void iser_task_set_status(struct iser_task_t *p_iser_task,	//!< IN - iSER task descriptor
			  enum iser_task_status_t status)	//!< IN - new task status
{
	ITRACE_ENTRY();

	if (p_iser_task == NULL)
		IPANIC("NULL task descriptor\n");

	atomic_set(&p_iser_task->status, status);

	ITRACE_EXIT();
}				/* iser_task_set_status */

/**
 * iser_task_free - Frees all task resources
 */
void iser_task_free(struct iser_task_t *p_iser_task)	//!< IN - task descriptor
{
	ITRACE_ENTRY();

	if (p_iser_task == NULL) {
		IPANIC("NULL task descriptor\n");
	}
	ITRACE(ISER_TRACE_TASK, "Freeing iSER task: 0x%p, itt=%#x\n",
	       p_iser_task, p_iser_task->itt);

	BUG_ON(atomic_read(&p_iser_task->refs) != 0);

	// Remove the task from the hash, if it is found there.
	if (hash_find_iser_task(p_iser_task->p_conn, p_iser_task->itt))
		hash_delete_iser_task(p_iser_task);

	// Remove from the connection's task list
	if (p_iser_task->p_conn != NULL) {
		iser_conn_delete_task(p_iser_task->p_conn, p_iser_task);
	}
	// Detach from tasks hash
	//    hash_delete_iser_task(p_iser_task);
	// Free all rcv buffers from the list
	iser_task_release_recv_buffers(p_iser_task);
	// Free all sent buffers from the list
	iser_task_release_send_buffers(p_iser_task);

	// Return the memory to slab
	my_kmem_cache_free(iser_global.task_mem_cache, p_iser_task);

	ITRACE_EXIT();
}				/* iser_task_free */

/**
 * Gets a reference to the given task.
 */
void iser_task_get(struct iser_task_t *p_task)
{
	TRACE(TRACE_ISER, "Getting task %p\n", p_task);
	atomic_inc(&p_task->refs);
}

/**
 * Releases a reference to the given task.
 * When all of the refs are depleted, the task is freed.
 */
void iser_task_put(struct iser_task_t *p_task)
{
	TRACE(TRACE_ISER, "Putting task %p\n", p_task);
	if (atomic_dec_and_test(&p_task->refs)
	    && atomic_read(&p_task->status) == ISER_TASK_STATUS_COMPLETED) {
		iser_task_free(p_task);
	}
}
