/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2007-2008 UNH InterOperability Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 */

#if defined(__KERNEL__)
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/io.h>		/* struct iovec */
#include <asm/scatterlist.h>	/* struct scatterlist */
#endif				/* __KERNEL__ */

#include "../common/scatterlist.h"

#include "iser.h"
#include "iser_conn.h"
#include "iser_utils.h"
#include "iser_platform.h"
#include "iser_trace.h"


/* ---------------------------------------------------------------------
 * GENERIC HASH MANAGEMENT
 * ------------------------------------------------------------------ */

#define HASH_TABLE_MASK 0xFF

/**
 * hash_init - Initializes a hash table
 */
void hash_init(struct hash_table_t *hash_table)	//!< IN - hash table to initialize
{
	int i;

	ITRACE_ENTRY();

	for (i = 0; i < HASH_TABLE_SIZE; i++)
		INIT_LIST_HEAD(&(hash_table->bucket_head[i]));

	spin_lock_init(&hash_table->lock);

	ITRACE_EXIT();
}				/* hash_init */

/**
 * hash_func - Compute hash value for a given 32-bit key. XOR all bytes.
 *
 * returns hash value
 */
static int hash_func(int key)	//!< IN - key to be hashed
{
	// Shift-right the 32-bit integer by 0,1,2,3 bytes, then XOR and
	// mask out the LSB, which is the resulting hash value.
	int hash_val =
	    (key ^ (key >> 8) ^ (key >> 16) ^ (key >> 24)) &
	    HASH_TABLE_MASK;

	return hash_val;
}				/* hash_func */

/* ---------------------------------------------------------------------
 * ISER TASK-SPECIFIC HASH MANAGEMENT
 * ------------------------------------------------------------------ */

/**
 * hash_add_iser_task - Add iSER task descriptor to the hash table
 */
void hash_add_iser_task(struct iser_task_t *iser_task)	//!< IN - task to add
{
	unsigned long flags;
	int hash_val;
	int itt = iser_task->itt;
	struct iser_task_t *double_iser_task = NULL;
	struct list_head *p_bucket;
	struct list_head *p_list;

	ITRACE_ENTRY();

	iser_task_get(iser_task);

	hash_val = hash_func(itt ^ (u32) (long) iser_task->p_conn);
	p_bucket = &(iser_global.task_hash.bucket_head[hash_val]);

	ITRACE(ISER_TRACE_HASHTABLES,
	       "p_task: 0x%p, p_conn: 0x%p, itt: %d, hash_val = %d\n",
	       iser_task, iser_task->p_conn, itt, hash_val);

	spin_lock_irqsave(&iser_global.task_hash.lock, flags);
	INIT_LIST_HEAD(&iser_task->hash_list);
	p_list = p_bucket->next;
	while (p_list != p_bucket) {
		double_iser_task =
		    list_entry(p_list, struct iser_task_t, hash_list);
		if (double_iser_task->itt == itt
		    && double_iser_task->p_conn == iser_task->p_conn) {
			break;
		}
		double_iser_task = NULL;
		p_list = p_list->next;
	}
	if (double_iser_task != NULL) {
		IPANIC("Double task with same itt:0x%X in hash\n", itt);
	}
	list_add_tail(&iser_task->hash_list, p_bucket);
	spin_unlock_irqrestore(&iser_global.task_hash.lock, flags);

	ITRACE_EXIT();
}				/* hash_add_iser_task */

/**
 * hash_find_iser_task - Find am iSER task descriptor in task hash table.
 *
 *  Use connection descriptor and ITT combination as the key.
 *
 * returns found task decriptor or NULL
 */
struct iser_task_t *hash_find_iser_task(struct iser_conn_t *iser_conn,	//!< IN - part of hash key
					u32 itt)	//!< IN -  part of hash key
{
	unsigned long flags;
	int hash_val;
	struct list_head *p_bucket = NULL;
	struct list_head *p_list;
	struct iser_task_t *iser_task = NULL;

	ITRACE_ENTRY();

	hash_val = hash_func(itt ^ (u32) (long) iser_conn);
	p_bucket = &(iser_global.task_hash.bucket_head[hash_val]);

	//ITRACE(ISER_TRACE_TASK,"Trying to get task lock\n");
	spin_lock_irqsave(&iser_global.task_hash.lock, flags);
	//ITRACE(ISER_TRACE_TASK,"Got task lock\n");
	p_list = p_bucket->next;
	while (p_list != p_bucket) {
		iser_task =
		    list_entry(p_list, struct iser_task_t, hash_list);
		if (iser_task->itt == itt
		    && iser_task->p_conn == iser_conn) {
			break;
		}
		iser_task = NULL;
		p_list = p_list->next;
	}
	spin_unlock_irqrestore(&iser_global.task_hash.lock, flags);
	//ITRACE(ISER_TRACE_TASK,"Released task lock\n");

	ITRACE(ISER_TRACE_HASHTABLES,
	       "p_conn: 0x%p, itt: %d, hash_val = %d, found p_task: 0x%p\n",
	       iser_conn, itt, hash_val, iser_task);

	ITRACE_EXIT();
	return iser_task;
}				/* hash_find_iser_task */

/**
 * hash_delete_iser_task - Removes an iSER task from hash
 */
void hash_delete_iser_task(struct iser_task_t *iser_task)	//!< IN - task descriptor
{
	unsigned long flags;
	ITRACE_ENTRY();

	spin_lock_irqsave(&iser_global.task_hash.lock, flags);
	if (!list_empty(&iser_task->hash_list)) {
		list_del_init(&iser_task->hash_list);
		ITRACE(ISER_TRACE_HASHTABLES,
		       "deleted task:0x%p p_conn:0x%p itt:0x%0X\n",
		       iser_task, iser_task->p_conn, iser_task->itt);
	} else {
		ITRACE(ISER_TRACE_HASHTABLES,
		       "task:0x%p not in hash to delete\n", iser_task);
	}
	spin_unlock_irqrestore(&iser_global.task_hash.lock, flags);

	iser_task_put(iser_task);

	ITRACE_EXIT();
}				/* hash_delete_iser_task */

/* ---------------------------------------------------------------------
 * ISER CONNECTION-SPECIFIC HASH MANAGEMENT
 * ------------------------------------------------------------------ */

/**
 * hash_add_iser_conn - Add iSER connection descriptor to the connections hash table
 */
void hash_add_iser_conn(struct iser_conn_t *iser_conn)	//!< IN - connection to add
{
	unsigned long flags;
	int hash_val;

	ITRACE_ENTRY();

	TRACE(TRACE_ISER, "%s: adding iser_conn=%p iscsi_conn_h=%p\n",
	      __func__, iser_conn, iser_conn->iscsi_conn_h);

	iser_conn_get(iser_conn);
	hash_val = hash_func((u32) (long) iser_conn->iscsi_conn_h);

	spin_lock_irqsave(&iser_global.conn_hash.lock, flags);
	INIT_LIST_HEAD(&iser_conn->hash_list);
	list_add_tail(&iser_conn->hash_list,
		      &(iser_global.conn_hash.bucket_head[hash_val]));
	spin_unlock_irqrestore(&iser_global.conn_hash.lock, flags);

	ITRACE_EXIT();
}				/* hash_add_iser_conn */

/**
 * hash_find_iser_conn - Find am iSER connection descriptor in the connections hash table.
 *
 *  Use connection handle supplied by iSCSI as the key.
 *
 * returns found connection decriptor or NULL.
 */
struct iser_conn_t *hash_find_iser_conn(iser_conn_handle_t iscsi_conn_h)	//!< IN - issci-supplied connection handle
{
	unsigned long flags;
	int hash_val;
	struct list_head *p_bucket;
	struct list_head *p_list;
	struct iser_conn_t *iser_conn = NULL;

	ITRACE_ENTRY();

	hash_val = hash_func((u32) (long) iscsi_conn_h);
	p_bucket = &(iser_global.conn_hash.bucket_head[hash_val]);

	TRACE(TRACE_ISER, "%s: looking for iscsi_conn_h=%p\n",
	      __func__, iscsi_conn_h);

	spin_lock_irqsave(&iser_global.conn_hash.lock, flags);
	p_list = p_bucket->next;
	while (p_list != p_bucket) {
		iser_conn =
		    list_entry(p_list, struct iser_conn_t, hash_list);
		TRACE(TRACE_ISER,
		      "%s: iser_conn->iscsi_conn_h=%p, looking for %p\n",
		      __func__, iser_conn->iscsi_conn_h, iscsi_conn_h);
		if (iser_conn->iscsi_conn_h == iscsi_conn_h)
			break;
		iser_conn = NULL;
		p_list = p_list->next;
	}
	spin_unlock_irqrestore(&iser_global.conn_hash.lock, flags);

	TRACE(TRACE_ISER, "%s: found iser_conn=%p for iscsi_conn_h=%p\n",
	      __func__, iser_conn, iscsi_conn_h);

	ITRACE_EXIT();
	return iser_conn;
}				/* hash_find_iser_conn */

/** --------------------------------------------------------------------
 Function:   hash_delete_iser_conn

 Parameters:
     iser_conn - IN - connection descriptor

 Description:
     Removes an iSER connection from the connections hash
 -------------------------------------------------------------------- */
void hash_delete_iser_conn(struct iser_conn_t *iser_conn)
{
	unsigned long flags;
	ITRACE_ENTRY();

	spin_lock_irqsave(&iser_global.conn_hash.lock, flags);
	if (!list_empty(&iser_conn->hash_list)) {
		list_del_init(&iser_conn->hash_list);
	}
	spin_unlock_irqrestore(&iser_global.conn_hash.lock, flags);
	iser_conn_put(iser_conn);

	ITRACE_EXIT();
}				/* hash_delete_iser_conn */

/**
 * iser_get_data_total_length - Calculates the total buffer length for different buffer
 * descriptor types
 *
 * returns total buffer length in bytes
 */
unsigned long iser_get_data_total_length(struct iser_data_buf_t *p_data)
{
	ITRACE_ENTRY();

	if (p_data == NULL) {
		IPANIC("NULL data buffer descriptor\n");
	}
	switch (p_data->type) {
	case ISER_BUF_TYPE_SINGLE:
		return p_data->size;

	case ISER_BUF_TYPE_IOVEC_VIRT: {
			unsigned long total_len = 0;
			struct iovec *p_iovec =
			    (struct iovec *) p_data->p_buf;
			int i;

			if (p_iovec == NULL) {
				IPANIC("NULL data buffer's iovec\n");
			}
			for (i = 0; i < p_data->size; i++) {
				total_len +=
				    (unsigned long) p_iovec[i].iov_len;
			}
			return total_len;
		}

	case ISER_BUF_TYPE_SCATTERLIST:{
			int i;
			unsigned long total_len = 0;
			struct scatterlist *p_sg;

			if (!p_data->p_buf)
				IPANIC("%s: NULL data buffer\n", __func__);
			for_each_sg (p_data->p_buf, p_sg, p_data->size, i)
				total_len += (unsigned long) scatterlist_length(p_sg);

			return total_len;
		}

	default:
		IPANIC("Unsupported buffer type: %d\n", p_data->type);
		return 0;
	}

	ITRACE_EXIT();
}				/* iser_get_data_total_length */

/**
 * iser_iovec_virt_to_mem_t - Translates virtual addresses in a struct iovec array to
 *          an iser_mem_t structure.
 * returns the length of resulting array (may be less than the original
 *          due to possible compaction).
 */
int
iser_iovec_virt_to_mem_t(struct iser_task_t *p_task,
			 enum dma_data_direction dir,
			 struct iser_data_buf_t *p_data,
			 struct iser_mem_t *p_phys,
			 unsigned long *p_total_sz)
{
	struct iovec *p_iovec_virt;
	void *virt_addr;
	unsigned int cur_phys = 0;
	unsigned long total_sz = 0;
	uint64_t cur_addr;
	int i;

	ITRACE_ENTRY();

	if (p_data == NULL) {
		IPANIC("NULL data buffer descriptor\n");
	}
	if (p_data->type != ISER_BUF_TYPE_IOVEC_VIRT) {
		IPANIC("Unexpected buffer type: %d\n", p_data->type);
	}
	if (p_phys == NULL) {
		IPANIC("NULL physical mem vector\n");
	}

	/* Get the input iovec array */
	p_iovec_virt = (struct iovec *) p_data->p_buf;
	if (p_iovec_virt == NULL) {
		IPANIC("NULL data buffer's virt iovec\n");
	}

	ITRACE(ISER_TRACE_SCATTERLIST,
	       "Translating data: 0x%p, virt iovec: 0x%p, entries: %d\n",
	       p_data, p_iovec_virt, p_data->size);

	for (i = 0; i < p_data->size; i++) {
		virt_addr = p_iovec_virt[i].iov_base;
		cur_addr = (unsigned long) virt_addr;

		if (cur_phys > 0
		    && (p_phys[cur_phys - 1].addr +
			p_phys[cur_phys - 1].size) == cur_addr) {
			p_phys[cur_phys - 1].size +=
			    p_iovec_virt[i].iov_len;

			ITRACE(ISER_TRACE_SCATTERLIST,
			       "IOVEC VIRT[%d] virt: 0x%08lX, sz: 0x%05lX -> 0x%08lX, MERGED   phys[%ld]\n",
			       i, (unsigned long) p_iovec_virt[i].iov_base,
			       (unsigned long) p_iovec_virt[i].iov_len,
			       (unsigned long) cur_addr,
			       (long) (cur_phys - 1));
		} else {
			p_phys[cur_phys].addr = cur_addr;
			p_phys[cur_phys].size = p_iovec_virt[i].iov_len;

			ITRACE(ISER_TRACE_SCATTERLIST,
			       "IOVEC VIRT[%d] virt: 0x%08lX, sz: 0x%05lX -> 0x%08lX, ADDED as phys[%ld]\n",
			       i, (unsigned long) p_iovec_virt[i].iov_base,
			       (unsigned long) p_iovec_virt[i].iov_len,
			       (unsigned long) cur_addr, (long) cur_phys);

			cur_phys++;
		}
		total_sz += p_iovec_virt[i].iov_len;
	}

	if (p_total_sz != NULL)
		*p_total_sz = total_sz;

	ITRACE_EXIT();
	return cur_phys;
}				/* iser_iovec_virt_to_mem_t */

/**
 * iser_scatterlist_virt_to_mem_t - Translates scatterlist entries to an iser_mem_t.
 * returns the length of resulting address array (may be less than the original
 *         due to possible compaction).
 */
int
iser_scatterlist_to_mem_t(struct iser_task_t *p_task,
			  enum dma_data_direction cmd_dir,
			  struct iser_data_buf_t *p_data,
			  struct iser_mem_t *p_phys,
			  unsigned long *p_total_sz)
{
	struct scatterlist *p_sg;
	unsigned int cur_phys = 0;
	unsigned long total_sz = 0;
	uint64_t cur_addr;
	int i;

	ITRACE_ENTRY();

	if (p_data == NULL) {
		IPANIC("NULL data buffer descriptor\n");
	}
	if (p_data->type != ISER_BUF_TYPE_SCATTERLIST) {
		IPANIC("Unexpected buffer type: %d\n", p_data->type);
	}
	if (p_phys == NULL) {
		IPANIC("NULL output phys iovec\n");
	}
	if (!p_data->p_buf)
		IPANIC("%s: NULL data buffer's scatterlist array\n", __func__);

	ITRACE(ISER_TRACE_SCATTERLIST,
	       "Translating data: 0x%p, scatterlist: 0x%p, entries: %d\n",
	       p_data, p_data->p_buf, p_data->size);

	for_each_sg (p_data->p_buf, p_sg, p_data->size, i) {

		TRACE(TRACE_ISER, "%s: entry=%d\n", __func__, i);

#if defined(__KERNEL__)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,20)
		cur_addr = (unsigned long) sg_virt(p_sg);
#else /* linux > 2.6.20 */
		cur_addr = (unsigned long) page_address(p_sg->page) + p_sg->offset;
#endif /* linux > 2.6.20 */
#else  /* __KERNEL__ */
		cur_addr = (unsigned long) sg_virt(p_sg)
                           + scatterlist_offset(p_sg);
#endif /* __KERNEL__ */

		if (cur_phys > 0
		    && (p_phys[cur_phys - 1].addr +
			p_phys[cur_phys - 1].size) == cur_addr) {
			p_phys[cur_phys - 1].size += scatterlist_length(p_sg);

			TRACE(TRACE_ISER_MEM,
			      "%s: SCATTER[%d] addr: 0x%08lX + 0x%05X, sz: 0x%05X) -> 0x%08lX, MERGED   phys[%d]\n",
			      __func__,
			      i, (unsigned long) sg_virt(p_sg),
			      scatterlist_offset(p_sg), scatterlist_length(p_sg),
			      (unsigned long) cur_addr, cur_phys - 1);
		} else {
			p_phys[cur_phys].addr = cur_addr;
			p_phys[cur_phys].size = scatterlist_length(p_sg);

			TRACE(TRACE_ISER_MEM,
			      "%s: SCATTER[%d] addr: 0x%08lX + 0x%05X, sz: 0x%05X -> 0x%08lX, ADDED as phys[%d]\n",
			      __func__,
			      i, (unsigned long) sg_virt(p_sg),
			      scatterlist_offset(p_sg), scatterlist_length(p_sg),
			      (unsigned long) cur_addr, cur_phys);

			cur_phys++;
		}
		total_sz += scatterlist_length(p_sg);
	}

	if (p_total_sz != NULL)
		*p_total_sz = total_sz;

	TRACE(TRACE_ISER_MEM, "%s: total_sz=%lu\n", __func__, total_sz);

	ITRACE_EXIT();
	return cur_phys;
}				/* iser_scatterlist_to_mem_t */


#define MASK_4K   ((1UL << iser_buffer_aligment_shift) - 1)	// 0xFFF
//#define IS_PAGE_ALIGNED(addr) ((((unsigned long)addr) & ~PAGE_MASK) == 0)
#define IS_PAGE_ALIGNED(addr) ((((unsigned long)addr) & MASK_4K) == 0)

/**
 * iser_iovec_aligned_length - Tries to determine the maximal correctly aligned sub-list of a
 * scatter-gather list of memory buffers
 * returns the number of entries which are aligned correctly
 */
unsigned int
iser_iovec_aligned_length(struct iser_mem_t *p_phys,
			  unsigned int phys_vec_len,
			  unsigned long *total_data_sz)
{
	unsigned int ret_len = phys_vec_len;
	unsigned long ret_total_sz = 0;
	unsigned long end_addr;
	int i;

	ITRACE_ENTRY();

	if (p_phys == NULL) {
		IPANIC("NULL physical array for alignment check\n");
	}
	if (phys_vec_len == 0) {
		IPANIC("zero physical array length for alignment check\n");
	}

	for (i = 0; i < phys_vec_len; i++) {
		ret_total_sz += p_phys[i].size;
		ITRACE(ISER_TRACE_SCATTERLIST,
		       "Checking buf [%d]: 0x%08lX sz: %ld\n", i,
		       (unsigned long) p_phys[i].addr,
		       (unsigned long) p_phys[i].size);
		if ((i > 0) && !IS_PAGE_ALIGNED(p_phys[i].addr)) {
			ret_len = i + 1;
			break;
		}
		if (i + 1 < phys_vec_len) {
			end_addr =
			    ((unsigned long) p_phys[i].addr) +
			    p_phys[i].size;
			if (!IS_PAGE_ALIGNED(end_addr)) {
				ret_len = i + 1;
				break;
			}
		}
	}
	if (total_data_sz) {
		*total_data_sz = ret_total_sz;
	}

	ITRACE(ISER_TRACE_SCATTERLIST,
	       "Found %d aligned entries out of %d in phys.array:0x%p, total sz: %ld\n",
	       ret_len, phys_vec_len, p_phys, ret_total_sz);

	ITRACE_EXIT();
	return ret_len;
}				/* iser_iovec_aligned_length */

/**
 * iser_iovec_virt_entry_addr - Retrieves address of a single IOVEC_VIRT entry
 * returns IOVEC_VIRT entry address
 */
inline void *iser_iovec_virt_entry_addr(void *p_iovec_virt, int i)
{
	struct iovec *p_iovec;

	BUG_ON(!p_iovec_virt);

	p_iovec = (struct iovec *) p_iovec_virt;
	return p_iovec[i].iov_base;
}

/**
 * iser_scatterlist_entry_to_virt - Translates address of a single SCATTERLIST entry to virtual kernel address
 * returns SCATTERLIST entry virtual address
 */
inline void *iser_scatterlist_entry_to_virt(void *p_scatterlist, int ind)
{
	int i;
	struct scatterlist *p_sg;

	p_sg = (struct scatterlist *) p_scatterlist;
	for (i = 0; i < ind; i += 1)
		p_sg = sg_next(p_sg);

	return (void*) ((unsigned long) sg_virt(p_sg) + scatterlist_offset(p_sg));
}

/**
 * iser_iovec_entry_len - Retrieves length of a single IOVEC_PHYS/VIRT entry
 * returns IOVEC_PHYS/VIRT entry lengths
 */
inline unsigned long iser_iovec_entry_len(void *p_iovec_arr, int i)
{
	struct iovec *p_iovec;
	p_iovec = (struct iovec *) p_iovec_arr;
	return (unsigned long) p_iovec[i].iov_len;
}

/**
 * iser_scatterlist_entry_len - Retrieves length of a single SCATTERLIST entry
 * returns SCATTERLIST entry lengths
 */
inline unsigned long iser_scatterlist_entry_len(void *p_scatterlist, int ind)
{
	int i;
	struct scatterlist *p_sg;

	p_sg = (struct scatterlist *) p_scatterlist;
	for (i = 0; i < ind; i += 1)
		p_sg = sg_next(p_sg);

	return (unsigned long) scatterlist_length(p_sg);
}

/**
 * iser_data_buf_memcpy - Copies arbitrary data buffer to a contiguous memory region
 */
void
iser_data_buf_memcpy(struct iser_data_buf_t *p_dst_buf,
		     struct iser_data_buf_t *p_src_data,
		     unsigned long *p_total_copied_sz)
{
	unsigned char *p_dst;
	unsigned int dst_sz;

	ITRACE_ENTRY();

	if (p_dst_buf == NULL) {
		IPANIC("Trying to copy to NULL dst buffer\n");
	}
	if (p_dst_buf->type != ISER_BUF_TYPE_SINGLE) {
		IPANIC
		    ("Trying to copy to destination which is not a SINGLE buffer: 0x%p\n",
		     p_dst_buf);
	}
	p_dst = p_dst_buf->p_buf;
	dst_sz = p_dst_buf->size;

	if (p_src_data->type == ISER_BUF_TYPE_SINGLE) {
		ITRACE(ISER_TRACE_SCATTERLIST,
		       "copy SINGLE virt: 0x%p -> 0x%p, sz: %d\n",
		       p_src_data->p_buf, p_dst, p_src_data->size);
		if (dst_sz < p_src_data->size) {
			IPANIC
			    ("Destination buffer too small for copy: %d\n",
			     dst_sz);
		}
		memcpy(p_dst, p_src_data->p_buf, p_src_data->size);
		if (p_total_copied_sz != NULL) {
			*p_total_copied_sz = p_src_data->size;
		}
	} else {
		unsigned char *chunk_addr = 0;
		unsigned int chunk_size = 0;
		int i;

		if (p_total_copied_sz != NULL) {
			*p_total_copied_sz = 0;
		}

		for (i = 0; i < p_src_data->size; i++) {
			switch (p_src_data->type) {
			case ISER_BUF_TYPE_IOVEC_VIRT:
				chunk_addr = (unsigned char *)
				    iser_iovec_virt_entry_addr(p_src_data->
							       p_buf, i);
				chunk_size =
				    iser_iovec_entry_len(p_src_data->p_buf,
							 i);
				break;

			case ISER_BUF_TYPE_SCATTERLIST:
				chunk_addr = (unsigned char *)
				    iser_scatterlist_entry_to_virt
				    (p_src_data->p_buf, i);
				chunk_size =
				    iser_scatterlist_entry_len(p_src_data->
							       p_buf, i);
				break;

			default:
				IPANIC("Unexpected buffer type\n");
				break;
			}

			ITRACE(ISER_TRACE_SCATTERLIST,
			       "copy IOVEC[%d] virt: 0x%p -> 0x%p, sz: %d\n",
			       i, chunk_addr, p_dst, chunk_size);
			if (dst_sz < chunk_size) {
				IPANIC
				    ("Destination buffer too small for copy: %d\n",
				     dst_sz);
			}
			memcpy(p_dst, chunk_addr, chunk_size);
			p_dst += chunk_size;
			dst_sz -= chunk_size;
			if (p_total_copied_sz != NULL) {
				*p_total_copied_sz += chunk_size;
			}
		}
	}

	ITRACE_EXIT();
}				/* iser_data_buf_memcpy */

/**
 * iser_data_buf_get_type_name - Retrieves name of a iSER data buffer type
 * returns name string
 */
static char *iser_data_buf_type_name[ISER_BUF_TYPES_NUM + 1] = {
	"SINGLE",		// ISER_BUF_TYPE_SINGLE
	"IOVEC_VIRT",		// ISER_BUF_TYPE_IOVEC_VIRT
	"SCATTERLIST",		// ISER_BUF_TYPE_SCATTERLIST

	"ILLEGAL"		// ISER_BUF_TYPES_NUM
};

char *iser_data_buf_get_type_name(struct iser_data_buf_t *p_data)
{
	if (p_data != NULL && p_data->type < ISER_BUF_TYPES_NUM) {
		return iser_data_buf_type_name[p_data->type];
	} else {
		return iser_data_buf_type_name[ISER_BUF_TYPES_NUM];
	}
}				/* iser_data_buf_get_type_name */
