
/*
 * Copyright (c) 2004, 2005, 2006 Voltaire, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if defined(__KERNEL__)
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <dat/kdat.h>
#include <asm/io.h>
#else
#include <dat/udat.h>
#endif				/* __KERNEL__ */

#include "iser.h"
#include "iser_bhs.h"
#include "iser_global.h"
//#include "iser_initiator.h"
#include "iser_target.h"
#include "iser_conn.h"
#include "iser_task.h"
#include "iser_dto.h"
#include "iser_kdapl.h"
#include "iser_utils.h"
#include "iser_memory.h"
#include "iser_procfs.h"
#include "iser_trace.h"
#include "iser_platform.h"

/* ---------------------------------------------------------------------
 * CONSTANTS & MACROS
 * ------------------------------------------------------------------ */
#define ISER_MAX_ASYNC_QLEN	8
// #define ISER_MAX_QLEN       ((ISER_MAX_INITIATOR_ENTITIES * ISER_INITIATOR_MAX_QLEN) + 
//                  (ISER_MAX_TARGET_ENTITIES * ISER_TARGET_MAX_QLEN))
#define ISER_MAX_QLEN		(ISER_MAX_TARGET_ENTITIES * ISER_TARGET_MAX_QLEN)

/* Service ID generation */
#define CONN_QUAL_ISER_MASK	0xFEFEFEFE
#define CONN_QUAL_PORT_MASK	0x0000FFFF

/* the prefix used by the openib gen2 CMA to translate from 16bit */
/* port to 64bit IB SID                                           */
#define IB_OPENIB_OUI  (0x001405)
#define RDMA_PS_TCP    (0x0106)

/* Logging helper strings */
static char *dto_status_name[DAT_RMR_OPERATION_FAILED + 2] = {
	/* DAT_DTO_COMPLETION_STATUS does not have max value */
	"SUCCESS",
	"ERR_FLUSHED",
	"ERR_LOCAL_LENGTH",
	"ERR_LOCAL_EP",
	"ERR_LOCAL_PROTECTION",
	"ERR_BAD_RESPONSE",
	"ERR_REMOTE_ACCESS",
	"ERR_REMOTE_RESPONDER",
	"ERR_TRANSPORT",
	"ERR_RECEIVER_NOT_READY",
	"ERR_PARTIAL_PACKET",
	"RMR_OPERATION_FAILED",

	"ILLEGAL"
};

static char *dto_type_name[ISER_DTO_TYPES_NUM + 1] = {
	"RECV",
	"SEND",
	"RDMA_W",
	"RDMA_R",
	"PASSIVE",

	"ILLEGAL"
};

/* ---------------------------------------------------------------------
 * STATIC FUNCTIONS DECLARATION
 * ------------------------------------------------------------------ */
static void
iser_evd_upcall(DAT_PVOID inst,
		const DAT_EVENT * ev, DAT_BOOLEAN has_more);

static int
iser_handle_single_kdapl_event(struct iser_adaptor_t *p_adaptor);

/* ---------------------------------------------------------------------
 * FUNCTIONS
 * ------------------------------------------------------------------ */

/**
 * print_dat_error - Prints out DAT error message
 */
static void print_dat_error(const char *msg,	//<! IN - message text
			    DAT_RETURN err)	//<! IN - error code
{
	const char *major_msg;
	const char *minor_msg;

	dat_strerror(err, &major_msg, &minor_msg);
	IERROR("DAT - %s (%s %s)\n", msg, major_msg, minor_msg);
}				/* print_dat_error */

/**
 * iser_create_ia_pz_evd - Creates global Interface Adapter (IA), Protection Zone (PZ), Event Dispatcher (EVD).
 *
 * returns 0 on success, -1 on failure
 */
int iser_create_ia_pz_evd(struct iser_adaptor_t *p_adaptor)	//<! IN - iSER adaptor descriptor
{
	DAT_UPCALL_OBJECT upcall;
	DAT_RETURN dat_ret;

	DAT_EVD_HANDLE async_evd_hdl;
	//DAT_IA_ATTR         dat_ia_attr;
	DAT_PROVIDER_ATTR dat_provider_attr;

	// Open the Interface Adaptor
	dat_ret = dat_ia_open((DAT_NAME_PTR) p_adaptor->name, ISER_MAX_ASYNC_QLEN, &p_adaptor->async_evd_hdl, &p_adaptor->ia_handle);	/* IA returned here */
	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ia_open failed!", dat_ret);
		return -1;
	}

	/* Get the adaptor attributes */
	dat_ret = dat_ia_query(p_adaptor->ia_handle,	// IN          DAT_IA_HANDLE
			       &async_evd_hdl,	// OUT     DAT_EVD_HANDLE *
			       DAT_IA_ALL,	// IN          DAT_IA_ATTR_MASK
			       &p_adaptor->dat_ia_attr,	// OUT     DAT_IA_ATTR *
			       DAT_PROVIDER_FIELD_ALL,	// IN      DAT_PROVIDER_ATTR_MASK
			       &dat_provider_attr);	// OUT         DAT_PROVIDER_ATTR *

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ia_query failed!", dat_ret);
		return -1;
	}


	/* Create a protection zone */
	p_adaptor->pz_handle = DAT_HANDLE_NULL;
	dat_ret =
	    dat_pz_create(p_adaptor->ia_handle, &p_adaptor->pz_handle);
	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_pz_create failed!", dat_ret);
		return -1;
	}

	/* Create event dispatcher */

	/* Instance allows working with multiple IAs */
	upcall.instance_data = (DAT_PVOID) p_adaptor;
	upcall.upcall_func = iser_evd_upcall;

	p_adaptor->gnrl_evd_hdl = DAT_HANDLE_NULL;
	ITRACE(ISER_TRACE_KDAPL, "Creating IA with QLEN: %d\n",
	       ISER_MAX_QLEN);
	dat_ret = dat_evd_kcreate(p_adaptor->ia_handle,	/* IA handle */
				  ISER_MAX_QLEN,	/* Queue length */
				  DAT_UPCALL_SINGLE_INSTANCE,	/* call at most 1 time */
				  &upcall,	/* initialized above */
				  DAT_EVD_CR_FLAG | DAT_EVD_CONNECTION_FLAG | DAT_EVD_DTO_FLAG, &p_adaptor->gnrl_evd_hdl);	/* EVD handle returned here */

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_evd_create failed!\n", dat_ret);
		p_adaptor->gnrl_evd_hdl = DAT_HANDLE_NULL;
		return -1;
	}

	return 0;
}				/* iser_create_ia_pz_evd */

/**
 * iser_free_ia_pz_evd - Frees global Interface Adapter (IA), Protection Zone (PZ), Event Dispatcher (EVD).
 *
 * returns 0 on success, -1 on failure
 */
int iser_free_ia_pz_evd(struct iser_adaptor_t *p_adaptor)	//<! IN - iSER adaptor descriptor
{
	DAT_RETURN dat_ret;

	/* Release EVD */
	if (p_adaptor->gnrl_evd_hdl != DAT_HANDLE_NULL) {
		ITRACE(ISER_TRACE_KDAPL, "Releasing evd: 0x%p\n",
		       p_adaptor->gnrl_evd_hdl);
		dat_ret = dat_evd_free(p_adaptor->gnrl_evd_hdl);
		if (dat_ret != DAT_SUCCESS) {
			print_dat_error("free_evds failed!", dat_ret);
			return -1;
		}
	}

	/* Release PZ */
	if (p_adaptor->pz_handle != DAT_HANDLE_NULL) {
		ITRACE(ISER_TRACE_KDAPL, "Releasing pz: 0x%p\n",
		       p_adaptor->pz_handle);
		dat_ret = dat_pz_free(p_adaptor->pz_handle);
		if (dat_ret != DAT_SUCCESS) {
			print_dat_error("dat_pz_free failed!", dat_ret);
			return -1;
		}
	}

	/* Release IA */
	if (p_adaptor->ia_handle != DAT_HANDLE_NULL) {
		ITRACE(ISER_TRACE_KDAPL, "Releasing ia: 0x%p\n",
		       p_adaptor->ia_handle);
		dat_ret =
		    dat_ia_close(p_adaptor->ia_handle,
				 DAT_CLOSE_GRACEFUL_FLAG);
		if (dat_ret != DAT_SUCCESS) {
			print_dat_error("dat_ia_close failed!", dat_ret);
			return -1;
		}
	}

	return 0;
}				/* iser_free_ia_pz_evd */

/**
 * iser_create_ep - Creates an End-Point (EP)
 *
 * returns 0 on success, -1 on failure
 */
int iser_create_ep(struct iser_conn_t *p_iser_conn,	//<! IN - iSER connection descriptor
		   unsigned long max_rdma_sz,	//<! IN - max size of RDMA-ed data
		   unsigned int max_rdma_read_in,	//<! IN - max number of incoming RDMA-read reqs
		   unsigned int max_rdma_read_out,	//<! IN - max number of outgoing RDMA-read reqs
		   unsigned int max_recv_dtos,	//<! IN - max number of receive-type DTOs
		   unsigned int max_request_dtos,	//<! IN - max number of request-type DTOs
		   unsigned int max_request_iov)	//<! IN - max length of a request data IOV
{
	struct iser_adaptor_t *p_adaptor;
	DAT_EP_ATTR ep_attr;
	DAT_RETURN dat_ret;

	ITRACE_ENTRY();

	if (p_iser_conn->p_adaptor == NULL) {
		IERROR("NULL adaptor in connection, p_conn: 0x%p\n",
		       p_iser_conn);
		ITRACE_EXIT();
		return -1;
	}
	p_adaptor = p_iser_conn->p_adaptor;

	memset(&ep_attr, 0, sizeof(DAT_EP_ATTR));

	ep_attr.service_type = DAT_SERVICE_TYPE_RC;	// type: DAT_SERVICE_TYPE

	ep_attr.max_mtu_size = 2 * 1024;	// type: DAT_VLEN
	ep_attr.max_rdma_size = max_rdma_sz;	// type: DAT_VLEN

	ep_attr.qos = DAT_QOS_BEST_EFFORT;	// type: DAT_QOS

	ep_attr.recv_completion_flags = DAT_COMPLETION_DEFAULT_FLAG;	// type: DAT_COMPLETION_FLAGS
	ep_attr.request_completion_flags = DAT_COMPLETION_DEFAULT_FLAG;	// type: DAT_COMPLETION_FLAGS

	ep_attr.max_recv_dtos = max_recv_dtos;	// type: DAT_COUNT
	ep_attr.max_request_dtos = max_request_dtos;	// type: DAT_COUNT

	ep_attr.max_recv_iov = 2;	// type: DAT_COUNT; header and data
	ep_attr.max_request_iov = max_request_iov;	// type: DAT_COUNT

	ep_attr.max_rdma_read_in = MIN(max_rdma_read_in, p_adaptor->dat_ia_attr.max_rdma_read_per_ep_in);	// DAT_COUNT
	ep_attr.max_rdma_read_out = MIN(max_rdma_read_out, p_adaptor->dat_ia_attr.max_rdma_read_per_ep_out);	// DAT_COUNT

	ITRACE(ISER_TRACE_KDAPL, "Creating EP with\n"
	       "\tmax_mtu_sz: %ld, max_rdma_sz: %ld\n"
	       "\tmax_recv_dtos: %d, max_req_dtos: %d\n"
	       "\tmax_recv_iov: %d, max_req_iov: %d\n"
	       "\tmax_rdma_in: %d, max_rdma_out: %d\n",
	       (unsigned long) ep_attr.max_mtu_size,
	       (unsigned long) ep_attr.max_rdma_size,
	       ep_attr.max_recv_dtos, ep_attr.max_request_dtos,
	       ep_attr.max_recv_iov, ep_attr.max_request_iov,
	       ep_attr.max_rdma_read_in, ep_attr.max_rdma_read_out);

	/* Create end-point */
	p_iser_conn->ep_handle = DAT_HANDLE_NULL;

	dat_ret = dat_ep_create(p_adaptor->ia_handle,
				p_adaptor->pz_handle,
				p_adaptor->gnrl_evd_hdl,
				p_adaptor->gnrl_evd_hdl,
				p_adaptor->gnrl_evd_hdl,
				&ep_attr, &p_iser_conn->ep_handle);

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ep_create failed!", dat_ret);
		IERROR("%s failed for conn: with parameters0x%p, "
		       "max_rdma_sz: %ld, max_rdma_read_in: %d, max_rdma_read_out: %d, "
		       "max_recv_dtos: %d, max_request_dtos: %d, max_request_iov: %d\n",
		       __func__, p_iser_conn,
		       max_rdma_sz, max_rdma_read_in, max_rdma_read_out,
		       max_recv_dtos, max_request_dtos, max_request_iov);

		IERROR("Creating EP with\nttributes"
		       "\tmax_mtu_sz: %ld, max_rdma_sz: %ld\n"
		       "\tmax_recv_dtos: %d, max_req_dtos: %d\n"
		       "\tmax_recv_iov: %d, max_req_iov: %d\n"
		       "\tmax_rdma_in: %d, max_rdma_out: %d\n",
		       (unsigned long) ep_attr.max_mtu_size,
		       (unsigned long) ep_attr.max_rdma_size,
		       ep_attr.max_recv_dtos, ep_attr.max_request_dtos,
		       ep_attr.max_recv_iov, ep_attr.max_request_iov,
		       ep_attr.max_rdma_read_in,
		       ep_attr.max_rdma_read_out);

		return -1;
	}

	ITRACE_EXIT();
	return 0;
}				/* iser_create_ep */

/**
 * iser_free_ep - Releases the end point object.
 *
 * Also starts the connection termination. May be called more than
 * once, actually initiates releasing EP only for the first call.
 *
 * returns 0 on success, -1 on failure
 */
int iser_free_ep(struct iser_conn_t *p_iser_conn)	//<! IN - iSER connection descriptor
{
	int ret_val = 0;

	ITRACE_ENTRY();

	if (p_iser_conn != NULL) {
		/* Release EP */
		if (p_iser_conn->ep_handle != DAT_HANDLE_NULL) {
			DAT_RETURN dat_ret =
			    dat_ep_free(p_iser_conn->ep_handle);
			if (dat_ret == DAT_SUCCESS) {
				p_iser_conn->ep_handle = DAT_HANDLE_NULL;
			} else {
				print_dat_error("dat_ep_free failed!",
						dat_ret);
				ret_val = -1;
			}
		}
	} else {
		IERROR("NULL connection descriptor to free\n");
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_free_ep */

/**
 * iser_connect - Connects an End-Point
 *
 * returns 0 on success, -1 on failure
 */
int iser_connect(struct iser_conn_t *p_iser_conn,	//<! IN - iSER connection descriptor
		 struct sockaddr_in *dst_addr,	//<! IN - destination (remote) IP addr
		 struct sockaddr_in *src_addr)	//<! IN - source (local) IP addr
{
	DAT_CONN_QUAL conn_qualifier;
	DAT_RETURN dat_ret;
	int ret_val = 0;
	// unsigned char               hello_private_data[ISER_HELLO_LEN];

	ITRACE_ENTRY();

	if (!dst_addr) {
		IPANIC("NULL destination IP address\n");
	}
	// set the listening port as the 16 LS-Bits of connection qualifier
	conn_qualifier =
	    (DAT_CONN_QUAL) (dst_addr->sin_port & CONN_QUAL_PORT_MASK);
	// set the iSER service qualifier as the 32 MS-Bits
	conn_qualifier |= ((DAT_CONN_QUAL) CONN_QUAL_ISER_MASK) << 32;

	ITRACE(ISER_TRACE_KDAPL, "port: %d, conn. qualifier: 0x%llX\n",
	       dst_addr->sin_port, (long long) conn_qualifier);

	dat_ret = dat_ep_connect(p_iser_conn->ep_handle, (DAT_IA_ADDRESS_PTR) dst_addr,	/* destination address */
				 conn_qualifier,	/* Connection qualifier */
				 DAT_TIMEOUT_INFINITE,	/* timeout */
				 0,	// ISER_HELLO_LEN, /* private data size */
				 (DAT_PVOID) 0,	// hello_private_data /* private data */
				 DAT_QOS_BEST_EFFORT,	/* QoS */
				 DAT_CONNECT_DEFAULT_FLAG);	/* flags */

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ep_connect failed!", dat_ret);
		dat_ep_free(p_iser_conn->ep_handle);
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_connect */

/**
 * iser_disconnect - Disconnects an End-Point
 *
 * returns 0 on success, -1 on failure
 */
int iser_disconnect(struct iser_conn_t *p_iser_conn)	//<! IN - iSER connection descriptor
{
	int ret_val = 0;
	DAT_RETURN dat_ret;

	ITRACE_ENTRY();

	dat_ret =
	    dat_ep_disconnect(p_iser_conn->ep_handle,
			      DAT_CLOSE_ABRUPT_FLAG);
	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ep_disconnect failed!", dat_ret);
		// if disconnect failed, try freeing EP as a brute-force method
		iser_free_ep(p_iser_conn);
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_disconnect */


/**
 * iser_create_psp_by_type - Creates a Public Service Point (PSP) for a specific conn qualifier.
 *
 * returns 0 on success, -1 on failure
 */
int
iser_create_psp_by_type(struct iser_entity_t *p_entity,
			struct iser_adaptor_t *p_adaptor,
			DAT_PSP_HANDLE * p_psp_handle,
			DAT_CONN_QUAL conn_qualifier)
{
	DAT_RETURN dat_ret;

	*p_psp_handle = DAT_HANDLE_NULL;	// Init the handle before creating it
	dat_ret = dat_psp_create(p_adaptor->ia_handle,	/* adaptor */
				 conn_qualifier, p_adaptor->gnrl_evd_hdl,	/* EVD */
				 DAT_PSP_CONSUMER_FLAG,	/* EP is allocated by iSER */
				 p_psp_handle);	/* PSP handle returned here */

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_psp_create failed!", dat_ret);
		p_entity->port = 0;
		*p_psp_handle = DAT_HANDLE_NULL;	// Init the handle again because it wasn't created
		if (iser_free_psp_by_entity(p_entity) != 0) {
			// ToDo: If this call fails, some PSPs could not be deallocated. We should panic.
		}
		return -1;
	}
	return 0;
}

/**
 * iser_create_psp - Creates a Public Service Point (PSP).
 *
 * returns 0 on success, -1 on failure
 */
int iser_create_psp(struct iser_entity_t *p_entity,	//<! IN - entity descriptor
		    unsigned int port)	//<! IN - number of the listening port
{
	DAT_CONN_QUAL conn_qualifier;
	int i;

	if (port == p_entity->port) {
		IERROR("PSP for port %d already exists\n", port);
		return 0;
	}
	// Free previously created PSP
	if (iser_free_psp_by_entity(p_entity) != 0)
		return -1;

	// Create PSP
	for (i = 0; i < iser_global.num_adaptors; i++) {
		ITRACE(ISER_TRACE_KDAPL, "Creating ibhost psp\n");
		conn_qualifier =
		    (DAT_CONN_QUAL) (port & CONN_QUAL_PORT_MASK);
		// set the iSER service qualifier as the 32 MS-Bits
		conn_qualifier |=
		    ((DAT_CONN_QUAL) CONN_QUAL_ISER_MASK) << 32;

		if (iser_create_psp_by_type
		    (p_entity, &iser_global.adaptor[i],
		     &p_entity->ibhost_psp_handle[i], conn_qualifier)) {
			IERROR
			    ("iser_create_psp_by_type (ibhost conn qualifier) failed\n");
			return -1;
		}

		ITRACE(ISER_TRACE_KDAPL, "Creating openib psp\n");
		conn_qualifier = ((u64) RDMA_PS_TCP << 16) + port;
		if (iser_create_psp_by_type
		    (p_entity, &iser_global.adaptor[i],
		     &p_entity->openib_psp_handle[i], conn_qualifier)) {
			IERROR
			    ("iser_create_psp_by_type (openib conn qualifier) failed\n");
			return -1;
		}
	}
	p_entity->port = port;

	return 0;
}				/* iser_create_psp */

/**
 * iser_free_psp_by_entity - Frees all Public Service Points (PSPs) for an iSCSI entity.
 *
 * returns 0 on success, -1 on failure
 */
int iser_free_psp_by_entity(struct iser_entity_t *p_entity)	//<! IN - iSCSI layer entity
{
	DAT_RETURN dat_ret;
	int i;

	/* Release PSPs of this entity on all adaptors */
	for (i = 0; i < iser_global.num_adaptors; i++) {
		if (p_entity->ibhost_psp_handle[i] != DAT_HANDLE_NULL) {
			ITRACE(ISER_TRACE_KDAPL,
			       "Releasing ibhost psp for entity %d: %s - %s\n",
			       i, p_entity->provider_name,
			       iser_get_entity_type_name(p_entity->type));
			dat_ret =
			    dat_psp_free(p_entity->ibhost_psp_handle[i]);
			if (dat_ret != DAT_SUCCESS) {
				print_dat_error
				    ("dat_psp_free failed (ibhost conn qualifier)!",
				     dat_ret);
				return -1;
			}
			p_entity->ibhost_psp_handle[i] = DAT_HANDLE_NULL;
		}

		if (p_entity->openib_psp_handle[i] != DAT_HANDLE_NULL) {
			ITRACE(ISER_TRACE_KDAPL,
			       "Releasing openib psp for entity %d: %s - %s\n",
			       i, p_entity->provider_name,
			       iser_get_entity_type_name(p_entity->type));
			dat_ret =
			    dat_psp_free(p_entity->openib_psp_handle[i]);
			if (dat_ret != DAT_SUCCESS) {
				print_dat_error
				    ("dat_psp_free failed (openib conn qualifier)!",
				     dat_ret);
				return -1;
			}
			p_entity->openib_psp_handle[i] = DAT_HANDLE_NULL;
		}
	}

	return 0;
}				/* iser_free_psp_by_entity */

/**
 * iser_free_psp_by_adaptor - Frees a Public Service Points (PSPs) for an adaptor object.
 *
 * returns 0 on success, -1 on failure
 */
int iser_free_psp_by_adaptor(struct iser_adaptor_t *p_adaptor)	//<! IN - adaptor descriptor
{
	struct iser_entity_t *p_entity;
	int i;
	DAT_RETURN dat_ret;

	i = p_adaptor - iser_global.adaptor;	/* pointer arithmetic returns adaptor's index */

	// Release PSPs of this adaptor on all entities
	for (i = 0; i < iser_global.num_entities; i++) {
		p_entity = &iser_global.entity[i];
		if (p_entity->ibhost_psp_handle[i] != DAT_HANDLE_NULL) {
			ITRACE(ISER_TRACE_KDAPL,
			       "Releasing ibhost psp for entity %d: %s - %s\n",
			       i, p_entity->provider_name,
			       iser_get_entity_type_name(p_entity->type));
			dat_ret =
			    dat_psp_free(p_entity->ibhost_psp_handle[i]);
			if (dat_ret != DAT_SUCCESS) {
				print_dat_error
				    ("dat_psp_free failed (ibhost)!",
				     dat_ret);
				return -1;
			}
			p_entity->ibhost_psp_handle[i] = DAT_HANDLE_NULL;
		}

		if (p_entity->openib_psp_handle[i] != DAT_HANDLE_NULL) {
			ITRACE(ISER_TRACE_KDAPL,
			       "Releasing openib psp for entity %d: %s - %s\n",
			       i, p_entity->provider_name,
			       iser_get_entity_type_name(p_entity->type));
			dat_ret =
			    dat_psp_free(p_entity->openib_psp_handle[i]);
			if (dat_ret != DAT_SUCCESS) {
				print_dat_error
				    ("dat_psp_free failed (openib)!",
				     dat_ret);
				return -1;
			}
			p_entity->openib_psp_handle[i] = DAT_HANDLE_NULL;
		}
	}

	return 0;
}				/* iser_free_psp_by_adaptor */

/**
 * iser_cr_accept - Accepts an incoming connection request.
 *
 * returns 0 on success, -1 on failure
 */
int iser_cr_accept(struct iser_conn_t *p_iser_conn,	//<! iSER connection descriptor
		   struct iser_data_buf_t *p_priv_data)	//<! Private data buffer
{
	DAT_RETURN dat_ret;
	unsigned long priv_data_size = 0;
	unsigned char *p_priv_data_buf = NULL;

	if (p_priv_data != NULL) {
		if (p_priv_data->type != ISER_BUF_TYPE_SINGLE) {
			IPANIC
			    ("Private data passed not in a single buffer, unsupported\n");
		}
		priv_data_size = p_priv_data->size;
		p_priv_data_buf = p_priv_data->p_buf;
	}

	dat_ret = dat_cr_accept(p_iser_conn->cr_handle, p_iser_conn->ep_handle, priv_data_size,	/* private data size */
				p_priv_data_buf);	/* private data */
	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_cr_accept failed", dat_ret);
		return -1;
	}

	return 0;
}				/* iser_cr_accept */

/**
 * iser_cr_reject - Rejects an incoming connection request.
 *
 * returns 0 on success, -1 on failure
 */
int iser_cr_reject(struct iser_conn_t *p_iser_conn)	//<! iSER connection descriptor
{
	DAT_RETURN dat_ret;

	dat_ret = dat_cr_reject(p_iser_conn->cr_handle);
	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_cr_reject failed", dat_ret);
		return -1;
	}

	return 0;
}				/* iser_cr_reject */

/**
 * iser_register_virt_mem - Register a memory buffer.
 *
 * returns: 0 on success, -1 on failure
 */
int iser_register_virt_mem(struct iser_adaptor_t *p_adaptor,	//<! IN - iSER adaptor descriptor
			   void *p_buf,	//<! IN - memory to register
			   unsigned long data_sz,	//<! IN - size of the data buffer
			   DAT_MEM_PRIV_FLAGS priv_flags,	//<! IN - DAT privilege flags
			   struct iser_mem_handles_t *p_mem_reg,	//<! OUT - local registration handles
			   DAT_RMR_CONTEXT * p_rmr_ctxt)	//<! OUT - remote registration handles
{
	DAT_REGION_DESCRIPTION mem_region;
	DAT_RETURN dat_ret;
	DAT_RMR_CONTEXT dummy_rmr;
	int ret_val = 0;

	ITRACE_ENTRY();

	mem_region.for_va = p_buf;
#ifdef _DAT1_2_
	dat_ret = dat_lmr_kcreate(p_adaptor->ia_handle, DAT_MEM_TYPE_VIRTUAL, mem_region,	// single contiguous memory region
				  data_sz,	// its size
				  p_adaptor->pz_handle,
				  priv_flags,
				  DAT_MEM_OPTIMIZE_DONT_CARE,
				  &p_mem_reg->lmr_handle,
				  &p_mem_reg->lmr_triplet.lmr_context,
				  p_rmr_ctxt ==
				  NULL ? &dummy_rmr : p_rmr_ctxt,
				  &p_mem_reg->lmr_triplet.segment_length,
				  &p_mem_reg->lmr_triplet.virtual_address);
#else				//for dat1.1
	dat_ret = dat_lmr_kcreate(p_adaptor->ia_handle, DAT_MEM_TYPE_VIRTUAL, mem_region,	// single contiguous memory region
				  data_sz,	// its size
				  p_adaptor->pz_handle,
				  priv_flags,
				  DAT_MEM_OPT_DONT_CARE,
				  &p_mem_reg->lmr_handle,
				  &p_mem_reg->lmr_triplet.lmr_context,
				  p_rmr_ctxt ==
				  NULL ? &dummy_rmr : p_rmr_ctxt,
				  &p_mem_reg->lmr_triplet.segment_length,
				  &p_mem_reg->lmr_triplet.virtual_address);

#endif

	ITRACE(ISER_TRACE_PHYS_MEM_REG,
	       "VIRTUAL Mem.register [SINGLE p_buf: 0x%p, sz: %ld] -> [lkey: 0x%08X lmr_h: 0x%p va: 0x%08lX sz: %ld]\n",
	       p_buf, data_sz,
	       (unsigned int) p_mem_reg->lmr_triplet.lmr_context,
	       p_mem_reg->lmr_handle,
	       (unsigned long) p_mem_reg->lmr_triplet.virtual_address,
	       (unsigned long) p_mem_reg->lmr_triplet.segment_length);

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_lmr_kcreate VIRTUAL failed", dat_ret);
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_register_virt_mem */


/**
 * iser_register_ia_mem - Register a virtual memory buffer using IA method.
 * 
 * returns: 0 on success, -1 on failure
 */
#ifdef _DAT1_2_
int iser_register_ia_mem(struct iser_adaptor_t *p_adaptor,	//<! IN - iSER adaptor descriptor
			 void *virt_addr,	//<! IN - virtual mem address to register
			 unsigned long data_sz,	//<! IN - size of the data buffer
			 DAT_MEM_PRIV_FLAGS priv_flags,	//<! IN - DAT privilege flags
			 struct iser_mem_handles_t *p_mem_reg,	//<! OUT - local registration handles
			 DAT_RMR_CONTEXT * p_rmr_ctxt)	//<! OUT - remote registration handles
{
	DAT_REGION_DESCRIPTION mem_region;
	DAT_RETURN dat_ret;
	DAT_RMR_CONTEXT dummy_rmr;
	int ret_val = 0;

	ITRACE_ENTRY();

	if (virt_addr)
		mem_region.for_pa = virt_to_phys(virt_addr);	// DAN
	else
		mem_region.for_pa = 0;

	dat_ret = dat_lmr_kcreate(p_adaptor->ia_handle, DAT_MEM_TYPE_IA,	// DAN
				  mem_region,	// single contig mem region (from get_free_pages)
				  data_sz,	// its size
				  p_adaptor->pz_handle,
				  priv_flags,
				  DAT_MEM_OPTIMIZE_DONT_CARE,
				  &p_mem_reg->lmr_handle,
				  &p_mem_reg->lmr_triplet.lmr_context,
				  p_rmr_ctxt ==
				  NULL ? &dummy_rmr : p_rmr_ctxt,
				  &p_mem_reg->lmr_triplet.segment_length,
				  &p_mem_reg->lmr_triplet.virtual_address);

	ITRACE(ISER_TRACE_PHYS_MEM_REG,
	       "IA Mem.register [SINGLE virt_addr: 0x%p, sz: %ld] -> [lkey: 0x%08X lmr_h: 0x%p va: 0x%08lX sz: %ld]\n",
	       virt_addr, data_sz,
	       (unsigned int) p_mem_reg->lmr_triplet.lmr_context,
	       p_mem_reg->lmr_handle,
	       (unsigned long) p_mem_reg->lmr_triplet.virtual_address,
	       (unsigned long) p_mem_reg->lmr_triplet.segment_length);

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_lmr_kcreate IA failed", dat_ret);
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_register_ia_mem */
#endif

/**
 * iser_register_phys_mem - Register a memory buffer.
 *
 * returns: 0 on success, -1 on failure
 */
int iser_register_phys_mem(struct iser_adaptor_t *p_adaptor,	//<! IN - iSER adaptor descriptor
			   struct iser_phys_mem_t *p_phys_vec,	//<! IN - phys memory descriptor (array)
			   unsigned int phys_vec_len,	//<! IN - memory descriptor array length
			   unsigned long total_sz,	//<! IN - total size of memory to register
			   DAT_MEM_PRIV_FLAGS priv_flags,	//<! IN - DAT privilege flags
			   struct iser_mem_handles_t *p_mem_reg,	//<! OUT - local registration handles
			   DAT_RMR_CONTEXT * p_rmr_ctxt,	//<! OUT - remote registration handles
			   enum iser_entity_type entity_type)	//<! IN - initiator or target
{
	DAT_REGION_DESCRIPTION mem_region;
	DAT_RETURN dat_ret;
	DAT_RMR_CONTEXT dummy_rmr;
	int ret_val = 0;

	ITRACE_ENTRY();

	/* Register iovec array */
	mem_region.for_va = p_phys_vec;

#ifdef _DAT1_2_
	dat_ret = dat_lmr_kcreate(p_adaptor->ia_handle, iser_platform_reg_mem_type(total_sz), mem_region,	// array of struct iovec entries
				  phys_vec_len,	// its size
				  p_adaptor->pz_handle,
				  priv_flags,
				  DAT_MEM_OPTIMIZE_DONT_CARE,
				  &p_mem_reg->lmr_handle,
				  &p_mem_reg->lmr_triplet.lmr_context,
				  p_rmr_ctxt ==
				  NULL ? &dummy_rmr : p_rmr_ctxt,
				  &p_mem_reg->lmr_triplet.segment_length,
				  &p_mem_reg->lmr_triplet.virtual_address);
#else
	dat_ret = dat_lmr_kcreate(p_adaptor->ia_handle, iser_platform_reg_mem_type(total_sz), mem_region,	// array of struct iovec entries
				  phys_vec_len,	// its size
				  p_adaptor->pz_handle,
				  priv_flags,
				  DAT_MEM_OPT_DONT_CARE,
				  &p_mem_reg->lmr_handle,
				  &p_mem_reg->lmr_triplet.lmr_context,
				  p_rmr_ctxt ==
				  NULL ? &dummy_rmr : p_rmr_ctxt,
				  &p_mem_reg->lmr_triplet.segment_length,
				  &p_mem_reg->lmr_triplet.virtual_address);
#endif

	ITRACE(ISER_TRACE_PHYS_MEM_REG,
	       "PHYSICAL Mem.register, "
	       "[PHYS total: %ld, array:(0x%p,%d), entry[0]:(0x%08lx,%ld)] -> "
	       "[lkey: 0x%08X lmr_h: 0x%p va: 0x%08lX sz: %ld]\n",
	       total_sz, p_phys_vec, phys_vec_len,
	       (unsigned long) p_phys_vec[0].addr,
	       (unsigned long) p_phys_vec[0].size,
	       (unsigned int) p_mem_reg->lmr_triplet.lmr_context,
	       p_mem_reg->lmr_handle,
	       (unsigned long) p_mem_reg->lmr_triplet.virtual_address,
	       (unsigned long) p_mem_reg->lmr_triplet.segment_length);


	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_lmr_kcreate PHYSICAL failed",
				dat_ret);
		ret_val = -1;
	}

	ITRACE_EXIT();
	return ret_val;
}				/* iser_register_phys_mem */

/**
 * iser_unregister_memory - Unregister (previosuly registered) memory.
 *
 * returns 0 on success, -1 on failure
 */
int iser_unregister_memory(struct iser_mem_handles_t *mem_reg)	//<! IN - registered memory handles
{
	int ret_val = 0;
	DAT_RETURN dat_ret;

	ITRACE_ENTRY();

	if (mem_reg->lmr_handle == DAT_HANDLE_NULL) {
		IERROR("Attempt to deregister NULL LMR Handle\n");
		ret_val = -1;
		goto unregister_memory_exit;
	}

	ITRACE(ISER_TRACE_PHYS_MEM_REG, "Mem.unregister, lmr_h: 0x%p\n",
	       mem_reg->lmr_handle);

	dat_ret = dat_lmr_free(mem_reg->lmr_handle);

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_lmr_kcreate failed", dat_ret);
		ret_val = -1;
		goto unregister_memory_exit;
	}
	mem_reg->lmr_handle = DAT_HANDLE_NULL;
	ITRACE(ISER_TRACE_PHYS_MEM_REG, "Mem.unregister succeeded\n");

      unregister_memory_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_unregister_memory */

/**
 * iser_dto_to_iov - Posts a receive buffer.
 *
 * returns 0 on success, -1 on failure
 */
static void iser_dto_to_iov(struct iser_dto_t *p_dto,	//!< IN - PostReceive DTO descriptor
			    unsigned int trace_ctxt,	//!< IN - trace context for printouts
			    DAT_LMR_TRIPLET * iov)	//!< OUT - I/O vector for DAT
{
	int i;

	/* Copy all LMR triplets from the registered buffer descriptors
	   and adjust for the offsets within the buffers and lesser sizes */
	for (i = 0; i < p_dto->regd_vector_len; i++) {
		/* Sanity check - registered buffer descriptor shuld not be NULL */
		if (p_dto->regd[i] == NULL) {
			IPANIC
			    ("NULL registered buffer p_dto: 0x%p, entry: %d\n",
			     p_dto, i);
		}
		/* Copy the original registered buffer descriptor */
		memcpy(&iov[i],
		       &p_dto->regd[i]->mem_reg.lmr_triplet,
		       sizeof(DAT_LMR_TRIPLET));
		/* Sanity check - segment length should be non-zero */
		if (iov[i].segment_length == 0) {
			IPANIC
			    ("Zero segment length in p_dto: 0x%p, entry: %d, va: 0x%08X\n",
			     p_dto, i,
			     (unsigned int) iov[i].virtual_address);
		}
		/* Adjust size */
		if (p_dto->used_sz[i] > 0) {
			iov[i].segment_length = p_dto->used_sz[i];
		}
		/* Sanity check - offset and length should not exceed the registered segment length */
		if (p_dto->offset[i] + iov[i].segment_length >
		    p_dto->regd[i]->mem_reg.lmr_triplet.segment_length) {
			IPANIC
			    ("Used length: %ld, and offset: %d, exceed total reg.seg.length: %ld, in p_dto: 0x%p, entry: %d, va: 0x%08lX\n",
			     (unsigned long) iov[i].segment_length,
			     p_dto->offset[i],
			     (unsigned long) p_dto->regd[i]->mem_reg.
			     lmr_triplet.segment_length, p_dto, i,
			     (unsigned long) iov[i].virtual_address);
		}
		/* Adjust offset */
		iov[i].virtual_address += p_dto->offset[i];

		ITRACE(trace_ctxt,
		       "IOV[%d] va: 0x%08lX sz: %ld lkey:0x%08X lmr_h: 0x%p\n",
		       i, (unsigned long) iov[i].virtual_address,
		       (unsigned long) iov[i].segment_length,
		       (unsigned int) iov[i].lmr_context,
		       p_dto->regd[i]->mem_reg.lmr_handle);
	}
}				/* iser_dto_to_iov */

/**
 * iser_post_recv - Posts a receive buffer.
 *
 * returns 0 on success, -1 on failure
 */
int iser_post_recv(struct iser_dto_t *p_recv_dto)	//!< IN - PostReceive DTO descriptor
{
	unsigned long flags;
	int ret_val = 0;
	DAT_DTO_COOKIE cookie;
	DAT_RETURN dat_ret;
	struct iser_conn_t *p_iser_conn;
	DAT_LMR_TRIPLET iov[MAX_REGD_BUF_VECTOR_LEN];

	ITRACE_ENTRY();

	/* Create a cookie */
	cookie.as_ptr = p_recv_dto;

	/* Retrieve connection */
	p_iser_conn = p_recv_dto->p_conn;
	if (p_iser_conn == NULL) {
		IPANIC("NULL p_conn in dto: 0x%p\n", p_recv_dto);
	}

	/* Translate into I/O vector */
	if (p_recv_dto->regd_vector_len > MAX_REGD_BUF_VECTOR_LEN) {
		IPANIC("DTO regd_vector_len exceeds maximal IOV len\n");
	}

	iser_dto_to_iov(p_recv_dto, ISER_TRACE_RECV_DTO, iov);

//    iser_cache_invalidate_dto(p_recv_dto);
	// Post rcv buffers
	spin_lock_irqsave(&p_iser_conn->post_rx_lock, flags);
	dat_ret = dat_ep_post_recv(p_iser_conn->ep_handle,
				   p_recv_dto->regd_vector_len,
				   iov,
				   cookie, DAT_COMPLETION_DEFAULT_FLAG);
	spin_unlock_irqrestore(&p_iser_conn->post_rx_lock, flags);

	if (dat_ret != DAT_SUCCESS) {
		print_dat_error("dat_ep_post_recv failed!", dat_ret);
		ret_val = -1;
		goto post_recv_exit;
	}

	ITRACE(ISER_TRACE_RECV_DTO,
	       "Posted recv buf, p_dto: 0x%p, IOV len: %d\n", p_recv_dto,
	       p_recv_dto->regd_vector_len);

      post_recv_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_post_recv */

/**
 * iser_start_dto - Initiate a DTO operation (PostSend, RDMA-W, RDMA_RMAX_REGD_BUF_VECTOR_LEN)
 *
 * returns 0 on success, -1 on failure
 */
int iser_start_dto(struct iser_dto_t *p_dto)	//<! IN - iSER DTO descriptor
{
	unsigned long flags;
	int ret_val = 0;
	DAT_DTO_COOKIE cookie;
	struct iser_conn_t *p_iser_conn;
	DAT_RETURN dat_ret;
	DAT_COMPLETION_FLAGS completion_flags;
	DAT_LMR_TRIPLET iov[MAX_REGD_BUF_VECTOR_LEN];

	ITRACE_ENTRY();

	if (p_dto == NULL) {
		IPANIC("NULL p_dto\n");
	}
	/* Initialize a cookie */
	cookie.as_ptr = p_dto;

	/* Retrieve connection */
	p_iser_conn = p_dto->p_conn;
	if (p_iser_conn == NULL) {
		IPANIC("NULL p_conn in dto: 0x%p\n", p_dto);
	}
	/* Check task */
#if 0
	if (p_dto->p_task == NULL) {
		IPANIC("NULL p_task in dto: 0x%p\n", p_dto);
	}
#endif

	completion_flags = p_dto->notify_enable ?
	    DAT_COMPLETION_DEFAULT_FLAG : DAT_COMPLETION_SUPPRESS_FLAG;

	/* Translate into I/O vec iser_complete_connection_terminationtor */
	if (p_dto->regd_vector_len > MAX_REGD_BUF_VECTOR_LEN) {
		IPANIC("DTO regd_vector_len exceeds maximal IOV len\n");
	}


	iser_dto_to_iov(p_dto, ISER_TRACE_SEND_DTO, iov);

	switch (p_dto->type) {
	case ISER_DTO_SEND:
//#ifdef NO_BUFF_POOL
		// flush the cache 
		iser_cache_flush_dto(p_dto);
//#endif
		spin_lock_irqsave(&p_iser_conn->post_tx_lock, flags);
		dat_ret = dat_ep_post_send(p_iser_conn->ep_handle, p_dto->regd_vector_len,	/* I/O vector sz */
					   iov,	/* LMR triplets I/O vector */
					   cookie, completion_flags);
		spin_unlock_irqrestore(&p_iser_conn->post_tx_lock, flags);

		if (dat_ret != DAT_SUCCESS) {
			IERROR
			    ("Failed to start SEND DTO, p_dto: 0x%p, IOV len: %d\n",
			     p_dto, p_dto->regd_vector_len);
			print_dat_error("dat_ep_post_send failed!",
					dat_ret);
			ret_val = -1;
			goto start_dto_exit;
		}

		atomic_inc(&p_iser_conn->post_send_buf_count);

		ITRACE(ISER_TRACE_SEND_DTO,
		       "Started SEND DTO, p_dto: 0x%p, IOV len: %d, Notify: %d, Posted sends: %d\n",
		       p_dto, p_dto->regd_vector_len, p_dto->notify_enable,
		       atomic_read(&p_iser_conn->post_send_buf_count));
		break;

	case ISER_DTO_RDMA_W:
		spin_lock_irqsave(&p_iser_conn->post_tx_lock, flags);
		dat_ret = dat_ep_post_rdma_write(p_iser_conn->ep_handle, p_dto->regd_vector_len,	/* I/O vector sz */
						 iov,	/* LMR triplets I/O vector - local */
						 cookie, &p_dto->as.rdma.remote,	/* remote buffer */
						 completion_flags);
		spin_unlock_irqrestore(&p_iser_conn->post_tx_lock, flags);

		if (dat_ret != DAT_SUCCESS) {
			IERROR
			    ("Failed to start RDMA-W DTO, p_dto: 0x%p, IOV len: %d, RMR: 0x%08X, VA: 0x%08lX, len: %ld\n",
			     p_dto, p_dto->regd_vector_len,
			     p_dto->as.rdma.remote.rmr_context,
			     (unsigned long) p_dto->as.rdma.remote.
			     target_address,
			     (unsigned long) p_dto->as.rdma.remote.
			     segment_length);
			print_dat_error("dat_ep_post_send failed!",
					dat_ret);
			ret_val = -1;
			goto start_dto_exit;
		}
		ITRACE(ISER_TRACE_SEND_DTO,
		       "Started RDMA-W DTO, p_dto: 0x%p, IOV len: %d, RMR: 0x%08X, VA: 0x%08lX, len: %ld\n",
		       p_dto, p_dto->regd_vector_len,
		       p_dto->as.rdma.remote.rmr_context,
		       (unsigned long) p_dto->as.rdma.remote.
		       target_address,
		       (unsigned long) p_dto->as.rdma.remote.
		       segment_length);
		break;

	case ISER_DTO_RDMA_R:
		spin_lock_irqsave(&p_iser_conn->post_tx_lock, flags);
		dat_ret = dat_ep_post_rdma_read(p_iser_conn->ep_handle, p_dto->regd_vector_len,	/* I/O vector sz */
						iov,	/* LMR triplets I/O vector - local */
						cookie, &p_dto->as.rdma.remote,	/* remote buffer */
						completion_flags);
		spin_unlock_irqrestore(&p_iser_conn->post_tx_lock, flags);

		if (dat_ret != DAT_SUCCESS) {
			IERROR
			    ("Failed to start RDMA-R DTO, p_dto: 0x%p, IOV len: %d, RMR: 0x%08X, VA: 0x%08lX, len: %ld\n",
			     p_dto, p_dto->regd_vector_len,
			     p_dto->as.rdma.remote.rmr_context,
			     (unsigned long) p_dto->as.rdma.remote.
			     target_address,
			     (unsigned long) p_dto->as.rdma.remote.
			     segment_length);
			print_dat_error("dat_ep_post_send failed!",
					dat_ret);
			ret_val = -1;
			goto start_dto_exit;
		}
		ITRACE(ISER_TRACE_SEND_DTO,
		       "Started RDMA-R DTO, p_dto: 0x%p, IOV len: %d, RMR: 0x%08X, VA: 0x%08lX, len: %ld\n",
		       p_dto, p_dto->regd_vector_len,
		       p_dto->as.rdma.remote.rmr_context,
		       (unsigned long) p_dto->as.rdma.remote.
		       target_address,
		       (unsigned long) p_dto->as.rdma.remote.
		       segment_length);
		break;

	default:
		IPANIC("Illegal DTO type for iser_start_dto\n");

	}			/* switch p_dto->type */

      start_dto_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_start_dto */

/**
 * iser_evd_upcall - Upcall of the global iSER EVD. Handles all events.
 *
 * returns 0 on success, -1 on failure
 */
static void iser_evd_upcall(DAT_PVOID inst,	//<! instance data passed during upcall registration
			    const DAT_EVENT * event,	//<! event descriptor
			    DAT_BOOLEAN has_more)	//<! non-zero if has more events pending
{
	struct iser_adaptor_t *p_adaptor;
	DAT_UPCALL_OBJECT upcall;

	ITRACE_ENTRY();

	p_adaptor = (struct iser_adaptor_t *) inst;
	upcall.instance_data = (DAT_PVOID) p_adaptor;
	upcall.upcall_func = iser_evd_upcall;

	ITRACE(ISER_TRACE_EVENT_UPCALL,
	       "ev->event_number = 0x%X,calling  modify upcall to DISABLE\n",
	       event->event_number);

	dat_evd_modify_upcall(p_adaptor->gnrl_evd_hdl, DAT_UPCALL_DISABLE,	// don't call more
			      &upcall);

	// Save the first event
	if (!p_adaptor->has_first_event) {
		memcpy(&p_adaptor->event, event, sizeof(DAT_EVENT));
		p_adaptor->has_first_event = 1;
	} else {
		IERROR("Upcall found has_first_event set!\n");
	}

	// Wake up the events thread
	wake_up_interruptible(&p_adaptor->dat_events_wait_q);

	ITRACE_EXIT();
}				/* iser_evd_upcall */

/**
 * iser_event_handler_thread - Event handling thread.
 *
 * returns 0 on success, -1 on failure
 */
int iser_event_handler_thread(void *adaptor)	//<! thread argument, adaptor descriptor is passed
{
	struct iser_adaptor_t *p_adaptor;
	DAT_RETURN dat_ret;
	DAT_UPCALL_OBJECT upcall;
	int evt_handle_err;
	int process_cnt = 0;
	int schedule_cnt = 0;

	DAEMONIZE(current, "iser_events");
	current->policy = 0;
	current->flags |= PF_MEMALLOC;
	siginitsetinv(&current->blocked,
		      sigmask(SIGKILL) | sigmask(SIGINT) |
		      sigmask(SIGTERM));

	p_adaptor = (struct iser_adaptor_t *) adaptor;
	upcall.instance_data = (DAT_PVOID) p_adaptor;
	upcall.upcall_func = iser_evd_upcall;

	up(&p_adaptor->startstop_sem);

	while (!signal_pending(current)) {
		if (atomic_read(&p_adaptor->in_callback) == 0) {

			ITRACE(ISER_TRACE_EVENT_THREAD,
			       "waiting for 1st event, pid:%d\n",
			       current->pid);
			wait_event_interruptible(p_adaptor->
						 dat_events_wait_q,
						 p_adaptor->
						 has_first_event);
			if (signal_pending(current))
				continue;

			ITRACE(ISER_TRACE_EVENT_THREAD,
			       "woke up with 1st event, pid:%d\n",
			       current->pid);
			evt_handle_err =
			    iser_handle_single_kdapl_event(p_adaptor);
			if (evt_handle_err != 0)
				IERROR("failed to handle 1st event\n");

			if (signal_pending(current)) {
				ITRACE(ISER_TRACE_EVENT_THREAD,
				       "Thread killed after handling 1st event, pid:%d\n",
				       current->pid);
				continue;
			}
			process_cnt = 1;
		} else
			ITRACE(ISER_TRACE_EVENT_THREAD,
			       "New co-thread started, pid:%d\n",
			       current->pid);

		do {
			if (process_cnt > 6) {
				process_cnt = 0;
				ITRACE(ISER_TRACE_EVENT_THREAD,
				       "schedule(), processing break\n");
				schedule();
			}

			/* Try to retrieve more events */
			dat_ret =
			    dat_evd_dequeue(p_adaptor->gnrl_evd_hdl,
					    &p_adaptor->event);

			if (dat_ret == DAT_SUCCESS) {	/* got another event */
				evt_handle_err =
				    iser_handle_single_kdapl_event
				    (p_adaptor);
				if (evt_handle_err != 0)
					IERROR
					    ("failed to handle polled event, pid:%d\n",
					     current->pid);

				if (signal_pending(current)) {
					ITRACE(ISER_TRACE_EVENT_THREAD,
					       "Thread killed after handling polled event, pid:%d\n",
					       current->pid);
					break;
				}

				process_cnt++;
				schedule_cnt = 0;
			} else if (DAT_GET_TYPE(dat_ret) == DAT_QUEUE_EMPTY) {	/* no more events */
				DAT_RETURN modify_upcall_ret;

				if (schedule_cnt < 4) {
					schedule_cnt++;
					process_cnt = 0;
					ITRACE(ISER_TRACE_EVENT_THREAD,
					       "schedule() #%d, polling\n",
					       schedule_cnt);
					schedule();
					dat_ret = DAT_SUCCESS;
					continue;
				} else {
					schedule_cnt = 0;
				}

				/* Re-enable upcalls */
				p_adaptor->has_first_event = 0;
				modify_upcall_ret =
				    dat_evd_modify_upcall(p_adaptor->
							  gnrl_evd_hdl,
							  DAT_UPCALL_SINGLE_INSTANCE,
							  &upcall);
				if (modify_upcall_ret == DAT_SUCCESS)
					ITRACE(ISER_TRACE_EVENT_THREAD,
					       "No more events, pid:%d\n",
					       current->pid);
				else if (DAT_GET_TYPE(modify_upcall_ret) ==
					 DAT_INTERRUPTED_CALL) {
					ITRACE(ISER_TRACE_EVENT_THREAD,
					       "more events pending, continue processing, pid:%d\n",
					       current->pid);
					dat_ret = DAT_SUCCESS;	/* more events became pending, continue */
				} else
					IPANIC
					    ("Failed to modify upcall to SINGLE INSTANCE, rc:%X, pid:%d\n",
					     modify_upcall_ret,
					     current->pid);
			} else
				IPANIC("Failed to dequeue event\n");

		} while (dat_ret == DAT_SUCCESS);
	}			/* thread main loop */

	if (p_adaptor->event_thrd_pid == current->pid) {
		lock_kernel();
		mb();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
		reparent_to_init();
#endif
		up(&p_adaptor->startstop_sem);
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "Active event thread terminated, pid:%d\n",
		       current->pid);
	} else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
		reparent_to_init();
#endif
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "Inactive event thread terminated, pid:%d\n",
		       current->pid);
	}

	return 0;
}				/* iser_event_handler_thread */

/**
 * iser_spawn_event_thread - Spawns a new isntance of the event thread
 *
 * returns 0 on success, -1 on failure
 */
int iser_spawn_event_thread(struct iser_adaptor_t *p_adaptor)	//!< IN - iSER adaptor descriptor
{
	int ret_val = 0;

	ITRACE_ENTRY();

	init_MUTEX_LOCKED(&p_adaptor->startstop_sem);
	p_adaptor->event_thrd_pid = kernel_thread(iser_event_handler_thread, p_adaptor,	/* passed to the thread function */
						  CLONE_VM | CLONE_FS |
						  CLONE_FILES |
						  CLONE_SIGHAND);
	if (p_adaptor->event_thrd_pid > 0)
		down(&p_adaptor->startstop_sem);	/* wait for thread to start */
	else
		ret_val = -1;

	ITRACE_EXIT();
	return ret_val;
}				/* iser_spawn_event_thread */


/**
 * iser_suicide_thread - send kill signal to current thread
 *
 */
void iser_suicide_thread()
{
	ITRACE(ISER_TRACE_EVENT_THREAD,
	       "Sending kill signal to self, pid:%d\n", current->pid);
	kill_proc(current->pid, SIGKILL, 1);
}				/* iser_suicide_thread */

/**
 * iser_kill_event_thread - Spawns a new isntance of the event thread
 *
 */
void iser_kill_event_thread(struct iser_adaptor_t *p_adaptor)	//!< IN - iSER adaptor descriptor
{
	ITRACE(ISER_TRACE_EVENT_THREAD,
	       "Sending kill signal to active event thread, pid:%d\n",
	       p_adaptor->event_thrd_pid);
	lock_kernel();
	init_MUTEX_LOCKED(&p_adaptor->startstop_sem);
	mb();
	kill_proc(p_adaptor->event_thrd_pid, SIGKILL, 1);
	down(&p_adaptor->startstop_sem);
	unlock_kernel();
	ITRACE(ISER_TRACE_EVENT_THREAD, "event thread terminated\n");
}				/* iser_suicide_thread */

/**
 * iser_handle_connection_event - Handles a connection event
 *
 * returns 0 on success, -1 on failure
 */
static int iser_handle_connection_event(struct iser_adaptor_t *p_adaptor,	//!< IN - iSER adaptor descriptor
					DAT_EP_HANDLE ep_handle,	//!< IN - Connection End-Point handle
					DAT_EVENT_NUMBER event_number)	//!< IN - DAT conn. event number
{
	struct iser_conn_t *p_iser_conn;
	int ret_val = 0;

	ITRACE_ENTRY();

	p_iser_conn =
	    iser_adaptor_find_connection(p_adaptor, (void *) ep_handle);
	if (p_iser_conn == NULL) {
		IERROR("Connection not found, EP handle= 0x%p\n",
		       ep_handle);
		ret_val = -1;
		goto handle_connection_event_exit;
	}

	switch (event_number) {
	case DAT_CONNECTION_EVENT_ESTABLISHED:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection established\n");
		break;
	case DAT_CONNECTION_EVENT_DISCONNECTED:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection disconnected\n");
		break;
	case DAT_CONNECTION_EVENT_PEER_REJECTED:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection peer rejected\n");
		break;
	case DAT_CONNECTION_EVENT_NON_PEER_REJECTED:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection non-peer rejected\n");
		break;
	case DAT_CONNECTION_EVENT_ACCEPT_COMPLETION_ERROR:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection accept compl. error\n");
		break;
	case DAT_CONNECTION_EVENT_BROKEN:
		ITRACE(ISER_TRACE_EVENT_THREAD, "connection broken\n");
		break;
	case DAT_CONNECTION_EVENT_TIMED_OUT:
		ITRACE(ISER_TRACE_EVENT_THREAD, "connection timed-out\n");
		break;
	case DAT_CONNECTION_EVENT_UNREACHABLE:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection unreachable\n");
		break;
	default:
		ITRACE(ISER_TRACE_EVENT_THREAD,
		       "connection event: irrelevant\n");
		break;
	}

	switch (event_number) {
	case DAT_CONNECTION_EVENT_ESTABLISHED:
		/* Set the new connection state */
		atomic_set(&p_iser_conn->state, ISER_CONN_UP);

		/* Notify about connection establishment */
		atomic_inc(&p_adaptor->in_callback);
		p_iser_conn->p_entity->ext_api_cb.
		    connection_establish_notify(p_iser_conn->iscsi_conn_h,
						1);
		atomic_dec(&p_adaptor->in_callback);

		break;

	case DAT_CONNECTION_EVENT_DISCONNECTED:
	case DAT_CONNECTION_EVENT_BROKEN:
		p_iser_conn->disc_evt_flag = 1;

		/* If this event is unsolicited this means that the connection is being
		   terminated asynchronously from the iSCSI layer's perspective. */
		if (atomic_read(&p_iser_conn->state) == ISER_CONN_PENDING) {
			atomic_set(&p_iser_conn->state, ISER_CONN_DOWN);

			atomic_inc(&p_adaptor->in_callback);
			p_iser_conn->p_entity->ext_api_cb.
			    connection_establish_notify(p_iser_conn->
							iscsi_conn_h, 0);
			atomic_dec(&p_adaptor->in_callback);

			iser_conn_free(p_iser_conn);
		} else {
			if (atomic_read(&p_iser_conn->state) ==
			    ISER_CONN_UP) {
				atomic_set(&p_iser_conn->state,
					   ISER_CONN_ASYNC_TERM);
				ITRACE(ISER_TRACE_CONN,
				       "Conn. 0x%p is being terminated asynchronously\n",
				       p_iser_conn);
			}
			/* Complete the termination process if possible (no events pending) */
			iser_complete_connection_termination(p_iser_conn);
		}
		break;

	case DAT_CONNECTION_EVENT_PEER_REJECTED:
	case DAT_CONNECTION_EVENT_NON_PEER_REJECTED:
	case DAT_CONNECTION_EVENT_ACCEPT_COMPLETION_ERROR:
	case DAT_CONNECTION_EVENT_TIMED_OUT:
	case DAT_CONNECTION_EVENT_UNREACHABLE:
		if (atomic_read(&p_iser_conn->state) == ISER_CONN_PENDING) {
			atomic_set(&p_iser_conn->state, ISER_CONN_DOWN);

			atomic_inc(&p_adaptor->in_callback);
			p_iser_conn->p_entity->ext_api_cb.
			    connection_establish_notify(p_iser_conn->
							iscsi_conn_h, 0);
			atomic_dec(&p_adaptor->in_callback);

			iser_conn_free(p_iser_conn);
			break;
		} else {
			IERROR
			    ("Event unexpected in current conn. state: %s\n",
			     iser_conn_get_state_name(p_iser_conn));
		}

	default:
		IERROR("Connection event unhandled, %d\n", event_number);
		break;
	}

      handle_connection_event_exit:
	ITRACE_EXIT();
	return ret_val;
}				/* iser_handle_connection_event */

static int iser_handle_async_event(DAT_EVENT * ev)
{
	ITRACE_ENTRY();
	switch (ev->event_number) {
	case DAT_ASYNC_ERROR_EVD_OVERFLOW:
		IERROR("DAT_ASYNC_ERROR_EVD_OVERFLOW\n");
		break;
	case DAT_ASYNC_ERROR_IA_CATASTROPHIC:
		IERROR("DAT_ASYNC_ERROR_IA_CATASTROPHIC\n");
		break;
	case DAT_ASYNC_ERROR_EP_BROKEN:
		IERROR("DAT_ASYNC_ERROR_EP_BROKEN\n");
		break;
	case DAT_ASYNC_ERROR_TIMED_OUT:
		IERROR("DAT_ASYNC_ERROR_TIMED_OUT\n");
		break;
	case DAT_ASYNC_ERROR_PROVIDER_INTERNAL_ERROR:
		IERROR("DAT_ASYNC_ERROR_PROVIDER_INTERNAL_ERROR\n");
		break;
	default:
		IPANIC("Unexpected event number: 0x%X(%d)\n",
		       ev->event_number, ev->event_number);
	}

	// ToDo: release adaptor by
	// ev->event_data.asynch_error_event_data.ia_handle
	// or at least all its connections:
	// iser_conn_async_terminate for all

	ITRACE_EXIT();
	return 0;
}				/* iser_handle_async_event */

/**
 * iser_handle_single_kdapl_event - Single event handler.
 *
 * returns 0 on success, -1 on failure
 */
static int iser_handle_single_kdapl_event(struct iser_adaptor_t *p_adaptor)	//<! iSER adaptor with event pending
{
	DAT_EVENT *ev;
	DAT_RETURN dat_ret;
	int ret_val = 0;

	ITRACE_ENTRY();

	/* Retrieve the last event */
	ev = &p_adaptor->event;

	switch (ev->event_number) {
		/* --------------------------------------------------------------
		   CONNECTION REQUEST EVENTS
		   ---------------------------------------------------------------- */
	case DAT_CONNECTION_REQUEST_EVENT:
		{
			DAT_CR_HANDLE cr_handle =
			    ev->event_data.cr_arrival_event_data.cr_handle;
			struct sockaddr_in *local_addr =
			    (struct sockaddr_in *) ev->event_data.
			    cr_arrival_event_data.local_ia_address_ptr;
			unsigned int port =
			    ((unsigned int) ev->event_data.
			     cr_arrival_event_data.
			     conn_qual) & CONN_QUAL_PORT_MASK;

			ret_val =
			    iser_target_handle_connection_request
			    (p_adaptor, cr_handle, local_addr, port);
			if (ret_val != 0) {
				IERROR
				    ("Failed to handle connection request event, cr_h: 0x%p, port: %d\n",
				     cr_handle, port);
				dat_ret = dat_cr_reject(cr_handle);
				if (dat_ret != DAT_SUCCESS) {
					print_dat_error
					    ("dat_cr_reject failed",
					     dat_ret);
				}
			}
		}
		break;

		/* --------------------------------------------------------------
		   CONNECTION EVENTS
		   ---------------------------------------------------------------- */
	case DAT_CONNECTION_EVENT_ESTABLISHED:
	case DAT_CONNECTION_EVENT_PEER_REJECTED:
	case DAT_CONNECTION_EVENT_NON_PEER_REJECTED:
	case DAT_CONNECTION_EVENT_ACCEPT_COMPLETION_ERROR:
	case DAT_CONNECTION_EVENT_DISCONNECTED:
	case DAT_CONNECTION_EVENT_BROKEN:
	case DAT_CONNECTION_EVENT_TIMED_OUT:
	case DAT_CONNECTION_EVENT_UNREACHABLE:
		ret_val = iser_handle_connection_event(p_adaptor,
						       ev->event_data.
						       connect_event_data.
						       ep_handle,
						       ev->event_number);
		if (ret_val != 0) {
			IERROR
			    ("Failed to handle connection event no.: %d\n",
			     ev->event_number);
		}
		break;

		/* --------------------------------------------------------------
		   DTO COMPLETION EVENTS
		   ---------------------------------------------------------------- */
	case DAT_DTO_COMPLETION_EVENT:{
			struct iser_dto_t *p_dto;
			struct iser_conn_t *p_iser_conn;
			unsigned long dto_xfer_len;

			p_dto =
			    (struct iser_dto_t *) ev->event_data.
			    dto_completion_event_data.user_cookie.as_ptr;
			dto_xfer_len =
			    (unsigned long) ev->event_data.
			    dto_completion_event_data.transfered_length;

			if (p_dto == NULL) {
				IPANIC("NULL p_dto\n");
			}
			p_iser_conn = p_dto->p_conn;
			if (p_iser_conn == NULL) {
				IPANIC("NULL connection in p_dto: 0x%p\n",
				       p_dto);
			}
			if (p_dto->type >= ISER_DTO_PASSIVE) {
				IPANIC
				    ("Unexpected type: %d in p_dto: 0x%p\n",
				     p_dto->type, p_dto);
			}

			switch (ev->event_data.dto_completion_event_data.
				status) {
			case DAT_DTO_SUCCESS:
				ITRACE(ISER_TRACE_EVENT_THREAD,
				       "DTO completion success, Cookie(DTO): 0x%p, Type: %s, Xfer Len: %ld, ep: 0x%p\n",
				       p_dto,
				       dto_type_name[p_dto->type],
				       dto_xfer_len,
				       ev->event_data.
				       dto_completion_event_data.
				       ep_handle);

				/* Handle successful DTO completion for target or initiator */
				ret_val =
				    p_iser_conn->p_entity->
				    dto_completion_cb(p_dto, dto_xfer_len);

				if (ret_val != 0) {
					IERROR
					    ("Failed to handle successful DTO completion\n");
				}
				break;

			case DAT_DTO_ERR_FLUSHED:
			case DAT_DTO_ERR_LOCAL_LENGTH:
			case DAT_DTO_ERR_LOCAL_EP:
			case DAT_DTO_ERR_LOCAL_PROTECTION:
			case DAT_DTO_ERR_BAD_RESPONSE:
			case DAT_DTO_ERR_REMOTE_ACCESS:
			case DAT_DTO_ERR_REMOTE_RESPONDER:
			case DAT_DTO_ERR_TRANSPORT:
			case DAT_DTO_ERR_RECEIVER_NOT_READY:
			case DAT_DTO_ERR_PARTIAL_PACKET:
			case DAT_RMR_OPERATION_FAILED:

				ITRACE(ISER_TRACE_ERR_DTO,
				       "DTO completion error, Cookie(DTO): 0x%p, Type: %s, Status: %s (%d), Xfer Len: %ld, ep: 0x%p\n",
				       p_dto,
				       dto_type_name[p_dto->type],
				       (ev->event_data.
					dto_completion_event_data.status <=
					DAT_RMR_OPERATION_FAILED ?
					dto_status_name[ev->event_data.
							dto_completion_event_data.
							status] :
					"UNKNOWN"),
				       ev->event_data.
				       dto_completion_event_data.status,
				       dto_xfer_len,
				       ev->event_data.
				       dto_completion_event_data.
				       ep_handle);

				/* Since dapl doesn't notify us about DAT_CONNECTION_EVENT_BROKEN event,
				   we need to initiate a disconnection */
				if (ev->event_data.
				    dto_completion_event_data.status !=
				    DAT_DTO_ERR_FLUSHED) {
					ITRACE(ISER_TRACE_CONN,
					       "Got dto event 0x%X. Calling iser_conn_async_terminate\n",
					       ev->event_data.
					       dto_completion_event_data.
					       status);
					ret_val =
					    iser_conn_async_terminate
					    (p_iser_conn);
					if (ret_val != 0) {
						IERROR
						    ("Failed to terminate the connection asyncronously\n");
					}
				}

				/* If this event is unsolicited this means that the connection is being
				   terminated asynchronously from the iSCSI layer's perspective. */
				if (atomic_read(&p_iser_conn->state) ==
				    ISER_CONN_UP) {
					atomic_set(&p_iser_conn->state,
						   ISER_CONN_ASYNC_TERM);
					ITRACE(ISER_TRACE_CONN,
					       "Conn. 0x%p is being terminated asynchronously\n",
					       p_iser_conn);
				}

				ret_val = iser_dto_completion_error(p_dto);
				if (ret_val != 0) {
					IERROR
					    ("Failed to handle ERROR DTO completion\n");
				}
				break;

			default:
				IPANIC
				    ("Unknown KDAPL DTO completion status: %d\n",
				     ev->event_data.
				     dto_completion_event_data.status);
				break;
			}	/* switch dto_status */
		}
		break;

		/* --------------------------------------------------------------
		   ASYNCHRONOUS EVENTS
		   ---------------------------------------------------------------- */
	case DAT_ASYNC_ERROR_EVD_OVERFLOW:
	case DAT_ASYNC_ERROR_IA_CATASTROPHIC:
	case DAT_ASYNC_ERROR_EP_BROKEN:
	case DAT_ASYNC_ERROR_TIMED_OUT:
	case DAT_ASYNC_ERROR_PROVIDER_INTERNAL_ERROR:
		ret_val = iser_handle_async_event(ev);

		if (ret_val != 0) {
			IERROR("Failed to handle Async Event\n");
		}
		break;

		/* --------------------------------------------------------------
		   SOFTWARE-GENERATED EVENTS
		   ---------------------------------------------------------------- */
	case DAT_SOFTWARE_EVENT:
		ITRACE(ISER_TRACE_EVENT_THREAD, "DAT_SOFTWARE_EVENT\n");
		break;

		/* --------------------------------------------------------------
		   UNHANDLED EVENTS
		   ---------------------------------------------------------------- */
	default:
		IERROR("Unexpected event number: 0x%X\n",
		       ev->event_number);
		break;

	}			/* switch event number */

	ITRACE_EXIT();
	return ret_val;
}				/* iser_handle_single_kdapl_event */
