/*	do_scsi_command.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 */

#define _POSIX_C_SOURCE 200112L

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <pthread.h>

#if !defined(REPORT_LUNS)
#define	REPORT_LUNS	0xa0
#endif

#include "../initiator/iscsi_initiator.h"

#define	BYTES_PER_BLOCK		512

/* maximum numer of outstanding commands to queue with iSCSI using a
 * `repeat'. */
#define OUTSTANDING_CMNDS 10

/**
 * The total length (in bytes) of data transferred by read and write
 * commands over a repeat loop.
 */
unsigned long long total_data_xfer_len;

/* number of completed commands in a repeat situation */
unsigned int completed_commands;

/* Don't queue all the commands at once (for memory purposes).  This
 * condition wakes us up when we are ready to queue more. */
pthread_cond_t queue_more_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t queue_more_mutex = PTHREAD_MUTEX_INITIALIZER;

static void print_short_capacity(void *data)
{
	printf("\tREADCAPACITY returned logical block address %0x08x\n",
		ntohl(((unsigned int *)data)[0]));
	printf("\tREADCAPACITY returned block length in bytes %u (%0x08x)\n",
		ntohl(((unsigned int *)data)[1]),
		ntohl(((unsigned int *)data)[1]));
}


static void print_long_capacity(void *data)
{
	int			n;
	unsigned long long int	value;

	value = 0LL;
	for( n = 0; n < 8; n++ )
		value = (value << 8) + ((unsigned char *)data)[n];
	printf("\tREADCAPACITY returned logical block address 0x%16llx\n",
		value);;
	printf("\tREADCAPACITY returned block length in bytes %u (0x%08x)\n",
		ntohl(((unsigned int *)data)[2]),
		ntohl(((unsigned int *)data)[2]));
}


/*
 * if count is non-zero, prints number of duplicates of previously printed buf1
 * then always prints buf2 and copies it into buf1.
 */
static void print_a_line(int count, char *buf1, char *buf2)
{
	if (count > 0) {
		/* at least one duplicate followed previously printed buf1 */
		if (count > 1) {
			printf("previous line repeated %d more times\n", count);
		} else
			printf("%s", buf1);
	}
	printf("%s", buf2);
	strcpy(buf1, buf2);
}


/* dump out the data just read, 16 characters per line in hex and ascii,
 * but don't print duplicate lines
 */
void done_reading(void *buffer, unsigned int bufflen, char *message)
{
	unsigned int	k;		/* number of chars in buf2 so far */
	unsigned int	j;		/* print position of this char, 0..15 */
	unsigned int	count;		/* number of duplicate lines so far */
	unsigned int	i;		/* index of this char in buffer */
	unsigned char	*cptr;		/* pointer to this char in buffer */
	char		ascii[17], buf1[200], buf2[200];

	if (!TRACE_TEST(TRACE_BUF))
		return;

	if ((cptr = buffer) != NULL) {
		printf("\tread %u bytes of %s\n", bufflen, message);
		ascii[16] = '\0';
		count = 0;
		buf1[0] = '\0';
		k = 0;
		/* once around this loop for each character in buffer */
		for (i = 0; i < bufflen; i++) {
			j = i & 0xf;
			if (j == 0) {
				/* ith char is first char on next line */
				if (i > 0) {
					/* this line in buf2 is full */
					k+=sprintf(&buf2[k],"   |%s|\n",ascii);
					if (strcmp(&buf1[11], &buf2[11]) == 0) {
						/* new line identical to prev */
						count++;
					} else {
						/* this line differs from prev*/
						print_a_line(count, buf1, buf2);
						count = 0;
					}
				}
				/* start a new line */
				k = sprintf(buf2, "%10u:", i);
			}
			k += sprintf(&buf2[k], " %02x", *cptr);
			if (isprint(*cptr))
				ascii[j] = *cptr;
			else
				ascii[j] = ' ';
			cptr++;
		}
		/* finished complete buffer, may have partial last line */
		if (i > 0) {
			/* part of a line already built, pad it out and print */
			while( (j = i & 0xf) != 0) {
				k += sprintf(&buf2[k], " %2s", "");
				ascii[j] = ' ';
				i++;
			}
			k += sprintf(&buf2[k], "   |%s|\n", ascii);
			print_a_line(count, buf1, buf2);
		}
	}
}

void print_report_luns_list( unsigned long long *buffer, unsigned int len )
{
	unsigned int	j;
	len /= 8;
	for( j = 0; j < len; j++ )
		printf("\tlun[%u] 0x%0llx\n", j, buffer[j]);
}

void print_sense_data( unsigned char *buffer, int len )
{
	printf("\tsense key 0x%02x\n", buffer[2] & 0x0f);
	printf("\tadditional sense length %d\n", buffer[7]);
	printf("\tadditional sense code 0x%02x\n", buffer[12]);
	printf("\tadditional sense code qualifier 0x%02x\n", buffer[13]);
	done_reading(buffer, len, "sense data");
}


void done(struct scsi_cmnd *Cmnd)
{
	unsigned int	len;

	TRACE(TRACE_DEBUG, "in \"done\" for Cmnd %p, result 0x%08x\n", Cmnd, Cmnd->result);
	if (Cmnd->result & (DRIVER_SENSE << 24)) {
		TRACE(TRACE_DEBUG, "\tdriver sense data present\n");
		print_sense_data(Cmnd->sense_buffer, SCSI_SENSE_BUFFERSIZE);
	}
	switch (Cmnd->cmnd[0])
	{
	case TEST_UNIT_READY:	/* TESTUNITREADY 0x00 */
		TRACE(TRACE_DEBUG, "in \"done\" for TESTUNITREADY\n");
		break;

	case INQUIRY:		/* INQUIRY 0x12 */
		TRACE(TRACE_DEBUG, "in \"done\" for INQUIRY\n");
		if (Cmnd->request_buffer != NULL) {
			TRACE(TRACE_DEBUG, "\tINQUIRY peripheral qualifier %d\n",
				((unsigned char *)Cmnd->request_buffer)[0]>>5);
			TRACE(TRACE_DEBUG, "\tINQUIRY peripheral device type %d\n",
				((unsigned char *)Cmnd->request_buffer)[0]&0x1f);
			TRACE(TRACE_DEBUG, "\tINQUIRY version %u\n",
				((unsigned char *)Cmnd->request_buffer)[2]);
			TRACE(TRACE_DEBUG, "\tINQUIRY response data format %d\n",
				((unsigned char *)Cmnd->request_buffer)[3]&0x0f);
			TRACE(TRACE_DEBUG, "\tINQUIRY additional length %u\n",
				((unsigned char *)Cmnd->request_buffer)[4]);
			free(Cmnd->request_buffer);
		}
		break;

	case REPORT_LUNS:	/* REPORT LUNS 0xa0 */
		TRACE(TRACE_DEBUG, "in \"done\" for REPORTLUNS\n");
		if (Cmnd->request_buffer != NULL) {
			len = ntohl(*(unsigned int *)Cmnd->request_buffer);
			TRACE(TRACE_DEBUG, "\tREPORTLUNS returned length %u\n", len);
			print_report_luns_list(
				&((unsigned long long*)Cmnd->request_buffer)[1],
				len);
			free(Cmnd->request_buffer);
		}
		break;

	case REQUEST_SENSE:	/* REQUEST SENSE 0x03 */
		TRACE(TRACE_DEBUG, "in \"done\" for REQUEST SENSE\n");
		if (Cmnd->request_buffer != NULL) {
			len = ntohl(*(unsigned short *)Cmnd->request_buffer);
			TRACE(TRACE_DEBUG, "\tREQUESTSENSE returned length %u\n", len);
			print_sense_data(&(
				(unsigned char *)Cmnd->request_buffer)[2], len);
			free(Cmnd->request_buffer);
		}
		break;

	case MODE_SENSE:	/* MODE SENSE 0x1a */
		TRACE(TRACE_DEBUG, "in \"done\" for MODE SENSE\n");
		if (Cmnd->request_buffer != NULL) {
			TRACE(TRACE_DEBUG, "\tMODESENSE returned mode data length %u\n",
				((unsigned char *)Cmnd->request_buffer)[0]);
			TRACE(TRACE_DEBUG, "\tmedium type %u\n",
				((unsigned char *)Cmnd->request_buffer)[1]);
			TRACE(TRACE_DEBUG, "\tdevice-specific parameter %u\n",
				((unsigned char *)Cmnd->request_buffer)[2]);
			TRACE(TRACE_DEBUG, "\tblock descriptor length %u\n",
				len=((unsigned char *)Cmnd->request_buffer)[3]);
			done_reading(&((unsigned char*)Cmnd->request_buffer)[4],
				len, "mode parameter block descriptor");
			free(Cmnd->request_buffer);
		}
		break;

	case READ_CAPACITY:	/* READ CAPACITY 0x25 */
		TRACE(TRACE_DEBUG, "in \"done\" for READCAPACITY\n");
		if (Cmnd->request_buffer != NULL) {
			if (Cmnd->cmnd[1] & 0x02) {	/* long form of data */
				print_long_capacity(Cmnd->request_buffer);
			} else {			/* short form of data */
				print_short_capacity(Cmnd->request_buffer);
			}
			free(Cmnd->request_buffer);
		}
		break;

	case READ_6:		/* READ6 0x08 */
		TRACE(TRACE_DEBUG, "in \"done\" for READ6\n");
		done_reading(Cmnd->request_buffer,Cmnd->request_bufflen,"data");
		free(Cmnd->request_buffer);
		break;
	case READ_10:		/* READ10 0x28 */
		TRACE(TRACE_DEBUG, "in \"done\" for READ10\n");
		done_reading(Cmnd->request_buffer,Cmnd->request_bufflen,"data");
		free(Cmnd->request_buffer);
		break;
	case READ_12:		/* READ12 0xa8 */
		TRACE(TRACE_DEBUG, "in \"done\" for READ12\n");
		done_reading(Cmnd->request_buffer,Cmnd->request_bufflen,"data");
		free(Cmnd->request_buffer);
		break;
	case 0x88:		/* READ16 0x88 */
		TRACE(TRACE_DEBUG, "in \"done\" for READ16\n");
		done_reading(Cmnd->request_buffer,Cmnd->request_bufflen,"data");
		free(Cmnd->request_buffer);
		break;
	case WRITE_6:		/* WRITE6 0x0a */
		TRACE(TRACE_DEBUG, "in \"done\" for WRITE6\n");
		free(Cmnd->request_buffer);
		break;
	case WRITE_10:		/* WRITE10 0x2a */
		TRACE(TRACE_DEBUG, "in \"done\" for WRITE10\n");
		free(Cmnd->request_buffer);
		break;
	case WRITE_12:		/* WRITE12 0xaa */
		TRACE(TRACE_DEBUG, "in \"done\" for WRITE12\n");
		free(Cmnd->request_buffer);
		break;
	case 0x8a:		/* WRITE16 0x8a */
		TRACE(TRACE_DEBUG, "in \"done\" for WRITE16\n");
		free(Cmnd->request_buffer);
		break;
	default:
		TRACE(TRACE_DEBUG, "in \"done\" for unkown command, opcode 0x%02x\n",
			Cmnd->cmnd[0]);
	}	/* switch */

	free(Cmnd);

	completed_commands += 1;
	pthread_cond_signal(&queue_more_cond);
}


static struct scsi_cmnd *set_up_cmnd( int argc, char *argv[] )
{
	struct scsi_cmnd	*Cmnd;

	Cmnd = calloc(1, sizeof(*Cmnd));
	if (Cmnd == NULL) {
		printf("No space for \"%s\" struct scsi_cmnd size %lu\n",
			argv[0], (unsigned long) sizeof(*Cmnd));
	} else {
		/* fill in basic stuff that is common to a lot of commands */
		Cmnd->device = global_host->device;
		Cmnd->cmd_len = Cmnd->old_cmd_len = 6;
		Cmnd->sc_data_direction = SCSI_DATA_READ;	 /* NONE ?? */
		Cmnd->sc_old_data_direction = Cmnd->sc_data_direction;
		Cmnd->scsi_done = done;
	}
	return Cmnd;
}


static struct scsi_cmnd *set_up_read_capacity( int argc, char *argv[],
						int longlba )
{
	int			size;
	struct scsi_cmnd	*Cmnd;

	if (argc > 1) {
		printf("Extra token(s) after \"%s\"\n", argv[0]);
		return NULL;
	}
	Cmnd = set_up_cmnd(argc,argv);
	if (Cmnd == NULL) {
		return NULL;
	}
	Cmnd->cmd_len = Cmnd->old_cmd_len = 10;
	Cmnd->cmnd[0] = READ_CAPACITY;	/* READ CAPACITY */
	if (longlba) {			/* long read capacity data */
		size = 12;
		Cmnd->cmnd[1] = 0x02;	/* LONGLBA bit */
	} else {			/* short read capacity data */
		size = 8;
	}
	Cmnd->request_bufflen = size;	/* actual request size */
	if ((Cmnd->request_buffer = calloc(1, size)) == NULL) {
		printf("No space for \"%s\" data size %d\n",
			argv[0], size);
		free(Cmnd);
		return NULL;
	}
	return Cmnd;
}


static struct scsi_cmnd *set_up_request_sense( int argc, char *argv[] )
{
	int			size;
	struct scsi_cmnd	*Cmnd;

	if (argc > 1) {
		printf("Extra token(s) after \"%s\"\n", argv[0]);
		return NULL;
	}
	Cmnd = set_up_cmnd(argc,argv);
	if (Cmnd == NULL) {
		return NULL;
	}
	Cmnd->cmd_len = Cmnd->old_cmd_len = 6;
	Cmnd->cmnd[0] = REQUEST_SENSE;		/* REQUEST SENSE 6 */
	Cmnd->cmnd[4] = size = 16;		/* actual request size */
	Cmnd->request_bufflen = size;		/* actual request size */
	if ((Cmnd->request_buffer = calloc(1, size)) == NULL) {
		printf("No space for \"%s\" requestsense size %d\n",
			argv[0], size);
		free(Cmnd);
		return NULL;
	}
	return Cmnd;
}


static struct scsi_cmnd *set_up_mode_sense( int argc, char *argv[] )
{
	int			size;
	struct scsi_cmnd	*Cmnd;

	if (argc > 1) {
		printf("Extra token(s) after \"%s\"\n", argv[0]);
		return NULL;
	}
	Cmnd = set_up_cmnd(argc,argv);
	if (Cmnd == NULL) {
		return NULL;
	}
	Cmnd->cmd_len = Cmnd->old_cmd_len = 6;
	Cmnd->cmnd[0] = MODE_SENSE;		/* MODE SENSE 6 */
	Cmnd->cmnd[4] = size = 12;		/* actual request size */
	Cmnd->request_bufflen = size;		/* actual request size */
	if ((Cmnd->request_buffer = calloc(1, size)) == NULL) {
		printf("No space for \"%s\" modesense size %d\n",
			argv[0], size);
		free(Cmnd);
		return NULL;
	}
	return Cmnd;
}


/* when called, opcode is READ_6 (0x08) for READ, WRITE_6 (0x0a) for WRITE */
static struct scsi_cmnd *set_up_read_write( int argc, char *argv[], int opcode,
						int size )
{
	char			*ptr;
	unsigned char		*bptr;
	int			nblocks, nbytes, blockno;
	struct scsi_cmnd	*Cmnd;
	unsigned char		fill_value;

	if (argc <= 1) {
		printf("Missing number-of-blocks after \"%s\"\n", argv[0]);
		return NULL;
	}
	if (argc > 2) {
		printf("Extra token(s) after \"%s\"\n", argv[0]);
		return NULL;
	}
	errno = 0;
	nblocks = strtol(argv[1], &ptr, 0);
	if (errno != 0 || ptr == argv[1] || nblocks < 0
		|| strspn(ptr, WHITE_SPACE) != strlen(ptr)) {
		printf("Illegal number-of-blocks \"%s\"\n", argv[1]);
		return NULL;
	}
	Cmnd = set_up_cmnd(argc,argv);
	if (Cmnd == NULL) {
		return NULL;	/* fatal */
	}
	total_data_xfer_len += nblocks * 512;
	switch (size) {
	case 0:	/* size depends on number-of-blocks */
		if (nblocks > 0 && nblocks <= 256) {
			size = 6;
			Cmnd->cmnd[4] = nblocks & 0xff;
		} else if (nblocks <= 65535) {
			size = 10;
			opcode |= 0x20;			/* READ10, WRITE10 */
			Cmnd->cmnd[7] = nblocks >> 8;
			Cmnd->cmnd[8] = nblocks & 0xff;
		} else {
			size = 16;
			opcode |= 0x80;			/* READ16, WRITE16 */
			*(unsigned int *)(&Cmnd->cmnd[10]) = htonl(nblocks);
		}
		break;
	case 6:	/* explicit READ6, WRITE6 */
		if (nblocks > 256) {
			printf("Number-of-blocks too big, limit is %d\n", 256);
			free(Cmnd);
			return NULL;
		}
		Cmnd->cmnd[4] = nblocks & 0xff;
		break;
	case 10:/* explicit READ10, WRITE10 */
		if (nblocks > 65535) {
			printf("Number-of-blocks too big, limit is %d\n",65535);
			free(Cmnd);
			return NULL;
		}
		opcode |= 0x20;				/* READ10, WRITE10 */
		Cmnd->cmnd[7] = nblocks >> 8;
		Cmnd->cmnd[8] = nblocks & 0xff;
		break;
	case 12:/* explicit READ12, WRITE12 */ 
		opcode |= 0xA0;				/* READ12, WRITE12 */
		*(unsigned int *)(&Cmnd->cmnd[6]) = htonl(nblocks);
		break;
	case 16:/* explicit READ16, WRITE16 */ 
		opcode |= 0x80;				/* READ16, WRITE16 */
		*(unsigned int *)(&Cmnd->cmnd[10]) = htonl(nblocks);
		break;
	default:/* should never happen */
		printf("Illegal size %d\n", size);
		return NULL;
		break;
	}	/* switch */
	Cmnd->cmd_len = Cmnd->old_cmd_len = size;
	Cmnd->cmnd[0] = opcode;
	nbytes = nblocks * BYTES_PER_BLOCK;
	if ((nbytes / BYTES_PER_BLOCK) != nblocks) {
		printf("Number-of-blocks %d too big for 32-bit cpu\n", nblocks);
		free(Cmnd);
		return NULL;
	}
	Cmnd->request_bufflen = nbytes;		/* actual request size */
	if ((bptr = Cmnd->request_buffer = calloc(1, nbytes)) == NULL) {
		printf("No space for \"%s\" buffer size %d\n",
			argv[0], nbytes);
		free(Cmnd);
		return NULL;	/* fatal */
	}
	if ((opcode & 0x0f) == WRITE_6) {
		/* this is a write of some type, fill in blocks with "data" */
		fill_value = 0xff;
		for (blockno = 0; blockno < nblocks; blockno++) {
			memset(bptr, fill_value, BYTES_PER_BLOCK);
			bptr += BYTES_PER_BLOCK;
			fill_value--;
		}
		/* change default direction of data flow */
		Cmnd->sc_data_direction = SCSI_DATA_WRITE;
	}
	return Cmnd;
}


void do_help(void)
{
	printf("Possibilites for \"do xxx\" are:\n");
	printf("\tdo testunitready\n");
	printf("\tdo inquiry\n");
	printf("\tdo readcapacity\n");
	printf("\tdo readcapacityshort\n");
	printf("\tdo readcapacitylong\n");
	printf("\tdo reportluns\n");
	printf("\tdo modesense\n");
	printf("\tdo requestsense\n");
	printf("\tdo read <number-of-blocks>\n");
	printf("\tdo read6 <number-of-blocks>\n");
	printf("\tdo read10 <number-of-blocks>\n");
	printf("\tdo read12 <number-of-blocks>\n");
	printf("\tdo read16 <number-of-blocks>\n");
	printf("\tdo write <number-of-blocks>\n");
	printf("\tdo write6 <number-of-blocks>\n");
	printf("\tdo write10 <number-of-blocks>\n");
	printf("\tdo write12 <number-of-blocks>\n");
	printf("\tdo write16 <number-of-blocks>\n");
}

/**
 * Builds a new SCSI command from the given argument vector.
 * \param argc The number of elements in the argv array.
 * \param argv The argument vector.
 * \param Cmnd The command buffer (used to return a pointer to the new
 *             command)
 * \return 1 if the function had a fatal error, 0 if not.
 */
static int new_scsi_command(int argc, char *argv[], struct scsi_cmnd **Cmnd)
{
	if (strcasecmp(argv[0], "HELP") == 0) {
		do_help();
		return 0;
	} else if (strcasecmp(argv[0], "TESTUNITREADY") == 0) {
		if (argc > 1) {
			printf("Extra token(s) after \"%s\"\n", argv[0]);
			return 0;
		}
		(*Cmnd) = set_up_cmnd(argc,argv);
		if ((*Cmnd) == NULL) {
			return 1;	/* fatal */
		}
		/* this is the easiest possible SCSI command to build --
		 * almost everything in the CDB and the struct scsi_cmnd is 0
		 */
	} else if (strcasecmp(argv[0], "INQUIRY") == 0) {
		if (argc > 1) {
			printf("Extra token(s) after \"%s\"\n", argv[0]);
			return 0;
		}
		(*Cmnd) = set_up_cmnd(argc,argv);
		if ((*Cmnd) == NULL) {
			return 1;	/* fatal */
		}
		(*Cmnd)->cmnd[0] = INQUIRY;		/* INQUIRY */
		(*Cmnd)->cmnd[4] = 96;		/* allocation length */
		(*Cmnd)->request_bufflen = 96;	/* actual request size */
		if (((*Cmnd)->request_buffer = calloc(1, 96)) == NULL) {
			printf("No space for \"%s\" data size %d\n",
				argv[0], 96);
			free((*Cmnd));
			return 1;	/* fatal */
		}
	} else if (strcasecmp(argv[0], "REPORTLUNS") == 0) {
		int size;

		if (argc > 1) {
			printf("Extra token(s) after \"%s\"\n", argv[0]);
			return 0;
		}
		(*Cmnd) = set_up_cmnd(argc,argv);
		if ((*Cmnd) == NULL) {
			return 1;	/* fatal */
		}
		(*Cmnd)->cmd_len = (*Cmnd)->old_cmd_len = 12;
		(*Cmnd)->cmnd[0] = REPORT_LUNS;		/* REPORTLUNS */
		/* allocation length field in CDB is 4 bytes
		 * we cheat a bit by only allowing for a max of 10 luns,
		 * each lun requiring 8 bytes in the reply
		 * there is also a leading 8 bytes of list information
		 * therefore, this total fits in the low-order byte
		 * of the 4-byte allocation length field
		 * we could handle up to 30 luns this way,
		 * since 8 + 8*30 = 248 < 256 fits in one byte
		 */
		size = 8 + 8*10;
		(*Cmnd)->cmnd[9] = size;		/* allocation length */
		(*Cmnd)->request_bufflen = size;	/* actual request size */
		if (((*Cmnd)->request_buffer = calloc(1, size)) == NULL) {
			printf("No space for \"%s\" data size %d\n",
				argv[0], size);
			free((*Cmnd));
			return 1;	/* fatal */
		}
	} else if (strcasecmp(argv[0], "READCAPACITY") == 0 ||
				strcasecmp(argv[0], "READCAPACITYSHORT") == 0) {
		(*Cmnd) = set_up_read_capacity(argc, argv, 0);
	} else if (strcasecmp(argv[0], "READCAPACITYLONG") == 0) {
		(*Cmnd) = set_up_read_capacity(argc, argv, 2);
	} else if (strcasecmp(argv[0], "REQUESTSENSE") == 0) {
		(*Cmnd) = set_up_request_sense(argc, argv);
	} else if (strcasecmp(argv[0], "MODESENSE") == 0) {
		(*Cmnd) = set_up_mode_sense(argc, argv);
	} else if (strcasecmp(argv[0], "READ") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, READ_6, 0);
	} else if (strcasecmp(argv[0], "READ6") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, READ_6, 6);
	} else if (strcasecmp(argv[0], "READ10") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, READ_6, 10);
	} else if (strcasecmp(argv[0], "READ12") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, READ_6, 12);
	} else if (strcasecmp(argv[0], "READ16") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, READ_6, 16);
	} else if (strcasecmp(argv[0], "WRITE") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, WRITE_6, 0);
	} else if (strcasecmp(argv[0], "WRITE6") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, WRITE_6, 6);
	} else if (strcasecmp(argv[0], "WRITE10") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, WRITE_6, 10);
	} else if (strcasecmp(argv[0], "WRITE12") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, WRITE_6, 12);
	} else if (strcasecmp(argv[0], "WRITE16") == 0) {
		(*Cmnd) = set_up_read_write(argc, argv, WRITE_6, 16);
	} else {
		printf("unknown SCSI command \"%s\"\n", argv[0]);
		do_help();
		return 0;
	}

	return 0;
}

/* __do_scsi_command
 * uses "argc" tokens in "argv" to
 * create a SCSI CDB and accompanying info and hands it off to
 * iscsi_initiator_queuecommand() for execution by the user-space
 * initiator.
 * returns 1 to exit, 0 otherwise (error or ok)
 */
int __do_scsi_command( int argc, char *argv[], int rep )
{
	struct scsi_cmnd	*Cmnd;
	int err, ret, i;
	int trace_mask;

	ret = 0;

	TRACE_GET(trace_mask);
	/*
	TRACE_SET(TRACE_DEBUG);
	*/
	TRACE_SET(trace_mask);

	if (argc <= 0) {
		printf("Missing command name after \"do\"\n");
		do_help();
		ret = 0;
		goto out;
	}

	completed_commands = 0;
	TRACE(TRACE_DEBUG, "calling queuecommand for \"%s\" "
	      "with Cmnd %p\n", argv[0], Cmnd);

	pthread_mutex_lock(&queue_more_mutex);
	for (i = 0; i < rep; i += 1) {
		err = new_scsi_command(argc, argv, &Cmnd);
		if (Cmnd == NULL) {
			ret = err;
			goto out;
		}
		if (global_host->hostt->queuecommand(Cmnd, done) != 0) {
			ret = 1;
			goto out;
		}

		while (i - completed_commands > OUTSTANDING_CMNDS)
			pthread_cond_wait(&queue_more_cond,
					  &queue_more_mutex);


		TRACE(TRACE_DEBUG, "queuing cmnd %d, %d completed, "
		      "%d outstanding\n",
		      i, completed_commands,
		      i - completed_commands);

	}
	pthread_mutex_unlock(&queue_more_mutex);

	/* wait for the remaining commands to complete. */
	pthread_mutex_lock(&queue_more_mutex);
	while (completed_commands < rep)
		pthread_cond_wait(&queue_more_cond, &queue_more_mutex);
	pthread_mutex_unlock(&queue_more_mutex);
 out:
 	TRACE_SET(trace_mask);
	return ret;
}

/* __do_repeat
 * Repeat a command multiple times.
 *
 * repind specifies the index into the repv vector of repeat counts
 * that is being used for this repeat.
 *
 * repv specifies a vector of repeat counts.
 *
 * This function calls itself recursively with incrementing repind
 * values until the final repeat value is reached.
 */
int __do_repeat(int argc, char *argv[], int repc, int repv[], int repind)
{
	int i, ret;

	BUG_ON(repind == repc - 1);

	TRACE(TRACE_DEBUG, "%s: repind=%d, repv[%d]=%d\n",
	      __func__, repind, repind, repv[repind]);

	for(i = 0; i < repv[repind]; i += 1) {
		if (repind == repc - 2)
			ret = __do_scsi_command(argc, argv, repv[repc - 1]);
		else
			ret = __do_repeat(argc, argv, repc,
					  repv, repind + 1);

		if (ret == 1)
			return 1;
	}

	return 0;
}

/* do_scsi_command
 *
 * Possibly repeat a scsi command.
 *
 * uses "argc" tokens in "argv" to
 * create a SCSI CDB and accompanying info and hands it off to
 * iscsi_initiator_queuecommand() for execution by the user-space
 * initiator.
 *
 * repc specifies the number of elements in the repeat count vector
 * for nested repeat commands.
 *
 * recv is a vector of integers.  Each element of this vector
 * specifies a repeat count.  Statistics are gathered for the outter
 * most repeat.
 *
 * returns 1 to exit, 0 otherwise (error or ok)
 */
int do_scsi_command( int argc, char *argv[], int repc, int repv[] )
{

	BUG_ON(repc < 0);

	if (repc == 0)
		return __do_scsi_command(argc, argv, 1);
	else if(repc == 1)
		return __do_scsi_command(argc, argv, repv[0]);
	else
		return __do_repeat(argc, argv, repc, repv, 0);

}
