/*	initiator/initiator_tx.c
 * 
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the functions for iscsi initiator code that are
 *	responsible for building and sending PDUS to a target.
 *
 *	Copyright (C) 2001-2006 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
*/
#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "../common/unh_threads.h"


extern struct iscsi_hostdata *global_hostdata;
extern int get_immed_length(struct session *current_session,
			    struct connection *current_connection,
			    struct scsi_cmnd *Cmnd);

/*
 * executed by rx thread or tx thread or command process.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called by drive_text_negotiate() and tx_nopout().
 * add padding bytes, if necessary, to buffer which already has size bytes
 * update the command's iov and msghdr to point at this buffer
 * then compute and add data digest, if necessary
 *
 * Draft 20, Section 10.1 iSCSI PDU Length and Padding
 * "iSCSI PDUs are padded to the closest integer number of
 * four byte words.  The padding bytes SHOULD be sent as 0."
 */
static void
add_padding(struct connection *current_connection,
	    struct command *new_command, char *buffer, int size)
{
	int n;
	struct iovec *iov;

	n = (-size) & 3;
	if (unlikely(n != 0)) {
		TRACE(TRACE_ISCSI_FULL, "Attach %u pad bytes\n", n);
		memset(buffer + size, 0, n);
		size += n;
	}

	/* get ptr to i/o vector which is already set up for header */
	iov = new_command->tx_iov + new_command->tx_iovlen;
	iov->iov_base = buffer;
	iov->iov_len = size;
	new_command->tx_size += size;
	new_command->tx_iovlen++;

	/* compute the final data digest if we are using them */
	if (current_connection->connection_flags & USE_DATADIGEST) {
		new_command->data_checksum = 0;
		do_crc(iov->iov_base, iov->iov_len,
		       &new_command->data_checksum);
		iov++;
		iov->iov_base = &new_command->data_checksum;
		iov->iov_len = CRC_LEN;
		new_command->tx_size += CRC_LEN;
		new_command->tx_iovlen++;
		TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
	}
}

/*
 * executed by rx thread or tx thread or command process.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called to send a NopOut to target as:
 * 	a new nop (nopin_cmd == NULL, size == 0 )
 * or
 * 	a new ping (nopin_cmd == NULL, size != 0 )
 * or
 * 	a response to a NopIn (nopin_cmd != NULL).
 * Buffer points to data to send, and size is number of bytes
 * of data to send.
 * If size > 0 then better have buffer != NULL.
 * Returns 1 if all ok, else 0.
 */
int
tx_nopout(struct connection *current_connection, __u32 i_bit,
	  struct iscsi_targ_nopin *nopin_cmd, char **buffer, int size)
{
	int need_itt, retval = 1;
	struct command *new_command;
	struct iscsi_init_nopout *nopout_cmd;

	TRACE(TRACE_ENTER_LEAVE, "Enter tx_nopout, size = %d\n", size);

	new_command = setup_command(NULL, current_connection);
	if (unlikely(new_command == NULL)) {
		/* will not be able to send the nop */
		retval = 0;
		goto out;
	}

	/* set up the NopOut header (Draft 20, Section 10.18)
	 * to be sent to the target
	 */
	nopout_cmd = (struct iscsi_init_nopout *) &new_command->iscsi_cmd;
	nopout_cmd->opcode = ISCSI_INIT_NOP_OUT | i_bit;
	nopout_cmd->flags = F_BIT;	/* always set on NopOut */

	if (size > 0) {
		/* we have data to send to target */
		nopout_cmd->length = htonl(size);

		/* now add padding bytes and data digest, if necessary */
		add_padding(current_connection, new_command, *buffer,
			    size);

		/* mark that buffer will be freed in tx thread,
		 * after NopOut is sent
		 */
		new_command->buffer_to_free = *buffer;
		*buffer = NULL;
	}

	/* RFC 3720 Section 10.18.2 Target Transfer Tag (for NopOut)
	 * "The NOP-Out MUST only have the Target Transfer Tag set if it is
	 * issued in response to a NOP-In with a valid Target Transfer Tag.
	 * In this case, it copies the Target Transfer Tag from the NOP-In PDU.
	 * Otherwise, the Target Transfer Tag MUST be set to 0xffffffff.
	 *
	 * When the Target Transfer Tag is set to a value other than 0xffffffff,
	 * the LUN field MUST also be copied from the NOP-In."
	 */
	if (nopin_cmd != NULL) {
		/* this is a response to a NopIn from the target */
		nopout_cmd->target_xfer_tag = nopin_cmd->target_xfer_tag;
		nopout_cmd->lun = nopin_cmd->lun;
		need_itt = 0;	/* need to set ITT to 0xffffffff */
	} else {
		/* this is a new nop or nop ping command to the target */
		nopout_cmd->target_xfer_tag = ALL_ONES;

		/*  RFC 3720 Section 10.18.1 Initiator Task Tag (for NopOut)
		 * "The NOP-Out MUST have the Initiator Task Tag set to
		 * a valid value only if a response in the form of
		 * NOP-In is requested (i.e., the NOP-Out is used as a
		 * ping request).  Otherwise, the Initiator Task
		 * Tag MUST be set to 0xffffffff."
		 */
		if (size == 0) {
			need_itt = 0;
		} else {
			need_itt = 1;	/* set ITT to next in sequence */
		}
	}

	/* attach the new NopOut command to end of pending commands list.
	 * This also sets the ITT, the I bit if needed, and the CmdSN.
	 */
	attach_pending_command(new_command, current_connection, need_itt);
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave tx_nopout\n");
	return retval;
}

static void deal_with_tx_timer(unsigned long data);

/*
 * (re)start the tx timer -- period is always 1 second
 */
static void __attribute__ ((no_instrument_function))
    restart_tx_timer(struct connection *conn)
{
	if (likely(!timer_pending(&conn->tx_timer)
		   && !(conn->connection_flags & TX_TIMER_OFF))) {
		TRACE(TRACE_TIMERS,
		      "Start timer for %u ticks for conn %u\n", HZ,
		      conn->connection_id);
		conn->tx_timer.expires = jiffies + HZ;
		conn->tx_timer.data = (unsigned long) conn;
		conn->tx_timer.function = deal_with_tx_timer;
		add_timer(&conn->tx_timer);
	} else {
		/* no timer or timer already going or timer should be
		 * stopped, don't (re)start it now
		 */
		TRACE(TRACE_TIMERS, "Timer not restarted for conn %u\n",
		      conn->connection_id);
	}
}

/*
 * Called by kernel when periodic session timer expires
 */
static void deal_with_tx_timer(unsigned long data)
{
	struct connection *conn;

	conn = (struct connection *) data;

	/* get the tx thread started, if necessary */
	/* if( sem_getcount(&conn->tx_sem) <= 0 ) */
	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_TIMERS,
			"tx_timer: wake up tx thread on tx_sem %p\n",
			&conn->tx_sem);
		up(&conn->tx_sem);
	}

	/* mark that the timer went off and keep it going */
	atomic_inc(&conn->tx_timer_went_off);
	restart_tx_timer(conn);
}

/*
 * executed by tx thread or command process.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called by do_test(), iscsi_initiator_tx_thread(), tx_retransmit_state2().
 * drives sending of a nop out to the target
 * returns 0 on success, -1 on error
 */
int
drive_nopout(struct connection *conn, struct session *current_session,
	     __u32 i_bit, __u32 want_data)
{
	char *buffer;
	int size;

	TRACE(TRACE_ENTER_LEAVE, "Enter drive_nopout\n");

	if (!want_data) {
		/* don't want any data attached to this nop out */
		buffer = NULL;
		if (i_bit)
			size = 0;	/* send it immediate */
		else
			size = -1;	/* send it normal (forces empty ping) */
	} else if ((buffer = my_kmalloc(32, "NopOut data")) == NULL) {
		size = 0;
	} else {
		/* send the '\0' at the end of the string too */
		size = 1 + sprintf(buffer, "NopOut_Ping_Number=%u",
				   conn->nop_counter++);

		TRACE(TRACE_DEBUG, "Send \"%s\", size %d\n", buffer, size);
	}

	tx_nopout(conn, i_bit, NULL, &buffer, size);

	TRACE(TRACE_ENTER_LEAVE, "Leave drive_nopout\n");

	return 0;
}

/*
 * executed by tx thread or command process.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called by do_test(), create_session().
 * drives a text negotiation exchange on this "conn" in this "current_session"
 * returns 0 on success, 1 on failure
 */
int
drive_text_negotiate(struct connection *conn,
		     struct session *current_session, __u32 i_bit,
		     enum text_nego_type nego_type)
{
	struct command *new_command;
	struct iscsi_init_text_cmnd *text_cmd;
	int retval, size;
	char *buffer, *val_ptr = NULL, local_buf[16];
	struct parameter_type *param = NULL;
	struct iscsi_targetdata *targ_data;

	TRACE(TRACE_ENTER_LEAVE, "Enter drive_text_negotiate type %d\n",
	      nego_type);

	new_command = setup_command(NULL, conn);
	if (unlikely(new_command == NULL)) {
		/* will not be able to send the text request */
		retval = 1;
		goto out;
	}

	/* set up the Text Request header (Draft 20, Section 10.10)
	 * to be sent to the target
	 */
	text_cmd = (struct iscsi_init_text_cmnd *) &new_command->iscsi_cmd;
	text_cmd->opcode = ISCSI_INIT_TEXT_CMND | i_bit;
	text_cmd->flags = F_BIT;	/* we have only 1 command */
	text_cmd->target_xfer_tag = ALL_ONES;	/* reset the target */
	new_command->new_max_recv_length = 0;	/* didn't negotiate this yet */
	new_command->new_max_send_length = 0;	/* nor this */
	/* LUN is 0 by default */

	targ_data = &global_hostdata->target_data
	    [current_session->scsi_target_id];
	switch (nego_type) {

	case DISCO_TEXT_NEGO:
		/* send a SendTargets key to get back discovery info */
		param = find_flag_parameter(SENDTARGETS_FLAG,
					    targ_data->param_tbl);
		if (param) {
			val_ptr = param->str_value;
		}
		break;

	case TARGET_TEXT_NEGO:
		/* send an empty text request to allow target to start nego */
		break;

	case INI_TEXT_NEGO:
		/* send new MaxRecvDataSegmentLength from initiator to target */
		param = find_flag_parameter(MAXRECVDATASEGMENTLENGTH_FLAG,
					    targ_data->param_tbl);
		if (param) {
			/* remember this new value so if text negotiation
			 * is successful, we will change our operational value
			 * on the final text response from target
			 */
			new_command->new_max_recv_length =
			    param->int_value;
			sprintf(local_buf, "%u", param->int_value);
			val_ptr = local_buf;
		}
		break;

	}			/* switch */

	if (param) {
		/* send an '=' and '\0' at the end of the string too */
		size = 2 + strlen(param->parameter_name);
		if (val_ptr)
			size += strlen(val_ptr);

		/* must allocate an extra 3 bytes in case of padding */
		buffer = my_kmalloc(size + 3, "Text data");
		if (unlikely(buffer == NULL)) {
			/* no space for the text request,
			 * free up the command structure
			 */
			endup_command(new_command, current_session);
			retval = 1;
			goto out;
		}

		/* save ptr to this buffer so we can free it in
		 * rx_text_rsp when done
		 */
		new_command->buffer_to_free = buffer;

		strcpy(buffer, param->parameter_name);
		strcat(buffer, "=");
		if (val_ptr)
			strcat(buffer, val_ptr);

		TRACE(TRACE_ISCSI, "Attach key: %s\n", buffer);

		text_cmd->length = htonl(size);

		/* now add padding bytes and data digest, if needed */
		add_padding(conn, new_command, buffer, size);
	}
	/* attach the new Text Request command to end of pending commands list.
	 * this also sets the ITT and the CmdSN
	 */
	attach_pending_command(new_command, conn, 1);

	/* return success */
	retval = 0;

      out:
	TRACE(TRACE_ENTER_LEAVE,
	      "Leave drive_text_negotiate, retval = %d\n", retval);
	return retval;
}

/*
 * executed by tx thread or rx thread or init_recovery thread or command process
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called by do_test(), iscsi_initiator_tx_thread(), init_recovery_thread(),
 *	rx_async_msg(), rx_text_rsp()
 * drives the logout of a session
 * returns 1 on success, 0 on failure
 * Modified to include connection logout - SAI
 * for legal "reason" values, see Draft 20, Section 10.14.1 Reason Code
 */
int
drive_logout(struct session *current_session, struct connection *conn,
	     int reason)
{
	struct command *new_command;
	struct iscsi_init_logout_cmnd *logout_cmd;
	int retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter drive_logout\n");

	new_command = setup_command(NULL, conn);
	if (unlikely(new_command == NULL)) {
		/* will not be able to send the logout request */
		retval = 0;
		goto out;
	}

	/* set up the Logout Request header (Draft 20, Section 10.14)
	 * to send to the target
	 */
	logout_cmd =
	    (struct iscsi_init_logout_cmnd *) &new_command->iscsi_cmd;
	logout_cmd->opcode = ISCSI_INIT_LOGOUT_CMND | I_BIT;
	/* use the reason code, always with F bit set to 1 - SAI */
	logout_cmd->flags = F_BIT | reason;

	/* RFC 3720 Section 10.14.3 CID
	 * "This is the connection ID of the connection to be closed.
	 * This field is only valid if the reason code is not 'close the
	 * session'."
	 */
	if (reason != LOGOUT_CLOSE_SESSION)
		logout_cmd->cid = cpu_to_be16(conn->connection_id);

	/* attach the new Logout Request command to end of pending commands list
	 * this also sets the ITT and the CmdSN
	 */
	attach_pending_command(new_command, conn, 1);

	/* return success */
	retval = 1;

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave drive_logout, retval = %d\n",
	      retval);
	return retval;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by check_for_retransmit().
 * Enter the xmit state of a command into state 1
 * by rebuilding the cmd and setting up a resend
 */
static int
tx_retransmit_state1(struct session *current_session,
		     struct connection *current_connection,
		     struct command *current_command)
{
	struct r2t_cookie *cookie;

	/* never got anything back from target, rexmit */
	if (current_command->cmd_state & CMD_STATE_REXMIT_ORIG) {
		/* data out iovec overwrote original command iovec */
		if (!list_empty(&current_command->r2t_cookies)) {
			cookie =
			    list_entry(current_command->r2t_cookies.next,
				       struct r2t_cookie, link);
			free_r2t_cookie(current_connection,
					current_command, cookie);
		}
		build_write_command(current_session, current_connection,
				    current_command);
	}
	TRACE(TRACE_ERROR_RECOVERY, "Re-Transmitting PDU, "
	      "ITT %u, size %u\n", current_command->init_task_tag,
	      current_command->retransmit_tx_size);

	current_command->tx_sent_so_far = 0;
	/* start again */
	current_command->tx_size = current_command->retransmit_tx_size;
	current_command->tx_wait_to_send = jiffies;
	current_command->time_stamp = jiffies;

	/* RDR - don't come back here a second time */
	current_command->activity_flg = 1;
	return 1;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by check_for_retransmit().
 * Enter the xmit state of a command into state 2
 * by marking the time and sending another nop
 */
static int
tx_retransmit_state2(struct session *current_session,
		     struct connection *current_connection,
		     struct command *current_command)
{
	struct init_error_rec err_rec;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	current_command->activity_flg = 2;
	current_command->time_stamp = jiffies;
	if (current_session->oper_param.ErrorRecoveryLevel &&
	    current_command->SCpnt->sc_data_direction == SCSI_DATA_WRITE) {
		/* a WRITE, assume lost R2T
		 * RFC 3720 Section 10.16.6 BegRun
		 * "BegRun 0 when used in conjunction with
		 * RunLength 0 means resend all unacknowledged
		 * Data-In, R2T or Response PDUS."
		 */
		err_rec.curr_conn = current_connection;
		err_rec.related_cmd = current_command;
		printk("%s No response to WRITE\n", current->comm);
		init_send_snack(&err_rec, 0, 0, DATA_R2T_SNACK);
	} else {
		drive_nopout(current_connection, current_session, I_BIT,
			     0);
	}
	return 1;

}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by check_for_retransmit().
 * Enter the xmit state of a command into state 3
 * by issuing SNACK requests
 */
static int
tx_retransmit_state3(struct session *current_session,
		     struct connection *current_connection,
		     struct command *current_command)
{
	int DInOrder = current_session->oper_param.DataSequenceInOrder;
	int PInOrder = current_session->oper_param.DataPDUInOrder;
	struct scsi_cmnd *SCpnt = current_command->SCpnt;
	struct init_error_rec err_rec;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	current_command->activity_flg = 3;
	if (likely(current_command->SCpnt != NULL)) {
		do {
			err_rec.curr_conn = current_connection;
			err_rec.related_cmd = current_command;
			if (current_command->SCpnt->sc_data_direction
			    == SCSI_DATA_READ) {
				/* take care of out-of-order PDUs too --
				 * new 18_04 SAI
				 */
				if (likely(DInOrder && PInOrder)) {
					if (current_command->data_offset
					    != scsi_bufflen(SCpnt)) {
						printk
						    ("%s No response to READ "
						     "ITT %u "
						     "expecting DataSN %u\n",
						     current->comm,
						     current_command->
						     init_task_tag,
						     current_command->
						     data_in_sn);
						init_send_snack(&err_rec,
								current_command->
								data_in_sn,
								0,
								DATA_R2T_SNACK);
						break;
					}
				} else {
					if (check_range_list_complete
					    (&current_command->
					     r2t_range_list,
					     ntohl(current_command->
						   iscsi_cmd.cmd_sn),
					     current_command->
					     init_task_tag) > 0) {
						printk
						    ("%s No response to READ "
						     "ITT %u "
						     "expecting DataSN %u\n",
						     current->comm,
						     current_command->
						     init_task_tag,
						     current_command->
						     data_in_sn);
						init_send_snack(&err_rec,
								current_command->
								data_in_sn,
								0,
								DATA_R2T_SNACK);
						break;
					}
				}
			}
			init_send_snack(&err_rec,
					current_connection->exp_stat_sn, 1,
					STATUS_SNACK);
		} while (0);

		/* Timestamp command transmission time - SAI */
		current_command->time_stamp = jiffies;
	}
	return 1;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_tx_thread().
 * command and determines if the commands needs re-sent
 * during 3 states:
 * State 0 -	Entered after original cmd is sent but NO reply is recv.
 *		State 0 sets up the command again and re-xmits.
 * State 1/2 -	Entered after state 0. Is sets up a no-op ping to see
 *		if the target is alive. This state is executed twice if
 *		necessary.
 * State 3 -	State 3 uses SNACKS requests to fill -in lost or dropped
 *		data PDU's.
 *
 * 0:  * returns 1 if ok, 0 if nothing works and thread should terminate
 * Retransmit a loss/dropped command
 */
static int __attribute__ ((no_instrument_function))
    check_for_retransmit(struct session *current_session,
		     struct connection *current_connection,
		     struct command *current_command)
{
	int retval = 1;
	long delta_time;

	TRACE(TRACE_ENTER_LEAVE, "Enter check_for_retransmit ITT %u, "
	      "state 0x%08x, still_to_send %u\n",
	      current_command->init_task_tag,
	      current_command->cmd_state, current_command->still_to_send);

	if (current_session->oper_param.ErrorRecoveryLevel
	    && current_session->retran_period) {
		/* If timer expired and no activity, assume command was lost
		 * and re-transmit command - SAI
		 */
		delta_time = jiffies - (current_command->time_stamp
					+
					((current_session->retran_period) *
					 HZ));
		if (delta_time >= 0) {
			/* nothing received for this command during last
			 * retran_period seconds
			 */
			/* TRACE(TRACE_ERROR_RECOVERY, */
			printk
			    ("Timeout, ITT %u, time_stamp %lu, jiffies %lu, "
			     "delta %ld retran_period %u, HZ %u, "
			     "activity_flg %u\n",
			     current_command->init_task_tag,
			     current_command->time_stamp, jiffies,
			     delta_time, current_session->retran_period,
			     HZ, current_command->activity_flg);

			if (!current_command->activity_flg) {
				/* no reply from target, rexmit */
				tx_retransmit_state1(current_session,
						     current_connection,
						     current_command);

			} else if (current_command->activity_flg == 1) {
				/* first timeout, allow 1 more
				 * due to freezing
				 */
				tx_retransmit_state2(current_session,
						     current_connection,
						     current_command);

			} else if (current_command->activity_flg == 2) {
				/* for read commands, the data_in or
				 * the scsi response could be lost
				 * due to packet drops because of
				 * bad connection or during connection
				 * recovery. This function pro-actively
				 * fills the gap in data-in pdus and
				 * scsi responses. This uses the command
				 * re-transmit timer for checking
				 * and filling gaps - SAI
				 */
				retval =
				    tx_retransmit_state3(current_session,
							 current_connection,
							 current_command);
			} else {
				/*
				 * current_command->activity_flg == 3
				 * Retransmit, Nop and SNACK all failed
				 * nothing works, shut this tx thread down
				 */
				retval = 0;
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave check_for_retransmit retval %d\n",
	      retval);
	return retval;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called by ready_to_xmit() and iscsi_initiator_tx_thread().
 * returns 0 if command is immediate or its CmdSN is NOT beyond MaxCmdSN
 *		(i.e., it is ok to send it to target now)
 * returns 1 if command is non-immediate and its CmdSN is beyond MaxCmdSN
 */
static int
beyond_window(struct session *current_session,
	      struct connection *current_connection,
	      struct command *current_command)
{
	int retval = 0, delta;

	if (!(current_command->iscsi_cmd.opcode & I_BIT)) {
		/* non-immediate, is its sequence number
		 * beyond target's window limit?
		 */
		delta = current_session->max_cmd_sn
		    - ntohl(current_command->iscsi_cmd.cmd_sn);
		if (delta < 0) {
			/* yes, mark that tx thread needs to be awakened */
			current_connection->connection_flags |=
			    NEED_TX_WAKEUP;
			/* can't send this command now */
			TRACE(TRACE_ISCSI, "Delay sending ITT %u: CmdSN "
			      "%u beyond MaxCmdSN %u\n",
			      current_command->init_task_tag,
			      ntohl(current_command->iscsi_cmd.cmd_sn),
			      current_session->max_cmd_sn);
			retval = 1;
		}
	}
	return retval;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_tx_thread().
 * returns 0 if nothing to xmit for this command,
 * else size to xmit
 * returns -1 if this command was cut short
 */
static int
ready_to_xmit(struct session *current_session,
	      struct connection *current_connection,
	      struct command *current_command)
{
	int size;
	long delta_time;
	struct r2t_cookie *cookie;

	if (unlikely(current_connection->short_command != NULL)) {
		/* in the middle of a partial send */
		if (current_connection->short_command != current_command) {
			/* not a partial send of this command, leave now */
			size = 0;
			goto out;
		}
		delta_time = jiffies - current_command->tx_wait_to_send;
		if (delta_time < 0) {
			/* this command has not waited long enuf since
			 * last partial send
			 */
			size = -1;
			current_connection->connection_flags |=
			    NEED_TX_WAKEUP;
			goto out;
		}
		/* ready to resume after partial send,
		 * everything already set up
		 */
		size = current_command->tx_size;
	} else
	    if (unlikely(current_command->cmd_state & CMD_STATE_ABORTING))
	{
		/* current command is being aborted, do not xmit anything */
		size = 0;
	} else
	    if (unlikely
		(current_command->cmd_state & CMD_STATE_ABORT_FINISH)) {
		/* task management command timed out or was replied to */
		finish_task_mgt(current_connection, current_command);
		size = 0;
	} else if ((size = current_command->tx_size) == 0) {
		/* current command has already been sent */
		if (!list_empty(&current_command->r2t_cookies)) {
			/* this command has an unsatisfied R2T ready to go */
			cookie =
			    list_entry(current_command->r2t_cookies.next,
				       struct r2t_cookie, link);
			if (&cookie->order_link ==
			    current_connection->r2t_list.next) {
				/* this r2t is next to send on this connection */
				if ((int) cookie->r2t_xfer_length > 0) {
					/* still data left to send on this R2T */
					size =
					    setup_dataoutpdu
					    (current_connection,
					     current_command, cookie);
				} else {
					/* should never happen */
					TRACE_ERROR
					    ("Cookie r2t_xfer_length %d "
					     "not positive, ITT %u\n",
					     (int) cookie->r2t_xfer_length,
					     current_command->
					     init_task_tag);
					free_r2t_cookie(current_connection,
							current_command,
							cookie);
					goto out;
				}
			}
		}
	}

      out:
	return size;
}

//TODO -- in case of error return appropriate value <0 (-EINTR or otherwise)
// each causes diff behaviour in do_tx_send
static inline
    int handle_init_sendmsg(struct session *sess,
			    struct connection *conn,
			    struct command *current_command)
{
	int size, sent;
#ifdef __KERNEL__
	mm_segment_t oldfs;
#endif
#ifdef ISER
	__u32 opcode;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef ISER
	opcode = current_command->header_opcode & ISCSI_OPCODE;
#endif
	sent = size = current_command->tx_size;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_init_sendmsg \n");

	if (!current_command) {
		TRACE_ERROR("%s: passed a NULL current_command\n", __func__);
		return -1;
	}

#ifdef ISER
	if (sess->oper_param.RDMAExtensions) {
#ifdef RDMA_MODE
		struct iser_data_buf_t *buf;
		struct iser_send_pdu_t send_pdu;
		iser_status_t status;
		int error = 0;

		switch (opcode) {

		case ISCSI_INIT_SCSI_CMND:

			send_pdu.p_bhs = (union iser_pdu_bhs_t *)
			    &current_command->iscsi_cmd;
			if (current_command->header_flags & R_BIT) {

				buf = &send_pdu.data.command.buf_in;
				buf->p_buf =
				    scsi_buffer(current_command->SCpnt);

				if (scsi_sg_count(current_command->SCpnt) > 0) {
					buf->type = ISER_BUF_TYPE_SCATTERLIST;
					buf->size = scsi_sg_count(current_command->SCpnt);
				} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
					BUG(); /* This can't happen
						* >= 2.6.25 */
#else
					buf->type = ISER_BUF_TYPE_SINGLE;
					buf->size = scsi_bufflen(current_command->SCpnt);
#endif
				}
			} else if (current_command->header_flags & W_BIT) {
				int immed_len;

				if(sess->oper_param.ImmediateData) {
					immed_len = get_immed_length(sess, conn,
							     current_command->
							     SCpnt);
				} else {
					immed_len = 0;
				}

				send_pdu.data.command.immediate_sz =
				    immed_len;

				TRACE(TRACE_DEBUG, "%s: immediate_sz=%lu\n",
				      __func__, (unsigned long) immed_len);

				if (!sess->oper_param.InitialR2T) {
					send_pdu.data.command.unsolicited_sz =
					    sess->oper_param.FirstBurstLength
					    - immed_len;
				} else {
					send_pdu.data.command.unsolicited_sz = 0;
				}

				/*
				 * Becides unsolicited data (that is
				 * send with DataOut PDUs), we consider
				 * everything else as _sent_.  This is
				 * because: immediate data is sent along
				 * with the command, done-and-done...
				 * and solicited data is handled by RDMA
				 * and we need to trust that that is
				 * completed with out us being notified
				 * of it (we consider it done).
				 */
				current_command->still_to_send =
					send_pdu.data.command.unsolicited_sz;

				buf = &send_pdu.data.command.buf_out;

				if (!scsi_sg_count(current_command->SCpnt)) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
					BUG(); /* This can't happen
						* >= 2.6.25 */
#else
					buf->p_buf = scsi_buffer(current_command->SCpnt);
					buf->size = scsi_bufflen(current_command->SCpnt);
					buf->type = ISER_BUF_TYPE_SINGLE;
#endif
				} else {
					buf->p_buf = scsi_buffer(current_command->SCpnt);
					buf->size = scsi_sg_count(current_command->SCpnt);
					buf->type = ISER_BUF_TYPE_SCATTERLIST;
				}

			} else {
				buf = &send_pdu.data.command.buf_in;
				buf->p_buf = NULL;
				buf->type = ISER_BUF_TYPE_SINGLE;
				buf->size = 0;
			}
			break;
		case ISCSI_INIT_SCSI_DATA_OUT:
			send_pdu.p_bhs = (union iser_pdu_bhs_t *)
			    &current_command->data;
			buf = &send_pdu.data.tx.buf;

			buf->p_buf = current_command->tx_iov;
			buf->size = current_command->tx_iovlen;
			buf->type = ISER_BUF_TYPE_IOVEC_VIRT;
			break;

		case ISCSI_INIT_TASK_MGMT_CMND:
			/* TODO - Added for iSER - Minti 
			 * iSER doesn't handle task mgmt cmds yet
			 */
			send_pdu.p_bhs = (union iser_pdu_bhs_t *)
			    &current_command->iscsi_cmd;

			break;

		case ISCSI_INIT_SNACK:
			TRACE_ERROR
			    ("####Disable SNACKs in RDMA mode###\n");
			error = 1;
			break;

		case ISCSI_INIT_NOP_OUT:
		case ISCSI_INIT_LOGIN_CMND:
		case ISCSI_INIT_TEXT_CMND:
		case ISCSI_INIT_LOGOUT_CMND:
			send_pdu.p_bhs = (union iser_pdu_bhs_t *)
			    &current_command->iscsi_cmd;
			buf = &send_pdu.data.tx.buf;

			buf->p_buf = current_command->tx_iov;
			buf->size = current_command->tx_iovlen;
			buf->type = ISER_BUF_TYPE_IOVEC_VIRT;
			break;

		default:
			TRACE_ERROR
			    ("####Unknown opcode in handle_sendmsg in RDMA mode\n");
			error = 1;
			break;
		}

		if (!error) {
			status =
			    global_hostdata->iser_api.
			    send_control((iser_conn_handle_t) conn,
					 &send_pdu);

			if (status != ISER_SUCCESS) {
				TRACE_ERROR
				    ("iser_send_control failed %d\n",
				     status);
				/* Note that we never return 0 (incomplete) */
				sent = -1;
			}
#if 0
			TRACE(TRACE_ISER,
			      "handle_init_sendmsg: waiting for"
			      "send completion semaphore %p\n",
			      &conn->send_done_sem);
			if (down(&conn->send_done_sem) != 0) {
				sent = -1;
				goto out;
			}
#endif
		}
		//else                  
#else				/* ifdef RDMA_MODE */
#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
#endif
		sent = sock_sendmsg(conn->sock,
				    &current_command->tx_msghdr, size);
#ifdef __KERNEL__
		set_fs(oldfs);
#endif				/* ifdef __KERNEL__ */
#endif				/* ifdef RDMA_MODE */
	} else
#endif
	{
		TRACE(TRACE_NET, "%s sock_sendmsg size %d, flags 0x%08x\n",
		      current->comm, size,
		      current_command->tx_msghdr.msg_flags);
#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
		sent =
		    sock_sendmsg(conn->sock, &current_command->tx_msghdr,
				 size);
		set_fs(oldfs);
#else				/* ifdef __KERNEL__ */
		TRACE(TRACE_NET, "%s Call writev, fd %d, %d vectors\n",
		       current->comm, conn->sock->fd,
		       current_command->tx_iovlen);
		sent =
		    writev(conn->sock->fd, current_command->tx_iov,
			   current_command->tx_iovlen);
		TRACE(TRACE_NET, "%s Done writev, fd %d, result%d\n", current->comm,
		       conn->sock->fd, sent);
#endif				/* ifdef __KERNEL__ */
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_init_sendmsg \n");
	return sent;
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by do_tx_send().
 * Finalize a PDU header just before it is about to be transmitted.
 * This requires inserting the latest ExpStatSN on this connection
 * and then computing the header CRC if enabled.
 * The PDU header is always pointed to by the first slot in tx_iov
 * and the slot for the CRC is always at the end of that header.
 */
static void __attribute__ ((no_instrument_function))
    finalize_header(struct connection *conn, struct command *command)
{
	struct pdu_hdr *pdu;

	pdu = (struct pdu_hdr *) command->tx_iov->iov_base;

	/* SNACK request freezes ExpStatSN early */
	if ((pdu->opcode & ISCSI_OPCODE) != ISCSI_INIT_SNACK)
		pdu->exp_stat_sn = htonl(conn->exp_stat_sn);

	/* compute and add the header digest if we are using them */
	if (conn->connection_flags & USE_HEADERDIGEST) {
		pdu->header_digest = 0;
		do_crc((__u8 *) pdu, ISCSI_HDR_LEN, &pdu->header_digest);
		TRACE(TRACE_ISCSI_FULL, "Attach HeaderDigest\n");
	}

	command->tx_sent_so_far = 0;	/* nothing sent so far */
	command->retransmit_tx_size = command->tx_size;
	command->header_complete = 1;	/* don't come back here again */
}

/*
 * executed only by tx thread.
 * the current_session->sess_lock MUST be held by the calling process/thread.
 * Called only by iscsi_initiator_tx_thread().
 * We have a command that needs to be sent on this connection
 * and target window is open enough to receive it.
 * session lock is held on entry and exit,
 * but is released during call to sock_sendmsg().
 * returns : 1 on success,
 * 	     0 on incomplete send,
 * 	    -1 on fatal error
 */
static int
do_tx_send(struct session *current_session,
	   struct connection *current_connection,
	   struct command *current_command)
{
	int size, sent, retval = 1, iovlen;
	struct iovec *iov;
	__u32 opcode;
#if defined(CONFIG_ISCSI_DEBUG)
	char op_buf[16];
#endif
# if defined (K26) && !defined (RDMA_MODE)
	mm_segment_t oldfs;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!current_command->header_complete) {
		finalize_header(current_connection, current_command);
	}

	/* mark that this command was sent to target */
	current_command->cmd_state |= CMD_STATE_TXSTARTED;

	opcode = current_command->header_opcode & ISCSI_OPCODE;

	TRACE(TRACE_DEBUG, "%s Header opcode 0x%.2x flags 0x%.2x\n",
	      current->comm, current_command->header_opcode,
	      current_command->header_flags);

	iov = current_command->tx_iov;
	iovlen = current_command->tx_iovlen;
	size = current_command->tx_size;
	if ((sent = current_command->tx_sent_so_far) == 0) {
		/* first time sending this command, use original iovec */
		if (likely(opcode != ISCSI_INIT_SCSI_DATA_OUT)) {
			TRACE(TRACE_ISCSI, "%s Send %s ITT %u "
			      "CmndSn %u ExpStatSN %u length %d\n",
			      current->comm,
			      printable_iscsi_op(iov->iov_base, op_buf),
			      ntohl(current_command->iscsi_cmd.
				    init_task_tag),
			      ntohl(current_command->iscsi_cmd.cmd_sn),
			      ntohl(current_command->iscsi_cmd.
				    exp_stat_sn), size);
		} else {
			TRACE(TRACE_ISCSI, "%s Send %s ITT %u "
			      "CmndSn %u ExpStatSN %u DataSN %u length %u "
			      "offset %u\n", current->comm,
			      printable_iscsi_op(iov->iov_base, op_buf),
			      ntohl(current_command->data.init_task_tag),
			      ntohl(current_command->iscsi_cmd.cmd_sn),
			      ntohl(current_command->data.exp_stat_sn),
			      ntohl(current_command->data.data_sn), size,
			      ntohl(current_command->data.offset));
		}
		if (unlikely(TRACE_TEST(TRACE_ISCSI_FULL))) {
			print_iscsi_command(iov->iov_base);
		}
	} else {
		/* previously did partial send, may need to use copy of iovec */
		TRACE(TRACE_ISCSI, "%s More 0x%02x, ITT %u, %u bytes\n",
		      current->comm, opcode,
		      current_command->init_task_tag, size);
		while (iov->iov_len <= sent) {
			/* all of this iov already sent */
			sent -= iov->iov_len;
			iov++;
			iovlen--;
		}
		if (sent > 0) {
			/* copy the remaining iovecs so we can modify 1st one */
			memcpy(current_command->tx_iov_copy, iov,
			       iovlen * sizeof(struct iovec));
			/* update copied iov to account for previous sends */
			iov = current_command->tx_iov_copy;
			iov->iov_base += sent;
			iov->iov_len -= sent;
		}
	}

	/* set up the message header for sendmsg */
	memset(&current_command->tx_msghdr, 0, sizeof(struct msghdr));
	current_command->tx_msghdr.msg_iov = iov;
	current_command->tx_msghdr.msg_iovlen = iovlen;
	current_command->tx_msghdr.msg_flags =
	    current_command->cmd_msg_flags;

	/***** uncomment this to put tx thread into non-blocking mode *****
	current_command->tx_msghdr.msg_flags |= MSG_NOSIGNAL | MSG_DONTWAIT;
	*****/

	TRACE(TRACE_DEBUG, "--sending %d, ITT %u\n",
	      size, current_command->init_task_tag);

	/* Timestamp the command transmission time - SAI */
	current_command->time_stamp = jiffies;

	/* allow other threads to run while we send */
	UNH_UNLOCK(&current_session->sess_lock,
		   current_connection->tx_lock_flags);

# if defined (K26) && !defined (RDMA_MODE)
	if (use_sendpage(current_connection, current_command) &&
	    current_command->pindex > 0) {
		struct socket *sock = current_connection->sock;
		struct scatterlist *sg;
		int res = 0, total_len, total_sent = 0, first_offset;
		int left_in_header;

		int flags = MSG_MORE, padding = current_command->padding;

		total_len = size - padding;

		left_in_header = current_connection->basic_hdr_len
		    - current_command->tx_sent_so_far;

		if (left_in_header > 0) {
			current_command->tx_msghdr.msg_iovlen = 1;
			oldfs = get_fs();
			set_fs(get_ds());
			res = sock_sendmsg(current_connection->sock,
					   &current_command->tx_msghdr,
					   left_in_header);
			set_fs(oldfs);

			if (res != left_in_header) {
				sent = res;
				goto finish;
			}

			total_len -= res;
			total_sent += res;
			first_offset = 0;
		} else {
			first_offset = sent;
		}

		sg = current_command->sg;
		while (total_len > 0) {
			struct page *page = sg_page(sg);
			int i, len;
			int sg_len = sg->length;
			int nr =
			    (sg->offset + sg->length +
			     (PAGE_SIZE - 1)) >> PAGE_SHIFT;
			int offset = sg->offset + first_offset;

			sg_len -= first_offset;
			first_offset = 0;	/* rdr */
			i = offset >> PAGE_SHIFT;
			offset = offset & (PAGE_SIZE - 1);
			page += i;	/* rdr */
			/* rdr This loop seems to needlessly duplicate the code
			 * in sendpage.  Why?  Because this goes through the
			 * page array and calls sendpage with only one page at
			 * a time.  But sendpage is itself capable of going
			 * through a page array and sending multiple pages!
			 */
			for (; total_len > 0 && i < nr; i++) {
				if (sg_len + offset > PAGE_SIZE)
					len = PAGE_SIZE - offset;
				else
					len = sg_len;

				if (len > total_len) {
					len = total_len;
				}

				if (total_len == len)
					flags = 0;

				res =
				    sock->ops->sendpage(sock, page, offset,
							len, flags);

				if (len != res) {
					if (res > 0) {
						sent = total_sent + res;	/* rdr */
						goto finish;
					} else if (res == -EAGAIN) {
						sent = total_sent;
						goto eagain;
					} else {
						sent = res;
						goto finish;
					}
				}

				page++;
				offset = 0;
				total_len -= res;
				sg_len -= res;
				total_sent += res;
			}
			sg = sg_next(sg);
		}

		if (unlikely(padding > 0)) {
			__u32 pad_bytes = 0;
			struct iovec iov = { &pad_bytes, padding };
			struct msghdr msg;

			memset(&msg, 0, sizeof(msg));
			msg.msg_iov = &iov;
			msg.msg_iovlen = 1;

			oldfs = get_fs();
			set_fs(get_ds());
			res = sock_sendmsg(sock, &msg, padding);
			set_fs(oldfs);

			if (res != padding) {
				if (res > 0) {
					sent = total_sent + res;	/* rdr */
					current_command->padding -= res;	/*rdr */
					goto finish;
				} else if (res == -EAGAIN) {
					sent = total_sent;
					goto eagain;
				} else {
					sent = res;
					goto finish;
				}
			}
		}

		sent = size;
	} else {
		oldfs = get_fs();
		set_fs(get_ds());
		sent = sock_sendmsg(current_connection->sock,
				    &current_command->tx_msghdr, size);
		set_fs(oldfs);
	}
      finish:
#else				//Either 2.4 kernel or K26 in rdma mode
	/* Added for iSER - Minti */
	TRACE(TRACE_ISER, "Before handle_init_sendmsg \n");
	sent =
	    handle_init_sendmsg(current_session, current_connection,
				current_command);
#endif
	/* now get back exclusive access to this session iscsi structures */
	UNH_LOCK(&current_session->sess_lock,
		 current_connection->tx_lock_flags);

	TRACE(TRACE_DEBUG, "--sent %d, ITT %u resume\n",
	      sent, current_command->init_task_tag);

	if (unlikely(sent < size)) {
		/* did not send what we expected to send, why? */
		if (current_connection->tx_thread == NULL
		    || sent == -EINTR) {
			/* thread was killed as part of session shutdown */
			retval = -ESRCH;	/* No such process */
			goto out;
		}

		TRACE(TRACE_DEBUG,
		      "%s sock_sendmsg returned %d, expected %d\n",
		      current->comm, sent, size);

		if (sent < 0) {
			if (sent != -EAGAIN) {
				/* hard error -- report it, then get out */
				TRACE_ERROR("%s sock_sendmsg error %d\n",
					    current->comm, -sent);
				retval = sent;
				goto out;
			}
			sent = 0;	/* got nothing on this send */
		}
# if defined (K26) && !defined (RDMA_MODE)
	      eagain:
#endif
		/* need to try again later */
		current_command->tx_size -= sent;
		current_command->tx_sent_so_far += sent;
		current_command->tx_wait_to_send = jiffies + (HZ >> 6);
		current_connection->connection_flags |= NEED_TX_WAKEUP;

		/* couldn't send all of this command now */
		TRACE(TRACE_ISCSI, "%s Part 0x%02x, ITT %u, "
		      "%u bytes unsent\n",
		      current->comm, opcode,
		      current_command->init_task_tag,
		      current_command->tx_size);

		/* do nothing more on this command or connection */
		current_connection->short_command = current_command;
		retval = 0;
		goto out;
	}

	/* sent everything we wanted to send, this pdu is completely sent */
	current_connection->short_command = NULL;

	if (unlikely(TRACE_TEST(TRACE_DEBUG)
		     && opcode == ISCSI_INIT_SCSI_CMND)) {
		/* this was a SCSI Request */
		TRACE(TRACE_DEBUG, "%s Sent SCSI %s ITT %u "
		      "CmndSn %u exp_stat_sn %u length %d\n",
		      current->comm,
		      printable_scsi_op(current_command->iscsi_cmd.cdb[0],
					op_buf),
		      ntohl(current_command->iscsi_cmd.init_task_tag),
		      ntohl(current_command->iscsi_cmd.cmd_sn),
		      ntohl(current_command->iscsi_cmd.exp_stat_sn), size);
	}

	/* Timestamp the command transmission time - SAI */
	current_command->time_stamp = jiffies;

	/* Mark that current PDU has been completely sent.  This allows
	 * new transfers to be set up in this command's pdu slot.
	 */
	current_command->tx_sent_so_far += sent;
	current_command->tx_size = 0;

	if (!(current_command->header_flags & F_BIT)) {
		/* F bit is 0, more DataOut PDUs to be sent, keep loop going */
		if (atomic_read(&current_connection->tx_sem.count) <= 0) {
			TRACE(TRACE_SEM,
				"do_tx_send: wake up tx thread on tx_sem %p\n",
				&current_connection->tx_sem);
			up(&current_connection->tx_sem);
		}
	} else if (opcode == ISCSI_INIT_SCSI_DATA_OUT) {
		/* this was the last DataOut PDU in an R2T sequence */
		if (!list_empty(&current_connection->r2t_list)) {
			/* first DataOut PDU of next R2T ready to be sent */
			if (atomic_read(&current_connection->tx_sem.count)
			    <= 0) {
				TRACE(TRACE_SEM,
				"do_tx_send: wake up tx thread on tx_sem %p\n",
				&current_connection->tx_sem);
				up(&current_connection->tx_sem);
			}
		}
	} else if (opcode == ISCSI_INIT_NOP_OUT) {
		/* this was the last (only) NopOut PDU sent */
		if (current_command->init_task_tag == ALL_ONES) {
			/* this was a ping response to target's NopIn
			 * Free up the command itself since we will
			 * get no further response from target.
			 */
			current_command->cmd_state
			    |= CMD_STATE_TXDONE | CMD_STATE_RXDONE;
		}
	} else if (opcode == ISCSI_INIT_SNACK) {
		/* Free SNACK Request from pending commands list - SAI */
		current_command->cmd_state
		    |= CMD_STATE_TXDONE | CMD_STATE_RXDONE;
	}
#ifdef ISCSI_STATS
	UPDATE_TX_SESS_STATS(opcode, current_command,
			     current_session->sess_stats,
			     current_connection);
#endif

      out:
	return retval;
}

/*
 * executed only by tx thread.
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called only once by iscsi_initiator_tx_thread() to shutdown the tx thread,
 */
static void tx_shutdown_thread(void *ptr)
{
	struct connection *conn;
	struct session *sess;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "%s shutting down\n", current->comm);

	conn = ptr;
	sess = conn->my_session;

	/* get exclusive access to this session iscsi structures */
	UNH_LOCK(&sess->sess_lock, conn->tx_lock_flags);

	/* do not allow tx_timer to restart itself */
	conn->connection_flags |= TX_TIMER_OFF;

	/* indicate that this thread is shutting itself down */
	conn->tx_thread = NULL;

	/* tell anyone waiting that this tx thread is finished */
	TRACE(TRACE_SEM,
			"tx_shutdown: wake up on tx_done_sem %p\n",
			&conn->tx_done_sem);
	up(&conn->tx_done_sem);

	/* finally, release our exclusive access to the iscsi structures */
	UNH_UNLOCK(&sess->sess_lock, conn->tx_lock_flags);

	printk("%s Exiting pid %d conn %p\n", current->comm, current->pid,
	       conn);
}

/****************************************************************************
 * The tx thread is woken up anytime someone has something
 * that needs to be sent.  each time it is woken up it searches through the
 * command table for stuff that it needs to send.
 * There is one tx thread for each connection in a session.
 ******************************************************************************/
void iscsi_initiator_tx_thread(void *arg)
{
	struct connection *conn;
	struct session *sess;
	struct command *command;
	int err;
	long delta_time;
	__u32 nop_timer_expired;
	struct list_head *lptr, *next;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	conn = (struct connection *) arg;
	sess = conn->my_session;

	/* Lock kernel to daemonize thread and update tx thread pointer */
	lock_kernel();

	unh_daemonize("tx-%u-%02u", sess->scsi_target_id,
	              conn->connection_id);

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this tx thread is alive */
	conn->tx_thread = current;

	unlock_kernel();

	printk("%s Starting pid %d conn %p\n", current->comm, current->pid,
	       conn);

	/* set up counter so NopOut pings can be sent to target periodically */
	conn->nop_remaining = sess->nop_period;
	TRACE(TRACE_TIMERS, "%s nop_remaining %u\n", current->comm,
	      conn->nop_remaining);

	/* Always start the tx timer now */
	restart_tx_timer(conn);

	/* signal anyone waiting that the tx thread is up and running */
	TRACE(TRACE_SEM,
			"tx_thread: wake up on tx_done_sem %p\n",
			&conn->tx_done_sem);
	up(&conn->tx_done_sem);

	for (;;) {
#ifndef __KERNEL__
		pthread_cleanup_push(tx_shutdown_thread, conn);
#endif
		TRACE(TRACE_SEM,
		      "%s blocked on tx_sem %p\n",
		      current->comm, &conn->tx_sem);

		err = down_interruptible(&conn->tx_sem);

		TRACE(TRACE_SEM,
		      "%s unblocked on tx_sem %p, err %d\n",
		      current->comm, &conn->tx_sem, err);
#ifndef __KERNEL__
		pthread_cleanup_pop(0);
#endif
		if (unlikely(err < 0)) {
			/* did not get the sem but got signalled */
			TRACE(TRACE_ISCSI, "%s Killed by signal\n",
			      current->comm);
			goto leave;
		}

		/* now get exclusive access to the iscsi structures */
		UNH_LOCK(&sess->sess_lock, conn->tx_lock_flags);

		/* tx thread just awakened, so no further wakeups needed yet */
		conn->connection_flags &= ~NEED_TX_WAKEUP;

		nop_timer_expired = 0;
		if (unlikely(atomic_read(&conn->tx_timer_went_off))) {
			/* periodic tx timer went off */
			TRACE(TRACE_TIMERS, "%s tx_timer_went_off %u, "
			      "nop_remaining %u\n",
			      current->comm,
			      atomic_read(&conn->tx_timer_went_off),
			      conn->nop_remaining);

			atomic_dec(&conn->tx_timer_went_off);
			if (conn->nop_remaining) {
				/* sending periodic nop pings is enabled */
				conn->nop_remaining--;
				if (conn->nop_remaining == 0) {
					/* time to send a periodic nop
					 * on inactive connections
					 */
					nop_timer_expired = 1;
					conn->nop_remaining =
					    sess->nop_period;
				}
			}
		}

		/* tx thread IS dependent on rx thread */
		if (unlikely(conn->rx_thread == NULL)) {
			/* no rx thread, connection not up yet
			 * or was brought down unexpectedly.
			 * In either case, don't try to use it!
			 */
			printk("%s Exiting, no rx thread!\n",
			       current->comm);
			goto out;
		}

		/* See if this connection got logged out */
		if (unlikely
		    (conn->connection_state == CONNECTION_LOGGED_OUT)) {
			printk("%s Logged out\n", current->comm);
			goto out;
		}

		/* if timer went off and this connection has had no
		 * activity since the last timer then send a
		 * NopOut ping to the target
		 */
		if (unlikely(nop_timer_expired
			     && list_empty(&conn->pending_commands))) {
			/* timer went off and nothing pending
			 * on this connection
			 */
			if (!(conn->connection_flags & GOT_ACTIVITY)) {
				/* no activity since
				 * last timer timed out,
				 * send nop ping
				 */
				drive_nopout(conn, sess, 0, 1);
			} else {
				/* had activity, turn off
				 * the flag to start next period
				 */
				conn->connection_flags &= ~GOT_ACTIVITY;
			}
		}

		list_for_each_safe(lptr, next, &conn->pending_commands) {
			command = list_entry(lptr, struct command, link);

			/* bypass this command if target window closed to it */
			if (beyond_window(sess, conn, command))
				continue;

			if (command->need_write_built) {
				/* A write request.
				 * Need to construct the infrastructure 
				 */
				TRACE(TRACE_DEBUG,
				      "command %p marked for write\n",
				      command);
				build_write_command(sess, conn, command);
				command->need_write_built = 0;
			}

			if (unlikely(TRACE_TEST(TRACE_DEBUG))) {
				delta_time = jiffies - command->time_stamp;
				TRACE(TRACE_DEBUG,
				      "%s for command %p, tx_size %u, "
				      "ITT %u, time_stamp %lu, "
				      "delta %ld\n",
				      current->comm, command,
				      command->tx_size,
				      command->init_task_tag,
				      command->time_stamp, delta_time);
			}

			if (unlikely(conn->short_command != NULL)) {
				if (conn->short_command != command) {
					/* when we have a partial send
					 * outstanding, don't look at any other
					 * command until that is resumed
					 */
					break;
				}
			}
#ifndef RDMA_MODE
			if (unlikely(!conn->sock))
				goto out;
#else
			if (conn->connection_state ==
			    CONNECTION_TERMINATED)
				goto out;
#endif

			/* Check if command re-transmission required
			 * -- SAI
			 */
			if (unlikely(!command->tx_size
				     && !check_for_retransmit(sess, conn,
							      command))) {
				/* nothing works on rexmit, shut
				 * this tx thread down
				 */
				printk
				    ("%s Exiting, bad retransmit, ITT %u\n",
				     current->comm,
				     command->init_task_tag);
				goto out;
			}

			if ((err =
			     ready_to_xmit(sess, conn, command)) <= 0) {
				if (err < 0) {
					/* do nothing more on
					 * this command or connection
					 */
					break;
				}
				/* this command not ready to send */
				continue;
			}

			/* have a command that needs to be
			 * sent on this connection
			 * and target window is open
			 * enough to receive it
			 */
			err = do_tx_send(sess, conn, command);
			if (unlikely(err <= 0)) {
				if (err == 0) {
					/* do nothing more on
					 * this command or connection
					 */
					break;
				}
				/* fatal error */
				printk("%s Exiting due to error %d\n",
				       current->comm, -err);
				goto out;
			}

			/* in case this changed while we gave
			 * up the lock
			 */
			next = lptr->next;

			if (command->still_to_send == 0)
				command->cmd_state |= CMD_STATE_TXDONE;
			free_pending_command(command, conn);
		}		/* end of for command loop */

		/* Check if a Connection Recovery Test needs
		 * to be done - SAI
		 */
		if (unlikely(conn->rec_tests == 8)) {
			if (sess->nconnections <= 1) {
				UNH_UNLOCK(&sess->sess_lock,
					   conn->tx_lock_flags);
				create_rec_conn(sess->connection_head);
				UNH_LOCK(&sess->sess_lock,
					 conn->tx_lock_flags);
			}
			drive_logout(sess, conn,
				     sess->oper_param.ErrorRecoveryLevel);
			conn->rec_tests = 0;
		}

		/* release our exclusive access to the session
		 * iscsi structures and resume infinite loop
		 */
		UNH_UNLOCK(&sess->sess_lock, conn->tx_lock_flags);
	}			/* end of infinite loop */

	/* NOT REACHED */

      out:
	/* release our exclusive access to the iscsi session structures */
	UNH_UNLOCK(&sess->sess_lock, conn->tx_lock_flags);

      leave:
	tx_shutdown_thread(conn);
}
