/*	initiator/iscsi_initiator.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the core functions for iscsi initiator code that are
 *	responsible for connecting and maintaining sessions between initiator
 *	and target. The object file generated is the loadable kernel module.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
*/

#include "iscsi_initiator.h"
#include "initiator_utilities.h"
#include "initiator_error_rec.h"
#include "initiator_proc_iface.h"
#include "../common/lun_packing.h"
#include "version.h"

#ifdef __KERNEL__

#include <linux/reboot.h>

#include "../common/kernel_workqueue.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 15)
typedef struct scsi_host_template Scsi_Host_Template;
#endif

#else

/* always need this in user space */
typedef struct scsi_host_template Scsi_Host_Template;

#include "../common/user_space.h"

#endif

/* max number of redirects to follow during login on any connection */
#define	MAX_REDIRECTION_RETRIES	2

/* Holds pointer to private data part of Scsi_Host structure.
 * This is set up once, by iscsi_detect()
 * which is called when the initiator module is loaded
 */
struct iscsi_hostdata *global_hostdata = NULL;

/* Holds pointer to the registered Scsi_Host */
struct Scsi_Host *global_host = NULL;

/* Used for all access to the shared structures, and during the operation
 * of any single thread to prevent things from disappearing from under
 * the thread.
*/
UNH_LOCK_TYPE host_data_lock = UNH_LOCK_UNLOCKED;


/* dummy that always returns false */
int target_in_use(__u32 id)
{
	return 0;
}

static void close_connection(struct connection *conn);

#ifdef __KERNEL__

/* Mike Christie, mikenc@us.ibm.com */

static int iscsi_initiator_halt(struct notifier_block *nb, ulong event,
				void *buf);

/* Mike Christie, mikenc@us.ibm.com */
/* reboot notifier */
static struct notifier_block iscsi_initiator_notifier = {
	iscsi_initiator_halt, NULL, 0
};

#endif

/*
 * either the session->sess_lock MUST be held by the calling process/thread
 * or the session structure must not be active yet.
 * increase maximum number command structures session may use to:
 * 	(MAX_COMMANDS_PER_LUN + 1) * n_luns_in_session
 * (one extra in case of task-management and/or nopout commands)
 * Returns 1 on success, 0 on failure.
 */
static int __attribute__ ((no_instrument_function))
    preallocate_commands(struct session *sess)
{
	__u32 limit;
	struct command *new_command;

	limit = (MAX_COMMANDS_PER_LUN + 1) * sess->n_luns_in_session;
	while (sess->n_commands_alloced < limit) {
		new_command =
		    my_kmalloc(sizeof(struct command), "command");
		if (unlikely(new_command == NULL)) {
			return 0;
		}
		/* each new command goes at end of session's free list */
		endup_command(new_command, sess);
		sess->n_commands_alloced++;
	}
	TRACE(TRACE_DEBUG,
	      "%u commands now allocated to session %p target %u\n",
	      sess->n_commands_alloced, sess, sess->scsi_target_id);
	return 1;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * reduce maximum number command structures session may use to:
 * 	(MAX_COMMANDS_PER_LUN + 1) * n_luns_in_session
 */
static void __attribute__ ((no_instrument_function))
    unallocate_commands(struct session *sess)
{
	__u32 limit;
	struct command *old_command;
	struct list_head *lptr;

	limit = (MAX_COMMANDS_PER_LUN + 1) * sess->n_luns_in_session;
	while (sess->n_commands_alloced > limit
	       && !list_empty(&sess->free_commands)) {
		lptr = sess->free_commands.next;
		list_del(lptr);
		old_command = list_entry(lptr, struct command, link);
		my_kfree((void **) &old_command, "command");
		sess->n_commands_alloced--;
	}
	TRACE(TRACE_DEBUG,
	      "%u commands now allocated to session %p target %u\n",
	      sess->n_commands_alloced, sess, sess->scsi_target_id);
}

/* initialize all fields in struct iscsi_targetdata for one target */
void init_targetdata(struct iscsi_targetdata *targ_data)
{
	/* Initialize all fields to 0 or NULL */
	memset(targ_data, 0, sizeof(struct iscsi_targetdata));

	/* now go back and initialize those that are not 0 or NULL by default */
	targ_data->isid_type = DEFAULT_ISID_TYPE;
	targ_data->isid_number = DEFAULT_ISID_NUMBER;
	targ_data->isid_qualifier = DEFAULT_ISID_QUALIFIER;

	targ_data->auth_parameter.chap_local_ctx =
	    CHAP_InitializeContext();
	targ_data->auth_parameter.chap_peer_ctx = CHAP_InitializeContext();
	targ_data->auth_parameter.srp_ctx = SRP_InitializeContext();

	/* Initialize parameter list to a copy of default set of parameters */
	param_tbl_init(targ_data->param_tbl);
}

/* uninitialize fields in struct iscsi_targetdata that need freeing */
void uninit_targetdata(struct iscsi_targetdata *targ_data)
{
	param_tbl_uncpy(targ_data->param_tbl);

	SRP_FinalizeContext(targ_data->auth_parameter.srp_ctx);
	CHAP_FinalizeContext(targ_data->auth_parameter.chap_peer_ctx);
	CHAP_FinalizeContext(targ_data->auth_parameter.chap_local_ctx);
}

/***************************************************************************
* Called from iscsi_initiator_detect() and unh_iscsi_probe().
* Initializes the hostdata structure.
* Creates the parameter table for this hostdata.
* Returns non-NULL pointer to parameter table if ok, NULL if error.
***************************************************************************/
static int __attribute__ ((no_instrument_function))
    init_initiator(struct iscsi_hostdata *hostdata)
{
	int i;
	struct iscsi_targetdata *targ_data;
	int retval = 1;
#ifdef RDMA_MODE
	/* Added for iSER - Minti */
	struct iser_api_cb_t iser_api_cb;
	struct iser_ext_api_cb_t iser_ext_api_cb;
	iser_status_t status;
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter init_initiator, hostdata %p\n",
	      hostdata);

	/* Initialize all fields to 0 or NULL */
	memset(hostdata, 0, sizeof(struct iscsi_hostdata));

	/* now go back and initialize those that are not 0 or NULL by default */
	INIT_LIST_HEAD(&hostdata->session_list);

	for (i = 0, targ_data = hostdata->target_data;
	     i < MAX_TARGET_IDS; targ_data++, i++) {
		init_targetdata(targ_data);
	}

#ifdef RDMA_MODE
	/* if the DM is not loaded its still possible to operate in TCP mode.
	 * However the module won't load if it can't locate iser_api_register.
	 * Need to make it conditional during compile time - otherwise DM will have
	 * to be loaded whether it is used or not. -Minti
	 */
	iser_api_cb.control_notify = control_notify;
	iser_api_cb.conn_terminate_notify = control_terminate_notify;
	iser_ext_api_cb.connection_establish_notify =
	    connection_establish_notify;
//      iser_api_cb.send_notify = send_notify;

	/* These needed only at target */
	iser_api_cb.data_completion_notify = NULL;
	iser_api_cb.data_ack_notify = NULL;
	iser_ext_api_cb.connection_request_notify = NULL;

	status = iser_api_register_initiator("unh_iscsi_initiator",
					     &global_hostdata->iser_api,
					     &iser_api_cb,
					     &global_hostdata->
					     iser_ext_api,
					     &iser_ext_api_cb,
					     &global_hostdata->
					     iser_handle);
	if (status != ISER_SUCCESS) {
		TRACE_ERROR("iser_api_register failed %d\n", status);
		/* We could also NOT return an error - its still possible to operate in
		 * TCP streaming mode - notify the user that RDMA mode will never work
		 */
		retval = -1;
	}
#endif

	TRACE(TRACE_ENTER_LEAVE, "Leave init_initiator\n");
	return retval;
}

/***************************************************************************
* Called by the scsi subsystem when the iscsi_initiator module is loaded.
* (via the detect field in the Scsi_Host_Template driver_template)
* This allows the iscsi_initiator driver code to register itself with the
* SCSI Mid-level as being the driver for this type of host adapter.
* It registers space for iscsi_hostdata struct with the SCSI Mid-level.
* When called, host_lock is held by the midlevel caller.
* It returns:
* 	n > 0 if ok, where n is number of host adapters of this type
*		on this system (all will be driven by this code);
* 	n < 0 on error.
***************************************************************************/
static int iscsi_initiator_detect(Scsi_Host_Template * tmpt)
{
	int retval = 1;
	struct Scsi_Host *host = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_detect, tmpt %p\n",
	      tmpt);

	/* scsi_register() is defined in linux/drivers/scsi/hosts.c */
	host = scsi_register(tmpt, sizeof(struct iscsi_hostdata));

	TRACE(TRACE_DEBUG, "Back from scsi_register, host %p\n", host);

	if (unlikely(host == NULL)) {
		retval = -1;
		goto out;
	}

	TRACE(TRACE_ISCSI, "Registered UNH iSCSI host number %d\n",
	      host->host_no);

	/* setup the security key hash table */
	setup_security_hash_table();

	/* three standard default values which can be overridden:
	 * host->max_channel = 0;
	 * host->max_id = 8;
	 * host->max_lun = 8;
	 */

	/* set the number of allowed "targets" this adaptor "card" can handle.
	 *  the "target" numbers will be [0..max_id-1].
	 *   each "target" will have LUN numbers in [0..max_lun-1].
	 */
	host->max_id = MAX_TARGET_IDS;
	host->max_lun = MAX_LUNS;

	/* set the maximum number of bytes we will accept in a scsi cdb */
	host->max_cmd_len = MAX_CDB_LEN;

	/* scsi_register() allocated space for our private structure at the end
	 *  of the host structure. This global variable is cheating!
	 *  Potentially each (physical) adaptor card can have its own hostdata.
	 *  We get away with having a single global variable here because
	 *  we know we will never have more than 1 adaptor of this type ever.
	 */

	global_hostdata = (struct iscsi_hostdata *) host->hostdata;
	global_host = host;

	/* Initialize the adapter's private data structure */
	retval = init_initiator(global_hostdata);

	(void) initiator_register_device();

#ifdef __KERNEL__

	/* Mike Christie, mikenc@us.ibm.com */
	register_reboot_notifier(&iscsi_initiator_notifier);

#else

	/* rdr added this */
	global_host->device = calloc(1, sizeof(struct scsi_device));
	if (global_host->device == NULL) {
		printf("No space for scsi_device %lu bytes\n",
			(unsigned long) sizeof(struct scsi_device));
		return 1;
	}
	global_host->device->host = global_host;
	/* for now, assume id and lun are always 0 */

#endif

	TRACE(TRACE_DEBUG, "Register iSCSI host adapter \"%s\"\n",
	      tmpt->proc_name);

      out:
//      mod_pri_for_print();
	TRACE(TRACE_ENTER_LEAVE,
	      "Leave iscsi_initiator_detect, retval %d\n", retval);

	return retval;
}

/* called only when session->sched_scheme is CONN_SCHED_LUN.
 * the session->sess_lock MUST be held by the calling process/thread.
 * the host_data_lock MAY be held by the calling process/thread.
 * find the connection with the fewest number of LUNs assigned to it
 * and then to assign this lun to it
 */
static void __attribute__ ((no_instrument_function))
    locked_add_lun_to_sess(__u32 lun, struct session *session)
{
	__u32 min_count;
	struct connection *conn, *sofar;

	/* find the connection with the fewest luns assigned to it */
	min_count = MAX_LUNS;
	sofar = NULL;
	for (conn = session->connection_head;
	     conn != NULL; conn = conn->next) {
		if (conn->connection_state == CONNECTION_FULL_FEATURE_PHASE
		    && min_count > conn->assigned_lun_count) {
			min_count = conn->assigned_lun_count;
			sofar = conn;
		}
	}

	if (sofar) {
		session->lun_assignments[lun] = sofar;
		sofar->assigned_lun_count++;
		printk
		    ("Assign lun %u to connection %u, %u luns assigned\n",
		     lun, sofar->connection_id, sofar->assigned_lun_count);
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * find the connection with the fewest number of LUNs assigned to it
 * and then to assign this lun to it
 */
static void __attribute__ ((no_instrument_function))
    add_lun_to_sess(__u32 lun, struct session *session)
{
	unsigned long lock_flags;

	UNH_LOCK(&session->sess_lock, lock_flags);

	/* find the connection with the fewest luns assigned to it */
	locked_add_lun_to_sess(lun, session);

	UNH_UNLOCK(&session->sess_lock, lock_flags);
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * called only when this lun is assigned to a specific connection
 * remove this lun from this session's table of assigned connections.
*/
static void __attribute__ ((no_instrument_function))
    locked_drop_lun_from_sess(__u32 lun, struct session *sess,
			  int free_inquiry)
{
	struct connection *conn;

	/* this lun was assigned to this connection, unassign it now */
	conn = sess->lun_assignments[lun];
	sess->lun_assignments[lun] = NULL;
	if (free_inquiry)
		free_inquiry_stuff(sess, lun);
	conn->assigned_lun_count--;
	printk("Unassign lun %u from connection %u, %u luns assigned\n",
	       lun, conn->connection_id, conn->assigned_lun_count);
}

/*
 * the host_data_lock MUST be held by the calling process/thread.
 * remove this lun from this session's table of assigned connections.
 */
static void __attribute__ ((no_instrument_function))
    drop_lun_from_sess(__u32 lun, struct session *session)
{
	unsigned long lock_flags;

	if (session->lun_assignments[lun]) {
		UNH_LOCK(&session->sess_lock, lock_flags);

		/* this lun was assigned to this connection, unassign it now */
		locked_drop_lun_from_sess(lun, session, 1);

		UNH_UNLOCK(&session->sess_lock, lock_flags);
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * find a lun that is already in use in this session.
 * returns -1 on error, else lun number in range [0..31]
 */
int find_used_lun(struct session *sess)
{
	int lun;

	if (unlikely(sess->lun_bits == 0)) {
		TRACE_ERROR
		    ("No LUNs currently in use by session %p target %u\n",
		     sess, sess->scsi_target_id);
		lun = -1;
	} else {
		/* at least 1 bit set in lun_bits, find the smallest
		 * numbered 1 bit
		 */
		for (lun = 0; !(sess->lun_bits & (1 << lun)); lun++) {
		}
	}

	return lun;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Create a new recovery connection to keep the iSCSI session active.
 * Remember, iscsi requires at least one connection in a session to be active
 * for the session to be alive - SAI
 * Returns 1 on success, 0 on failure.
 */
int create_rec_conn(struct connection *conn)
{
	__u32 connid;
	int lun, retval;
	struct connection *this_conn;
	struct session *sess = conn->my_session;
	struct iscsi_targetdata *targ_data;
	struct sockaddr *new_ip_address;
	int new_ip_length;
	unsigned long flags, lock_flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Entering create_rec_conn\n",
	      current->comm);

	connid = conn->connection_id;
	TRACE(TRACE_DEBUG, "Only connection available is %p, id %u -"
	      "Create new connection for recovery\n", conn, connid);

	targ_data = &global_hostdata->target_data[sess->scsi_target_id];
	restore_global_data(targ_data);
	targ_data->nop_period = 0;
	targ_data->init_sess_flags = 0;
	targ_data->init_conn_flags = 0;
	global_hostdata->init_sessrec_flg = 0;

	/* find a connection id that is not yet used in this session */
      retry:
	connid++;
	for (this_conn = sess->connection_head; this_conn;
	     this_conn = this_conn->next) {
		if (this_conn->connection_id == connid) {
			/* this id already in use, try the next one */
			goto retry;
		}
	}

	/* find a lun that is already in use in this session */
	lun = find_used_lun(sess);
	if (unlikely(lun < 0)) {
		retval = 0;
		goto out;
	}

	/* got the lun, now create a new recovery session */
	if (dup_inet_struct(conn->ip_address, &new_ip_address,
			    &new_ip_length) < 0
	    || create_session(sess->scsi_target_id, new_ip_address,
			      new_ip_length, lun, connid,
			      global_host) != 0) {
		TRACE_ERROR("Unable to create recovery connection\n");
		retval = 0;
		goto out;
	}

	/* if any luns were assigned to old connection, reassign them */
	if (conn->assigned_lun_count) {
		UNH_LOCK(&host_data_lock, flags);
		UNH_LOCK(&sess->sess_lock, lock_flags);
		for (lun = 0; lun < MAX_LUNS; lun++) {
			if (conn == sess->lun_assignments[lun]) {
				locked_drop_lun_from_sess(lun, sess, 0);
				locked_add_lun_to_sess(lun, sess);
			}
		}
		UNH_UNLOCK(&sess->sess_lock, lock_flags);
		UNH_UNLOCK(&host_data_lock, flags);
	}

	retval = 1;
      out:
	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leaving create_rec_conn retval = %d\n", current->comm,
	      retval);
	return retval;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * release a socket and free the connection structure
 */
void __attribute__ ((no_instrument_function))
    free_connection(struct connection *conn)
{
#if !defined(RDMA_MODE)
	char ip_string[INET6_ADDRSTRLEN + 2], port_string[8];
#endif /* !RDMA_MODE */

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#if !defined(RDMA_MODE)
	if (conn->sock != NULL) {
		TRACE(TRACE_NET, "Releasing socket %p\n", conn->sock);
		if (cnv_inet_to_string(conn->ip_address, ip_string,
				       port_string) > 0) {
			printk("Disconnected from %s:%s\n", ip_string,
			       port_string);
		}
		sock_release(conn->sock);
		TRACE(TRACE_NET, "Released socket %p\n", conn->sock);
		conn->sock = NULL;
	}
#else  /* RDMA_MODE */
	iser_conn_terminate(conn);
	queue_free(&conn->rx_pdu_q);
#endif /* !RDMA_MODE */

	my_kfree((void **) &conn->isvalid, "isvalid");
	my_kfree((void **) &conn->accepted_stat_sn, "accepted stat sn");

	TRACE(TRACE_NET, "Free ip_address %p\n", conn->ip_address);
	my_kfree((void **) &conn->ip_address, "ip_address");

	TRACE(TRACE_NET, "Free local_ip_address %p\n",
	      conn->local_ip_address);
	my_kfree((void **) &conn->local_ip_address, "local_ip_address");

	TRACE(TRACE_DEBUG, "%s Free conn %p\n", current->comm, conn);
	my_kfree((void **) &conn, "connection");
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * clear disconnected and unused connections - SAI
 */
void __attribute__ ((no_instrument_function))
    clear_free_conn(struct session *sess)
{
	struct connection *prev, *conn;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!sess->connection_head) {
		TRACE_ERROR("No Valid connections available\n");
		return;
	}

	if (sess->nconnections <= 1
	    && sess->connection_head->connection_state ==
	    CONNECTION_RECOVERING) {
		/* No connections available - Need to create a new one - SAI */
		create_rec_conn(sess->connection_head);
	}

	for (prev = NULL, conn = sess->connection_head;
	     conn != NULL; conn = conn->next) {
		if ((conn->connection_state == CONNECTION_NOT_PRESENT)) {
			TRACE(TRACE_ISCSI, "%s Releasing connection %p\n",
			      current->comm, conn);
			if (prev == NULL)
				sess->connection_head = conn->next;
			else
				prev->next = conn->next;

			/* Adjust the round robin connection head - SAI */
			if (conn == sess->rr_conn_head) {
				sess->rr_conn_head = prev;
			}

			free_connection(conn);
			break;
		}
		prev = conn;
	}

	return;
}

/*
 * executed by scsi midlevel process and rx thread
 * the session->sess_lock MUST be held by the calling process/thread.
 * performs round-robin scheduling by picking up at last connection we
 * scheduled and moving to next one that is in full feature phase.
 * Returns pointer to next connection to schedule, and leaves session's
 * rr_conn_head pointing to it.
 */
static struct connection * __attribute__ ((no_instrument_function))
    do_round_robin(struct session *session)
{
	int nconn;
	struct connection *conn;

	if ((conn = session->rr_conn_head) == NULL
	    || (conn = conn->next) == NULL)
		conn = session->connection_head;

	for (nconn = 0; nconn < session->nconnections; nconn++) {
		if (conn->connection_state ==
		    CONNECTION_FULL_FEATURE_PHASE)
			break;

		if ((conn = conn->next) == NULL)
			conn = session->connection_head;
	}
	return session->rr_conn_head = conn;
}

/*
 * executed by scsi midlevel process and rx thread
 * the session->sess_lock MUST be held by the calling process/thread.
 * Multiple Connection Scheduling Algorithm - SAI
 */
struct connection * __attribute__ ((no_instrument_function))
    get_connection(struct scsi_cmnd *Cmnd, struct session *sess)
{
	struct connection *conn;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Entering get_connection\n",
	      current->comm);

	switch (sess->sched_scheme) {

	case CONN_SCHED_RR:
		/* get the next connection in the round-robin */

		if (sess->nconnections == 1)
			conn = sess->connection_head;
		else {
			conn = do_round_robin(sess);
		}
		break;

	case CONN_SCHED_LUN:
		/* get the connection assigned to this LUN */

#if !defined(__KERNEL__) || defined(K26)
		conn = sess->lun_assignments[Cmnd->device->lun];
#else
		conn = sess->lun_assignments[Cmnd->lun];
#endif

		break;

	default:
		/* use only one connection */
		conn = sess->connection_head;
		break;
	}

	TRACE(TRACE_ENTER_LEAVE, "%s Leaving get_connection\n",
	      current->comm);
	return conn;
}

static int iscsi_initiator_do_queuecommand(struct scsi_cmnd *Cmnd,
					   void (*done) (struct scsi_cmnd *));

#if defined(__KERNEL__)
/**
 * data for the queuecommand work function.
 */
struct iscsi_queuecommand_data {
	struct scsi_cmnd *cmnd;
	void (*done)(struct scsi_cmnd *);
	struct work_struct work;
};

/**
 * The queuecommand deferred work function.
 */
static void iscsi_queuecommand_worker(work_func_arg *a)
{
	struct work_struct *work;
	struct iscsi_queuecommand_data *data;

	work = GET_WORK_STRUCT_P(a);
	data = container_of(work, struct iscsi_queuecommand_data, work);

	iscsi_initiator_do_queuecommand(data->cmnd, data->done);

	kfree(data);
}
#endif /* __KERNEL__ */

/**
 * The SCSI midlevel queuecommand callback.
 */
static int iscsi_initiator_queuecommand(struct scsi_cmnd *cmnd,
					void (*done)(struct scsi_cmnd*))
{
#if defined(__KERNEL__)

	struct iscsi_queuecommand_data *data;

	data = kmalloc(sizeof(*data), GFP_ATOMIC);
	if (!data) {
		TRACE_ERROR("%s: kmalloc failed\n", __func__);
		return -ENOMEM;
	}

	data->cmnd = cmnd;
	data->done = done;
	KERNEL_INIT_WORK(&data->work, iscsi_queuecommand_worker);

	schedule_work(&data->work);

	return 0;

#else /* !__KERNEL__ */

	/* in user-space, just call through to the
	 * iscsi_initiator_do_queuecommand() function */

	return iscsi_initiator_do_queuecommand(cmnd, done);

#endif /* __KERNEL__ */
}

/*****************************************************************************
* This function is called by a workqueue.
* This function will set up TCP buffers for sending SCSI
* Command and, if the negotiated ImmediateData is 1, ImmediateData. After
* setting up the TCP buffers, the tx thread is woken up which is waiting on
* tx_sem.
* Parameters: struct scsi_cmnd,
*	  done function which is called when any SCSI Command is satisfied.
* Return Value: 0, on success
* 		-1, on failure (any non-zero value seems to work)
*
* Note that the done() function should not be called until queuecommand returns!
*
*****************************************************************************/
static int
iscsi_initiator_do_queuecommand(struct scsi_cmnd *Cmnd,
				void (*done) (struct scsi_cmnd *))
{
	int retval = 0;		/* assume this function will succeed */
	__u32 target_id;
	struct iscsi_hostdata *hostdata;
	struct session *current_session;
	unsigned long flags, lock_flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

/* Ming Zhang, mingz@ele.uri.edu */
#if !defined(__KERNEL__) || defined(K26)
	target_id = Cmnd->device->id;
#else
	target_id = Cmnd->target;
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter iscsi_initiator_queuecommand target id %u\n",
	      current->comm, target_id);

	/* get exclusive access to the iscsi structures */
	UNH_LOCK(&host_data_lock, flags);

	if (unlikely(global_host == NULL || global_hostdata == NULL)) {
		TRACE_ERROR("queuecommand called after shutdown\n");
		retval = -1;
		goto out;
	}

	/* get pointer to hostdata struct registered to the SCSI Mid-level */

/* Ming Zhang, mingz@ele.uri.edu */
#if !defined(__KERNEL__) || defined(K26)
	hostdata = (struct iscsi_hostdata *) Cmnd->device->host->hostdata;
#else
	hostdata = (struct iscsi_hostdata *) Cmnd->host->hostdata;
#endif

	if (unlikely(hostdata != global_hostdata)) {
		TRACE_ERROR
		    ("Hostdata struct %p not same as global_hostdata %p\n",
		     hostdata, global_hostdata);
		retval = -1;
		goto out;
	}

	/* tell Mid-level when I am done receiving the SCSI Cmnd info. */
	Cmnd->scsi_done = done;

	/* Find out the session corresponding to the target_id */
	current_session = find_session_by_id(target_id, hostdata);

	/* Check that we have finished the Login Phase
	 * with the concerned target and are now in the
	 * full feature phase of a normal session
	 */
	if (unlikely(current_session == NULL
		     || current_session->session_state !=
		     SESSION_FULL_FEATURE_PHASE
		     || current_session->connection_head == NULL
		     || current_session->connection_head->rx_thread == NULL
		     || current_session->connection_head->tx_thread == NULL
		     || current_session->oper_param.SessionType)) {

		/* Check to see if session recovery is active - SAI */
		if (global_hostdata && global_hostdata->init_sessrec_flg) {
			/* changes for session recovery new 18_04 - SAI */

			TRACE(TRACE_ISCSI,
			      "%s Recovery in progress for target " "%u\n",
			      current->comm, target_id);

			/* create a new session recovery structure or
			 * get the already existing session recovery
			 * structure depending on whether one or
			 * multiple sessions participate in session
			 * recovery - new 18_04 SAI
			 */

			/* NOTE XXX jpd 7/18/2003
			 * this section has changed in ref20_17 compared
			 * to ref20_10 (RPM 1.2.0) .
			 * Needs investigations for lock spliting
			 * per-session. I don't know why these are commented
			 * out .
			 */
			/* create and store in a recovery session - SAI */
			set_rec_group(Cmnd);
			goto out;
		}

		printk("iscsi Target %u not logged in\n", target_id);

		Cmnd->result = DID_NO_CONNECT << 16;
		if (likely(done != NULL))
			/* should not do this before queuecommand returns! */
			done(Cmnd);
		retval = -1;
		goto out;
	}

	UNH_LOCK(&current_session->sess_lock, lock_flags);
	retval = scsi_to_iscsi(Cmnd, current_session);
	UNH_UNLOCK(&current_session->sess_lock, lock_flags);

      out:
	/* release our exclusive access to the iscsi structures */
	UNH_UNLOCK(&host_data_lock, flags);

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave iscsi_initiator_queuecommand retval %d\n",
	      current->comm, retval);

	return retval;
}

/*****************************************************************************
 * executed by scsi midlevel process and init_recovery thread
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * This function takes the command and data from the scsi buffers, forms a new
 * iSCSI PDU with the corresponding command and data filled-in and attaches the
 * command to the end of the pending commands list for the session - SAI
 * scsi midlevel's SPIN_LOCK host_lock must NOT be held
*****************************************************************************/
int scsi_to_iscsi(struct scsi_cmnd *Cmnd, struct session *current_session)
{
	int retval = 0;		/* assume this function succeeds */
	__u32 lun;
	struct iscsi_init_scsi_cmnd *iscsi_cmd;
	struct command *new_command;
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter scsi_to_iscsi\n");

/* Ming Zhang, mingz@ele.uri.edu */
#if !defined(__KERNEL__) || defined(K26)
	lun = Cmnd->device->lun;
#else
	lun = Cmnd->lun;
#endif

	if (current_session->nconnections <= 1)
		conn = current_session->connection_head;
	else {
		/* multiple connections usage - SAI */
		conn = get_connection(Cmnd, current_session);
	}
	if (unlikely(conn == NULL)) {
		TRACE_ERROR("Session %u has no connections for lun %u\n",
			    current_session->scsi_target_id, lun);
		retval = -1;
		goto out;
	}

	/* make sure connection is still in FFP */
	if (unlikely
	    (conn->connection_state != CONNECTION_FULL_FEATURE_PHASE)) {
		/* rx_thread gone -- peer probably reset the connection */
		Cmnd->result = DID_NO_CONNECT << 16;

		if (likely(Cmnd->scsi_done != NULL)) {
			/* can we do this before queuecommand returns?! */
			call_back_scsi_done(Cmnd);
		}
		goto out;
	}

	/* get a new struct command to send SCSI Request PDU to target */
	new_command = setup_command(Cmnd, conn);
	if (unlikely(new_command == NULL)) {
		retval = -1;
		goto out;
	}
	new_command->lun = lun;

	/* set up the SCSI_Cmnd header that is to be sent to the target */
	iscsi_cmd = &new_command->iscsi_cmd;
	iscsi_cmd->opcode = ISCSI_INIT_SCSI_CMND;
	iscsi_cmd->flags = current_session->lun_attr[lun];

	/* SAM-2, section 4.12.2 LUN 0 address
	 * "To address the LUN 0 of a SCSI device the peripheral device
	 * address method shall be used."
	 *
	 * What this means is that whenever the LUN is 0, the full 8 bytes of
	 * the LUN field in the iscsi PDU will also be 0. Because we have
	 * already zeroed out the entire PDU header in setup_command() above,
	 * we don't need to do anything else here in this case.
	 */
	if (lun) {
		pack_lun(lun,
			 conn->connection_flags & USE_FLAT_SPACE_LUN,
			 (__u8 *) & iscsi_cmd->lun);
	}
	iscsi_cmd->xfer_len = htonl(scsi_bufflen(Cmnd));

	TRACE(TRACE_DEBUG, "Xfer len for queuecommand %d\n",
	      scsi_bufflen(Cmnd));

	/* copy the SCSI CDB into the PDU */
	memcpy(iscsi_cmd->cdb, Cmnd->cmnd, Cmnd->cmd_len);

	if (Cmnd->sc_data_direction == SCSI_DATA_WRITE) {
		/* SCSI Command has write data to be sent.
		 * Mark the cmd to be built when xmit_thread wakes up in order
		 * to avoid any potential sleep hazards in
		 * build_write_command(), since we may be in an interrupt
		 * context here.
		 */
		new_command->need_write_built |= CMD_NEEDS_WRITE_BUILT;
		TRACE(TRACE_DEBUG, "command %p marked for write\n",
		      new_command);
	} else {
		/* not a write, probably a read, but check out of curiousity */

		/* always set the F_BIT if it is not a WRITE SCSI Commamnd */
		if (likely(Cmnd->sc_data_direction == SCSI_DATA_READ)) {
			iscsi_cmd->flags |= F_BIT | R_BIT;
		} else {
			/* direction given as neither Read nor Write */
			TRACE(TRACE_DEBUG,
			      "sc_data_direction %u not READ or WRITE\n",
			      Cmnd->sc_data_direction);
			if (unlikely(scsi_bufflen(Cmnd) != 0)) {
				TRACE_ERROR
				    ("request_bufflen %u not zero when "
				     "sc_data_direction %u not READ or "
				     "WRITE\n", scsi_bufflen(Cmnd),
				     Cmnd->sc_data_direction);
				iscsi_cmd->flags |= F_BIT | R_BIT;
			} else {
				/* no data expected to be transfered,
				 * not READ or WRITE, so only set the F bit
				 */
				iscsi_cmd->flags |= F_BIT;
			}
		}
	}

	/* attach the new SCSI Request command to end of pending commands list
	 * this also sets the ITT and the CmdSN.
	 */
	attach_pending_command(new_command, conn, 1);
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave scsi_to_iscsi, retval %d\n",
	      retval);
	return retval;
}

/****************************************************************************
* This function is called by the system when the abort_timer set up in
* iscsi_initiator_abort() expires. This function marks the
* task management command as "abort finished" and wakes up the
* tx thread to process it.
****************************************************************************/
static void abort_timer_function(unsigned long data)
{
	struct command *tm_command = (struct command *) data;
	struct connection *conn;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Enter abort_timer_function\n",
	      current->comm);
	if (tm_command
	    && (tm_command->task_mgt_response == FUNCTION_REJECTED)
	    && (conn = tm_command->my_conn)
	    && tm_command->header_opcode ==
	    (ISCSI_INIT_TASK_MGMT_CMND | I_BIT)) {
		/* mark this task management command as "abort finished"
		 * after making it look like a reply was received from the
		 * target (for purposes of dealing with initiator's outstanding
		 * commands), then wake up the tx thread for this connection to
		 * finish the task management processing
		 */
		tm_command->stat_sn = conn->last_rcvd_stat_sn;
		tm_command->cmd_state |= CMD_STATE_ABORT_FINISH;
		printk("%s TM request timed out, ITT %u, RTT %u\n",
		       current->comm, tm_command->init_task_tag,
		       tm_command->ref_task_tag);
		if (atomic_read(&conn->tx_sem.count) <= 0) {
			TRACE(TRACE_ISCSI_FULL,
				"%s wake up tx_thread on tx_sem %p\n",
				current->comm, &conn->tx_sem);
			up(&conn->tx_sem);
		}
	}
	TRACE(TRACE_ENTER_LEAVE, "%s Leave abort_timer_function\n",
	      current->comm);
}

static void
dump_command_info(struct command *related_cmd,
		  struct connection *related_connection,
		  struct session *related_session)
{
	long delta_time;
	struct r2t_cookie *cookie;

	if (related_cmd->SCpnt) {
		printk(" SCpnt %p request_bufflen %u request_buf "
		       "%p\n", related_cmd->SCpnt,
		       scsi_bufflen(related_cmd->SCpnt),
		       scsi_buffer(related_cmd->SCpnt));
	}
	delta_time = jiffies - related_cmd->time_stamp;
	printk(" op 0x%02x, flags 0x%02x, time_stamp %lu, delta_time "
	       "%ld\n", related_cmd->header_opcode,
	       related_cmd->header_flags, related_cmd->time_stamp,
	       delta_time);
	printk(" recvd_length %u, data_offset %u, still_to_send %u, "
	       "wait_to_send %lu, cmd_state 0x%02x\n",
	       related_cmd->recvd_length,
	       related_cmd->data_offset, related_cmd->still_to_send,
	       related_cmd->tx_wait_to_send, related_cmd->cmd_state);
	if (related_cmd->SCpnt) {
		printk(" request_bufflen %u, recv_so_far %u, "
		       "data_in_sn %u, direction %c\n",
		       scsi_bufflen(related_cmd->SCpnt),
		       related_cmd->r2t_range_list.limit,
		       related_cmd->data_in_sn,
		       related_cmd->SCpnt->sc_data_direction
		       == SCSI_DATA_WRITE ? 'W' : 'R');
	}
	printk
	    (" tx_size %u, tx_sent_so_far %u, r2t_sn %u, n_r2t_cookies %u, "
	     "msg_flags 0x%08x\n", related_cmd->tx_size,
	     related_cmd->tx_sent_so_far, related_cmd->r2t_sn,
	     related_cmd->n_r2t_cookies, related_cmd->cmd_msg_flags);
	printk
	    (" short_command %p, max_cmd_sn %u, cmd_sn %u, difference %d\n",
	     related_connection->short_command,
	     related_session->max_cmd_sn,
	     ntohl(related_cmd->iscsi_cmd.cmd_sn),
	     related_session->max_cmd_sn -
	     ntohl(related_cmd->iscsi_cmd.cmd_sn));
	if (list_empty(&related_cmd->r2t_cookies)) {
		printk(" No R2T cookies\n");
	} else {
		list_for_each_entry(cookie, &related_cmd->r2t_cookies,
				    link) {
			printk
			    ("cookie TTT %u, data_out_sn %u, r2t_xfer_length "
			     "%d, offset %u\n",
			     ntohl(cookie->target_xfer_tag),
			     cookie->data_out_sn,
			     (int) cookie->r2t_xfer_length,
			     cookie->offset);
		}
	}
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Searches iscsi structures for Cmnd
 * Returns 1 on success, 0 on failure to find this command
 * On success, it has filled in the last 3 parameters :
 *	session, connection, and related command
 *	with session->sess_lock LOCKED.
 * On failure, these last 3 parameters are unchanged /undefined
 * and session->sess_lock is NOT LOCKED.
 */
static int
find_aborted_command(struct scsi_cmnd *Cmnd,
		     struct session **found_session,
		     struct connection **found_connection,
		     struct command **found_command,
		     unsigned long *lock_flags)
{
	struct session *related_session;
	struct connection *related_connection = NULL;
	struct command *related_cmd;
	struct iscsi_hostdata *hostdata;
	int retval = 0;		/* assume failure */
	__u32 target_id;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	UNH_LOCK(&host_data_lock, flags);

/* Ming Zhang, mingz@ele.uri.edu */
#if !defined(__KERNEL__) || defined(K26)
	target_id = Cmnd->device->id;
	hostdata = (struct iscsi_hostdata *) Cmnd->device->host->hostdata;
#else
	target_id = Cmnd->target;
	hostdata = (struct iscsi_hostdata *) Cmnd->host->hostdata;
#endif

	if (unlikely(hostdata != global_hostdata)) {
		TRACE_ERROR
		    ("Hostdata struct %p not same as global_hostdata %p\n",
		     hostdata, global_hostdata);
		UNH_UNLOCK(&host_data_lock, flags);
		goto out;
	}

	/* Find out the session that corresponds to this target ID */
	related_session = find_session_by_id(target_id, hostdata);

	if (unlikely(related_session == NULL)) {
		TRACE_ERROR("No session for target %u\n", target_id);
		UNH_UNLOCK(&host_data_lock, flags);
		goto out;
	}

	UNH_LOCK(&related_session->sess_lock, *lock_flags);
	UNH_UNLOCK(&host_data_lock, flags);

	/* check if the session is in full feature phase or not */
	if (unlikely
	    (related_session->session_state != SESSION_FULL_FEATURE_PHASE
	     || (related_connection =
		 related_session->connection_head) == NULL
	     || related_connection->rx_thread == NULL
	     || related_connection->tx_thread == NULL
	     || related_connection->connection_state !=
	     CONNECTION_FULL_FEATURE_PHASE)) {

		TRACE_ERROR("No connection to target %u\n", target_id);
		UNH_UNLOCK(&related_session->sess_lock, *lock_flags);
		goto out;
	}

	/* search for the PDU containing the SCSI command
	 * which has to be aborted
	 */
	for (; related_connection;
	     related_connection = related_connection->next) {
		list_for_each_entry(related_cmd,
				    &related_connection->pending_commands,
				    link) {
			if (related_cmd->SCpnt == Cmnd) {
				if (likely((related_cmd->iscsi_cmd.
					    opcode & ISCSI_OPCODE)
					   == ISCSI_INIT_SCSI_CMND))
					goto found_command_match;
				TRACE_ERROR
				    ("Related Cmnd %p for target %u has "
				     "opcode 0x%02x, expected 0x%02x\n",
				     Cmnd, target_id,
				     related_cmd->iscsi_cmd.
				     opcode & ISCSI_OPCODE,
				     ISCSI_INIT_SCSI_CMND);
				UNH_UNLOCK(&related_session->sess_lock,
					   *lock_flags);
				goto out;
			}
		}
	}

	/* did not find the command to abort, dump pending cmnds */
	TRACE_ERROR("No related SCSI Cmnd %p for target %u\n", Cmnd,
		    target_id);

	for (related_connection = related_session->connection_head;
	     related_connection != NULL;
	     related_connection = related_connection->next) {
		printk
		    ("   pending cmds for connection %p, CID %u, jiffies %lu\n",
		     related_connection, related_connection->connection_id,
		     jiffies);
		list_for_each_entry(related_cmd,
				    &related_connection->pending_commands,
				    link) {
			dump_command_info(related_cmd, related_connection,
					  related_session);
		}
	}
	UNH_UNLOCK(&related_session->sess_lock, *lock_flags);
	goto out;

      found_command_match:
	/* successfully found the matching command to be aborted
	 * session->sess_lock is LOCKED
	 */
	TRACE_ERROR("%s Aborting SCSI Cmnd with ITT %u, CmndSN %u, "
		    "jiffies %lu\n", current->comm,
		    related_cmd->init_task_tag,
		    ntohl(related_cmd->iscsi_cmd.cmd_sn), jiffies);
	if (TRACE_TEST(TRACE_TASK_MGT)) {
		dump_command_info(related_cmd, related_connection,
				  related_session);
	}

	*found_session = related_session;
	*found_connection = related_connection;
	*found_command = related_cmd;
	retval = 1;

      out:
	return retval;
}

/*
 * the session->sess_lock MUST be held by the calling process/thread.
 * This function is called only from iscsi_initiator_abort()
 * executed by the SCSI Mid-level thread.
 * Allocates, sets up and queues a task management pdu to target,
 * Also starts timer in case target never replies to task mgt command.
 */
static void
send_task_mgt_cmd(struct scsi_cmnd *Cmnd,
		  struct connection *conn, struct command *related_cmd)
{
	struct command *tm_command;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	tm_command = setup_task_mgt_command(Cmnd, conn, related_cmd,
					    TMF_ABORT_TASK);
	if (unlikely(tm_command == NULL)) {
		goto out;
	}

	TRACE(TRACE_DEBUG, "%s Start timer for task mgt rsp from target\n",
	      current->comm);

	/* set up timer and go away until the timer expires
	 * or we receive the task management response from the target
	 */
	tm_command->abort_timer.expires = jiffies + ABORT_TIMEOUT;
	tm_command->abort_timer.data = (long) tm_command;
	tm_command->abort_timer.function = abort_timer_function;
	add_timer(&tm_command->abort_timer);

      out:
	return;
}


/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * This function is called by the SCSI Mid-level when it has to abort any
 * SCSI Command. This function will check if the command has been sent to the
 * target for satisfying the request. If it has been sent, then this function
 * will set up TCP buffers to send out a Task Mgt. Command.
 * When called by the SCSI Mid-level (from linux/drivers/scsi/scsi_error.c)
 * the host_lock (or equivalent) is held by the caller.
 *
 ****************************************************************************/
static int iscsi_initiator_abort(struct scsi_cmnd *Cmnd)
{
	unsigned long lock_flags;
	int retval;
	struct session *sess;
	struct connection *conn;
	struct command *related_cmd;
#if defined(CONFIG_ISCSI_DEBUG)
	__u32 save_trace;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE_GET(save_trace);
	if (TRACE_TEST(TRACE_TASK_MGT)) {
		/* turn on (almost) all tracing while processing an abort */
		TRACE_SET(TRACE_ALL
			  & ~(TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE
			      | TRACE_MY_MEMORY));
	}

	TRACE(TRACE_ENTER_LEAVE, "%s Enter iscsi_initiator_abort\n",
	      current->comm);

/* RDR */
#if !defined(__KERNEL__) || defined(K26)
#else
	/* unlock the lock obtained by the scsi midlevel before calling us */
	spin_unlock(&Cmnd->device->host->host_lock);
#endif

	/* gets exclusive access to the iscsi structures when no error */
	if (!find_aborted_command(Cmnd, &sess, &conn, &related_cmd,
				  &lock_flags)) {
		retval = FAILED;
		goto out;
	}

	/* Successfully found the matching command to be aborted.
	 * Corresponding session, connection and command are valid.
	 * Session lock is held.  Mark this command as being aborted.
	 */
	related_cmd->cmd_state |= CMD_STATE_ABORTING;

	if (related_cmd->cmd_state & CMD_STATE_TXSTARTED) {
		/* this command was previously sent to target, so we now have
		 * to send a task management function command to the target
		 */
		send_task_mgt_cmd(Cmnd, conn, related_cmd);
	}
	/* else this command was never sent to target, so don't send a
	 * task management function command to target now
	 */
	/*
	 * Now abort the command, after making it look like we
	 * got a reply from the target (for purposes of dealing with
	 * initiator's outstanding commands)
	 */
	related_cmd->stat_sn = conn->last_rcvd_stat_sn;
	free_aborted_command(related_cmd);
	retval = SUCCESS;
	UNH_UNLOCK(&sess->sess_lock, lock_flags);

      out:

/* RDR */
#if !defined(__KERNEL__) || defined(K26)
#else
	/* get back lock so caller in scsi midlevel can release it! */
	spin_lock(&Cmnd->device->host->host_lock);
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave iscsi_initiator_abort, 0x%04x\n", current->comm,
	      retval);

	if (TRACE_TEST(TRACE_TASK_MGT)) {
		/* restore tracing to what it was when we came in here */
		TRACE_SET(save_trace);
	}

	return retval;
}

/*****************************************************************************
*	This function is called by the SCSI Mid-level.
*	(via the eh_device_reset_handler field in the
*	Scsi_Host_Template driver_template)
*	Still need to implement this function!!
*****************************************************************************/
static int iscsi_initiator_reset(struct scsi_cmnd *Cmnd)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_reset\n");
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_reset, 0x%04x\n",
	      SUCCESS);
	return SUCCESS;
}


static int
scan_redirect_address(char *string, struct sockaddr **ip_address,
		      int *ip_length)
{
	int retval;
	char *ptr;

	if (*string == '[') {
		/* start of a bracketed IPv6 address */
		if ((ptr = strchr(string, ']')) == NULL) {
			TRACE_ERROR("Illegal bracketed IPv6 address %s\n",
				    string);
			retval = -ECONNABORTED;
		} else {
			/* see if a port was supplied or not */
			if (*(ptr + 1) == ':') {
				ptr += 2;
			} else {
				ptr = ISCSI_WKP_STRING;
			}
			if ((retval = cnv_string_to_inet(string + 1, ptr,
							 ip_address,
							 ip_length)) !=
			    AF_INET6) {
				if (retval >= 0) {
					TRACE_ERROR
					    ("Illegal bracketed IPv6 "
					     "address %s\n", string);
					retval = -ECONNABORTED;
				}
			}
		}
	} else {
		/* must be start of an IPv4 address */
		if ((ptr = strchr(string, ':'))) {
			ptr++;
		} else {
			ptr = ISCSI_WKP_STRING;
		}
		if ((retval = cnv_string_to_inet(string, ptr, ip_address,
						 ip_length)) != AF_INET) {
			if (retval >= 0) {
				TRACE_ERROR("Illegal IPv4 address %s\n",
					    string);
				retval = -ECONNABORTED;
			}
		}
	}
	return retval;
}

/* set up the 6-byte isid from values configured for this target
 * configured values limited to 2 bits for type, 24 bits for number & qualifier
 *
 * RFC 3720 Section 10.12.5 ISID
 * T = 00b
 *	A&B - 22-bit OUI
 *	C&D - 24-bit qualifier
 * T = 01b
 *	A - 6-bit Reserved
 *	B&C - 24-bit IANA Enterprise Number
 *	D - 16-bit Qualifier
 * T = 10b
 *	A - 6-bit Reserved
 *	B&C - 24-bit Random number
 *	D - 16-bit Qualifier
 * T = 11b
 *	A, B, C, D - Reserved
 */
static void __attribute__ ((no_instrument_function))
    setup_isid(struct iscsi_targetdata *targ_data, __u32 * isid)
{
	__u32 temp;
	__u16 qual, *ptr16 = (__u16 *) & isid[1];

	temp = targ_data->isid_type << 30;
	qual = targ_data->isid_qualifier;

	switch (targ_data->isid_type) {
	case 0:		/* OUI Format */
		temp |= ((targ_data->isid_number & MASK_22_BITS) << 8)
		    | ((targ_data->isid_qualifier >> 16) & MASK_8_BITS);
		break;
	case 1:		/* EN Format */
	case 2:		/* Random Format */
		temp |= targ_data->isid_number;
		break;
	case 3:		/* Reserved Format */
		qual = 0;
		break;
	}			/* switch */

	*isid = htonl(temp);
	*ptr16 = htons(qual);
}

/*****************************************************************************
*	Sets the session operation parameters and changes the session
*	state to LOGGED_IN, meaning the initiator is in done with login phase.
*	Called only from init_connection()
*	Return value:	=1 on redirection
*			=0 on any other success
*			<0 on failure
*****************************************************************************/
static int
iscsi_initiator_login(struct connection *current_connection,
		      __u32 when_called,
		      struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		      struct auth_parameter_type auth_param,
		      struct iscsi_targetdata *targ_data,
		      struct sockaddr **redirect_ip_address,
		      int *redirect_ip_length)
/* chap and srp support - CHONG */
{
	int retval = 0, cid;
	int login_task_tag;
	__u32 isid[2];
	struct session *current_session;
	__u32 pdu_version_number = 0, original_tsih;
	struct parameter_type *p;
#ifdef ISCSI_STATS
	unsigned long flags;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter iscsi_initiator_login, current connection %p\n",
	      current->comm, current_connection);
	current_session = current_connection->my_session;

	/* set up the 6-byte isid */
	setup_isid(targ_data, isid);
	cid = current_connection->connection_id;	/* for offset 20 */

	/* initialize task tag for the whole login process */
	/* then update the session's task tag counter */
	login_task_tag = current_session->cur_task_tag++;

	TRACE(TRACE_DEBUG, "iscsi_initiator_login pdu_version_number %u\n",
	      pdu_version_number);


	/* In login phase always disable Header & Data Digests */
	current_connection->connection_flags &= ~USE_HEADERDIGEST;
	current_connection->connection_flags &= ~USE_DATADIGEST;
	current_connection->basic_hdr_len = ISCSI_HDR_LEN;

	original_tsih = current_session->tsih;
	retval = initiator_parameter_negotiate(current_connection,
					       p_param_tbl,
					       auth_param,
					       &current_session->
					       oper_param,
					       &current_session->
					       cur_cmd_sn, login_task_tag,
					       isid,
					       &current_session->tsih, cid,
					       &current_connection->
					       connection_flags,
					       pdu_version_number,
					       when_called,
					       &current_connection->
					       exp_stat_sn,
					       &current_connection->
					       max_send_length,
					       &current_connection->
					       max_recv_length);

	if (unlikely(retval < 0)) {
		TRACE(TRACE_ISCSI, "%s Parameter negotiation failed\n",
		      current->comm);
		/* software caused connection abort */
		retval = -ECONNABORTED;
		goto out;
	} else if (likely(retval == 0)) {
		/* new connection negotiation finished successfully */
		current_connection->last_rcvd_stat_sn =
		    current_connection->exp_stat_sn - 1;
		if (current_connection->
		    connection_flags & USE_HEADERDIGEST)
			current_connection->basic_hdr_len += CRC_LEN;
		current_connection->connection_state =
		    CONNECTION_LOGGED_IN;
		if (original_tsih == 0) {
			/* successfully created new connection for new session
			 * Set up the session-wide negotiated operational
			 * parameters for this new session
			 */
			set_session_parameters(&current_session->
					       oper_param,
					       *current_session->
					       session_params);
		}
#ifdef ISCSI_STATS
		UNH_LOCK(&host_data_lock, flags);
		global_hostdata->instance_stats.
		    iscsi_intr_login_accept_rsp++;
		UNH_UNLOCK(&host_data_lock, flags);
#endif
	} else {
		/* have a redirection attempt */
#ifdef ISCSI_STATS
		UNH_LOCK(&host_data_lock, flags);
		global_hostdata->instance_stats.
		    iscsi_intr_login_redirect_rsp++;
		UNH_UNLOCK(&host_data_lock, flags);
#endif
		p = find_parameter(targetaddress_word, p_param_tbl);
		if (unlikely(p == NULL)) {
			/* could not find TargetAddress key */
			TRACE_ERROR("Cannot find %s key\n",
				    targetaddress_word);
			retval = -ECONNABORTED;
		} else if (!IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info)) {
			/* target never sent us this key */
			TRACE_ERROR
			    ("Target never sent TargetAddr key on Redirection\n");
			retval = -ECONNABORTED;
		} else if (scan_redirect_address(p->str_value,
						 redirect_ip_address,
						 redirect_ip_length) < 0) {
			retval = -ECONNABORTED;
		}
	}

      out:
#ifdef ISCSI_STATS
	if (retval == -ECONNABORTED) {
		struct timeval tv;
		char *old_name = NULL, *new_name = NULL;
		struct sockaddr *old_addr, *new_addr = NULL;
		/*
		 * Do do_gettimeofday() and find_parameter()
		 * before holding the lock to reduce the lock hold
		 * up time.
		 */
		do_gettimeofday(&tv);
		p = find_parameter(targetname_word, p_param_tbl);
		if (p && p->str_value && strlen(p->str_value))
			new_name = my_strdup(p->str_value);
		new_addr =
		    my_kmalloc(sizeof(struct sockaddr), "failure addr");

		UNH_LOCK(&host_data_lock, flags);
		global_hostdata->instance_stats.iscsi_intr_login_failure++;
		global_hostdata->instance_stats.
		    iscsi_intr_last_failure_time = tv;
		if (new_name) {
			old_name = global_hostdata->
			    instance_stats.
			    iscsi_intr_last_tgt_failure_name;
			global_hostdata->instance_stats.
			    iscsi_intr_last_tgt_failure_name = new_name;
		}
		if (current_connection->ip_address) {
			old_addr = global_hostdata->
			    instance_stats.
			    iscsi_intr_last_tgt_failure_address;
			if (!old_addr) {
				global_hostdata->
				    instance_stats.
				    iscsi_intr_last_tgt_failure_address =
				    old_addr = new_addr;
				new_addr = NULL;
			}
			if (old_addr)
				memcpy(old_addr,
				       current_connection->ip_address,
				       sizeof(struct sockaddr));
		}
		UNH_UNLOCK(&host_data_lock, flags);

		my_kfree((void **) &new_addr, "failure addr");
		if (old_name)
			kfree(old_name);
	}
#endif
	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave iscsi_initiator_login retval %d\n", current->comm,
	      retval);
	return retval;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Initializes the fields of the first connection in a new session,
 * connects to a remote target, logs in to it,
 * and creates the rx_thread for that connection
 * if full feature phase has been attained successfully.
 * Called by create_session() when adding new connection to existing session
 * and by init_session() when this is first connection in a new session.
 * Return value: =1 for redirection,
 *		 =0 for any other success,
 *		 <0 for error
 ****************************************************************************/
static int
init_connection(struct connection *conn,
		__u32 when_called,
		struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
		struct iscsi_targetdata *targ_data,
		struct sockaddr **redirect_ip_address,
		int *redirect_ip_length)
{
	int retval;
	/* auth support- CHONG */
	struct auth_parameter_type new_authparam;
	char ip_string[INET6_ADDRSTRLEN + 2], port_string[8];
	char local_ip_string[INET6_ADDRSTRLEN + 2], local_port_string[8];
	int local_addr_len;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* extract address family into retval, ip address and port in strings */
	retval =
	    cnv_inet_to_string(conn->ip_address, ip_string, port_string);
	if (unlikely(retval < 0)) {
		TRACE_ERROR("%s\n", ip_string);
		goto out0;
	}

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter init_connection conn %p sess %p\n", current->comm,
	      conn, conn->my_session);

	TRACE(TRACE_ISCSI, "%s Create connection %p to target %u CID %u "
	      "at %s:%s family %d\n", current->comm, conn,
	      conn->my_session->scsi_target_id, conn->connection_id,
	      ip_string, port_string, retval);

#ifdef RDMA_MODE
	/*
	 * Added for iSER - Minti
	 */
	if (conn->connection_flags & USE_RDMA_LOGIN) {
		iser_status_t status;

		TRACE(TRACE_ISER, "Calling connection_establish %p\n",
		      &conn);

		status =
		    global_hostdata->iser_ext_api.
		    connection_establish(global_hostdata->iser_handle,
					 (iser_conn_handle_t) conn,
					 /*Change to sockaddr */
					 (struct sockaddr_in *) conn->
					 ip_address,
					 (struct sockaddr_in *) NULL);
		TRACE(TRACE_ISER, "iser connection_establish status %d\n",
		      status);

		if (status != ISER_SUCCESS) {
			TRACE_ERROR
			    ("iser connection_establish failed %d\n",
			     status);
			retval = -1;
			goto out0;
		}

		/* Wait till the conn is setup otherwise the first send may fail
		 * May be better to check this in sock_sendmsg() for efficiency
		 */
		TRACE(TRACE_ISER,
		      "After initiator connection_establish\n");
		TRACE(TRACE_SEM,
			"init_connection waiting on iser_conn_sem %p\n",
			&conn->iser_conn_sem);
		down(&conn->iser_conn_sem);
		TRACE(TRACE_SEM,
			"init_connection awake after wait on "
			"iser_conn_sem %p\n", &conn->iser_conn_sem);

		if (conn->connection_state == CONNECTION_TERMINATED) {
			retval = -1;
			TRACE(TRACE_ISER,
			      "Failed connection_establish -ret\n");
			goto out0;
		}
		TRACE(TRACE_ISER,
		      "After connection_establish - state not conn_term\n");
	} else
#endif
	{
#ifdef FC2
		retval =
		    sock_create(retval, SOCK_STREAM, 0, &conn->sock, 0);
#else
		retval = sock_create(retval, SOCK_STREAM, 0, &conn->sock);
#endif
		if (unlikely(retval < 0)) {
			TRACE_ERROR("Can't create socket, error %d\n",
				    -retval);
			goto out0;
		}

		TRACE(TRACE_NET, "%s Created socket %p\n", current->comm,
		      conn->sock);

		/* Ming Zhang, mingz@ele.uri.edu */
#if !defined(__KERNEL__) || defined(K26)
#else
		/* to keep tcp from strangling due to lack of memory, force tcp's
		   calls to alloc_skb() to use atomic allocation, not GFP_KERNEL */
		conn->sock->sk->allocation = GFP_ATOMIC;
#endif

		retval =
		    conn->sock->ops->connect(conn->sock, conn->ip_address,
					     conn->ip_length, 0);

		printk("%s to target %u cid %u at %s:%s conn %p\n",
		       retval <
		       0 ? "***ERROR*** can't connect" : "Connected",
		       conn->my_session->scsi_target_id,
		       conn->connection_id, ip_string, port_string, conn);


		if (unlikely(retval < 0)) {
			/* turn on (almost) all tracing after this call */
			if (TRACE_TEST(TRACE_DEBUG)) {
				TRACE_SET(TRACE_ALL
					  & ~(TRACE_ISCSI_FULL |
					      TRACE_MY_MEMORY));
			}
			goto out1;
		}

		/* Turn off the Nagle Algorithm on this connection */
		tcp_nagle_off(conn->sock);

		/*
		 * Get the socket details - local IP/port
		 * You should't fail here unless you have failed in
		 * connect() call in which case, you won't be here.
		 *
		 * However, be sure to allocate enough space for
		 * "local_ip_address", since sockaddr_in6 is bigger than sockaddr!!
		 */
		local_addr_len = conn->ip_length;	/* value-result parameter */
		if (conn->local_ip_address == NULL)
			conn->local_ip_address =
			    my_kmalloc(sizeof(struct sockaddr_in6),
				       "local_ip_address");
		if (conn->local_ip_address != NULL
		    && conn->sock->ops->getname(conn->sock,
						conn->local_ip_address,
						&local_addr_len, 0) == 0) {
			/* extract ip address and port into strings for printing */
			if (cnv_inet_to_string
			    (conn->local_ip_address, local_ip_string,
			     local_port_string) > 0)
				TRACE(TRACE_DEBUG,
				      "iscsi local address %s:%s\n",
				      local_ip_string, local_port_string);
		}
	}
#ifdef ISCSI_STATS
	add_iscsi_proc_entry((void *) conn, CONNECTION);
#endif

#if !defined(RDMA_MODE)
	/*  Allocate buffer into which rx_thread will read
	 *      PDU headers from target.  For now, we ignore AHS,
	 *      so space is needed only for BHS and digest.
	 */
	conn->rx_buf =
	    my_kmalloc(ISCSI_HDR_LEN + CRC_LEN, "rx_buf");
	if (unlikely(conn->rx_buf == NULL)) {
	        /* no memory for rx_buf */
	        retval = -ENOMEM;
	        goto out2;
	}
#endif /* !RDMA_MODE */


	TRACE(TRACE_ISER, "Before iscsi_initiator_login\n");

	/* we are connected to a target */
	conn->connection_state = CONNECTION_CONNECTED;

	/* chap and srp support - CHONG */
	memset(&new_authparam, 0, sizeof(new_authparam));
	new_authparam.auth_flags = targ_data->auth_parameter.auth_flags;
	new_authparam.chap_local_ctx
	    = CHAP_CloneContext(targ_data->auth_parameter.chap_local_ctx);
	new_authparam.chap_peer_ctx
	    = CHAP_CloneContext(targ_data->auth_parameter.chap_peer_ctx);
	new_authparam.srp_ctx
	    = SRP_CloneContext(targ_data->auth_parameter.srp_ctx);

	TRACE(TRACE_ISER, "Call iscsi_initiator_login\n");

	/* now go do the login phase for this new connection */
	retval = iscsi_initiator_login(conn, when_called, p_param_tbl,
				       new_authparam, targ_data,
				       redirect_ip_address,
				       redirect_ip_length);

	/* chap and srp support - CHONG */
	CHAP_FinalizeContext(new_authparam.chap_peer_ctx);
	CHAP_FinalizeContext(new_authparam.chap_local_ctx);
	SRP_FinalizeContext(new_authparam.srp_ctx);

	if (unlikely(retval != 0)) {
		goto out3;
	}
	TRACE(TRACE_DEBUG,
	      "iscsi_initiator_login ok - create iscsi_rx_thd\n");

	/* Create connection's rx_thread only if get into Full Feature Phase */
	kernel_thread((int (*)(void *)) iscsi_initiator_rx_thread,
		      (void *) conn, 0);

	TRACE(TRACE_DEBUG, "Created iscsi_rx_thread -- wait for it\n");

#ifndef __KERNEL__
	sched_yield();
#endif

	/* wait for the rx thread to set itself up */
	TRACE(TRACE_SEM,
		"init_connection waiting on rx_done_sem %p\n",
		&conn->rx_done_sem);
	down(&conn->rx_done_sem);
	TRACE(TRACE_SEM,
		"init_connection awake after wait on rx_done_sem %p\n",
		&conn->rx_done_sem);

	TRACE(TRACE_DEBUG, "%s Waited for iscsi_rx_thread %p\n",
	      current->comm, conn->rx_thread);

	if (conn->rx_thread == NULL) {
		/* the rx thread has already shut itself down */
		retval = -ECONNRESET;
		goto out4;
	}

	TRACE(TRACE_DEBUG, "Create iscsi_tx_thread\n");

	/* start up the tx thread for this new connection */
	kernel_thread((int (*)(void *)) iscsi_initiator_tx_thread,
		      (void *) conn, 0);

	TRACE(TRACE_DEBUG, "Created iscsi_tx_thread -- wait for it\n");

#ifndef __KERNEL__
	sched_yield();
#endif

	/* wait for the tx thread to set itself up */
	TRACE(TRACE_SEM,
		"init_connection waiting on tx_done_sem %p\n",
		&conn->tx_done_sem);
	down(&conn->tx_done_sem);
	TRACE(TRACE_SEM,
		"init_connection awake after wait on tx_done_sem %p\n",
		&conn->tx_done_sem);

	TRACE(TRACE_DEBUG, "Waited for iscsi_tx_thread %p\n",
	      conn->tx_thread);

	if (conn->rx_thread == NULL || conn->tx_thread == NULL) {
		/* the rx and/or tx thread have already shut themselves down */
		close_connection(conn);
		retval = -ECONNRESET;
		goto out0;
	}

	/* both connections are now up and running
	   the rx thread should be waiting on a recv_msg
	   the tx thread should be waiting on the conn->tx_sem semaphore
	 */
	conn->connection_state = CONNECTION_FULL_FEATURE_PHASE;

      out0:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave init_connection retval %d\n",
	      current->comm, retval);
	return retval;

      out4:
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *) conn, CONNECTION);
#endif
	/* rx thread was started, so cleanup is different */
	goto out0;

      out3:
	conn->connection_state = CONNECTION_NOT_PRESENT;
#if !defined(RDMA_MODE)
	my_kfree((void **) &conn->rx_buf, "rx_buf");

      out2:
#endif /* !RDMA_MODE */

	printk("%s Disconnected from %s:%s\n", current->comm, ip_string,
	       port_string);
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *) conn, CONNECTION);
#endif

      out1:
#ifndef RDMA_MODE
	sock_release(conn->sock);
	conn->sock = NULL;
#endif
	goto out0;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called only by create_session() to
 * 1.	initialize a new session structure
 * 2.	create a new first connection for the session
 * 3.	connect to the target
 * 4.	conduct the complete login phase with the target
 *
 * Return value: =1 on redirection
 *		 =0 on all other successes,
 *		 <0 on failure
 ****************************************************************************/
static int
init_session(struct session *sess,
	     struct connection *conn,
	     __u32 lun,
	     struct iscsi_targetdata *targ_data,
	     struct sockaddr **redirect_ip_address,
	     int *redirect_ip_length)
{
	int retval = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Enter init_session, sess %p\n",
	      current->comm, sess);

	TRACE(TRACE_ISCSI,
	      "%s Create session for target %u lun %u CID %u\n",
	      current->comm, sess->scsi_target_id, lun,
	      conn->connection_id);

#ifdef ISCSI_STATS
	add_iscsi_proc_entry((void *) sess, SESSION);
#endif

	/* Connect to the target and go through the complete login phase */
	/* chap and srp support - CHONG */
	retval = init_connection(conn, LEADING_ONLY | INITIAL_ONLY | ALL,
				 *sess->session_params,
				 targ_data,
				 redirect_ip_address, redirect_ip_length);
	if (unlikely(retval != 0))
		goto out2;

	sess->session_state = SESSION_LOGGED_IN;

	TRACE(TRACE_DEBUG, "ErrorRecoveryLevel %u\n",
	      sess->oper_param.ErrorRecoveryLevel);

	/* force use of snack when error recovery is in effect - RDR
	 *      Draft 20, Section 10.16 SNACK Request
	 *      "If the implementation supports ErrorRecoveryLevel greater than
	 *      zero, it MUST support all SNACK types."
	 */
	if (sess->oper_param.ErrorRecoveryLevel) {
		sess->init_snack_flag
		    |= (DATA_SNACK_ENABLE | STATUS_SNACK_ENABLE);
	}

      out:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave init_session retval %d\n",
	      current->comm, retval);

	return retval;

      out2:
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *) sess, SESSION);
#endif
	goto out;
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * free up the memory allocated to a session struct, but not any connection
 * structs it might point to, which must already have been freed up.
 * Except for those connection structs, undoes the allocations in
 * build_session_skeleton() and those that might come later,
 * in the reverse order
 */
static void unbuild_session_skeleton(struct session *old_session)
{
	struct command *command;
	struct r2t_cookie *cookie;
	struct list_head *lptr, *next;

	/* done with unused r2t cookie structures */
	list_for_each_safe(lptr, next, &old_session->free_r2t_cookies) {
		list_del(lptr);
		cookie = list_entry(lptr, struct r2t_cookie, link);
		my_kfree((void **) &cookie, "r2t cookie");
	}

	/* done with unused command structures */
	list_for_each_safe(lptr, next, &old_session->free_commands) {
		list_del(lptr);
		command = list_entry(lptr, struct command, link);
		my_kfree((void **) &command, "command");
	}

#ifdef ISCSI_STATS
	my_kfree((void **) &old_session->sess_stats, "session_stats");
#endif

	if (old_session->session_params) {
		/* done with the session's parameter table */
		param_tbl_uncpy(*old_session->session_params);
		my_kfree((void **) &old_session->session_params,
			 "session param_tbl");
	}

	/* done with the session structure itself */
	my_kfree((void **) &old_session, "session");
}

/*
 * when called, NO LOCKS SHOULD BE LOCKED!
 * create and initialize a session struct and its first connection struct
 */
static struct session *build_session_skeleton(__u32 target,
					      struct sockaddr *ip_address,
					      int ip_length, __u32 lun,
					      __u32 cid)
{
	struct session *new_session;
	struct connection *conn;
	struct iscsi_targetdata *targ_data;

	new_session = my_kmalloc(sizeof(struct session), "session");
	if (unlikely(new_session == NULL)) {
		goto out1;
	}

	/* reset all fields in new struct session to 0 */
	memset(new_session, 0, sizeof(struct session));

	/* initialize most of the new struct session fields here */

#ifdef __KERNEL__
	new_session->sess_lock = UNH_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&new_session->sess_lock, NULL);
#endif

	INIT_LIST_HEAD(&new_session->free_commands);
	INIT_LIST_HEAD(&new_session->free_r2t_cookies);
	new_session->scsi_target_id = target;
	new_session->n_luns_in_session = 1;
	new_session->lun_bits = 1 << lun;
	new_session->session_state = SESSION_NOT_PRESENT;
	new_session->cur_task_tag = 88888;	/* something unique */
	new_session->cur_cmd_sn = 22222;	/* something unique */
	new_session->max_cmd_sn = new_session->cur_cmd_sn;
	new_session->exp_cmd_sn = new_session->cur_cmd_sn;
	init_timer(&new_session->connrec_timer);

	targ_data = &global_hostdata->target_data[target];

	/* use the configured values to initialize following fields */
	new_session->session_flags = targ_data->init_sess_flags;
	new_session->init_conn_flags = targ_data->init_conn_flags;
#if defined(RDMA_MODE)
	new_session->init_conn_flags |= USE_RDMA_LOGIN;
#endif				/* RDMA_MODE */
	new_session->nop_period = targ_data->nop_period;

	TRACE(TRACE_TIMERS, "new session nop_period %u\n",
	      new_session->nop_period);

	/* Added data and status snack support - SAI */
	new_session->init_snack_flag = targ_data->init_snack_flag;

	/* Added the retransmit wait period - SAI */
	new_session->retran_period = targ_data->retran_period;

	/* Store the connection scheduling scheme - SAI */
	new_session->sched_scheme = targ_data->sched_scheme;

	/* allocate space for configured parameters */
	new_session->session_params
	    = my_kmalloc(MAX_CONFIG_PARAMS * sizeof(struct parameter_type),
			 "session param_tbl");
	if (unlikely(new_session->session_params == NULL)) {
		goto out_err;
	}

	/* init session parameters with copy of configured parameter table */
	param_tbl_cpy(*new_session->session_params, targ_data->param_tbl);

#ifdef ISCSI_STATS
	/* get stats structure */
	new_session->sess_stats =
	    my_kmalloc(sizeof(struct iscsi_session_stats),
		       "session_stats");
	if (unlikely(new_session->sess_stats == NULL)) {
		goto out_err;
	}
	/* reset all the stats field */
	memset(new_session->sess_stats, 0,
	       sizeof(struct iscsi_session_stats));
#endif

	/* create the first connection on this session */
	conn = my_kmalloc(sizeof(struct connection), "connection");
	if (unlikely(conn == NULL)) {
		/* no memory for new connection structure */
		goto out_err;
	}

	/* reset all fields in new struct connection */
	memset(conn, 0, sizeof(struct connection));

	conn->connection_id = cid;
	/* target's default MaxRecvDataSegmentLength */
	conn->max_send_length = 8192;
	/* initiator's MaxRecvDataSegmentLength */
	conn->max_recv_length = 8192;
	conn->nop_counter = 1;
	conn->my_session = new_session;
	conn->connection_flags = new_session->init_conn_flags;
	conn->connection_state = CONNECTION_NOT_PRESENT;
	init_MUTEX_LOCKED(&conn->rx_done_sem);
	init_MUTEX_LOCKED(&conn->tx_sem);
	init_MUTEX_LOCKED(&conn->tx_done_sem);
#ifdef RDMA_MODE
	queue_init(&conn->rx_pdu_q);
	init_MUTEX_LOCKED(&conn->iser_notify_sem);
	init_MUTEX_LOCKED(&conn->iser_conn_sem);
	init_MUTEX_LOCKED(&conn->send_done_sem);
#endif
	init_timer(&conn->tx_timer);
	INIT_LIST_HEAD(&conn->pending_commands);
	INIT_LIST_HEAD(&conn->r2t_list);
	conn->ip_address = ip_address;
	conn->ip_length = ip_length;
	conn->last_accepted = -1;

	/* point fields in new session at new connection */
	new_session->nconnections = 1;
	new_session->connection_head = conn;
	new_session->rr_conn_head = conn;

	/* pre-allocate maximum number command structures session may use
	 * (one extra in case of task-management and/or nopout commands)
	 */
	if (unlikely(preallocate_commands(new_session) == 0)) {
		goto out_err;
	}
      out:
	return new_session;

      out_err:
	unbuild_session_skeleton(new_session);
      out1:
	TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
	my_kfree((void **) &ip_address, "ip_address");
	goto out;
}

static int
add_session_to_hostdata(struct connection *new_connection, __u32 lun,
			struct session *new_session,
			struct Scsi_Host *host);

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * executed by rx thread or tx thread or init_recovery thread or command process
 * Creates and initializes a new session and its first connection struct by
 * calling corresponding functions. It then tries to connect to the target.
 * Called from iscsi_initiator_proc_info() when a line such as
 * 	"iscsi_initiator ip 1234abcd port 4000 target 1 lun 2 cid 3"
 * is written to /proc/scsi/iscsi_initiator/0
 * also called from create_rec_conn() to create a recovery connection
 * Return value:
 * 	=1 on success -- new session added, logged in, now in FFP
 * 	=2 on success -- new lun added to existing session
 * 	=0 on success -- new connection added, logged in, now in FFP
 *			 or new discovery session added, now in FFP
 * 	<0 on failure -- value is errno negated
 ****************************************************************************/
int
create_session(__u32 target, struct sockaddr *ip_address, int ip_length,
	       __u32 lun, __u32 cid, struct Scsi_Host *host)
{
	unsigned long lock_flags;
	int i, retval, redirect_count = 0;
	struct sockaddr *redirect_ip_address;
	int redirect_ip_length;
	struct session *new_session;
	struct connection *conn;
	struct iscsi_hostdata *hostdata
	    = (struct iscsi_hostdata *) host->hostdata;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (unlikely(TRACE_TEST(TRACE_ENTER_LEAVE))) {
		char ip_string[INET6_ADDRSTRLEN + 2];
		char port_string[8];

		if (cnv_inet_to_string(ip_address, ip_string, port_string)
		    < 0) {
			/* should never happen */
			TRACE_ERROR("bad ip_address\n");
			TRACE(TRACE_NET, "Free ip_address %p\n",
			      ip_address);
			my_kfree((void **) &ip_address, "ip_address");
			/* software caused connection abort */
			return -ECONNABORTED;
		}

		TRACE(TRACE_ENTER_LEAVE,
		      "%s Enter create_session to %s:%s, "
		      "target %u, lun %u, cid %u\n", current->comm,
		      ip_string, port_string, target, lun, cid);
	}

      retry:			/* here to try again on a redirection */

	new_session = build_session_skeleton(target, ip_address, ip_length,
					     lun, cid);
	if (unlikely(new_session == NULL)) {
		retval = -ENOMEM;
		goto out;
	}

	/* add the session and/or connection to the global hostdata structure */
	conn = new_session->connection_head;
	retval = add_session_to_hostdata(conn, lun, new_session, host);

	/* retval < 0: error, everything already exists or too many sessions
	 * retval = 0: adding a new connection (only) to an existing session
	 * retval = 1: adding complete new session, lun and connection
	 * retval = 2: adding a new lun (only) to an existing session
	 */
	switch (retval) {
	case 2:
		/* adding a new lun (only) to an existing session */
		/* fall through to default case */

	default:
		/* error, everything already exits or too many sessions
		 * free up newly created connection and session structures
		 */
		goto out3;

	case 0:
		/* adding a new connection (only) to an existing session */
		retval = init_connection(conn, INITIAL_ONLY | ALL,
					 *new_session->session_params,
					 &hostdata->target_data[target],
					 &redirect_ip_address,
					 &redirect_ip_length);
		if (unlikely(retval != 0)) {
			/* could not bring up the new connection,
			 * remove it from session
			 */
			goto out4;
		}

		/* connection is up and running
		 * don't need the session structure or the
		 * session parameter table
		 */
		goto out2;

	case 1:
		/* adding complete new session, lun and connection */
		retval = init_session(new_session, conn, lun,
				      &hostdata->target_data[target],
				      &redirect_ip_address,
				      &redirect_ip_length);
		if (unlikely(retval != 0)) {
			/* initialization or login failed or got redirection */
			goto out4;
		}

		/* everything is now up and running in the new session */
		new_session->session_state = SESSION_FULL_FEATURE_PHASE;

		if (new_session->sched_scheme == CONN_SCHED_LUN) {
			add_lun_to_sess(lun, new_session);
		}

		if (new_session->oper_param.SessionType) {
			/* a discovery session, we have to drive it */
			TRACE(TRACE_ISCSI,
			      "%s Enter discovery session Full Feature Phase\n",
			      current->comm);
			UNH_LOCK(&new_session->sess_lock, lock_flags);
			/* gives back 1 on error, 0 on success */
			retval = drive_text_negotiate(conn, new_session, 0,
						      DISCO_TEXT_NEGO);
			UNH_UNLOCK(&new_session->sess_lock, lock_flags);
		} else {
			TRACE(TRACE_ISCSI,
			      "%s Enter normal session Full Feature Phase\n",
			      current->comm);
			retval = 1;
		}
#ifdef K26			/* Hot-plug modifications for 2.6.0 kernel */
		scsi_add_device(host, 0, target, lun);
#endif

	}			/* switch */

      out:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave create_session retval %d\n",
	      current->comm, retval);
	return retval;

      out4:
	/* Here to recover from a late error, after host's list was updated. */
	if ((i =
	     remove_session_from_hostdata(target, lun, cid,
					  hostdata)) == 1) {
		/* memory for entire session already freed, just leave */
		goto out0;
	} else if (i == 0) {
		/* only memory for new connection freed, get rid
		 * of new session too
		 */
		goto out2;
	}

	/* should never happen */
	TRACE_ERROR("non-existent target %u LUN %u CID %u\n", target, lun,
		    cid);

      out3:
	/* here to free the new session struct and its connection struct */
	my_kfree((void **) &new_session->connection_head, "connection");
	TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
	my_kfree((void **) &ip_address, "ip_address");

      out2:
	unbuild_session_skeleton(new_session);

      out0:
	/* now check to see if failure was caused by request for redirection */
	if (retval == STAT_CLASS_REDIRECTION) {
		if (redirect_count++ < MAX_REDIRECTION_RETRIES) {
			TRACE(TRACE_ISCSI,
			      "%s Attempting Redirection number " "%u\n",
			      current->comm, redirect_count);
			ip_address = redirect_ip_address;
			ip_length = redirect_ip_length;
			goto retry;
		}
		TRACE_ERROR("Too many redirection attempts, max of %d\n",
			    MAX_REDIRECTION_RETRIES);
		TRACE(TRACE_NET, "Free ip_address %p\n", ip_address);
		my_kfree((void **) &redirect_ip_address, "ip_address");
		retval = -ECONNABORTED;	/* software caused connection abort */
	}
	goto out;
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * returns 0 if ok, <0 on error
 */
static int __attribute__ ((no_instrument_function))
    add_conn_to_sess(struct connection *new_connection,
		 __u32 lun, struct session *old_session)
{
	int retval;
	struct connection *conn, *prev;

	retval = 0;
	prev = NULL;
	for (conn = old_session->connection_head;
	     conn != NULL; conn = conn->next) {
		if (unlikely(conn->connection_id
			     == new_connection->connection_id)) {
			TRACE_ERROR
			    ("Already have target %u LUN %u CID %u\n",
			     old_session->scsi_target_id, lun,
			     new_connection->connection_id);
			retval = -EBUSY;
			goto out;
		}
		prev = conn;
	}

	if (unlikely(old_session->nconnections
		     >= old_session->oper_param.MaxConnections)) {
		TRACE_ERROR("Already have maximum of %u connections"
			    " for target %u\n",
			    old_session->nconnections,
			    old_session->scsi_target_id);
		retval = -EBUSY;
		goto out;
	}

	/* ok to add new connection at end of session's connection list */
	old_session->nconnections++;
	if (prev == NULL)
		old_session->connection_head = new_connection;
	else
		prev->next = new_connection;

	/* override initializations that depend on session */
	new_connection->my_session = old_session;
	new_connection->connection_flags = old_session->init_conn_flags;

      out:
	return retval;
}

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * executed by rx thread or tx thread or init_recovery thread or command process
 * Called only from create_session().
 * Return value:
 * 	=1 on success -- new_session added to end of hostdata's session_list ok
 * 	=2 on success -- adding new lun to existing session
 * 	=0 on success -- adding new connection to existing session
 * 	<0 on failure -- value is errno negated
 ****************************************************************************/
static int
add_session_to_hostdata(struct connection *new_connection, __u32 lun,
			struct session *new_session,
			struct Scsi_Host *host)
{
	unsigned long flags, lock_flags;
	struct session *old_session;
	int retval;
	struct iscsi_hostdata *hostdata
	    = (struct iscsi_hostdata *) host->hostdata;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter add_session_to_hostdata new_session %p\n",
	      current->comm, new_session);

	UNH_LOCK(&host_data_lock, flags);

	/* see if a session with the same target id already exists */
	old_session =
	    find_session_by_id(new_session->scsi_target_id, hostdata);

	if (old_session != NULL) {
		/* already have a session with this target id */
		UNH_LOCK(&old_session->sess_lock, lock_flags);
		if (!(old_session->lun_bits & new_session->lun_bits)) {
			/* adding new lun to existing session */
			old_session->lun_bits |= new_session->lun_bits;
			old_session->n_luns_in_session++;

			printk("Add LUN %u to existing session "
			       "%p, target %u, now %u LUNS in session\n",
			       lun, old_session,
			       old_session->scsi_target_id,
			       old_session->n_luns_in_session);

			preallocate_commands(old_session);

			if (old_session->sched_scheme == CONN_SCHED_LUN) {
				locked_add_lun_to_sess(lun, old_session);
			}
			retval = 2;
		} else {
			/* this lun already part of session,
			 * must be adding a new connection
			 */
			retval = add_conn_to_sess(new_connection,
						  lun, old_session);
		}
		UNH_UNLOCK(&old_session->sess_lock, lock_flags);
		goto out;
	}

	/* no session with this target id exists yet, add it */
	if (hostdata->nsessions < host->max_id) {
		/* ok to add another session */
		list_add_tail(&new_session->link, &hostdata->session_list);
		hostdata->nsessions++;
		retval = 1;
	} else {
		/* host already has maximum number of sessions allowed */
		TRACE_ERROR("Max of %u sessions, can't add session for "
			    "target %u LUN %u CID %u\n",
			    host->max_id, new_session->scsi_target_id, lun,
			    new_connection->connection_id);
		retval = -EBUSY;
	}

      out:
	UNH_UNLOCK(&host_data_lock, flags);

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave add_session_to_hostdata, retval %d\n",
	      current->comm, retval);

	return retval;
}

/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called from close_session() and remove_session_from_hostdata().
 * Cleans up a connection struct whenever a connection is ended.
 * This does NOT close the socket or free the connection structure
 * and does NOT reduce the session's "nconnections" counter.
 ***************************************************************************/
static void close_connection(struct connection *conn)
{
	void *temp;
	struct session *session;
	unsigned long lock_flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Enter close_connection %p\n",
	      current->comm, conn);

	if (unlikely(conn == NULL
		     || conn->connection_state ==
		     CONNECTION_NOT_PRESENT)) {
		goto out;
	}

	TRACE(TRACE_ISCSI,
	      "%s Closing connection %p for target %u CID %u\n",
	      current->comm, conn, conn->my_session->scsi_target_id,
	      conn->connection_id);

	session = conn->my_session;

	/* get exclusive access to the iscsi session structure */

	UNH_LOCK(&session->sess_lock, lock_flags);

	/* do not allow tx_timer to restart itself */
	conn->connection_flags |= TX_TIMER_OFF;

	/* now delete the tx_timer (del_timer checks for timer active or not) */
	del_timer_sync(&conn->tx_timer);

	if (conn->rx_thread != NULL) {
		/* rx_thread still up and running, kill it */
		TRACE(TRACE_ISCSI, "%s Killing rx-%u-%u\n", current->comm,
		      conn->my_session->scsi_target_id,
		      conn->connection_id);
		temp = conn->rx_thread;
		conn->rx_thread = NULL;
		/*      Mike Christie, mikenc@us.ibm.com */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, temp, 1);

		/* release our exclusive access to the iscsi structures */

		UNH_UNLOCK(&session->sess_lock, lock_flags);

#if defined(__KERNEL__)
		TRACE(TRACE_SEM,
			"close_connection waiting on rx_done_sem %p\n",
			&conn->rx_done_sem);
		down(&conn->rx_done_sem);
		TRACE(TRACE_SEM,
		      "close_connection awake after wait on rx_done_sem %p\n",
		      &conn->rx_done_sem);
#else
		pthread_join(((struct task_struct *) temp)->threadid, NULL);
#endif /* __KERNEL__ */
#if !defined(RDMA_MODE)
	my_kfree((void **) &conn->rx_buf, "rx_buf");
#endif /* !RDMA_MODE */


		/* now get back exclusive access to the iscsi structures */
		UNH_LOCK(&session->sess_lock, lock_flags);
	}

	conn->connection_state = CONNECTION_NOT_PRESENT;
#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *) conn, CONNECTION);
#endif

	if (conn->tx_thread) {
		int ret;
		/* indicate that this thread is going to do the shut down */
		TRACE(TRACE_ISCSI, "%s Killing tx-%u-%u\n", current->comm,
		      conn->my_session->scsi_target_id,
		      conn->connection_id);
		temp = conn->tx_thread;
		conn->tx_thread = NULL;
		/*      Mike Christie, mikenc@us.ibm.com */
		ret = send_sig(ISCSI_SHUTDOWN_SIGNAL, temp, 1);

		/* then release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&session->sess_lock, lock_flags);

		/* then wait for tx thread to actually die */
#if defined(__KERNEL__)
		TRACE(TRACE_SEM,
			"close_connection waiting on tx_done_sem %p, ret %x\n",
			&conn->tx_done_sem, ret);
		down(&conn->tx_done_sem);
		TRACE(TRACE_SEM,
			"close_connection awake after wait on tx_done_sem %p\n",
			&conn->tx_done_sem);
#else
		pthread_join(((struct task_struct *) temp)->threadid, NULL);
#endif /* __KERNEL__ */

	} else {
		/* some other thread is already shutting this session down */

		/* just release our exclusive access to the iscsi structures */
		UNH_UNLOCK(&session->sess_lock, lock_flags);
	}

	TRACE(TRACE_ISCSI, "%s Closed connection for target %u CID %u\n",
	      current->comm, session->scsi_target_id, conn->connection_id);

      out:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave close_connection\n",
	      current->comm);
}

/***************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called only from remove_session_from_hostdata()
 * Cleans up session struct whenever a session has to be ended.
 * When called, this session struct must NOT be linked into the global list.
 * On exit, this session struct is NOT freed.
 ***************************************************************************/
static void close_session(struct session *sess)
{
	struct connection *conn;
	struct connection *next_conn;
	__u32 lun;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Enter close_session\n",
	      current->comm);

	TRACE(TRACE_ISCSI, "%s Closing session %p for target %u\n",
	      current->comm, sess, sess->scsi_target_id);

	/* close all connections, but do not release them or their sockets yet */
	for (conn = sess->connection_head; conn != NULL; conn = next_conn) {
		next_conn = conn->next;
		TRACE(TRACE_DEBUG,
		      "Going to close_connection(%p), next_conn %p\n",
		      conn, next_conn);
		/* close_connection is called without session lock held */
		close_connection(conn);
		TRACE(TRACE_DEBUG,
		      "Back from close_connection(%p), next_conn %p\n",
		      conn, next_conn);
	}

	TRACE(TRACE_ISCSI,
	      "%s All connections closed in session for target " "%u\n",
	      current->comm, sess->scsi_target_id);

	/* delete connection recovery timer - SAI */
	del_timer_sync(&sess->connrec_timer);

	/* go back and release all the connections' sockets and structures */
	while ((conn = sess->connection_head) != NULL) {
		sess->connection_head = conn->next;
		free_connection(conn);
	}

#ifdef ISCSI_STATS
	remove_iscsi_proc_entry((void *) sess, SESSION);
#endif

	for (lun = 0; lun < MAX_LUNS; lun++) {
		free_inquiry_stuff(sess, lun);
	}

	TRACE(TRACE_ISCSI, "%s Closed session for target %u\n",
	      current->comm, sess->scsi_target_id);

	TRACE(TRACE_ENTER_LEAVE, "%s Leave close_session\n",
	      current->comm);
}

/*
 * host_data_lock and sess->sess_lock MUST be held by the calling process/thread
 * remove this connection from this session's list of connections.
 * return 1 if this is the last connection in the session,
 * 0 if more connections remain in the session.
 */
static int __attribute__ ((no_instrument_function))
    drop_conn_from_session(struct connection *conn,
		       struct connection *prev_conn, struct session *sess)
{
	__u32 lun;

	if (sess->nconnections > 1) {
		/* at least 1 more connection will remain in this session */
		sess->nconnections--;
		if (prev_conn == NULL) {
			sess->connection_head = conn->next;
		} else {
			prev_conn->next = conn->next;
		}

		/* Adjust the round robin connection head - SAI */
		if (conn == sess->rr_conn_head) {
			sess->rr_conn_head = prev_conn;
		}

		/* remove all lun assignments, if any */
		if (conn->assigned_lun_count) {
			for (lun = 0; lun < MAX_LUNS; lun++) {
				if (sess->lun_assignments[lun] == conn) {
					/* this lun was assigned to this
					 * connection, so assign it to the
					 * remaining connection that has
					 * the fewest luns assigned to it
					 */
					locked_drop_lun_from_sess(lun,
								  sess, 0);
					locked_add_lun_to_sess(lun, sess);
				}
			}
		}
		return 0;
	} else {
		/* this is last connection, remove entire session */
		sess->lun_bits = 0;
		return 1;
	}
}

/****************************************************************************
* when called, NO LOCKS SHOULD BE LOCKED!
* Removes session, connection or lun for this target
*
* Return value:	=1 on success -- complete session removed from host list
* 		=2 on success -- old lun removed from old session
* 		=0 on success -- old connection removed from old session
* 		<0 on failure -- value is errno negated
*
*  	On entry: No locks held
*  	On exit: No locks held
*
* Called indirectly via create_session() from /proc interface
*		(scan_ip_stuff()) on a "up target lun cid" that fails
* Called directly from /proc interface (scan_ip_stuff())
*		on a "down target lun cid"
* Called indirectly via remove_luns() from /proc interface
*		(iscsi_initiator_release()) on module unload
* Called directly from rx thread (iscsi_initiator_rx_thread())
*		when rx thread terminates a discovery session
* Called indirectly via deal_with_connrec() from rx thread
*		to remove a failed connection after connection recovery
* Called indirectly via create_rec_conn() and create_session() from
*		rx thread to remove a connection after error on creation
* Called indirectly via create_session() from recovery_thread
*		(init_recovery_thread()) to remove a connection after error on
*		creation
* Called indirectly via remove_luns() from recovery_thread
*		(init_recovery_thread()) to remove a failed session
* Called indirectly via remove_luns() from tx thread (tx_shutdown_thread()
*		when tx thread terminates itself
* Called indirectly via create_rec_conn() and create_session() from
*		tx thread (iscsi_initiator_tx_thread()) on test 8 --
*		force session recovery logout
* Called indirectly via remove_luns() from reboot notifier callback
*		(iscsi_initiator_halt())
*
****************************************************************************/
int
remove_session_from_hostdata(__u32 target, __u32 lun, __u32 cid,
			     struct iscsi_hostdata *hostdata)
{
	unsigned long flags, lock_flags;
	int retval = -ENODEV;	/* assume failure */
	struct session *sess;
	struct connection *conn = NULL, *prev_conn;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter remove_session_from_hostdata target %u lun %u cid %u\n",
	      current->comm, target, lun, cid);

	UNH_LOCK(&host_data_lock, flags);

	sess = find_session_by_id(target, hostdata);
	if (likely(sess != NULL)) {
		/* this is the matching session structure, see if lun matches */
		if (sess->lun_bits & (1 << lun)) {
			/* lun matches, see if cid matches */
			/* We need to get the session lock here.
			 * As a rule, host_data_lock is used to protect
			 * global_hostdata and hence all the sessions
			 * attached to it. However, individual conns
			 * in a session are protected by sess_lock on
			 * each session structure. It is possible that
			 * queuecommand/get_connection hold this
			 * sess_lock and search the connection list
			 * when we are deleting/removing the same
			 * connection -AJ
			 */
			UNH_LOCK(&sess->sess_lock, lock_flags);
			prev_conn = NULL;
			for (conn = sess->connection_head;
			     conn != NULL; conn = conn->next) {
				if (conn->connection_id == cid) {
					if (drop_conn_from_session(conn,
								   prev_conn,
								   sess)) {
						/* last conn in sess */
						conn = NULL;
					} else {
						/* conn removed ok */
						retval = 0;
					}
					break;
				}
				prev_conn = conn;
			}
			UNH_UNLOCK(&sess->sess_lock, lock_flags);
			if (conn == NULL) {
				/* did not find matching connection,
				 * or matching connection is last in
				 * session, turn off bit for lun
				 */
				sess->lun_bits &= ~(1 << lun);
				sess->n_luns_in_session--;
				if (sess->lun_bits == 0) {
					/* no luns remain,
					 * remove entire session
					 */
					list_del(&sess->link);
					hostdata->nsessions--;
					retval = 1;
				} else {
					/* other luns remain, just
					 * remove the connection
					 * assignment for this lun
					 */
					drop_lun_from_sess(lun, sess);
					unallocate_commands(sess);
					retval = 2;
				}
			}
		} else {
			/* lun does not match, remove entire session */
			sess->lun_bits = 0;
			list_del(&sess->link);
			hostdata->nsessions--;
			retval = 1;
		}
	}

	UNH_UNLOCK(&host_data_lock, flags);

	switch (retval) {
	case 0:
		/* close only this connection (which has already been removed
		 * from session's connection list.  This has not
		 * released the socket yet.
		 */
		close_connection(conn);

		/* go back and release the socket and free the conn structure */
		free_connection(conn);
		break;

	case 1:
		/* close this entire session, including all
		 * remaining connections
		 */
		TRACE(TRACE_ISCSI,
		      "%s Remove entire session for target %u\n",
		      current->comm, target);

		/* free up any storage connected to this old session */
		if (sess->session_state == SESSION_NOT_PRESENT) {
			/* session's connection structure is empty,
			 * just free any connection it has
			 */
			if (sess->connection_head)
				free_connection(sess->connection_head);
		} else {
			close_session(sess);
		}

		/* finally free up the old struct session itself */
		unbuild_session_skeleton(sess);
		break;

	case 2:
		/* just removed a lun, do nothing further here */
		TRACE(TRACE_ISCSI,
		      "%s Remove LUN from session for target %u lun %u\n",
		      current->comm, target, lun);
		break;

	default:
		/* < 0, error somewhere */
		TRACE(TRACE_ISCSI,
		      "%s Non-existent target %u LUN %u CID %u\n",
		      current->comm, target, lun, cid);
		break;

	}			/* switch */

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave remove_session_from_hostdata retval %d\n",
	      current->comm, retval);
	return retval;
}

/*
 * executed only by command process.
 * the host_data_lock MUST be held by calling thread/process.
 * Returns	=0 on success (and sets old_connection, old_session),
 * 		<0 on failure (and leaves old_connection, old_session unchanged)
 */
int
find_connection(__u32 target,
		__u32 cid,
		struct iscsi_hostdata *hostdata,
		struct connection **old_connection,
		struct session **old_session)
{
	struct session *session;
	struct connection *conn;
	int retval = -ENODEV;	/* assume failure */

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter find_connection, target %u cid %u\n", target, cid);

	session = find_session_by_id(target, hostdata);
	if (likely(session != NULL)) {
		/* this is the matching session structure,
		 * try to find matching cid
		 */
		for (conn = session->connection_head; conn != NULL;
		     conn = conn->next) {
			if (conn->connection_id == cid) {
				/* found a matching connection,
				   pass back answers */
				*old_connection = conn;
				*old_session = session;
				retval = 0;
				break;
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave find_connection, retval %d\n",
	      retval);

	return retval;
}

/*	Mike Christie, mikenc@us.ibm.com
 *
 * when called, NO LOCKS SHOULD BE LOCKED!
 * remove all valid luns in session (last lun will take out session)
 * returns smallest lun number removed, -1 if none removed.
 */
int
remove_luns(struct session *sess, struct Scsi_Host *host,
	    struct iscsi_hostdata *hostdata)
{
	__u32 lun;
	int retval = -1;

	if (unlikely(!sess || sess->lun_bits == 0 || !host || !hostdata))
		goto out;

	for (lun = 0; lun < host->max_lun; lun++) {
		if (sess->lun_bits & (1 << lun)) {
			if (retval < 0)
				retval = lun;
			if (remove_session_from_hostdata
			    (sess->scsi_target_id, lun, ALL_ONES,
			     hostdata) == 1) {
				/* just removed the whole session structure */
				break;
			}
		}
	}
      out:
	return retval;
}

#ifdef __KERNEL__

/*	Mike Christie, mikenc@us.ibm.com
 *
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called when the system is shutdown. Closes all sessions.
 * Does not care about memory menagement
*/
static int
iscsi_initiator_halt(struct notifier_block *nb, ulong event, void *buf)
{
	struct session *sess;
	struct list_head *lptr, *next;

	if (event != SYS_RESTART && event != SYS_HALT
	    && event != SYS_POWER_OFF)
		return NOTIFY_DONE;

	if (global_host == NULL || global_hostdata == NULL)
		return NOTIFY_OK;

	/* Clear the session struct linked list */
	list_for_each_safe(lptr, next, &global_hostdata->session_list) {
		sess = list_entry(lptr, struct session, link);
		if (sess->lun_bits != 0)
			remove_luns(sess, global_host, global_hostdata);
	}
	global_host = NULL;
	global_hostdata = NULL;

	return NOTIFY_OK;
}

#endif

/****************************************************************************
 * when called, NO LOCKS SHOULD BE LOCKED!
 * Called by the SCSI Mid-level when the iscsi_initiator module is removed.
 * (via the release field in the Scsi_Host_Template driver_template)
 * When called, the midlevel holds the big kernel lock, but it does
 * not hold the host_lock.
 * Parameter: Ptr. to Scsi_Host
 * Return Value: 0, on success
 * 		-1, on error
 ****************************************************************************/
static int iscsi_initiator_release(struct Scsi_Host *host)
{
	int retval = 0, i;
	struct iscsi_hostdata *hostdata;
	struct session *sess;
	struct list_head *lptr, *next;
#ifdef RDMA_MODE
	iser_status_t status;
#endif
	TRACE(TRACE_ENTER_LEAVE,
	      "Enter iscsi_initiator_release, host %p\n", host);

	TRACE(TRACE_ISCSI, "Releasing UNH iSCSI initiator module\n");

	/* If the host struct is NULL, it is an error and we want to return */
	if (unlikely(host == NULL || global_host == NULL
		     || global_hostdata == NULL)) {
		retval = -1;
		goto out;
	}

	/* If the hostdata struct is not what we expect, we want to return */
	if (unlikely((hostdata = (struct iscsi_hostdata *) host->hostdata)
		     != global_hostdata)) {
		retval = -1;
		goto out;
	}

	/* Clear the session struct linked list */
	list_for_each_safe(lptr, next, &hostdata->session_list) {
		sess = list_entry(lptr, struct session, link);
		if (sess->lun_bits == 0) {
			/* all luns already removed, just release
			 * this structure
			 */
			list_del(lptr);
			unbuild_session_skeleton(sess);
		} else {
			/* remove all valid luns in session
			 * (last lun will take out session)
			 */
			remove_luns(sess, host, hostdata);
		}
	}

#ifdef RDMA_MODE
	status = iser_api_unregister(global_hostdata->iser_handle);
	if (status != ISER_SUCCESS) {
		TRACE_ERROR("iser_api_unregister failed %d\n", status);
		/* Continue with the release even in error */
	}
#endif

	/* global pointer is no longer valid after unregister */
	if (global_hostdata) {
		for (i = 0; i < MAX_TARGET_IDS; i++) {
			param_tbl_uncpy(global_hostdata->target_data[i].
					param_tbl);
			CHAP_FinalizeContext(global_hostdata->
					     target_data[i].auth_parameter.
					     chap_peer_ctx);
			CHAP_FinalizeContext(global_hostdata->
					     target_data[i].auth_parameter.
					     chap_local_ctx);
			SRP_FinalizeContext(global_hostdata->
					    target_data[i].auth_parameter.
					    srp_ctx);
		}
#ifdef ISCSI_STATS
		if (global_hostdata->instance_stats.
		    iscsi_intr_last_tgt_failure_name) {
			kfree(global_hostdata->instance_stats.
			      iscsi_intr_last_tgt_failure_name);
		}
		my_kfree((void **) &global_hostdata->instance_stats.
			 iscsi_intr_last_tgt_failure_address,
			 "failure address");
#endif
	}
#ifdef K26
	printk("unh_iscsi_initiator removed\n");
#else
	scsi_unregister(host);

	initiator_unregister_device();
#endif

	global_hostdata = NULL;
	global_host = NULL;

	/* check for any unfreed memory */
	my_kempty();

	/* print statistics about io vector usage */
	endup_iovs();

	TRACE(TRACE_ISCSI, "Unregister UNH iSCSI host number %d\n",
	      host->host_no);

#ifdef __KERNEL__

	unregister_reboot_notifier(&iscsi_initiator_notifier);

#endif

	TRACE(TRACE_ISCSI, "Released UNH iSCSI module\n");

      out:
	TRACE(TRACE_ENTER_LEAVE,
	      "Leave iscsi_initiator_release, retval %d\n", retval);
	return retval;
}

/*	Note the name "driver_template" is required for "scsi_module.c" to work! */
/*	Used by: "/usr/src/linux/drivers/scsi/scsi_module.c" */
/*	Wrap this in an ifdef because "scsi_module.c" says to! */
#ifdef MODULE

MODULE_AUTHOR("Bob Russell");
MODULE_DESCRIPTION(OUR_NAME);
MODULE_SUPPORTED_DEVICE("sd,sr,sg,st");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#endif				/* ifdef MODULES */

/*
 * initializor for static Scsi_Host_Template struct as defined in
 *			/usr/src/linux/include/scsi/scsi_host.h
 *	(where this is defined as "struct scsi_host_template"
 *	and here there are good comments about each field).
 *
 *	There is one such struct for each different type of host adaptor,
 *	and since there is only one iscsi type, we only need 1 static instance.
 *	(This is normal for each different adaptor type.)
*/

/* Make max_sectors a module parameter!  This may need to be changed by
 * the user when loading the module if they are connecting to a DISKIO
 * target where the physical target needs a smaller max_sectors value */
unsigned short max_sectors = 4 * (MAX_IOV_SLOTS - 4);
#if defined(__KERNEL__)
module_param(max_sectors, ushort, S_IRUGO);
#endif	/* __KERNEL__ */

/*
 * A single static instance of this structure for this adaptor type
*/
static Scsi_Host_Template driver_template = {
#ifdef __KERNEL__
      proc_info:iscsi_initiator_proc_info,
      module:THIS_MODULE,
#endif
      name:OUR_NAME,
      detect:iscsi_initiator_detect,
      release:iscsi_initiator_release,
#ifdef __KERNEL__
      info:NULL,
      ioctl:NULL,
#endif
      queuecommand:iscsi_initiator_queuecommand,

#ifdef __KERNEL__
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 17)
      eh_strategy_handler:NULL,
#endif
#endif
      eh_abort_handler:iscsi_initiator_abort,
      eh_device_reset_handler:iscsi_initiator_reset,
      eh_bus_reset_handler:NULL,
      eh_host_reset_handler:NULL,

#ifdef __KERNEL__
/* use default scsicam_bios_param() in linux/drivers/scsi/scsicam.c */
      bios_param:NULL,
#endif

/* max no. of simultaneously active SCSI commands driver can accept */
      can_queue:MAX_COMMANDS_PER_LUN * MAX_LUNS,

      this_id:-1,		/* our host has no id on the SCSI bus */

/* max no. of simultaneously active SCSI commands driver can accept */
      sg_tablesize:MAX_IOV_SLOTS - 4,

#ifdef __KERNEL__
/* max no. of sectors driver can accept in 1 SCSI READ/WRITE command */
/* SET BY A MODULE PARAMETER, see max_sectors above */
      max_sectors: 0,
#endif

/* max no. of simultaneously outstanding commands per LUN */
      cmd_per_lun:MAX_COMMANDS_PER_LUN,

#ifdef __KERNEL__
/* no. of cards of this type found on this machine */
      present:0,
      unchecked_isa_dma:0,
/* midlevel should merge requests for adjacent blocks of memory */
      use_clustering:ENABLE_CLUSTERING,

/* The version that introduces this is greater than 2.6.18, but may be
 * less than 2.6.20 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,23)
      /** This is an initiator device. */
      supported_mode:MODE_INITIATOR,
#endif

      /** There is not settleing. */
      skip_settle_delay:1,

#ifndef K26
      use_new_eh_code:1,	/* 2.4.x kernels only */
#endif
      emulated:0,
#endif

/* proc_name xxx becomes the name for the directory /proc/scsi/xxx */
      proc_name:"iscsi_initiator"
};


#ifdef __KERNEL__
#ifndef K26

/*
 * 2.4.x HBA module usage
 */

#include <scsi_module.c>

#endif
#else

#include "user_module.c"

#endif


int max_outstanding_commands(void)
{
	return driver_template.can_queue;
}


#ifdef __KERNEL__
#ifdef K26

/* New 2.6.0 HotPlug modifications */

#define to_iscsi_hostdata(d)       \
        container_of(d, struct iscsi_hostdata, dev)

/* Function declaration. */
static int unh_iscsi_probe(struct device *);
static int add_HBA(void);
static int unh_iscsi_bus_match(struct device *, struct device_driver *);
static int unh_iscsi_remove(struct device *);
static void unh_iscsi_device_release(struct device *);
static void remove_HBA(void);

static char unh_iscsi_proc_name[] = "unh_iscsi_initiator";

/* Place holder for all the HBA */
static LIST_HEAD(unh_iscsi_host_list);

/*
 * This describes each of the pseudo HBA
 */
struct unh_iscsi_host {
	struct list_head host_list;
	struct Scsi_Host *shost;
	struct device dev;
	struct list_head dev_info_list;
};

/*
 * This describes each of the devices attached
 * this is equivalent of sessions
 */
struct unh_iscsi_dev {
	struct list_head dev_list;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;
	struct unh_iscsi_host *iscsi_host;
};


/*
 * This describes the bus type
 */
static struct bus_type unh_iscsi_bus = {
	.name = "unh_iscsi_pseudo",
	.match = unh_iscsi_bus_match,
};

/*
 * The driver structure
 */
struct device_driver unh_iscsi_driver = {
	.name = unh_iscsi_proc_name,
	.bus = &unh_iscsi_bus,
	.probe = unh_iscsi_probe,
	.remove = unh_iscsi_remove,
};

static struct device unh_iscsi_pseudo = {
	.bus_id = "unh_iscsi_pseudo_0",
	.release = unh_iscsi_device_release,
};


/*
 * probe- does most of the stuff that was part of _detect().
 */
static
int unh_iscsi_probe(struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;
	struct Scsi_Host *hostptr;
	int retval = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_probe\n");

	iscsi_host = to_iscsi_hostdata(dev);

	driver_template.max_sectors = max_sectors;
	hostptr = scsi_host_alloc(&driver_template,
				  sizeof(struct iscsi_hostdata));

	if (unlikely(hostptr == NULL)) {
		printk("scsi_register failure\n");
		return -ENODEV;
	}

	/* setup the security key hash table */
	setup_security_hash_table();

	/* set the number of allowed "targets" this adaptor "card" can handle.
	 * the "target" numbers will be [0..max_id-1].
	 * each "target" will have LUN numbers in [0..max_lun-1].
	 */
	hostptr->max_id = MAX_TARGET_IDS;
	hostptr->max_lun = MAX_LUNS;

	/* set the maximum number of bytes we will accept in a scsi cdb */
	hostptr->max_cmd_len = MAX_CDB_LEN;

	iscsi_host->shost = hostptr;
	*((struct iscsi_hostdata **) hostptr->hostdata) = iscsi_host;

	retval = scsi_add_host(hostptr, &iscsi_host->dev);
	if (unlikely(retval)) {
		printk("scsi_add_host failed for unh\n");
		retval = -ENODEV;
		scsi_host_put(hostptr);
	} else {
		/* This should go away in future: TODO -AJ */
		global_hostdata =
		    (struct iscsi_hostdata *) hostptr->hostdata;
		global_host = hostptr;

		/* Initialize the adapter's private data structure */
		if (init_initiator(global_hostdata) < 0)
			retval = -1;

#ifdef __KERNEL__

		/* Register reboot notifier */
		register_reboot_notifier(&iscsi_initiator_notifier);

#endif

	}

	TRACE(TRACE_ENTER_LEAVE, "Leave unh_iscsi_probe\n");
	return retval;
}

static
int unh_iscsi_remove(struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;

	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_remove\n");

	iscsi_host = to_iscsi_hostdata(dev);

	if (unlikely(iscsi_host == NULL)) {
		printk("iscsi_host is NULL\n");
		return -ENODEV;
	}

	if (unlikely(iscsi_host->shost == NULL)) {
		printk("iscsi_host->shost is NULL !!\n");
		return -ENODEV;
	}

	iscsi_initiator_release(iscsi_host->shost);
	scsi_remove_host(iscsi_host->shost);
	scsi_host_put(iscsi_host->shost);

	TRACE(TRACE_ENTER_LEAVE, "Leave unh_iscsi_remove\n");
	return 0;
}

/*
 * Just to silence the warnings
 */
static
void unh_iscsi_device_release(struct device *dev)
{
	TRACE(TRACE_ENTER_LEAVE,
	      "unh_iscsi_device_release not implemented\n");
	return;
}

static int unh_iscsi_bus_match(struct device *dev,
			       struct device_driver *dev_driver)
{
	TRACE(TRACE_ENTER_LEAVE, "unh_iscsi_bus_match not implemented\n");
	return 1;
}

static
void unh_iscsi_release(struct device *dev)
{
	struct iscsi_hostdata *iscsi_host;
	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_release \n");
	/* Free all memory allocation */
	iscsi_host = to_iscsi_hostdata(dev);
	kfree(iscsi_host);
	return;
}


/*
 * add_HBA.
 * In our design, we add a new HBA on the pseudo
 * bus that we create. As of now only one HBA is
 * added, changes can be easily made to add more
 * As of now there is no pressing reason to add
 * more than one (pseudo)HBA.
 */
static
int add_HBA()
{
	struct iscsi_hostdata *iscsi_host;
	int retval = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter add_HBA \n");

	iscsi_host = kmalloc(sizeof(struct iscsi_hostdata), GFP_KERNEL);
	if (unlikely(iscsi_host == NULL)) {
		printk("add_HBA: allocated failed !\n");
		return -ENOMEM;
	}

	memset(iscsi_host, 0, sizeof(struct iscsi_hostdata));
/* Add locking code here */
	list_add_tail(&iscsi_host->host_list, &unh_iscsi_host_list);
/* Add unlocking code here */

	iscsi_host->dev.bus = &unh_iscsi_bus;
	iscsi_host->dev.parent = &unh_iscsi_pseudo;
	iscsi_host->dev.release = &unh_iscsi_release;
	sprintf(iscsi_host->dev.bus_id, "unh_iscsi%d", 0);

	retval = device_register(&iscsi_host->dev);
	if (unlikely(retval)) {
		printk("add_HBA:device_register () failed %d\n", retval);
		kfree(iscsi_host);
		return retval;
	}

	return retval;
}


/*
 * remove_HBA.
 * We remove the HBA that we had added.
 */
static
void remove_HBA()
{
	struct iscsi_hostdata *iscsi_host = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter remove_HBA \n");
	if (!list_empty(&unh_iscsi_host_list)) {
		iscsi_host = list_entry(unh_iscsi_host_list.prev,
					struct iscsi_hostdata, host_list);
		list_del(&iscsi_host->host_list);
	}
	if (unlikely(iscsi_host == NULL)) {
		printk("iscsi_host is null!!!!, not unregistering\n");
		return;
	}
	device_unregister(&iscsi_host->dev);
}

static int
__init unh_iscsi_init(void)
{
	int err;

	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_init \n");

#if defined(RDMA_MODE)
	{
		extern int iser_init_module(void);

		if (iser_init_module() < 0)
			TRACE_ERROR("iser_init_module failed\n");
	}
#endif /* RDMA_MODE */

	err = device_register(&unh_iscsi_pseudo);
	if (err)
		return err;
	err = bus_register(&unh_iscsi_bus);
	if (err)
		return err;
	err = driver_register(&unh_iscsi_driver);
	if (err)
		return err;

	add_HBA();

	return 0;
}

static void
__exit unh_iscsi_exit(void)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter unh_iscsi_exit \n");
	remove_HBA();
	driver_unregister(&unh_iscsi_driver);
	bus_unregister(&unh_iscsi_bus);
	device_unregister(&unh_iscsi_pseudo);

#if defined(RDMA_MODE)
	{
		extern void iser_cleanup_module(void);

		iser_cleanup_module();
	}
#endif /* RDMA_MODE */
}

device_initcall(unh_iscsi_init);
module_exit(unh_iscsi_exit);

#else				/* K26 hot-plug modifications */

/* Ming Zhang, mingz@ele.uri.edu */
EXPORT_NO_SYMBOLS;
#endif				/* K26 hot-plug modifications */

#endif				/* ifdef __KERNEL__ */
