/*	initiator/initiator_rx.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the functions for iscsi initiator code that are
 *	responsible for reading and processing PDUs sent by the target.
 *
 *	Copyright (C) 2001-2006 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote 
 *	products derived from this software without specific prior 
 * 	written permission.
 */

#include "iscsi_initiator.h"

#ifdef RDMA_MODE
//#include <linux/timex.h>
//#include <linux/delay.h>

static void iser_enqueue_pdu(struct connection *conn_p,
                             struct iser_recv_pdu_t *pdu_p);

#endif


/* Frees passed in buffer to the local memory or releases iSER's buffer */
void free_buffer(struct connection *conn, char **buffer)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;

	if (conn->my_session->oper_param.RDMAExtensions) {
		/* Free iSER buffer */
		if (conn->rx_pdu_front) {
			__u8 *bhs = (__u8 *) conn->rx_pdu_front->p_bhs->buf;
			__u8 *data =
			    (__u8 *) conn->rx_pdu_front->rx_data.p_buf;

			if (((__u8 *) conn->rx_buf != bhs) ||
			    (*buffer && (__u8 *) * buffer != data)) {
				TRACE_ERROR
				    ("free_iser_login_buffer called with incorrect args"
				     "%p %p %p %p\n", conn->rx_buf, bhs,
				     *buffer, data);
				return;
			}

			iser_retval =
			    global_hostdata->iser_api.
			    release_control((iser_conn_handle_t) conn,
					    conn->rx_pdu_front);
			if (iser_retval != ISER_SUCCESS)
				TRACE_ERROR
				    ("iser_release_control returned error %d",
				     iser_retval);

			iser_dequeue_pdu(conn);
			conn->rx_pdu_front = NULL;
		}
	} else
#endif
	{
		/* We are now finished with any locally allocated data buffer
		 * that is not still in use (i.e., being sent in NopOut).
		 */
		if (unlikely(buffer != NULL))
			my_kfree((void **) buffer, "local buf");
	}
}

#ifdef ISER

#ifdef TCP_ISER
int recv_entire_pdu(struct connection *current_connection, char **buffer,
		    int *received);
#endif


/*
 *	Returns > 0 if ok, 0 on err & still hold session->sess_lock,
 *	& for TCP_ISER only - <0 on err and lost session->sess_lock. 
 */
int recv_control_notifications(struct connection *conn,
			       char **buffer, int *received, int take_lock)
{
	int ret;

#ifdef TCP_ISER
	return recv_entire_pdu(conn, buffer, received);
#else
	TRACE(TRACE_ENTER_LEAVE, "Entering recv_control_notifications "
	      "conn=%p\n", conn);
	/* Allow other threads to run while we wait on a sem till 
	 * notifications arrive from either conn_notify or 
	 * conn_terminate_notify 
	 */
	*received = 0;
	if (take_lock)
		UNH_UNLOCK(&conn->my_session->sess_lock,
			   conn->rx_lock_flags);
	ret = down_interruptible(&conn->iser_notify_sem);
	if (take_lock)
		UNH_LOCK(&conn->my_session->sess_lock,
			 conn->rx_lock_flags);
	if(ret != 0) {
		TRACE_ERROR("%s: down_interruptible failed\n", __func__);
		*received = -1;
		goto out_ret;
	}

	if (conn->connection_state == CONNECTION_TERMINATED) {
		TRACE(TRACE_DEBUG, "recv_control_notifications: connection"
		      "terminated\n");
		*received = -1;
		goto out_ret;
	}

	iser_front_pdu(conn, &conn->rx_pdu_front);
	conn->rx_buf = conn->rx_pdu_front->p_bhs->buf;

	/* Other buffer types not handled right now - Minti */
	if (conn->rx_pdu_front->rx_data.type == ISER_BUF_TYPE_SINGLE) {
		*buffer = conn->rx_pdu_front->rx_data.p_buf;
		*received = conn->rx_pdu_front->rx_data.size;
	} else {
		TRACE_ERROR
		    ("Error in recv_control_notifications, invalid buffer type"
		     "%d\n", conn->rx_pdu_front->rx_data.type);
	}

      out_ret:
	TRACE(TRACE_ENTER_LEAVE, "Leaving recv_control_notifications %d\n",
	      *received);
	return *received;
#endif
}

#ifdef RDMA_MODE
iser_status_t connection_establish_notify(iser_conn_handle_t cid_h,
					  int established)
{
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE,
	      "!!!!!!Enter connection_establish_notify %p\n", cid_h);
#if 1

	conn = find_conn_by_iser_conn_handle(cid_h);
	TRACE(TRACE_ISER, "found conn %p\n", conn);

	if (conn == NULL) {
		TRACE_ERROR
		    ("Error in connection_establish_notify, invalid conn handle"
		     "%p\n", cid_h);
		return ISER_FAILURE;
	}

	if (!established)
		conn->connection_state = CONNECTION_TERMINATED;

	TRACE(TRACE_ISER, "before up %d\n", established);

	up(&conn->iser_conn_sem);
#endif
	TRACE(TRACE_ENTER_LEAVE, "Leave connection_establish_notify \n");
	return ISER_SUCCESS;
}

void control_notify(iser_conn_handle_t cid_h,
		    struct iser_recv_pdu_t *p_ctrl_pdu)
{
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter control_notify %p\n", cid_h);

	if ((conn = find_conn_by_iser_conn_handle(cid_h)) == NULL) {
		TRACE_ERROR
		    ("Error in control_notify, invalid conn handle %p\n",
		     cid_h);
		return;
	}

	iser_enqueue_pdu(conn, p_ctrl_pdu);

	up(&conn->iser_notify_sem);
	TRACE(TRACE_ENTER_LEAVE,
	      "Leave control_notify rx_iser_pdu:%p bhs:%p\n", p_ctrl_pdu,
	      p_ctrl_pdu->p_bhs->buf);
	return;
}

void send_notify(iser_conn_handle_t cid_h)
{
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter send_notify\n");

	if ((conn = find_conn_by_iser_conn_handle(cid_h)) == NULL) {
		TRACE_ERROR
		    ("Error in control_notify, invalid conn handle %p\n",
		     cid_h);
		return;
	}
	up(&conn->send_done_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave send_notify\n");
}

void control_terminate_notify(iser_conn_handle_t cid_h)
{
	struct connection *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter control_terminate_notify\n");

	if ((conn = find_conn_by_iser_conn_handle(cid_h)) == NULL) {
		TRACE_ERROR
		    ("Error in control_notify, invalid conn handle %p\n",
		     cid_h);
		return;
	}
	conn->connection_state = CONNECTION_TERMINATED;
	/* These appear to be counting semaphores so there shd be no problem */
	up(&conn->iser_notify_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave control_terminate_notify\n");

	return;
}

void free_iser_login_buffer(struct connection *current_connection,
			    __u32 flags, struct generic_pdu *inpdu)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter free_iser_login_buffer\n");

	if (flags & USE_RDMA_LOGIN) {
		if (current_connection->rx_pdu_front) {
			iser_status_t iser_retval;
			__u8 *bhs =
			    (__u8 *) current_connection->rx_pdu_front->
			    p_bhs->buf;
			__u8 *data =
			    (__u8 *) current_connection->rx_pdu_front->
			    rx_data.p_buf;

			if (((__u8 *) inpdu->hdr != bhs) ||
			    ((__u8 *) inpdu->text != data)) {
				TRACE_ERROR
				    ("free_iser_login_buffer called with incorrect args"
				     "%p %p %p %p\n", inpdu->hdr, bhs,
				     inpdu->text, data);
				return;
			}

			iser_retval =
			    global_hostdata->iser_api.
			    release_control((iser_conn_handle_t)
					    current_connection,
					    current_connection->
					    rx_pdu_front);
			if (iser_retval != ISER_SUCCESS)
				TRACE_ERROR
				    ("iser_release_control returned error %d",
				     iser_retval);

			iser_dequeue_pdu(current_connection);
			current_connection->rx_pdu_front = NULL;
		}
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave free_iser_login_buffer\n");
}
#endif



int iser_alloc_conn_resources(struct connection *conn,
			      struct parameter_type
			      p_param_tbl[MAX_CONFIG_PARAMS])
{
#ifdef RDMA_MODE
	int max_outstanding = 1;
	struct iser_conn_resources_t conn_resources;
	__u32 max_send_length, max_recv_length;
	iser_status_t iser_retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_alloc_conn_resources\n");

	// Gets the lengths based on whether the IRDSL or TRDSL were used
	set_connection_length(p_param_tbl, &max_recv_length,
			      &max_send_length, INITIATOR, 1);

	conn_resources.conn_type = ISER_TRANSPORT_CONN_IB;
	conn_resources.api_h = global_hostdata->iser_handle;
	conn_resources.max_outstand_cmds = max_outstanding_commands();

	/* Not used in the iSER DM code right now
	 * Need to chk that FBL < MRDSL and within bounds if using (i.e. 
	 * do check_integrity_rules etc. first)
	 */
	conn_resources.first_burst_length = 0;
	conn_resources.max_recv_dsl = max_recv_length;

	//Still used in the code although headers claim it needs deleting - fill in 
	conn_resources.max_recv_pdu_sz = max_recv_length;	//Same as max_recv_dsl        

	/* Don't need to pre-register buffers */
	conn_resources.num_reg_buf = 0;
	conn_resources.reg_buf = NULL;

	iser_retval =
	    global_hostdata->iser_api.
	    allocate_connection_resources((iser_conn_handle_t) conn,
					  &conn_resources,
					  &max_outstanding);
	if (iser_retval != ISER_SUCCESS) {
		TRACE_ERROR
		    ("iser_allocate_connection_resources: Failed %d\n",
		     iser_retval);
		return -1;
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_alloc_conn_resources\n");
#endif
	return 0;
}

int iser_dealloc_conn_resources(struct connection *conn)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;
	TRACE(TRACE_ENTER_LEAVE, "Enter iser_dealloc_conn_resources\n");

	iser_retval =
	    global_hostdata->iser_api.
	    deallocate_connection_resources((iser_conn_handle_t) conn);
	if (iser_retval != ISER_SUCCESS) {
		TRACE_ERROR("iser_dealloc_conn_resources: Failed %d\n",
			    iser_retval);
		return -1;
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_dealloc_conn_resources\n");
#endif
	return 0;
}

/* TODO 
 * deallocate_task_resources, 
 * notice_key_values - integrity chks of keys done later than notice_key_values
 * ... - how do we deal with this?
 */


int iser_enable_datamover(struct connection *conn)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_enable_datamover\n");

	iser_retval =
	    global_hostdata->iser_api.
	    enable_datamover((iser_conn_handle_t) conn, NULL, NULL);
	if (iser_retval != ISER_SUCCESS) {
		TRACE_ERROR("iser_enable_datamover: Failed %d\n",
			    iser_retval);
		return -1;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave iser_enable_datamover\n");
#endif
	return 0;
}

void iser_conn_terminate(struct connection *conn)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_conn_terminate\n");

	iser_retval =
	    global_hostdata->iser_api.
	    connection_terminate((iser_conn_handle_t) conn);
	if (iser_retval != ISER_SUCCESS) {
		TRACE_ERROR("iser_conn_terminate: Failed %d\n",
			    iser_retval);
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_conn_terminate\n");
	return;
#endif
}

int iser_send_control(struct connection *conn)
{
	TRACE_ERROR("Not implemented iser_send_control\n");
	return 0;
}

static void iser_enqueue_pdu(struct connection *conn_p,
                             struct iser_recv_pdu_t *pdu_p)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	queue_enqueue(&conn_p->rx_pdu_q, pdu_p);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

void iser_dequeue_pdu(struct connection *conn_p)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	BUG_ON(queue_empty(&conn_p->rx_pdu_q));
	queue_dequeue(&conn_p->rx_pdu_q);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

void iser_front_pdu(struct connection *conn_p,
                    struct iser_recv_pdu_t **pdu_p)
{
	int err;
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	if (queue_empty(&conn_p->rx_pdu_q))
		goto error;

	err = queue_front(&conn_p->rx_pdu_q, (void**) pdu_p);
	if (err)
		goto error;

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return;

 error:
	*pdu_p = NULL;
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s in error\n", __func__);
}

#endif
