/* unh_main_initiator.c */


#define _POSIX_C_SOURCE	200112L

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sched.h>
#include <pthread.h>

#include "../initiator/iscsi_initiator.h"
#include "../initiator/do_scsi_command.h"
#include "../common/param_scan.h"
#include "../common/config_scan.h"
#include "../common/test_scan.h"
#include "../common/user_space.h"
#include "../common/unh_threads.h"
#include "../common/print_time.h"

extern int module_init_init_this_scsi_driver(void);
extern void module_exit_exit_this_scsi_driver(void);

#define BUFSIZE	512

/* maximum number of allowed, nested repeat commands. */
#define MAX_REPEAT_NESTING 4


/* the global key for each thread's private "current" */
pthread_key_t current_key;


/* cleanup handler for the global key,
 * called on the exit of thread having non-null value of its private "current"
 */
extern void free_current(void *current);

/* tokenizes inbuf into at most "limit" tokens, storing pointers to them in
 * argv[].  counting starts from "start", but a
 * token starting with '#' ends the input prematurely.
 * returns final value of "start" or 0 if no tokens or more than "limit" tokens.
 */
int tokenize(char *inbuf, char *argv[], int start, int limit)
{
	char *next;

	if ((next = strtok(inbuf, WHITE_SPACE)) == NULL || *next == '#') {
		/* no tokens at all in the string */
		return 0;
	}

	while (start < limit) {
		argv[start++] = next;
		if ((next = strtok(NULL, WHITE_SPACE)) == NULL
		    || *next == '#')
			return start;
	}

	/* too many tokens, indicate none at all */
	return 0;
}

int process_input(char *inbuf);

/* returns 1 if should exit, 0 otherwise (just end of file) */
int read_command_file(FILE * fptr)
{
	char buffer[BUFSIZE];

	while (fgets(buffer, BUFSIZE, fptr) != NULL) {
		if (process_input(buffer))
			return 1;
	}
	return 0;
}

/* returns 1 if should exit, 0 otherwise (error or ok) */
int process_input(char *inbuf)
{
	static int repv[MAX_REPEAT_NESTING];
	static FILE *fptr = NULL;
	static int repc = 0;
	int argc, i, updown, ret;
	char text_line[4096];
	char *argv[10];
	char *start;
	char *endptr;
	time_type start_time, end_time;

	/*****
	printf("process_input: %s", inbuf);
	*****/

	argv[0] = "iscsi_manage";
	argv[1] = "init";

	/* first, tokenize the inbuf into argv */
	if ((argc = tokenize(inbuf, argv, 2, 10)) <= 2) {
		/* either no tokens or too many tokens */
		return 0;
	}

	printf("%d tokens\n", argc);
	for (i = 0; i < argc; i++)
		printf("%8d: %s\n", i, argv[i]);

	if (strcasecmp(argv[2], "repeat") == 0) {
		if (repc == MAX_REPEAT_NESTING) {
			printf("too many nested repeat commands\n");
			return 0;
		}
		if (argc < 4) {
			printf("invalid repeat command [%s]\n", inbuf);
			return 0;
		} else if (argc == 5) {
			if (repc != 0) {
				printf("an output file may only be "
				       "given on the outter most "
				       "repeat command\n");
				return 0;
			}
			fptr = fopen(argv[4], "a");
			if (!fptr) {
				perror(argv[4]);
				return 0;
			}
		} else if (repc == 0) {
			fptr = stdout;
		}

		repv[repc++] = strtol(argv[3], &endptr, 0);
		if (*endptr != '\0') {
			perror(argv[3]);
			return 0;
		}
		return 0;
	}

	if (strcasecmp(argv[2], "do") == 0) {
		if (repc == 0) {
			printf("using stdout output file for repeat\n");
			fptr = stdout;
		}

		total_data_xfer_len = 0;
		get_time(&start_time);
		if (do_scsi_command(argc - 3, argv + 3, repc, repv)) {
			printf("%s: do_scsi_command failed after\n",
			       __func__);
			ret = 1;
			goto rep_cleanup;
		}
		get_time(&end_time);

		/* repeat output */
		fprintf(fptr, "%-u\t%-20llu\t", repv[0],
			total_data_xfer_len);
		print_time(fptr, &start_time, &end_time, "\n");

		ret = 0;
	rep_cleanup:
		if (fptr != stdout)
			fclose(fptr);
		repc = 0;

		return ret;
	}

	if (repc) {
		printf("warning repeat command given before "
		       "non-'do' command: ignoring the repeat\n");
		if (fptr && fptr != stdout)
			fclose(fptr);
		repc = 0;
	}

	if (strcasecmp(argv[2], "echo") == 0) {
		if (argc < 3) {
			printf("invalid echo command [%s]\n", inbuf);
			return 0;
		}
		fptr = fopen(argv[3], "a");
		if (!fptr) {
			perror(argv[3]);
			return 0;
		} else {
			for (i = 4; i < argc; i++) {
				fprintf(fptr, "%s", argv[i]);
				if (i < argc - 1)
					fprintf(fptr, " ");
			}
			fprintf(fptr, "\n");
			fclose(fptr);
			return 0;
		}
	}

	if (argc == 3) {
		if (strcasecmp(argv[2], "quit") == 0)
			return 1;
		if (strcasecmp(argv[2], "exit") == 0)
			return 1;
	}

	if (argc == 4 && strcasecmp(argv[2], "include") == 0) {
		if ((fptr = fopen(argv[3], "r")) == NULL) {
			perror(argv[3]);
			return 0;
		} else {
			i = read_command_file(fptr);
			fclose(fptr);
			return i;
		}
	}

	/* next, scan the tokens into proper line format */
	if (strcasecmp(argv[2], "test") == 0) {
		argv[2] = "iscsi_test";
		if (test_scan(argc - 2, argv + 2, 0, text_line))
			return 0;
	} else if (strcasecmp(argv[2], "up") == 0
		   || strcasecmp(argv[2], "down") == 0) {
		updown = (strcasecmp(argv[2], "down") == 0);
		argv[1] = "iscsi_config";
		if (config_scan(argc - 1, argv + 1, 0, text_line))
			return 0;
	} else {
		updown = -1;
		if (param_scan(argc, argv, 0, text_line))
			return 0;
	}

	printf("text_line: %s\n", text_line);

	/* next, get the initiator to act on the command line */
	if (iscsi_initiator_proc_info(text_line, &start, 0,
				      strlen(text_line), 0, 1))
		return 0;

	switch (updown) {
	default:		/* manage */
		break;
	case 0:		/* config up */
		break;
	case 1:		/* config down */
		break;
	}			/* switch */

	return 0;
}


int main(void)
{
	int err;
	struct task_struct *current;

	/* first, set up this main thread */
	printf("unh_main_initiator Starting thread id %#08lx\n",
	       pthread_self());
	fflush(stdout);

	/* second create global key for the private-to-each thread "current" */
	if ((err = pthread_key_create(&current_key, free_current)) != 0) {
		fprintf(stderr, "pthread_key_create current_key: %s\n",
			strerror(err));
		exit(EXIT_FAILURE);
	}

	/* third give this thread its private copy of "current" and init it */
	unh_daemonize("unh_main_initiator");

#if defined(RDMA_MODE)
	{
		extern int iser_init_module(void);

		if (iser_init_module() < 0)
			TRACE_ERROR("iser_init_module failed");
	}
#endif				/* RDMA_MODE */

	/* fourth, initialize the iSCSI initiator module */
	printf("unh_main_initiator calling "
	       "module_init_init_this_scsi_driver\n");
	module_init_init_this_scsi_driver();
	fflush(stdout);
	sched_yield();

	/* fifth, go into loop (until read EOF) to allow initiator to function */
	read_command_file(stdin);

	/* sixth, clean up the iSCSI initiator module */
	printf("unh_main_initiator calling "
	       "module_exit_exit_this_scsi_driver\n");
	fflush(stdout);
	module_exit_exit_this_scsi_driver();
	fflush(stdout);
	sched_yield();

#if defined(RDMA_MODE)
	{
		extern void iser_cleanup_module(void);

		iser_cleanup_module();
	}
#endif				/* RDMA_MODE */

	/* seventh, free this thread's private "current" structure */
	free_current(current);
	if ((err = pthread_setspecific(current_key, current)) != 0) {
		fprintf(stderr, "pthread_setspecific current_key: %s\n",
			strerror(err));
		exit(EXIT_FAILURE);
	}


	/* eighth, delete global key for the private-to-each thread "current" */
	if ((err = pthread_key_delete(current_key)) != 0) {
		fprintf(stderr, "pthread_key_delete current_key: %s\n",
			strerror(err));
	}

	/* nineth, exit this thread (should be only one left running) */
	printf("unh_main_initiator Exiting thread id %#08lx\n",
	       pthread_self());
	fflush(stdout);
	pthread_exit((void *) 0);
}
