/*
 * target/iscsi_cmnd.c -- Contains function declarations for the iscsi_cmnd
 * 			  structure.
 *
 * Created Mon Feb 26 19:45:53 2007 -- Ethan Burns
 */

#include "iscsi_cmnd.h"
#include "target_error_rec.h"


const char *cmnd_state_name(__u32 state)
{
	static const char *cmnd_state_names[] = {
		"<zero>",		/* 0 */
		"ISCSI_CMND_RECEIVED",
		"ISCSI_NEW_CMND",
		"ISCSI_BUFFER_RDY",
		"ISCSI_DONE",
		"ISCSI_SENT",		/* 5 */
		"ISCSI_DEQUEUE",
		"ISCSI_AWAIT_BUFFER",
		"ISCSI_AWAIT_BUFFER_QUEUE",
		"ISCSI_UNSOLICITED_DATA_IN",
		"ISCSI_DATA_IN",	/* 10 */
		"ISCSI_MGT_FN_DONE",
		"ISCSI_SEND_TEXT_RESPONSE",
		"ISCSI_LOGOUT",
		"ISCSI_PING",
		"ISCSI_QUEUE_CMND",	/* 15 */
		"ISCSI_QUEUE_CMND_RDY",
		"ISCSI_QUEUE_OTHER",
		"ISCSI_NOPIN_SENT",
		"ISCSI_RESEND_STATUS",
		"ISCSI_ASK_FOR_MORE_TEXT", /* 20 */
		"ISCSI_AWAIT_MORE_TEXT",
		"ISCSI_BLOCKED_SENDING_TEXT",
		"ISCSI_AWAIT_DATA_ACK",
		"ISCSI_GOT_DATA_ACK",
		"ISCSI_REDO_DATA",	/* 25 */
		"ISCSI_RESEND_TEXT_RESPONSE",
		"ISCSI_RESEND_NOP_IN"
	};

	if (state > 0 && state <= ISCSI_CMND_STATE_MAX)
		return cmnd_state_names[state];

	return "<unknown state>";
}

/*	RFC 3720 Section 3.2.2.1 Command Numbering and Acknowledging
 *	...
 *	For non-immediate commands, the CmdSN field can take any value from
 *	ExpCmdSN to MaxCmdSN inclusive.  The target MUST silently ignore
 *	any non-immediate command outside of this range or non-immediate
 *	duplicates within the range.  The CmdSN carried by immediate commands
 *	may lie outside the ExpCmdSN to MaxCmdSN range.
 *
 *	check that cmnd is immediate (I bit is set) or cmnd's cmd_sn is within
 *  range [ExpCmdSN .. MaxCmdSN] for this session
 *	returns 0 if equal to exp_cmd_sn (so this is next command to perform)
 *		1 if greater than exp_cmd_sn but in range (should be delayed)
 *		-1 if outside the range (and hence should be ignored)
 *		0 for all immediate commands, regardless of cmd_sn value
 *  in all cases, neither exp_cmd_sn nor max_cmd_sn for session are incremented
 *  at this time, but fields are set in the cmnd for later incrementing
 *  when called, session->cmnd_lock must be held so access to session-wide
 *  max_cmd_sn and exp_cmd_sn fields is atomic.
 */
int inline __attribute__ ((no_instrument_function))
    check_cmd_sn(struct iscsi_cmnd *cmnd, void *ptr,
	     struct iscsi_session *session)
{
	struct pdu_hdr *pdu = ptr;
	int delta;

	if (!(pdu->opcode & I_BIT)) {
		/* this is an non-immediate command */
		delta = session->max_cmd_sn - pdu->cmd_sn;
		if (delta < 0) {	/* cmd_sn is greater than MaxCmdSN */
			return -1;
		}

		delta = pdu->cmd_sn - session->exp_cmd_sn;
		if (delta < 0) {	/* cmd_sn is less than ExpCmdSN */
			return -1;
		}

		if (delta > 0) {	/* cmd_sn is in range, but out of order */
			return 1;
		}

		/* non-immediate command is in expected order, so it is the next
		 * command to perform -- fall thru to return 0
		 */
	}

	return 0;
}

/* called only when a session is being released,
 * so there should be no active commands or connections in this session
 */
void release_free_commands(struct iscsi_session *session)
{
	struct list_head *list_ptr, *list_temp;
	struct iscsi_cmnd *cmnd;
	__u32 count;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	count = 0;
	list_for_each_safe(list_ptr, list_temp,
			   &session->free_command_list) {
		cmnd =
		    list_entry(list_ptr, struct iscsi_cmnd, command_link);
		count++;
		TEST_LOOP_COUNTER(count, "free_command_list");
		list_del(list_ptr);
		TRACE(TRACE_BUF, "%s %2u: release command %p\n",
		      current->comm, count, cmnd);
		my_kfree((void **) &cmnd, "command");
	}

	if (count == session->command_list_size) {
		if (count) {
			TRACE(TRACE_BUF, "%s release %u iscsi commands\n",
			      current->comm, count);
		}
	} else {
		TRACE_WARNING
		    ("%s release %u iscsi commands but %u allocated\n",
		     current->comm, count, session->command_list_size);
	}
	session->command_list_size = 0;
}

struct iscsi_cmnd *alloc_one_free_command(struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	cmnd = my_kmalloc(sizeof(struct iscsi_cmnd), "command");
	if (!cmnd)
		return NULL;
	session->command_list_size++;
	TRACE(TRACE_BUF, "%s allocate iscsi command %p, now have %u\n",
	      current->comm, cmnd, session->command_list_size);
	return cmnd;
}

/* called only from handle_login after login negotiation for new session
 * to allocate session-wide command structures
 * so there should be no active commands or connections in this session
 * allocate as many as possible up to computed maximum
 */
void setup_free_commands(struct iscsi_session *session)
{
	__u32 ncmnds, i;
	struct iscsi_cmnd *cmnd;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	ncmnds = (session->max_queue_depth_allowed + 1)
	    * session->oper_param->MaxConnections;

	/* allocate commands and thread them onto session's list */
	for (i = 0; i < ncmnds; i++) {
		cmnd = alloc_one_free_command(session);
		if (!cmnd)
			goto out;
		list_add_tail(&cmnd->command_link,
			      &session->free_command_list);
	}
      out:
	TRACE(TRACE_BUF,
	      "%s allocate %u iscsi commands of %u bytes each\n",
	      current->comm, session->command_list_size,
	      (int) sizeof(struct iscsi_cmnd));
}

/* return first free command from session's free command list
 * (and remove it from the list)
 * also initialize fields in the command, and check for cmd_sn ordering
 * return NULL if list is empty
 */
struct iscsi_cmnd *get_free_command(__u8 * buffer,
				    struct iscsi_session *session,
				    int *err)
{
	struct iscsi_cmnd *cmnd = NULL;
	struct pdu_hdr *pdu;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	spin_lock_irqsave(&session->cmnd_lock, flags);

	if (list_empty(&session->free_command_list)) {
		cmnd = alloc_one_free_command(session);
		if (cmnd == NULL)
			goto out_unlock;
/*
		TRACE_WARNING("%s allocate additional iscsi command, "
			      "now have %u\n", current->comm,
			      session->command_list_size);
*/
	} else {
		cmnd = list_entry(session->free_command_list.next,
				  struct iscsi_cmnd, command_link);
		list_del(&cmnd->command_link);
		TRACE(TRACE_DEBUG, "%s get free command %p\n",
		      current->comm, cmnd);
	}

	/* zero out just the first part of the command structure */
	memset(cmnd, 0, (long) ((struct iscsi_cmnd *) NULL)->tx_iov);

	/* Initialize the command's magic numbers for debugging
	 * clobbered commands */
	cmnd_init_magic(cmnd);

	/* assign a non-zero "unique" sequence number to every active command */
	if (!(cmnd->active_seqno = session->active_seqno++))
		cmnd->active_seqno = session->active_seqno++;

	INIT_LIST_HEAD(&cmnd->command_link);
	INIT_LIST_HEAD(&cmnd->unsolicited_data_list);
	cmnd->state = ISCSI_NEW_CMND;
	cmnd->session = session;
	if ((pdu = (struct pdu_hdr *) buffer)) {
		/* save fields that are common to all PDUs sent by initiator */
		cmnd->opcode_byte = pdu->opcode;
		cmnd->command_flags = pdu->flags;
		cmnd->init_task_tag = pdu->init_task_tag
		    = be32_to_cpu(pdu->init_task_tag);
		cmnd->cmd_sn = pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
		cmnd->stat_sn = pdu->exp_stat_sn
		    = be32_to_cpu(pdu->exp_stat_sn);
	}

	if (err) {
		/* check cmd_sn for command ordering */
		*err = check_cmd_sn(cmnd, pdu, session);
	} else {
		/* generate the next TTT, in case we need it */
		cmnd->target_xfer_tag = generate_next_ttt(session);
	}

      out_unlock:
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
	return cmnd;
}

void put_free_command(struct iscsi_cmnd *cmnd,
		      struct iscsi_session *session)
{
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "%s put free command %p\n", current->comm,
	      cmnd);
	spin_lock_irqsave(&session->cmnd_lock, flags);
	list_add_tail(&cmnd->command_link, &session->free_command_list);
	cmnd->active_seqno = 0;
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
}

/*
 * called only from iscsi_dequeue()
 * frees all cmnd's structures and resets its active sequence number
 * (but does not free the cmnd itself, since it is going onto free list next)
 */
void __attribute__ ((no_instrument_function))
    release_command(struct iscsi_cmnd *cmnd)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "%s release_command %p CmdSN %u ITT %u\n",
	      current->comm, cmnd, cmnd->cmd_sn, cmnd->init_task_tag);

	/* Free the R2T cookie if any - SAI */
	free_r2t_cookie(cmnd);
	/* free udata list if any, cdeng */
	free_udata_list(cmnd);
	my_kfree((void **) &cmnd->in_progress_buffer, "text buf");
	my_kfree((void **) &cmnd->ping_data, "data_buf");
	cmnd->active_seqno = 0;
}


void cmnd_print(const struct iscsi_cmnd *cmnd)
{
#if defined(__KERNEL__)
	printk(
#else
	fprintf(stderr,
#endif /* __KERNEL__ */
		"----------------------------------------\n"
		"%20s = %p\n"	/* command */
		"%20s = %u\n"	/* init_task_tag */
		"%20s = %u\n"	/* target_xfer_tag */
		"%20s = %u\n"	/* cmd_sn */
		"%20s = %u\n"	/* stat_sn */
		"%20s = %p\n"	/* conn */
		"%20s = %p\n"	/* session */
		"%20s = %u (%s)\n"	/* state */
		"%20s = %u\n"	/* active_seqno */
		"%20s = %p\n"	/* (scsi) cmnd */
		"%20s = %u\n"	/* (scsi) cmnd->state */
		"%20s = %u\n"	/* opcode_byte */
		"%20s = %u\n"	/* data_length */
		"%20s = %u\n"	/* data_done */
		"----------------------------------------\n",
		"command", cmnd,
		"init_task_tag", cmnd->init_task_tag,
		"target_xfer_tag", cmnd->target_xfer_tag,
		"cmd_sn", cmnd->cmd_sn,
		"stat_sn", cmnd->stat_sn,
		"conn", cmnd->conn,
		"session", cmnd->session,
		"state", cmnd->state, cmnd_state_name(cmnd->state),
		"active_seqno", cmnd->active_seqno,
		"(scsi) cmnd", cmnd->cmnd,
		"(scsi) cmnd->state",
		cmnd->cmnd ? cmnd->cmnd->state : -1,
		"opcode_byte", cmnd->opcode_byte,
		"data_length", cmnd->data_length,
		"data_done", cmnd->data_done);
}


#if defined(CMND_CLOBBER_DEBUG)
static spinlock_t cmnd_check_lock = SPIN_LOCK_UNLOCKED;

void cmnd_check_magic(struct iscsi_session *sess, struct iscsi_cmnd * cmnd)
{
	long flags;
	struct iscsi_cmnd *c;
	struct list_head *node;


	if (unlikely(cmnd->magic0 != CMND_MAGIC
		     || cmnd->magic1 != CMND_MAGIC
		     || cmnd->magic2 != CMND_MAGIC)) {
		spin_lock_irqsave(&cmnd_check_lock, flags);
		TRACE_ERROR("%s: bad command magic numbers:\n"
			    "\tCMND_MAGIC=%#lx (%lu)\n"
			    "\t    magic0=%#lx (%lu)\n"
			    "\t    magic1=%#lx (%lu)\n"
			    "\t    magic2=%#lx (%lu)\n",
			    __func__,
			    CMND_MAGIC, CMND_MAGIC,
			    cmnd->magic0, cmnd->magic0,
			    cmnd->magic1, cmnd->magic1,
			    cmnd->magic2, cmnd->magic2);

		cmnd_print(cmnd);

		TRACE_ERROR("%s: session=%p active commands:\n", __func__, sess);
		list_for_each(node, &sess->active_command_list) {
			c = list_entry(node, struct iscsi_cmnd, command_link);
			cmnd_print(c);
		}

		spin_unlock_irqrestore(&cmnd_check_lock, flags);

		/* doesn't something already do this? */
#if defined(__KERNEL__)
		panic("cmnd_check_magic failed");
#else
		abort();
#endif /* __KERNEL__ */
	}
}
#endif /* CMND_CLOBBER_DEBUG */

/* vi: set tabstop=8 textwidth=78: */
