/*
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * target/iscsi_cmnd.h -- Contains declarations for the iscsi_cmnd structure.
 *
 * Created Mon Feb 26 19:35:40 2007 -- Ethan Burns
 */
#ifndef _ISCSI_CMND_H_
#define _ISCSI_CMND_H_

/* debug clobbered iscsi_cmnd structures using magic numbers. */
#define CMND_CLOBBER_DEBUG

#if defined(CMND_CLOBBER_DEBUG)
#define CMND_MAGIC 0xEABEABEBUL
#endif /* CMND_CLOBBER_DEBUG */

#include "iscsi_target.h"

#if defined(__KERNEL__)
#include <linux/list.h>
#else
#include "../common/user_list.h"
#endif				/* __KERNEL__ */

#include "../common/iscsi_common.h"
#include "../common/range.h"

/* values for "state" in struct iscsi_cmnd */
#define ISCSI_CMND_RECEIVED			1
#define ISCSI_NEW_CMND				2
#define ISCSI_BUFFER_RDY			3
#define ISCSI_DONE				4
#define ISCSI_SENT				5
#define ISCSI_DEQUEUE				6
#define ISCSI_AWAIT_BUFFER			7
#define ISCSI_AWAIT_BUFFER_QUEUE		8
#define ISCSI_UNSOLICITED_DATA_IN		9	/* unused */
#define ISCSI_DATA_IN				10
#define ISCSI_MGT_FN_DONE			11
#define ISCSI_SEND_TEXT_RESPONSE		12
#define ISCSI_LOGOUT				13
#define ISCSI_PING				14
#define ISCSI_QUEUE_CMND			15
#define ISCSI_QUEUE_CMND_RDY			16
#define ISCSI_QUEUE_OTHER			17
#define ISCSI_NOPIN_SENT			18
#define ISCSI_RESEND_STATUS			19
#define ISCSI_ASK_FOR_MORE_TEXT			20
#define ISCSI_AWAIT_MORE_TEXT			21
#define ISCSI_BLOCKED_SENDING_TEXT		22
#define ISCSI_AWAIT_DATA_ACK			23
#define ISCSI_GOT_DATA_ACK			24
#define ISCSI_REDO_DATA				25
#define ISCSI_RESEND_TEXT_RESPONSE		26
#define ISCSI_RESEND_NOP_IN			27

#define ISCSI_CMND_STATE_MAX                    27
/***
 * UPDATE THE cmdn_state_names array in iscsi_cmnd.c when adding a
 * new state
 ***/


/* bits that can be set in cmd->seq_flags */
#define UNDERFLOW_FLAG	0x0001
#define OVERFLOW_FLAG	0x0002
#define SEND_SENSE_FLAG	0x0004
#define LAST_SEQ_FLAG	0x0010

/* stores everything related to a SCSI command received */
struct iscsi_cmnd {
#if defined(CMND_CLOBBER_DEBUG)
	unsigned long magic0;
#endif

	/* command_link: thread for session's free and active command lists */
	struct list_head command_link;

	/* conn: connection on which this command was received */
	struct iscsi_conn *conn;

	/* session: session on which this command was received */
	struct iscsi_session *session;

	/* state: execution state of the command (see defines above) */
	__u8 state;

	/* response: task_mgt_function response */
	__u8 response;

	/* ref_function: Referenced function for task management */
	__u8 ref_function;

	/* flags from command pdu */
	__u8 command_flags;

	/* active_seqno: id assigned to each active command, reset when free */
	__u32 active_seqno;

	/* ping_data: data get from nopout and send for nopin */
	char *ping_data;

	/* in_progress_buffer: accumulated data during text negotiations */
	char *in_progress_buffer;

	/* cmnd: cmnd corresponding to this struct */
	Target_Scsi_Cmnd *cmnd;

	/* message: message corresponding to this struct */
	Target_Scsi_Message *message;

	/* set to 1 if max_cmd_sn needs to be incremented by this command else 0 */
	__u8 need_max_cmd_sn_inc;

	/* unsolicited_data_present: */
	__u8 unsolicited_data_present;

	/* Data PDU re-transmission on Data SNACK - SAI */
	__u8 retransmit_flg;

	/* opcode_byte: command's opcode byte (including I bit) */
	__u8 opcode_byte;

	/* data_length: total number of bytes of data to be transferred by
	 * command */
	__u32 data_length;

	/* first_burst_len: data associated with the command */
	__u32 first_burst_len;

	/* first_solicited_offset: offset to first solicited data byte in
	 * command */
	__u32 first_solicited_offset;

	/* max_burst_len: data associated with one R2T */
	__u32 next_burst_len;

	/* data_done: number of bytes of data transferred so far in a command */
	__u32 data_done;

	/* r2t_data_total: total amount of data to be solicited by r2ts */
	int r2t_data_total;

	/* r2t_data_left: data length left for the next r2t -- can go
	 * negative! */
	int r2t_data_left;

	/* scatter_list_offset:when processing a WRITE command, is byte offset
	 * in  current sg list item of where to start storing next  data byte
	 * from a DataOut pdu if pdus, seqs are in order when processing a
	 * READ command, is byte offset in current sg list iemt of where to
	 * start sending next data byte in a DataOut pdu if pdus, seqs are
	 * in order
	 */
	__u32 scatter_list_offset;

	/* scatter_entry_count:when processing a WRITE command, is index into
	 * current sg list of item of where to start storing next
	 * data byte from a DataOut pdu if pdus, seqs are in order
	 * when processing a READ command, is byte offset in
	 * current sg list item of where to start sending next
	 * data byte in a DataOut pdu if pdus, seqs are in order
	 */
	__u32 scatter_list_count;

#if defined(CMND_CLOBBER_DEBUG)
	unsigned long magic1;
#endif /* CMND_CLOBBER_DEBUG */

	/* init_task_tag: Initiator Tag */
	__u32 init_task_tag;

	/* target_xfer_tag: Target Assigned Tag */
	__u32 target_xfer_tag;

	/* ref_task_tag: Referenced Task Tag used when aborting tasks */
	__u32 ref_task_tag;

	/* ref_cmd_sn: Referenced CmdSN used when aborting tasks */
	__u32 ref_cmd_sn;

	/* ref_active_seqno: Referenced active_seqno used when aborting tasks */
	__u32 ref_active_seqno;

	/* ref_command: Referenced command used when aborting luns */
	struct iscsi_cmnd *ref_command;

	/* lun and cdb: as their names imply */
	__u64 lun;
	__u8 cdb[16];

	/* cmd_sn of this command and stat_sn of this command response */
	__u32 cmd_sn;
	__u32 stat_sn;

	/* used by SCSI Command pdus only */
	__u32 cmd_sn_increment;

	/* Command WIDE COUNTERS: data_sn and r2t_sn */
	__u32 data_sn;
	__u32 r2t_sn;
	__u32 outstanding_r2t;

	/* for out-of-order sequence and DataOut PDUs */
	struct order_range seq_range_list;
	struct order_range pdu_range_list;

	/* for out-of-order command */
	__u8 hdr[ISCSI_HDR_LEN];
	struct list_head unsolicited_data_list;

	/* Added r2t cookie details - SAI */
	struct iscsi_cookie *first_r2t_cookie;
	struct iscsi_cookie *last_r2t_cookie;

	/* queue out-of-order data pdus - SAI */
	struct iscsi_cookie *first_data_q;
	struct iscsi_cookie *last_data_q;

	/* Added timestamp for r2t retransmissions - SAI */
	int timestamp;

	/* Store the retransmission limits for data_sn set by a SNACK,
	 * used when retransmit_flg set */
	__u32 startsn, endsn;

	/* following 5 values used to process sequences of DataIn pdus during
	 * READ */

	/* total number of bytes remaining to be sent in following sequences */
	__u32 data_length_left;

	/* offset to lowest byte in next sequence */
	__u32 seq_low_byte;

	/* offset to highest byte plus one in next sequence */
	__u32 seq_limit_byte;

	/* total number of bytes in this sequence */
	__u32 seq_length;

	/* flags used to determine end-of-sequence stuff */
	__u8 seq_flags;

	/* saved state of command when a SNACK is received for last Text
	 * Response or NopIn and the command needs to be retransmitted
	 */
	__u8 reply_state;

	/* flag for errorRecovery of r2t */
	__u8 recovery_r2t;

	/* dummy to keep __u8 fields in groups of 4 */
	__u8 dummy_0;

	/* time at which this command was last sent */
	unsigned long time_stamp;

	/* the following 6 values are saved at the start of every DataIn
	 * sequence when the A-bit will be set at the end of intermediate
	 * sequences, and restored if a Data/R2T SNACK requesting
	 * retransmission of DataIn pdus from that sequence is received at
	 * the end of the sequence.  When the A-bit is not set at the end
	 * of intermediate sequences, these 6 values are set to the initial
	 * values at the start of the command, and are restored if a Data/R2T
	 * SNACK requests retransmission of DataIn pdus from anywhere in
	 * the command, regardless of the sequence.
	 */

	/* data_sn at start of this sequence */
	__u32 start_data_sn;

	/* data_length_left at start of this sequence */
	__u32 start_data_length_left;

	/* seq_low_byte at start of this sequence */
	__u32 start_low_byte;

	/* seq_limit_byte at start of this sequence */
	__u32 start_limit_byte;

	/* scatter_list_count at start of this sequence */
	__u32 start_sg_list_count;

	/* scatter_list_offset at start of this sequence */
	__u32 start_sg_list_offset;

	/* header generated for last Text Response or NopIn (so we can find it
	 * in case it is needed for a retransmission triggered by a
	 * STATUS SNACK)
	 */
	__u8 reply_hdr[ISCSI_HDR_LEN];

	/* pointer to data segment for last Text Response (so we can find it
	 * in case it is needed for a retransmission triggered by a
	 * STATUS SNACK)
	 */
	void *reply_data;

	/* number of bytes in data segment for last Text Response (so we can
	 * find it in case it is needed for a retransmission triggered by a
	 * STATUS SNACK)
	 */
	__u32 reply_size;

	/* new MaxRecvDataSegmentLength from initiator in FFP Text
	 * negotiation */
	__u32 new_max_send_length;

#ifdef RDMA_MODE
	struct iser_recv_pdu_t *rx_iser_pdu;	/* Points to rcvd iSER PDU */
	void *iser_task_h;	/* iSER task handle */
#endif

	/* structures from this point on are NOT zeroed on command setup!
	 * Therefore, do NOT move the position of "tx_iov" unless you change
	 * the memset() in get_free_command()!
	 */

	/* place to build io vector for sendmsg */
	struct iovec tx_iov[MAX_IOV_SLOTS];
	struct iovec tx_iov_copy[MAX_IOV_SLOTS];

	/* place to build io vector for recvmsg */
	struct iovec rx_iov[MAX_IOV_SLOTS];
	struct iovec rx_iov_copy[MAX_IOV_SLOTS];

#if defined(CMND_CLOBBER_DEBUG)
	unsigned long magic2;
#endif /* CMND_CLOBBER_DEBUG */
};

extern int __attribute__ ((no_instrument_function))
    check_cmd_sn(struct iscsi_cmnd *cmnd, void *ptr,
	     struct iscsi_session *session);

extern void release_free_commands(struct iscsi_session *session);

extern struct iscsi_cmnd *alloc_one_free_command(struct iscsi_session
						 *session);

extern void setup_free_commands(struct iscsi_session *session);

extern struct iscsi_cmnd *get_free_command(__u8 * buffer,
					   struct iscsi_session *session,
					   int *err);

extern void
put_free_command(struct iscsi_cmnd *cmnd, struct iscsi_session *session);

extern void __attribute__ ((no_instrument_function))
    release_command(struct iscsi_cmnd *cmnd);

/* prints a command */
void cmnd_print(const struct iscsi_cmnd *cmnd);
/* get a command state name */
const char *cmnd_state_name(__u32 state);

#if defined(CMND_CLOBBER_DEBUG)

void cmnd_check_magic(struct iscsi_session *sess, struct iscsi_cmnd * cmnd);

static inline void cmnd_init_magic(struct iscsi_cmnd *cmnd)
{
	cmnd->magic0 = CMND_MAGIC;
	cmnd->magic1 = CMND_MAGIC;
	cmnd->magic2 = CMND_MAGIC;
}

#else  /* !CMND_CLOBBER_DEBUG */

#define cmnd_check_magic(x, y) /* */
#define cmnd_init_magic(x) /* */

#endif /* CMND_CLOBBER_DEBUG */

#endif				/* !_ISCSI_CMND_H_ */
/* vi: set tabstop=8 textwidth=78: */
