/*
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * target/iscsi_rx.c -- Contains function definitions used in the
 * iscsi_rx_thread() function and the iscsi_rx_thread() function itself.
 *
 * Created Fri 02 Mar 2007 12:25:44 PM EST -- Ethan Burns
 */

#include "iscsi_target.h"
#include "iscsi_cmnd.h"
#include "target_error_rec.h"
#include "../common/unh_threads.h"

#ifndef __KERNEL__
#include "../common/user_signal.h"
#endif				/* __KERNEL__ */

/*
 * Static declarations
 */
static int handle_login(struct iscsi_conn *conn,
			struct generic_pdu *inputpdu);

static int handle_text_request(struct iscsi_conn *conn,
			       struct iscsi_session *session,
			       __u8 * buffer, __u8 ** data_buffer);

static int handle_logout(struct iscsi_conn *conn,
			 struct iscsi_session *session,
			 __u8 * buffer, __u8 ** data_buffer);

static int handle_nopout(struct iscsi_conn *conn,
			 struct iscsi_session *session,
			 __u8 * buffer, __u8 ** data_buffer);

static int handle_cmnd(struct iscsi_conn *conn,
		       struct iscsi_session *session,
		       __u8 * buffer, __u8 * data_buffer);

static int handle_task_mgt_command(struct iscsi_conn *conn,
				   struct iscsi_session *session,
				   __u8 * buffer, __u8 ** data_buffer);

static int handle_data(struct iscsi_conn *conn,
		       struct iscsi_session *session,
		       __u8 * buffer, __u8 * data_buffer);

static int handle_snack(struct iscsi_conn *conn,
			struct iscsi_session *session,
			__u8 * buffer, __u8 * data_buffer);

static int handle_iscsi_rx_data(struct iscsi_conn *conn, __u8 ** pdu_bhs,
				__u8 ** data_buffer);

static void setup_unsolicited_buffers(struct iscsi_session *session);

static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_by_ttt(struct iscsi_conn *conn, struct iscsi_session *session,
	      __u32 this_ttt);

static int read_list_data_seg(struct pdu_hdr *hdr, struct iscsi_cmnd *cmd,
			      struct scatterlist *st_list, __u32 offset,
			      __u8 * data_buffer);

static int __attribute__ ((no_instrument_function))
    do_iscsi_rx_data(struct iscsi_conn *conn, __u8 ** pdu_bhs,
		 __u8 ** data_buffer);

static int save_unsolicited_data(struct iscsi_cmnd *cmnd, __u32 offset,
				 struct pdu_hdr *hdr, __u8 * data_buf,
				 void *iser_buf);

static int out_of_order_cmnd(struct iscsi_conn *conn,
			     struct iscsi_session *session, __u8 * buffer,
			     struct iscsi_cmnd *cmnd, int err,
			     __u8 * data_buffer);

static void __attribute__ ((no_instrument_function))
    merge_out_of_order(struct iscsi_init_scsi_data_out *hdr,
		   struct iscsi_cmnd *cmd);

static void __attribute__ ((no_instrument_function))
    restore_data_in_seq(struct iscsi_cmnd *cmd);

static struct udata_list *get_unsolicited_buffer(struct iscsi_session
						 *session);

static void __attribute__ ((no_instrument_function))
    update_after_read(struct pdu_hdr *hdr, struct iscsi_cmnd *cmnd);

static int
read_single_data_seg(__u8 * buffer, struct iscsi_conn *conn,
		     int bufsize, __u8 * existing_buf, __u8 ** result);

static int ack_sent_cmnds(struct iscsi_conn *conn,
			  struct iscsi_cmnd *cmnd, __u32 exp_stat_sn,
			  int add_cmnd_to_queue);

static struct udata_list *alloc_one_unsolicited_buffer(struct iscsi_session
						       *session);

/**
 * Performs the bottom-half of the rx_thread cleanup.
 * This is broken up into two proceedures since sometimes the top half is not
 * suppose to execute.
 */
static void rx_thread_out(struct iscsi_conn *conn)
{
#ifndef __KERNEL__
	struct task_struct *current;
	current = pthread_getspecific(current_key);
#endif
	conn->active = 0;
	conn->rx_thread = NULL;
	if (conn->tx_thread) {
		/* tx thread still going, shut it down */
#ifndef __KERNEL__
		TRACE(TRACE_ISCSI,
		      "sending SIGKILL (%d) to thread id %#08lx\n",
		      SIGKILL, conn->tx_thread->threadid);
#endif				/* __KERNEL__ */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->tx_thread, 1);

#ifndef __KERNEL__
		/* in user space, send_sig() yields to tx_thread's cleanup handler */
		conn->tx_thread = NULL;	/* so tx_thread is truly dead now */
#endif

	} else {
		/* tx thread has shut down, if socket is still open, release it */
		iscsi_release_socket(conn);
	}
	up(&conn->kill_rx_sem);
#ifndef __KERNEL__
	printk("%s Exiting thread id %#08lx\n", current->comm,
	       current->threadid);
#endif				/* __KERNEL__ */
}

/**
 * Holds the data that needs to be cleaned up by the rx_shutdown_thread()
 * function.
 */
struct rx_thread_cleanup_data {
	struct iscsi_conn *conn;
	struct generic_pdu *inputpdu;
	__u8 **data_buffer;
	__u8 **buffer;
};

/**
 * Performs the top-half of the rx_thread cleanup, and calls the bottom-half
 * routine: rx_thread_out().
 */
static void rx_shutdown_thread(void *aux)
{
	struct rx_thread_cleanup_data *data;

	data = aux;

	TRACE(TRACE_DEBUG, "free hdr%p text%p \n", data->inputpdu->hdr,
	      data->inputpdu->text);
	if (!(data->conn->connection_flags & USE_RDMA_LOGIN)) {
		login_pdu_free(data->inputpdu, "Input pdu");
	}
	free_targ_buf(data->conn, data->buffer, data->data_buffer);
	TRACE(TRACE_ISER, "After free_targ_buf\n");
}

/*
 * iscsi_rx_thread: This thread is responsible for receiving iSCSI PDUs
 * and messages from the Initiator and then figuring out what to do with
 * them
 */
int iscsi_rx_thread(void *param)
{
	struct iscsi_conn *conn = (struct iscsi_conn *) param;
	/* read 48-byte PDU header in for TCP mode */
	__u8 pdu_bhs[ISCSI_HDR_LEN];
	/* This points to pdu_bhs or to iSER's mem */
	__u8 *buffer = pdu_bhs;
	/* extract 6-bit PDU opcode into here */
	__u32 opcode;
	/* convert 4-byte PDU ITT into here */
	__u32 local_itt;
	__u8 *data_buffer = NULL;
	struct generic_pdu inputpdu;
	int free_pdu = 0;
	struct rx_thread_cleanup_data cleanup_data;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	lock_kernel();

	unh_daemonize("iscsi_rx_%d", conn->conn_id);

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this rx thread is alive */
	conn->rx_thread = current;

	unlock_kernel();

#ifndef __KERNEL__
	printk("%s Starting thread id %#08lx\n", current->comm,
	       current->threadid);
#endif				/* __KERNEL__ */

	report_peer_up_down(conn->ip_address, "connected to");

	if (!(conn->connection_flags & USE_RDMA_LOGIN)) {
		if (login_pdu_malloc(&inputpdu, "Input pdu") < 0)
			goto out;
		TRACE(TRACE_DEBUG, "malloc hdr %p \n", inputpdu.hdr);
	}


	/* receive loop */
	while (1) {
		/* Minti */
		if (!(conn->session->tsih)) {

			if (handle_recv_msg(conn, &inputpdu) < 0)
				goto iscsi_rx_thread_out;

			TRACE_BUFFER(TRACE_BUF, inputpdu.hdr,
				     ISCSI_HDR_LEN, "%s Got PDU header\n",
				     current->comm);

			opcode = inputpdu.hdr->opcode & ISCSI_OPCODE;

			local_itt = ntohl(inputpdu.hdr->init_task_tag);

			/*
			 * not in full feature phase yet, only accept Login
			 * request pdus
			 */
			if (opcode != ISCSI_INIT_LOGIN_CMND) {
				TRACE_ERROR("%s ITT %u has opcode 0x%02x, "
					    "expected 0x%02x\n",
					    current->comm, local_itt,
					    opcode, ISCSI_INIT_LOGIN_CMND);
				/* MUST immediately terminate the connection */
				goto iscsi_rx_thread_out;
			}

			TRACE(TRACE_ISCSI,
			      "%lu %s Got Login request ITT %u\n", jiffies,
			      current->comm, local_itt);
			if (handle_login(conn, &inputpdu) < 0) {
				TRACE_ERROR("%s Trouble in handle_login\n",
					    current->comm);
				goto iscsi_rx_thread_out;
			}
			continue;
		}

#if !defined(__KERNEL__)
	cleanup_data.conn = conn;
		cleanup_data.inputpdu = &inputpdu;
		cleanup_data.data_buffer = &data_buffer;
		cleanup_data.buffer = &buffer;
		pthread_cleanup_push(rx_shutdown_thread, &cleanup_data);
#endif /* !__KERNEL__ */

		/* Receives entire pdu (hdr & data) - except for Data-out pdus
		 */
		if (handle_iscsi_rx_data(conn, &buffer, &data_buffer) < 0) {
			TRACE(TRACE_DEBUG,
			      "Trouble in handle_iscsi_rx_data\n");
			goto iscsi_rx_thread_out;
		}

#if !defined(__KERNEL__)
		pthread_cleanup_pop(0);
#endif /* !__KERNEL__ */


		/* keep a copy of this header in case we need it in a later
		 * reject */
		memcpy(conn->bad_hdr, buffer, ISCSI_HDR_LEN);

		opcode = buffer[0] & ISCSI_OPCODE;
		local_itt =
		    ntohl(((struct pdu_hdr *) buffer)->init_task_tag);

		/* connection belongs to session that is in full feature
		 * phase */
		if (conn->session->oper_param->SessionType) {
			/* RFC 3720 Section 3.3. iSCSI Session Types
			 * "Discovery-session - a session only opened for
			 * target discovery.  The target MUST ONLY accept text
			 * requests with the SendTargets key and a logout
			 * request with the reason 'close the session'.
			 * All other requests MUST be rejected."
			 */
			if (opcode != ISCSI_INIT_TEXT_CMND
			    && (opcode != ISCSI_INIT_LOGOUT_CMND
				|| (((struct iscsi_init_logout_cmnd *)
				     buffer)->
				    flags & LOGOUT_REASON_MASK) !=
				LOGOUT_CLOSE_SESSION)) {
				if (enqueue_reject
				    (conn, REASON_PROTOCOL_ERR) < 0) {
					free_pdu = 1;
					goto iscsi_rx_thread_out;
				}
				free_targ_buf(conn, &buffer, &data_buffer);
				continue;
			}
		}

		switch (opcode) {

		case ISCSI_INIT_LOGIN_CMND:
			{
				/* got Login request while in full feature
				 * phase */
				TRACE_ERROR
				    ("%s Got Login request ITT %u in "
				     "full feature phase\n", current->comm,
				     local_itt);
				goto iscsi_rx_thread_out;
			}

		case ISCSI_INIT_TEXT_CMND:
			{
				TRACE(TRACE_ISCSI,
				      "%lu %s Got Text request "
				      "ITT %u\n", jiffies, current->comm,
				      local_itt);

				if (handle_text_request(conn,
							conn->session,
							buffer,
							&data_buffer) <
				    0) {
					TRACE_ERROR("%s Trouble in "
						    "handle_text_request, "
						    "ITT %u\n",
						    current->comm,
						    local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_SCSI_CMND:
			{
				TRACE(TRACE_ISCSI, "%lu %s Got SCSI "
				      "command CmdSN %u ITT %u "
				      "flags 0x%02x DSL %u EDTL %u "
				      "ExpStatSn %u\n",
				      jiffies, current->comm,
				      ntohl(((struct pdu_hdr *) buffer)->
					    cmd_sn), local_itt,
				      ((struct pdu_hdr *) buffer)->flags,
				      ntohl(((struct pdu_hdr *) buffer)->
					    length),
				      ntohl(((struct iscsi_init_scsi_cmnd
					      *) buffer)->xfer_len),
				      ntohl(((struct pdu_hdr *) buffer)->
					    exp_stat_sn));

				if (handle_cmnd
				    (conn, conn->session, buffer,
				     data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_cmnd, ITT %u\n",
					     current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_SCSI_DATA_OUT:
			{
				TRACE(TRACE_ISCSI,
				      "%lu %s Got DataOut ITT %u DSL %u offset %u"
				      "\n", jiffies, current->comm,
				      local_itt,
				      ntohl(((struct pdu_hdr *) buffer)->
					    length),
				      ntohl(((struct pdu_hdr *) buffer)->
					    offset));

				if (handle_data
				    (conn, conn->session, buffer,
				     data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_data, ITT %u\n",
					     current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_TASK_MGMT_CMND:
			{
				TRACE(TRACE_ISCSI,
				      "%lu %s Got Task Mgt command ITT %u\n",
				      jiffies, current->comm, local_itt);

				if (handle_task_mgt_command
				    (conn, conn->session, buffer,
				     &data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_task_mgt_cmnd, ITT %u\n",
					     current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_LOGOUT_CMND:
			{
				TRACE(TRACE_ISCSI,
				      "%lu %s Got Logout request ITT %u\n",
				      jiffies, current->comm, local_itt);

				if (handle_logout
				    (conn, conn->session, buffer,
				     &data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_logout, ITT %u\n",
					     current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_INIT_NOP_OUT:
			{
				if (handle_nopout
				    (conn, conn->session, buffer,
				     &data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_nopout, ITT %u\n",
					     current->comm, local_itt);
					goto iscsi_rx_thread_out;
				}

				break;
			}

			/* SNACK Handling by Target - SAI */
		case ISCSI_INIT_SNACK:
			{
				TRACE(TRACE_ERROR_RECOVERY,
				      "%s Got SNACK Request\n",
				      current->comm);

				if (handle_snack
				    (conn, conn->session, buffer,
				     data_buffer) < 0) {
					TRACE_ERROR
					    ("%s Trouble in handle_snack\n",
					     current->comm);
					goto iscsi_rx_thread_out;
				}

				break;
			}

		case ISCSI_TARG_NOP_IN:
		case ISCSI_TARG_SCSI_RSP:
		case ISCSI_TARG_TASK_MGMT_RSP:
		case ISCSI_TARG_LOGIN_RSP:
		case ISCSI_TARG_TEXT_RSP:
		case ISCSI_TARG_SCSI_DATA_IN:
		case ISCSI_TARG_LOGOUT_RSP:
		case ISCSI_TARG_R2T:
		case ISCSI_TARG_ASYNC_MSG:
		case ISCSI_TARG_RJT:
			{
				TRACE_ERROR
				    ("%s Initiator sent a target opcode %.2x, ITT %u\n",
				     current->comm, opcode, local_itt);
				/*****
				TRACE_SET(TRACE_DEBUG | TRACE_ISCSI | TRACE_ERROR_RECOVERY
							  | TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE);
				*****/

				/* Send a Reject PDU and escalate to session recovery */
				enqueue_reject(conn, REASON_PROTOCOL_ERR);
				targ_session_recovery(conn);
				goto iscsi_rx_thread_out;
			}

		default:
			{
				TRACE_ERROR
				    ("%s Got unknown opcode %.2x ITT %u\n",
				     current->comm, opcode, local_itt);
				/*****
				TRACE_SET(TRACE_DEBUG | TRACE_ISCSI | TRACE_ERROR_RECOVERY
							  | TRACE_ISCSI_FULL | TRACE_ENTER_LEAVE);
				*****/

				/* Send a Reject PDU and escalate to session recovery */
				enqueue_reject(conn,
					       REASON_COMMAND_NOT_SUPPORTED);
				targ_session_recovery(conn);
				goto iscsi_rx_thread_out;
			}
		}		/* switch */
		if ((opcode != ISCSI_INIT_SCSI_CMND) &&
		    (opcode != ISCSI_INIT_SCSI_DATA_OUT))
			free_targ_buf(conn, &buffer, &data_buffer);

		/* in case there are any out-of-order commands that are now in-order */
		check_queued_cmnd(conn->session);


	}			/* while(1) */

      iscsi_rx_thread_out:
	cleanup_data.conn = conn;
	cleanup_data.inputpdu = &inputpdu;
	cleanup_data.data_buffer = &data_buffer;
	cleanup_data.buffer = &buffer;
	rx_shutdown_thread(&cleanup_data);
      out:
	rx_thread_out(conn);
	return 0;
}

/*
 * Static definitions.
 */

/*
 * called only from rx thread.
 * handle_login: This function is responsible for interpreting the login
 * message and then making sure the necessary details get filled in, a
 * new session is opened if required by the login and the response is returned
 *
 * INPUT: Details of the connection, buffer containing the login
 * OUTPUT: 0 if everything is okay, < 0 if there is a problem
 * Note a 0 value returned may still mean that a connection is rejected
 */
static int handle_login(struct iscsi_conn *conn,
			struct generic_pdu *inputpdu)
{
	struct iscsi_init_login_cmnd *pdu;
	struct iscsi_session *session, *temp;
	struct iscsi_conn *temp_conn;
	__u32 when_called;
	int found, retval = -1, stat_detail = -1;
	struct auth_parameter_type auth_param;	/*chap and srp support - CHONG */
	struct parameter_type *p;
	struct parameter_type (*this_param_tbl)[MAX_CONFIG_PARAMS];
	struct parameter_type (*temp_params)[MAX_CONFIG_PARAMS];
	struct iscsi_global *host;
	struct generic_pdu outputpdu;
	DECLARE_LOOP_COUNTER(count);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	pdu = (struct iscsi_init_login_cmnd *) inputpdu->hdr;
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_login_cmnd(pdu);

	/* this will be non-NULL if the parameter table should be freed when done */
	temp_params = NULL;

	if ((session = conn->session) == NULL) {
		/* should never happen */
#ifdef RMDA_MODE
		free_iser_login_buffer(conn, inputpdu);
#endif
		goto out_quick;
	}

	if ((host = session->devdata) == NULL) {
		/* should never happen */
#ifdef RMDA_MODE
		free_iser_login_buffer(conn, inputpdu);
#endif
		goto out_quick;
	}

	if (login_pdu_malloc(&outputpdu, "Output pdu") < 0) {
#ifdef RMDA_MODE
		free_iser_login_buffer(conn, inputpdu);
#endif
		goto out_quick;
	}

	/* set up the "unchanging" fields in the Login Response pdu */
	memset(outputpdu.hdr, 0, ISCSI_HDR_LEN);
	outputpdu.hdr->opcode = ISCSI_TARG_LOGIN_RSP;
	outputpdu.hdr->version_max = ISCSI_MAX_VERSION;
	outputpdu.hdr->version_active = ISCSI_MIN_VERSION;
	memcpy(outputpdu.hdr->isid, pdu->isid, 6);
	outputpdu.hdr->tsih = pdu->tsih;
	outputpdu.hdr->init_task_tag = pdu->init_task_tag;

	/* always use clean copy of configured parameter table for negotiations */
	this_param_tbl = session->session_params;

	/* Added for iSER - In place modifcn of pkt - shdn't be a problem with iSER
	 * memory */
	pdu->length = be32_to_cpu(pdu->length);
	pdu->tsih = be16_to_cpu(pdu->tsih);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->cid = be16_to_cpu(pdu->cid);
	pdu->cmd_sn = be32_to_cpu(pdu->cmd_sn);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);

	/* destructive access to session lists */
	write_lock(&host->session_rwlock);

	if (pdu->tsih == 0) {
		/* a new session, the iscsi_session structure is already set up */
		conn->cid = pdu->cid;
		conn->stat_sn = pdu->exp_stat_sn;
		session->cmd_sn = pdu->cmd_sn;
		session->exp_cmd_sn = pdu->cmd_sn;
		session->max_cmd_sn =
		    pdu->cmd_sn + session->max_queue_depth_allowed;

		memcpy(session->isid, pdu->isid, 6);

		/* set up the operational parameters from the global structure */
		set_session_parameters(session->oper_param,
				       *session->session_params);

		/* add new session to the end of the global session list
		 * with tsih = 0 to indicate login not finished
		 */
		list_add_tail(&session->sess_link, &host->session_list);

		/* set flags to allow keys for new session (leading) login */
		when_called = LEADING_ONLY | INITIAL_ONLY | ALL;

		TRACE(TRACE_ISCSI, "%lu %s Start new session %p\n",
		      jiffies, current->comm, session);
	} else {
		/* existing session, check through the session list to find it */
		found = 0;
		INIT_LOOP_COUNTER(count);
		list_for_each_entry(session, &conn->dev->session_list,
				    sess_link) {
			BUMP_LOOP_COUNTER(count, "session_list");
			if (session->tsih == pdu->tsih) {
				found = 1;
				break;
			}
		}

		if (!found) {
			TRACE_ERROR("%s No existing session with TSIH %u, "
				    "terminate this connection\n",
				    current->comm, pdu->tsih);
			goto err_conn_out;
		}

		if (conn->portal_group_tag != session->portal_group_tag) {
			TRACE_ERROR
			    ("%s Portal group tag %u for new connection does "
			     "not match portal group tag %u of session\n",
			     current->comm, conn->portal_group_tag,
			     session->portal_group_tag);
			stat_detail = STAT_DETAIL_NOT_INCLUDED;
			goto err_conn_out;
		}

		/* check isid */
		if (memcmp(pdu->isid, session->isid, 6) != 0) {
			TRACE_ERROR("%s The session has a different ISID, "
				    "terminate the connection\n",
				    current->comm);
			stat_detail = STAT_DETAIL_ERR;
			goto err_conn_out;
		}

		conn->cid = pdu->cid;
		conn->stat_sn = pdu->exp_stat_sn;

		/* check cid, and if it already exists then release old connection */
		INIT_LOOP_COUNTER(count);
		list_for_each_entry(temp_conn, &session->conn_list,
				    conn_link) {
			BUMP_LOOP_COUNTER(count, "conn_list");
			if (temp_conn->cid == conn->cid) {
				TRACE(TRACE_ISCSI,
				      "%lu %s connection reinstatement with "
				      "cid %u\n", jiffies, current->comm,
				      conn->cid);

				if (iscsi_release_connection(temp_conn) <
				    0) {
					TRACE_ERROR
					    ("%s Error releasing connection\n",
					     current->comm);
				}
				break;
			}
		}

		TRACE(TRACE_ISCSI,
		      "%lu %s new connection cid %u attached to "
		      "existing session tsih %u, IP 0x%08x\n", jiffies,
		      current->comm, conn->cid, pdu->tsih,
		      ((struct sockaddr_in *) conn->ip_address)->sin_addr.
		      s_addr);

		/* add new connection to end of connection list for existing session */
		temp = conn->session;
		conn->session = session;
		list_del(&conn->conn_link);
		temp->nconn = 0;
		list_add_tail(&conn->conn_link, &session->conn_list);
		session->nconn++;

		/* use clean parameter table for negotiations, free it later */
		temp_params = this_param_tbl;
		temp->session_params = NULL;

		/* free up the no-longer-needed session structure */
		iscsi_release_session(temp);

		/* set back the leading-only keys if these keys were set to
		 * " KEY_TO_BE_NEGOTIATED" in the leading connection negotation.*/
		reset_parameter_table(*this_param_tbl);

		/* set flags to allow keys for new connection (only) login */
		when_called = INITIAL_ONLY | ALL;
	}

	/* by the time we get here, the connection is attached to the correct
	 * session and that session is threaded onto the devdata->session_list.
	 */

	write_unlock(&host->session_rwlock);

	/* chap and srp support - CHONG */
	auth_param.auth_flags = 0;
	auth_param.chap_local_ctx =
	    CHAP_CloneContext(conn->dev->auth_parameter.chap_local_ctx);
	auth_param.chap_peer_ctx =
	    CHAP_CloneContext(conn->dev->auth_parameter.chap_peer_ctx);
	auth_param.srp_ctx =
	    SRP_CloneContext(conn->dev->auth_parameter.srp_ctx);

	if ((p = find_flag_parameter(TARGETPORTALGROUPTAG_FLAG,
				     *conn->session->session_params))) {
		/* force the TargetPortalGroupTag key to have as its value the
		   portal group tag for this session */
		p->int_value = conn->session->portal_group_tag;
	}

	if (parameter_negotiate
	    (conn, *this_param_tbl, inputpdu, &outputpdu, when_called,
	     auth_param) < 0) {
		TRACE_ERROR("%s Parameter negotiation failed\n",
			    current->comm);
		/* chap and srp support - CHONG */
		CHAP_FinalizeContext(auth_param.chap_local_ctx);
		CHAP_FinalizeContext(auth_param.chap_peer_ctx);
		SRP_FinalizeContext(auth_param.srp_ctx);
		goto out;
	}

	/* Bjorn Thordarson, 10-May-2004 -- start -- */
	if (TRACE_TEST(TRACE_ISCSI_FULL)) {
		int k;

		for (k = 0; k < MAX_CONFIG_PARAMS; k++) {
			if ((*this_param_tbl)[k].parameter_name) {
				if ((*this_param_tbl)[k].type & NUMBER)
					printk("PARAM: %s = %u\n",
					       (*this_param_tbl)[k].
					       parameter_name,
					       (*this_param_tbl)[k].
					       int_value);
				else if ((*this_param_tbl)[k].str_value)
					printk("PARAM: %s = %s\n",
					       (*this_param_tbl)[k].
					       parameter_name,
					       (*this_param_tbl)[k].
					       str_value);
			}
		}
	}
	/* Bjorn Thordarson, 10-May-2004 -- end -- */

	/* chap and srp support - CHONG */
	CHAP_FinalizeContext(auth_param.chap_local_ctx);
	CHAP_FinalizeContext(auth_param.chap_peer_ctx);
	SRP_FinalizeContext(auth_param.srp_ctx);

	conn->stat_sn++;

	if (conn->stat_sn < 0) {
		TRACE_ERROR
		    ("%s Parameter negotiation with error stat_sn\n",
		     current->comm);
		goto out;
	}

	/* set the operational parameters to the negotiated value */
	if (pdu->tsih == 0) {
		/* this is a new connection in a new session */
		set_session_parameters(session->oper_param,
				       *session->session_params);

		/* preallocate iscsi command structures for this session */
		setup_free_commands(session);

		/* preallocate buffers for unsolicited data (if session can use them) */
		setup_unsolicited_buffers(session);

		/* Initialize R2T timer - ref18_09c SAI */
		if (session->oper_param->ErrorRecoveryLevel
		    && session->r2t_period) {
			/* error recovery ver ref18_04 */

			/* create a retransmit_thread for handling error recovery */
			if (kernel_thread
			    (iscsi_retran_thread, (void *) session,
			     0) < 0) {
				TRACE_ERROR
				    ("%s Unable to create retran_thread\n",
				     current->comm);
				session->r2t_period = 0;
			} else {
				restart_r2t_timer(session);
				TRACE(TRACE_ISCSI,
				      "%lu %s R2T timer %p started for session "
				      "%p\n", jiffies, current->comm,
				      session->r2t_timer, session);
			}
		}
	}

	/* we are now in Full Feature Phase */
	conn->hdr_crc =
	    (conn->connection_flags & USE_HEADERDIGEST) ? 1 : 0;
	conn->data_crc = (conn->connection_flags & USE_DATADIGEST) ? 1 : 0;

	/* all ok */
	retval = 0;

      out:
	/* here when nothing further can or needs to be done
	 * to the session or connection structs
	 */
	if (temp_params) {
		param_tbl_uncpy(*temp_params);
		my_kfree((void **) &temp_params, "temp_params");
	}
	login_pdu_free(&outputpdu, "Output pdu");
      out_quick:
	return retval;

	/* here on a fatal error detected before session is linked into devdata */
      err_conn_out:

	/* put this session onto end of "bad-session" list for cleanup later */
	session = conn->session;
	TRACE(TRACE_DEBUG, "add to list bad session %p, conn %p\n",
	      session, conn);

	list_add_tail(&session->sess_link, &host->bad_session_list);

	write_unlock(&host->session_rwlock);

	if (stat_detail >= 0) {
		login_reject(conn, STAT_CLASS_INITIATOR, stat_detail,
			     &outputpdu);
	}
#ifdef RMDA_MODE
	/* Added for iSER - free inputpdu here - In all other cases, it has
	 * already been freed */
	free_iser_login_buffer(conn, inputpdu);
#endif
	iscsi_release_socket(conn);
	goto out;
}

/*
 * executed only by the rx thread.
 * Called when a Text Request pdu has just been received.
 * Currently it only supports the SendTarget command.
 * INPUT: connection, session and the buffer containing the text pdu.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble.
 */
static int handle_text_request(struct iscsi_conn *conn,
			       struct iscsi_session *session,
			       __u8 * buffer, __u8 ** data_buffer)
{
	struct iscsi_init_text_cmnd *pdu =
	    (struct iscsi_init_text_cmnd *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_text_cmnd(pdu);

	if (!(cmnd = get_free_command(buffer, session, &err))) {
		return -1;
	}

	cmnd->conn = conn;
	cmnd->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);
	cmnd->first_burst_len = be32_to_cpu(pdu->length);

	if (cmnd->first_burst_len) {
		cmnd->ping_data = (char *) *data_buffer;
	} else {
		if (conn->text_in_progress == NULL ||
		    ((struct iscsi_cmnd *) conn->text_in_progress)->state
		    != ISCSI_BLOCKED_SENDING_TEXT) {
			/* empty text request, might be just to reset negotiations */
			if (cmnd->target_xfer_tag != ALL_ONES) {
				/* we could remove this warning after testing */
				TRACE_WARNING
				    ("%s Empty Text Request CmdSN %u ExpCmdSN %u "
				     "ITT %u opcode 0x%02x\n",
				     current->comm, cmnd->cmd_sn,
				     session->exp_cmd_sn,
				     cmnd->init_task_tag,
				     cmnd->opcode_byte);
			}
		}
	}

	if (err < 0) {
		/* out of range, silently ignore it */
		TRACE_ERROR
		    ("%s ignoring out of range CmdSN %u ExpCmdSN %u ITT %u "
		     "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		     session->exp_cmd_sn, cmnd->init_task_tag,
		     cmnd->opcode_byte);
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 0);
		put_free_command(cmnd, session);
	} else {
#ifndef RDMA_MODE
		*data_buffer = NULL;
#endif
		if (err > 0) {
#ifdef RDMA_MODE
			/* Copy into a new buffer so iSER's buffer can be freed - Text req
			 * and NopOuts are infrequent enough to do so - Minti 5/9/05 */
			cmnd->ping_data =
			    my_kmalloc(cmnd->first_burst_len, "data_buf");
			memcpy(cmnd->ping_data, *data_buffer,
			       cmnd->first_burst_len);
#endif
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 1);
		if (err == 0) {
			/* do immediate command or expected non-immediate command now */
			do_text_request(cmnd, conn, session);
		}
	}

	return 0;
}

/*
 * executed only by the rx thread.
 * called when a Logout Request has just been received.
 * INPUT: connection, session and the buffer containing the logout pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_logout(struct iscsi_conn *conn,
			 struct iscsi_session *session,
			 __u8 * buffer, __u8 ** data_buffer)
{
	struct iscsi_init_logout_cmnd *pdu;
	struct iscsi_cmnd *cmnd;
	int err;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	pdu = (struct iscsi_init_logout_cmnd *) buffer;
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_logout_cmnd(pdu);

	if (!(cmnd = get_free_command(buffer, session, &err))) {
		return -1;
	}

	cmnd->conn = conn;
	cmnd->command_flags &= LOGOUT_REASON_MASK;
	cmnd->data_length = be32_to_cpu(pdu->length);

	printk("%s Logout CmdSN %u ITT %u reason %u cid %u\n",
	       current->comm, cmnd->cmd_sn, cmnd->init_task_tag,
	       cmnd->command_flags, be16_to_cpu(pdu->cid));

	if (cmnd->data_length) {
		/* read length bytes of text data into a newly-allocated buffer */
		TRACE_ERROR
		    ("%s Data in logout request CmdSN %u ExpCmdSN %u ITT %u "
		     "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		     session->exp_cmd_sn, cmnd->init_task_tag,
		     cmnd->opcode_byte);
		/* Minti - logout has no data attached */
		cmnd->ping_data = NULL;
	}

	if (err < 0) {
		/* out of range, silently ignore it */
		TRACE_ERROR
		    ("%s ignoring out of range CmdSN %u ExpCmdSN %u ITT %u "
		     "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		     session->exp_cmd_sn, cmnd->init_task_tag,
		     cmnd->opcode_byte);
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 0);
		put_free_command(cmnd, session);
	} else {
		if (err == 0) {
			/* do immediate command or expected non-immediate command now */
			cmnd->state = ISCSI_LOGOUT;
		} else {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 1);
	}
	return 0;
}

/*
 * executed only by the rx thread.
 * called when a NopOut pdu has just been received.
 * there are 4 times initiator sends one of these:
 * 1. when ITT == 0xfffffff and TTT == 0xfffffffff
 *		then this is not a reply to a previous NopIn from target,
 *		and the initiator expects no reply
 *		the I bit must be 1 in this case
 *		there can be no data in this case
 * 2. when ITT != 0xffffffff and TTT == 0xffffffff
 *		then this is not a reply to a previous NopIn from target,
 *		but the initiator expects a NopIn in reply
 *		there may be data in this case
 * 3. when ITT == 0xffffffff and TTT != 0xffffffff
 *		then this is a reply to a previous NopIn from target,
 *		and the initiator expects no reply
 *		the I bit must be 1 in this case
 *		there can be no data in this case
 * 4. when ITT != 0xffffffff and TTT != 0xffffffff
 *		then this is a reply to a previous NopIn from target,
 *		and the initiator expects a NopIn in reply
 *		there may be data in this case
 * INPUT: connection, session and the buffer containing the NopOut pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_nopout(struct iscsi_conn *conn,
			 struct iscsi_session *session,
			 __u8 * buffer, __u8 ** data_buffer)
{
	struct iscsi_init_nopout *pdu =
	    (struct iscsi_init_nopout *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ISCSI,
	      "%lu %s Got NopOut ITT %d TTT %d CmdSN %u ExpStatSn %u" "\n",
	      jiffies, current->comm, be32_to_cpu(pdu->init_task_tag),
	      be32_to_cpu(pdu->target_xfer_tag), be32_to_cpu(pdu->cmd_sn),
	      be32_to_cpu(pdu->exp_stat_sn));

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_nopout(pdu);

	pdu->length = be32_to_cpu(pdu->length);
	pdu->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);

	if (pdu->init_task_tag == ALL_ONES) {
		/* Draft 20, Section 10.18.1 Initiator Task Tag
		 * "If the Initiator Task Tag contains 0xffffffff, the I bit MUST be
		 * set to 1 and the CmdSN is not advanced after this PDU is sent."
		 *
		 * there must be no data attached, and initiator expects no reply
		 */
		if (!(pdu->opcode & I_BIT)) {
			TRACE_ERROR
			    ("%s NopOut with ITT 0x%08x but I bit not set\n",
			     current->comm, ALL_ONES);
			pdu->opcode |= I_BIT;	/* so CmdSN is not advanced below */
		}
		if (pdu->length) {
			/* should not have ping data if not expecting a response */
			TRACE_ERROR
			    ("%s NopOut with ITT 0x%08x but DSL %u\n",
			     current->comm, ALL_ONES, pdu->length);
		}
	}

	if (pdu->target_xfer_tag != ALL_ONES) {
		/* this is a reply to earlier NopIn sent from target
		 * should now find that existing NopIn command in our queue
		 * (which we could only send with ITT = 0xffffffff)
		 */
		cmnd =
		    search_tags(conn, ALL_ONES, pdu->target_xfer_tag, 0);
		if (!cmnd) {
			TRACE_ERROR
			    ("%s No NopIn command found with TTT 0x%08x\n",
			     current->comm, pdu->target_xfer_tag);
		} else {
			/* this is the reply we expected.  Check the LUN and then free the
			 * NopIn cmnd (which should not have any ping data attached to it).
			 */
			cmnd->state = ISCSI_DEQUEUE;
			atomic_dec(&conn->outstanding_nopins);
		}
	}

	/* now set up a command structure for the new NopOut we just received */
	if (!(cmnd = get_free_command(buffer, session, &err))) {
		return -1;
	}

	cmnd->state = ISCSI_PING;
	cmnd->conn = conn;
	cmnd->target_xfer_tag = ALL_ONES;
	cmnd->data_length = pdu->length;

	if (cmnd->data_length) {
		cmnd->ping_data = (char *) *data_buffer;
	}

	if (err < 0 || (err == 0 && cmnd->init_task_tag == ALL_ONES)) {
		/* NopOut is out of range, or is in order but is not used as a ping
		 * request (so I == 1 and no reply NopIn expected), just drop this
		 * command structure without ever enqueuing it
		 */
		if (err < 0) {
			/* out of range, silently ignore it */
			TRACE_ERROR
			    ("%s ignoring out of range CmdSN %u ExpCmdSN %u "
			     "ITT %u opcode 0x%02x\n", current->comm,
			     cmnd->cmd_sn, session->exp_cmd_sn,
			     cmnd->init_task_tag, cmnd->opcode_byte);
		}		/* else the NopOut pdu had to be immediate (because ITT == 0xffffffff,
				 * so CmdSN was not advanced and MaxCndSN is ok as is
				 */
		TRACE(TRACE_DEBUG,
		      "%s tossing CmdSN %u ExpCmdSN %u ITT %u "
		      "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		      session->exp_cmd_sn, cmnd->init_task_tag,
		      cmnd->opcode_byte);
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 0);
		put_free_command(cmnd, session);
	} else {
		if (err > 0) {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}
		TRACE(TRACE_DEBUG,
		      "%s queueing CmdSN %u ExpCmdSN %u ITT %u "
		      "opcode 0x%02x, state %u, data_length %u\n",
		      current->comm, cmnd->cmd_sn, session->exp_cmd_sn,
		      cmnd->init_task_tag, cmnd->opcode_byte, cmnd->state,
		      cmnd->data_length);
		*data_buffer = NULL;
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 1);
	}
	return 0;
}

/*
 * executed only by the rx thread.
 * handle_cmnd: this function is responsible for processing a SCSI command
 * just received from an initiator, making the mid-level aware of the
 * command and then queueing up the command in the command queue for the
 * session
 * INPUT: connection and the buffer containing the just-received PDU header
 * OUTPUT: >= 0 if everything is okay, < 0 if there is trouble
 */
static int handle_cmnd(struct iscsi_conn *conn,
		       struct iscsi_session *session,
		       __u8 * buffer, __u8 * data_buffer)
{
	struct iscsi_init_scsi_cmnd *pdu =
	    (struct iscsi_init_scsi_cmnd *) buffer;
	struct iscsi_cmnd *cmnd;
	int err;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_cmnd\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_cmnd(pdu);

	/* 3th parameter is NULL because we cannot check command ordering until
	 * after any immediate data has been read in.
	 * This causes command list to become blocked if this command is in order.
	 * It also means we generate a new target_task_tag for this command.
	 */
	if (!(cmnd = get_free_command(buffer, session, NULL))) {
		err = -1;
		goto out;
	}

	cmnd->conn = conn;
	cmnd->data_length = be32_to_cpu(pdu->xfer_len);
	cmnd->lun = pdu->lun;
	cmnd->cdb[0] = pdu->cdb[0];
	cmnd->cdb[4] = pdu->cdb[4];
	cmnd->first_burst_len = pdu->length = be32_to_cpu(pdu->length);

#if defined(RDMA_MODE)
	iser_front_pdu(conn, &cmnd->rx_iser_pdu, &cmnd->iser_task_h);
#endif				/* RDMA_MODE */

	/* check data length, immediate data should not bigger than
	 * MaxRecvPDULength or FirstBurstLength
	 */
	if (cmnd->first_burst_len) {
		/* there is data attached to this PDU, it must be immediate data */
		if (cmnd->first_burst_len > conn->max_recv_length
		    || cmnd->first_burst_len >
		    session->oper_param->FirstBurstLength) {
			printk("%s %s %d\n", current->comm,
			       maxburstlength_word, conn->max_recv_length);
			printk("%s %s %d\n", current->comm,
			       firstburstlength_word,
			       session->oper_param->FirstBurstLength);
			TRACE_WARNING("%s DSL %u too big\n", current->comm,
				      cmnd->first_burst_len);
		}
	}

	if (!(cmnd->command_flags & F_BIT)) {
		/* F bit not set, so unsolicited DataOut PDUs must follow this PDU,
		 * any and all solicited data must follow a full first burst
		 */
		cmnd->unsolicited_data_present = 1;
		cmnd->first_solicited_offset =
		    session->oper_param->FirstBurstLength;
		if (!session->oper_param->DataPDUInOrder) {
			/* DataPDUInOrder==No, so DataIn PDUs can be out-of-order */
			cmnd->pdu_range_list.limit = cmnd->first_burst_len;
			merge_offset_length(&cmnd->pdu_range_list, 0,
					    cmnd->first_burst_len);
		}
	} else {
		/* no unsolicited DataIn PDUs, all remaining data is solicited */
		if ((cmnd->first_solicited_offset = cmnd->first_burst_len))
			/* there is some unsolicited immediate data however, and
			 * unsolicited data should always be the first sequence
			 */
			if (!session->oper_param->DataSequenceInOrder) {
				/* DataSequenceInOrder==No, so sequences can be out-of-order */
				cmnd->seq_range_list.limit =
				    cmnd->first_burst_len;
				merge_offset_length(&cmnd->seq_range_list,
						    0,
						    cmnd->first_burst_len);
			}
	}

	/* NOTE: this subtraction can produce a negative value! */
	cmnd->r2t_data_total =
	    cmnd->data_length - cmnd->first_solicited_offset;

	/* store "length" bytes of immediate data attached to this command
	 * the iser buffer is saved in cmnd - pass NULL here
	 */
	if ((err =
	     save_unsolicited_data(cmnd, 0, (struct pdu_hdr *) buffer,
				   data_buffer, NULL))) {
		/* When this fails for non RDMA_MODE, the data_buf is already cleaned up
		 * so no freeing is required.
		 */
#ifdef RDMA_MODE
		targ_release_control(conn, &buffer,
				     (__u8 **) & data_buffer);
#endif
		put_free_command(cmnd, session);
		goto out1;
	}

	/* now check cmd_sn for command ordering, but don't bump
	 * exp_cmd_sn for the session until after this command has been delivered
	 * to the mid-level in proper order.
	 * (bumping exp_cmd_sn now might let other commands jump ahead of this)
	 */
	err = check_cmd_sn(cmnd, pdu, session);

	if (err) {
		/* command received is not in expected CmdSN order */
		err =
		    out_of_order_cmnd(conn, session, buffer, cmnd, err,
				      data_buffer);
		/* cmnd has now been queued or freed, as appropriate */
		goto out1;
	}

	/* else command is immediate or (is in-order and active_list is blocked) */

	TRACE(TRACE_DEBUG, "%s unsolicited_data_present %d, err %d, flags "
	      "0x%02x\n", current->comm, cmnd->unsolicited_data_present,
	      err, cmnd->command_flags);

	cmnd->state = ISCSI_AWAIT_BUFFER;
	if (ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 1)) {
		/* this command's ITT duplicates one already in queue */
		goto out2;
	}

	/* make sure that the command is still good while going to
	 * the midlevel. */
	cmnd_check_magic(session, cmnd);

	/* pass this command on to the midlevel so it can set up the buffers */
	err = deliver_to_midlevel(cmnd, pdu, session);

	/* in case there are any out-of-order commands that are now in-order */
	check_queued_cmnd(session);

      out:
#ifdef RDMA_MODE
	free_targ_buf(conn, &buffer, &data_buffer);
#endif
      out1:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_cmnd, err = %d\n", err);
	return err;

      out2:
	/* here for errors that are detected after cmnd is added to queue */
	cmnd->state = ISCSI_DEQUEUE;
	goto out;
}

/*
 * executed only by the rx thread.
 * called when a Task Management Request has just been received.
 * INPUT: connection, session and the buffer containing the TM pdu
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_task_mgt_command(struct iscsi_conn *conn,
				   struct iscsi_session *session,
				   __u8 * buffer, __u8 ** data_buffer)
{
	struct iscsi_init_task_mgt_command *pdu
	    = (struct iscsi_init_task_mgt_command *) buffer;
	struct iscsi_cmnd *cmnd;
	int err = 0;
#ifdef CONFIG_ISCSI_DEBUG
	__u32 save_trace;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE_GET(save_trace);
	if (TRACE_TEST(TRACE_TASK_MGT)) {
		/* turn on (almost) all tracing while processing a TM command */
		TRACE_SET(TRACE_ALL
			  & ~(TRACE_NET | TRACE_BUF | TRACE_MY_MEMORY |
			      TRACE_TIMERS));
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_task_mgt_command(pdu);

	if (!(cmnd = get_free_command(buffer, session, &err))) {
		err = -1;
		goto out;
	}

	cmnd->conn = conn;
	cmnd->ref_task_tag = be32_to_cpu(pdu->ref_task_tag);
	cmnd->ref_cmd_sn = be32_to_cpu(pdu->ref_cmd_sn);
	cmnd->ref_function = pdu->function & TMF_FUNCTION_MASK;
	cmnd->response = FUNCTION_COMPLETE;

	TRACE_ERROR
	    ("%s Got TM Req CmdSN %u, ref CmdSN %u, ITT %u, RTT %u, "
	     "ExpCmdSN %u, ExpStatSN %u\n", current->comm, cmnd->cmd_sn,
	     cmnd->ref_cmd_sn, cmnd->init_task_tag, cmnd->ref_task_tag,
	     session->exp_cmd_sn, cmnd->stat_sn);

	if (!(pdu->function & F_BIT)) {
		TRACE_ERROR("%s task mgt request with F_BIT 0\n",
			    current->comm);
	}

	/* RFC 3720 Section 10.5.4 Referenced Task Tag
	 * "The Initiator Task Tag of the task to be
	 * aborted for the ABORT TASK function or reassigned for the TASK REASSIGN
	 * function.  For all the other functions this field MUST be set to the
	 * reserved value 0xffffffff."
	 */
	if (cmnd->ref_task_tag == ALL_ONES) {
		if ((cmnd->ref_function == TMF_ABORT_TASK)
		    || (cmnd->ref_function == TMF_TASK_REASSIGN)) {
			TRACE_ERROR
			    ("%s task mgt request has RTT 0x%08x but should refer "
			     "to a task when function is %d\n",
			     current->comm, cmnd->ref_task_tag,
			     cmnd->ref_function);
			err = -1;
		}
	} else if ((cmnd->ref_function != TMF_ABORT_TASK)
		   && (cmnd->ref_function != TMF_TASK_REASSIGN)) {
		TRACE_ERROR
		    ("%s task mgt request has RTT %u but should be 0x%08x "
		     "when function is %d\n", current->comm,
		     cmnd->ref_task_tag, ALL_ONES, cmnd->ref_function);
		err = -1;
	}

	if (pdu->length > 0) {
		/* read length bytes of text data into a newly-allocated buffer */
		TRACE_ERROR
		    ("%s Data attached to task mgt request, CmdSN %u, ExpCmdSN "
		     "%u, ITT %u, opcode 0x%02x\n", current->comm,
		     cmnd->cmd_sn, session->exp_cmd_sn,
		     cmnd->init_task_tag, cmnd->opcode_byte);
		/* Minti - Task management commands have no data attached */
		cmnd->ping_data = NULL;
	}

	if (err < 0) {
		/* out of range or some sort of error, silently ignore it */
		TRACE_ERROR
		    ("%s Drop out of range task mgt request CmdSN %u, "
		     "ExpCmdSN %u\n", current->comm, cmnd->cmd_sn,
		     session->exp_cmd_sn);
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 0);
		put_free_command(cmnd, session);
	} else {
		if (err > 0) {
			/* within range but out of order, queue it for later */
			cmnd->state = ISCSI_QUEUE_OTHER;
		}
		ack_sent_cmnds(conn, cmnd, cmnd->stat_sn, 1);
		if (err == 0) {
			/* is immediate command or expected non-immediate command */
			do_task_mgt(conn, cmnd);
		}
	}
	err = 0;
      out:
	if (TRACE_TEST(TRACE_TASK_MGT)) {
		/* restore tracing to what it was when we came in here */
		TRACE_SET(save_trace);
	}

	return err;
}

/*
 * executed only by the rx thread.
 * handle_data: This function is responsible for receiving DataOut pdus that
 * correspond to a WRITE command. It receives the data in the buffers that
 * are allocated by the STML. It then has to change the state of the
 * command and inform the STML about the data received
 * INPUT: connection and buffer containing the ISCSI_DATA (write) PDU
 * OUTPUT: 0, if everything is okay, < 0 otherwise
 */
static int handle_data(struct iscsi_conn *conn,
		       struct iscsi_session *session,
		       __u8 * buffer, __u8 * data_buffer)
{
	struct iscsi_init_scsi_data_out *hdr =
	    (struct iscsi_init_scsi_data_out *) buffer;
#if defined(RDMA_MODE)
	struct iser_recv_pdu_t *rx_pdu_buff;
	void *rx_handle_buff;
#endif				/* RDMA_MODE */
	struct iscsi_cmnd *cmd;
	struct targ_error_rec err_rec;
	struct scatterlist *st_list;
	int err = 0, giveback = 0;
	__u32 i, offset;
	__u32 data_sn;
	int free_pdu = 0;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_data\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_data_out(hdr);

	hdr->length = be32_to_cpu(hdr->length);
	hdr->init_task_tag = be32_to_cpu(hdr->init_task_tag);
	hdr->target_xfer_tag = be32_to_cpu(hdr->target_xfer_tag);
	hdr->exp_stat_sn = be32_to_cpu(hdr->exp_stat_sn);
	hdr->offset = be32_to_cpu(hdr->offset);
	data_sn = be32_to_cpu(hdr->data_sn);

	/* check data length, should not be bigger than MaxRecvDataSegmentLength */
	if (hdr->length > conn->max_recv_length) {
		TRACE_WARNING
		    ("%s DataOut ITT %u, DataSN %u, DSL %u exceeds "
		     "%s %u\n", current->comm, hdr->init_task_tag, data_sn,
		     hdr->length, maxrecvdatasegmentlength_word,
		     conn->max_recv_length);
		/* now ignore this, since we can deal with it */
	}

	/* search command queue to match command */
	cmd =
	    search_tags(conn, hdr->init_task_tag, hdr->target_xfer_tag, 0);

	if (!cmd) {
		TRACE_ERROR
		    ("%s DataOut ITT %u, DataSN %u, TTT %u, No matching "
		     "command\n", current->comm, hdr->init_task_tag,
		     data_sn, hdr->target_xfer_tag);
		TRACE(TRACE_ISCSI,
		      "%lu %s Probably SCSI cmnd PDU lost - drop it\n",
		      jiffies, current->comm);
#ifndef RDMA_MODE
		targ_drop_pdu_data(conn, hdr->length);
#endif
		free_pdu = 1;
		goto end_handle_data;
	}

	/* check that the command is good before giving it data. */
	cmnd_check_magic(session, cmd);

	/* DataOut has no CmdSN but does have an ExpStatSN so use that now */
	ack_sent_cmnds(conn, NULL, hdr->exp_stat_sn, 0);

	/* update last time this command got some data back from initiator */
	cmd->timestamp = jiffies;

#ifdef ISER
	/* iSER spec - 9.3.4 - Initiator may send solicited Data-out PDUs
	 * via send_control if requested to. This is done here so that data_sn
	 * checks are not done for iSER solicited data.
	 */
	if (session->oper_param->RDMAExtensions
	    && !cmd->unsolicited_data_present) {
#ifndef TCP_ISER
		TRACE_ERROR("Received solicited DataOut in iSER mode\n");
		err = -1;
		free_pdu = 1;
		goto end_handle_data;
#endif
	}
#endif
	/* These tests should pass for iSER - Minti */
	if (session->oper_param->DataPDUInOrder
	    && session->oper_param->DataSequenceInOrder) {
		/* error recovery ver ref18_04 - SAI */
		if (cmd->data_done > hdr->offset) {
			TRACE_WARNING
			    ("%s Dropping duplicate DataOut ITT %u, DataSN %u, "
			     "Offset %u\n", current->comm,
			     hdr->init_task_tag, data_sn, hdr->offset);
#ifndef RDMA_MODE
			targ_drop_pdu_data(conn, hdr->length);
#endif
			goto update_handle_data;
		} else if (cmd->data_done < hdr->offset) {
			/* offset error - expected packets might have been dropped */
			TRACE_ERROR
			    ("%s DataOut ITT %u, DataSN %u, Offset %u bigger than "
			     "expected %u\n", current->comm,
			     hdr->init_task_tag, data_sn, hdr->offset,
			     cmd->data_done);

			TRACE(TRACE_ERROR_RECOVERY,
			      "Start sequence error recovery\n");
			err_rec.curr_conn = conn;
			err_rec.pdu_hdr = (struct pdu_hdr *) hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = SEQUENCE_ERR;
			if ((err = targ_do_error_recovery(&err_rec)) < 0)
				goto end_handle_data;

			if (session->oper_param->ErrorRecoveryLevel !=
			    SESSION_RECOVERY) {
				/* What is this - Minti ?? */
				check_r2t_done(cmd, hdr);
			}
			goto end_handle_data;
		} else if (data_sn != cmd->data_sn) {
			TRACE_WARNING
			    ("%s DataOut ITT %u, got DataSN %u, expected %u\n",
			     current->comm, hdr->init_task_tag, data_sn,
			     cmd->data_sn);
			/* now ignore this, since we can deal with it */
		}
	}

	cmd->recovery_r2t = 0;
	cmd->data_sn = data_sn + 1;

	/* check FirstBurstLength and MaxBurstLength */
	if (cmd->unsolicited_data_present) {
		/* this DataOut PDU is part of the unsolicited first burst */
		cmd->seq_range_list.offset = 0;

		if (!session->oper_param->DataPDUInOrder) {
			/* DataPDUInOrder==No, so PDU can be out-of-order */
			merge_out_of_order(hdr, cmd);
		}

		/* number of bytes read so far in this unsolicited first burst */
		cmd->first_burst_len += hdr->length;

		if (cmd->first_burst_len >
		    session->oper_param->FirstBurstLength) {
			TRACE_WARNING
			    ("%s DataOut ITT %u, DataSN %u, data length %u "
			     "exceeds FirstBurstLength %u\n",
			     current->comm, hdr->init_task_tag, data_sn,
			     cmd->first_burst_len,
			     session->oper_param->FirstBurstLength);
			/* now ignore this, since we can deal with it */
		}

		if (hdr->flags & F_BIT) {
			/* end of the unsolicited first burst */
			if (!session->oper_param->DataPDUInOrder) {
				/* DataPDUInOrder==No, PDUs can be out-of-order */
				check_range_list_complete(&cmd->
							  pdu_range_list,
							  cmd->cmd_sn,
							  cmd->
							  init_task_tag);
				free_range_list(&cmd->pdu_range_list);
			}

			/* unsolicited data should always be the first sequence */
			if (!session->oper_param->DataSequenceInOrder) {
				/* DataSequenceInOrder==No, sequence can be out-of-order */
				merge_offset_length(&cmd->seq_range_list,
						    0,
						    cmd->first_burst_len);
			}
		}
	} else {
		/* this DataOut PDU is part of a solicited burst */
		if (cmd->next_burst_len == 0) {
			/* this is the first PDU in a solicited burst */
			cmd->seq_range_list.offset = hdr->offset;
			if (!session->oper_param->DataPDUInOrder) {
				cmd->pdu_range_list.offset = ALL_ONES;
				cmd->pdu_range_list.limit = 0;
			}
		}

		if (!session->oper_param->DataPDUInOrder) {
			/* DataPDUInOrder==No, so PDU can be out-of-order */
			merge_out_of_order(hdr, cmd);
		}

		/* number of bytes read so far in this solicited burst */
		cmd->next_burst_len += hdr->length;

		if (cmd->next_burst_len >
		    session->oper_param->MaxBurstLength) {
			TRACE_WARNING
			    ("%s DataOut ITT %u, DataSN %u, data length %u "
			     "exceeds MaxBurstLength %u\n", current->comm,
			     hdr->init_task_tag, data_sn,
			     cmd->next_burst_len,
			     session->oper_param->MaxBurstLength);
			/* now ignore this, since we can deal with it */
		}

		if (hdr->flags & F_BIT) {
			/* this is the end of the current solicited burst */
			if (!session->oper_param->DataPDUInOrder) {
				/* DataPDUInOrder==No, PDUs can be out-of-order */
				check_range_list_complete(&cmd->
							  pdu_range_list,
							  cmd->cmd_sn,
							  cmd->
							  init_task_tag);
				cmd->seq_range_list.offset =
				    cmd->pdu_range_list.offset;
				free_range_list(&cmd->pdu_range_list);
			}

			if (!session->oper_param->DataSequenceInOrder) {
				/* DataSequenceInOrder==No, sequence can be out-of-order */
				merge_offset_length(&cmd->seq_range_list,
						    cmd->seq_range_list.
						    offset,
						    cmd->next_burst_len);
			}

			/* reset counters for start of next solicited burst (if any) */
			cmd->next_burst_len = 0;
			cmd->data_sn = 0;

			/* this is the end of a solicited burst,
			 * so current R2T is no longer outstanding
			 * we can now generate another R2T
			 * to ask for remaining solicited data.
			 * (but not until after this data has actually been read)
			 */
			giveback = 1;
		}
	}

	if (cmd->unsolicited_data_present) {
		/* cdeng, store unsolicited data for queued command */
#ifdef RDMA_MODE
		iser_front_pdu(conn, &rx_pdu_buff, &rx_handle_buff);
		err =
		    save_unsolicited_data(cmd, hdr->offset,
					  (struct pdu_hdr *) hdr,
					  data_buffer, rx_pdu_buff);
#else
		err =
		    save_unsolicited_data(cmd, hdr->offset,
					  (struct pdu_hdr *) hdr,
					  data_buffer, NULL);
#endif
		if (hdr->length == 0) {
			free_pdu = 1;
			/* Want iSER to get its buffer back, but not free the locally
			 * malloced TCP_ISER mode buffer in free_targ_buf */
			data_buffer = NULL;
		}
		goto end_handle_data;
	} else if (cmd->state != ISCSI_BUFFER_RDY) {
		TRACE_ERROR
		    ("%lu %s in handle_data CmdSN %u ITT %u got state %u "
		     "expected %u\n", jiffies, current->comm, cmd->cmd_sn,
		     cmd->init_task_tag, cmd->state, ISCSI_BUFFER_RDY);
		err = -1;
		free_pdu = 1;
		goto end_handle_data;
	}

	/* receive this data */
	st_list = (struct scatterlist *) cmd->cmnd->req->sr_buffer;

	if (session->oper_param->DataPDUInOrder
	    && session->oper_param->DataSequenceInOrder) {
		/* everything is in order, pick up where we left off after reading data
		 * from DataOut pdu (or from data segment of a WRITE containing
		 * immediate data)
		 */
		offset = cmd->scatter_list_offset;
		st_list += cmd->scatter_list_count;
		TRACE(TRACE_DEBUG, "scatter list offset %u, count %u\n",
		      cmd->scatter_list_offset, cmd->scatter_list_count);
	} else {
		offset = skip_thru_sg_list(st_list, &i, hdr->offset);
		st_list += i;
		TRACE(TRACE_DEBUG, "scatter list offset %u, index %u\n",
		      offset, i);
	}

	TRACE(TRACE_DEBUG, "handle_data: receiving data for cmd_sn %.8x "
	      "init_task_tag %.8x target_xfer_tag %.8x\n",
	      cmd->cmd_sn, cmd->init_task_tag, cmd->target_xfer_tag);

	/* read data directly into midlevel buffers */
	if ((err = read_list_data_seg((struct pdu_hdr *) hdr, cmd, st_list,
				      offset, data_buffer)) <= 0) {
		free_pdu = 1;
		goto end_handle_data;
	}

	/* Minti */
	/* error recovery ver ref18_04 - SAI */
	if (cmd->first_data_q)
		search_data_q(cmd);

	if (cmd->data_done >= cmd->data_length) {
		/* we have just finished reading everything the command expected */
		if (!(hdr->flags & F_BIT)) {
			TRACE_ERROR
			    ("%s DataOut ITT %u DataSN %u F-bit=0 but data done "
			     "%u >= expected data length %u\n",
			     current->comm, hdr->init_task_tag, data_sn,
			     cmd->data_done, cmd->data_length);
			hdr->flags |= F_BIT;	/* so we wakeup tx thread below */
		}

		if (!session->oper_param->DataSequenceInOrder) {
			cmd->seq_range_list.offset = 0;
			cmd->seq_range_list.limit = cmd->data_done;
			check_range_list_complete(&cmd->seq_range_list,
						  cmd->cmd_sn,
						  cmd->init_task_tag);
			free_range_list(&cmd->seq_range_list);
		}

		/* tell the scsi midlevel that all the expected data has arrived */
		spin_lock_irqsave(&session->cmnd_lock, flags);
		cmd->state = ISCSI_DATA_IN;
		TRACE(TRACE_ISCSI,
		      "%lu %s All data read CmdSN %u ITT %u\n", jiffies,
		      current->comm, cmd->cmd_sn, cmd->init_task_tag);
		err = scsi_rx_data(cmd->cmnd);
		spin_unlock_irqrestore(&session->cmnd_lock, flags);
	}

	/* We are here because we have rxd into midlevel buffers and the buffers
	 * can be freed - Minti */
	free_pdu = 1;

      update_handle_data:
	if (hdr->flags & F_BIT) {
		/* end of DataOut pdu sequence, tx thread may need to do something */
		cmd->outstanding_r2t -= giveback;
		if (atomic_read(&conn->tx_sem.count) <= 0) {
			TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
			      "handle_data\n", jiffies, current->comm);
			up(&conn->tx_sem);
		}
	}

/* Added for Error Recovery - SAI */
      end_handle_data:
#ifdef RDMA_MODE
	/* If there is data attached */
	if (free_pdu) {
		free_targ_buf(conn, &buffer, &data_buffer);
	}
#endif

	/* check that the command is good after giving it data. */
	cmnd_check_magic(session, cmd);

	/* update last time this command got some data back from initiator */
	if (cmd)
		cmd->timestamp = jiffies;

	TRACE(TRACE_ENTER_LEAVE, "Leave handle_data, err = %d\n", err);
	return err;
}

/*
 * executed only by the rx thread.
 * Added by SAI for Handling SNACK requests from Initiator as
 * part of Error Recovery
 * For details, see Draft 20, Section 10.16 SNACK Request
 */
static int handle_snack(struct iscsi_conn *conn,
			struct iscsi_session *session,
			__u8 * buffer, __u8 * data_buffer)
{
	__u32 runlen, begrun, reason = REASON_DATA_SNACK;
	int delta1, delta2;
	struct iscsi_init_snack *pdu = (struct iscsi_init_snack *) buffer;
	struct iscsi_cmnd *cmd = NULL;
	struct iscsi_cmnd *related_command;
	struct list_head *list_ptr, *list_temp;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_snack\n");

	if (TRACE_TEST(TRACE_ERROR_RECOVERY))
		print_init_snack(pdu);

	pdu->flags &= SNACK_TYPE;
	/* length should be 0 on a SNACK -- in any case we ignore it */
	pdu->length = be32_to_cpu(pdu->length);
	pdu->init_task_tag = be32_to_cpu(pdu->init_task_tag);
	pdu->target_xfer_tag = be32_to_cpu(pdu->target_xfer_tag);
	pdu->exp_stat_sn = be32_to_cpu(pdu->exp_stat_sn);
	begrun = be32_to_cpu(pdu->begrun);
	runlen = be32_to_cpu(pdu->runlen);

	if (pdu->length > 0) {
		TRACE_ERROR("Data attached to SNACK request ITT %u TTT %u"
			    "ExpStatSN %u\n", pdu->init_task_tag,
			    pdu->target_xfer_tag, pdu->exp_stat_sn);
	}

	/* SNACK has no CmdSN but does have an ExpStatSN so use that now */
	ack_sent_cmnds(conn, NULL, pdu->exp_stat_sn, 0);

	if (!session->oper_param->ErrorRecoveryLevel) {
		TRACE_WARNING
		    ("%s Got SNACK type %u when ErrorRecoveryLevel=0\n",
		     current->comm, pdu->flags);
	}

	switch (pdu->flags) {

	case DATA_R2T_SNACK:
		{
			if (pdu->init_task_tag == ALL_ONES) {
				TRACE_ERROR
				    ("%s Data/R2T SNACK with ITT 0x%08x, reject "
				     "SNACK\n", current->comm,
				     pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag != ALL_ONES) {
				TRACE_ERROR
				    ("%s Data/R2T SNACK with TTT %u, reject SNACK\n",
				     current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else
			    if (!
				(session->
				 targ_snack_flg & DATA_SNACK_ENABLE)
&& !session->oper_param->ErrorRecoveryLevel) {
				TRACE_ERROR
				    ("%s Data/R2T SNACK not enabled, reject SNACK\n",
				     current->comm);
				goto out_reject;
			} else if (session->
				   targ_snack_flg & DATA_SNACK_REJECT) {
				TRACE(TRACE_ERROR_RECOVERY,
				      "handle_snack: Send Data SNACK Reject for SNACK\n");
				goto out_reject;
			} else {
				TRACE_WARNING
				    ("%s Got Data/R2T SNACK for ITT %u begrun %u "
				     "runlen %u\n", current->comm,
				     pdu->init_task_tag, begrun, runlen);
			}
		}
		break;

	case STATUS_SNACK:
		{
			if (pdu->init_task_tag != ALL_ONES) {
				TRACE_ERROR
				    ("%s Status SNACK with ITT %u, reject SNACK\n",
				     current->comm, pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag != ALL_ONES) {
				TRACE_ERROR
				    ("%s Status SNACK with TTT %u, reject SNACK\n",
				     current->comm, pdu->target_xfer_tag);
				goto out_reject;
			} else
			    if (!
				(session->
				 targ_snack_flg & STATUS_SNACK_ENABLE)
&& !session->oper_param->ErrorRecoveryLevel) {
				TRACE_ERROR
				    ("%s Status SNACK not enabled, reject SNACK\n",
				     current->comm);
				goto out_reject;
			} else {
				TRACE_WARNING("%s Got Status SNACK\n",
					      current->comm);
			}
		}
		break;

	case DATACK_SNACK:
		{
			if (pdu->init_task_tag != ALL_ONES) {
				TRACE_ERROR
				    ("%s DataAck SNACK with ITT %u, expected 0x%08x, "
				     "reject SNACK\n", current->comm,
				     pdu->init_task_tag, ALL_ONES);
				goto out_reject;
			} else if (pdu->target_xfer_tag == ALL_ONES) {
				TRACE_ERROR
				    ("%s DataACK SNACK with TTT 0x%08x, reject "
				     "SNACK\n", current->comm,
				     pdu->target_xfer_tag);
				goto out_reject;
			} else if (runlen != 0) {
				TRACE_ERROR
				    ("%s DataACK SNACK with RunLength %u, reject "
				     "SNACK\n", current->comm, runlen);
				goto out_reject;
			} else {
				TRACE(TRACE_ERROR_RECOVERY,
				      "%s Got DataACK SNACK TTT %u\n",
				      current->comm, pdu->target_xfer_tag);
			}
			/* try to find a command on this conn with this TTT */
			related_command =
			    search_by_ttt(conn, session,
					  pdu->target_xfer_tag);
			if (related_command) {
				if (related_command->state ==
				    ISCSI_AWAIT_DATA_ACK) {
					delta1 = related_command->data_sn;
					if (delta1 != begrun) {
						TRACE_WARNING
						    ("%s DataACK SNACK TTT %u has BegRun %u, "
						     "expected %u\n",
						     current->comm,
						     pdu->target_xfer_tag,
						     begrun, delta1);
					}
					related_command->retransmit_flg =
					    0;
					move_to_next_seq(related_command,
							 session);
					related_command->state =
					    ISCSI_GOT_DATA_ACK;
					/* wake up the transmit thread */
					if (atomic_read
					    (&related_command->conn->
					     tx_sem.count) <= 0) {
						TRACE(TRACE_SEM,
						      "%lu %s up on tx_sem in "
						      "handle_snack\n",
						      jiffies,
						      current->comm);
						up(&related_command->conn->
						   tx_sem);
					}
				}
			} else {
				TRACE_WARNING
				    ("%s DataACK SNACK with ITT %u matches no "
				     "command\n", current->comm,
				     pdu->target_xfer_tag);
			}
			goto out;
		}

	case R_DATA_SNACK:
		{
			if (pdu->init_task_tag == ALL_ONES) {
				TRACE_ERROR
				    (" %s R-Data SNACK with ITT 0x%08x, reject "
				     "SNACK\n", current->comm,
				     pdu->init_task_tag);
				goto out_reject;
			} else if (pdu->target_xfer_tag == 0
				   || pdu->target_xfer_tag == ALL_ONES) {
				TRACE_ERROR
				    ("%s R-Data SNACK with TTT 0x%08x, reject "
				     "SNACK\n", current->comm,
				     pdu->target_xfer_tag);
				goto out_reject;
			} else if (begrun != 0) {
				TRACE_ERROR
				    ("%s R-Data SNACK with BegRun %u, reject SNACK\n",
				     current->comm, begrun);
				goto out_reject;
			} else if (runlen != 0) {
				TRACE_ERROR
				    ("%s R-Data SNACK with RunLength %u, reject "
				     "SNACK\n", current->comm, runlen);
				goto out_reject;
			}
			TRACE_ERROR
			    ("%s R-Data SNACK, type %u, ITT %u, not implemented, "
			     "reject SNACK\n", current->comm, pdu->flags,
			     pdu->init_task_tag);
			goto out_reject;
		}

	default:
		{
			TRACE_ERROR
			    ("%s Invalid SNACK type %u, ITT %u, reject SNACK\n",
			     current->comm, pdu->flags,
			     pdu->init_task_tag);
			goto out_reject;
		}
	}			/* switch */


	/* here only if SNACK type is DATA_R2T_SNACK or STATUS_SNACK */
	spin_lock_irqsave(&session->cmnd_lock, flags);
	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp,
			   &session->active_command_list) {
		cmd =
		    list_entry(list_ptr, struct iscsi_cmnd, command_link);
		BUMP_LOOP_COUNTER(count, "active_command_list");
		delta1 = cmd->stat_sn - begrun;
		delta2 = delta1 - runlen + 1;
		if (cmd->init_task_tag == pdu->init_task_tag
		    && pdu->flags == DATA_R2T_SNACK
		    && !cmd->retransmit_flg) {
			if (cmd->state == ISCSI_SENT
			    || cmd->state == ISCSI_AWAIT_DATA_ACK) {
				/* Command was a READ, SNACK must be to retransmit DataIn pdus */
				if (cmd->data_sn >= begrun) {
					/* have at least 1 DataIn pdu to retransmit for this READ */
					TRACE(TRACE_ERROR_RECOVERY,
					      "handle_snack: Re-Transmit Data on SNACK Request\n");
					cmd->retransmit_flg = 1;
					/* set limits for which DataIn pdus should be resent */
					cmd->startsn = begrun;
					if (!runlen)
						cmd->endsn = cmd->data_sn;
					else
						cmd->endsn =
						    begrun + runlen - 1;

					/* backup to start of current sequence or whole command */
					restore_data_in_seq(cmd);

					/* signal the tx thread to do the retransmission */
					cmd->state = ISCSI_REDO_DATA;
					printk
					    ("%s backup ITT %u due to Data SNACK, "
					     "startsn %u, endsn %u "
					     "state %d, DataSN %u, "
					     "list_count %u, list_offset %u\n",
					     current->comm,
					     cmd->init_task_tag,
					     cmd->startsn, cmd->endsn,
					     cmd->state, cmd->data_sn,
					     cmd->scatter_list_count,
					     cmd->scatter_list_offset);
					if (atomic_read
					    (&cmd->conn->tx_sem.count) <=
					    0) {
						TRACE(TRACE_SEM,
						      "%lu %s up on tx_sem in "
						      "handle_snack\n",
						      jiffies,
						      current->comm);
						up(&cmd->conn->tx_sem);
					}
				}
			} else if (cmd->state == ISCSI_BUFFER_RDY) {
				/* Command was a WRITE, SNACK must be to retransmit R2T pdus */
				if (cmd->r2t_sn >= begrun) {
					/* Found the corresponding R2T PDU - re-transmit it */
					TRACE(TRACE_ERROR_RECOVERY,
					      "handle_snack: Re-Transmit R2T on SNACK Request\n");

					cmd->retransmit_flg = 1;
					cmd->startsn = begrun;
					if (!runlen) {
						cmd->endsn = cmd->r2t_sn;
					} else
						cmd->endsn =
						    begrun + runlen - 1;
					/* signal the tx thread to do the retransmission */
					if (atomic_read
					    (&cmd->conn->tx_sem.count) <=
					    0) {
						TRACE(TRACE_SEM,
						      "%lu %s up on tx_sem in "
						      "handle_snack\n",
						      jiffies,
						      current->comm);
						up(&cmd->conn->tx_sem);
					}
				}
			} else {
				/* RFC 3720 Section 10.16. SNACK Request
				 * "Any SNACK that requests a numbered-resonse, Data, or R2T
				 * that was not sent by the target or was already acknowledged
				 * by the initiator, MUST be rejected with a reason code of
				 * 'Protocol error'."
				 */
				reason = REASON_PROTOCOL_ERR;
				printk
				    ("%s opcode 0x%02x ITT %d TTT %d data_done %d "
				     "xfer length %d stat_sn %u data_sn %u state %u\n",
				     current->comm, cmd->opcode_byte,
				     cmd->init_task_tag,
				     cmd->target_xfer_tag, cmd->data_done,
				     cmd->data_length, cmd->stat_sn,
				     cmd->data_sn, cmd->state);
				spin_unlock_irqrestore(&session->cmnd_lock,
						       flags);
				goto out_reject;
			}
			goto out_unlock;
		} else if (pdu->flags == STATUS_SNACK
			   && delta1 >= 0 && (runlen == 0
					      || delta2 <= 0)) {
			if ((cmd->opcode_byte & ISCSI_OPCODE) ==
			    ISCSI_INIT_TEXT_CMND) {
				TRACE(TRACE_ERROR_RECOVERY,
				      "Retransmit Text Response on SNACK Request\n");
				cmd->reply_state = cmd->state;
				cmd->state = ISCSI_RESEND_TEXT_RESPONSE;
				if (atomic_read(&cmd->conn->tx_sem.count)
				    <= 0) {
					TRACE(TRACE_SEM,
					      "%lu %s up on tx_sem in "
					      "handle_snack\n", jiffies,
					      current->comm);
					up(&cmd->conn->tx_sem);
				}
			} else if ((cmd->opcode_byte & ISCSI_OPCODE) ==
				   ISCSI_INIT_NOP_OUT) {
				TRACE(TRACE_ERROR_RECOVERY,
				      "Retransmit NopIn on SNACK Request\n");
				cmd->reply_state = cmd->state;
				cmd->state = ISCSI_RESEND_NOP_IN;
				if (atomic_read(&cmd->conn->tx_sem.count)
				    <= 0) {
					TRACE(TRACE_SEM,
					      "%lu %s up on tx_sem in "
					      "handle_snack\n", jiffies,
					      current->comm);
					up(&cmd->conn->tx_sem);
				}
			} else if (cmd->state == ISCSI_SENT) {
				/* Found the corresponding Status PDU - re-transmit it */
				TRACE(TRACE_ERROR_RECOVERY,
				      "Retransmit iscsi Response on SNACK Request\n");
				cmd->retransmit_flg = 1;
				cmd->state = ISCSI_RESEND_STATUS;
				if (atomic_read(&cmd->conn->tx_sem.count)
				    <= 0) {
					TRACE(TRACE_SEM,
					      "%lu %s up on tx_sem in "
					      "handle_snack\n", jiffies,
					      current->comm);
					up(&cmd->conn->tx_sem);
				}
			}
			goto out_unlock;
		}
	}

	/* RFC 3720 Section 10.16. SNACK Request
	 * "Any SNACK that requests a numbered-resonse, Data, or R2T
	 * that was not sent by the target or was already acknowledged
	 * by the initiator, MUST be rejected with a reason code of
	 * 'Protocol error'."
	 */
	reason = REASON_PROTOCOL_ERR;
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
	goto out_reject;

      out_unlock:
	spin_unlock_irqrestore(&session->cmnd_lock, flags);

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_snack\n");
	return 0;

      out_reject:
	enqueue_reject(conn, reason);
	goto out;
}

static int handle_iscsi_rx_data(struct iscsi_conn *conn, __u8 ** pdu_bhs,
				__u8 ** data_buffer)
{
	int retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_iscsi_rx_data\n");

	/* no buffer allocated yet for pdu payload, and won't be if pdu's DSL = 0 */
	*data_buffer = NULL;

#ifdef ISER
	if (conn->session->oper_param->RDMAExtensions) {
#ifdef TCP_ISER
		retval = do_iscsi_rx_data(conn, pdu_bhs, data_buffer);
#else
		retval =
		    recv_target_control_notifications(conn, pdu_bhs,
						      data_buffer);
#endif
	} else
#endif
	{
		retval = do_iscsi_rx_data(conn, pdu_bhs, data_buffer);
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_iscsi_rx_data\n");
	return retval;
}

/* called only from handle_login after login negotiation for new session
 * to allocate session-wide buffers for unsolicited data (if needed)
 * so there should be no active commands or connections in this session
 * allocate as many as possible up to computed maximum
 */
static void setup_unsolicited_buffers(struct iscsi_session *session)
{
	struct session_operational_parameters *oper_param_entry;
	__u32 size, fbl, min_fbl, mrdsl, min_mrdsl, nbufs, i;
	struct parameter_type *p;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	oper_param_entry = session->oper_param;

	if (oper_param_entry->ImmediateData
	    || !oper_param_entry->InitialR2T) {
		/* Immediate data or unsolicited DataOut pdus or both are allowed */
		/* each pdu can carry min(FirstBurstLength, MaxRecvDSL) bytes of data */

		/* min_fbl is min(FirstBurstLength, MAX_KMALLOC_SIZE) */
		min_fbl = fbl = oper_param_entry->FirstBurstLength;
		if (min_fbl > MAX_KMALLOC_SIZE)
			min_fbl = MAX_KMALLOC_SIZE;

		/* min_mrdsl is min(MaxRecvDataSegmentLength, MAX_KMALLOC_SIZE) */
		if ((p = find_flag_parameter(MAXRECVDATASEGMENTLENGTH_FLAG,
					     *session->session_params))) {
			min_mrdsl = mrdsl = p->int_value;
			if (min_mrdsl > MAX_KMALLOC_SIZE)
				min_mrdsl = MAX_KMALLOC_SIZE;
		} else
			/* should never happen */
			min_mrdsl = mrdsl = MAX_KMALLOC_SIZE;

		/* size is number of bytes per buffer = min(FirstBurstLength,
		   MaxRecvDataSegmentLength, MAX_KMALLOC_SIZE) */
		size = min_fbl;
		if (size > min_mrdsl)
			size = min_mrdsl;

		if ((session->unsolicited_buffer_size = size)) {
			/* first compute in nbufs the number of buffers per command */
			if (mrdsl >= fbl)
				/* everything can fit into one pdu */
				nbufs = (fbl + min_fbl - 1) / min_fbl;
			else if (!oper_param_entry->InitialR2T)
				/* unsolicited DataOut pdus can be used */
				nbufs = (fbl + min_mrdsl - 1) / min_mrdsl;
			else
				/* only immediate data can be used, so only one pdu */
				nbufs =
				    (mrdsl + min_mrdsl - 1) / min_mrdsl;

			/* now compute in nbufs the total no. of buffers for all commands */
			nbufs *= (session->max_queue_depth_allowed + 1)
			    * oper_param_entry->MaxConnections;

			/* allocate buffers and thread udata items onto session's list */
			for (i = 0; i < nbufs; i++) {
				if (!alloc_one_unsolicited_buffer(session))
					break;
			}
		}
	}
	if (session->unsolicited_list_size) {
		printk("%s allocate %u udata buffers of %u bytes each\n",
		       current->comm, session->unsolicited_list_size,
		       session->unsolicited_buffer_size);
	}
}

static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_by_ttt(struct iscsi_conn *conn, struct iscsi_session *session,
	      __u32 this_ttt)
{
	struct iscsi_cmnd *cmd = NULL;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

	spin_lock_irqsave(&session->cmnd_lock, flags);
	INIT_LOOP_COUNTER(count);
	list_for_each_entry(cmd, &session->active_command_list,
			    command_link) {
		BUMP_LOOP_COUNTER(count, "active_command_list");
		if (cmd->conn == conn && cmd->target_xfer_tag == this_ttt) {
			goto out;
		}
	}
	cmd = NULL;
      out:
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
	return cmd;
}

/*
 * executed only by the rx thread.
 * read data segment for command into list of buffers at given offset
 * this can be either immediate data in a write, or the payload of a DataIn
 * returns > 0 if ok, < 0 if error, = 0 if error recovery completed ok
 */
static int read_list_data_seg(struct pdu_hdr *hdr, struct iscsi_cmnd *cmd,
			      struct scatterlist *st_list, __u32 offset,
			      __u8 * data_buffer)
{
	struct targ_error_rec err_rec;
	struct iovec *iov, *iov_copy;
	struct iovec *local_iov = NULL, *local_iov_copy = NULL;
	int size, niov, padding, err, sgindex;
	__u32 digest, pad_bytes;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if ((size = hdr->length) <= 0) {
		update_after_read(hdr, cmd);
		return 1;
	}
	niov = find_iovec_needed(size, st_list, offset);
	if (niov <= 0) {
		TRACE_ERROR("%s Trouble in find_iovec_needed\n",
			    current->comm);
		return -1;
	}

	if (niov + 2 > MAX_IOV_SLOTS) {
		TRACE_WARNING
		    ("%s CmdSN %u ITT %u needs %d iov slots, command max %d\n",
		     current->comm, cmd->cmd_sn, cmd->init_task_tag,
		     niov + 2, MAX_IOV_SLOTS);
		iov = local_iov =
		    my_kmalloc((niov + 2) * sizeof(struct iovec), "iovec");
		iov_copy = local_iov_copy =
		    my_kmalloc((niov + 2) * sizeof(struct iovec), "iovec");
		if (!local_iov || !local_iov_copy) {
			err = -ENOMEM;
			goto out1;
		}
	} else {
		/* use the io vector preallocated in the command structure */
		iov = cmd->rx_iov;
		iov_copy = cmd->rx_iov_copy;
	}

	/* gives back number of st_list elements used */
	sgindex = fill_iovec(iov, 0, niov, st_list, &offset, size);

	padding = (-size) & 3;
	if (padding) {
		iov[niov].iov_base = &pad_bytes;
		iov[niov].iov_len = padding;
		niov++;
		size += padding;
	}

	if (cmd->conn->data_crc) {
		iov[niov].iov_base = &digest;
		iov[niov].iov_len = CRC_LEN;
		niov++;
		size += CRC_LEN;
	}
#ifdef ISER
	/* Minti */
	if (cmd->conn->session->oper_param->RDMAExtensions) {
#ifdef TCP_ISER
		err = iscsi_rx_data(cmd->conn, iov, iov_copy, niov, size);
#else
		TRACE(TRACE_ISER, "Doing memcpy in read_list_data_seg\n");
		memcpy(iov[0].iov_base, data_buffer, size);
		err = size;
#endif
	} else
#endif
	{
		err = iscsi_rx_data(cmd->conn, iov, iov_copy, niov, size);
	}

	if (err == size) {
		/* we received everything we expected to receive
		 * Store scatter list count and offset for recovery purposes - SAI
		 */
		cmd->scatter_list_count += sgindex;
		cmd->scatter_list_offset = offset;
		update_after_read(hdr, cmd);
	} else {
		/* Payload Digest Error Recovery - SAI */
		if (err == PAYLOAD_DIGERR) {
			TRACE(TRACE_ERROR_RECOVERY,
			      "Start payload digest error recovery\n");
			err_rec.curr_conn = cmd->conn;
			err_rec.pdu_hdr = hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = PAYLOAD_DIGERR;
			err = targ_do_error_recovery(&err_rec);
		}
	}
      out1:
	my_kfree((void **) &local_iov_copy, "iovec");
	my_kfree((void **) &local_iov, "iovec");
	return err;
}

/* Minti */
/*
 * Called only from handle_iscsi_rx_data()
 * For efficiency, we still want to read immediate and unsolictd SCSI write data
 * directly into midlevel buffers (for both TCP_ISER mode and non-iSER mode).
 * This will become a copy to the mid-level buffers once iSER is in place.
 */
static int __attribute__ ((no_instrument_function))
    do_iscsi_rx_data(struct iscsi_conn *conn, __u8 ** pdu_bhs,
		 __u8 ** data_buffer)
{
	struct pdu_hdr *pdu;
	__u32 hdr_crc;		/* compute PDU header digest in here */
	__u32 digest;		/* read 4-byte header digest in here */
	struct targ_error_rec err_rec;
	int len, err;
	__u32 opcode;
	struct iovec iov, iov_copy;
	int retval = 1;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter do_iscsi_rx_data\n");
	iov.iov_base = *pdu_bhs;
	iov.iov_len = ISCSI_HDR_LEN;

	if (iscsi_rx_data(conn, &iov, &iov_copy, 1, ISCSI_HDR_LEN) !=
	    ISCSI_HDR_LEN) {
		retval = -1;
		goto out;
	}

	pdu = (struct pdu_hdr *) iov.iov_base;
	len = be32_to_cpu(pdu->length);
	opcode = pdu->opcode & ISCSI_OPCODE;

	if (conn->hdr_crc) {
		/* read and check header digest on this pdu */
		TRACE(TRACE_DEBUG,
		      "iscsi_rx_thread: Header digest check\n");
		hdr_crc = 0;
		do_crc((__u8 *) pdu, ISCSI_HDR_LEN, &hdr_crc);

		iov.iov_len = CRC_LEN;
		iov.iov_base = &digest;

		if ((err =
		     iscsi_rx_data(conn, &iov, &iov_copy, 1,
				   CRC_LEN)) != CRC_LEN) {
			retval = -1;
			goto out;
		}

		if (hdr_crc != digest) {
			TRACE_ERROR
			    ("Expected header crc 0x%08x, got 0x%08x\n",
			     ntohl(hdr_crc), ntohl(digest));
			/* Error Recovery Code Added by SAI */
			err_rec.curr_conn = conn;
			err_rec.pdu_hdr = pdu;
			err_rec.pdu_hdr->length =
			    be32_to_cpu(err_rec.pdu_hdr->length);
			err_rec.err_type = HEADER_DIGERR;
			if ((err = targ_do_error_recovery(&err_rec)) < 0) {
				retval = -1;
				goto out;
			}
			retval = 0;

		} else {
			TRACE(TRACE_ISCSI_FULL,
			      "%lu %s Got header crc 0x%08x\n", jiffies,
			      current->comm, ntohl(digest));
		}
	}

	if (len > 0) {
		if ((opcode != ISCSI_INIT_SCSI_CMND) &&
		    (opcode != ISCSI_INIT_SCSI_DATA_OUT)) {
			/* read length bytes of text data into a newly-allocated buffer */
			err =
			    read_single_data_seg((__u8 *) pdu, conn, len,
						 NULL, data_buffer);
			/* Will indicate #of bytes read */
			if (err <= 0)
				retval = err;
		}
	}
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave do_iscsi_rx_data\n");
	return retval;
}

/*
 * executed only by the rx thread.
 * cdeng: called by handle_cmnd and handle_data if the
 * command is out of order.  Store the data into udata_list in
 * iSCSI layer rather than the st_list in SCSI layer.
 * Returns 0 if all ok, -1 on any error (no memory for buffers).
 */
static int save_unsolicited_data(struct iscsi_cmnd *cmnd, __u32 offset,
				 struct pdu_hdr *hdr, __u8 * data_buf,
				 void *iser_buf)
{
	int err;
	struct udata_list *udata;
	__u32 total_length, length;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	total_length = hdr->length;

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter save_unsolicited_data offset %u length " "%u\n",
	      current->comm, offset, total_length);

	while ((length = total_length)) {
		if (length > cmnd->conn->session->unsolicited_buffer_size)
			length =
			    cmnd->conn->session->unsolicited_buffer_size;

		if (!(udata = get_unsolicited_buffer(cmnd->conn->session)))
			return -1;

		udata->offset = offset;
		udata->length = length;

#ifdef ISER
		if (cmnd->conn->session->oper_param->RDMAExtensions) {
			/* Minti */
#ifdef TCP_ISER
			err =
			    read_single_data_seg((__u8 *) hdr, cmnd->conn,
						 length, udata->buffer,
						 NULL);
			if (err <= 0) {
				put_unsolicited_buf(udata,
						    cmnd->conn->session);
				return err;
			}
#else
			/* Save data_buf into data->buffer - dont need to
			 * worry about padding since the PDU has already been read
			 */
			TRACE(TRACE_ISER,
			      "Saving data_buf in save_unsolicited_data\n");
			udata->buffer = data_buf;
			udata->rx_iser_pdu = iser_buf;
#endif
		} else
#endif
		{
			/* receive unsolicited data into a pre-allocated buffer */
			TRACE(TRACE_ISCSI_FULL,
			      "%lu %s ITT %u recv udata %p buffer %p "
			      "offset " "%u length %u\n", jiffies,
			      current->comm, cmnd->init_task_tag, udata,
			      udata->buffer, offset, length);
			err =
			    read_single_data_seg((__u8 *) hdr, cmnd->conn,
						 length, udata->buffer,
						 NULL);
			if (err <= 0) {
				put_unsolicited_buf(udata,
						    cmnd->conn->session);
				return err;
			}
		}

		/* add the new buffer to end of list of unsolicted data buffers */
		list_add_tail(&udata->udata_link,
			      &cmnd->unsolicited_data_list);

		offset += length;
		total_length -= length;
	}

	update_after_read(hdr, cmnd);


	if (atomic_read(&cmnd->conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "save_unsolicited_data\n", jiffies, current->comm);
		up(&cmnd->conn->tx_sem);
	}

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Leave save_unsolicited_data offset %u length %u"
	      " unsolicited data present %u\n", current->comm, offset,
	      total_length, cmnd->unsolicited_data_present);
	return 0;
}

/**
 * Searches the active command list on the given session for a
 * command with the given cmd_sn.
 */
/*
static struct iscsi_cmnd *find_cmnd_by_cmdsn(struct iscsi_session *sess,
					     __u32 cmd_sn)
{
	long flags;
	struct iscsi_cmnd *c;
	struct list_head *node;

	spin_lock_irqsave(&sess->cmnd_lock, flags);
	list_for_each(node, &sess->active_command_list) {
		c = list_entry(node, struct iscsi_cmnd, command_link);
		if (c->cmd_sn == cmd_sn)
			break;
	}
	spin_unlock_irqrestore(&sess->cmnd_lock, flags);

	if (c->cmd_sn == cmd_sn)
		return c;
	return NULL;
}
*/

/*
 * executed only by the rx thread.
 * called only by handle_cmnd() when processing an ISCSI Command pdu
 * when called, err is +1 if cmnd is in range but out of order
 *				err is -1 if cmnd is out of range and should be ignored
 * returns	0 if command is queued ok or freed because it was out of range
 *			-1 if there was an error (no memory to save immediate data)
 */
static int out_of_order_cmnd(struct iscsi_conn *conn,
			     struct iscsi_session *session, __u8 * buffer,
			     struct iscsi_cmnd *cmnd, int err,
			     __u8 * data_buffer)
{
	struct iscsi_init_scsi_cmnd *hdr =
	    (struct iscsi_init_scsi_cmnd *) buffer;
	int retval = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	memcpy(cmnd->hdr, buffer, ISCSI_HDR_LEN);
	cmnd->state = ISCSI_QUEUE_CMND;
#ifdef RDMA_MODE
	iser_front_pdu(conn, &cmnd->rx_iser_pdu, &cmnd->iser_task_h);
#endif

	/* last param is 0 (don't queue) if cmnd is out of range (err == -1),
	 * 2 (queue) if cmnd is in range but out of order (err == +1)
	 */
	ack_sent_cmnds(conn, cmnd, hdr->exp_stat_sn, err + 1);

	if (err < 0) {
		/* out of range, silently ignore it */
		TRACE_ERROR
		    ("%s ignore out of range CmdSN %u, ExpCmdSN %u, ITT %u, "
		     "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		     session->exp_cmd_sn, cmnd->init_task_tag,
		     cmnd->opcode_byte);
		/* command was not put into queue */
		free_udata_list(cmnd);
		put_free_command(cmnd, session);
	} else {
#if 0
		struct iscsi_cmnd *exp_cmd;
#endif

		/* within range but out of order, we queued it for later (above) */
		TRACE(TRACE_ISCSI,
		      "%lu %s o-o-o CmdSN %u exceeds ExpCmdSN %u\n",
		      jiffies, current->comm, cmnd->cmd_sn,
		      cmnd->session->exp_cmd_sn);

#if 0
		/*
		 *
		 * Lets see what the command with the ExpCmdSN is
		 * doing right now
		 *
		 */
		exp_cmd = find_cmnd_by_cmdsn(cmnd->session,
					     cmnd->session->exp_cmd_sn);
		if (exp_cmd) {
			cmnd_print(exp_cmd);
		} else {
			TRACE_WARNING("%s: Command with exp_cmd_sn is not "
				      "active\n", __func__);
		}
#endif

#ifdef RDMA_MODE
		/* If there is data attached, the data list structure points to it in
		 * save_unsolicited_data. Otherwise, we have already made a copy of
		 * the bhs with memcpy above
		 */
		if (hdr->length == 0)
			targ_release_control(conn, &buffer, &data_buffer);
#endif
	}
	return retval;
}

/* Called when DataPDUInOrder==No (so DataOut PDU can be out-of-order)
 * to merge the offset and length of the DataOut pdu pointed to by hdr
 * into the WRITE cmd's pdu_range_list.
 * Maintains in pdu_range_list.offset the smallest offset of any pdu so far.
 * Maintains in pdu_range_list.limit the largest limit of any pdu so far.
 */
static void __attribute__ ((no_instrument_function))
    merge_out_of_order(struct iscsi_init_scsi_data_out *hdr,
		   struct iscsi_cmnd *cmd)
{
	__u32 limit;

	if (cmd->pdu_range_list.offset > hdr->offset)
		cmd->pdu_range_list.offset = hdr->offset;
	limit = cmd->pdu_range_list.offset + hdr->length;
	if (cmd->pdu_range_list.limit < limit)
		cmd->pdu_range_list.limit = limit;
	merge_offset_length(&cmd->pdu_range_list, hdr->offset,
			    hdr->length);
}

/* restore values to start of current DataIn sequence or whole READ command */
static void __attribute__ ((no_instrument_function))
    restore_data_in_seq(struct iscsi_cmnd *cmd)
{
	cmd->data_sn = cmd->start_data_sn;
	cmd->data_length_left = cmd->start_data_length_left;
	cmd->seq_low_byte = cmd->start_low_byte;
	cmd->seq_limit_byte = cmd->start_limit_byte;
	cmd->scatter_list_count = cmd->start_sg_list_count;
	cmd->scatter_list_offset = cmd->start_sg_list_offset;
	cmd->seq_flags &= ~LAST_SEQ_FLAG;
}

/* called only from save_unsolicited_data() in rx thread
 * return first free udata item from session's unsolicited data list
 * (and remove it from the list)
 * return NULL if list is empty
 */
static struct udata_list *get_unsolicited_buffer(struct iscsi_session
						 *session)
{
	struct udata_list *udata = NULL;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	spin_lock_irqsave(&session->unsolicited_data_lock, flags);

	if (list_empty(&session->unsolicited_data_list)) {
		udata = alloc_one_unsolicited_buffer(session);
		if (!udata) {
			TRACE_ERROR
			    ("%s All %u udata buffers in use, can't allocate more\n",
			     current->comm,
			     session->unsolicited_list_size);
			goto out_unlock;
		}
		TRACE_WARNING
		    ("%s allocate additional udata buffer, now have %u\n",
		     current->comm, session->unsolicited_list_size);
	}

	udata =
	    list_entry(session->unsolicited_data_list.next,
		       struct udata_list, udata_link);
	list_del(&udata->udata_link);
	TRACE(TRACE_BUF, "%s get udata %p, buffer %p\n",
	      current->comm, udata, udata->buffer);

      out_unlock:
	spin_unlock_irqrestore(&session->unsolicited_data_lock, flags);
	return udata;
}

/* called after reading unsolicited data attached to a WRITE SCSI command pdu,
 * or solicited data attached to a DataOut pdu.
 */
static void __attribute__ ((no_instrument_function))
    update_after_read(struct pdu_hdr *hdr, struct iscsi_cmnd *cmnd)
{
	__update_after_read(hdr->length, hdr->flags & F_BIT, cmnd);
}

void __attribute__ ((no_instrument_function))
    __update_after_read(unsigned int length, unsigned int fbit,
		    struct iscsi_cmnd *cmnd)
{
	cmnd->data_done += length;
	if (fbit) {
		/* end of this (solicited or unsolicited) sequence,
		 * reset the counters for the next solicited burst (if any)
		 */
		cmnd->data_sn = 0;
		cmnd->unsolicited_data_present = 0;
	}
}

/* read data segment from cmd's connection into a single buffer
 * that is newly allocated to hold bufsize bytes of data
 * returns > 0 total number of bytes read (including pad & crc) if all ok,
 *		   = 0 if recovered from payload digest error ok
 *		   < 0 error
 * if ok, sets result to newly allocated buffer, else leaves result unchanged
 */
static int
read_single_data_seg(__u8 * buffer, struct iscsi_conn *conn,
		     int bufsize, __u8 * existing_buf, __u8 ** result)
{
	struct iovec iov[3], iov_copy[3];
	int size, niov = 1, padding, err;
	__u32 digest, pad_bytes;
	__u8 *data_buf;
	struct targ_error_rec err_rec;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	size = bufsize;

	padding = (-size) & 3;
	if (padding) {
		iov[niov].iov_base = &pad_bytes;
		iov[niov].iov_len = padding;
		size += padding;
		niov++;
	}

	if (conn->data_crc) {
		iov[niov].iov_base = &digest;
		iov[niov].iov_len = CRC_LEN;
		size += CRC_LEN;
		niov++;
	}

	if (!(data_buf = existing_buf)) {
		data_buf = my_kmalloc(bufsize, "data_buf");
	}
	if (data_buf) {
		iov[0].iov_base = data_buf;
		iov[0].iov_len = bufsize;

		err = iscsi_rx_data(conn, iov, iov_copy, niov, size);

		if (err != size) {
			/* Payload Digest Error Recovery - SAI */
			if (err == PAYLOAD_DIGERR) {
				TRACE(TRACE_ERROR_RECOVERY,
				      "Start payload digest error recovery\n");
				err_rec.curr_conn = conn;
				err_rec.pdu_hdr =
				    (struct pdu_hdr *) buffer;
				/* This function will only be called for non DataOut PDUs,
				 * do payload recovery isn't needed and therefore cmd is NULL
				 */
				err_rec.cmd = NULL;
				err_rec.err_type = PAYLOAD_DIGERR;
				err = targ_do_error_recovery(&err_rec);
			}
			if (!existing_buf)
				my_kfree((void **) &data_buf, "data_buf");
		} else if (result) {
			*result = data_buf;
		}
	} else {
		TRACE_ERROR("%s no buffer\n", current->comm);
		err = -1;
	}
	return err;
}

/*
static void print_dequeue_list(struct list_head *dequeue_list, unsigned int len)
{
	struct iscsi_cmnd *cmnd;
	unsigned int i = 0;

	printk("%s: printing dequeue list %p, len=%u\n", __func__,
	       dequeue_list, len);
	list_for_each_entry(cmnd, dequeue_list, command_link) {
		if (i >= len) {
			TRACE_ERROR("%s: too many items on list\n", __func__);
			break;
		}
		cmnd_print(cmnd);
		i += 1;
	}
}
*/

/*
 * executed only by the rx thread.
 * find commands for this connection already in the queue that have already
 * sent everything and are being ACKed by this command in order to mark them
 * for dequeueing.
 * Then adds this command to end of queue if flag is set.
 * Finally, if any command's state was changed, or if this command was added
 * to the queue, wake up the tx thread.
 * NOTE: cmnd MAY be NULL, and if it is, add_cmnd_to_queue MUST be 0!
 * returns 0 on success, -1 on error (i.e., a duplicate ITT).
 */
static int ack_sent_cmnds(struct iscsi_conn *conn,
			  struct iscsi_cmnd *cmnd,
			  __u32 exp_stat_sn, int add_cmnd_to_queue)
{
	struct iscsi_cmnd *temp;
	int delta, changed_something, retval;
	struct list_head *list_ptr, *list_temp;
	__u32 count;
	struct iscsi_session *session;
	unsigned long flags;

	TRACE(TRACE_ENTER_LEAVE, "%s: entering ack_sent_cmnds\n", __func__);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif


	changed_something = add_cmnd_to_queue;

	session = conn->session;
	spin_lock_irqsave(&session->cmnd_lock, flags);

	count = 0;
	retval = 0;
	list_for_each_safe(list_ptr, list_temp,
			   &session->active_command_list) {
		count++;
		TEST_LOOP_COUNTER(count, "active_command_list");
		temp =
		    list_entry(list_ptr, struct iscsi_cmnd, command_link);
		if (temp->conn == conn) {
			if (temp->state == ISCSI_SENT) {
				delta = temp->stat_sn - exp_stat_sn;
				if (delta < 0) {
					TRACE(TRACE_DEBUG, "%s: "
					      "set dequeue command=%p statsn %d, "
					      "received exp_stat_sn %d, command state %d\n",
					      __func__,
					      temp,
					      temp->stat_sn, exp_stat_sn,
					      temp->state);
					temp->state = ISCSI_DEQUEUE;
					changed_something = 1;
				}
			}
			if (add_cmnd_to_queue
			    && temp->init_task_tag ==
			    cmnd->init_task_tag) {
				retval = -1;
				cmnd->state = ISCSI_DEQUEUE;
				printk
				    ("%s duplicate ITT %u in queue, q-CmdSN %u, "
				     "n-CmdSN %u, expected CmdSN %u, q-increment %u, "
				     "n-increment %u, q-state %d, n-state %d, "
				     "q-opcode 0x%02x, n-opcode 0x%02x, count %u\n",
				     current->comm, cmnd->init_task_tag,
				     temp->cmd_sn, cmnd->cmd_sn,
				     session->exp_cmd_sn,
				     temp->cmd_sn_increment,
				     cmnd->cmd_sn_increment, temp->state,
				     cmnd->state, temp->opcode_byte,
				     cmnd->opcode_byte, count);
				target_info();
			}
		}
	}

	if (add_cmnd_to_queue) {
		/* Add this command to the session's active cmmand queue */
		TRACE(TRACE_DEBUG,
		      "%s add command %p to queue CmdSN %u ITT %u "
		      "state %d count %u\n", current->comm, cmnd,
		      cmnd->cmd_sn, cmnd->init_task_tag, cmnd->state,
		      count);
		count++;
		if (!(cmnd->opcode_byte & I_BIT)) {
			/* this command is not immediate, so we will need to advance
			 * ExpCmdSN when we start working on it, and
			 * MaxCmdSN when we finish working on it
			 */
			cmnd->need_max_cmd_sn_inc = 1;
			cmnd->cmd_sn_increment = 1;
		}
		list_add_tail(&cmnd->command_link,
			      &session->active_command_list);
	}

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	/* tell the tx thread it has something to do */
	if (changed_something) {
		if (atomic_read(&conn->tx_sem.count) <= 0) {
			TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
			      "ack_sent_cmnds, %u cmnds in queue\n",
			      jiffies, current->comm, count);
			up(&conn->tx_sem);
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "%s: leaving ack_sent_cmnds\n", __func__);


	return retval;
}

/* allocate one unsolicited buffer and thread it onto session's list */
static struct udata_list *alloc_one_unsolicited_buffer(struct iscsi_session
						       *session)
{
	struct udata_list *udata;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	udata = my_kmalloc(sizeof(struct udata_list), "udata slot");
	if (!udata)
		return NULL;
	memset(udata, 0, sizeof(struct udata_list));
#ifndef TCP_ISER
	udata->buffer =
	    my_kmalloc(session->unsolicited_buffer_size, "udata buffer");
	if (!udata->buffer) {
		my_kfree((void **) &udata, "udata slot");
		return NULL;
	}
	TRACE(TRACE_BUF, "%s %2u: udata %p, new buffer %p\n",
	      current->comm, session->unsolicited_list_size + 1,
	      udata, udata->buffer);
#endif
	list_add_tail(&udata->udata_link, &session->unsolicited_data_list);
	session->unsolicited_list_size++;
	return udata;
}

/* vi: set tabstop=8 textwidth=78: */
