/*	target/iscsi_target.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This is the iscsi target front-end which interfaces with the STML.
 *	The front-end has been written to the specifications of RFC 3720 for the
 *	iSCSI spec.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 *  written permission.
*/

#if !defined(MANGLE_INQUIRY_DATA)
#define MANGLE_INQUIRY_DATA
#endif

#include "iscsi_target.h"
#include "iscsi_cmnd.h"
#include "target_error_rec.h"
#include "../common/ffp_negotiate.h"
#include "../common/unh_threads.h"

#ifdef __KERNEL__

#include <net/sock.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <net/if_inet6.h>
#include <net/ipv6.h>
#include <net/addrconf.h>

#else

rwlock_t dev_base_lock = RW_LOCK_UNLOCKED;
rwlock_t inetdev_lock = RW_LOCK_UNLOCKED;

#endif

/* 4 bytes of zeroes that can be sent as pad bytes when needed */
__u32 pad_bytes_for_sending = 0;

/* statistics counters to keep track of iov size in sendmsg and recvmsg */
int max_send_niov = 0;
int max_recv_niov = 0;

/* added by RDR when rcu_read_lock(), rcu_read_ulock() came in with 2.6.10
 * These macros define it the way it was in 2.4 and in earlier versions of 2.6
 *
 * eaburns - Changed macro names so that they didn't conflict with
 * rcu_read_(un)lock().
 */
#ifndef rcu_read_lock
#define inetdev_read_lock()	read_lock(&inetdev_lock)
#else
#define inetdev_read_lock()	rcu_read_lock()
#endif				/* !rcu_read_lock */

#ifndef rcu_read_unlock
#define inetdev_read_unlock()	read_unlock(&inetdev_lock)
#else
#define inetdev_read_unlock()	rcu_read_unlock()
#endif				/* !rcu_read_unlock */

/*
 * eaburns -- These macros abstract the addrconf read locking mechanism for
 * IPv6 between rcu_read_lock() and read_lock(&addrconf_lock).
 */
#if defined(CONFIG_IPV6) && defined(__KERNEL__)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
#define addrconf_read_lock()	rcu_read_lock()
#define addrconf_read_unlock()	rcu_read_unlock()
#else
#define addrconf_read_lock()	read_lock(&addrconf_lock)
#define addrconf_read_unlock()	read_unlock(&addrconf_lock)
#endif				/* >=linux-2.6.19 */
#endif				/* CONFIG_IPV6 && __KERNEL__ */

static void deal_with_nop_timer(unsigned long data);
static struct iscsi_conn *alloc_conn(struct portal_group *pg);
static void dealloc_conn(struct iscsi_conn *conn);

struct iscsi_global *devdata;

int bring_up_portal(struct portal_group *ptr)
{
	ptr->family = cnv_string_to_inet(ptr->ip_string,
					 ptr->port_string,
					 &ptr->ip_address,
					 &ptr->ip_length);
	if (ptr->family < 0)
		return -1;

#if !defined(RDMA_MODE)
	return iscsi_bring_up_portal(ptr);
#else
	return iser_bring_up_portal(ptr);
#endif
}

/* returns 0 on success, -1 on failure */
int iscsi_bring_up_portal(struct portal_group *ptr)
{
	int backlog = 5;
	struct socket *sockptr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "create server_thread on ip %s port %s\n",
	      ptr->ip_string, ptr->port_string);

	/* create a socket */
#ifdef FC2
	if (sock_create(ptr->family, SOCK_STREAM, 0, &sockptr, 0) < 0)
#else
	if (sock_create(ptr->family, SOCK_STREAM, 0, &sockptr) < 0)
#endif
	{
		TRACE_ERROR("%s Could not create socket on %s:%s\n",
			    current->comm, ptr->ip_string,
			    ptr->port_string);
		goto out1;
	}

	TRACE(TRACE_NET, "Socket %p created\n", sockptr);

	/*  turn on the option to reuse this socket's port quickly.
	 *  Must be done BEFORE calling bind()!
	 */
	tcp_reuse_port(sockptr);

	/* bind */
	if (sockptr->ops->bind(sockptr, ptr->ip_address, ptr->ip_length)) {
		TRACE_ERROR("%s Could not bind socket on %s:%s\n",
			    current->comm, ptr->ip_string,
			    ptr->port_string);
		goto out2;
	}

	TRACE(TRACE_NET, "Socket %p bound\n", sockptr);

	/* listen */
	if ((sockptr->ops->listen(sockptr, backlog))) {
		TRACE_ERROR("%s Could not listen with socket on %s:%s\n",
			    current->comm, ptr->ip_string,
			    ptr->port_string);
		goto out2;
	}

	devdata->server_socket[ptr - iscsi_portal_groups] = sockptr;

	/* create a server_thread that can accept connections */
	if (kernel_thread(iscsi_server_thread, (void *) ptr, 0) < 0) {
		TRACE_ERROR("%s Unable to create server thread\n",
			    current->comm);
		devdata->server_socket[ptr - iscsi_portal_groups] = NULL;
		goto out2;
	}

	TRACE(TRACE_NET, "Server_thread %lu spawned for socket %p\n",
	      (unsigned long) (ptr - iscsi_portal_groups), sockptr);

	/* wait for that server_thread to come up before continuing */
	if (!down_interruptible(&devdata->server_sem))
		return 0;

      out2:
	sock_release(sockptr);

      out1:
	my_kfree((void **) &ptr->ip_address, "ip_address");

	return -1;
}

/* print socket connect/disconnect message */
void report_peer_up_down(struct sockaddr *ip_address, char *message)
{
	char ip_string[INET6_ADDRSTRLEN + 2], port_string[8];

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (cnv_inet_to_string(ip_address, ip_string, port_string) > 0) {
		printk("%s %s %s:%s\n", current->comm, message,
		       ip_string, port_string);
	}
}

/*
 * generate the next TTT in a session
 * must be called with session->cmnd_lock held
 */
__u32 __attribute__ ((no_instrument_function))
    generate_next_ttt(struct iscsi_session *session)
{
	__u32 retval;

	retval = session->cmnd_id++;
	if (session->cmnd_id == 0) {
		/* just assigned -1 to retval, which is reserved
		 * so do it again, since 0 is not reserved
		 */
		retval = session->cmnd_id++;
	}
	return retval;
}

/*
 * called only from iscsi_server_thread() when a new connection is accepted
 * allocates all the structures necessary for a new connection and a
 * new session.  If later we find out that this connection belongs to
 * an existing session, the session structure alloced here will be
 * freed.  Its just easier to do everything here in one place, and it
 * makes it easier to recover from errors if everything is all set up
 * (i.e., there are no partial structures)
 * returns NULL if anything fails, in which case everything is freed
 *		   pointer to new iscsi_conn, which points to new iscsi_sess, if all ok
 */
int build_conn_sess(struct iscsi_conn *conn, struct portal_group *ptr)
{
	struct iscsi_session *session;

	TRACE(TRACE_DEBUG, "new conn %p\n", conn);

	INIT_LIST_HEAD(&conn->conn_link);

	INIT_LIST_HEAD(&conn->reject_list);
#ifdef __KERNEL__
	conn->reject_lock = SPIN_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&conn->reject_lock, NULL);
#endif

	conn->active = 1;
	conn->phase_collapse = devdata->phase_collapse;
	conn->conn_status = CONN_UP;
	conn->conn_id = ++devdata->conn_id;
	conn->dev = devdata;
	conn->max_send_length = 8192;
	conn->max_recv_length = 8192;
	conn->portal_group_tag = ptr->tag;
	conn->connection_flags = devdata->force;
#ifdef RDMA_MODE
	conn->conn_terminated = 0;
	conn->connection_flags |= USE_RDMA_LOGIN;
#endif
	conn->nop_period = devdata->nop_period;
	TRACE(TRACE_TIMERS, "%s: initializing nop_timer %p\n",
	      __func__, &conn->nop_timer);
	init_timer(&conn->nop_timer);

	init_MUTEX_LOCKED(&conn->tx_sem);
	init_MUTEX_LOCKED(&conn->kill_rx_sem);
	init_MUTEX_LOCKED(&conn->kill_tx_sem);
#ifdef __KERNEL__
	conn->sock_lock = SPIN_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&conn->sock_lock, NULL);
#endif
#ifdef RDMA_MODE
	init_MUTEX_LOCKED(&conn->data_in_sem);
	init_MUTEX_LOCKED(&conn->iser_notify_sem);
	init_MUTEX_LOCKED(&conn->send_done_sem);
	init_MUTEX_LOCKED(&conn->iser_conn_done_sem);
#endif
#ifdef __KERNEL__
	conn->text_in_progress_lock = SPIN_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&conn->text_in_progress_lock, NULL);
#endif

	session = my_kmalloc(sizeof(struct iscsi_session), "session");
	if (!session) {
		goto out4;
	}

	memset(session, 0, sizeof(struct iscsi_session));

	INIT_LIST_HEAD(&session->sess_link);
	INIT_LIST_HEAD(&session->active_command_list);
	INIT_LIST_HEAD(&session->free_command_list);
	INIT_LIST_HEAD(&session->conn_list);
	INIT_LIST_HEAD(&session->unsolicited_data_list);
#ifdef __KERNEL__
	session->unsolicited_data_lock = SPIN_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&session->unsolicited_data_lock, NULL);
#endif

	list_add_tail(&conn->conn_link, &session->conn_list);

	conn->session = session;
	session->nconn = 1;
	session->devdata = devdata;
	session->portal_group_tag = ptr->tag;
	session->version_max = ISCSI_MAX_VERSION;
	session->version_min = ISCSI_MIN_VERSION;

	if ((session->r2t_timer =
	     my_kmalloc(sizeof(struct timer_list), "r2t timer")) == NULL) {
		goto out5;
	}
	init_timer(session->r2t_timer);
	TRACE(TRACE_DEBUG, "Allocated R2T timer %p for session %p\n",
	      session->r2t_timer, session);

	if (!(session->session_params = my_kmalloc(MAX_CONFIG_PARAMS *
						   sizeof(struct
							  parameter_type),
						   "session_params"))) {
		goto out6;
	}

	if (!(session->oper_param = my_kmalloc(MAX_CONFIG_PARAMS *
					       sizeof(struct
						      session_operational_parameters),
					       "oper_param"))) {
		goto out7;
	}

	/* copy the parameters information from the global structure */
	param_tbl_cpy(*session->session_params, *devdata->param_tbl);
	session->r2t_period = devdata->r2t_period;
	session->max_queue_depth_allowed =
	    devdata->max_queue_depth_allowed;

	/* Store SNACK flags as part of Session - SAI */
	session->targ_snack_flg = devdata->targ_snack_flg;

#ifdef __KERNEL__
	session->cmnd_lock = SPIN_LOCK_UNLOCKED;
#else
	pthread_mutex_init(&session->cmnd_lock, NULL);
#endif
	init_MUTEX_LOCKED(&session->retran_sem);
	init_MUTEX_LOCKED(&session->thr_kill_sem);

#if defined(RDMA_MODE)
	/*
	 * Initialize the iSER receive PDU queues used to pass PDUs between the
	 * iSER callback functions and the processing threads.
	 */
	TRACE(TRACE_DEBUG, "%s: initializing queues for conn=%p\n",
	      __func__, conn);
	queue_init(&conn->rx_pdu_q);
	queue_init(&conn->rx_handle_q);
#endif				/* RDMA_MODE */

	/* all ok */
	goto out;

      out7:
	my_kfree((void **) &session->session_params, "session_params");

      out6:
	TRACE(TRACE_DEBUG, "Releasing R2T timer %p for session %p\n",
	      session->r2t_timer, session);
	my_kfree((void **) &session->r2t_timer, "r2t timer");

      out5:
	my_kfree((void **) &session, "session");

      out4:
	my_kfree((void **) &conn->local_ip_address, "local ip address");
	my_kfree((void **) &conn->ip_address, "ip address");
	my_kfree((void **) &conn, "iscsi_conn");
#ifndef RDMA_MODE
	sock_release(conn->conn_socket);
#endif
	return -1;
      out:
	return 0;
}

#ifndef __KERNEL__
/* cleanup handler called if thread is cancelled while sitting on an accept() */
void sock_alloc_cleaner(void *ptr)
{
	struct iscsi_conn *conn;

	TRACE(TRACE_NET, "in sock_alloc_cleaner, thread id %#08lx\n",
	      pthread_self());
	if ((conn = ptr) != NULL) {
		sock_release(conn->conn_socket);
		my_kfree((void **) &conn->local_ip_address,
			 "local ip address");
		my_kfree((void **) &conn->ip_address, "ip address");
		my_kfree((void **) &conn, "iscsi_conn");
	}
}
#endif

/*
 * Called only from iscsi_server_thread
 */
int accept_conn(struct iscsi_conn *conn, struct portal_group *ptr)
{
#ifdef RDMA_MODE
	iser_status_t status;
#else
	int result;
	int index;
	int addr_len;

	struct socket *sock;
	struct socket *newsock;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	index = ptr - iscsi_portal_groups;
	sock = devdata->server_socket[index];
#endif				/* RDMA_MODE */

	TRACE(TRACE_ENTER_LEAVE, "Entering accept_conn\n");

#ifdef RDMA_MODE
	/* Wait here on a sem that conn_req_notify will up after doing an accept
	 * Technically, we must check the devdata->force for USE_RDMA_LOGIN. But
	 * since DAPL can't do byte steam to RDMA mode switch right now, its OK
	 */

	TRACE(TRACE_DEBUG, "RDMA_MODE in accept for pending conn=%p\n", conn);
	TRACE(TRACE_DEBUG, "iser_conn_sem count=%d\n",
	      ptr->iser_conn_sem.count.counter);
	if(down_interruptible(&ptr->iser_conn_sem)) {
		TRACE_ERROR("%s: down_interruptible failed\n", __func__);
		goto err;
	}

	TRACE(TRACE_DEBUG, "RDMA_MODE Got a connection %p %p!!!!\n", conn,
	      ptr->conn_req_h);
	status = devdata->iser_ext_api.connection_accept(ptr->conn_req_h,
							 (iser_conn_handle_t)
							 conn, 1);

	if (status != ISER_SUCCESS) {
		TRACE_ERROR("iser_connection_accept: Failed %d\n", status);
		TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
		return -1;
	}
	TRACE(TRACE_ISER,
	      "Done iser_connection_accept - waiting for accept completion\n");

	if(down_interruptible(&conn->iser_conn_done_sem) != 0) {
		TRACE_ERROR("iser_connection_accept callback interrupted\n");
		TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
		return -1;
	}
	if (conn->conn_status == CONN_SETUP_FAIL) {
		TRACE_ERROR
		    ("iser_connection_accept callback - conn not establised\n");
		TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
		return -1;
	}
	conn->conn_status = CONN_UP;
	TRACE(TRACE_ISER, "iser_connection_accept callback received\n");

	//conn->iser_handle = ptr->iser_handle;

	if (ptr->dst_addr)
		memcpy(conn->ip_address, ptr->dst_addr,
		       sizeof(struct sockaddr));
	else {
		my_kfree((void **) &conn->ip_address, "ip address");
		conn->ip_address = NULL;
	}

	if (ptr->src_addr)
		memcpy(conn->local_ip_address, ptr->src_addr,
		       sizeof(struct sockaddr));
	else {
		my_kfree((void **) &conn->local_ip_address, "ip address");
		conn->local_ip_address = NULL;
	}
	TRACE(TRACE_ISER, "After waking up ..2\n");

	my_kfree((void **) &(ptr->src_addr), "sockaddr");
	my_kfree((void **) &(ptr->dst_addr), "sockaddr");

	TRACE(TRACE_ISER, "Done accept_conn\n");
	TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
	return 0;
#else				/* RDMA_MODE */

#ifdef __KERNEL__

/* fix for 2.6.10 contributed by Charles Coffing, ccoffing@novell.com */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 9)
	newsock = sock_alloc();
	TRACE(TRACE_DEBUG, "NONRDMA_MODE in accept\n");
	if (!newsock)
#else
	if (sock_create_kern(sock->sk->sk_family, sock->type,
			     sock->sk->sk_protocol, &newsock))
#endif

#else

	newsock = sock_alloc();
	if (!newsock)
#endif

	{
		TRACE_ERROR("%s Could not allocate socket for accept\n",
			    current->comm);
		goto err;
	}

	newsock->type = sock->type;
	newsock->ops = sock->ops;

	TRACE(TRACE_NET, "%s before accept: sock %p newsock %p\n",
	      current->comm, sock, newsock);

#ifndef __KERNEL__
	conn->conn_socket = newsock;
	pthread_cleanup_push(sock_alloc_cleaner, conn);
#endif

	result = sock->ops->accept(sock, newsock, 0);

#ifndef __KERNEL__
	pthread_cleanup_pop(0);
#endif

	if (result < 0) {
		goto out4;
	}

	/* turn off the Nagle Algorithm on this socket */
	tcp_nagle_off(newsock);

	conn->conn_socket = newsock;

	addr_len = ptr->ip_length;
	if (newsock->ops->
	    getname(newsock, conn->ip_address, &addr_len, 1) < 0) {
		TRACE_ERROR("%s Could not get peer name for socket %p\n",
			    current->comm, newsock);
		goto out4;
	}

	addr_len = ptr->ip_length;
	if (newsock->ops->
	    getname(newsock, conn->local_ip_address, &addr_len, 0) < 0) {
		TRACE_ERROR("%s Could not get local name for socket %p\n",
			    current->comm, newsock);
		goto out4;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
	return 0;

      out4:
	sock_release(newsock);
#endif				/* RDMA_MODE */

      err:
	TRACE(TRACE_ENTER_LEAVE, "Leaving accept_conn\n");
	return -1;
}

/**
 * Cleans up the server thread.
 */
static void server_shutdown_thread(void *aux)
{
	struct portal_group *ptr;

#ifndef __KERNEL__
	struct task_struct *current;
	current = pthread_getspecific(current_key);
#endif

	ptr = aux;

	/* mark that this server thread is no longer alive */
	devdata->server_thr[ptr - iscsi_portal_groups] = NULL;

	printk("%s closed %s:%s\n", current->comm,
	       ptr->ip_string, ptr->port_string);

	/* exit this thread after notifying whoever might be waiting */
	up(&devdata->server_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
}

/*
 * iscsi_server_thread: This thread is responsible for accepting
 * connections
 */
int iscsi_server_thread(void *param)
{
	struct iscsi_conn *new_conn;
	struct portal_group *ptr = param;
	int ret;
	long i;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	i = ptr - iscsi_portal_groups;

	lock_kernel();

	unh_daemonize("iscsi_thr_%ld", i);

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this server thread is alive */
	devdata->server_thr[i] = current;

	unlock_kernel();

#ifndef __KERNEL__
	printk("%s Starting thread id %#08lx\n", current->comm,
	       current->threadid);
#endif				/* __KERNEL__ */

	printk("%s Listening on %s:%s\n", current->comm,
	       ptr->ip_string, ptr->port_string);
	/* notify our parent that this thread is up */
	up(&devdata->server_sem);

	/* loop forever accepting new connections from initiators */
	for (;;) {
		new_conn = alloc_conn(ptr);
		BUG_ON(!new_conn);

#if !defined(__KERNEL__)
		pthread_cleanup_push(server_shutdown_thread, ptr);
#endif /* !__KERNEL__ */
		ret = accept_conn(new_conn, ptr);
#if !defined(__KERNEL__)
		pthread_cleanup_pop(0);
#endif /* !__KERNEL__ */
		if (ret == -1) {
			dealloc_conn(new_conn);
			TRACE_ERROR("%s: accept_conn failed\n", __func__);
			goto iscsi_server_thread_out;
		}
		/* re-copy the parameters to this new connection */
		param_tbl_uncpy(*new_conn->session->session_params);
		param_tbl_cpy(*new_conn->session->session_params,
		              *new_conn->session->devdata->param_tbl);
		new_conn->session->r2t_period = devdata->r2t_period;
		new_conn->session->max_queue_depth_allowed = devdata->max_queue_depth_allowed;
		new_conn->session->targ_snack_flg = devdata->targ_snack_flg;

		TRACE(TRACE_DEBUG, "accept_conn returned OK %p\n",
		      new_conn);

		/* clear bad sessions and connections here */
		clean_bad_stuff();

		TRACE(TRACE_DEBUG, "clean_bad_stuff returned OK\n");

		TRACE(TRACE_DEBUG, "build_conn_sess returned OK\n");

/*		TRACE(TRACE_ISCSI, "Connect socket %p on conn %p, cid %d, session %p\n",
			  newsock, new_conn, new_conn->conn_id, new_conn->session);*/
		/*
		 * NOTE THAT WE CANNOT DO MUCH OF ANYTHING WITH THIS
		 * CONNECTION AS WE DO NOT KNOW WHAT SESSION IT BELONGS
		 * TO. THIS WILL HAVE TO BE A PART OF THE RX_THREAD
		 */

		/* start the tx thread */
		if (kernel_thread(iscsi_tx_thread, (void *) new_conn, 0) <
		    0) {
			TRACE_ERROR("%s unable to create tx_thread\n",
				    current->comm);
			goto out1;
		}

		/* start the rx thread */
		if (kernel_thread(iscsi_rx_thread, (void *) new_conn, 0) <
		    0) {
			TRACE_ERROR("%s unable to create rx_thread\n",
				    current->comm);
			if (new_conn->tx_thread) {
				/*  Mike Christie mikenc@us.ibm.com */
				TRACE(TRACE_ISCSI,
				      "Releasing iscsi_tx_%d\n",
				      new_conn->conn_id);
				send_sig(ISCSI_SHUTDOWN_SIGNAL,
					 new_conn->tx_thread, 1);
			}
			goto out1;
		}
	}

      out1:

      iscsi_server_thread_out:
	server_shutdown_thread(ptr);
	return 0;
}

/* not used currently */
#if 0
void dump_unsolicited_buffers(struct iscsi_session *session)
{
	struct list_head *list_ptr, *list_temp;
	struct udata_list *udata;
	__u32 count;
	unsigned long flags;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	printk("%u udata buffers, %u bytes each\n",
	       session->unsolicited_list_size,
	       session->unsolicited_buffer_size);

	count = 0;
	list_for_each_safe(list_ptr, list_temp,
			   &session->unsolicited_data_list) {
		udata =
		    list_entry(list_ptr, struct udata_list, udata_link);
		count++;
		TEST_LOOP_COUNTER(count, "unsolicited_data_list");
		printk("%2u: udata %p, buffer %p\n",
		       count, udata, udata->buffer);
	}
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
}
#endif				/* 0 */

/* called only when a session structure is being released,
 * so there should be no active commands or connections in that session
 */
static void release_unsolicited_buffers(struct iscsi_session *session)
{
	struct list_head *list_ptr, *list_temp;
	struct udata_list *udata;
	__u32 count;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_BUF, "%s release %u udata buffers\n",
	      current->comm, session->unsolicited_list_size);

	count = 0;
	list_for_each_safe(list_ptr, list_temp,
			   &session->unsolicited_data_list) {
		udata =
		    list_entry(list_ptr, struct udata_list, udata_link);
		count++;
		TEST_LOOP_COUNTER(count, "unsolicited_data_list");
		list_del(list_ptr);
#ifndef TCP_ISER
		TRACE(TRACE_BUF, "%s %2u: udata %p, free buffer %p\n",
		      current->comm, count, udata, udata->buffer);
		my_kfree((void **) &udata->buffer, "udata buffer");
#endif
		my_kfree((void **) &udata, "udata slot");
	}

	if (count == session->unsolicited_list_size) {
		if (count) {
			printk("%s released %u udata buffers\n",
			       current->comm, count);
		}
	} else {
		printk("%s released %u udata buffers but allocated %u\n",
		       current->comm, count,
		       session->unsolicited_list_size);
	}
	session->unsolicited_list_size = 0;
}

void put_unsolicited_buf(struct udata_list *udata,
			 struct iscsi_session *session)
{
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_BUF, "%s put udata %p, buffer %p\n",
	      current->comm, udata, udata->buffer);
	spin_lock_irqsave(&session->unsolicited_data_lock, flags);
	list_add_tail(&udata->udata_link, &session->unsolicited_data_list);
	spin_unlock_irqrestore(&session->unsolicited_data_lock, flags);
}

/* not used currently */
#if 0
void reset_unsolicited_buffers(struct iscsi_session *session)
{
	struct session_operational_parameters *oper_param_entry;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	oper_param_entry = session->oper_param;

	if (!oper_param_entry->ImmediateData
	    && oper_param_entry->InitialR2T) {
		/* after negotiation, no Immediate data or unsolicited DataOut pdus */
		TRACE(TRACE_BUF, "%s negotiation frees %u udata buffers\n",
		      current->comm, session->unsolicited_list_size);
		release_unsolicited_buffers(session);
	}
}
#endif				/* 0 */

/*
 * for each session on the list,
 * release inactive connections and the session itself if no conns remain.
 * devdata->session_sem MUST be locked before this routine is called.
 */
static void __attribute__ ((no_instrument_function))
    clean_session_list(struct list_head *list)
{
	struct iscsi_session *session;
	struct iscsi_conn *conn;
	struct list_head *list_ptr1, *list_temp1;
	struct list_head *list_ptr2, *list_temp2;
	DECLARE_LOOP_COUNTER(count);
	DECLARE_LOOP_COUNTER(count2);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr1, list_temp1, list) {
		session =
		    list_entry(list_ptr1, struct iscsi_session, sess_link);
		BUMP_LOOP_COUNTER(count, "list");

		INIT_LOOP_COUNTER(count2);
		list_for_each_safe(list_ptr2, list_temp2,
				   &session->conn_list) {
			conn =
			    list_entry(list_ptr2, struct iscsi_conn,
				       conn_link);
			BUMP_LOOP_COUNTER(count2, "conn_list");

			if (conn->active == 0) {
				if (iscsi_release_connection(conn) < 0) {
					TRACE_ERROR
					    ("%s Error releasing connection\n",
					     current->comm);
				}
			}
		}

		if (list_empty(&session->conn_list)) {
			iscsi_release_session(session);
		}
	}
}

void __attribute__ ((no_instrument_function))
    clean_bad_stuff(void)
{
	/* destructive access to session lists */
	write_lock(&devdata->session_rwlock);
	clean_session_list(&devdata->bad_session_list);
	clean_session_list(&devdata->session_list);
	write_unlock(&devdata->session_rwlock);
}

/*
 * iscsi_release_session: This function is responsible for closing out
 * a session and removing it from whatever list it is on.
 * host->session_sem MUST be locked before this routine is called.
 * INPUT: session to be released
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
int iscsi_release_session(struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;
	struct iscsi_conn *conn;
	struct list_head *list_ptr, *list_temp;
	DECLARE_LOOP_COUNTER(count);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!session) {
		TRACE_ERROR("%s Cannot release a NULL session\n",
			    current->comm);
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI)) {
		print_isid_tsih_message(session, "Release session with ");
	}

	TRACE(TRACE_DEBUG, "Releasing R2T timer %p for session %p\n",
	      session->r2t_timer, session);

	/* Delete r2t timer - SAI */
	if (session->r2t_timer) {
		TRACE(TRACE_DEBUG, "Deleting r2t timer %p\n",
		      session->r2t_timer);
		del_timer_sync(session->r2t_timer);
		TRACE(TRACE_DEBUG, "Deleted r2t timer\n");
		my_kfree((void **) &session->r2t_timer, "r2t timer");
	}

	/* free mid-level structure for any active commands */
	INIT_LOOP_COUNTER(count);
	list_for_each_entry(cmnd, &session->active_command_list,
			    command_link) {
		BUMP_LOOP_COUNTER(count, "active_command_list");
		if (cmnd->cmnd != NULL) {
			if (scsi_release(cmnd->cmnd) < 0) {
				TRACE_ERROR
				    ("%s Trouble releasing command, opcode 0x%02x, "
				     "ITT %u, state 0x%x\n", current->comm,
				     cmnd->opcode_byte,
				     cmnd->init_task_tag, cmnd->state);
			}
			cmnd->cmnd = NULL;
		}
	}

	/* now free all active commands */
	iscsi_dequeue_list(&session->active_command_list, session);

	/* free connections */
	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp, &session->conn_list) {
		conn = list_entry(list_ptr, struct iscsi_conn, conn_link);
		BUMP_LOOP_COUNTER(count, "conn_list");
		TRACE(TRACE_ISCSI, "iscsi%d: releasing connection %d\n",
		      (int)
		      session->devdata->device->dev_id, conn->conn_id);

		if (iscsi_release_connection(conn) < 0) {
			TRACE_ERROR("%s Trouble releasing connection\n",
				    current->comm);
		}
	}

	/* dequeue session if it is linked into some list */
	if (!list_empty(&session->sess_link)) {
		list_del(&session->sess_link);

		/* error recovery ver new 18_04 - SAI */
		if (session->retran_thread) {
			/*  Mike Christie mikenc@us.ibm.com */
			TRACE(TRACE_ISCSI, "Releasing retran_thread\n");
			send_sig(ISCSI_SHUTDOWN_SIGNAL,
				 session->retran_thread, 1);
			down_interruptible(&session->thr_kill_sem);
		}
	}

	/* free session structures */
	release_unsolicited_buffers(session);

	release_free_commands(session);

	my_kfree((void **) &session->session_params, "session_params");

	my_kfree((void **) &session->oper_param, "oper_param");

	my_kfree((void **) &session, "session");

	return 0;
}

/*
 * called by both rx and tx threads, but only after the rx thread has been
 * started on a connection (so the connected message has been given).
 * if socket for this conn has not already been released, do it now and
 * print a message that says we disconnected.
 * devdata->session_sem MAY be locked before this routine is called.
 * Also NULL out the conn's socket so we don't try to do this again later.
 */
void iscsi_release_socket(struct iscsi_conn *conn)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	/* Need a spin_lock to avoid a race if more than one thread tries to shut
	 * down simultaneously on a multiprocessor, but do not disable interrupts
	 * during socket release or the kernel gets unhappy
	 */
	spin_lock(&conn->sock_lock);
	if (conn->conn_status) {
		/* Release socket */
		TRACE(TRACE_ISCSI,
		      "%lu %s disconnect conn %p cid %d session %p\n",
		      jiffies, current->comm, conn, conn->conn_id,
		      conn->session);

		report_peer_up_down(conn->ip_address, "disconnected from");
#ifndef RDMA_MODE
		sock_release(conn->conn_socket);
#else
		iser_conn_terminate(conn);
#endif
		conn->conn_socket = NULL;
		conn->conn_status = CONN_DOWN;
	}
	spin_unlock(&conn->sock_lock);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/*
 * iscsi_release_connection: releases all the stuff related to a conn.
 * Right now, it does not do anything for releasing commands specific
 * to a connection since commands are more specific to a session as
 * against a connection. It also dequeues the connection from the list
 */
int iscsi_release_connection(struct iscsi_conn *conn)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!conn) {
		TRACE_ERROR("%s Cannot release NULL connection/n",
			    current->comm);
		return -1;
	}

	/* dequeue any unsent reject messages (do not send them) */
	dequeue_reject(conn, 0);

	/* Kill the rx and tx thread */

	if (conn->rx_thread) {
		/*  Mike Christie mikenc@us.ibm.com */
		TRACE(TRACE_ISCSI, "Releasing iscsi_rx_%d\n",
		      conn->conn_id);
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->rx_thread, 1);
#if defined(__KERNEL__)
		down_interruptible(&conn->kill_rx_sem);
#else
		pthread_join(conn->rx_thread->threadid, NULL);
#endif /* __KERNEL__ */
	}

	if (conn->tx_thread) {
		/*  Mike Christie mikenc@us.ibm.com */
		TRACE(TRACE_ISCSI, "Releasing iscsi_tx_%d\n",
		      conn->conn_id);
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->tx_thread, 1);
#if defined(__KERNEL__)
		down_interruptible(&conn->kill_tx_sem);
#else
		pthread_join(conn->tx_thread->threadid, NULL);
#endif /* __KERNEL__ */
	}

	/* Release socket */
	iscsi_release_socket(conn);

	TRACE(TRACE_ISCSI_FULL, "Dequeue connection conn->cid %u\n",
	      conn->conn_id);

	/* Dequeue connection */
	list_del(&conn->conn_link);
	conn->session->nconn--;

	/* Free connection */
	dealloc_conn(conn);

	return 0;
}

/*
 * executed only by the rx thread.
 * Allocates new reject item, fills it in with header of rejected pdu,
 * and enqueues it for tx thread on reject_list of this connection.
 * Returns 0 if all ok, -1 on error (i.e., no memory)
 */
int enqueue_reject(struct iscsi_conn *conn, __u8 reason)
{
	int err = 0;
	struct reject_item *new_item;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter enqueue_reject, reason %u\n",
	      reason);

	if ((new_item =
	     my_kmalloc(sizeof(struct reject_item), "reject item"))
	    == NULL) {
		err = -1;
		goto out;
	}
	memcpy(new_item->bad_header, conn->bad_hdr, ISCSI_HDR_LEN);
	new_item->reason = reason;
	spin_lock_irqsave(&conn->reject_lock, flags);
	list_add_tail(&new_item->reject_link, &conn->reject_list);
	spin_unlock_irqrestore(&conn->reject_lock, flags);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "enqueue_reject\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave enqueue_reject, err %d\n", err);

	return err;
}

/*
 * executed only by the tx thread.
 * locks connection's queue of pending rejects, then takes each item off
 * queue in order and uses it to generate and send a Reject pdu to initiator.
 */
int dequeue_reject(struct iscsi_conn *conn, int sendit)
{
	int err = 0;
	struct reject_item *item;
	struct list_head *list_ptr, *list_temp;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

	INIT_LOOP_COUNTER(count);

      restart:
	spin_lock_irqsave(&conn->reject_lock, flags);
	list_for_each_safe(list_ptr, list_temp, &conn->reject_list) {
		item =
		    list_entry(list_ptr, struct reject_item, reject_link);
		BUMP_LOOP_COUNTER(count, "reject_list");
		list_del(list_ptr);
		if (sendit) {
			spin_unlock_irqrestore(&conn->reject_lock, flags);
			err =
			    iscsi_tx_rjt(conn, item->bad_header,
					 item->reason);
			my_kfree((void **) &item, "reject item");
			if (err)
				goto out;
			goto restart;
		}
		my_kfree((void **) &item, "reject item");
	}
	spin_unlock_irqrestore(&conn->reject_lock, flags);
      out:
	return err;
}

int handle_recv_msg(struct iscsi_conn *conn, struct generic_pdu *inputpdu)
{
	struct socket *sock = conn->conn_socket;
	int padding = 0;
	int retval = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "%s Enter handle_recv_msg \n",
	      current->comm);
#ifdef RDMA_MODE
	if (conn->connection_flags & USE_RDMA_LOGIN) {
		if (recv_target_control_notifications(conn,
						      (__u8 **) &
						      (inputpdu->hdr),
						      (__u8 **) &
						      (inputpdu->text)) < 0) {
			retval = -1;
			goto out;
		}

		inputpdu->text_length = be32_to_cpu(inputpdu->hdr->length);
		if (inputpdu->text_length > MAX_TEXT_LEN) {
			TRACE_ERROR("DSL %u greater than default "
				    "MaxRecvDataSegmentLength %d\n",
				    inputpdu->text_length, MAX_TEXT_LEN);
			retval = -1;
		}
	} else
#endif
	{
		if (iscsi_recv_msg
		    (sock, ISCSI_HDR_LEN, (char *) inputpdu->hdr,
		     conn->connection_flags) < 0) {
			TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
			retval = -1;
			goto out;
		}

		inputpdu->text_length = be32_to_cpu(inputpdu->hdr->length);
		if (inputpdu->text_length > 0) {
			padding = (-inputpdu->text_length) & 3;
			if (inputpdu->text_length <= MAX_TEXT_LEN) {
				if (iscsi_recv_msg
				    (sock, inputpdu->text_length + padding,
				     inputpdu->text,
				     conn->connection_flags) < 0) {
					TRACE(TRACE_DEBUG,
					      "iscsi_recv_msg failed\n");
					retval = -1;
					goto out;
				}
			} else {
				TRACE_ERROR
				    ("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
				     inputpdu->text_length, MAX_TEXT_LEN);
				retval = -1;
				goto out;
			}
		}
	}
	TRACE_BUFFER(TRACE_BUF, (char *) inputpdu->hdr, ISCSI_HDR_LEN,
		     "Received header\n");
	if (inputpdu->text_length > 0)
		TRACE_BUFFER(TRACE_BUF, (char *) inputpdu->text,
			     inputpdu->text_length,
			     "Received data, length: %d\n",
			     inputpdu->text_length);
      out:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave handle_recv_msg %d",
	      current->comm, retval);
	return retval;
}

/* The nop_timer runs continuously, even when NopIn pings are not enabled.
 * It is used to periodically wake up the tx thread,
 * which enables the tx thread to generate NopIn pings when initiator
 * has not responded to a StatSN update within a reasonable amount of time
 * (or has not given the target a chance to convey the StatSN update to it).
 * The target needs this in low-memory situations, since it has to hang onto
 * allocated memory, even after sending the scsi response, in case the
 * initiator asks for a retransmit.  Therefore, the target often needs to
 * force the initiator to acknowledge the StatSN sent in the last response.
 */
void restart_nop_timer(struct iscsi_conn *conn)
{
	conn->nop_timer.expires = jiffies + TIMER_PERIOD;
	conn->nop_timer.data = (unsigned long) conn;
	conn->nop_timer.function = deal_with_nop_timer;
	add_timer(&conn->nop_timer);
	/*
	TRACE(TRACE_TIMERS, "%s: restarting nop_timer %p\n",
	      __func__, &conn->nop_timer);
	 */
}

/* called by kernel when nop periodic timer expires on a connection
 * Always turn on the silence bit,
 * and also turn on the need nopin bit if silence bit was on before the set.
*/
static void deal_with_nop_timer(unsigned long data)
{
	struct iscsi_conn *conn;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	conn = (struct iscsi_conn *) data;
	if (conn->nop_period_remaining > TIMER_PERIOD) {
		conn->nop_period_remaining -= TIMER_PERIOD;
	} else {
		if ((conn->nop_period_remaining = conn->nop_period)) {
			if (test_and_set_bit(SILENCE_BIT, &conn->control)) {
				/*
				TRACE(TRACE_TIMERS,
				      "%s timer set NEED_NOPIN\n",
				      current->comm);
				 */
				set_bit(NEED_NOPIN_BIT, &conn->control);
			}
		}
	}
	restart_nop_timer(conn);
	if (atomic_read(&conn->tx_sem.count) <= 0) {
		/*
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "deal_with_nop_timer\n", jiffies, current->comm);
		 */
		up(&conn->tx_sem);
	}
}

/*
 * iscsi_rx_data: will receive a fixed amount of data.
 * INPUT: struct iscsi_conn (what connection), struct iovec, number of iovecs,
 * total amount of data in bytes
 * OUTPUT: > 0 total bytes read if everything is okay
 * 	       < 0 if trouble (-ECONNRESET means end-of-file)
 *		   = 0 for PAYLOAD_DIGERR
 */
int
iscsi_rx_data(struct iscsi_conn *conn, struct iovec *iov,
	      struct iovec *iov_copy, int niov, int data)
{
	struct msghdr msg;
	int msg_iovlen;
	int rx_loop, total_rx;
	__u32 checksum, data_crc;
	int i;
	struct iovec *iov_ptr;
#ifdef __KERNEL__
	mm_segment_t oldfs;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	total_rx = 0;

	TRACE(TRACE_ENTER_LEAVE,
	      "%s Enter iscsi_rx_data niov %d data %d base %p " "len %u\n",
	      current->comm, niov, data, iov->iov_base,
	      (unsigned int) iov->iov_len);

	if (!conn->conn_status) {
		TRACE_ERROR("%s Transport endpoint is not connected\n",
			    current->comm);
		total_rx = -ENOTCONN;	/* Transport endpoint is not connected */
		goto out;
	}

	while (total_rx < data) {
		/* get a clean copy of the original io vector to work with */
		memcpy(iov_copy, iov, niov * sizeof(struct iovec));
		msg_iovlen = niov;
		iov_ptr = iov_copy;

		if ((rx_loop = total_rx)) {
			/* have done only partial read, recalculate iov and niov -- cdeng */
			TRACE(TRACE_ISCSI, "%lu %s Recompute iov"
			      "data %d, received so far %d\n",
			      jiffies, current->comm, data, total_rx);
			do {
				if (iov_ptr->iov_len <= rx_loop) {
					rx_loop -= iov_ptr->iov_len;
					iov_ptr++;
					msg_iovlen--;
				} else {
					iov_ptr->iov_base += rx_loop;
					iov_ptr->iov_len -= rx_loop;
					rx_loop = 0;
				}
			} while (rx_loop);
		}

		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_iovlen = msg_iovlen;
		msg.msg_iov = iov_ptr;

		/* keep track of maximum number of iovs in any received message */
		if (max_recv_niov < niov)
			max_recv_niov = niov;

		TRACE(TRACE_NET,
		      "%s before sock_recvmsg size %d niov %d\n",
		      current->comm, (data - total_rx), niov);

#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
		rx_loop =
		    sock_recvmsg(conn->conn_socket, &msg,
				 (data - total_rx), MSG_WAITALL);
		set_fs(oldfs);
#else
		TRACE(TRACE_NET, "%s Call readv, fd %d, %d vectors\n",
		      current->comm, conn->conn_socket->fd, msg_iovlen);
		rx_loop =
		    readv(conn->conn_socket->fd, iov_ptr, msg_iovlen);
		TRACE(TRACE_NET, "%s Done readv, fd %d, result %d\n",
		      current->comm, conn->conn_socket->fd, rx_loop);
#endif

		TRACE(TRACE_NET, "%s after sock_recvmsg ret %d size %d\n",
		      current->comm, rx_loop, (data - total_rx));

		/* this receive from initiator broke the silence */
		clear_bit(SILENCE_BIT, &conn->control);

		if (rx_loop <= 0) {
			total_rx = -ECONNRESET;	/* Connection reset by peer */
			goto out;
		}

		total_rx += rx_loop;
		TRACE(TRACE_DEBUG,
		      "iscsi_rx_data: rx_loop %d total_rx %d\n", rx_loop,
		      total_rx);
	}

	if (niov > 1 && conn->data_crc) {
		/* this really is for a data segment, and data digests are in effect */
		data_crc = 0;
		for (i = 0; i < niov - 1; i++)
			do_crc(iov[i].iov_base, iov[i].iov_len, &data_crc);
		checksum = *(__u32 *) (iov[niov - 1].iov_base);
		if (checksum != data_crc) {
			TRACE_ERROR
			    ("%s Got data crc 0x%08x expected 0x%08x\n",
			     current->comm, ntohl(checksum),
			     ntohl(data_crc));
			/* Data Digest Error Recovery - SAI */
			total_rx = PAYLOAD_DIGERR;
			goto out;
		} else {
			TRACE(TRACE_ISCSI_FULL,
			      "%lu %s Got data crc 0x%08x\n", jiffies,
			      current->comm, ntohl(checksum));
		}
	}
      out:
	TRACE(TRACE_ENTER_LEAVE, "%s Leave iscsi_rx_data, total_rx %d\n",
	      current->comm, total_rx);

	return total_rx;
}

/*
 * executed only by the tx thread.
 * called only from handle_sendmsg().
 */
static int
do_sock_send(struct iovec *iov, struct iovec *iov_copy, int niov,
	     int data, struct socket *socket)
{
	struct msghdr msg;
	int msg_iovlen;
	int total_tx, tx_loop;
	struct iovec *iov_ptr;
#ifdef __KERNEL__
	mm_segment_t oldfs;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	total_tx = 0;
	while (total_tx < data) {
		/* get a clean copy of the original io vector to work with */
		memcpy(iov_copy, iov, niov * sizeof(struct iovec));
		msg_iovlen = niov;
		iov_ptr = iov_copy;

		if ((tx_loop = total_tx)) {
			TRACE(TRACE_ISCSI, "%lu %s Recompute iov "
			      "d: data %d not completed\n", jiffies,
			      current->comm, data);
			do {
				if (iov_ptr->iov_len <= tx_loop) {
					tx_loop -= iov_ptr->iov_len;
					iov_ptr++;
					msg_iovlen--;
				} else {
					iov_ptr->iov_base += tx_loop;
					iov_ptr->iov_len -= tx_loop;
					tx_loop = 0;
				}
			} while (tx_loop);
			TRACE(TRACE_ISCSI,
			      "%lu %s sock_sendmsg total_tx %d data %d "
			      "niov %d msg_iovlen %d\n", jiffies,
			      current->comm, total_tx, data, niov,
			      msg_iovlen);
		}

		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_iov = iov_ptr;
		msg.msg_iovlen = msg_iovlen;
		msg.msg_flags = MSG_NOSIGNAL;

		TRACE(TRACE_DEBUG,
		      "do_sock_send: niov %d, data %d, total_tx %d\n",
		      niov, data, total_tx);

		/* keep track of maximum number of iovs in any sent message */
		if (max_send_niov < niov)
			max_send_niov = niov;

#ifdef __KERNEL__
		oldfs = get_fs();
		set_fs(get_ds());
		tx_loop = sock_sendmsg(socket, &msg, (data - total_tx));
		set_fs(oldfs);
#else
		TRACE(TRACE_NET, "%s Call writev, fd %d, %d vectors\n",
		      current->comm, socket->fd, msg_iovlen);
		tx_loop = writev(socket->fd, iov_ptr, msg_iovlen);
		TRACE(TRACE_NET, "%s Done writev, fd %d, result %d\n",
		      current->comm, socket->fd, tx_loop);
#endif

		if (tx_loop <= 0) {
			struct pdu_hdr *pdu =
			    (struct pdu_hdr *) iov[0].iov_base;
			TRACE_ERROR
			    ("%s sock_sendmsg error %d, total_tx %d, data %d, niov "
			     "%d, msg_iovlen %d, op 0x%02x, flags 0x%02x, ITT %u\n",
			     current->comm, tx_loop, total_tx, data, niov,
			     msg_iovlen, pdu->opcode, pdu->flags,
			     ntohl(pdu->init_task_tag));
			return tx_loop;
		}

		total_tx += tx_loop;
		TRACE(TRACE_DEBUG,
		      "do_sock_send: tx_loop %d total_tx %d\n", tx_loop,
		      total_tx);
	}

	return total_tx;
}

/*
 * executed only by the tx thread.
 * called only from iscsi_tx_data().
 */
static inline int
handle_sendmsg(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
	       struct iovec *iov, struct iovec *iov_copy, int niov,
	       int data, int dm_flag)
{
	int retval = data;

	/* mark time at which something was last sent on this connection & command */
	conn->time_stamp = jiffies;
	if (cmnd)
		cmnd->time_stamp = jiffies;

#ifdef ISER
	if (conn->session->oper_param->RDMAExtensions) {
#ifdef RDMA_MODE
		struct iser_data_buf_t *databuf;
		struct iser_send_pdu_t send_pdu;
		iser_status_t status;
		struct pdu_hdr *bhs = (struct pdu_hdr *) iov[0].iov_base;
		struct iovec *iov_ptr = iov;
		int error = 0;

		send_pdu.p_bhs = iov[0].iov_base;
		iov_ptr++;

		switch (bhs->opcode) {
		case ISCSI_TARG_SCSI_RSP:
			send_pdu.data.response.buf_status.p_buf = iov_ptr;
			send_pdu.data.response.buf_status.size = niov - 1;
			TRACE(TRACE_ISER, "%s: niov=%d\n", __func__, niov);
			send_pdu.data.response.buf_status.type =
			    ISER_BUF_TYPE_IOVEC_VIRT;
			break;

		case ISCSI_TARG_NOP_IN:
		case ISCSI_TARG_TASK_MGMT_RSP:
		case ISCSI_TARG_TEXT_RSP:
		case ISCSI_TARG_LOGOUT_RSP:
		case ISCSI_TARG_LOGIN_RSP:
		case ISCSI_TARG_ASYNC_MSG:
		case ISCSI_TARG_RJT:
			send_pdu.data.tx.buf.p_buf = iov_ptr;
			send_pdu.data.tx.buf.size = niov - 1;
			TRACE(TRACE_ISER, "%s: niov=%d\n", __func__, niov);
			send_pdu.data.tx.buf.type =
			    ISER_BUF_TYPE_IOVEC_VIRT;
			break;

		case ISCSI_TARG_SCSI_DATA_IN:
			send_pdu.data.data_in.src_buf.p_buf = iov_ptr;
			send_pdu.data.data_in.src_buf.size = niov - 1;
			TRACE(TRACE_ISER, "%s: niov=%d\n", __func__, niov);
			send_pdu.data.data_in.src_buf.type =
			    ISER_BUF_TYPE_IOVEC_VIRT;
			TRACE(TRACE_ISER, "%s: calling put_data with "
			      "iSCSI cmnd=%p\n", __func__, cmnd);
			status =
			    devdata->iser_api.put_data(cmnd->iser_task_h,
						       &send_pdu, 1);
			if (status != ISER_SUCCESS) {
				TRACE_ERROR("put_data failed %d\n",
					    status);
				retval = -1;
			}
			goto out;
		case ISCSI_TARG_R2T:
			/* the data element of the iovec is really a
			 * scatterlist */
			databuf = iov_ptr->iov_base;

			send_pdu.data.r2t.dst_buf.p_buf = databuf->p_buf;
			send_pdu.data.r2t.dst_buf.size = databuf->size;
			send_pdu.data.r2t.dst_buf.type = databuf->type;
			TRACE(TRACE_ISER, "%s: niov=%d\n", __func__, niov);
			status =
			    devdata->iser_api.get_data(cmnd->iser_task_h,
						       &send_pdu, 1);
			if (status != ISER_SUCCESS) {
				TRACE_ERROR("get_data failed %d\n",
					    status);
				retval = -1;
			}
			goto out;
		default:
			TRACE_ERROR
			    ("####Unknown opcode in handle_sendmsg in RDMA mode\n");
			error = 1;
			break;
		}

		status =
		    devdata->iser_api.
		    send_control((iser_conn_handle_t) conn, &send_pdu);
		if (status != ISER_SUCCESS) {
			TRACE_ERROR("iser_send_control failed %d\n",
				    status);
			retval = -1;
		}
	      out:
		return retval;
#elif TCP_ISER
		retval =
		    do_sock_send(iov, iov_copy, niov, data,
				 conn->conn_socket);
#endif
	} else
#endif
	{
		retval =
		    do_sock_send(iov, iov_copy, niov, data,
				 conn->conn_socket);
	}
	return retval;
}

/*
 * executed only by the tx thread.
 * iscsi_tx_data: will transmit a fixed-size PDU of any type.
 * INPUT: struct iscsi_conn (what connection), struct iovec, number of iovecs,
 * total amount of data in bytes,
 * iovec[0] must be the 48-byte PDU header.
 * iovec[1] (if header digests are in use, calculate it in this function)
 * iovec[...] the data
 * iovec[niov-1] (if data digests are in use, calculate it in this function)
 * OUTPUT: int total bytes written if everything is okay
 * 	       < 0 if trouble
 */
int iscsi_tx_data(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		  struct iovec *iov, struct iovec *iov_copy, int niov,
		  int data, int dm_flag)
{

# ifdef DEBUG_DATA
	int i, j;
	struct iovec *debug_iov;
	__u8 *to_print;
# endif

	__u32 hdr_crc, data_crc;
	int data_len, k;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!conn->conn_status) {
		TRACE_ERROR("%s NULL conn_socket\n", current->comm);
		return -1;
	}
# ifdef DEBUG_DATA
	TRACE(TRACE_DEBUG, "iscsi_tx_data: iovlen %d\n", niov);

	debug_iov = iov;

	for (i = 0; i < niov; i++) {
		to_print = (__u8 *) debug_iov->iov_base;

		for (j = 0; ((j < debug_iov->iov_len) && (j < 64)); j++) {
			TRACE(TRACE_DEBUG, "%.2x ", to_print[j]);

			if (((j + 1) % 16) == 0)
				TRACE(TRACE_DEBUG, "\n");
			else if (((j + 1) % 4) == 0)
				TRACE(TRACE_DEBUG, "    ");
		}

		TRACE(TRACE_DEBUG, "\n");

		debug_iov++;
	}
# endif

	/* compute optional header digest */
	if (conn->hdr_crc) {
		hdr_crc = 0;
		do_crc(iov[0].iov_base, ISCSI_HDR_LEN, &hdr_crc);

		iov[1].iov_base = &hdr_crc;
		iov[1].iov_len = CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "%lu %s Send header crc 0x%08x\n",
		      jiffies, current->comm, ntohl(hdr_crc));
	}

	/* compute optional data digest */
	if (conn->data_crc && niov > conn->hdr_crc + 2) {
		data_len = 0;
		data_crc = 0;
		for (k = conn->hdr_crc + 1; k < niov - 1; k++) {
			do_crc(iov[k].iov_base, iov[k].iov_len, &data_crc);
			data_len += iov[k].iov_len;
		}

		iov[niov - 1].iov_base = &data_crc;
		iov[niov - 1].iov_len = CRC_LEN;
		TRACE(TRACE_ISCSI_FULL,
		      "%lu %s Send data len %d, data crc 0x%08x\n",
		      jiffies, current->comm, data_len, ntohl(data_crc));
	}

	return handle_sendmsg(cmnd, conn, iov, iov_copy, niov, data,
			      dm_flag);
}

/*
 * executed only by the tx thread.
 * RDR
 */
int send_hdr_plus_1_data(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
			 void *iscsi_hdr, void *data_buf,
			 int data_len, int dm_flag)
{
	struct iovec iov[5], iov_copy[5];
	int niov, total_size, padding, retval;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Entering %s, data_len=%d\n", __func__,
	      data_len);

#ifdef K26
	if (irqs_disabled()) {
		TRACE_WARNING
		    ("%lu %s send_hdr_plus_1_data CmdSN %u ITT %u "
		     "irqs disabled\n", jiffies, current->comm,
		     cmnd ? cmnd->cmd_sn : 0,
		     cmnd ? cmnd->init_task_tag : 0);
	}
#endif

	/* set up the header in the first iov slot */
	iov[0].iov_base = iscsi_hdr;
	iov[0].iov_len = ISCSI_HDR_LEN;
	total_size = ISCSI_HDR_LEN;
	niov = 1;		/* one for the header */

	if (conn->hdr_crc) {
		/* set up the header digest in the second iov slot */
		iov[niov].iov_len = CRC_LEN;
		total_size += CRC_LEN;
		niov++;		/* one for header digest */
	}

	if (data_len) {
		/* set up the data in the next iov slot */
		iov[niov].iov_base = data_buf;
		iov[niov].iov_len = data_len;
		total_size += data_len;
		niov++;		/* one for data */

		padding = (-data_len) & 3;
		if (padding) {
			/* set up the data padding in the next iov slot */
			iov[niov].iov_base = &pad_bytes_for_sending;
			iov[niov].iov_len = padding;
			total_size += padding;
			niov++;	/* one for padding */
			TRACE(TRACE_DEBUG, "padding attached: %d bytes\n",
			      padding);
		}

		if (conn->data_crc) {
			/* set up the data digest in the next iov slot */
			iov[niov].iov_len = CRC_LEN;
			total_size += CRC_LEN;
			niov++;	/* one for data digest */
		}
	}

	retval =
	    iscsi_tx_data(cmnd, conn, iov, iov_copy, niov, total_size,
			  dm_flag);

	if (retval != total_size) {
		TRACE_ERROR
		    ("%s Trouble in iscsi_tx_data, expected %d bytes, got %d\n",
		     current->comm, total_size, retval);
		retval = -1;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return retval;
}


/* aparantly unused */
#if 0
static inline int
send_fpp_hdr_only(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		  void *iscsi_hdr)
{
	return send_hdr_plus_1_data(cmnd, conn, iscsi_hdr, NULL, 0,
				    USE_SEND_CONTROL);
}
#endif				/* 0 */


/*
 * executed by the tx thread and handle_login in the rx thread.
 * RDR
 */
int send_hdr_only(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		  void *iscsi_hdr)
{
	return send_hdr_plus_1_data(cmnd, conn, iscsi_hdr, NULL, 0,
				    USE_SEND_CONTROL);
}


#ifdef __KERNEL__

/* 
 * helper routine for print_expanded_address_any() 
 * Only called from print_expanded_address_any()
 */
static int
convert_and_print_ip_stuff(char *ptr, struct sockaddr *real_ip_address,
			   struct portal_group *pg_ptr)
{
	char ip_string[INET6_ADDRSTRLEN + 2], port_string[8];
	int k = 0;

	if (cnv_inet_to_string(real_ip_address, ip_string, port_string) >=
	    0) {
		k = 1 + sprintf(ptr, "%s=%s:%s,%u", targetaddress_word,
				ip_string, port_string, pg_ptr->tag);
		TRACE(TRACE_DEBUG, "Expand %s to %s\n", pg_ptr->ip_string,
		      ptr);
	}
	return k;
}


/* called only from do_text_request()
 * when we find a portal_group entry with IP address INADDR_ANY
 * in order to expand it to all IP addresses on this host and print them
 * as values of TargetAddress== keys.
 *
 * this code is modelled on the functions inet_select_addr() and
 * inet_dump_ifaddr() in the file net/ipv4/devinet.c for IPV4; and on
 * the function ipv6_get_saddr() in the file net/ipv6/addrconf.c for IPV6.
 */
static int
print_expanded_address_any(char *ptr, struct portal_group *pg_ptr)
{
	struct net_device *dev;
	struct in_device *in_dev;
	struct in_ifaddr *ifa;
	struct inet6_dev *idev;
	struct inet6_ifaddr *ifp;
	struct sockaddr real_ip_address;
	struct sockaddr_in *iptr = (struct sockaddr_in *) &real_ip_address;
	struct sockaddr_in6 *i6ptr =
	    (struct sockaddr_in6 *) &real_ip_address;
	int n = 0;
	unsigned long rw_flags;

	if (pg_ptr->family != AF_INET && pg_ptr->family != AF_INET6)
		return n;

	/* search the list of all known network devices */
	read_lock_irqsave(&dev_base_lock, rw_flags);

#ifdef CONFIG_IPV6
	if (pg_ptr->family == AF_INET6)
		addrconf_read_lock();
#endif
	inetdev_read_lock();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	for (dev = dev_base; dev != NULL; dev = dev->next) {
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	for_each_netdev(dev) {
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24) */
	for_each_netdev(&init_net, dev) {
#endif
		memset(&real_ip_address, 0, pg_ptr->ip_length);
		real_ip_address.sa_family = pg_ptr->family;
		if (pg_ptr->family == AF_INET) {
			/* this is an IPv4 address */
			if ((in_dev = in_dev_get(dev)) == NULL)
				continue;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
			read_lock(&in_dev->lock);
#endif

			for (ifa = in_dev->ifa_list; ifa != NULL;
			     ifa = ifa->ifa_next) {
				/* do not advertize the loopback address */
				if (ifa->ifa_local ==
				    htonl(INADDR_LOOPBACK))
					continue;
				iptr->sin_addr.s_addr = ifa->ifa_local;
				iptr->sin_port
				    =
				    ((struct sockaddr_in *) pg_ptr->
				     ip_address)->sin_port;
				n += convert_and_print_ip_stuff(ptr + n,
								&real_ip_address,
								pg_ptr);
			}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
			read_unlock(&in_dev->lock);
#endif

		} else {
			/* this is an IPv6 address */
			if ((idev = __in6_dev_get(dev)) == NULL)
				continue;
			read_lock(&idev->lock);
			for (ifp = idev->addr_list; ifp != NULL;
			     ifp = ifp->if_next) {
				/* do not advertize link-local or loopback addresses */
#ifdef CONFIG_IPV6
				if (ipv6_addr_type(&ifp->addr)
				    & (IPV6_ADDR_LINKLOCAL |
				       IPV6_ADDR_LOOPBACK))
					continue;
#endif
				ipv6_addr_copy(&i6ptr->sin6_addr,
					       &ifp->addr);
				i6ptr->sin6_port =
				    ((struct sockaddr_in6 *) pg_ptr->
				     ip_address)->sin6_port;
				n += convert_and_print_ip_stuff(ptr + n,
								&real_ip_address,
								pg_ptr);
			}
			read_unlock(&idev->lock);
		}
		/* else ignore this address family */
	}

#ifdef CONFIG_IPV6
	if (pg_ptr->family == AF_INET6)
		addrconf_read_unlock();
#endif				/* CONFIG_IPV6 */

	inetdev_read_unlock();

	read_unlock_irqrestore(&dev_base_lock, rw_flags);
	return n;
}

#else

/* for now in user space -- needs to be filled in */
static int
print_expanded_address_any(char *ptr, struct portal_group *pg_ptr)
{
	return 0;
}

#endif


/*
 * executed only by the rx thread.
 * called only from do_text_request() with conn's text_in_progress_lock held
 * returns 0 if ok, != 0 is reason for reject
 */
static int
accumulate_text_input(struct iscsi_cmnd *cmnd,
		      struct iscsi_cmnd *in_progress)
{
	__u32 size;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if ((size = cmnd->first_burst_len)) {
		/* command has some data attached, must accumulate it */
		if (in_progress->in_progress_buffer == NULL) {
			in_progress->in_progress_buffer
			    = my_kmalloc(MAX_TEXT_LEN, "text buf");
			if (in_progress->in_progress_buffer == NULL) {
				return REASON_OUT_OF_RESOURCES;
			}
		}

		size += in_progress->data_length;
		if (size > MAX_TEXT_LEN) {
			TRACE_ERROR
			    ("%s too many total bytes %u in text request, max %u\n",
			     current->comm, size, MAX_TEXT_LEN);
			return REASON_OUT_OF_RESOURCES;
		}

		memcpy(in_progress->in_progress_buffer +
		       in_progress->data_length, cmnd->ping_data,
		       cmnd->first_burst_len);
		in_progress->data_length = size;
		my_kfree((void **) &cmnd->ping_data, "data_buf");
	}
	return 0;
}

/*
 * executed only by the rx thread.
 * auxilliary function called only from do_text_request()
 * called when the Response to a Text Request should be generated.
 * connection's text_in_progress_lock MUST be held by caller.
 */
static void
generate_text_response(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		       struct iscsi_session *session)
{
	enum text_types text_type;
	char *buffer = NULL, *ptr, *ip_ptr, *val_string, *name;
	int i, j, n, maxt;
	__u32 reason, size, newlen, new_max_send_length = 0;
	struct portal_group *pg_ptr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter generate_text_response\n");

	/* allocate new buffer to hold response keys */
	if ((buffer = my_kmalloc(MAX_TEXT_LEN, "discovery buf")) == NULL) {
		reason = REASON_OUT_OF_RESOURCES;
		goto outbad;
	}

	reason = parse_text_buffer(cmnd->ping_data, cmnd->data_length,
				   *session->session_params,
				   session->oper_param->SessionType,
				   USE_BY_INITIATOR,
				   buffer, MAX_TEXT_LEN,
				   &text_type, &new_max_send_length, &size,
				   &val_string);
	if (reason) {
		goto outbad;
	}

	if (text_type == TEXT_EMPTY || text_type == TEXT_OTHER) {
		/* all ok but did not find a SendTargets key */
		goto outok;
	}

	/* generate the SendTargets response in buffer */
	/* send the '\0' at the end of each key=value string too */
	ptr = buffer;
	maxt = MAX_TARGETS;
	if (maxt < target_count)
		maxt = target_count;
	for (i = 0; i < maxt; i++) {
		if (!target_in_use(i) || (text_type == TEXT_ST_NULL
					  && i !=
					  session->oper_param->TargetName))
			continue;
		/* include '=', '\0', room for 3-digit number */
		newlen =
		    size + targetname_size + targetname_header_size + 3;
		if (newlen > MAX_TEXT_LEN) {
			reason =
			    give_text_overflow_message(newlen,
						       MAX_TEXT_LEN);
			goto outbad;
		}
		n = sprintf(ptr, "%s=", targetname_word);
		name = ptr + n;
		n += 1 + sprintf(name, "%s%d", targetname_header_word, i);
		if (text_type == TEXT_ST_TN) {
			/* need to match the name in SendTargets=<target-name> */
			if (strcmp(val_string, name) != 0) {
				continue;
			}
		}
		ptr += n;
		size += n;

		for (j = 0; j < MAX_PORTAL; j++) {

			pg_ptr = &iscsi_portal_groups[j];
			if (!pg_ptr->in_use)
				continue;

			if (size > (MAX_TEXT_LEN - TEXT_FUDGE_LEN)) {
				reason = give_text_overflow_message(size,
								    (MAX_TEXT_LEN
								     -
								     TEXT_FUDGE_LEN));
				goto outbad;
			}
			/* expand the address "any" to all IP addresses on this host */
			ip_ptr = pg_ptr->ip_string;
			if (strcmp(ip_ptr, INADDR_ANY_STRING) == 0
			    || strcmp(ip_ptr, IN6ADDR_ANY_STRING) == 0) {
				n = print_expanded_address_any(ptr,
							       pg_ptr);
			} else {
				n = 1 + sprintf(ptr, "%s=%s:%s,%u",
						targetaddress_word, ip_ptr,
						pg_ptr->port_string,
						pg_ptr->tag);
			}
			ptr += n;
			size += n;
		}
	}

      outok:
	if ((cmnd->data_length = size) == 0) {
		/* no keys need to be sent back */
		my_kfree((void **) &buffer, "discovery buf");
	}
	my_kfree((void **) &cmnd->ping_data, "data_buf");
	cmnd->ping_data = buffer;
	cmnd->data_done = 0;
	cmnd->new_max_send_length = new_max_send_length;
	cmnd->state = ISCSI_SEND_TEXT_RESPONSE;
	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "generate_text_response\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}
      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave generate_text_response\n");
	return;

      outbad:
	my_kfree((void **) &buffer, "discovery buf");
	cmnd->state = ISCSI_DEQUEUE;
	enqueue_reject(conn, reason);
	goto out;
}

/*
 * called only from do_text_request()
 */
static inline void __attribute__ ((no_instrument_function))
    copy_in_progress_stuff(struct iscsi_cmnd *cmnd,
		       struct iscsi_cmnd *in_progress)
{
	cmnd->state = ISCSI_DEQUEUE;
	in_progress->command_flags = cmnd->command_flags;
	in_progress->cmd_sn = cmnd->cmd_sn;
	in_progress->stat_sn = cmnd->stat_sn;
}

/*
 * executed only by the rx thread.
 * called when a Text Request is the next command to be processed.
 */
void do_text_request(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		     struct iscsi_session *session)
{
	int reason;
	struct iscsi_cmnd *in_progress;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter do_text_request\n");

	spin_lock(&conn->text_in_progress_lock);

	if (cmnd->init_task_tag == ALL_ONES) {
		TRACE_ERROR("%s Text Request with reserved ITT=0x%08x\n",
			    current->comm, ALL_ONES);
		goto outbadprotocol;
	}

	/*      RFC 3720 Section 10.10.2 C (Continue) Bit
	 *      "A Text Request with the C bit set to 1 MUST have the F bit set to 0."
	 */
	if ((cmnd->command_flags & (C_BIT | F_BIT)) == (C_BIT | F_BIT)) {
		TRACE_ERROR
		    ("%s Text Request with ITT %u has C=1 and F=1\n",
		     current->comm, cmnd->init_task_tag);
		goto outbadprotocol;
	}

	in_progress = (struct iscsi_cmnd *) conn->text_in_progress;
	if (in_progress) {
		/* Text negotiation already in progress, must be same task:
		 *
		 * RFC 3720 Section 10.10.3 Initiator Task Tag
		 * "If the command is sent as part of a sequence of text requests
		 * and responses, the Initiator Task Tag MUST be the same for all the
		 * requests within the sequence (similar to linked SCSI commands).
		 * The I bit for all requests in a sequence also MUST be the same.
		 */
		if (cmnd->init_task_tag != in_progress->init_task_tag) {
			TRACE_ERROR
			    ("%s Text Request has ITT=%u, expected %u\n",
			     current->comm, cmnd->init_task_tag,
			     in_progress->init_task_tag);
			goto outbadprotocol;
		}

		if (cmnd->opcode_byte != in_progress->opcode_byte) {
			TRACE_ERROR
			    ("%s Text Request with ITT %u has I=%u, expected %u\n",
			     current->comm, cmnd->init_task_tag,
			     (cmnd->opcode_byte & I_BIT) == I_BIT,
			     (in_progress->opcode_byte & I_BIT) == I_BIT);
			goto outbadprotocol;
		}

		/*
		 * RFC 3720 Section 10.10.4 Target Transfer Tag
		 * "When the Target Transfer Tag is set to the reserved value
		 * 0xffffffff, it tells the target that this is a new request and
		 * the target resets any internal state associated with the
		 * Initiator Task Tag (resets the current negotiation state)
		 *
		 * The target sets the Target Transfer Tag in a text response to
		 * a value other than the reserved value 0xffffffff whenever it
		 * indicates that it has more data to send or more operations to
		 * perform that are associated with the specified Initiator Task
		 * Tag.  It MUST do so whenever it sets the F bit to 0 in the
		 * response.  By copying the Target Transfer Tag from the response
		 * to the next Text Request, the initiator tells the target to
		 * continue the operation for the specific Initiator Task Tag."
		 */
		if (cmnd->target_xfer_tag == ALL_ONES) {
			/* initiator is resetting an in-progress text exchange
			 * just forget everything that happened up to this point
			 */
			printk("iscsi %s ITT %u reset\n",
			       session->oper_param->
			       SessionType ? discovery_word :
			       "Text request", in_progress->init_task_tag);
			in_progress->state = ISCSI_DEQUEUE;
			in_progress->init_task_tag = ALL_ONES;
			conn->text_in_progress = in_progress = NULL;
			/* now fall thru to start again as if a new command */
		} else {
			/* initiator is continuing an in-progress text exchange
			 * the Target Transfer Tag must match the one we sent earlier
			 */
			if (cmnd->target_xfer_tag !=
			    in_progress->target_xfer_tag) {
				TRACE_ERROR
				    ("%s Text Request with ITT %u has TTT=%u, "
				     "expected %u\n", current->comm,
				     cmnd->init_task_tag,
				     cmnd->target_xfer_tag,
				     in_progress->target_xfer_tag);
				goto outbadprotocol;
			}

			if (in_progress->state ==
			    ISCSI_BLOCKED_SENDING_TEXT) {
				/* The last Text Response we sent had C = 1 or this connection
				 * has USE_ONE_KEY_PER_TEXT set.  We expect this Text Request
				 * should not contain any data and should have its C = 0
				 */
				if (cmnd->first_burst_len) {
					TRACE_ERROR
					    ("%s Text Request with ITT %u TTT %u has DSL=%u,"
					     " expected 0\n",
					     current->comm,
					     cmnd->init_task_tag,
					     cmnd->target_xfer_tag,
					     cmnd->first_burst_len);
					goto outbadprotocol;
				}
				if (cmnd->command_flags & C_BIT) {
					TRACE_ERROR
					    ("%s Text Request with TTT %u has C=1, "
					     "expected 0\n", current->comm,
					     cmnd->target_xfer_tag);
					goto outbadprotocol;
				}
				/* ok to send our next text response pdu (done with this pdu) */
				in_progress->state =
				    ISCSI_SEND_TEXT_RESPONSE;
				copy_in_progress_stuff(cmnd, in_progress);
				goto out;
			}
			/* Our last Text Response must have had C = 0 (empty or complete) */
			if (in_progress->state != ISCSI_AWAIT_MORE_TEXT) {
				/* this should never happen */
				TRACE_ERROR
				    ("%s Text in_progress with ITT %u has state=%u, "
				     "expected %u\n", current->comm,
				     in_progress->init_task_tag,
				     in_progress->state,
				     ISCSI_AWAIT_MORE_TEXT);
				reason = REASON_OUT_OF_RESOURCES;
				goto outbad;
			}
		}
	} else if (cmnd->target_xfer_tag != ALL_ONES) {
		TRACE_ERROR("%s Text Request with ITT %u has TTT=%u, "
			    "expected 0xffffffff\n", current->comm,
			    cmnd->init_task_tag, cmnd->target_xfer_tag);
		goto outbadprotocol;
	}

	if (!in_progress) {
		in_progress = cmnd;
	}
	if ((reason = accumulate_text_input(cmnd, in_progress)))
		goto outbad;

	if (in_progress == cmnd) {
		conn->text_in_progress = in_progress;
	} else {
		copy_in_progress_stuff(cmnd, in_progress);
	}
	if (cmnd->command_flags & C_BIT) {
		in_progress->state = ISCSI_ASK_FOR_MORE_TEXT;
	} else {
		/* have accumulated a complete Text Request, generate response to it */
		in_progress->ping_data = in_progress->in_progress_buffer;
		in_progress->in_progress_buffer = NULL;

		generate_text_response(in_progress, conn, session);
	}

      out:
	spin_unlock(&conn->text_in_progress_lock);
	TRACE(TRACE_ENTER_LEAVE, "Leave do_text_request\n");
	return;

      outbadprotocol:
	reason = REASON_PROTOCOL_ERR;
      outbad:
	cmnd->state = ISCSI_DEQUEUE;
	cmnd->init_task_tag = ALL_ONES;
	enqueue_reject(conn, reason);
	goto out;
}

/* computes the abort response when the referenced task is not found:
 * RFC 3720 section 10.6.1 Response
 * For the ABORT TASK function,
 * ...
 * "b) If the Referenced Task Tag does not identify an existing task,
 * but if the CmdSN indicated by the RefCmdSN field in the Task
 * Management function request is within the valid CmdSN window and less
 * than the CmdSN of the Task Management function request itself,
 * then targets must consider the CmdSN received and return the
 * 'Function complete' response.
 * "c) If the Referenced Task Tag does not identify an existing task
 * and if the CmdSN indicated by the RefCmdSN field in the Task
 * Management function request is outside the valid CmdSN window,
 * then targets must return the 'Task does not exist' response.
 */
static int __attribute__ ((no_instrument_function))
    get_abort_response(struct iscsi_session *session, struct iscsi_cmnd *cmnd)
{
	int delta, retval = TASK_DOES_NOT_EXIST;
	unsigned long flags;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	delta = session->max_cmd_sn - cmnd->ref_cmd_sn;
	if (delta < 0) {
		/* ref_cmd_sn is greater than MaxCmdSN, so outside valid CmdSN window */
		goto out_unlock;
	}

	delta = cmnd->ref_cmd_sn - session->exp_cmd_sn;
	if (delta < 0) {
		/* ref_cmd_sn is less than ExpCmdSN, so outside valid CmdSN window */
		goto out_unlock;
	}

	delta = cmnd->cmd_sn - cmnd->ref_cmd_sn;
	if (delta <= 0) {
		/* ref_cmd_sn is NOT less than cmd_sn */
		goto out_unlock;
	}

	/* target must consider the CmdSN received, and return FUNCTION_COMPLETE */
	session->exp_cmd_sn++;
	session->max_cmd_sn++;	/* don't let command window shrink! */
	retval = FUNCTION_COMPLETE;

      out_unlock:
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
	return retval;
}

void __attribute__ ((no_instrument_function))
    do_task_mgt(struct iscsi_conn *conn, struct iscsi_cmnd *tm_cmnd)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (tm_cmnd->ref_function == TMF_ABORT_TASK) {
		TRACE(TRACE_TASK_MGT, "%s ABORT TASK with ITT %u\n",
		      current->comm, tm_cmnd->ref_task_tag);
		tm_cmnd->ref_command =
		    search_tags(conn, tm_cmnd->ref_task_tag, ALL_ONES,
				TRACE_TEST(TRACE_TASK_MGT));
		if (tm_cmnd->ref_command == NULL) {
			/* can't find command with this ITT, check the RefCmdSN field */
			TRACE_ERROR("%s No command with ITT %u\n",
				    current->comm, tm_cmnd->ref_task_tag);
			tm_cmnd->response =
			    get_abort_response(conn->session, tm_cmnd);
			tm_cmnd->state = ISCSI_MGT_FN_DONE;
		} else {
			TRACE_ERROR
			    ("%s Aborting opcode 0x%02x, ITT %u, xfer_len %u, "
			     "data_done %u, r2t_data_left %d, r2t_sn %u, state %u\n",
			     current->comm,
			     tm_cmnd->ref_command->opcode_byte,
			     tm_cmnd->ref_command->init_task_tag,
			     tm_cmnd->ref_command->data_length,
			     tm_cmnd->ref_command->data_done,
			     tm_cmnd->ref_command->r2t_data_left,
			     tm_cmnd->ref_command->r2t_sn,
			     tm_cmnd->ref_command->state);
			tm_cmnd->ref_active_seqno =
			    tm_cmnd->ref_command->active_seqno;
			tm_cmnd->message =
			    rx_task_mgmt_fn(conn->dev->device,
					    tm_cmnd->ref_function,
					    tm_cmnd->ref_command->cmnd);
			if (!tm_cmnd->message) {
				/* midlevel could not even start on this */
				tm_cmnd->response = FUNCTION_REJECTED;
				tm_cmnd->state = ISCSI_MGT_FN_DONE;
			}
		}
	} else {
		/* we don't deal with this task management function (yet) */
		tm_cmnd->response = TASK_MANAGEMENT_FUNCTION_NOT_SUPPORTED;
		tm_cmnd->state = ISCSI_MGT_FN_DONE;
	}
}

/* Deliver this command to the midlevel and see if it was accepted.  If midlevel
 * can't handle it, we are finished with it, throw it away
 * (for now, without sending a reply to initiator -- should be changed!!)
 * returns 0 if all ok, -1 on error with cmnd->state changed to ISCSI_DEQUEUE
 */
int deliver_to_midlevel(struct iscsi_cmnd *cmnd,
			struct iscsi_init_scsi_cmnd *pdu,
			struct iscsi_session *session)
{
	unsigned long flags;
	int err = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ISCSI_FULL,
	      "%lu %s deliver CmdSN %u ITT %u to target\n", jiffies,
	      current->comm, cmnd->cmd_sn, cmnd->init_task_tag);

	/* pass this command on to the midlevel so it can set up the buffers */
	rx_cmnd(cmnd->conn->dev->device, session->oper_param->TargetName,
		pdu->lun, pdu->cdb, ISCSI_CDB_LEN, cmnd->data_length,
		cmnd->command_flags, &cmnd->cmnd);

	if (!(cmnd->cmnd)) {
		/* midlevel couldn't handle this command, so we are finished with it,
		 * throw it away (without sending a reply to initiator)
		 */
		TRACE_ERROR("%s rx_cmnd returned NULL, ITT %u\n",
			    current->comm, cmnd->init_task_tag);
		spin_lock_irqsave(&session->cmnd_lock, flags);
		cmnd->state = ISCSI_DEQUEUE;
		/* have to bump these on error */
		session->exp_cmd_sn += cmnd->cmd_sn_increment;
		cmnd->cmd_sn_increment = 0;
		session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
		cmnd->need_max_cmd_sn_inc = 0;
		spin_unlock_irqrestore(&session->cmnd_lock, flags);
		err = -1;
	}
	return err;
}

/*
 * executed by the rx thread and tx thread.
 * auxilliary function called only from check_queued_cmnd()
 * called when out-of-order queued commands (except for SCSI Commands)
 * become in-order and have to be "delivered"
 * Note: these commands cannot be immediate, since immediate commands are
 *		 never out of order
 */
static void __attribute__ ((no_instrument_function))
    deliver_queue_other(struct iscsi_cmnd *cmnd, struct iscsi_session *session)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (cmnd->opcode_byte == ISCSI_INIT_NOP_OUT) {
		/* this was an out-of-order NopOut that expects a reply --
		 * just change state here and let tx thread deal with it later
		 */
		cmnd->state = ISCSI_PING;
		TRACE(TRACE_DEBUG,
		      "%s Ping back CmdSN %u ExpCmdSN %u ITT %u"
		      " opcode 0x%02x, state %u, data_length " "%u\n",
		      current->comm, cmnd->cmd_sn, session->exp_cmd_sn,
		      cmnd->init_task_tag, cmnd->opcode_byte, cmnd->state,
		      cmnd->data_length);
	} else if (cmnd->opcode_byte == ISCSI_INIT_TEXT_CMND) {
		/* this was an out-of-order Text Request */
		do_text_request(cmnd, cmnd->conn, session);
	} else if (cmnd->opcode_byte == ISCSI_INIT_LOGOUT_CMND) {
		/* this was an out-of-order Logout Request */
		cmnd->state = ISCSI_LOGOUT;
	} else if (cmnd->opcode_byte == ISCSI_INIT_TASK_MGMT_CMND) {
		/* this was an out-of-order Task Management Request */
		do_task_mgt(cmnd->conn, cmnd);
	} else {
		/* should never happen */
		TRACE_ERROR
		    ("%s Unexpected queue other CmdSN %u ITT 0x%08x "
		     "opcode 0x%02x\n", current->comm, cmnd->cmd_sn,
		     cmnd->init_task_tag, cmnd->opcode_byte);
	}
}

/*	called by any rx_thread or tx_thread.
 *	Checks session's command queue to see if any out-of-order command is now
 *	in order.  If so, and it is a SCSI command,
 *	call rx_cmnd for it as the same as the handle_cmnd does
 */
void check_queued_cmnd(struct iscsi_session *session)
{
	struct iscsi_init_scsi_cmnd *pdu;
	struct iscsi_cmnd *cmnd;
	struct list_head *list_ptr, *list_temp;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* need to restart search loop whenever ExpCmdSN is incremented,
	 * which happens when we find and deliver a previously out-of-order
	 * command because it is now in order.
	 */
      restart:
	/* lock the session-wide list of commands while searching it */
	spin_lock_irqsave(&session->cmnd_lock, flags);

	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp,
			   &session->active_command_list) {
		cmnd =
		    list_entry(list_ptr, struct iscsi_cmnd, command_link);
		BUMP_LOOP_COUNTER(count, "active_command_list");
		if (cmnd->cmd_sn == session->exp_cmd_sn
		    && !(cmnd->opcode_byte & I_BIT)) {

			/* ok now to deliver this command to the target in proper order */
			if (cmnd->state == ISCSI_QUEUE_OTHER) {
				/* this is a non-SCSI command, deliver it to us, not target */
				TRACE(TRACE_ISCSI_FULL,
				      "%lu %s CmdSN %u ITT %u now in order\n",
				      jiffies, current->comm, cmnd->cmd_sn,
				      cmnd->init_task_tag);
				spin_unlock_irqrestore(&session->cmnd_lock,
						       flags);
				deliver_queue_other(cmnd, session);
				/* command should have changed state,
				 * tx thread may need to take further action on it
				 */
				if (atomic_read(&cmnd->conn->tx_sem.count)
				    <= 0) {
					TRACE(TRACE_SEM,
					      "%lu %s up tx_sem in check_queued_cmnd\n",
					      jiffies, current->comm);
					up(&cmnd->conn->tx_sem);
				}
				goto restart;	/* exp_cmd_sn was bumped */
			} else if (cmnd->state == ISCSI_QUEUE_CMND) {
				/* this was an out-of-order (non-immediate) SCSI Command that
				 * just now is in-order, deliver it to target after changing
				 * its state until delivery is confirmed
				 */
				cmnd->state = ISCSI_AWAIT_BUFFER_QUEUE;
				spin_unlock_irqrestore(&session->cmnd_lock,
						       flags);
				pdu =
				    (struct iscsi_init_scsi_cmnd *) cmnd->
				    hdr;
				if (deliver_to_midlevel
				    (cmnd, pdu, session)) {
					/* command changed state to ISCSI_DEQUEUE on error,
					 * tx thread needs to take further action on it
					 */
					if (atomic_read
					    (&cmnd->conn->tx_sem.count) <=
					    0) {
						TRACE(TRACE_SEM,
						      "%lu %s up tx_sem in "
						      "check_queued_cmnd\n",
						      jiffies,
						      current->comm);
						up(&cmnd->conn->tx_sem);
					}
					goto restart;	/* exp_cmd_sn was bumped on error */
				}
				/* this SCSI command was successfully delivered to midlevel, so
				 * give other threads a chance to act on it before we deliver
				 * any subsequent commands.  (In any case, exp_cmd_sn will not
				 * be changed until the midlevel calls us back, so unless the
				 * midlevel thread gets control now, searching through the queue
				 * again will be a waste of time!)
				 */
				return;
			}	/* else this command was previously delivered to us or to the
				 * midlevel, and we are still waiting for tx-thread action or a
				 * callback from midlevel before we bump ExpCmdSN (have to wait
				 * so other commands don't jump ahead of this one in the
				 * session-wide command sequence).  No other command in queue
				 * should have this cmd_sn, so leave now.
				 */
			break;
		}
	}
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
}

/* called from rx or tx thread to free all unsolicited data buffers
 * allocated to this command
 */
void free_udata_list(struct iscsi_cmnd *cmnd)
{
	struct list_head *list_ptr, *list_temp;
	struct udata_list *udata;
	DECLARE_LOOP_COUNTER(count);

	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp,
			   &cmnd->unsolicited_data_list) {
		udata =
		    list_entry(list_ptr, struct udata_list, udata_link);
		BUMP_LOOP_COUNTER(count, "unsolicited_data_list");
		list_del(list_ptr);
#ifdef RDMA_MODE
		free_queued_targ_buf(cmnd, (char **) &udata->buffer,
				     (void **) &udata->rx_iser_pdu);
#endif
		put_unsolicited_buf(udata, cmnd->conn->session);
	}
}

__u32 __attribute__ ((no_instrument_function))
    get_sglen(struct scatterlist *st_list)
{
	return scatterlist_length(st_list);
}

__u8 * __attribute__ ((no_instrument_function))
    get_sgbuf(struct scatterlist * st_list)
{
	return sg_virt(st_list);
}

/* skip thru sg list to appropriate offset for data buffer
 * i is index to sg list item
 * offset is total number of bytes to skip
 * returns number of bytes remaining to be skipped in ith sg list item
 */
__u32 __attribute__ ((no_instrument_function))
    skip_thru_sg_list(struct scatterlist * st_list, __u32 * i, __u32 offset)
{
	__u32 sglen;

	*i = 0;			/* index to current sg list item */
	while (offset > 0) {
		sglen = get_sglen(st_list);
		if (offset < sglen) {
			/* copying starts at "offset" bytes in ith sg list item */
			break;
		}
		/* need to skip over this entire sg list item */
		offset -= sglen;
		st_list++;
		*i += 1;
	}
	return offset;
}

/* until the last data_out received, it's ready for recovery r2t */
void
check_r2t_done(struct iscsi_cmnd *cmd,
	       struct iscsi_init_scsi_data_out *hdr)
{
	struct iscsi_cookie *next;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* in case we lost the last data_out of unsolicited data */
	if (hdr->offset > cmd->session->oper_param->FirstBurstLength)
		cmd->unsolicited_data_present = 0;

	if (cmd->unsolicited_data_present) {
		if (hdr->flags & F_BIT) {
			cmd->data_sn = 0;
			cmd->unsolicited_data_present = 0;
		}
	} else {
		if (hdr->flags & F_BIT) {
			cmd->next_burst_len = 0;
			cmd->data_sn = 0;
			/* unhook data_q now */
			while (cmd->first_data_q) {
				next = cmd->first_data_q->next;
				my_kfree((void **) &cmd->first_data_q,
					 "first_data_q");
				cmd->first_data_q = next;
			}
			cmd->last_data_q = NULL;
			if (!cmd->recovery_r2t)
				cmd->outstanding_r2t--;
			if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
				TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
				      "check_r2t_done\n", jiffies,
				      current->comm);
				up(&cmd->conn->tx_sem);
			}
		}
	}
}

/* move to next sequence of DataIn pdus for this READ command */
void __attribute__ ((no_instrument_function))
    move_to_next_seq(struct iscsi_cmnd *cmnd, struct iscsi_session *session)
{
	cmnd->data_length_left -= cmnd->seq_length;
	if (session->oper_param->DataSequenceInOrder)
		cmnd->seq_low_byte = cmnd->seq_limit_byte;	/* in order */
	else
		cmnd->seq_limit_byte = cmnd->seq_low_byte;	/* reverse order */
}

int get_r2t_xfer_len(struct iscsi_cmnd *cmd, int r2tsn, int *len)
{
	int retval = -1;
	struct iscsi_cookie *cookie = cmd->first_r2t_cookie;

	TRACE(TRACE_ENTER_LEAVE, "Enter %s\n", __func__);

	while (cookie) {
		if (cookie->seq == r2tsn) {
			*len = cookie->xfer_len;
			retval = 1;
			break;
		}
		cookie = cookie->next;
	}

	if (retval == -1) {
		TRACE_ERROR
		    ("Unidentified R2T seq. no %d retransmission request received \n",
		     r2tsn);
	}
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return retval;
}

/*
 * search_tags: This goes through the command queue
 * for the session to which conn belongs to look for a match to
 * a given set of tags.
 * INPUT: connection, initiator and target tag
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_tags(struct iscsi_conn *conn, __u32 init_task_tag,
	    __u32 target_xfer_tag, int dumpall)
{
	struct iscsi_cmnd *cmd = NULL;
	struct iscsi_session *session;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!conn) {
		TRACE_ERROR("%s Cannot search a NULL connection\n",
			    current->comm);
		return NULL;
	}

	/* get session from conn structure directly to support multiple sessions */
	session = conn->session;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	if (dumpall || TRACE_TEST(TRACE_DEBUG)) {
		INIT_LOOP_COUNTER(count);
		list_for_each_entry(cmd, &session->active_command_list,
				    command_link) {
			BUMP_LOOP_COUNTER(count, "active_command_list");
			if (!cmd)
				continue;
			printk
			    ("%s scsi cmnd %p opcode 0x%02x CmdSN %u ITT %u TTT %u "
			     "xfer length %d data_done %d data_length_left %d StatSN %u "
			     "state %u\n", current->comm, cmd->cmnd,
			     cmd->opcode_byte, cmd->cmd_sn,
			     cmd->init_task_tag, cmd->target_xfer_tag,
			     cmd->data_length, cmd->data_done,
			     cmd->data_length_left, cmd->stat_sn,
			     cmd->state);
		}
	}

	INIT_LOOP_COUNTER(count);
	list_for_each_entry(cmd, &session->active_command_list,
			    command_link) {
		BUMP_LOOP_COUNTER(count, "active_command_list");
		if ((cmd->init_task_tag == init_task_tag)
		    && ((cmd->target_xfer_tag == target_xfer_tag)
			|| (target_xfer_tag == ALL_ONES))) {
			spin_unlock_irqrestore(&session->cmnd_lock, flags);

			TRACE(TRACE_DEBUG, "Search found the command\n");
			TRACE(TRACE_DEBUG,
			      "scsi cmnd %p, init_task_tag %u target_xfer_tag %u"
			      " data_done %d xfer length %d\n", cmd->cmnd,
			      cmd->init_task_tag, cmd->target_xfer_tag,
			      cmd->data_done, cmd->data_length);

			goto search_tags_out;
		}
	}
	cmd = NULL;
	spin_unlock_irqrestore(&session->cmnd_lock, flags);

      search_tags_out:
	return cmd;
}

/*
 * executed only by the tx thread.
 * auxilliary function called only by handle_discovery_rsp()
 */
int __attribute__ ((no_instrument_function))
    determine_text_to_send(struct iscsi_cmnd *cmnd,
		       struct iscsi_conn *conn,
		       struct iscsi_targ_text_rsp *hdr)
{
	int size, offset;
	int next_state;
	struct iscsi_cmnd *next_in_progress;

	/* get offset to, and amount of, data remaining to send back to initiator */
	offset = cmnd->data_done;
	size = cmnd->data_length - offset;
	cmnd->reply_data = cmnd->ping_data + offset;

	if (size > 0 && conn->connection_flags & USE_ONE_KEY_PER_TEXT) {
		/* send just one key per text response (for testing) */
		size = strlen(cmnd->reply_data) + 1;
		cmnd->data_done += size;
		next_state = ISCSI_BLOCKED_SENDING_TEXT;
		next_in_progress = cmnd;
	} else if (size > conn->max_send_length) {
		/* reply too big for one pdu, more text reply pdus follow this one */
		size = conn->max_send_length;
		cmnd->data_done += size;
		hdr->flags |= C_BIT;
		next_state = ISCSI_BLOCKED_SENDING_TEXT;
		next_in_progress = cmnd;
	} else if (cmnd->command_flags & F_BIT) {
		/* this is our last reply and initiator has no more to send */
		hdr->flags |= F_BIT;
		cmnd->target_xfer_tag = hdr->target_xfer_tag = ALL_ONES;
		next_state = ISCSI_SENT;
		next_in_progress = NULL;
	} else {
		/* this is our last reply but initiator has more to send */
		cmnd->data_length = 0;	/* reset for next Request PDU */
		cmnd->in_progress_buffer = cmnd->ping_data;
		cmnd->ping_data = NULL;
		next_state = ISCSI_AWAIT_MORE_TEXT;
		next_in_progress = cmnd;
	}
	cmnd->reply_size = size;
	cmnd->state = next_state;
	conn->text_in_progress = next_in_progress;
	hdr->length = cpu_to_be32(size);
	return size;
}

/*
 * executed only by the tx thread.
 * iscsi_dequeue: this function frees up a command's allocated memory
 * after it has been removed from the session list (before the call)
 * INPUT: command which has to be freed
 */
static inline void iscsi_dequeue(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn;

	struct list_head *orig_next;


#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif
	orig_next = cmnd->command_link.next;

	if (orig_next != cmnd->command_link.next)
		TRACE_ERROR("%s: next link clobbered before everythingy\n",
			    __func__);


	TRACE(TRACE_ISCSI, "%lu %s Dequeue %p CmdSN %u ITT %d\n",
	      jiffies, current->comm, cmnd, cmnd->cmd_sn,
	      cmnd->init_task_tag);

	if (cmnd->cmnd != NULL) {
		TRACE(TRACE_ISCSI, "%s: cmnd=%p CmdSn=%u ITT %d done\n",
		      __func__, cmnd, cmnd->cmd_sn, cmnd->init_task_tag);
		if (scsi_target_done(cmnd->cmnd) < 0) {
			TRACE_ERROR
			    ("%s scsi_target_done returned an error\n",
			     current->comm);
		}
	}


	if (orig_next != cmnd->command_link.next)
		TRACE_ERROR("%s: next link clobbered after scsi_target_done\n",
			    __func__);

	/* if this command was the text command currently in progress on this
	 * connection, unblock the connection so it can process more text commands.
	 */
	conn = cmnd->conn;
	if (conn) {
		spin_lock(&conn->text_in_progress_lock);
		if (conn->text_in_progress == cmnd)
			conn->text_in_progress = NULL;
		spin_unlock(&conn->text_in_progress_lock);
	}

	if (orig_next != cmnd->command_link.next)
		TRACE_ERROR("%s: next link clobbered after in progress stuff\n",
			    __func__);

	release_command(cmnd);

	if (orig_next != cmnd->command_link.next)
		TRACE_ERROR("%s: next link clobbered after release_command\n",
			    __func__);
}

/*
 * executed by the rx and tx threads.
 * free up all commands on (non-empty) dequeue list
 * then append dequeue list to end of session's free command list
 */
void iscsi_dequeue_list(struct list_head *dequeue_list,
			struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd, *next;
	DECLARE_LOOP_COUNTER(count);
	unsigned long flags;

	INIT_LOOP_COUNTER(count);
	list_for_each_entry_safe(cmnd, next, dequeue_list, command_link) {
		BUMP_LOOP_COUNTER(count, "dequeue_list");

		cmnd_check_magic(session, cmnd);

		iscsi_dequeue(cmnd);
	}

	/* now append the dequeue_list to end of free command list */
	spin_lock_irqsave(&session->cmnd_lock, flags);
	list_splice(dequeue_list, session->free_command_list.prev);
	INIT_LIST_HEAD(dequeue_list);
	spin_unlock_irqrestore(&session->cmnd_lock, flags);
}

/*
 * executed only by the tx thread.
 * no locks should be held by calling process/thread (since we do I/O here)
 * iscsi_tx_reject: this function transmits a reject to the Initiator
 * the attached 48-byte data segment is the header of the rejected PDU
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int iscsi_tx_rjt(struct iscsi_conn *conn, __u8 * bad_hdr, __u8 reason)
{
	int err = 0;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_rjt *hdr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_rjt *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_RJT;
	hdr->flags |= F_BIT;
	hdr->reason = reason;
	hdr->length = cpu_to_be32(ISCSI_HDR_LEN);
	/* init task tag must be all ones - new 18_04 SAI */
	hdr->init_task_tag = ALL_ONES;

	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	if (conn->session) {
		hdr->exp_cmd_sn = cpu_to_be32(conn->session->exp_cmd_sn);
		hdr->max_cmd_sn = cpu_to_be32(conn->session->max_cmd_sn);
	}

	if (send_hdr_plus_1_data
	    (NULL, conn, iscsi_hdr, bad_hdr, ISCSI_HDR_LEN,
	     USE_SEND_CONTROL) < 0) {
		err = -1;
		goto out;
	}

	TRACE_WARNING("%s Send Reject, reason 0x%02x\n", current->comm,
		      reason);

	if (TRACE_TEST(TRACE_ISCSI_FULL | TRACE_ERROR_RECOVERY))
		print_targ_rjt(hdr);

      out:
	return err;
}

/*
 * find_iovec_needed: this function returns the number of iovecs that
 * are needed to receive some data
 * INPUT: amount of data needed, pointer to scattergather list (from the
 * 	  point where the data is to start for this round), offset within
 * 	  the first scattergather element - in case it was not used up
 * 	  completely
 * OUTPUT: number of iovectors needed, else < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
    find_iovec_needed(__u32 data_len, struct scatterlist *st_list,
		  __u32 offset)
{
	int i = 0;
	__u32 sglen;

	while (data_len) {
		sglen = get_sglen(st_list);
		if (data_len > (sglen - offset))
			data_len -= (sglen - offset);
		else
			data_len = 0;

		i++;
		st_list++;
		offset = 0;
	}

	return (i);
}

/*
 * fill_iovec: This function fills in a given iovec structure so as to
 * reflect a given scattergather list upto the length that is required
 * No checks are made on the length on the iovec given except that it
 * needs to be greater than the size needed. Also, the scatterlist is
 * also not checked. The integer offset upon return reflects the posn of
 * the offset in the last scatterlist that was used for this function
 * INPUT: struct iovec, no. of iovecs, scatterlist, ptr to offset within
 * 	  the first element, and the total length required
 * OUTPUT: how many st_list elements are used, < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
    fill_iovec(struct iovec *iov, int p, int niov,
	   struct scatterlist *st_list, __u32 * offset, __u32 data)
{
	int count = 0;
	__u32 sglen;

	TRACE(TRACE_DEBUG, "offset: %d, data_len: %d\n", *offset, data);

	iov += p;
	while (data > 0 && p < niov) {
		iov->iov_base = (void *) (get_sgbuf(st_list) + (*offset));
		sglen = get_sglen(st_list);
		if ((sglen - *offset) > data) {
			/* more space left in this st list item than iovector can use */
			iov->iov_len = data;
			*offset += data;
		} else {
			/* iovector uses all the space in this st list item */
			iov->iov_len = sglen - *offset;
			*offset = 0;
			count++;
		}

		data -= iov->iov_len;

		TRACE(TRACE_DEBUG,
		      "iov %p, p %d, iov_base %p, iov_len %u\n", iov, p,
		      iov->iov_base, (unsigned int) iov->iov_len);

		p++;
		iov++;
		st_list++;
	}

	return count;
}

/**
 * \brief Allocates a new connection structure.
 * \return A newly allocated connection structure, or NULL on error.
 */
static struct iscsi_conn *alloc_conn(struct portal_group *pg)
{
	struct iscsi_conn *conn;

	conn = my_kmalloc(sizeof(struct iscsi_conn), "iscsi_conn");
	if (!conn)
		return NULL;
	memset(conn, 0, sizeof(struct iscsi_conn));

	/* get ip addressing info about both ends of this connection */
	conn->ip_address =
	    my_kmalloc(sizeof(struct sockaddr), "ip address");
	if (!conn->ip_address)
		goto err;
	conn->local_ip_address = my_kmalloc(sizeof(struct sockaddr),
					    "local ip address");
	if (!conn->local_ip_address)
		goto err1;
	conn->conn_socket = NULL;
	conn->conn_status = CONN_DOWN;

	/* allocate and fill in connection details */
	if (build_conn_sess(conn, pg) == -1)
		goto err1;

	return conn;

      err1:
	my_kfree((void **) &conn->ip_address, "ip address");
      err:
	my_kfree((void **) &conn, "iscsi_conn");
	return NULL;
}

/**
 * \brief Deallocates an iscsi_conn structure.
 * \param conn The connection to deallocate
 */
static void dealloc_conn(struct iscsi_conn *conn)
{
#if defined(RDMA_MODE)
	queue_free(&conn->rx_pdu_q);
	queue_free(&conn->rx_handle_q);
#endif /* RDMA_MODE  */
	TRACE(TRACE_TIMERS, "%s: deleting nop_timer %p\n",
	      __func__, &conn->nop_timer);
	del_timer_sync(&conn->nop_timer);
	param_tbl_uncpy(*conn->session->session_params);
	my_kfree((void **) &conn->local_ip_address, "local ip address");
	my_kfree((void **) &conn->ip_address, "ip address");
	my_kfree((void **) &conn, "iscsi_conn");
}

/*
 * Note the name "driver_template" is required for "scsi_module.c" to work!
 *      Used by: "/usr/src/linux/drivers/scsi/scsi_module.c"
 *      Wrap this in an ifdef because "scsi_module.c" says to!
 */
#ifdef MODULE

MODULE_AUTHOR("UNH IOL ISCSI");
MODULE_DESCRIPTION("UNH-IOL iSCSI reference target");
MODULE_SUPPORTED_DEVICE("sd");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#endif

/*
 * initializor for static Scsi_Target_Template
 *
 * These functions are defined in scsi_midlevel.c
 * 	-- eaburns
 */
#define ISCSI_TEMPLATE {\
  name: "iscsi_target", \
  detect:iscsi_detect, \
  release:iscsi_release,\
  xmit_response:iscsi_xmit_response, \
  rdy_to_xfer:iscsi_rdy_to_xfer,\
  proc_info:iscsi_proc_info, \
  task_mgmt_fn_done:iscsi_task_mgt_fn_done \
};

Scsi_Target_Template my_template = ISCSI_TEMPLATE;

/* EAB - I am not sure why this was *included* */
//#include "iscsi_target_module.c"
