/*	target/iscsi_target.h
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	header file for the iSCSI Target
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
*/

#ifndef	_ISCSI_TARGET_H
#define	_ISCSI_TARGET_H

#ifdef __KERNEL__

#include <linux/version.h>

#include <linux/proc_fs.h>

#else /* __KERNEL__ */

#include "../common/user_space.h"

#endif /* !__KERNEL__ */

#include "../common/scatterlist.h"
#include "../common/scsi_cmnd.h"

#include "scsi_target.h"

#include "../common/iscsi_common.h"
#include "../common/debug.h"
#include "../common/my_memory.h"
#include "../common/range.h"
#include "../common/crc.h"
#include "../common/tcp_utilities.h"

/* chap and srp support - CHONG */
#include "../security/misc/misc_func.h"
#include "../security/chap/chap.h"
#include "../security/srp/srp.h"

#include "../common/text_param.h"
#include "../common/target_negotiate.h"
#include "iscsi_portal_group.h"

/* a "forward declaration" of this structure type */
struct iscsi_cmnd;

/* 4 bytes of zeroes that can be sent as pad bytes when needed */
extern __u32 pad_bytes_for_sending;

/* Globals in iscsi_target.c */
extern int max_send_niov;
extern int max_recv_niov;

#ifdef RDMA_MODE
#include <iser_api.h>
#endif
#include "iser_target.h"

/* to limit all loops (for debugging) */
#ifdef MAX_LOOP_COUNT
#define DECLARE_LOOP_COUNTER(name)	__u32 name
#define INIT_LOOP_COUNTER(name)		name = 0
#define TEST_LOOP_COUNTER(name, mess)	\
			if (name >= MAX_LOOP_COUNT) { \
				TRACE_ERROR("%s max loop count %u, %s\n", \
					    current->comm, name, mess); \
				break; \
			}  else \
				(void)(name)
#define BUMP_LOOP_COUNTER(name, mess)	\
			name++; \
			TEST_LOOP_COUNTER(name, mess)
#else
#define DECLARE_LOOP_COUNTER(name)	struct name
#define INIT_LOOP_COUNTER(name)		do {} while(0)
#define TEST_LOOP_COUNTER(name, mess)	do {} while(0)
#define BUMP_LOOP_COUNTER(name, mess)	do {} while(0)
#endif

/* To identify which Datamover/Stream function to call */
#define USE_TCP_STREAM		0
#define USE_PUT_DATA		1
#define USE_GET_DATA		2
#define USE_SEND_CONTROL	3

/* maximum number of iov slots iscsi_target can set up for I/O */
#define MAX_IOV_SLOTS	260

extern struct iscsi_global *devdata;

int iscsi_detect(Scsi_Target_Template *);
int iscsi_release(Scsi_Target_Device *);
int iscsi_xmit_response(Target_Scsi_Cmnd *);
int iscsi_rdy_to_xfer(Target_Scsi_Cmnd *);
int iscsi_proc_info(char *, char **, off_t, int, int, int);
void iscsi_task_mgt_fn_done(Target_Scsi_Message * msg);

int iscsi_server_thread(void *);
int iscsi_rx_thread(void *);
int iscsi_tx_thread(void *);


#define PDU_SENSE_LENGTH_SIZE 2
#define SENSE_STRUCTURE_SIZE 18


/*
 * executed only by the rx thread, in spite of the tx in the name!
 * iscsi_tx_reject: this function transmits a reject to the Initiator
 * the attached 48-byte data segment is the header of the rejected PDU
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int iscsi_tx_rjt(struct iscsi_conn *conn, __u8 * bad_hdr, __u8 reason);

/*
 * executed only by the rx thread.
 * Allocates new reject item, fills it in with header of rejected pdu,
 * and enqueues it for tx thread on reject_list of this connection.
 */
int enqueue_reject(struct iscsi_conn *conn, __u8 reason);


/*
 * find_iovec_needed: this function returns the number of iovecs that
 * are needed to receive some data
 * INPUT: amount of data needed, pointer to scattergather list (from the
 * 	  point where the data is to start for this round, offset within
 * 	  the first scattergather element - in case it was not used up
 * 	  completely
 * OUTPUT: number of iovectors needed, else < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
    find_iovec_needed(__u32 data_len,
		  struct scatterlist *st_list, __u32 offset);


/*
 * fill_iovec: This function fills in a given iovec structure so as to
 * reflect a given scattergather list upto the length that is required
 * No checks are made on the length on the iovec given except that it
 * needs to be greater than the size needed. Also, the scatterlist is
 * also not checked. The integer offset upon return reflects the posn of
 * the offset in the last scatterlist that was used for this function
 * INPUT: struct iovec, no. of iovecs, scatterlist, ptr to offset within
 * 	  the first element, and the total length required
 * OUTPUT: how many st_list elements are used, < 0 if there is trouble
 */
int __attribute__ ((no_instrument_function))
    fill_iovec(struct iovec *iov, int i, int niov,
	   struct scatterlist *st_list, __u32 * offset, __u32 data);


/*
 * iscsi_rx_data: will receive a fixed amount of data.
 * INPUT: struct iscsi_conn (what connection), struct iovec, number of iovecs,
 * total amount of data in bytes
 * OUTPUT: int total bytes read if everything is okay
 * 	       < 0 if trouble
 */
int
iscsi_rx_data(struct iscsi_conn *conn, struct iovec *iov,
	      struct iovec *iov_copy, int niov, int data);

__u32 __attribute__ ((no_instrument_function))
    generate_next_ttt(struct iscsi_session *session);

void free_udata_list(struct iscsi_cmnd *cmnd);
int bring_up_portal(struct portal_group *ptr);
void clean_bad_stuff(void);
void report_peer_up_down(struct sockaddr *ip_address, char *message);
int handle_recv_msg(struct iscsi_conn *conn, struct generic_pdu *inputpdu);
void check_queued_cmnd(struct iscsi_session *session);
void iscsi_release_socket(struct iscsi_conn *conn);
int iscsi_release_connection(struct iscsi_conn *conn);
void do_text_request(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		     struct iscsi_session *session);

struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_tags(struct iscsi_conn *conn, __u32 init_task_tag,
	    __u32 target_xfer_tag, int dumpall);

/* XXX - maybe move to scsi_midlevel.c */
int deliver_to_midlevel(struct iscsi_cmnd *cmnd,
			struct iscsi_init_scsi_cmnd *pdu,
			struct iscsi_session *session);

void __attribute__ ((no_instrument_function))
    do_task_mgt(struct iscsi_conn *conn, struct iscsi_cmnd *tm_cmnd);

void check_r2t_done(struct iscsi_cmnd *cmd,
		    struct iscsi_init_scsi_data_out *hdr);

__u32 __attribute__ ((no_instrument_function))
    skip_thru_sg_list(struct scatterlist *st_list, __u32 * i, __u32 offset);

void __attribute__ ((no_instrument_function))
    move_to_next_seq(struct iscsi_cmnd *cmnd, struct iscsi_session *session);

void put_unsolicited_buf(struct udata_list *udata,
			 struct iscsi_session *session);

void iscsi_dequeue_list(struct list_head *dequeue_list,
			struct iscsi_session *session);

void restart_nop_timer(struct iscsi_conn *conn);
int dequeue_reject(struct iscsi_conn *conn, int sendit);

int send_hdr_plus_1_data(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
			 void *iscsi_hdr, void *data_buf,
			 int data_len, int dm_flag);

int __attribute__ ((no_instrument_function))
    determine_text_to_send(struct iscsi_cmnd *cmnd,
		       struct iscsi_conn *conn,
		       struct iscsi_targ_text_rsp *hdr);
int send_hdr_only(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		  void *iscsi_hdr);

__u32 get_sglen(struct scatterlist *st_list);
__u8 *get_sgbuf(struct scatterlist *st_list);

int iscsi_tx_data(struct iscsi_cmnd *cmnd, struct iscsi_conn *conn,
		  struct iovec *iov, struct iovec *iov_copy, int niov,
		  int data, int dm_flag);

int get_r2t_xfer_len(struct iscsi_cmnd *cmd, int r2tsn, int *len);

int iscsi_bring_up_portal(struct portal_group *ptr);

void __attribute__ ((no_instrument_function))
    __update_after_read(unsigned int length, unsigned int fbit,
		    struct iscsi_cmnd *cmnd);

#endif
