/* target/iscsi_tx.c -- Contains function definitions for iscsi_tx_thread()
 * and friends.
 *
 * vi: set autoindent tabstop=8 shiftwidth=8 textwidth=78 :
 *
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * Created Fri 02 Mar 2007 01:41:14 PM EST -- Ethan Burns
 */

#include "iscsi_target.h"
#include "iscsi_cmnd.h"
#include "target_error_rec.h"
#include "../common/unh_threads.h"

/* from /usr/src/linux/include/scsi/scsi.h */
#define host_byte(result)	(((result) >> 16) & 0xff)

/*
 * Static declarations
 */

static int handle_resend_text_rsp(struct iscsi_cmnd *cmnd,
				  struct iscsi_conn *conn,
				  struct iscsi_session *session);
static int handle_discovery_rsp(struct iscsi_cmnd *cmnd,
				struct iscsi_conn *conn,
				struct iscsi_session *session,
				__u32 expected_state);
static int handle_logout_rsp(struct iscsi_cmnd *cmnd,
			     struct iscsi_conn *conn,
			     struct iscsi_session *session);
static int handle_resend_nop_in(struct iscsi_cmnd *cmnd,
				struct iscsi_conn *conn,
				struct iscsi_session *session);

static int handle_iscsi_done(struct iscsi_cmnd *cmnd,
			     struct iscsi_conn *conn,
			     struct iscsi_session *session);

static int handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd,
				    struct iscsi_conn *conn,
				    struct iscsi_session *session);

static int iscsi_tx_r2t(struct iscsi_cmnd *cmnd,
			struct iscsi_conn *conn,
			struct iscsi_session *session,
			unsigned long flags, int *cmnd_lock_held);

static int
need_a_nop(struct iscsi_conn *conn, struct iscsi_session *session);

static int send_unsolicited_data(struct iscsi_cmnd *cmd,
				 struct iscsi_conn *conn,
				 struct iscsi_session *session);
static int send_iscsi_response(struct iscsi_cmnd *cmnd,
			       struct iscsi_conn *conn,
			       struct iscsi_session *session);
static int __attribute__ ((no_instrument_function))
    generate_nopin(struct iscsi_conn *conn, struct iscsi_session *session);

static int send_read_data(struct iscsi_cmnd *cmnd,
			  struct iscsi_conn *conn,
			  struct iscsi_session *session,
			  int *phase_collapse);
static int handle_nopin(struct iscsi_cmnd *cmnd,
			struct iscsi_conn *conn,
			struct iscsi_session *session);

static void __attribute__ ((no_instrument_function))
    save_data_in_seq(struct iscsi_cmnd *cmnd);

static __u32 __attribute__ ((no_instrument_function))
    do_command_status(struct iscsi_cmnd *cmnd, Scsi_Request * req);

static int fill_and_send_datain_iov(struct iscsi_cmnd *cmnd,
				    struct iscsi_conn *conn,
				    int data_payload_length,
				    struct scatterlist *st_list,
				    __u8 * iscsi_hdr);

static void
set_a_bit(struct iscsi_cmnd *cmnd, struct iscsi_targ_scsi_data_in *hdr);

#if defined(MANGLE_INQUIRY_DATA)
static int mangle_inquiry_data(struct iscsi_cmnd *iscsi_command,
			       struct iovec *iov, int start_iov,
			       int limit_iov);
#endif				/* MANGLE_INQUIRY_DATA */

/**
 * Executed by the tx thread when called there should be NO LOCKS HELD.
 */
static void tx_shutdown_thread(void *aux)
{
	long flags;
	__u32 count;
	struct iscsi_cmnd *cmnd;
	struct iscsi_conn *conn;
	struct iscsi_session *session;
#ifndef __KERNEL__
	struct task_struct *current;
	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	conn = aux;
	session = conn->session;

	if (TRACE_TEST(TRACE_DEBUG) && session) {
		/* lock the session-wide list of commands */
		spin_lock_irqsave(&session->cmnd_lock, flags);

		count = 0;
		list_for_each_entry(cmnd, &session->active_command_list,
				    command_link) {
			count++;
			TEST_LOOP_COUNTER(count, "active_command_list");
			printk
			    ("%s opcode 0x%02x ITT %u CmdSN %u TTT %u data_done %d "
			     "xfer length %d stat_sn %u data_sn %u state %u\n",
			     current->comm, cmnd->opcode_byte,
			     cmnd->init_task_tag, cmnd->cmd_sn,
			     cmnd->target_xfer_tag, cmnd->data_done,
			     cmnd->data_length, cmnd->stat_sn,
			     cmnd->data_sn, cmnd->state);
		}
		/* unlock the session-wide list of commands */
		spin_unlock_irqrestore(&session->cmnd_lock, flags);
	}
	conn->active = 0;
	conn->tx_thread = NULL;

	/*  This is deleted by the deallocation of the connection instead
	TRACE(TRACE_TIMERS, "%s: deleting nop_timer=%p\n",
	      __func__, &conn->nop_timer);
	del_timer_sync(&conn->nop_timer);
	*/

	if (conn->rx_thread) {
		/* rx thread still going, shut it down */
		send_sig(ISCSI_SHUTDOWN_SIGNAL, conn->rx_thread, 1);

#ifndef __KERNEL__
		/* in user space, send_sig() yields to rx_thread's cleanup handler */
		conn->rx_thread = NULL;	/* so rx_thread is truly dead now */
#endif

	} else {
		/* rx thread has shut down, if socket is still open, release it */
		iscsi_release_socket(conn);
	}
	up(&conn->kill_tx_sem);
#ifndef __KERNEL__
	printk("%s Exiting thread id %#08lx\n", current->comm,
	       current->threadid);
#endif				/* __KERNEL__ */

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/*
 * iscsi_tx_thread: This thread is responsible for the transmission of
 * responses and/or status information to the initiator
 */
int iscsi_tx_thread(void *param)
{
	struct iscsi_conn *conn = (struct iscsi_conn *) param;
	struct iscsi_session *session = NULL;
	struct iscsi_cmnd *cmnd;
	struct list_head dequeue_list;
	struct list_head *list_ptr, *list_temp;
	__u32 active_seqno;
	__u32 count, tied_up_count;
	int delta, cmnd_lock_held;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	lock_kernel();

	unh_daemonize("iscsi_tx_%d", conn->conn_id);

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	/* mark that this tx thread is alive */
	conn->tx_thread = current;

	unlock_kernel();

#ifndef __KERNEL__
	printk("%s Starting thread id %#08lx\n", current->comm,
	       current->threadid);
#endif				/* !__KERNEL__ */

	/* set up to start the nopin timer first time through the main loop */
	set_bit(NEED_NOPIN_BIT, &conn->control);
	set_bit(SILENCE_BIT, &conn->control);
	conn->nop_period_remaining = conn->nop_period;


	while (1) {
		/* nothing on dequeue list yet */
		INIT_LIST_HEAD(&dequeue_list);

		/* block until somebody adds/changes a command for this conn */
		TRACE(TRACE_SEM, "%lu %s wait on tx_sem\n", jiffies,
		      current->comm);

#if !defined(__KERNEL__)
		pthread_cleanup_push(tx_shutdown_thread, conn);
#endif /* !__KERNEL__ */

		if (down_interruptible(&conn->tx_sem)) {
			TRACE_ERROR("%s: interrupted\n", __func__);
			goto iscsi_tx_thread_out;
		}

#if !defined(__KERNEL__)
	pthread_cleanup_pop(0);
#endif /* !__KERNEL__ */

		TRACE(TRACE_SEM, "%lu %s after tx_sem\n", jiffies,
		      current->comm);

		/* can't set session sooner because it is not set up in conn until
		 * AFTER login is complete on this connection.  But nothing will wake
		 * up this tx thread until AFTER that login is complete, so we are safe.
		 */
		session = conn->session;

		TRACE(TRACE_DEBUG, "%s awake, session %p, conn %p\n",
		      current->comm, session, conn);

		if (test_and_clear_bit(NEED_NOPIN_BIT, &conn->control)) {
			/* bit was set, either first time through
			 * or timer expired after an interval of silence from initiator
			 */
			if (!timer_pending(&conn->nop_timer)) {
				/* first time thru here -- timer not going yet, start it up */
				restart_nop_timer(conn);
			} else {
				/* not first time thru here, timer must have expired */
				TRACE(TRACE_TIMERS,
				      "%s head of loop clear NEED_NOPIN\n",
				      current->comm);
				if (need_a_nop(conn, session) < 0) {
					goto iscsi_tx_thread_out;
				}
			}
		}

		if (!list_empty(&conn->reject_list)) {
			/* process connection's queue of pending rejects */
			if (dequeue_reject(conn, 1))
				goto iscsi_tx_thread_out;
		}
#ifdef K26
		if (irqs_disabled()) {
			TRACE_WARNING
			    ("%lu %s tx_thread irqs disabled prior to "
			     "spin_lock_irqsave\n", jiffies,
			     current->comm);
		}
#endif

		/* lock the session-wide list of commands */
		spin_lock_irqsave(&session->cmnd_lock, flags);

		/* look at each command in session to which this connection belongs */
		tied_up_count = count = 0;
		list_for_each_safe(list_ptr, list_temp,
				   &session->active_command_list) {
			cmnd =
			    list_entry(list_ptr, struct iscsi_cmnd,
				       command_link);
			count++;
			TEST_LOOP_COUNTER(count, "active_command_list");
			TRACE(TRACE_DEBUG, "%s pick up cmnd[%u] %p\n",
			      current->comm, count, cmnd);
			if (cmnd->conn == conn) {
				/* this command is for this connection, handle it */
				if (!(conn->connection_flags & CONN_LOGGED_OUT)) {
					TRACE(TRACE_DEBUG,
					      "%s handle cmnd no. %u, ITT %u, "
					      "opcode 0x%02x, state %d\n",
					      current->comm, count,
					      cmnd->init_task_tag,
					      cmnd->opcode_byte,
					      cmnd->state);
					TRACE(TRACE_DEBUG,
					      "%s UnsolData %u, data_len %u, "
					      "data_done %u, r2t_data_left %d\n",
					      current->comm,
					      cmnd->
					      unsolicited_data_present,
					      cmnd->data_length,
					      cmnd->data_done,
					      cmnd->r2t_data_left);
					active_seqno = cmnd->active_seqno;

					switch (cmnd->state) {

					case ISCSI_RESEND_TEXT_RESPONSE:
						{
							TRACE(TRACE_DEBUG,"%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_resend_text_rsp(cmnd, conn, session) < 0) {
								TRACE_ERROR
								    ("%s Trouble in resend_text_rsp\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_SEND_TEXT_RESPONSE:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_discovery_rsp(cmnd, conn, session, ISCSI_SEND_TEXT_RESPONSE) < 0) {
								TRACE_ERROR
								    ("%s Trouble in handle_discovery_rsp\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_ASK_FOR_MORE_TEXT:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_discovery_rsp(cmnd, conn, session, ISCSI_ASK_FOR_MORE_TEXT) < 0) {
								TRACE_ERROR
								    ("%s Trouble in handle_discovery_rsp\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_LOGOUT:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							/* move this command from active to dequeue list */
							list_move_tail
							    (list_ptr,
							     &dequeue_list);
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_logout_rsp(cmnd, conn, session) < 0) {
								TRACE_ERROR
								    ("%s Trouble in handle_logout_rsp\n",
								     current->
								     comm);
							}
							/* always exit the tx thread after a logout response */
							iscsi_dequeue_list
							    (&dequeue_list,
							     session);
							goto iscsi_tx_thread_out;
						}

					case ISCSI_RESEND_NOP_IN:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_resend_nop_in(cmnd, conn, session) < 0) {
								TRACE_ERROR
								    ("%s Trouble in resend_nop_in\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_PING:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_nopin
							    (cmnd, conn,
							     session) <
							    0) {
								TRACE_ERROR
								    ("%s Trouble in handle_nopin\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_REDO_DATA:
					case ISCSI_GOT_DATA_ACK:
					case ISCSI_DONE:
						{
							TRACE(TRACE_DEBUG, "%s: cmd_sn=%u, state=%s\n",
							       __func__,
							       cmnd->cmd_sn,
							       cmnd_state_name(cmnd->state));
							spin_unlock_irqrestore(&session->cmnd_lock, flags);
							if (handle_iscsi_done(cmnd, conn, session) < 0) {

								TRACE_ERROR("%s Trouble in handle_iscsi_done\n",
								     current->comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave(&session->cmnd_lock, flags);
							if (active_seqno != cmnd->active_seqno)
								goto restart_active_list;
							list_temp = list_ptr->next;
							break;
						}

					case ISCSI_RESEND_STATUS:
						{
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (send_iscsi_response(cmnd, conn, session) < 0) {
								TRACE_ERROR
								    ("%s Trouble in send_iscsi_response\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_MGT_FN_DONE:
						{
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							if (handle_iscsi_mgt_fn_done(cmnd, conn, session) < 0) {
								TRACE_ERROR
								    ("%s Trouble in iscsi_mgt_fn_done\n",
								     current->
								     comm);
								goto iscsi_tx_thread_out;
							}
							/* relock the session-wide list of commands */
							spin_lock_irqsave
							    (&session->
							     cmnd_lock,
							     flags);
							if (active_seqno !=
							    cmnd->
							    active_seqno)
								goto restart_active_list;
							list_temp =
							    list_ptr->next;
							break;
						}

					case ISCSI_BUFFER_RDY:
						{
							cmnd_lock_held = 1;
							if (iscsi_tx_r2t(cmnd, conn, session, flags, &cmnd_lock_held) < 0) {
								TRACE_ERROR("%s Trouble in iscsi_tx_r2t\n", current->comm);
								if (cmnd_lock_held)
									spin_unlock_irqrestore(&session->cmnd_lock, flags);
								goto iscsi_tx_thread_out;
							}
							if (!cmnd_lock_held) {
								/* relock the session-wide list of commands */
								spin_lock_irqsave(&session->cmnd_lock, flags);
								if (active_seqno != cmnd->active_seqno)
									goto restart_active_list;
								list_temp = list_ptr->next;
							}
							break;
						}

					case ISCSI_DEQUEUE:
						{
							TRACE(TRACE_DEBUG,
							      "dequeue command, ITT %u, CmdSN %u,"
							      " count %u\n",
							      cmnd->
							      init_task_tag,
							      cmnd->cmd_sn,
							      count);
							/* move this command from active to dequeue list */
							list_move_tail
							    (list_ptr,
							     &dequeue_list);
							break;
						}

					case ISCSI_QUEUE_CMND_RDY:
						{
							if (!cmnd->
							    unsolicited_data_present
							    &&
							    send_unsolicited_data
							    (cmnd, conn,
							     session) <
							    0) {
								TRACE_ERROR
								    ("%s Trouble in send_unsolicited_data\n",
								     current->
								     comm);
								spin_unlock_irqrestore
								    (&session->
								     cmnd_lock,
								     flags);
								goto iscsi_tx_thread_out;
							}
							break;
						}

					case ISCSI_SENT:
						{
							if ((cmnd->opcode_byte & ISCSI_OPCODE) != ISCSI_INIT_NOP_OUT)
							{
								delta = jiffies - cmnd->time_stamp;
								if (delta >= NOP_TIME_LIMIT)
								{
									tied_up_count++;
									set_bit(NEED_NOPIN_BIT, &conn-> control);
									TRACE
									    (TRACE_TIMERS,
									     "%s ISCSI_SENT set "
									     "NEED_NOPIN opcode 0x%02x ITT %d StatSN %u "
									     "count %u delta %d\n",
									     current->comm,
									     cmnd->opcode_byte & ISCSI_OPCODE,
									     cmnd->init_task_tag,
									     cmnd->stat_sn,
									     tied_up_count,
									     delta);
								}
								break;
							}
						}

					case ISCSI_QUEUE_CMND:
					case ISCSI_QUEUE_OTHER:
					case ISCSI_CMND_RECEIVED:
					case ISCSI_NEW_CMND:
					case ISCSI_NOPIN_SENT:
					case ISCSI_AWAIT_BUFFER:
					case ISCSI_AWAIT_BUFFER_QUEUE:
					case ISCSI_UNSOLICITED_DATA_IN:
					case ISCSI_DATA_IN:
					case ISCSI_BLOCKED_SENDING_TEXT:
					case ISCSI_AWAIT_MORE_TEXT:
					case ISCSI_AWAIT_DATA_ACK:
						{
							/* Not much to do */
							break;
						}

					default:	/* NEVER HERE */
						{
							spin_unlock_irqrestore
							    (&session->
							     cmnd_lock,
							     flags);
							TRACE_ERROR
							    ("%s Unknown command state %u\n",
							     current->comm,
							     cmnd->state);
							goto iscsi_tx_thread_out;
						}
					}	/* switch */
				}
			}	/* if */
		}		/* for_each */

	      restart_active_list:
		/* unlock the session-wide list of commands */
		spin_unlock_irqrestore(&session->cmnd_lock, flags);

		/* now go back and free up all commands on local dequeue list */
		if (!list_empty(&dequeue_list)) {
			iscsi_dequeue_list(&dequeue_list, session);
		}

		TRACE(TRACE_DEBUG, "%s handled %u commands\n",
		      current->comm, count);

		if (test_and_clear_bit(NEED_NOPIN_BIT, &conn->control)) {
			delta = jiffies - conn->time_stamp;
			TRACE(TRACE_TIMERS,
			      "%s restart_active clear NEED_NOPIN count "
			      "%u delta %d\n", current->comm,
			      tied_up_count, delta);
			if (delta >= NOP_TIME_LIMIT
			    || tied_up_count >=
			    (session->command_list_size >> 1)) {
				if (need_a_nop(conn, session) < 0) {
					goto iscsi_tx_thread_out;
				}
			}
		}

		/*
		 * Moved check to see if connection is active here so that the
		 * tx_thread can transmit one last packet, i.e., the logout response
		 */
		if (!conn->active)
			goto iscsi_tx_thread_out;
	}			/* while */

      iscsi_tx_thread_out:
	tx_shutdown_thread(conn);

	return 0;
}

/*
 * Static definitions
 */

/*
 * executed only by the tx thread.
 * called when a STATUS SNACK from initiator causes last Text Response pdu
 * to be resent (i.e., when cmnd stat is ISCSI_RESEND_TEXT_RESPONSE)
 * this function uses most of the Text Response pdu previously built
 */
static int handle_resend_text_rsp(struct iscsi_cmnd *cmnd,
				  struct iscsi_conn *conn,
				  struct iscsi_session *session)
{
	int retval = 0;
	struct iscsi_targ_text_rsp *hdr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	spin_lock(&conn->text_in_progress_lock);

	/* restore the command's state to what it was prior to this retransmit */
	cmnd->state = cmnd->reply_state;

	/* RFC 3720 Section 10.16 SNACK Request
	 * "The numbered-response(s) or R2T(s), requested by a SNACK, MUST be
	 * delivered as exact replicas of the ones that the target transmitted
	 * originally except for the fields ExpCmdSN, MaxCmdSN, and ExpDataSN,
	 * which MUST carry the current values."
	 */
	hdr = (struct iscsi_targ_text_rsp *) cmnd->reply_hdr;
	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	if (send_hdr_plus_1_data(cmnd, conn, hdr, cmnd->reply_data,
				 cmnd->reply_size, USE_SEND_CONTROL) < 0)
		goto out1;

	TRACE(TRACE_ISCSI, "%lu %s text response resent, ITT %u\n",
	      jiffies, current->comm, cmnd->init_task_tag);
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_text_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "handle_resend_text_rsp\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}

      out:
	spin_unlock(&conn->text_in_progress_lock);
	return retval;

      out1:
	cmnd->state = ISCSI_DEQUEUE;
	cmnd->init_task_tag = ALL_ONES;
	conn->text_in_progress = NULL;
	retval = -1;
	goto out;
}

/*
 * executed only by the tx thread.
 * called when it is ok to send a Text Response pdu for this command
 * (cmnd state is ISCSI_SEND_TEXT_RESPONSE or ISCSI_ASK_FOR_MORE_TEXT).
 * this function builds the complete Text Response pdu from scratch
 * in the command's reply_hdr area.
 */
static int handle_discovery_rsp(struct iscsi_cmnd *cmnd,
				struct iscsi_conn *conn,
				struct iscsi_session *session,
				__u32 expected_state)
{
	struct iscsi_targ_text_rsp *hdr;
	int size, retval = 0;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	spin_lock(&conn->text_in_progress_lock);

	/* under protection of this lock, make sure command has not been reset */
	if (cmnd->state != expected_state
	    || conn->text_in_progress != cmnd)
		goto out;

	hdr = (struct iscsi_targ_text_rsp *) cmnd->reply_hdr;
	memset(hdr, 0x0, ISCSI_HDR_LEN);

	hdr->opcode = ISCSI_TARG_TEXT_RSP;

	if (expected_state == ISCSI_ASK_FOR_MORE_TEXT) {
		/* Send back empty text response so initiator can send more text data.
		 * This text response pdu has C=0 F=0 DSL=0 ITT=from request,
		 * TTT=newly generated below
		 */
		size = 0;
		cmnd->state = ISCSI_AWAIT_MORE_TEXT;
	} else {
		size = determine_text_to_send(cmnd, conn, hdr);
	}

	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	cmnd->stat_sn = conn->stat_sn++;
	hdr->stat_sn = cpu_to_be32(cmnd->stat_sn);

	spin_lock_irqsave(&session->cmnd_lock, flags);

	/* generate next TTT if we expect initiator to send another text request */
	if (conn->text_in_progress) {
		cmnd->target_xfer_tag = generate_next_ttt(session);
		hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
	}

	/* if the Text command pdu was not immediate, CmdSN has advanced */
	session->exp_cmd_sn += cmnd->cmd_sn_increment;
	cmnd->cmd_sn_increment = 0;
	session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
	cmnd->need_max_cmd_sn_inc = 0;

	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	if (send_hdr_plus_1_data(cmnd, conn, hdr, cmnd->reply_data, size,
				 USE_SEND_CONTROL) < 0) {
		goto out1;
	}

	TRACE(TRACE_ISCSI, "%lu %s text response sent ITT %u\n",
	      jiffies, current->comm, cmnd->init_task_tag);
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_text_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "handle_discovery_rsp\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}

	/* here is where we change the max size of pdus initiator can receive */
	if (cmnd->new_max_send_length) {
		if (conn->max_send_length != cmnd->new_max_send_length) {
			TRACE(TRACE_ISCSI,
			      "%lu %s changed initiator's %s from %u to %u\n",
			      jiffies, current->comm,
			      maxrecvdatasegmentlength_word,
			      conn->max_send_length,
			      cmnd->new_max_send_length);
			conn->max_send_length = cmnd->new_max_send_length;
		}
	}
      out:
	spin_unlock(&conn->text_in_progress_lock);
	return retval;

      out1:
	cmnd->state = ISCSI_DEQUEUE;
	cmnd->init_task_tag = ALL_ONES;
	conn->text_in_progress = NULL;
	retval = -1;
	goto out;
}

/*
 * executed only by the tx thread.
 * called when cmnd->state == ISCSI_LOGOUT to generate logout response pdu.
*/
static int
handle_logout_rsp(struct iscsi_cmnd *cmnd,
		  struct iscsi_conn *conn, struct iscsi_session *session)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_logout_rsp *hdr;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_logout_rsp *) iscsi_hdr;

	hdr->opcode = ISCSI_TARG_LOGOUT_RSP;

	hdr->flags |= F_BIT;

	/* normally we send response = 0, connection or session closed ok */
	if (cmnd->command_flags == LOGOUT_REMOVE_CONN_FOR_RECOVERY
	    && session->oper_param->ErrorRecoveryLevel == 0) {
		/* See RFC 3720 Section 10.15.1 Response (field in Logout Response)
		 * initiator wanted to remove connection for recovery but ERL is zero
		 */
		hdr->response = LOGOUT_RESPONSE_CR_NOT_SUPPORTED;
	}

	/* connection now logged out, don't send any more commands after this */
	conn->connection_flags |= CONN_LOGGED_OUT;

	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
	hdr->stat_sn = cpu_to_be32(conn->stat_sn++);

	cmnd->state = ISCSI_SENT;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	/* if the Logout command pdu was not immediate, CmdSN has advanced */
	session->exp_cmd_sn += cmnd->cmd_sn_increment;
	cmnd->cmd_sn_increment = 0;
	session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
	cmnd->need_max_cmd_sn_inc = 0;

	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	if (send_hdr_only(cmnd, conn, iscsi_hdr) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%lu %s logout response sent\n", jiffies,
	      current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_logout_rsp(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "handle_logout_rsp\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}

	return 0;
}

/*
 * executed only by the tx thread.
 * called when a STATUS SNACK from initiator causes last NopIn pdu
 * to be resent (i.e., when cmnd state is ISCSI_RESEND_NOP_IN)
 * this function uses most of the NopIn pdu previously built
 */
static int handle_resend_nop_in(struct iscsi_cmnd *cmnd,
				struct iscsi_conn *conn,
				struct iscsi_session *session)
{
	int retval = 0;
	struct iscsi_targ_nopin *hdr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* restore the command's state to what it was prior to this retransmit
	 * (the prior state was probably ISCSI_SENT or ISCSI_NOPIN_SENT)
	 */
	cmnd->state = cmnd->reply_state;

	/* RFC 3720 Section 10.16 SNACK Request
	 * "The numbered-response(s) or R2T(s), requested by a SNACK, MUST be
	 * delivered as exact replicas of the ones that the target transmitted
	 * originally except for the fields ExpCmdSN, MaxCmdSN, and ExpDataSN,
	 * which MUST carry the current values."
	 */
	hdr = (struct iscsi_targ_nopin *) cmnd->reply_hdr;
	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	if (send_hdr_plus_1_data(cmnd, conn, hdr, cmnd->ping_data,
				 cmnd->data_length, USE_SEND_CONTROL) < 0)
		goto out1;

	TRACE(TRACE_ISCSI, "%lu %s NopIn resent, ITT %u\n",
	      jiffies, current->comm, cmnd->init_task_tag);
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_nopin(hdr);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "handle_resend_nop_in\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}

      out:
	return retval;

      out1:
	cmnd->state = ISCSI_DEQUEUE;
	retval = -1;
	goto out;
}

/*
 * executed only by the tx thread when state is
 * ISCSI_DONE or ISCSI_GOT_DATA_ACK or ISCSI_REDO_DATA
 * handle_iscsi_done: This is responsible for building the data (if any)
 * and response headers and transmitting them to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is done
 * executing
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_iscsi_done(struct iscsi_cmnd *cmnd,
			     struct iscsi_conn *conn,
			     struct iscsi_session *session)
{
	Scsi_Request *req;
	int err = 0;
	int phase_collapse = 0;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE,
	      "Enter handle_iscsi_done, cmnd %p, state %d\n", cmnd,
	      cmnd->state);

	if (cmnd->cmnd == NULL) {
		TRACE_ERROR("%s cmnd is NULL for cmnd %p\n", current->comm,
			    cmnd);
		err = -1;
		goto out;
	}

	req = cmnd->cmnd->req;

	if (req == NULL) {
		TRACE_ERROR("%s req is NULL for cmnd %p\n", current->comm,
			    cmnd);
		err = -1;
		goto out;
	}

	if ((req->sr_data_direction == SCSI_DATA_READ)
	    && host_byte(req->sr_result) == DID_OK) {
		err = send_read_data(cmnd, conn, session, &phase_collapse);
		if (err < 0 || cmnd->state == ISCSI_AWAIT_DATA_ACK)
			goto out;
	}

	if (cmnd->need_max_cmd_sn_inc) {
		/* the scsi command pdu was not immediate, CmdSN was advanced */
		spin_lock_irqsave(&session->cmnd_lock, flags);
		session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
		cmnd->need_max_cmd_sn_inc = 0;
		spin_unlock_irqrestore(&session->cmnd_lock, flags);
	}

	err = 0;

#ifdef ISER
	/* For Read command in iSER mode, we want to wait till all data-in PDUs
	 * are done
	 */
	if (conn->data_in_complete
	    && conn->session->oper_param->RDMAExtensions) {
		TRACE(TRACE_ISER, "%s: waiting for data_in_complete\n",
		      __func__);
		if(down_interruptible(&conn->data_in_sem))
			goto out;
	}
	conn->data_in_complete = 0;
#endif

	/* Separated Response for Error Recovery - SAI */
	if (cmnd->retransmit_flg) {
		cmnd->retransmit_flg = 0;
		cmnd->state = ISCSI_SENT;
	} else {
		if (!phase_collapse)
			/* Added for iSER - Minti
			 * This is just the response - send_control will do
			 */
			err = send_iscsi_response(cmnd, conn, session);
		else
			cmnd->state = ISCSI_SENT;
	}

	/* in case there are any out-of-order commands that are now in-order */
	check_queued_cmnd(session);

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave handle_iscsi_done, err = %d\n",
	      err);
	return err;
}

/*
 * executed only by the tx thread.
 * when state is ISCSI_MGT_FN_DONE:
 * handle_iscsi_mgt_fn_done: This is responsible for building the task mgt
 * response header and transmitting it to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is finished.
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd,
				    struct iscsi_conn *conn,
				    struct iscsi_session *session)
{
	struct iscsi_targ_task_mgt_response rsp;
	struct iscsi_cmnd *aborted_command;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* most fields in task management function response pdu go back as 0 */
	memset(&rsp, 0x0, ISCSI_HDR_LEN);

	rsp.opcode = ISCSI_TARG_TASK_MGMT_RSP;
	rsp.flags |= F_BIT;
	rsp.response = cmnd->response;
	rsp.init_task_tag = htonl(cmnd->init_task_tag);
	rsp.stat_sn = htonl(conn->stat_sn++);

	cmnd->state = ISCSI_DEQUEUE;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	if ((aborted_command = cmnd->ref_command)) {
		if (aborted_command->init_task_tag != cmnd->ref_task_tag
		    || aborted_command->cmd_sn != cmnd->ref_cmd_sn
		    || aborted_command->active_seqno !=
		    cmnd->ref_active_seqno) {
			TRACE_ERROR
			    ("%s aborted command ITT %u expected %u, "
			     "CmdSN %u expected %u, "
			     "active_seqno %u expected %u\n",
			     current->comm, aborted_command->init_task_tag,
			     cmnd->ref_task_tag, aborted_command->cmd_sn,
			     cmnd->ref_cmd_sn,
			     aborted_command->active_seqno,
			     cmnd->ref_active_seqno);
		} else {
			/* clear the aborted command struct entry in the dequeue routine */
			aborted_command->state = ISCSI_DEQUEUE;
			session->exp_cmd_sn +=
			    aborted_command->cmd_sn_increment;
			aborted_command->cmd_sn_increment = 0;
			session->max_cmd_sn +=
			    aborted_command->need_max_cmd_sn_inc;
			aborted_command->need_max_cmd_sn_inc = 0;
		}
	}

	/* if the TM command pdu was not immediate, CmdSN has advanced */
	session->exp_cmd_sn += cmnd->cmd_sn_increment;
	cmnd->cmd_sn_increment = 0;
	session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
	cmnd->need_max_cmd_sn_inc = 0;

	rsp.exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	rsp.max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	if (send_hdr_only(cmnd, conn, &rsp) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%lu %s task mgt response sent\n", jiffies,
	      current->comm);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_task_mgt_response(&rsp);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "handle_iscsi_mgt_fn_done\n", jiffies,
		      current->comm);
		up(&conn->tx_sem);
	}
	return 0;
}

#if defined(RDMA_MODE)
#ifndef MIN
#define MIN(x, y) (x > y? y : x)
#endif
#ifndef MAX
#define MAX(x, y) (x > y? x : y)
#endif
/**
 * Map a scatterlist to an R2T cookie.  The problem is that a single command's
 * data transfer may be broken up over multiple R2T's.  We need to build the
 * appropriate iovecs from the scatterlist using the R2T info in a given
 * cookie.
 */
void sg_map_to_cookie(struct scatterlist *sg_list, unsigned int sg_len,
                      struct iscsi_cookie *cookie)
{
	struct scatterlist *sg;
	struct iovec *iov = &cookie->vec[0];
	void *virt;
	unsigned int i, niov = 0;
	unsigned long bytes = 0, len;

	cookie->vec_len = 0;

	TRACE(TRACE_DEBUG, "need %u bytes\n", cookie->xfer_len);

	for_each_sg(sg_list, sg, sg_len, i) {
		len = scatterlist_length(sg);
		virt = sg_virt(sg);

		TRACE(TRACE_DEBUG,"sg_list[%d] base=%p, len=%lu\n",
		       i, virt, len);

		if (bytes >= cookie->offset + cookie->xfer_len) {
			TRACE(TRACE_DEBUG,"dropping out early with %lu bytes\n",
			       bytes);
			break;
		}

		if (bytes + len > cookie->offset) {

			if (niov == 0) {
			        TRACE(TRACE_DEBUG, "first iov at %lu bytes (%lu len)\n",
				       bytes, len);
			}

			if (bytes < cookie->offset) {
				iov->iov_base = virt + (cookie->offset - bytes);
				iov->iov_len = bytes + len - cookie->offset;
			} else {
				iov->iov_base = virt;
				iov->iov_len = MIN(cookie->offset
				                   + cookie->xfer_len
				                   - bytes, len);
			}

			TRACE(TRACE_DEBUG, "iov[%d] base=%p, len=%lu\n",
			      niov, iov->iov_base,
			      (unsigned long) iov->iov_len);

			iov += 1;
			niov += 1;
		}

		bytes += len;
	}

	cookie->vec_len = niov;
}
#endif /* RDMA_MODE */

/*
 * executed only by the tx thread when state is ISCSI_BUFFER_RDY.
 * iscsi_tx_r2t: this function transmits a ready-to-transfer to the
 * Initiator so that it can transmit the data necessary for a WRITE
 * INPUT: command for which R2T needs to be transferred
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 *
 * during a WRITE, the following fields in struct iscsi_cmnd maintain state:
 *
 *	data_length		The total number of bytes to be received in this WRITE
 *					(includes bytes in immediate and unsolicited pdus)
 *					Set from EDTL in iSCSI command and never changed
 *
 *	data_done		Number of bytes correctly received so far in this WRITE
 *					(includes bytes in immediate and unsolicited pdus)
 *					When sequences and PDUs are being sent in order,
 *					this value will be the expected offset of the next
 *					DataOut PDU we should receive from the initiator.
 *					WRITE should be finished when this equals data_length
 *
 *	first_burst_len	Number of unsolicited bytes received so far in this WRITE
 *					(counts only bytes in immediate and unsolicited pdus)
 *
 *	first_solicited_offset		Offset to first byte of any solicited WRITE data
 *
 *	unsolicited_data_present	1 until all unsolicited pdus have been read
 *
 *	next_burst_len	Number of solicited bytes received so far in this sequence
 *					reset to 0 on receiving last DataOut PDU in a sequence
 *					which is used to detect first DataOut PDU in next sequence
 *
 *	r2t_data_total	The total number of bytes to be solicited by R2Ts
 *					Set during iSCSI command processing and never changed
 *					Used only to initialize r2t_data_left on midlevel call back
 *					(can be negative)
 *
 *	r2t_data_left	The remaining number of bytes left to be solicited by R2Ts
 *					set to 0 until midlevel calls back iscsi_rdy_to_xfer()
 *					to indicate target's receive buffers have been set up
 *					updated as R2Ts are sent
 *					(can be negative)
 *
 *	retransmit_flg	1 when retransmitting a previously sent R2T in response to
 *					a timeout on target or an explicit SNACK from initiator
 *					Not set on a timeout when recovery_r2t is also non-zero.
 *
 *	recovery_r2t	non-zero when recovering from target-detected error in a
 *					DataOut received from initiator
 *					1	waiting to get DataOut with F=1 in current burst
 *					2	got DataOut with F=1 or got timeout, need to send
 *						the recovery R2T to initiator
 *					0	after recovery R2T has been sent
 *
 *	startsn			used only when retransmit_flg or recovery_r2t are non-zero
 *					holds R2TSN of next cookie to be used to send next R2T
 *
 *	endsn			used only when retransmit_flg or recovery_r2t are non-zero
 *					holds R2TSN of last cookie to be used to send R2Ts
 *
 *	timestamp		set to jiffies upon receipt of iSCSI command and DataOut
 *					pdus, also on sending of an R2T.  Used to check for
 *					initiator inactivity which provokes a timeout R2T retransmit
 *
 *	outstanding_r2t	keeps track of number of R2Ts sent to initiator but for
 *					which the end of the corresponding solicited sequence has
 *					not been detected (i.e., by receiving a DataOut pdu with
 *					the F bit set).  Initialized to 0, incremented for every
 *					new or recovery (i.e., not retransmitted) R2T sent,
 *					decremented for every DataOut pdu received with F bit set.
 *
 *	r2t_sn			counts number of new and recovery (i.e., not retransmitted)
 *					R2Ts sent so far to initiator for this command.  This value
 *					becomes the R2TSN in the next new or recovery R2T sent.
 *					Initialized to 0, incremented for every new or recovery R2T
 *					sent.
 *
 *	data_sn			number expected in DataSN field of next DataOut pdu received
 *					in current sequence.  Initialized to 0 and reset to 0 at the
 *					end of every sequence (i.e., upon receiving a DataOut with
 *					F bit set).  Incremented by one for every DataOut pdu
 *					received.
 */
static int iscsi_tx_r2t(struct iscsi_cmnd *cmnd,
			struct iscsi_conn *conn,
			struct iscsi_session *session,
			unsigned long flags, int *cmnd_lock_held)
{
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	int err = 0;
	struct iscsi_targ_r2t *hdr;
	int data_length_left;
	int max_burst_len = 0;
	struct iscsi_cookie *cookie = NULL;

#if defined(RDMA_MODE)
	struct iser_data_buf_t data_sink;
	unsigned long xfer_len;
#endif				/* RDMA_MODE */

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_tx_r2t, r2t_data_left %d, "
	      "retransmit_flg %u, outstanding_r2t %d, recovery_r2t %u\n",
	      cmnd->r2t_data_left, cmnd->retransmit_flg,
	      cmnd->outstanding_r2t, cmnd->recovery_r2t);

	if (!cmnd->unsolicited_data_present
	    && !list_empty(&cmnd->unsolicited_data_list)) {
		if ((err = send_unsolicited_data(cmnd, conn, session)) < 0
		    || cmnd->state != ISCSI_BUFFER_RDY) {
			goto r2t_out;
		}
	}

	if ((data_length_left = cmnd->r2t_data_left) <= 0
	    && !cmnd->retransmit_flg && cmnd->recovery_r2t != 2) {
		/* no need to send any (more) R2Ts (yet) */
		goto r2t_out;
	}

	/* try to send all the R2T's at once, but limited by MaxOutstandingR2T */
	do {
		if (session->oper_param->MaxOutstandingR2T <=
		    cmnd->outstanding_r2t && !cmnd->retransmit_flg
		    && cmnd->recovery_r2t != 2) {
			/* we have hit the limit on number of outstanding R2Ts allowed */
			goto r2t_out;
		}

		memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
		hdr = (struct iscsi_targ_r2t *) iscsi_hdr;

		hdr->opcode = ISCSI_TARG_R2T;
		hdr->flags |= F_BIT;
		hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);
		hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
		hdr->stat_sn = cpu_to_be32(conn->stat_sn);
		hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
		hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
		if ((cmnd->retransmit_flg || cmnd->recovery_r2t == 2)) {
			/* sending a recovery R2T or retransmitting R2Ts in response to
			 * an R2T SNACK from initiator
			 */
			for (cookie = cmnd->first_r2t_cookie; cookie;
			     cookie = cookie->next) {
				if (cookie->seq == cmnd->startsn) {
					/* this cookie lets us regenerate the missing r2t */
					if (cmnd->retransmit_flg) {
						TRACE_WARNING
						    ("%s Retransmit R2T, ITT %u R2TSN %u "
						     "Buffer Offset %u\n",
						     current->comm,
						     cmnd->init_task_tag,
						     cookie->seq,
						     cookie->offset);
					}
					hdr->r2t_sn =
					    cpu_to_be32(cookie->seq);
					hdr->offset =
					    cpu_to_be32(cookie->offset);
					hdr->xfer_len =
					    cpu_to_be32(cookie->xfer_len);
					if (cmnd->recovery_r2t == 2) {
						cmnd->r2t_data_left =
						    data_length_left =
						    cmnd->data_length -
						    (cookie->offset +
						     cookie->xfer_len);
					} else if (++cmnd->startsn <=
						   cmnd->endsn
						   && cookie->next) {
						/* more R2Ts to retransmit */
						data_length_left =
						    cookie->next->xfer_len;
					} else {
						/* last R2T to retransmit, go back to normal r2ts */
						data_length_left =
						    cmnd->r2t_data_left;
						cmnd->retransmit_flg = 0;
					}
					break;
				}
			}
			if (cmnd->recovery_r2t == 2) {
				/* done with recovery if get here */
				cmnd->recovery_r2t = 0;
			}
			if (!cookie) {
				/* no cookies to retransmit, go back to normal r2t processing */
				data_length_left = cmnd->r2t_data_left;
				cmnd->retransmit_flg = 0;
				continue;
			}
		} else {

			hdr->r2t_sn = cpu_to_be32(cmnd->r2t_sn++);
			hdr->offset =
			    cpu_to_be32(cmnd->data_length -
					data_length_left);

			max_burst_len =
			    session->oper_param->MaxBurstLength;

			/* form an R2T cookie to be used for re-transmissions - SAI
			 *
			 * This cookie is also used by iSER upon data completion to get
			 * the lenght of the R2T's requested data. -- EAB
			 */
			cookie = create_r2t_cookie(cmnd);

			if (cookie) {
				cookie->seq = cmnd->r2t_sn - 1;
				cookie->offset =
				    cmnd->data_length - data_length_left;
				cookie->xfer_len =
				    (data_length_left <= max_burst_len)
				    ? data_length_left : max_burst_len;
			}

			if (data_length_left <= max_burst_len) {
				/* this is the last R2T we need to send for this command */
				hdr->xfer_len =
				    cpu_to_be32(data_length_left);
				data_length_left = 0;
			} else {
				hdr->xfer_len = cpu_to_be32(max_burst_len);
				data_length_left -= max_burst_len;
			}
			cmnd->r2t_data_left = data_length_left;

			/* if DataSequenceInOrder is no, send r2t offset in reverse order */
			if (!session->oper_param->DataSequenceInOrder)
				hdr->offset = cpu_to_be32(data_length_left
							  +
							  cmnd->
							  first_solicited_offset);

			/* Increment the outstanding R2T for every R2T sent - SAI */
			cmnd->outstanding_r2t++;
		}

		if (*cmnd_lock_held) {
			*cmnd_lock_held = 0;
			spin_unlock_irqrestore(&session->cmnd_lock, flags);
		}
#ifdef K26
		if (irqs_disabled()) {
			TRACE_WARNING
			    ("%lu %s iscsi_tx_r2t irqs disabled on call to "
			     "send_hdr_only, cmnd lock held %d\n", jiffies,
			     current->comm, *cmnd_lock_held);
		}
#endif

#if defined(RDMA_MODE)
		sg_map_to_cookie(cmnd->cmnd->req->sr_buffer,
 		                 cmnd->cmnd->req->sr_use_sg,
		                 cookie);
		xfer_len = be32_to_cpu(hdr->xfer_len);
		data_sink.p_buf = &cookie->vec;
		data_sink.size = cookie->vec_len;
		data_sink.type = ISER_BUF_TYPE_IOVEC_VIRT;
		TRACE(TRACE_ISER, "R2T for buff=%p, len=%lu\n", &data_sink,
		      xfer_len);

		/*
		 * EAB - we are not really *sending* the `1 data', we are passing a
		 *       databuffer to iSER for an RDMA data sink.  We pass it the
		 *       SCSI buffer from the command as the sink.
		 */
		if (send_hdr_plus_1_data
		    (cmnd, conn, iscsi_hdr, &data_sink, xfer_len,
		     USE_SEND_CONTROL) < 0) {
#else
		if (send_hdr_only(cmnd, conn, iscsi_hdr) < 0) {
#endif				/* RDMA_MODE */
			/* Added for iSER - Minti
			 * Send R2Ts (iSCSI Data type PDUs) with get_data in iSER mode -TODO
			 */
			err = -1;
			goto r2t_out;
		}

		TRACE(TRACE_ISCSI,
		      "%lu %s Sent R2T ITT %u DDTL %u offset %d\n",
		      jiffies, current->comm, cmnd->init_task_tag,
		      be32_to_cpu(hdr->xfer_len),
		      be32_to_cpu(hdr->offset));

		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_targ_r2t(hdr);

		/* store the r2t time stamp - SAI */
		/* error recovery ver ref18_04 */
		cmnd->timestamp = jiffies;
	} while (data_length_left > 0);

      r2t_out:

	cmnd_check_magic(session, cmnd);

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_tx_r2t, r2t_data_left %d, "
	      "retransmit_flg %u, outstanding_r2t %d, recovery_r2t %u, err %d\n",
	      cmnd->r2t_data_left, cmnd->retransmit_flg,
	      cmnd->outstanding_r2t, cmnd->recovery_r2t, err);
	return err;
}

/*
 * Called only by iscsi_tx_thread
 */
static int
need_a_nop(struct iscsi_conn *conn, struct iscsi_session *session)
{
	__u32 count;
	int delta;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	count = atomic_read(&conn->outstanding_nopins);
	delta = jiffies - conn->time_last_nopin_sent;
	if (!count || delta >= PING_REPLY_TIME_LIMIT * count) {
		if (count >= MAX_OUTSTANDING_NOPINS) {
			TRACE_ERROR
			    ("%lu %s Hit limit of %u outstanding NopIns delta %d\n",
			     jiffies, current->comm, count, delta);
			return -1;
		}
		if (count > MAX_OUTSTANDING_NOPINS / 2) {
			TRACE_WARNING
			    ("%lu %s No reply to %u outstanding NopIns delta %d\n",
			     jiffies, current->comm, count, delta);
		}
		if (generate_nopin(conn, session) < 0)
			return -1;
		atomic_inc(&conn->outstanding_nopins);
		/* save when this NopIn was sent so we don't ping initiator too often */
		conn->time_last_nopin_sent = jiffies;
	}
	return 0;
}

/*
 * executed only by the tx thread when state is ISCSI_QUEUE_CMND_RDY or
 * ISCSI_BUFFER_RDY and all unsolicited data has been read.
 * copy the data in udata_list in iSCSI layer to st_list in SCSI middle level.
 * if all the data has been read at this time,
 * then change state to ISCSI_DATA_IN and call scsi_rx_data()
 * else change state to ISCSI_BUFFER_RDY
 * this thread is called with session->cmnd_lock held by caller (tx thread)
 * on successful return, the command's unsolicited_data_list is empty
 */
static int send_unsolicited_data(struct iscsi_cmnd *cmd,
				 struct iscsi_conn *conn,
				 struct iscsi_session *session)
{
	struct list_head *list_ptr, *list_temp;
	struct scatterlist *st_list;
	struct udata_list *udata;
	int err;
	__u32 expected_data_offset, offset, length, sglen, i, n;
	__u8 *buffer, *sgbuf;
	DECLARE_LOOP_COUNTER(count);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* all unsolicited data has been read */
	TRACE(TRACE_DEBUG, "%s send unsolicited CmdSN %u ITT %u\n",
	      current->comm, cmd->cmd_sn, cmd->init_task_tag);

	if (!cmd->cmnd) {
		TRACE_ERROR("%s no cmnd found\n", current->comm);
		return -1;
	}

	st_list = (struct scatterlist *) cmd->cmnd->req->sr_buffer;

	/* copy each buffer in udata list to appropriate place in sg list */
	expected_data_offset = offset = i = 0;
	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp,
			   &cmd->unsolicited_data_list) {
		udata =
		    list_entry(list_ptr, struct udata_list, udata_link);
		BUMP_LOOP_COUNTER(count, "unsolicited_data_list");
		/* if items in udata list are in correct order by offset,
		 * then "offset" and "i" are already positioned correctly in the sg list
		 */
		if (udata->offset != expected_data_offset) {
			offset =
			    skip_thru_sg_list(st_list, &i, udata->offset);
		}

		/* copy all "length" bytes of data from udata list "buffer" to sg list,
		 * starting at "offset" bytes in ith sg list item */
		buffer = udata->buffer;
		length = udata->length;
		expected_data_offset = udata->offset + length;
		while (length > 0) {
			sglen = get_sglen(st_list + i);
			sgbuf = get_sgbuf(st_list + i);
			n = sglen - offset;	/* no of bytes left in sg list item's buf */
			if (n > length)
				n = length;	/* only this many data bytes left to copy */
			memcpy(sgbuf + offset, buffer, n);
			length -= n;
			buffer += n;
			offset += n;
			if (offset >= sglen) {
				/* copied all bytes in this sg list item's buf, move to next */
				offset = 0;
				i++;
			}
		}

		/* "i" and "offset" positioned to fill next sg list slot in order */
		cmd->scatter_list_count = i;
		cmd->scatter_list_offset = offset;
	}

	/* finished with udata list */
	free_udata_list(cmd);

	if (cmd->data_done >= cmd->data_length) {
		/* already read all the data, which was all unsolicited */
		cmd->state = ISCSI_DATA_IN;
		TRACE(TRACE_ISCSI,
		      "%lu %s All data read CmdSN %u ITT %u\n", jiffies,
		      current->comm, cmd->cmd_sn, cmd->init_task_tag);
		if ((err = scsi_rx_data(cmd->cmnd)) < 0) {
			TRACE_ERROR("%s scsi_rx_data returned an error\n",
				    current->comm);
		}
	} else {
		/* more data left to read, which is all solicited */
		cmd->state = ISCSI_BUFFER_RDY;
		err = 0;
	}

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "send_unsolicited_data\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}

	return err;
}

/*
 * executed only by the tx thread
 * called directly when state is ISCSI_RESEND_STATUS
 * and called indirectly via handle_iscsi_done() when state is ISCSI_DONE
 * to send ISCSI Response PDU, with status information if needed.
 * Separated Data-In from Status Response PDU sending for Status
 * Error Recovery - SAI
 */
static int send_iscsi_response(struct iscsi_cmnd *cmnd,
			       struct iscsi_conn *conn,
			       struct iscsi_session *session)
{
	struct iscsi_targ_scsi_rsp *rsp;
	Scsi_Request *req;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	struct sense_data_buf {
		__u16 len;	/* added by iscsi standard */
		__u8 data[SCSI_SENSE_BUFFERSIZE];	/* direct from scsi (64 bytes) */
	} sense_data = {
	0};
	int residual_count = 0;
	__u32 length = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "send_scsi_response\n");

	memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	rsp = (struct iscsi_targ_scsi_rsp *) iscsi_hdr;
	rsp->opcode = ISCSI_TARG_SCSI_RSP;
	rsp->flags |= F_BIT;
	rsp->init_task_tag = cpu_to_be32(cmnd->init_task_tag);

	req = cmnd->cmnd->req;
	if ((req->sr_data_direction == SCSI_DATA_READ)
	    && host_byte(req->sr_result) == DID_OK) {
		residual_count = do_command_status(cmnd, req);
		rsp->exp_data_sn = cpu_to_be32(cmnd->data_sn);
	}

	/* cdeng 3/24/02 */
	if (cmnd->seq_flags & OVERFLOW_FLAG) {
		rsp->flags |= O_BIT;
		rsp->resid = cpu_to_be32(residual_count);
	} else if (cmnd->seq_flags & UNDERFLOW_FLAG) {
		rsp->flags |= U_BIT;
		rsp->resid = cpu_to_be32(residual_count);
	}

	/* RDR - do NOT assign and increment stat_sn on a retransmission */
	if (!cmnd->retransmit_flg) {
		cmnd->stat_sn = conn->stat_sn++;
	}
	rsp->stat_sn = cpu_to_be32(cmnd->stat_sn);
	rsp->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	rsp->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	if (cmnd->seq_flags & SEND_SENSE_FLAG) {
		/* sense data has to be sent as part of SCSI Response pdu */
		rsp->status = CHECK_CONDITION << 1;	/* why does Linux do this shift? */
		sense_data.len = cpu_to_be16(SCSI_SENSE_BUFFERSIZE);
		/* 2 more than actual sense data to include extra iscsi length header */
		length = SCSI_SENSE_BUFFERSIZE + 2;
		rsp->length = cpu_to_be32(length);
		if (cmnd->seq_flags & UNDERFLOW_FLAG) {
			TRACE(TRACE_DEBUG, "underflow is found\n");
			memset(sense_data.data, 0x0,
			       SCSI_SENSE_BUFFERSIZE);
			sense_data.data[0] = 0xf0;	/* scsi valid bit, code 70h */
			sense_data.data[2] = 0x20;	/* scsi ILI bit */
			sense_data.data[7] = 0x07;	/* scsi additional length = 7 */
			memcpy(sense_data.data + 3,
			       req->sr_sense_buffer + 3, 4);
		} else {
			TRACE(TRACE_DEBUG, "sense key 0x%x\n",
			      req->sr_sense_buffer[2] & 0xf);
			memcpy(sense_data.data, req->sr_sense_buffer,
			       SCSI_SENSE_BUFFERSIZE);
		}
	}

	TRACE(TRACE_DEBUG,
	      "send_iscsi_response: sending status for cmnd_rn %.8x "
	      "init_task_tag %.8x target_xfer_tag %.8x\n", cmnd->cmd_sn,
	      cmnd->init_task_tag, cmnd->target_xfer_tag);

	cmnd->retransmit_flg = 0;
	cmnd->state = ISCSI_SENT;

#if defined(RDMA_MODE)
	/* If RDMAExtensions=Yes, always send ExpDataSN of zero, yet
	 * we still need to maintain this state information. */
	if (cmnd->session->oper_param->RDMAExtensions)
		rsp->exp_data_sn = 0;
#endif /* RDMA_MODE */

	if (send_hdr_plus_1_data
	    (cmnd, conn, iscsi_hdr, &sense_data, length,
	     USE_SEND_CONTROL) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI,
	      "%lu %s Sent SCSI reply CmdSN %u ITT %u StatSN %u "
	      "ExpCmdSN %u MaxCmdSN %u\n", jiffies, current->comm,
	      cmnd->cmd_sn, cmnd->init_task_tag, cmnd->stat_sn,
	      ntohl(rsp->exp_cmd_sn), ntohl(rsp->max_cmd_sn));

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_scsi_rsp(rsp);

	if (atomic_read(&conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "send_iscsi_response\n", jiffies, current->comm);
		up(&conn->tx_sem);
	}
	return 0;
}

/*
 * executed only by the tx thread.
 * called only from need_a_nop()
 * to generate and send a new NopIn "ping" PDU to initiator
 * (that is not a reply to a NopOut sent by initiator)
 */
static int __attribute__ ((no_instrument_function))
    generate_nopin(struct iscsi_conn *conn, struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;
	unsigned long flags;

	/* NULL 3rd parameter generates a new TTT for new command */
	if (!(cmnd = get_free_command(NULL, session, NULL))) {
		return -1;
	}

	/* set up new command to look like an immediate NopOut sent by initiator
	 * with no reply expected by initiator
	 */
	cmnd->state = ISCSI_NOPIN_SENT;
	cmnd->conn = conn;
	cmnd->opcode_byte = ISCSI_INIT_NOP_OUT | I_BIT;
	cmnd->init_task_tag = ALL_ONES;

	spin_lock_irqsave(&session->cmnd_lock, flags);

	/* Add this command to end of the session's queue of active commands */
	list_add_tail(&cmnd->command_link, &session->active_command_list);

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	return handle_nopin(cmnd, conn, session);
}

/*
 * auxilliary function called only by handle_iscsi_done() on a READ-type command
 *
 * executed only by tx thread when state is
 * ISCSI_DONE or ISCSI_AWAIT_DATA_ACK or ISCSI_REDO_DATA.
 * at entry the command should be at the beginning of a sequence (the beginning
 * of the first sequence if this is the first call for this command).
 * sends all the DataIn pdus for this sequence to the initiator
 * and, if not setting the A bit at the end of non-final sequences,
 * sends all the sequences too.
 *
 * During this sending:
 *	data_length_left	number of bytes left to send in following sequences.
 *	seq_length_left		number of bytes left to send in this sequence.
 *	seq_low_byte		offset to lowest byte in this sequence.
 *	seq_limit_byte		offset to highest byte plus one in this sequence.
 *	pdu_offset			offset to first byte in this pdu.
 *	data_payload_length	number of data bytes to send in this pdu.
 *	total_data_length	total number of all bytes to send in this pdu
 */
static int send_read_data(struct iscsi_cmnd *cmnd,
			  struct iscsi_conn *conn,
			  struct iscsi_session *session,
			  int *phase_collapse)
{
	int err = 0;
	struct iscsi_targ_scsi_data_in *hdr;
	struct scatterlist *st_list;
	__u8 iscsi_hdr[ISCSI_HDR_LEN];
	int seq_length_left;
	int pdu_offset, pdu_data_sn;
	int data_payload_length;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_DEBUG, "sending sense data for cmnd_sn "
	      "%u, ITT %u, TTT %u\n", cmnd->cmd_sn, cmnd->init_task_tag,
	      cmnd->target_xfer_tag);

	if (cmnd->state == ISCSI_DONE) {
		/* first call for this command, set things up */
		do_command_status(cmnd, cmnd->cmnd->req);

		/* one of these changes before use, depending on DataSequenceInOrder */
		cmnd->seq_low_byte = 0;
		cmnd->seq_limit_byte = cmnd->data_length_left;

		/* always save the starting values for the whole READ command */
		save_data_in_seq(cmnd);
	}

	/* once around this loop for each sequence of DataIns sent to initiator */
	while (cmnd->data_length_left > 0) {

		/* save values at start of this sequence in case we need to restart it */
		if (session->targ_snack_flg & DATACK1_SNACK_ENABLE) {
			save_data_in_seq(cmnd);
		}

		/* find out how long next sequence will be */
		if (cmnd->data_length_left <=
		    session->oper_param->MaxBurstLength) {
			/* all remaining data fits in this last sequence */
			seq_length_left = cmnd->data_length_left;
			cmnd->seq_flags |= LAST_SEQ_FLAG;
		} else {
			/* another sequence will follow this one */
			seq_length_left =
			    session->oper_param->MaxBurstLength;
		}
		cmnd->seq_length = seq_length_left;

		if (session->oper_param->DataSequenceInOrder)
			/* sequences are being sent in order */
			cmnd->seq_limit_byte =
			    cmnd->seq_low_byte + seq_length_left;
		else
			/* sequences are being sent in reverse order */
			cmnd->seq_low_byte =
			    cmnd->seq_limit_byte - seq_length_left;

		TRACE(TRACE_ERROR_RECOVERY,
		      "%s new DataIn sequence, data_length_left "
		      "%u, seq_length %u, low_byte %u, limit_byte %u\n",
		      current->comm, cmnd->data_length_left,
		      cmnd->seq_length, cmnd->seq_low_byte,
		      cmnd->seq_limit_byte);

		if (session->oper_param->DataPDUInOrder) {
			/* pdus within each sequence are being sent in order */
			pdu_offset = cmnd->seq_low_byte;
		} else {
			/* pdus within each sequence are being sent in reverse order */
			pdu_offset = cmnd->seq_limit_byte;
		}

		/* once around this loop for each pdu in this sequence */
		while (seq_length_left > 0) {

			memset(iscsi_hdr, 0x0, ISCSI_HDR_LEN);
			hdr = (struct iscsi_targ_scsi_data_in *) iscsi_hdr;

			hdr->opcode = ISCSI_TARG_SCSI_DATA_IN;
			hdr->target_xfer_tag = ALL_ONES;

			if (seq_length_left > conn->max_send_length) {
				data_payload_length =
				    conn->max_send_length;
			} else {
				/* this is the last DataIn pdu in this sequence */
				data_payload_length = seq_length_left;
				hdr->flags |= F_BIT;

				if ((cmnd->seq_flags & LAST_SEQ_FLAG)
				    && !(cmnd->
					 seq_flags & SEND_SENSE_FLAG)) {
					/* this is also the last DataIn pdu for this command and
					 * we are not sending any sense data, try phase collapse.
					 * 0 never collapse, 1 always collapse,
					 * 2 collapse this time, -1 do not collapse this time
					 */
#ifdef RDMA_MODE
					conn->data_in_complete = 1;
#endif
					TRACE(TRACE_ISER,
					      "Final pdu in sequence--\n");

					if (conn->phase_collapse > 0) {
						/* do phase collapse, either always or just this time */
						hdr->flags |= S_BIT;
						*phase_collapse = 1;
						if (conn->phase_collapse >=
						    2)
							/* was just this time, don't do it next time */
							conn->
							    phase_collapse
							    = -1;
					} else if (conn->phase_collapse <
						   0) {
						/* didn't do it just this time, do it next time */
						conn->phase_collapse = 2;
					}
					/* else never do it */
					if (session->
					    targ_snack_flg &
					    DATACK2_SNACK_ENABLE) {
						/* set A bit on last DataIn pdu in last sequence */
						set_a_bit(cmnd, hdr);
					}
				} else if (session->
					   targ_snack_flg &
					   DATACK1_SNACK_ENABLE) {
					/* set A bit on each DataIn pdu that ends a sequence that
					 * is not also the last sequence for the whole command.
					 * target will wait for SNACK reply to this before sending
					 * more DataIn pdus.
					 */
					set_a_bit(cmnd, hdr);
					cmnd->state = ISCSI_AWAIT_DATA_ACK;
				}
			}

			/* when pdus are not in order within seq, send them in reverse order
			 * so bump this offset down from previous offset (if any)
			 */
			if (!session->oper_param->DataPDUInOrder)
				pdu_offset -= data_payload_length;

			TRACE(TRACE_ERROR_RECOVERY,
			      "%s new DataIn pdu, seq_length_left %d,"
			      " DSL %u, DataSN %u, offset %u\n",
			      current->comm, seq_length_left,
			      data_payload_length, cmnd->data_sn,
			      pdu_offset);

			hdr->length = cpu_to_be32(data_payload_length);
			hdr->init_task_tag =
			    cpu_to_be32(cmnd->init_task_tag);

			if (hdr->flags & S_BIT) {
				/* RDR - already assigned and incremented stat_sn before
				 * a retransmission
				 */
				if (!cmnd->retransmit_flg) {
					cmnd->stat_sn = conn->stat_sn++;
				}
				hdr->stat_sn = cpu_to_be32(cmnd->stat_sn);
				if (cmnd->need_max_cmd_sn_inc) {
					/* scsi command pdu was not immediate, CmdSN was advanced */
					spin_lock_irqsave(&session->
							  cmnd_lock,
							  flags);
					session->max_cmd_sn +=
					    cmnd->need_max_cmd_sn_inc;
					cmnd->need_max_cmd_sn_inc = 0;
					spin_unlock_irqrestore(&session->
							       cmnd_lock,
							       flags);
				}

			}

			hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
			hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);
			pdu_data_sn = cmnd->data_sn++;
			hdr->data_sn = cpu_to_be32(pdu_data_sn);
			hdr->offset = cpu_to_be32(pdu_offset);

			st_list =
			    (struct scatterlist *) cmnd->cmnd->req->
			    sr_buffer;

			/* if anything out of order, need to skip through command's
			 * scatter-gather list to find start of this pdu
			 */
			if (!session->oper_param->DataPDUInOrder
			    || !session->oper_param->DataSequenceInOrder) {
				cmnd->scatter_list_offset =
				    skip_thru_sg_list(st_list,
						      &cmnd->
						      scatter_list_count,
						      pdu_offset);
			}

			st_list += cmnd->scatter_list_count;

			TRACE(TRACE_DEBUG,
			      "offset into sg entry %d, offset into sg list"
			      " %d\n", cmnd->scatter_list_offset,
			      cmnd->scatter_list_count);

			/* skip data PDU re-transmissions that are not asked for - SAI */
			if (cmnd->retransmit_flg) {
				if (pdu_data_sn < cmnd->startsn
				    || pdu_data_sn > cmnd->endsn) {
					printk
					    ("skip rexmit of DataIn ITT %u, DataSN %u, "
					     "Buffer Offset %u\n",
					     cmnd->init_task_tag,
					     pdu_data_sn, pdu_offset);
					goto skip_transmission;
				}
				TRACE_WARNING
				    ("%s Retransmitting DataIn ITT %u, DataSN %u,"
				     " Buffer Offset %u\n", current->comm,
				     cmnd->init_task_tag, pdu_data_sn,
				     pdu_offset);
			}

			err =
			    fill_and_send_datain_iov(cmnd, conn,
						     data_payload_length,
						     st_list, iscsi_hdr);
			if (err <= 0)
				goto out;

			if (TRACE_TEST(TRACE_ISCSI_FULL)) {
				TRACE(TRACE_ISCSI_FULL,
				      "%lu %s DataIn sent, offset %u\n",
				      jiffies, current->comm, pdu_offset);
				print_targ_scsi_data_in(hdr);
			}

		      skip_transmission:	/* Label added for error recovery - SAI */

			/* end of this DataIn pdu, set up for next in same seq (if any) */

			/* when pdus are in order within seq, bump offset up for next pdu */
			if (session->oper_param->DataPDUInOrder)
				pdu_offset += data_payload_length;

			seq_length_left -= data_payload_length;

			TRACE(TRACE_DEBUG,
			      "data sent %d data left in seq %d sg_list_offset %d "
			      "sg_entry_offset %d\n", err, seq_length_left,
			      cmnd->scatter_list_count,
			      cmnd->scatter_list_offset);
		}

		/* end of this sequence of DataIn pdus */
		if (cmnd->state == ISCSI_AWAIT_DATA_ACK)
			/* don't send next sequence until initiator sends DataACK SNACK */
			break;

		/* set up for next seq (if any) */
		move_to_next_seq(cmnd, session);
	}

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave send_read_data, err = %d\n", err);
	return err;
}

/*
 * executed only by the tx thread.
 * sends a NopIn PDU to initiator
 * either in response to an earlier NopOut ping from initiator
 *	(state == ISCSI_PING)
 * or as a new NopIn ping generated by the target
 *	(state == ISCSI_NOPIN_SENT)
 */
static int handle_nopin(struct iscsi_cmnd *cmnd,
			struct iscsi_conn *conn,
			struct iscsi_session *session)
{
	struct iscsi_targ_nopin *hdr;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	hdr = (struct iscsi_targ_nopin *) cmnd->reply_hdr;
	memset(hdr, 0x0, ISCSI_HDR_LEN);

	hdr->opcode = ISCSI_TARG_NOP_IN;

	hdr->flags |= F_BIT;

	hdr->length = cpu_to_be32(cmnd->data_length);
	hdr->init_task_tag = cpu_to_be32(cmnd->init_task_tag);

	/* RFC 3720 Section 10.19.2 StatSN (for NOP-In)
	 * "The StatSN field will always contain the next StatSN. However,
	 * when the Initiator Task Tag is set to 0xffffffff, StatSN for the
	 * connection is not advanced after this PDU is sent."
	 */
	hdr->stat_sn = cpu_to_be32(conn->stat_sn);
	if (cmnd->init_task_tag != ALL_ONES)
		conn->stat_sn++;

	hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
	if (cmnd->target_xfer_tag == ALL_ONES) {
		/* target not expecting a reply from initiator */
		cmnd->state = ISCSI_SENT;
	}

	spin_lock_irqsave(&session->cmnd_lock, flags);

	/* if the NopOut command pdu was not immediate, CmdSN has advanced */
	session->exp_cmd_sn += cmnd->cmd_sn_increment;
	cmnd->cmd_sn_increment = 0;
	session->max_cmd_sn += cmnd->need_max_cmd_sn_inc;
	cmnd->need_max_cmd_sn_inc = 0;

	hdr->exp_cmd_sn = cpu_to_be32(session->exp_cmd_sn);
	hdr->max_cmd_sn = cpu_to_be32(session->max_cmd_sn);

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	if (send_hdr_plus_1_data(cmnd, conn, hdr, cmnd->ping_data,
				 cmnd->data_length,
				 USE_SEND_CONTROL) < 0) {
		return -1;
	}

	TRACE(TRACE_ISCSI, "%lu %s Sent NopIn ITT %d TTT %d ExpCmdSN %u "
	      "StatSN %u\n", jiffies, current->comm,
	      cmnd->init_task_tag,
	      cmnd->target_xfer_tag,
	      session->exp_cmd_sn, be32_to_cpu(hdr->stat_sn));

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_nopin(hdr);

	return 0;
}

/* save values at start of this DataIn sequence in case we need to restart it */
static void __attribute__ ((no_instrument_function))
    save_data_in_seq(struct iscsi_cmnd *cmnd)
{
	cmnd->start_data_sn = cmnd->data_sn;
	cmnd->start_data_length_left = cmnd->data_length_left;
	cmnd->start_low_byte = cmnd->seq_low_byte;
	cmnd->start_limit_byte = cmnd->seq_limit_byte;
	cmnd->start_sg_list_count = cmnd->scatter_list_count;
	cmnd->start_sg_list_offset = cmnd->scatter_list_offset;
}

static __u32 __attribute__ ((no_instrument_function))
    do_command_status(struct iscsi_cmnd *cmnd, Scsi_Request * req)
{
	int transfer = 0;
	__u32 residual_count = 0;

	cmnd->seq_flags = 0;

	cmnd->data_length_left = req->sr_bufflen;

	TRACE(TRACE_DEBUG,
	      "Sense: %02x %02x %02x %02x %02x %02x %02x %02x\n",
	      req->sr_sense_buffer[0], req->sr_sense_buffer[1],
	      req->sr_sense_buffer[2], req->sr_sense_buffer[3],
	      req->sr_sense_buffer[4], req->sr_sense_buffer[5],
	      req->sr_sense_buffer[6], req->sr_sense_buffer[7]);

	if ((req->sr_sense_buffer[0] & 0x7e) == 0x70) {
		/* current or deferred error -- else, we don't send the sense buffer */
		cmnd->seq_flags |= SEND_SENSE_FLAG;

		if (req->sr_sense_buffer[0] & 0x80) {
			/* bytes 3 thru 6 contain valid data */
			transfer = (req->sr_sense_buffer[3] << 24) |
			    (req->sr_sense_buffer[4] << 16) |
			    (req->sr_sense_buffer[5] << 8) | req->
			    sr_sense_buffer[6];
			TRACE(TRACE_DEBUG,
			      "information in sense data: %d\n", transfer);
		}

		if ((req->sr_sense_buffer[2] & 0x20) != 0) {	/* ILI bit set */
			cmnd->data_length_left -= transfer;
		}
	} else if (req->sr_command) {
		/* ensure correct DataSegmentLength and ResidualCount */
		TRACE(TRACE_DEBUG,
		      "data_length_left %u, sr_command->resid %d\n",
		      cmnd->data_length_left, scsi_get_resid(req->sr_command));
		cmnd->data_length_left -= scsi_get_resid(req->sr_command);
	}

	/* check overflow or underflow problem, cdeng 3/24/02 */
	if (cmnd->data_length_left > cmnd->data_length) {
		residual_count =
		    cmnd->data_length_left - cmnd->data_length;
		cmnd->data_length_left = cmnd->data_length;
		cmnd->seq_flags |= OVERFLOW_FLAG;
	} else if (cmnd->data_length_left < cmnd->data_length) {
		residual_count =
		    cmnd->data_length - cmnd->data_length_left;
		cmnd->seq_flags |= UNDERFLOW_FLAG;
	}

	TRACE(TRACE_DEBUG,
	      "data_length_left %u, residual_count %d, flags 0x%08x\n",
	      cmnd->data_length_left, residual_count, cmnd->seq_flags);

	return residual_count;
}

/* auxilliary function called only by send_read_data() */
static int fill_and_send_datain_iov(struct iscsi_cmnd *cmnd,
				    struct iscsi_conn *conn,
				    int data_payload_length,
				    struct scatterlist *st_list,
				    __u8 * iscsi_hdr)
{
	int err, total_data_length, niov;
#if !defined(RDMA_MODE)
	int padding;
#endif /* RDMA_MODE */
	struct iovec *iov, *local_iov = NULL;
	struct iovec *iov_copy, *local_iov_copy = NULL;

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
	int miov, siov;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* get number of iovec slots needed to send this DataIn payload */
	niov = find_iovec_needed(data_payload_length, st_list,
				 cmnd->scatter_list_offset);
	if (niov <= 0) {
		TRACE_ERROR("%s Trouble in find_iovec_needed\n",
			    current->comm);
		err = -1;
		goto out;
	}

	/* now add on number of additional iovec slots this pdu needs */
	niov++;			/* add one for header */
	total_data_length = data_payload_length + ISCSI_HDR_LEN;
	/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
	siov = 1;
#endif

	if (conn->hdr_crc) {
		niov++;		/* add one for hdr_crc */
		total_data_length += CRC_LEN;
		/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
		siov++;
#endif
	}

	/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
	miov = niov;
#endif

#if !defined(RDMA_MODE)
	padding = (-data_payload_length) & 3;
	if (padding) {
		niov++;		/* add one for padding */
		total_data_length += padding;
	}
#endif

	if (conn->data_crc) {
		niov++;		/* add one for data_crc */
		total_data_length += CRC_LEN;
	}

	if (niov > MAX_IOV_SLOTS) {
		TRACE_WARNING
		    ("%s CmdSN %u ITT %u needs %d iov slots, command max %d\n",
		     current->comm, cmnd->cmd_sn, cmnd->init_task_tag,
		     niov, MAX_IOV_SLOTS);
		iov = local_iov =
		    my_kmalloc(niov * sizeof(struct iovec), "iovec");
		iov_copy = local_iov_copy =
		    my_kmalloc(niov * sizeof(struct iovec), "iovec");
		if (!local_iov || !local_iov_copy) {
			err = -ENOMEM;
			goto out1;
		}
	} else {
		/* use the io vector preallocated in the command structure */
		iov = cmnd->tx_iov;
		iov_copy = cmnd->tx_iov_copy;
	}

	/* now go back and fill in the iovec slots */
	iov[0].iov_base = iscsi_hdr;
	iov[0].iov_len = ISCSI_HDR_LEN;

	cmnd->scatter_list_count +=
	    fill_iovec(iov, 1 + conn->hdr_crc, niov, st_list,
		       &cmnd->scatter_list_offset, data_payload_length);

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)
	if (cmnd->cdb[0] == INQUIRY
	    && mangle_inquiry_data(cmnd, iov, siov, miov)) {
		err = -1;
		goto out1;
	}
#endif

#if !defined(RDMA_MODE)
	if (padding) {
		iov[niov - 1 - conn->data_crc].iov_base =
		    &pad_bytes_for_sending;
		iov[niov - 1 - conn->data_crc].iov_len = padding;
		TRACE(TRACE_DEBUG, "padding attached: %d bytes\n",
		      padding);
	}
#endif

	if (conn->data_crc) {
		iov[niov - 1].iov_len = CRC_LEN;
	}

	/* stop sending more DataIn pdus in case command was aborted */
	if (cmnd->state == ISCSI_DEQUEUE) {
		err = 0;
		goto out1;
	}

	/* Added for iSER - Minti
	 * This is the DataIn PDU - use put_data
	 */
	err =
	    iscsi_tx_data(cmnd, conn, iov, iov_copy, niov,
			  total_data_length, USE_PUT_DATA);

	if (err != total_data_length) {
		TRACE_ERROR("%s Trouble in iscsi_tx_data, expected %d "
			    "bytes, got %d\n", current->comm,
			    total_data_length, err);
		err = -1;
	}

      out1:
	my_kfree((void **) &local_iov_copy, "iovec");
	my_kfree((void **) &local_iov, "iovec");
      out:
	return err;
}

/*
 * auxilliary function called only by send_read_data()
 * to set A bit on a DataIn pdu that ends a sequence.
 */
static void
set_a_bit(struct iscsi_cmnd *cmnd, struct iscsi_targ_scsi_data_in *hdr)
{
	/* RFC 3720 Section 10.7.4 Target Transfer Tag and LUN
	 *
	 * On incoming data, the Target Transfer Tag and LUN
	 * MUST be provided by the target if the A bit is set
	 * to 1; otherwise they are reserved.
	 */
	hdr->flags |= A_BIT;
	hdr->target_xfer_tag = cpu_to_be32(cmnd->target_xfer_tag);
	hdr->lun = cmnd->lun;
	TRACE(TRACE_ERROR_RECOVERY,
	      "Send DataIn, A=1, ITT %u, TTT %u, dataSN %u\n",
	      cmnd->init_task_tag, cmnd->target_xfer_tag, cmnd->data_sn);
}

/* Daren Hayward, darenh@4bridgeworks.com */
#if defined(MANGLE_INQUIRY_DATA)

/*
 * offsets to first byte and limit byte of version descriptors
 * section of standard INQUIRY data layout
 */
#define VER_DESC_START	58
#define VER_DESC_LIMIT	74

/*
 * Mangle the returned INQUIRY data to ensure compliance
 * INPUT: struct iscsi_cmnd, struct iovec,
 * number of first data iovec, number of limit data iovec
 * OUTPUT returns 0 if successful, -1 on error
 * 22nd July 2004 Daren Hayward, modified by rdr
 */
static int mangle_inquiry_data(struct iscsi_cmnd *iscsi_command,
			       struct iovec *iov, int start_iov,
			       int limit_iov)
{
	Scsi_Request *req;
	int data_size;
	int i, j, offset;
	int ver_desc_start;
	int ver_desc_limit;
	unsigned char *vdb;	/* Version Descriptor byte pointer */

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (iscsi_command->cmnd == NULL
	    || (req = iscsi_command->cmnd->req) == NULL
	    || req->sr_command == NULL)
		return 0;

	if (req->sr_cmnd[0] != INQUIRY || (req->sr_cmnd[1] & 0x3) != 0) {
		/* Not INQUIRY, or not Standard INQUIRY data format */
		return 0;
	}

	TRACE(TRACE_ISCSI, "%lu %s INQUIRY of STANDARD type\n",
	      jiffies, current->comm);
	TRACE(TRACE_DEBUG, "start_iov=%d, limit_iov=%d\n", start_iov,
	      limit_iov);

	/* Is it possible for data_size to be negative?? */
	data_size = req->sr_bufflen - req->sr_command->resid;
	TRACE(TRACE_DEBUG, "mangle_inquiry_data: sr_bufflen %d, resid %d, "
	      "data_size %d\n", req->sr_bufflen, req->sr_command->resid,
	      data_size);
	if (data_size <= VER_DESC_START) {
		/* Version Descriptors start 58 bytes in */
		return 0;
	}

	/* ensure iSCSI(0960) is in Version Descriptors */
	offset = 0;
	iov += start_iov;
	for (i = start_iov; i < limit_iov; iov++, i++) {
		if (offset >= VER_DESC_LIMIT) {
			/* we've gone past end of descriptor area */
			break;
		}
		if (offset + iov->iov_len <= VER_DESC_START) {
			/* not yet up to start of descriptor area */
			offset += iov->iov_len;
			continue;
		}
		if (offset < VER_DESC_START) {
			/* start part way into this item */
			ver_desc_start = VER_DESC_START - offset;
		} else {
			/* start at beginning of this item */
			ver_desc_start = 0;
		}
		offset += iov->iov_len;
		if (offset <= VER_DESC_LIMIT) {
			/* use this entire item */
			ver_desc_limit = iov->iov_len;
		} else {
			/* use only first part of this item */
			ver_desc_limit =
			    VER_DESC_LIMIT - (offset - iov->iov_len);
		}
		TRACE(TRACE_DEBUG,
		      "i=%d, iov_len=%u, start=%d, limit=%d\n", i,
		      (unsigned int) iov->iov_len, ver_desc_start,
		      ver_desc_limit);

		/* find first all-zero vector and change it to ISCSI */
		for (j = ver_desc_start; j < ver_desc_limit; j += 2) {
			vdb = (unsigned char *) iov->iov_base;
			TRACE(TRACE_ISCSI,
			      "%lu %s Version Descriptor %d = %.2x %.2x\n",
			      jiffies, current->comm, j, vdb[j],
			      vdb[j + 1]);
			/* Check to see if anything has already been set */
			if (vdb[j] == 0x09 && vdb[j + 1] == 0x60) {
				/* ISCSI code already set, leave now */
				TRACE(TRACE_ISCSI,
				      "%lu %s ISCSI Version Descriptor found\n",
				      jiffies, current->comm);
				return 0;
			}
			if (vdb[j] == 0 && vdb[j + 1] == 0) {
				/* Set empty Version Descriptor to be ISCSI */
				vdb[j] = 0x09;
				vdb[j + 1] = 0x60;
				TRACE(TRACE_ISCSI,
				      "%lu %s ISCSI Version Descriptor added\n",
				      jiffies, current->comm);
				return 0;
			}
		}
	}

	/* Success */
	return 0;
}
#endif
