/*	initiator/initiator_rx.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 *	This file contains the functions for iscsi initiator code that are
 *	responsible for reading and processing PDUs sent by the target.
 *
 *	Copyright (C) 2001-2006 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 * 	written permission.
 */
#include "iscsi_target.h"
#include "iscsi_cmnd.h"

#ifdef RDMA_MODE

static void iser_enqueue_pdu(struct iscsi_conn *conn_p,
								struct iser_recv_pdu_t *pdu_p,
								void *iser_task_h);

struct iscsi_conn *find_targ_conn_by_iser_handle(void *iser_conn_handle)
{
	struct iscsi_conn *conn = NULL, *result = NULL;
	struct iscsi_session *session;
	DECLARE_LOOP_COUNTER(count);
	DECLARE_LOOP_COUNTER(count2);

	TRACE(TRACE_ENTER_LEAVE, "Enter find_targ_conn_by_iser_handle %p\n",
					iser_conn_handle);

	/* non-destructive access to session lists */
	read_lock(&devdata->session_rwlock);

	INIT_LOOP_COUNTER(count);

	TRACE(TRACE_ISER, "Begin search \n");
	list_for_each_entry(session, &devdata->session_list, sess_link) {
		TRACE(TRACE_ISER, "find_targ_conn_by_iser_handle: sess%p ", session);
		BUMP_LOOP_COUNTER(count, "session_list");

		INIT_LOOP_COUNTER(count2);
		list_for_each_entry(conn, &session->conn_list, conn_link) {
			TRACE(TRACE_ISER, "conn %p ", conn);
			BUMP_LOOP_COUNTER(count2, "conn_list");
			if(conn == iser_conn_handle) {
				result = conn;
				TRACE(TRACE_ISER, "MAtch found \n");
				goto find_conn_out;
			}
		}
		TRACE(TRACE_ISER, "\n");
	}

find_conn_out:
	read_unlock(&devdata->session_rwlock);
	TRACE(TRACE_ENTER_LEAVE, "Leave find_targ_conn_by_iser_handle %p\n",
					result);
	return result;
}


int recv_target_control_notifications(struct iscsi_conn *conn,
				      __u8 **pdu_bhs,
				      __u8 **data_buffer)
{
	int retval = -1;
	struct iser_recv_pdu_t *rx_iser_pdu;
	void *rx_iser_task;

	TRACE(TRACE_ENTER_LEAVE, "Enter recv_target_control_notifications conn:%p"
					" sem:%p \n", conn, &conn->iser_notify_sem);
	if (!conn->conn_status) {
		TRACE_ERROR("Transport endpoint is not connected\n");
		goto out;
	}

	TRACE(TRACE_SEM, "Waiting on iser_notify_sem in "
			"recv_target_control_notifications %p\n",
			&conn->iser_notify_sem);
	if(down_interruptible(&conn->iser_notify_sem)) {
		TRACE(TRACE_DEBUG, "recv_target_control_notifications:"
							"failed on down_interruptible\n");
		goto out;
	}
	TRACE(TRACE_SEM, "Woke from iser_notify_sem in "
							"recv_target_control_notifications %p\n",
							&conn->iser_notify_sem);

	if(conn->conn_terminated) {
		TRACE(TRACE_DEBUG, "recv_target_control_notifications: "
							"conn terminated\n");
		goto out;
	}

	iser_front_pdu(conn, &rx_iser_pdu, &rx_iser_task);
	if(rx_iser_pdu && rx_iser_pdu->rx_data.type == ISER_BUF_TYPE_SINGLE)
	{
		*data_buffer = rx_iser_pdu->rx_data.p_buf;
		*pdu_bhs = rx_iser_pdu->p_bhs->buf;
	}
	else if (rx_iser_pdu->rx_data.type != ISER_BUF_TYPE_SINGLE)
	{
		TRACE_ERROR("Error in recv_control_notifications, invalid buffer type"
			    "%d\n", rx_iser_pdu->rx_data.type);
		goto out;
	}
	else
	{
		TRACE_ERROR("%s: No PDU on the queue\n", __func__);
		goto out;
	}

	/* this receive from initiator broke the silence */
	clear_bit(SILENCE_BIT, &conn->control);

	retval = 0;
	TRACE(TRACE_DEBUG, " data_buffer:0x%p pdu_bhs:0x%p datalen:%d\n",
			*data_buffer, *pdu_bhs, rx_iser_pdu->rx_data.size);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave recv_target_control_notifications "
							"retval:%d", retval);
	return retval;
}

void send_notify(iser_conn_handle_t cid_h)
{
	struct iscsi_conn *conn;
	TRACE(TRACE_ENTER_LEAVE, "Enter send_notify\n");

	conn = find_targ_conn_by_iser_handle(cid_h);
	if(conn == NULL) {
		TRACE_ERROR("Error in control_notify, invalid conn handle %p\n", cid_h);
		return;
	}

	up(&conn->send_done_sem);
	TRACE(TRACE_ENTER_LEAVE, "Leave send_notify\n");
}


void control_notify(iser_conn_handle_t cid_h, struct iser_recv_pdu_t *p_ctrl_pdu)
{
	struct iscsi_conn *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter control_notify conn:%p ctrl_pdu:%p\n",
			cid_h, p_ctrl_pdu);

	conn = (struct iscsi_conn *) cid_h;
	iser_enqueue_pdu(conn, p_ctrl_pdu, NULL);

	TRACE(TRACE_SEM, "%s: up on iser_notify_sem %p\n", __func__,
	      &conn->iser_notify_sem);
	up(&conn->iser_notify_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave control_notify %p\n",
									&conn->iser_notify_sem);
}

/**
 * \brief The command notify callback.
 * \param iscsi_conn_h This output parameter is used to return the
 *                     struct iscsi_conn for the connection that the command
 *                     was received on.  This is used later to retrieve the
 *                     connection when a datatransfer completes.
 */
void command_notify(iser_conn_handle_t cid_h,
					struct iser_recv_pdu_t *p_ctrl_pdu,
					void *iser_task_h, void **iscsi_conn_h)
{
	struct iscsi_conn *conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	conn = (struct iscsi_conn *) cid_h;
	iser_enqueue_pdu(conn, p_ctrl_pdu, iser_task_h);
	up(&conn->iser_notify_sem);

	*iscsi_conn_h = conn;

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

void control_terminate_notify(iser_conn_handle_t cid_h)
{
	struct iscsi_conn *conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter control_terminate_notify %p\n", cid_h);

	conn = find_targ_conn_by_iser_handle(cid_h);
	if(conn == NULL) {
		TRACE_ERROR("Error in control_terminate_notify, invalid conn handle"
					" %p\n", cid_h);
		return;
	}
	conn->conn_terminated = 1;
	up(&conn->iser_notify_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave control_terminate_notify\n");
}

void free_iser_login_buffer(struct iscsi_conn *conn, struct generic_pdu *inpdu)
{
	struct iser_recv_pdu_t *rx_iser_pdu;
	void *rx_iser_task;

	TRACE(TRACE_ENTER_LEAVE, "Enter free_iser_login_buffer \n");

	if(conn->connection_flags & USE_RDMA_LOGIN) {
		iser_front_pdu(conn, &rx_iser_pdu, &rx_iser_task);
		if(rx_iser_pdu) {
			iser_status_t iser_retval;
			__u8 *bhs = (__u8 *)rx_iser_pdu->p_bhs->buf;
			__u8 *data = (__u8 *)rx_iser_pdu->rx_data.p_buf;

			if(((__u8 *)inpdu->hdr != bhs) ||
				((__u8 *)inpdu->text != data))
			{
				TRACE_ERROR("free_iser_login_buffer called with incorrect args"
						"%p %p %p %p\n", inpdu->hdr, bhs, inpdu->text, data);
				return;
			}
			TRACE(TRACE_ISER, "Before release_control\n");
			iser_retval = devdata->iser_api.release_control(
										(iser_conn_handle_t)conn,
										rx_iser_pdu);
			if(iser_retval != ISER_SUCCESS)
				TRACE_ERROR("iser_release_control returned error %d",
							iser_retval);

			iser_dequeue_pdu(conn);
		}
		else {
			TRACE_ERROR("free_iser_login_buffer called with NULL buffer, %p\n",
						inpdu);
		}
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave free_iser_login_buffer\n");
}

void targ_release_control(struct iscsi_conn *conn, __u8 **bhs, __u8 **buffer)
{
	struct iser_recv_pdu_t *rx_iser_pdu;
	void *rx_iser_task;

	TRACE(TRACE_ENTER_LEAVE, "Enter targ_release_control\n");

	if(conn->session->oper_param->RDMAExtensions) {
		iser_status_t iser_retval;
		iser_front_pdu(conn, &rx_iser_pdu, &rx_iser_task);
		if(rx_iser_pdu != NULL) {
			__u8 *conn_bhs = (__u8 *)rx_iser_pdu->p_bhs->buf;
			__u8 *conn_data = (__u8 *)rx_iser_pdu->rx_data.p_buf;

			if(((__u8 *) conn_bhs != *bhs) ||
				(*buffer && conn_data != *buffer))
			{
				TRACE_ERROR("free_iser_login_buffer called with incorrect args"
						"%p %p %p %p\n", *bhs, conn_bhs, *buffer, conn_data);
				return;
			}

			iser_retval = devdata->iser_api.release_control(
							(iser_conn_handle_t)conn,
							rx_iser_pdu);
			if(iser_retval != ISER_SUCCESS)
				TRACE_ERROR("iser_release_control returned error %d",
												iser_retval);
			iser_dequeue_pdu(conn);
		}
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave targ_release_control\n");
}
#endif

void free_targ_buf(struct iscsi_conn *conn, __u8 **bhs, __u8 **buffer)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter free_targ_buf conn%p bhs%p buffer%p\n",
						conn, *bhs, *buffer);
#ifdef RDMA_MODE
	targ_release_control(conn, bhs, buffer);
#else
	my_kfree((void *)buffer, "data_buf");
#endif
	TRACE(TRACE_ENTER_LEAVE, "Leave free_targ_buf\n");
}
#if 0
void free_queued_targ_buf(struct iscsi_cmnd *cmnd, char **buffer)
{
#ifdef RDMA_MODE
	if(cmnd->session->oper_param->RDMAExtensions) {
		if(cmnd->rx_iser_pdu != NULL) {
			iser_status_t iser_retval;

			__u8 *data = (__u8 *)cmnd->rx_iser_pdu->rx_data.p_buf;
			if((*buffer != NULL) && (data != (__u8 *)*buffer) {
				TRACE_ERROR("free_iser_login_buffer called with incorrect args"
						"%p %p\n", *buffer, data);
				return;
			}
			iser_retval = devdata->iser_api.release_control(
							(iser_conn_handle_t)cmnd->conn,
							cmnd->rx_iser_pdu);
			if(iser_retval != ISER_SUCCESS)
				TRACE_ERROR("iser_release_control returned error %d", iser_retval);

			cmnd->rx_iser_pdu = NULL;
		}
	}
	else
#endif
	{
		my_kfree((void *)buffer, "data_buf");
	}
}
#endif



int iser_alloc_conn_resources(struct iscsi_conn *conn,
			struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
#ifdef RDMA_MODE
	int max_outstanding = 1;
	//u32 max_send_length = 1024;
	u32 max_recv_length = 1024;
	iser_status_t iser_retval;
	struct iser_conn_resources_t conn_resources;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_alloc_conn_resources\n");

	/* EAB -- This doesn't work since we really don't have any params.
	// Gets the lengths based on whether the IRDSL or TRDSL were used
	set_connection_length(p_param_tbl, &max_recv_length, &max_send_length,
				TARGET, 1);
	*/

	conn_resources.conn_type = ISER_TRANSPORT_CONN_IB;
	conn_resources.api_h = conn->iser_handle;
	TRACE(TRACE_ISER, "%s: conn_resources.api_h=%p\n", __func__,
										conn_resources.api_h);

	/* RDR - should change QUEUE_DEPTH_ALLOWED to get value
	 * set by "force Q=" in max_queue_depth_allowed field of global/session
	 */
	conn_resources.max_outstand_cmds = QUEUE_DEPTH_ALLOWED + 1;

	/* Not used in the iSER DM code right now
	 * Need to chk that FBL < MRDSL and within bounds if using (i.e.
	 * do check_integrity_rules etc. first)
	 */
	conn_resources.first_burst_length = 0;
	conn_resources.max_recv_dsl = max_recv_length;

	//Still used in the code although headers claim it needs deleting - fill in
	conn_resources.max_recv_pdu_sz = max_recv_length; //Same as max_recv_dsl

	/* Don't need to pre-register buffers */
	conn_resources.num_reg_buf = 0;
	conn_resources.reg_buf = NULL;

	iser_retval = devdata->iser_api.allocate_connection_resources((iser_conn_handle_t) conn,
								      &conn_resources,
								      &max_outstanding);
	if(iser_retval != ISER_SUCCESS)	{
		TRACE_ERROR("iser_allocate_connection_resources: Failed %d\n",
					iser_retval);
		return -1;
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_alloc_conn_resources\n");
#endif
	return 0;
}

int iser_enable_datamover(struct iscsi_conn *conn,
			  struct generic_pdu *final_loginreply)
{
#ifdef RDMA_MODE

	struct iser_send_pdu_t send_pdu;
	iser_status_t iser_retval;

	send_pdu.p_bhs = (void*) final_loginreply->hdr;
	send_pdu.data.tx.buf.p_buf = final_loginreply->text;
	send_pdu.data.tx.buf.size = final_loginreply->text_length;
	send_pdu.data.tx.buf.type = ISER_BUF_TYPE_SINGLE;

	iser_retval = devdata->iser_api.enable_datamover((iser_conn_handle_t)conn, NULL, &send_pdu);
	if(iser_retval != ISER_SUCCESS)	{
		TRACE_ERROR("iser_enable_datamover: Failed %d\n", iser_retval);
		return -1;
	}

#endif
	return 0;
}

int iser_dealloc_conn_resources(struct iscsi_conn *conn)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_dealloc_conn_resources\n");
	iser_retval = devdata->iser_api.deallocate_connection_resources(
										(iser_conn_handle_t)conn);
	if(iser_retval != ISER_SUCCESS)	{
		TRACE_ERROR("iser_dealloc_conn_resources: Failed %d\n", iser_retval);
		return -1;
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_dealloc_conn_resources\n");
#endif
	return 0;
}

int iser_conn_terminate(struct iscsi_conn *conn)
{
#ifdef RDMA_MODE
	iser_status_t iser_retval;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_conn_terminate\n");
	iser_retval = devdata->iser_api.connection_terminate(
										(iser_conn_handle_t)conn);
	if(iser_retval != ISER_SUCCESS)	{
		TRACE_ERROR("iser_connection_terminate: Failed %d\n", iser_retval);
		return -1;
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_conn_terminate\n");
#endif
	return 0;
}

void free_queued_targ_buf(struct iscsi_cmnd *cmnd, char **buffer,
							void **iser_buf)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter free_queued_targ_buf cmnd %p, buffer %p"
				"iser_buf%p\n", cmnd, *buffer, *iser_buf);
#ifdef RDMA_MODE
	if(cmnd->session->oper_param->RDMAExtensions) {
		if(*iser_buf != NULL) {
			iser_status_t iser_retval;

			iser_retval = devdata->iser_api.release_control(
							(iser_conn_handle_t)cmnd->conn,
							(struct iser_recv_pdu_t *)*iser_buf);
			if(iser_retval != ISER_SUCCESS)
				TRACE_ERROR("iser_release_control returned error %d",
							iser_retval);

			*iser_buf = NULL;
		}
	}
	else
#endif
	{
		my_kfree((void *)buffer, "data_buf");
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave free_queued_targ_buf\n");
}

#ifdef RDMA_MODE
iser_status_t connection_establish_notify(iser_conn_handle_t cid_h,
											int established)
{
	struct iscsi_conn *conn = (struct iscsi_conn *) cid_h;
	struct portal_group *pg;

	TRACE(TRACE_ENTER_LEAVE, "Enter connection_establish_notify %p %d\n",
													cid_h, established);

	pg = &iscsi_portal_groups[0];
	if(pg == NULL)
		goto out;

	if(conn == NULL) {
		TRACE_ERROR("Error in connection_establish_notify, invalid conn handle"
					"%p\n", cid_h);
			TRACE(TRACE_ENTER_LEAVE, "Leave connection_establish_notify\n");
			return ISER_FAILURE;
	}

	if(!established)
		conn->conn_status = CONN_SETUP_FAIL;
	conn->iser_handle = pg->iser_handle;
	up(&conn->iser_conn_done_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave connection_establish_notify\n");
	return ISER_SUCCESS;
 out:
	TRACE(TRACE_ENTER_LEAVE, "Leave %s\n", __func__);
	return ISER_SUCCESS;
}

iser_status_t connection_request_notify(iser_conn_request_t conn_req_h,
    									struct sockaddr_in *dst_addr,
    									struct sockaddr_in *src_addr)
{
	struct portal_group *pg;

	TRACE(TRACE_ENTER_LEAVE, "Enter iser_conn_request_notify\n");

	/* Find the portal that this connection is for (based on the handle
	that will be added in the future by iSER) */
	pg = &iscsi_portal_groups[0];
	if(pg == NULL) {
		goto out;
	}
	TRACE(TRACE_ISER, "In iser_conn_request_notify\n");


	pg->conn_req_h = conn_req_h;
	/* sockaddr_in is way smaller (8 bytes) compared to sockaddr (16 bytes) */
	TRACE(TRACE_ISER, "dst_addr %p, src_addr %p\n", dst_addr, src_addr);

	if(dst_addr) {
		pg->dst_addr = my_kmalloc(sizeof(struct sockaddr), "sockaddr");
		memcpy(pg->dst_addr, dst_addr, sizeof(struct sockaddr_in));
	}
	if(src_addr) {
		pg->src_addr = my_kmalloc(sizeof(struct sockaddr), "sockaddr");
		memcpy(pg->src_addr, src_addr, sizeof(struct sockaddr_in));
	}

	up(&pg->iser_conn_sem);
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iser_conn_request_notify %p\n", conn_req_h);
	return ISER_SUCCESS;
}

void data_ack_notify (void *iscsi_conn_h, int itt, int data_sn )
{
	TRACE(TRACE_ENTER_LEAVE, "Enter data_ack_notify - function not "
												"implemented\n");

	TRACE(TRACE_ENTER_LEAVE, "Leave data_ack_notify\n");
}

/**
 * \brief Called by iSER when a Data-In transfer is completed.
 * \param iscsi_conn_h A pointer to the struct iscsi_conn for the connection
 *                     that contains the data transfer task.  This, in
 *                     conjunction with the itt, is used to find the command
 *                     that initiated the data transfer.
 * \param itt The Initiator task tag for the task that this data transfer was
 *            for.
 * \param sn The sequence number (R2TSN?) of the completed transfer.
 */
void data_completion_notify(void *iscsi_conn_h, int itt, int sn)
{
	struct iscsi_conn *conn;
	struct iscsi_cmnd *cmd;
	__u8 opcode;
	int err = 0;
	unsigned long flags;

	TRACE(TRACE_ENTER_LEAVE, "Enter data_completion_notify\n");

	conn = iscsi_conn_h;
	if(conn == NULL) {
		TRACE_ERROR("data_completion_notify: conn %p not found\n",
														iscsi_conn_h);
	}

	cmd = search_tags(conn, itt, ALL_ONES, 0);
	if(cmd == NULL) {
		TRACE_ERROR("ITT %u, SN %u, No matching command\n",	itt, sn);
		return;
	}

	TRACE(TRACE_ISER, "%s: iSCSI cmnd=%p\n", __func__, cmd);

	opcode = cmd->opcode_byte & ISCSI_OPCODE;
	if(opcode != ISCSI_INIT_SCSI_CMND) {
		TRACE_ERROR("cmd is not a SCSI Command, itt %u, sn %u, opcode %u\n",
					itt, sn, opcode);
		return;
	}

	if(cmd->command_flags & R_BIT) {
		TRACE(TRACE_ISER, "%s: R_BIT is set\n", __func__);
		if(conn->data_in_complete) {
			TRACE(TRACE_ISER, "%s: upping data_in_sem\n", __func__);
			/* If this is the final data-in, wake up the tx thread so it can
			transmit the response */
			up(&conn->data_in_sem);
		} else {
			TRACE(TRACE_ISER, "%s: no need to up "
			"data_in_sem, data_in_complete is not set\n", __func__);
		}
	} else {
		TRACE(TRACE_ISER, "%s: R_BIT is not set\n", __func__);
	}

	if(cmd->command_flags & W_BIT) {
		int length;

		TRACE(TRACE_ISER, "%s: W_BIT is set\n", __func__);

		/* Find the cookie with this r2tsn to get the xfer_len */
		if(get_r2t_xfer_len(cmd, sn, &length) < 0) {
			TRACE_ERROR("%s: get_r2t_xfer_len() returned negative\n",
			            __func__);
			return;
		}

		/* This notify is called for a write only when the entire R2T sequence
		   has completed */
		__update_after_read(length, 1, cmd);

		cmd->outstanding_r2t--;
		if (cmd->r2t_data_left > 0) {
			if (atomic_read(&conn->tx_sem.count) <= 0)
				up(&conn->tx_sem);
		}

		TRACE(TRACE_ISER, "%s: cmd->data_done=%u, cmd->data_length=%u\n",
							__func__, cmd->data_done, cmd->data_length);

		if (cmd->data_done >= cmd->data_length) {
		/* we have just finished reading everything the command expected */

			/* tell the scsi midlevel that all the expected data has arrived */
			spin_lock_irqsave(&conn->session->cmnd_lock, flags);
			cmd->state = ISCSI_DATA_IN;
			err = scsi_rx_data(cmd->cmnd);
			spin_unlock_irqrestore(&conn->session->cmnd_lock, flags);
		}

		/* end of sequence of DataOut pdus, tx thread may need to do something*/
		if (atomic_read(&conn->tx_sem.count) <= 0)
			up(&conn->tx_sem);

	} else {
		TRACE(TRACE_ISER, "%s: W_BIT is not set\n", __func__);
	}
	TRACE(TRACE_ENTER_LEAVE, "Leave data_completion_notify\n");
	return;
}

/* Added for iSER - Minti */
int iser_bring_up_portal(struct portal_group *ptr)
{
	char targ_string[100];

	struct iser_api_cb_t iser_api_cb;
	struct iser_ext_api_cb_t iser_ext_api_cb;
	iser_status_t status;

	/* Needs a unique string for each registration */
	strcpy(targ_string, "unh_iscsi_target");
	strcat(targ_string, ptr->ip_string);
	strcat(targ_string, ptr->port_string);

	iser_api_cb.control_notify = control_notify;
	iser_api_cb.command_notify = command_notify;
	iser_api_cb.conn_terminate_notify = control_terminate_notify;
	iser_api_cb.data_ack_notify = data_ack_notify;
//	iser_api_cb.send_notify = send_notify;

	iser_api_cb.data_completion_notify = data_completion_notify; //TODO
	iser_ext_api_cb.connection_request_notify = connection_request_notify;
	iser_ext_api_cb.connection_establish_notify = connection_establish_notify;

	TRACE(TRACE_ISER, "connection_establish_notify%p \n", connection_establish_notify);

	status = iser_api_register_target(targ_string,
							&devdata->iser_api,
							&iser_api_cb,
							&devdata->iser_ext_api,
							&iser_ext_api_cb,
							ptr->ip_address,
							ptr->ip_length,
							&ptr->iser_handle);
	if(status != ISER_SUCCESS)
	{
		TRACE_ERROR("iser_api_register failed %d\n", status);
		return 1;
	}

	/* create a server_thread that can accept connections */
	if (kernel_thread(iscsi_server_thread, (void *)ptr, 0) < 0) {
		TRACE_ERROR("Unable to create server thread\n");
		ptr->iser_handle = NULL;
		goto out;
	}

	TRACE(TRACE_NET, "Server_thread spawned \n");

	/* wait for that server_thread to come up before continuing */
	if (!down_interruptible(&devdata->server_sem))
		return 0;	//Success
out:
	status = iser_api_unregister(ptr->iser_handle);
	if(status != ISER_SUCCESS)
		TRACE_ERROR("iser_api_unregister failed %d\n", status);

	my_kfree((void **)&ptr->ip_address, "ip_address");
	return 1;
}

/**
 * \brief Enqueues a newly received PDU for a given connection.
 * \param conn_p A pointer to the connection.
 * \param pdu_p A pointer to the PDU.
 * \param iser_task_h If the PDU was received with command_notify(), then this
 *                    parameter is the iser_task_h handle, or else it should
 *                    be NULL.
 */
static void iser_enqueue_pdu(struct iscsi_conn *conn_p,
								struct iser_recv_pdu_t *pdu_p,
								void *iser_task_h)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s, conn_p=%p, pdu_p=%p, "
				"iser_task_h=%p\n", __func__, conn_p, pdu_p, iser_task_h);

	queue_enqueue(&conn_p->rx_pdu_q, pdu_p);
	queue_enqueue(&conn_p->rx_handle_q, iser_task_h);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Removes the PDU at the front of the queue.
 * \param conn_p The connection to remove the PDU from.
 */
void iser_dequeue_pdu(struct iscsi_conn *conn_p)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	queue_dequeue(&conn_p->rx_pdu_q);
	queue_dequeue(&conn_p->rx_handle_q);

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

/**
 * \brief Gets the front PDU.
 * \param conn_p The connection to remove the PDU from.
 * \param rx_pdu_buff Buffer used to return the front PDU pointer.
 * \param rx_task_buff Buffer used to return the front iSER task handle.
 */
void iser_front_pdu(struct iscsi_conn *conn_p,
						struct iser_recv_pdu_t **rx_pdu_buff,
						void **rx_task_buff)
{
	int err;

	TRACE(TRACE_ENTER_LEAVE, "Entering %s\n", __func__);

	if (queue_empty(&conn_p->rx_pdu_q))
		goto error;
	err = queue_front(&conn_p->rx_pdu_q, (void**) rx_pdu_buff);
	if (err)
		goto error;
	err = queue_front(&conn_p->rx_handle_q, rx_task_buff);
	if (err)
		*rx_task_buff = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
	return;

 error:
	*rx_pdu_buff = NULL;
	*rx_task_buff = NULL;
	TRACE(TRACE_ENTER_LEAVE, "Leaving %s\n", __func__);
}

#endif /* RDMA_MODE */

/* vi: set tabstop=4: */
