/*
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * scsi_diskio_target.c --	Contains functions and declarations for
 * 				DISKIO scsi mode.
 *
 * Created Fri 23 Feb 2007 11:46:16 AM EST -- Ethan Burns
 */
#include "scsi_target.h"
# ifdef DISKIO

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
#include <scsi_device.h>
#endif				/* LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18) */

#include "../common/iscsi_common.h"

char device[BYTE * 5];
#ifdef TRUST_CDB
const char pr_label[] = "DISKIO TRUST_CDB";
#else
const char pr_label[] = "DISKIO not TRUST_CDB";
#endif

/*
 * Forward declarations
 */
static void __te_cmnd_processed(struct scsi_cmnd *sc_cmd,
				struct scsi_request *sreq);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
static void te_cmnd_processed(struct scsi_cmnd *);
#else
static void te_cmnd_processed(void *, char *, int, int);
#endif				/* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18) */

static int add_one_scsi_device(struct scsi_device *the_device)
{
	struct target_map_item *this_item;

	printk
	    ("iscsi target %d: scsi device type %s, host %d, channel %d, "
	     "targetid %d, lun %d\n", target_count,
	     printable_dev_type(the_device->type),
	     the_device->host->host_no, the_device->channel,
	     the_device->id, the_device->lun);

	write_lock(&target_map_rwlock);

#if defined(TRUST_CDB)
	/* always add this device to end of list of known target devices */
	this_item = kmalloc(sizeof(struct target_map_item), GFP_KERNEL);
	if (!this_item) {
		printk
		    ("add_one_scsi_device: kmalloc failed after %d targets\n",
		     target_count);
		return -1;
	}
	this_item->the_device = the_device;
	this_item->in_use = 1;
	this_item->target_id = target_count++;
	list_add_tail(&this_item->link, &target_map_list);
#else
	if (target_count >= MAX_TARGETS) {
		printk
		    ("More than MAX_TARGETS %d target devices, ignoring this\n",
		     MAX_TARGETS);
	} else if (the_device->lun >= MAX_LUNS) {
		printk("LUN %d >= MAX_LUNS %d, ignoring this\n",
		       the_device->lun, MAX_LUNS);
	} else {
		/* set up this element of the target_map array for this device */
		this_item = &target_map[target_count][the_device->lun];
		if (this_item->in_use) {
			printk
			    ("targetid %d lun %d already in use, ignoring this\n",
			     target_count, the_device->lun);
		} else {
			this_item->the_device = the_device;
			this_item->in_use = 1;
			target_count++;
		}
	}
#endif

	write_unlock(&target_map_rwlock);
	return 0;
}

int add_scsi_device_by_addr(int hostid, int chan, int id, int lun)
{
	struct Scsi_Host *host;
	int ret = -ESRCH;

	host = scsi_host_lookup(hostid);
	if (!host) {
		printk(KERN_ERR "Unable to find host %d\n", hostid);
	} else {
		struct scsi_device *dev;
		dev = scsi_device_lookup(host, chan, id, lun);
		if (!dev) {
			printk(KERN_ERR "Unable to find device "
			       "%d:%d:%d:%d\n", hostid, chan, id, lun);
		} else
			ret = add_one_scsi_device(dev);
	}

	return ret;
}

/* if we trust the information of iSCSI from initiator, then
 * set the req and malloc space using that information, else
 * we parse the cdb to get correct information ourselves */
#ifdef TRUST_CDB

/*
 * handle_cmd: This command is responsible for dealing with the received
 * command when a real is used by the scsi target emulator. It receives
 * new commands, and gets them to the state where they are processing or
 * pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 * REMARKS - ramesh@global.com : changed the code to not to parse the
 *                               input cdb. this makes the code to be
 *                               independent of devices.
 */
int handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = -1;
	int allocate_buff = 1;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering %s handle_cmd\n", pr_label);
# endif

	/* set data direction */
	cmnd->req->sr_data_direction = SCSI_DATA_NONE;
	if (cmnd->flags & R_BIT) {
		cmnd->req->sr_data_direction = SCSI_DATA_READ;
	} else if (cmnd->flags & W_BIT) {
		cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
	}
# ifdef DEBUG_HANDLE_CMD
	printk("%s handle_cmd: data direction %d data len %d\n", pr_label,
	       cmnd->req->sr_data_direction, cmnd->datalen);
# endif

	/*check the data length need to be allocated */
	if (cmnd->datalen) {
		if (cmnd->req->sr_data_direction == SCSI_DATA_WRITE &&
		    cmnd->state == ST_TO_PROCESS) {
			allocate_buff = 0;
		}

		if (allocate_buff) {
			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd, cmnd->datalen, 0)) {
				err = 0;	/* come back later */
				return err;
			}
		}
	} else {
		cmnd->req->sr_use_sg = 0;
		cmnd->req->sr_bufflen = 0;
	}
# ifdef DEBUG_HANDLE_CMD
	printk("%s handle_cmd: buffer length %d\n", pr_label,
	       cmnd->req->sr_bufflen);
# endif

	/* change status */
	/* for write needs more data to arive... */
	if (cmnd->req->sr_data_direction == SCSI_DATA_WRITE &&
	    cmnd->state == ST_NEW_CMND) {
		cmnd->state = ST_PENDING;
	} else {
		cmnd->state = ST_PROCESSING;
		/* Quick fix timeout, for tape is 60 * disk */
		/* hand it off to the mid-level to deal with */
		scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
			    cmnd->req->sr_bufflen, te_cmnd_processed,
			    cmnd->req->sr_device->type == TYPE_TAPE
			    ? TE_TIMEOUT * 60 : TE_TIMEOUT, TE_TRY);
# ifdef DEBUG_HANDLE_CMD
		printk("%s handle_cmd: handing it off to mid_level\n",
		       pr_label);
# endif
	}

	return 0;
}

#else				/* ifdef TRUST_CDB */

static int handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	int to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering %s handle_cmd with id %d\n", pr_label, cmnd->id);
# endif

	switch (cmnd->req->sr_cmnd[0]) {
	case READ_CAPACITY:	/* in diskio not trustcdb mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd, READ_CAP_LEN, 0)) {
				err = 0;	/* come back later */
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case INQUIRY:		/* in diskio not trustcdb mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			if ((err = setup_new_command(cmnd, &to_read)) <= 0)
				break;

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case REPORT_LUNS:	/* in diskio not trustcdb mode */
		{
			if ((to_read =
			     allocate_report_lun_space(cmnd)) < 0) {
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case MODE_SENSE:	/* in diskio not trustcdb mode */// cdeng    May 28, 2002
		{
# ifdef DEBUG_HANDLE_CMD
			printk("MODE_SENSE received\n");
# endif
			if ((err = setup_new_command(cmnd, &to_read)) <= 0)
				break;

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case TEST_UNIT_READY:	/* in diskio not trustcdb mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case VERIFY:		/* in diskio not trustcdb mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("VERIFY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case READ_6:		/* in diskio not trustcdb mode */
	case READ_10:		/* in diskio not trustcdb mode */
		{
			/* set up the newly-received diskio not trustcdb mode READ */
			if ((err = setup_new_read(cmnd)) <= 0)
				break;

			/* change_status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd,
				    cmnd->req->sr_buffer,
				    cmnd->req->sr_bufflen,
				    te_cmnd_processed, TE_TIMEOUT, TE_TRY);

			err = 0;
			break;
		}

	case WRITE_6:		/* in diskio not trustcdb mode */
	case WRITE_10:		/* in diskio not trustcdb mode */
		{
			if (cmnd->state == ST_NEW_CMND) {
				/* set up the newly-received diskio not trustcdb mode WRITE */
				err = setup_new_write(cmnd);
				break;
			} else if (cmnd->state == ST_TO_PROCESS) {
				cmnd->state = ST_PROCESSING;

				/* hand it off to the mid-level to deal with */
				scsi_do_req(cmnd->req, cmnd->cmd,
					    cmnd->req->sr_buffer,
					    cmnd->req->sr_bufflen,
					    te_cmnd_processed, TE_TIMEOUT,
					    TE_TRY);
			}
			err = 0;
			break;
		}

	default:
		{
			printk("%s handle_cmd: Unknown command 0x%02x\n",
			       pr_label, cmnd->req->sr_cmnd[0]);
			err = -1;
			break;
		}
	}
# ifdef DEBUG_HANDLE_CMD
	printk("%s handle_cmd: cmd id %d, state %d, err %d\n",
	       pr_label, cmnd->id, cmnd->state, err);
# endif
	return err;
}
#endif				/* end of else of ifdef TRUST_CDB */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
/*
 * te_cmnd_processed: This is the done function called by the mid-level
 * upon detecting that a particular command has been executed. This
 * function can be called from an interrupt context - it is responsible
 * for changing the state of the command and waking up the target
 * mid-level.
 */
static void te_cmnd_processed(struct scsi_cmnd *sc_cmd)
{
	Scsi_Request *req;
	Target_Scsi_Cmnd *te_cmd;
	int found;
	unsigned long flags;

	/* find the scsi_request */
	req = sc_cmd->sc_request;

	__te_cmnd_processed(sc_cmd, req);
}
#else				/* !LINUX_VERSION_CODE < KERNEL_VERISON(2, 6, 18) */
/*
 * This version of te_cmnd_processed() has a prototype compatable with 
 * scsi_execute_async().
 */
static void te_cmnd_processed(void *data, char *sense, int result,
			      int resid)
{
	struct scsi_request *sreq;

	sreq = (struct scsi_request *) data;
	memcpy(sreq->sr_sense_buffer, sense,
	       sizeof(sreq->sr_sense_buffer));
	sreq->sr_result = result;

	__te_cmnd_processed(NULL /* unused in >=2.6.18 */ , sreq);
}
#endif				/* LINUX_VERSION_CODE < KERNEL_VERISON(2, 6, 18) */

/*
 * The sc_cmd parameter is not used in kernel versions greater or equal to
 * 2.6.18, setting it to NULL should be fine.
 * - eaburns
 */
void __te_cmnd_processed(struct scsi_cmnd *sc_cmd,
			 struct scsi_request *req)
{
	int found;
	unsigned long flags;
	Target_Scsi_Cmnd *te_cmd;

	/* find the Target_Scsi_Cmnd containing the scsi_request */
	found = 0;
	spin_lock_irqsave(&target_data.cmd_queue_lock, flags);
	list_for_each_entry(te_cmd, &target_data.cmd_queue, link) {
		if (te_cmd->req == req) {
			found = 1;
			break;
		}
	}
	spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);

	if (!found) {
		/* we can just throw away the response */
		/* may be the command was aborted */
		printk
		    ("te_cmnd_processed: Could not find request in command queue "
		     "for cmd %p, id %d, state %d\n", te_cmd, te_cmd->id,
		     te_cmd->state);
		scsi_release_request(req);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
		sc_cmd->sc_request = NULL;
#endif				/* LINUX_VERSION_CODE < KEREL_VERSION(2, 6, 18) */
	} else {
		/* change its state */
# ifdef DEBUG_HANDLE_CMD
		printk("Got response for cmd %p, id %d, state %d\n",
		       te_cmd, te_cmd->id, te_cmd->state);
# endif

		/* if the lun of device is not the lun of request */
		if (te_cmd->lun != req->sr_device->lun
		    && req->sr_cmnd[0] == INQUIRY) {
			__u8 *inq;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			inq =
			    page_address(((struct scatterlist *) req->
					  sr_buffer)->page);
#else
			inq =
			    ((struct scatterlist *) req->sr_buffer)->
			    address;
#endif
			if (inq) {
				inq[0] = 0x7f;	// 011 + 1FH
			}
		}

		te_cmd->state = ST_DONE;
		/* wake up scsi_process_target_thread so it can dequeue stuff */
		if (atomic_read(&target_data.target_sem.count) <= 0) {
			up(&target_data.target_sem);
		}
	}

	return;
}

# endif
/* vi: set tabstop=8 textwidth=78: */
