/* -*- mode:c; tab-width:4; fill-column:78 -*-
 * scsi_fileio_target.c
 *
 *	vi: set autoindent tabstop=4 shiftwidth=8 textwidth=78 :
 *
 *	Contains functions and declarations for FILEIO
 * 				scsi emulation mode.
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 * Created Fri 23 Feb 2007 11:47:35 AM EST -- Ethan Burns
 */

#ifdef __KERNEL__

#else

#include "../common/user_space.h"

#endif

#include "../common/scatterlist.h"

#include "scsi_target.h"
#ifdef FILEIO
char device[BYTE * 5];
const char pr_label[] = "FILEIO";

#ifndef __KERNEL__
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
/**
 * Reads the requested number of bytes.
 */
ssize_t default_read(struct file *file, void *buf, size_t count,
		     loff_t * offs)
{
    size_t bytes_left;
    ssize_t nread;
    struct stat sbuf;
    int fd;

    fd = file->fd;

    fstat(fd, &sbuf);

    if (count + *offs > sbuf.st_size) {
	printf("%s: clamping read because count=%lu + offs=%lu > "
	       "file size=%lu\n", __func__, (unsigned long) count,
	       (unsigned long) *offs, (unsigned long) sbuf.st_size);
	count = sbuf.st_size - *offs;
    }
    bytes_left = count;

    if (*offs)
	lseek(fd, *offs, SEEK_SET);

    while (bytes_left) {
	nread = read(fd, buf + (count - bytes_left), bytes_left);
	bytes_left -= nread;
    }

    /*
     *offs += count;
     */

    return count;
}
#endif				/* !__KERNEL__ */


/* Bjorn Thordarson, 8-May-2004 -- start -- */
/* RDR */
/* set_up_bad_command_sense_response: This function fills up the req->sr_buffer
 * with sense data indicating a bad command (e.g., no fileio "disk" present)
 * See RFC 3720 page 109 section 9.2 Autosense and Auto Contingent Alliance
 *	"Autosense refers to the automatic return of sense data to the
 *	 initiator in case a command did not complete successfully.  iSCSI
 *	 initiators and targets MUST support and use autosense."
 *
 * See RFC 3720 page 43 section 3.5.1.2 SCSI-Response
 *	"For a valid status that indicates that the command has been
 *	 processed, but resulted in an exception (e.g., a SCSI CHECK
 *	 CONDITION), the PDU data segment contains the associated sense data.
 *	 The use of Autosense ([SAM2]) is REQUIRED by iSCSI."
 *
 * See RFC 3720 page 125 section 10.4.7. Data Segment - Sense and Response Data
 *	"iSCSI targets MUST support and enable autosense.  If Status is CHECK
 *	 CONTITION (0x02), then the Data Segment MUST contain sense data for
 *	 the failed command."
 *	 . . .
 *	 If the DataSegmentLength is not 0, the format of the Data Segment is
 *	 as follows:
 *  Byte/      0       |       1       |       2       |       3       |
 *      /              |               |               |               |
 *     |0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|
 *     +---------------+---------------+---------------+---------------+
 *    0| SenseLength                   | Sense Data		       |
 *     +---------------+---------------+---------------+---------------+
 *    x/ Sense Data						       /
 *     +---------------+---------------+---------------+---------------+
 *    y/ Response Data						       /
 *     /							       /
 *     +---------------+---------------+---------------+---------------+
 *    z|
 *
 * INPUT: Scsi_Request pointer to the bad command having any opcode
 *	  sense_key is put into the sense key field of the data
 *	  asc is put into the additional sense code field
 *	  ascq is put into the additional sense code qualifier field
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */

struct iscsi_sense_data {
    __u16 length;
    __u8 data[0];
} __packed;

int set_up_bad_command_sense_response(Scsi_Request * req, __u8 sense_key,
				      __u8 asc, __u8 ascq)
{
    struct iscsi_sense_data *sense;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
    __u8 *buffer =
	(void *) (unsigned long) sg_virt((struct scatterlist *) req->
					 sr_buffer);
#else
    __u8 *buffer =
	scatterlist_address((struct scatterlist *) req->sr_buffer);
#endif

    sense = (struct iscsi_sense_data *) buffer;

    sense->length = cpu_to_be16(14);
    memset(sense->data, 0, 14);
    sense->data[0] = 0xf0;	/* Valid bit set, Response code 0x70 */
    sense->data[2] = sense_key;
    sense->data[7] = 6;		/* Additional sense length */
    sense->data[12] = asc;	/* Additional sense code */
    sense->data[13] = ascq;	/* Additional sense code qualifier */

    req->sr_result = DID_OK << 16;
    return 0;
}

static int __attribute__ ((no_instrument_function))
    do_file_llseek(struct file *dev_file, __u64 lba_start_byte)
{
#ifdef __KERNEL__
    mm_segment_t old_fs;
#endif
    loff_t err;

#ifdef DEBUG_HANDLE_CMD
    printk
	("%s: enter do_file_llseek, lba %llu, dev_file->f_op %p\n",
	 __func__, lba_start_byte, dev_file->f_op);
#endif

    lock_kernel();

#ifdef __KERNEL__
    old_fs = get_fs();
    set_fs(get_ds());
#endif

    if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->llseek)) {
	err = dev_file->f_op->llseek(dev_file, lba_start_byte,
				     0 /*SEEK_SET */ );
    } else {
	err = default_llseek(dev_file, lba_start_byte, 0 /*SEEK_SET */ );
    }

#ifdef __KERNEL__
    set_fs(old_fs);
#endif

    unlock_kernel();

#ifdef DEBUG_HANDLE_CMD
    printk("%s: exit do_file_llseek, err %llu\n", __func__, err);
#endif

    if (err != lba_start_byte) {
	printk("%s handle_cmd: lseek got %llu, expected %llu\n",
	       pr_label, (unsigned long long) err,
	       (unsigned long long) lba_start_byte);
	return -1;
    }
    return 0;
}

static int __attribute__ ((no_instrument_function))
    do_file_read(struct file *dev_file, void *address, __u32 length)
{
#ifdef __KERNEL__
    mm_segment_t old_fs;
#endif
    loff_t err = 0;

#ifdef DEBUG_HANDLE_CMD
    printk
	("%s: enter do_file_read, dev_file->f_op %p, address %p, length "
	 "%u\n", __func__, dev_file->f_op, address, length);
#endif

    if ((dev_file->f_op) && (dev_file->f_op->read)) {

#ifdef __KERNEL__
	old_fs = get_fs();
	set_fs(get_ds());
#endif
	err =
	    dev_file->f_op->read(dev_file, (void *) address,
				 length, &dev_file->f_pos);

#ifdef __KERNEL__
	set_fs(old_fs);
#endif
    } else {

#ifdef DEBUG_HANDLE_CMD
	printk
	    ("Attempt default read for file %p, address %p, "
	     "length %u\n", dev_file, address, length);
#endif

#ifndef __KERNEL__
	err = default_read(dev_file, address, length, &dev_file->f_pos);
#else				/* __KERNEL__ */
	BUG();
#endif				/* __KERNEL__ */

#ifndef __KERNEL__
	if (err < 0)
	    err = -errno;
#endif
    }

#ifdef DEBUG_HANDLE_CMD
    printk("%s: exit do_file_read, result %llu\n", __func__, err);
#endif

    return err;
}


static int __attribute__ ((no_instrument_function))
    do_file_write(struct file *dev_file, void *address, __u32 length)
{
#ifdef __KERNEL__
    mm_segment_t old_fs;
#endif
    loff_t err = 0;

#ifdef DEBUG_HANDLE_CMD
    printk
	("%s: enter do_file_write, dev_file->f_op %p, address %p, length "
	 "%u\n", __func__, dev_file->f_op, address, length);
#endif

    if ((dev_file->f_op) && (dev_file->f_op->write)) {

#ifdef __KERNEL__
	old_fs = get_fs();
	set_fs(get_ds());
#endif
	err = dev_file->f_op->write(dev_file, address, length,
				    &dev_file->f_pos);

#ifdef __KERNEL__
	set_fs(old_fs);
#endif
    } else {

#ifdef DEBUG_HANDLE_CMD
	printk("Attempt default write for file %p, address %p, "
	       "length %u\n", dev_file, address, length);
#endif

#ifndef __KERNEL__
	err = default_write(dev_file, address, length, &dev_file->f_pos);
#else				/* __KERNEL__ */
	BUG();
#endif				/* __KERNEL__ */

#ifndef __KERNEL__
	if (err < 0)
	    err = -errno;
#endif
    }

#ifdef DEBUG_HANDLE_CMD
    printk("%s: exit do_file_write, result %llu\n", __func__, err);
#endif

    return err;
}

static __u64 __attribute__ ((no_instrument_function))
    get_logical_block_address(__u8 * cmd)
{
    switch (cmd[0]) {

    case WRITE_10:		/* in fileio mode */
    case READ_10:		/* in fileio mode */
	return (cmd[LBA_POSN_10] << (BYTE * 3)) +
	    (cmd[LBA_POSN_10 + 1] << (BYTE * 2)) +
	    (cmd[LBA_POSN_10 + 2] << (BYTE)) + cmd[LBA_POSN_10 + 3];

    case READ_6:		/* in fileio mode */
    case WRITE_6:		/* in fileio mode */
	return ((cmd[LBA_POSN_6] & 0x1F) << (BYTE * 2)) +
	    (cmd[LBA_POSN_6 + 1] << (BYTE)) + cmd[LBA_POSN_6 + 2];

    default:
	printk
	    ("get_logical_block_address: Unknown command 0x%02x\n",
	     cmd[0]);
	return 0;

    }				/* switch */
}

/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int handle_cmd(Target_Scsi_Cmnd * cmnd)
{
    __u64 lba_start;
    struct file *dev_file;
    int i;
    int to_read;
    loff_t err = 0;
    struct scatterlist *st_list;

#ifdef DEBUG_HANDLE_CMD
    printk("Entering %s handle_cmd, cmnd=%p, dev_file=%p\n",
	   pr_label, cmnd, cmnd->fd);
#endif

    /* Bjorn Thordarson, 9-May-2004 -- start -- */
    if (cmnd->fd == NULL) {
	if (cmnd->req->sr_cmnd[0] == INQUIRY) {
	    printk
		("INQUIRY received (invalid lun/target), responding with TYPE_NO_LUN\n");

	    if ((err = setup_new_command(cmnd, &to_read)) <= 0)
		return err;

	    /* get response */
	    if (get_inquiry_response(cmnd->req, to_read, TYPE_NO_LUN)) {
		printk
		    ("handle_command: get_inquiry_response returned error "
		     "for %d\n", cmnd->id);
		return -1;
	    }

	    /* change status */
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;
	    return 0;
	} else if (cmnd->req->sr_cmnd[0] == REPORT_LUNS)
	    printk("REPORTS_LUNS received (invalid lun/target)\n");
	else
	    printk
		("LUN or TARGET_ID IS INVALID -> Responding with "
		 "ILLEGAL_REQUEST\n");
	/* set data direction */
	cmnd->req->sr_data_direction = SCSI_DATA_READ;

	/* get length */
	to_read = sizeof(struct iscsi_sense_data) + 14;

	/* allocate space */
	if (get_space(cmnd, to_read, 0)) {
	    return 0;		/* come back later */
	}

	/* get response */
	if (set_up_bad_command_sense_response(cmnd->req,
					      ILLEGAL_REQUEST, 0x25,
					      0x0)) {
	    printk("handle_command: set_up_bad_command_sense_"
		   "response returned error for %d\n", cmnd->id);
	    return -1;
	}

	/* change status */
	cmnd->state = ST_DONE;
	cmnd->req->sr_result = DID_OK << 16;
	return 0;
    }
    /* Bjorn Thordarson, 9-May-2004 -- end -- */

    switch (cmnd->req->sr_cmnd[0]) {
    case READ_CAPACITY:	/* in fileio mode */
	{
#ifdef DEBUG_HANDLE_CMD
	    printk("READ_CAPACITY received\n");
#endif
	    /* perform checks on READ_CAPACITY - LATER */

	    /* set data direction */
	    cmnd->req->sr_data_direction = SCSI_DATA_READ;

	    /* allocate sg_list and get_free_pages */
	    if (get_space(cmnd, READ_CAP_LEN, 0)) {
		err = 0;	/* come back later */
		break;
	    }

	    /* get response */
	    /* Bjorn Thordarson, 9-May-2004 */
			/*****
			if (get_read_capacity_response(cmnd->req)) {
			*****/
	    if (get_read_capacity_response(cmnd)) {
		printk
		    ("handle_command: get_read_capacity_response returned error for %d\n",
		     cmnd->id);
		err = -1;
		break;
	    }

	    /* change status */
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;

	    err = 0;
	    break;
	}

    case INQUIRY:		/* in fileio mode */
	{
#ifdef DEBUG_HANDLE_CMD
	    printk("INQUIRY received\n");
#endif
	    if ((err = setup_new_command(cmnd, &to_read)) <= 0)
		break;

	    /* get response */
	    /* Bjorn Thordarson, 9-May-2004 */
			/*****
			if (get_inquiry_response(cmnd->req, to_read)) {
			*****/
	    if (get_inquiry_response(cmnd->req, to_read, TYPE_DISK)) {
		printk
		    ("handle_command: get_inquiry_response returned error for %d\n",
		     cmnd->id);
		err = -1;
		break;
	    }

	    /* change status */
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;
	    err = 0;
	    break;
	}

    case REPORT_LUNS:		/* in fileio mode */
	{
	    if ((to_read = allocate_report_lun_space(cmnd)) < 0) {
		err = -1;
		break;
	    }

	    /* get response */
	    if ((err = get_report_luns_response(cmnd, to_read))) {
		printk
		    ("handle_command: get_report_luns_response returned error for %d\n",
		     cmnd->id);
	    }
	    break;
	}

	/* cdeng, August 24 2002 */
    case MODE_SENSE:		/* in fileio mode */
	{
#ifdef DEBUG_HANDLE_CMD
	    printk("MODE_SENSE received\n");
#endif
	    if ((err = setup_new_command(cmnd, &to_read)) <= 0)
		break;

	    /* get response */
	    if (get_mode_sense_response(cmnd->req, to_read)) {
		printk
		    ("handle_command: get_mode_sense_response returned error for %d\n",
		     cmnd->id);
		err = -1;
		break;
	    }

	    /* change status */
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;
	    err = 0;
	    break;
	}

	/* cdeng, August 24 2002 */
    case VERIFY:		/* in fileio mode */
	{
#ifdef DEBUG_HANDLE_CMD
	    printk("VERIFY received\n");
#endif
	    /* perform checks on TEST UNIT READY */
	    cmnd->req->sr_data_direction = SCSI_DATA_NONE;
	    cmnd->req->sr_use_sg = 0;
	    cmnd->req->sr_bufflen = 0;
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;
	    err = 0;
	    break;
	}

    case START_STOP:		/* in fileio mode */
    case TEST_UNIT_READY:	/* in fileio mode */
	{
#ifdef DEBUG_HANDLE_CMD
	    printk("TEST UNIT READY received\n");
#endif
	    /* perform checks on TEST UNIT READY */
	    cmnd->req->sr_data_direction = SCSI_DATA_NONE;
	    cmnd->req->sr_use_sg = 0;
	    cmnd->req->sr_bufflen = 0;
	    cmnd->req->sr_result = DID_OK << 16;
	    cmnd->state = ST_DONE;
	    err = 0;
	    break;
	}

    case READ_6:		/* in fileio mode */
    case READ_10:		/* in fileio mode */
	{
	    /* set up the newly-received fileio mode READ */
	    if ((err = setup_new_read(cmnd)) <= 0)
		break;

	    /* data needs to be returned from the file */
	    lba_start = get_logical_block_address(cmnd->req->sr_cmnd);

	    st_list = (struct scatterlist *) cmnd->req->sr_buffer;

	    dev_file = cmnd->fd;
	    if (dev_file) {
		for (i = 0; i < cmnd->req->sr_use_sg; i++) {
		    if (do_file_llseek
			(dev_file, lba_start * BLOCKSIZE) < 0)
			break;

		    err = do_file_read(dev_file,
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
				       (void *) (unsigned long) sg_virt(st_list + i),
				       scatterlist_length(st_list + i)
#else
				       scatterlist_address(&st_list[i]),
				       scatterlist_length(&st_list[i])
#endif
			);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		    if (err < scatterlist_length(st_list + i))
#else
		    if (err < scatterlist_length(&st_list[i]))
#endif
		    {
			if (err < 0) {
			    if (err != -EAGAIN) {
				printk
				    ("%s handle_cmnd: read returned "
				     "%lld\n",
				     pr_label, (unsigned long long) err);
				break;
			    }
			    err = 0;
			}
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			memset((void *)
			       (unsigned long) sg_virt
			       (st_list + i) + err,
			       0, scatterlist_length(st_list + i) - err);
#else
			memset(scatterlist_address
			       (&st_list[i]) + err,
			       0, scatterlist_length(&st_list[i]) - err);
#endif
		    }
//                  if (err > 0)
//                      inode_dir_notify (dev_file->f_dentry->d_parent->d_inode, DN_ACCESS);

		    /*
		     * Assumption is that everything will be
		     * written in blocksize chunks. Also the
		     * blocksize is less than pagesize and
		     * the latter is a multiple of the first
		     */
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		    lba_start +=
			scatterlist_length(st_list + i) / BLOCKSIZE;
#else
		    lba_start += scatterlist_length(&st_list[i])
			/ BLOCKSIZE;
#endif
		}
	    }
	    /* change_status */
	    cmnd->state = ST_DONE;
	    cmnd->req->sr_result = DID_OK << 16;
	    err = 0;
	    break;
	}

    case WRITE_6:		/* in fileio mode */
    case WRITE_10:		/* in fileio mode */
	{
	    if (cmnd->state == ST_NEW_CMND) {
		/* set up the newly-received fileio mode WRITE */
		err = setup_new_write(cmnd);
		break;
	    } else if (cmnd->state == ST_TO_PROCESS) {
		/* data needs to be written to the file */
		lba_start = get_logical_block_address(cmnd->req->sr_cmnd);

		st_list = (struct scatterlist *) cmnd->req->sr_buffer;

		dev_file = cmnd->fd;
		if (dev_file) {
		    for (i = 0; i < cmnd->req->sr_use_sg; i++) {
			if (do_file_llseek
			    (dev_file, lba_start * BLOCKSIZE) < 0)
			    break;

			err = do_file_write(dev_file,
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
					    (void *) (unsigned long) sg_virt(st_list + i),
					    scatterlist_length(st_list + i)
#else
					    scatterlist_address(&st_list [i]),
					    scatterlist_length(&st_list[i])
#endif
			    );

			if ((err < 0)
			    && (err != -EAGAIN)) {
			    printk
				("%s handle_cmnd: write returned %lld\n",
				 pr_label, (unsigned long long) err);
			    break;
			}

			/*
			 * Assumption is that everything will be
			 * written in blocksize chunks. Also the
			 * blocksize is less than pagesize and
			 * the latter is a multiple of the first
			 */
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			lba_start +=
			    scatterlist_length(st_list + i) / BLOCKSIZE;
#else
			lba_start +=
			    scatterlist_length(&st_list[i]) / BLOCKSIZE;
#endif
		    }
		}
		cmnd->state = ST_DONE;
		cmnd->req->sr_result = DID_OK << 16;
	    }
	    err = 0;
	    break;
	}

    default:
	{
	    printk("%s handle_cmd: Unknown command 0x%02x\n",
		   pr_label, cmnd->req->sr_cmnd[0]);
	    break;
	}
    }				/* switch */
    return err;
}

#endif
