/*
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * scsi_genericio_target.c --	Contains functions and declarations for
 * 				GENERICIO scsi emulation mode.
 *
 * Created Fri 23 Feb 2007 11:46:58 AM EST -- Ethan Burns
 */

#include "scsi_target.h"
# ifdef GENERICIO

char device[BYTE];
const char pr_label[] = "GENERICIO";

static int fill_sg_hdr(Target_Scsi_Cmnd *, int);
void signal_process_thread(void *);

/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	mm_segment_t old_fs;
	struct file *dev_file;
	int to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering %s handle_cmd\n", pr_label);
# endif

	if (cmnd->fd)
		dev_file = cmnd->fd;
	else {
		printk("%s handle_cmd: No file descriptor available\n",
		       pr_label);
		return -1;
	}

	switch (cmnd->req->sr_cmnd[0]) {
	case READ_CAPACITY:	/* in genericio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd, READ_CAP_LEN, 0)) {
				err = 0;	/* come back later */
				break;
			}

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk
				    ("%s handle_cmd: fill_sg_hdr returned error\n",
				     pr_label);
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op)
			    && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err = dev_file->f_op->write(dev_file,
#ifdef K26
							    (char __user *)
							    cmnd->sg,
#else
							    (__u8 *) cmnd->
							    sg,
#endif
							    sizeof
							    (sg_io_hdr_t),
							    &dev_file->
							    f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk
					    ("%s handle_cmd: READ_CAPACITY: write returned %d\n",
					     pr_label, err);
					break;
				}
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			err = 0;
			break;
		}

	case INQUIRY:		/* in genericio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			if ((err = setup_new_command(cmnd, &to_read)) <= 0)
				break;

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk
				    ("%s handle_cmd: fill_sg_hdr returned error\n",
				     pr_label);
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op)
			    && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err = dev_file->f_op->write(dev_file,
#ifdef K26
							    (char __user *)
							    cmnd->sg,
#else
							    (__u8 *) cmnd->
							    sg,
#endif
							    sizeof
							    (sg_io_hdr_t),
							    &dev_file->
							    f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk("ioctl returned %d\n", err);
					break;
				}
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			err = 0;
			break;
		}

	case REPORT_LUNS:	/* in genericio mode */
		{
			if ((to_read =
			     allocate_report_lun_space(cmnd)) < 0) {
				err = -1;
				break;
			}

			/* get response */
			if ((err =
			     get_report_luns_response(cmnd, to_read))) {
				printk
				    ("handle_command: get_report_luns_response returned error for %d\n",
				     cmnd->id);
			}
			break;
		}

	case TEST_UNIT_READY:	/* in genericio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;

			if (fill_sg_hdr(cmnd, SG_DXFER_NONE)) {
				printk
				    ("%s handle_cmd: fill_sg_hdr returned error\n",
				     pr_label);
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op)
			    && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err = dev_file->f_op->write(dev_file,
#ifdef K26
							    (char __user *)
							    cmnd->sg,
#else
							    (__u8 *) cmnd->
							    sg,
#endif
							    sizeof
							    (sg_io_hdr_t),
							    &dev_file->
							    f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk
					    ("%s handle_cmd: TEST UNIT READY write returned %d\n",
					     pr_label, err);
					break;
				}
			}

			cmnd->state = ST_PROCESSING;
			err = 0;
			break;
		}

	case READ_6:		/* in genericio mode */
	case READ_10:		/* in genericio mode */
		{
			/* set up the newly-received genericio mode READ */
			if ((err = setup_new_read(cmnd)) <= 0)
				break;

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk
				    ("%s handle_cmd: fill_sg_hdr returned error\n",
				     pr_label);
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op)
			    && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err = dev_file->f_op->write(dev_file,
#ifdef K26
							    (char __user *)
							    cmnd->sg,
#else
							    (__u8 *) cmnd->
							    sg,
#endif
							    sizeof
							    (sg_io_hdr_t),
							    &dev_file->
							    f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk
					    ("%s handle_cmd: READ_CAPACITY: write returned %d\n",
					     pr_label, err);
					break;
				}
			}

			/* change_status */
			cmnd->state = ST_PROCESSING;
			err = 0;
			break;
		}

	case WRITE_6:		/* in genericio mode */
	case WRITE_10:		/* in genericio mode */
		{
			if (cmnd->state == ST_NEW_CMND) {
				/* set up the newly-received genericio mode WRITE */
				err = setup_new_write(cmnd);
				break;
			} else if (cmnd->state == ST_TO_PROCESS) {
				if (fill_sg_hdr(cmnd, SG_DXFER_TO_DEV)) {
					printk
					    ("%s handle_cmd: fill_sg_hdr returned error\n",
					     pr_label);
					return -1;
				}

				/* get response */
				if ((dev_file) && (dev_file->f_op)
				    && (dev_file->f_op-> /*ioctl */ write)) {
					old_fs = get_fs();
					set_fs(get_ds());
//                  err = dev_file->f_op->ioctl (dev_file->f_dentry->d_inode, dev_file, SG_IO, (unsigned long) cmnd->sg);
					err =
					    dev_file->f_op->write(dev_file,
#ifdef K26
								  (char
								   __user
								   *)
								  cmnd->sg,
#else
								  (__u8 *)
								  cmnd->sg,
#endif
								  sizeof
								  (sg_io_hdr_t),
								  &dev_file->
								  f_pos);
					set_fs(old_fs);
					if (err < 0) {
						printk
						    ("%s handle_cmd: WRITE_10: write returned %d\n",
						     pr_label, err);
						break;
					}
				}

				cmnd->state = ST_PROCESSING;
			}
			err = 0;
			break;
		}

	default:
		{
			printk("%s handle_cmd: Unknown command 0x%02x\n",
			       pr_label, cmnd->req->sr_cmnd[0]);
			break;
		}
	}

	return err;
}

/*
 * fill_sg_hdr: This function fills in the values required by the sg
 * struct. The sg_hdr is kmalloced, filled and returned to the calling
 * function.
 * INPUT: Target_Scsi_Cmnd to process, direction of transfer (use the
 * values specified by SCSI generic
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int fill_sg_hdr(Target_Scsi_Cmnd * cmnd, int direction)
{
	int i;
	sg_iovec_t *sg_iov;
	struct scatterlist *st_list;

	cmnd->sg = kmalloc(sizeof(sg_io_hdr_t), GFP_ATOMIC);
	if (!cmnd->sg) {
		printk("fill_sg_hdr: cmnd->sg: kmalloc failed\n");
		return -1;
	}

	cmnd->sg->interface_id = 'S';	// for SCSI generic
	cmnd->sg->dxfer_direction = direction;
	cmnd->sg->cmd_len = cmnd->len;
	cmnd->sg->mx_sb_len = MAX_SENSE_DATA;
	cmnd->sg->iovec_count = cmnd->req->sr_use_sg;
	cmnd->sg->dxfer_len = cmnd->req->sr_bufflen;
//  cmnd->sg->dxferp handled later
	cmnd->sg->cmdp = cmnd->req->sr_cmnd;
	cmnd->sg->sbp = kmalloc(MAX_SENSE_DATA, GFP_ATOMIC);
	cmnd->sg->timeout = TE_TIMEOUT;
	cmnd->sg->flags = 0;

	/* we need the data buffers to get pointed correctly */
	if (cmnd->req->sr_bufflen) {	// there is data

		sg_iov =
		    kmalloc(cmnd->sg->iovec_count * sizeof(sg_iovec_t),
			    GFP_ATOMIC);
		if (!sg_iov) {
			printk("fill_sg_hdr: sg_iov: kmalloc failed\n");
			return -1;
		}

		st_list = (struct scatterlist *) cmnd->req->sr_buffer;

		for (i = 0; i < cmnd->sg->iovec_count; i++) {

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			sg_iov[i].iov_base =
			    (void *) sg_virt(st_list + i);
			sg_iov[i].iov_len = scatterlist_length(st_list + i);
#else
			sg_iov[i].iov_base = st_list[i].address;
			sg_iov[i].iov_len = st_list[i].length;
#endif

# ifdef DEBUG_HANDLE_CMD
//          printk ("sg_iov[%d]: %d\n", i, sg_iov[i].iov_len);
# endif
		}
		cmnd->sg->dxferp = sg_iov;
	} else
		cmnd->sg->dxferp = NULL;

	return 0;
}

# define SETFL_MASK (O_APPEND | O_NONBLOCK | O_NDELAY | FASYNC)

/*
 * signal_process_thread: This thread is responsible for receiving SIGIO
 * signals when SCSI generic is used for processing. This receives the
 * SIGIO signal, changes the state of the oldest command that is in the
 * ST_PROCESSING state and goes back to sleep. I found this to be
 * necessary because signals tend to screw up the processing of the scsi
 * generic commands. Hopefully this will clear things up
 */
void signal_process_thread(void *param)
{
	int i, sigioflag = 0;
	struct file *dev_file;
	Target_Scsi_Cmnd *cmd_curr;
	unsigned long arg;
	unsigned long flags;
	__u32 targ, lun;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	lock_kernel();

	unh_daemonize("signal_thread");

	/* also we want to be able to receive SIGIO */
	siginitsetinv(&current->blocked, IO_SIGS);

	/* mark that this signal thread is alive */
	target_data.signal_id = current;

	unlock_kernel();

	printk("%s Starting pid %d\n", current->comm, current->pid);

	/*
	 * we must now prepare this thread to receive the SIGIO signal
	 * for the opened SCSI device(s)
	 */
	write_lock(&target_map_rwlock);
	for (targ = 0; targ < MAX_TARGETS; targ++) {
		for (lun = 0; lun < MAX_LUNS; lun++) {
			if ((dev_file = target_map[targ][lun].the_file)) {
				arg = dev_file->f_flags | FASYNC;
				lock_kernel();

				/* equal to fcntl (fd, F_SETOWN, getpid()) */
				dev_file->f_owner.pid = current->pid;
				dev_file->f_owner.uid = current->uid;
				dev_file->f_owner.euid = current->euid;

				/* equal to fcntl (fd,F_SETFL,flags|O_ASYNC) */
				if ((arg ^ dev_file->f_flags) & FASYNC) {
					if (dev_file->f_op
					    && dev_file->f_op->fasync) {
						i = dev_file->f_op->
						    fasync(0, dev_file,
							   (arg & FASYNC)
							   != 0);
						if (i < 0) {
							unlock_kernel();
							write_unlock
							    (&target_map_rwlock);
							printk
							    ("si..thread: fasync returned %d\n",
							     i);
							return;
						}
					}
				}

				/* required for strict SunOS emulation */
				/*
				   if (O_NONBLOCK != O_NDELAY)
				   if (arg & O_NDELAY)
				   arg |= O_NONBLOCK;
				 */
				dev_file->f_flags =
				    (arg & SETFL_MASK) | (dev_file->
							  f_flags &
							  ~SETFL_MASK);
				unlock_kernel();
			}
		}
	}
	write_unlock(&target_map_rwlock);

	while (1) {
		down_interruptible(&target_data.sig_thr_sem);

		sigioflag = 0;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 18)
		if (signal_pending(current)) {
			spin_lock_irq(&current->sigmask_lock);
			if (sigismember(&current->pending.signal, SIGIO)) {
# ifdef DEBUG_SCSI_THREAD
				printk("SIGIO received\n");
# endif
				sigioflag = 1;
				sigdelset(&current->pending.signal, SIGIO);
			}
			recalc_sigpending(current);
			spin_unlock_irq(&current->sigmask_lock);
			if (!sigioflag) {
				printk("si...thread: time to die\n");
				break;
			}
		}
#else
		if (signal_pending(current)) {
			spin_lock_irq(&current->sighand->siglock);
			if (sigtestsetmask
			    (&current->pending.signal, sigmask(SIGIO))) {
# ifdef DEBUG_SCSI_THREAD
				printk("SIGIO received\n");
# endif
				sigioflag = 1;
				sigdelsetmask(&current->pending.signal,
					      sigmask(SIGIO));
			}
			recalc_sigpending();
			spin_unlock_irq(&current->sighand->siglock);
			if (!sigioflag) {
				printk("si...thread: time to die\n");
				break;
			}
		}
#endif

		/* change the state of all PROCESSING commands */
		spin_lock_irqsave(&target_data.cmd_queue_lock, flags);
		list_for_each_entry(cmd_curr, &target_data.cmd_queue, link) {
			if (cmd_curr->state == ST_PROCESSING) {
				cmd_curr->state = ST_PROCESSED;
				/* wake up scsi_target_process_thread */
				if (atomic_read
				    (&target_data.target_sem.count) <= 0) {
					up(&target_data.target_sem);
				}
			}
		}
		spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);
	}
	up(&target_data.signal_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
}

# endif
/* vi: set tabstop=8 textwidth=78: */
