/*
 * Copyright (C) 2001-2007 InterOperability Lab (IOL)
 * University of New Hampshire (UNH)
 * Durham, NH 03824
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * The name of IOL and/or UNH may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 *
 * target/scsi_memoryio_target.c --	Contains functions and declarations
 * 					specific to the MEMORYIO target scsi
 * 					emulation layer.
 *
 * Created Fri 23 Feb 2007 11:13:27 AM EST -- Ethan Burns
 */

#include "scsi_target.h"
#if defined(MEMORYIO)

struct scatterlist memory_st_list;
char device[BYTE * 5];
const char pr_label[] = "MEMORYIO";

/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	int to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering %s handle_cmd\n", pr_label);
# endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif
	switch (cmnd->req->sr_cmnd[0]) {

	case READ_CAPACITY:	/* in memoryio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd, READ_CAP_LEN, 0)) {
				err = 0;	/* come back later */
				break;
			}

			/* get response */
			/* Bjorn Thordarson, 9-May-2004 */
			/*****
			if (get_read_capacity_response(cmnd->req)) {
			*****/
			if (get_read_capacity_response(cmnd)) {
				printk
				    ("handle_command: get_read_capacity_response returned error for %d\n",
				     cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;

			err = 0;
			break;
		}

	case INQUIRY:		/* in memoryio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			if ((err = setup_new_command(cmnd, &to_read)) <= 0) {
				err = -1;
				break;
			}

			/* get response */
			/* Bjorn Thordarson, 9-May-2004 */
			/*****
			if (get_inquiry_response(cmnd->req, to_read)) {
			*****/
			if (get_inquiry_response
			    (cmnd->req, to_read, TYPE_DISK)) {
				printk
				    ("handle_command: get_inquiry_response returned error for %d\n",
				     cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;

			err = 0;
			break;
		}

	case TEST_UNIT_READY:	/* in memoryio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->req->sr_result = DID_OK << 16;
			cmnd->state = ST_DONE;
			err = 0;
			break;
		}

	case REPORT_LUNS:	/* in memoryio mode */
		{
			if ((to_read =
			     allocate_report_lun_space(cmnd)) < 0) {
				err = -1;
				break;
			}

			/* get response */
			if ((err =
			     get_report_luns_response(cmnd, to_read))) {
				printk
				    ("handle_command: get_report_luns_response returned error for %d\n",
				     cmnd->id);
			}
			break;
		}

	case MODE_SENSE:	/* in memoryio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("MODE_SENSE received\n");
# endif
			if ((err = setup_new_command(cmnd, &to_read)) <= 0) {
				err = -1;
				break;
			}

			/* get response */
			if (get_mode_sense_response(cmnd->req, to_read)) {
				printk
				    ("handle_command: get_mode_sense_response returned error for %d\n",
				     cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case VERIFY:		/* in memoryio mode */
		{
# ifdef DEBUG_HANDLE_CMD
			printk("VERIFY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case READ_6:		/* in memoryio mode */
	case READ_10:		/* in memoryio mode */
		{
			/* set up the newly-received memoryio mode READ */
			if ((err = setup_new_read(cmnd)) <= 0)
				break;

			/* this data can just be returned from memory */
			/* change_status */
			cmnd->req->sr_result = DID_OK << 16;
			cmnd->state = ST_DONE;
			err = 0;
			break;
		}

	case WRITE_6:		/* in memoryio mode */
	case WRITE_10:		/* in memoryio mode */
		{
			if (cmnd->state == ST_NEW_CMND) {
				/* set up the newly-received memoryio mode WRITE */
				err = setup_new_write(cmnd);
				break;
			} else if (cmnd->state == ST_TO_PROCESS) {
				/*
				 * in memory mode of processing we do not
				 * care about the data received at all
				 */
# ifdef DEBUG_SCSI_THREAD
				printk
				    ("%s in %s handle_cmd, command %p id %d change "
				     "state from %d to %d\n",
				     current->comm, pr_label, cmnd,
				     cmnd->id, cmnd->state, ST_DONE);
# endif
				cmnd->state = ST_DONE;
				cmnd->req->sr_result = DID_OK << 16;
			}
			err = 0;
			break;
		}

	default:
		{
			printk("%s handle_cmd: Unknown command 0x%02x\n",
			       pr_label, cmnd->req->sr_cmnd[0]);
			break;
		}
	}			/* switch */
	return err;
}

#endif				/* MEMORYIO */
/* vi: set tabstop=8 textwidth=78: */
