/*	target/scsi_midlevel.c -- Contains functions that are called by the SCSI
 *	Mid-Level
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 textwidth=78 :
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 * Created Thu Mar  1 10:51:44 2007 -- Ethan Burns
 */

#include "scsi_midlevel.h"
#include "iscsi_target.h"
#include "iscsi_cmnd.h"
#include "target_error_rec.h"

#include "../common/target_negotiate.h"

/*
 * Static declarations
 */
static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_iscsi_cmnd(Target_Scsi_Cmnd * cmnd, unsigned long *flags,
		  struct iscsi_session **result_sess);

static struct iscsi_cmnd *search_task_mgt_command(Target_Scsi_Message *
						  message);

static int init_target(void);
static void bring_down_portals(void);
static void release_portals(void);
static int start_server_thread(char *ip_string, char *port_string,
			       int tag);
static int stop_server_thread(char *ip_string, char *port_string);
static void free_sock_data(int i);
static int iscsi_proc_info_read(char *buffer, char **start, off_t offset,
				int length, int inode);
static int iscsi_proc_info_write(char *buffer, char **start, off_t offset,
				 int length, int inode);
static void iscsi_proc_info_targ_manage(char *buffer, char **start,
					off_t offset, int length,
					int inode);
static int iscsi_proc_info_targ_ip(char *buffer, char **start,
				   off_t offset, int length, int inode);

/*
 * iscsi_detect: this function sets up the server so that it can accept
 * multiple connections.
 */
int iscsi_detect(Scsi_Target_Template * tmpt)
{
	long i;
	struct portal_group *ptr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Entering iSCSI detect tmpt %p\n", tmpt);

	devdata = my_kmalloc(sizeof(struct iscsi_global), "devdata");

	if (!devdata) {
		TRACE(TRACE_ENTER_LEAVE, "Leaving iSCSI detect tmpt %p\n",
		      tmpt);
		return -1;
	}

	TRACE(TRACE_DEBUG, "devdata %p\n", devdata);

	/* Initialize devdata */
	if (init_target() < 0) {
		TRACE_ERROR("%s Initialize devdata failed\n",
			    current->comm);
		my_kfree((void **) &devdata, "devdata");
		TRACE(TRACE_ENTER_LEAVE, "Leaving iSCSI detect tmpt %p\n",
		      tmpt);
		return -1;
	}

	/* setup the security key hash table */
	setup_security_hash_table();

	TRACE(TRACE_DEBUG,
	      "iSCSI initialization complete tmpt %p devdata %p "
	      "(size %d)\n", tmpt, devdata, (int) sizeof(*devdata));

	for (i = 0, ptr = iscsi_portal_groups;
	     i < MAX_PORTAL && ptr->tag != 0; i++, ptr++) {
		ptr->in_use = 1;
#ifdef RDMA_MODE
		ptr->src_addr = ptr->dst_addr = NULL;
		TRACE(TRACE_DEBUG, "Initializing iser_conn_sem: locked\n");
		init_MUTEX_LOCKED(&ptr->iser_conn_sem);
#endif

		if (bring_up_portal(ptr)) {
			TRACE(TRACE_DEBUG, "bring_up_portal failed\n");
			ptr->in_use = 0;
			goto out;
		}
	}

	/* register the front end now */
	devdata->device = register_target_front_end(tmpt);

	if (!devdata->device) {
		TRACE_ERROR("%s Device registration failed\n",
			    current->comm);
		goto out;
	}

	devdata->device->dev_specific = (void *) devdata;

	TRACE(TRACE_DEBUG,
	      "Device registration complete tmpt %p devdata %p\n", tmpt,
	      devdata);
#ifdef RDMA_MODE
//      mod_pri_for_print();
#endif
	TRACE(TRACE_ENTER_LEAVE, "Leaving iSCSI detect tmpt %p\n", tmpt);
	return 1;		/* one device detected */

      out:
	/* due to an error, shut down everything brought up so far */
	bring_down_portals();
	release_portals();

	TRACE(TRACE_ENTER_LEAVE, "Leaving iSCSI detect tmpt %p\n", tmpt);
	return -2;
}

/*
 * iscsi_release: function to release the iSCSI device as required by
 * the Mid-Level
 */
int iscsi_release(Scsi_Target_Device * device)
{
	struct iscsi_session *session;
	struct list_head *list_ptr, *list_temp;
	int err = 0;
	DECLARE_LOOP_COUNTER(count);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_release\n");

	if ((struct iscsi_global *) device->dev_specific != devdata) {
		TRACE_ERROR("%s This is not an iscsi device\n",
			    current->comm);
		err = -1;
		goto out;
	}

	/* stop accepting all new connections so no new sessions can be formed
	 */
	bring_down_portals();

	/*
	 * destructive access to session lists in order to terminate all
	 * sessions
	 */
	write_lock(&devdata->session_rwlock);
	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp, &devdata->session_list) {
		session =
		    list_entry(list_ptr, struct iscsi_session, sess_link);
		BUMP_LOOP_COUNTER(count, "session_list");
		iscsi_release_session(session);
	}
	INIT_LOOP_COUNTER(count);
	list_for_each_safe(list_ptr, list_temp, &devdata->bad_session_list) {
		session =
		    list_entry(list_ptr, struct iscsi_session, sess_link);
		BUMP_LOOP_COUNTER(count, "bad_session_list");
		TRACE(TRACE_DEBUG,
		      "iscsi%d: release bad session %p, tsih %u\n",
		      (int) devdata->device->dev_id, session,
		      session->tsih);
		iscsi_release_session(session);
	}
	write_unlock(&devdata->session_rwlock);

	release_portals();

      out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_release, err %d\n", err);

	return err;
}

/*
 * executed only by midlevel's target_process_thread as a call-back.
 * iscsi_xmit_response: responsible for transmitting data and scsi
 * response
 */
int iscsi_xmit_response(Target_Scsi_Cmnd * cmnd)
{
	struct iscsi_cmnd *cmd;
	struct iscsi_session *session;
	unsigned long flags;
	int retval = 0;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef K26
	TRACE(TRACE_ENTER_LEAVE, "%s enter iscsi_xmit_response with irqs "
	      "%sabled\n", current->comm, irqs_disabled()? "dis" : "en");
#else
	TRACE(TRACE_ENTER_LEAVE, "%s enter iscsi_xmit_response\n",
	      current->comm);
#endif

	cmd = search_iscsi_cmnd(cmnd, &flags, &session);

#ifdef K26
	TRACE(TRACE_ENTER_LEAVE,
	      "%s in iscsi_xmit_response after search with "
	      "irqs %sabled\n", current->comm,
	      irqs_disabled()? "dis" : "en");
#endif

	if (!cmd) {
		TRACE_ERROR
		    ("%s iscsi_xmit_response could not find cmnd %d\n",
		     current->comm, cmnd->id);
		retval = -1;
		goto out;
	}

	/* bump next expected cmd sequence number if necessary */
	session->exp_cmd_sn += cmd->cmd_sn_increment;
	cmd->cmd_sn_increment = 0;

	cmd->state = ISCSI_DONE;

	TRACE(TRACE_ISCSI_FULL,
	      "%lu %s Done CmdSN %u ITT %u ExpCmdSN %u\n", jiffies,
	      current->comm, cmd->cmd_sn, cmd->init_task_tag,
	      session->exp_cmd_sn);
	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	/* wake up the transmit thread */
	if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s cmnd %p state %d up on tx_sem in "
		      "iscsi_xmit_response\n", jiffies, current->comm,
		      cmd, cmd->state);
		up(&cmd->conn->tx_sem);
	}

      out:

#ifdef K26
	TRACE(TRACE_ENTER_LEAVE,
	      "%s exit iscsi_xmit_response retval %d with "
	      "irqs %sabled\n", current->comm, retval,
	      irqs_disabled()? "dis" : "en");
#else
	TRACE(TRACE_ENTER_LEAVE, "%s exit iscsi_xmit_response retval %d\n",
	      current->comm, retval);
#endif

	return retval;
}

/*
 * executed only by midlevel's target_process_thread as a call-back.
 * iscsi_rdy_to_xfer: function changes the state of the command
 * and wakes up the tx thread to send an R2T to the Initiator if needed
 * Returns 0 if all ok, -1 on error (couldn't find matching command).
 */
int iscsi_rdy_to_xfer(Target_Scsi_Cmnd * cmnd)
{
	struct iscsi_cmnd *cmd;
	struct iscsi_session *session;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	cmd = search_iscsi_cmnd(cmnd, &flags, &session);
	/* on successful return, the session->cmnd_lock is held */

	if (!cmd) {
		TRACE_ERROR
		    ("%s iscsi_rdy_to_xfer could not find cmnd %d\n",
		     current->comm, cmnd->id);
		return -1;
	}

#if defined(DEBUG_SCSI_THREAD)
	printk("%s: cmd=%p, cmd_sn=%u\n", __func__, cmd, cmd->cmd_sn);
#endif /* DEBUG_SCSI_THREAD */

	/*
	 * now that this command has been successfully accepted by the
	 * midlevel, bump the session-wide expected cmd sequence number as
	 * appropriate to allow the next command in this session to be
	 * delivered to midlevel.
	 */
	session->exp_cmd_sn += cmd->cmd_sn_increment;
	cmd->cmd_sn_increment = 0;

	/* allow r2t generation (if any) for this command */
	cmd->r2t_data_left = cmd->r2t_data_total;

	TRACE(TRACE_ISCSI, "%lu %s Buffers ready CmdSN %u ITT %u DSL %u\n",
	      jiffies, current->comm, cmd->cmd_sn, cmd->init_task_tag,
	      cmd->data_length);
	if (cmd->data_length == 0)
		/* this command expects no data from initiator, so it is
		 * done */
		cmd->state = ISCSI_DONE;
	else if (cmd->state == ISCSI_AWAIT_BUFFER_QUEUE)
		/* this command was queued, now it is queued and ready to
		 * receive */
		cmd->state = ISCSI_QUEUE_CMND_RDY;
	else {
		/* this command is ready to receive data from initiator */
		if (cmd->state != ISCSI_AWAIT_BUFFER) {
			TRACE_ERROR
			    ("%s iscsi_rdy_to_xfer for CmdSN %u ITT %u "
			     "opcode 0x%02x expected state %u got "
			     "state %u setting state %u\n", current->comm,
			     cmd->cmd_sn, cmd->init_task_tag,
			     cmd->opcode_byte, ISCSI_AWAIT_BUFFER,
			     cmd->state, ISCSI_BUFFER_RDY);
		}
		cmd->state = ISCSI_BUFFER_RDY;
	}

	spin_unlock_irqrestore(&session->cmnd_lock, flags);

	if (atomic_read(&cmd->conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "iscsi_rdy_to_xfer\n", jiffies, current->comm);
		up(&cmd->conn->tx_sem);
	}

	return 0;
}

/*
 * called when a file in the /proc/scsi_target/iscsi/ directory is read from
 * or written to.
 *
 * param inout	1 for writing
 * 		0 for reading
 */
int iscsi_proc_info(char *buffer, char **start, off_t offset,
		    int length, int inode, int inout)
{
	int result;
	TRACE(TRACE_DEBUG, "Entering iscsi_proc_info for %s length %d\n",
	      inout ? "write" : "read", length);

	if (inout == 1 /* 1 is write */ ) {
		result = iscsi_proc_info_write(buffer, start, offset,
					       length, inode);
	} else {
		/*
		 * write to this file, buffer should be "iscsi_initiator ip
		 * 1234abcd target 6 lun 0"
		 */
		if (buffer == NULL || length > PAGE_SIZE)
			result = -EINVAL;
		else
			result =
			    iscsi_proc_info_read(buffer, start, offset,
						 length, inode);
	}

	TRACE(TRACE_DEBUG, "Leaving iscsi_proc_info result %d\n", result);
	return result;
}

/*
 * iscsi_task_mgt_fn_done:
 * this is a call-back function from the target midlevel thread
 * after it has finished processing a task management request
 *
 */
void iscsi_task_mgt_fn_done(Target_Scsi_Message * msg)
{
	struct iscsi_cmnd *related_command;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	related_command = search_task_mgt_command(msg);

	if (!related_command) {
		TRACE_ERROR
		    ("%s Could not find message with fn %d, cmnd %p\n",
		     current->comm, msg->fn, msg->value);
		return;
	}

	/* wake up the transmit thread */
	if (atomic_read(&related_command->conn->tx_sem.count) <= 0) {
		TRACE(TRACE_SEM, "%lu %s up on tx_sem in "
		      "iscsi_task_mgt_fn_done\n", jiffies, current->comm);
		up(&related_command->conn->tx_sem);
	}
}

/*
 * Static function definitions.
 */

/*
 * executed only by midlevel's target_process_thread as part of a call-back
 * via iscsi_xmit_response() or iscsi_rdy_to_xfer().
 * search_iscsi_cmnd: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given command.
 * INPUT: Target_Scsi_Cmnd
 * OUTPUT: pointer to struct iscsi_cmnd if one is found, NULL otherwise
 *	   if return value is not NULL, result_sess has been filled in with
 *         the session, and the session->cmnd_lock is held
 */
static struct iscsi_cmnd * __attribute__ ((no_instrument_function))
    search_iscsi_cmnd(Target_Scsi_Cmnd * cmnd, unsigned long *flags,
		  struct iscsi_session **result_sess)
{
	struct iscsi_cmnd *cmd;
	struct iscsi_session *session;
	struct iscsi_global *host;
	DECLARE_LOOP_COUNTER(count);
	DECLARE_LOOP_COUNTER(count2);

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!cmnd) {
		TRACE_ERROR("%s Cannot search for a NULL command\n",
			    current->comm);
		return NULL;
	}

	host = (struct iscsi_global *) cmnd->device->dev_specific;

	/* non-destructive access to session lists */
	read_lock(&host->session_rwlock);

	INIT_LOOP_COUNTER(count);
	list_for_each_entry(session, &host->session_list, sess_link) {
		BUMP_LOOP_COUNTER(count, "session_list");
		spin_lock_irqsave(&session->cmnd_lock, *flags);
		INIT_LOOP_COUNTER(count2);
		list_for_each_entry(cmd, &session->active_command_list,
				    command_link) {
			BUMP_LOOP_COUNTER(count2, "active_command_list");
			if (cmd->cmnd == cmnd) {
				*result_sess = session;

				/* return non-NULL cmd */
				goto search_iscsi_cmnd_out;
			}
		}
		spin_unlock_irqrestore(&session->cmnd_lock, *flags);
	}
	cmd = NULL;		/* return NULL for not-found */

      search_iscsi_cmnd_out:
	read_unlock(&host->session_rwlock);
	return cmd;
}

/*
 * this is called only from iscsi_task_mgt_fn_done() on a call-back from
 * target midlevel thread.
 * search_iscsi_message: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given message.
 * INPUT: Target_Scsi_Message
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static struct iscsi_cmnd *search_task_mgt_command(Target_Scsi_Message *
						  message)
{
	struct iscsi_cmnd *related_task_mgt_command = NULL;
	struct iscsi_session *related_session;
	struct iscsi_global *host;
	DECLARE_LOOP_COUNTER(count);
	DECLARE_LOOP_COUNTER(count2);
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!message) {
		TRACE_ERROR("%s Cannot search for a NULL TM message\n",
			    current->comm);
		return NULL;
	}

	if (message->device == NULL
	    || (host =
		(struct iscsi_global *) message->device->dev_specific)
	    == NULL) {
		TRACE_ERROR("%s Empty fields in TM message\n",
			    current->comm);
		return NULL;
	}

	/* non-destructive access to session lists */
	read_lock(&host->session_rwlock);

	INIT_LOOP_COUNTER(count);
	list_for_each_entry(related_session, &host->session_list,
			    sess_link) {
		BUMP_LOOP_COUNTER(count, "session_list");
		spin_lock_irqsave(&related_session->cmnd_lock, flags);

		INIT_LOOP_COUNTER(count2);
		list_for_each_entry(related_task_mgt_command,
				    &related_session->active_command_list,
				    command_link) {
			BUMP_LOOP_COUNTER(count2, "active_command_list");
			if (related_task_mgt_command->message == message) {
				/* found the TM command, check the command it
				 * refers to */
				if (related_task_mgt_command->
				    ref_command == NULL
				    || related_task_mgt_command->
				    ref_command->cmnd != message->value) {
					/* should never happen */
					TRACE_ERROR
					    ("%s TM message value does "
					     "not match\n", current->comm);
				}
				related_task_mgt_command->state =
				    ISCSI_MGT_FN_DONE;
				spin_unlock_irqrestore(&related_session->
						       cmnd_lock, flags);
				goto out;
			}
		}
		related_task_mgt_command = NULL;
		spin_unlock_irqrestore(&related_session->cmnd_lock, flags);
	}

      out:
	read_unlock(&host->session_rwlock);

	return related_task_mgt_command;
}

/*
 * initialize stuff in devdata.
 */
static int init_target(void)
{
	memset(devdata, 0, sizeof(struct iscsi_global));

	devdata->max_queue_depth_allowed = QUEUE_DEPTH_ALLOWED;

	INIT_LIST_HEAD(&devdata->session_list);
	INIT_LIST_HEAD(&devdata->bad_session_list);

	rwlock_init(&devdata->session_rwlock);
	init_MUTEX_LOCKED(&devdata->server_sem);

	devdata->param_tbl = my_kmalloc(MAX_CONFIG_PARAMS
					* sizeof(struct parameter_type),
					"param_tbl");
	if (!(devdata->param_tbl)) {
		return -1;
	}

	/* Copy the default parameters */
	param_tbl_init(*devdata->param_tbl);

	/* chap and srp support - CHONG */
	devdata->auth_parameter.chap_local_ctx = CHAP_InitializeContext();
	devdata->auth_parameter.chap_peer_ctx = CHAP_InitializeContext();
	devdata->auth_parameter.srp_ctx = SRP_InitializeContext();

	return 0;
}

/* signal all the server threads to terminate, which stops them from
 * listening for new connections from initiators
 */
static void bring_down_portals(void)
{
	int i;
	struct portal_group *ptr;

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL; i++, ptr++) {
		if (ptr->in_use == 0)
			continue;

		if (devdata->server_thr[i]) {
			/*  Mike Christie mikenc@us.ibm.com */
			TRACE(TRACE_DEBUG, "Sending kill to server thread %d\n", i);
			send_sig(ISCSI_SHUTDOWN_SIGNAL, devdata->server_thr[i], 1);
#if defined(__KERNEL__)
			down_interruptible(&devdata->server_sem);
#else
			pthread_join(devdata->server_thr[i]->threadid, NULL);
#endif /* __KERNEL__ */
		}
	}
}

/* contributed by Charles Coffing to avoid a race during shutdown
 * this releases all listening sockets (threads listening on them
 * already killed)
 */
static void release_portals(void)
{
	int i;
	struct portal_group *ptr;
#ifdef RDMA_MODE
	iser_status_t status;
#endif

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL; i++, ptr++) {
		if (ptr->in_use == 0)
			continue;

#ifdef RDMA_MODE
		status = iser_api_unregister(ptr->iser_handle);
		if (status != ISER_SUCCESS) {
			TRACE_ERROR("iser_api_unregister failed %d\n",
				    status);
			/* Continue with the release even in error */
		}
#endif
		free_sock_data(i);
	}

	if (devdata->param_tbl) {
		param_tbl_uncpy(*devdata->param_tbl);
		my_kfree((void **) &devdata->param_tbl, "param_tbl");
	}

	/* chap and srp support - CHONG */
	CHAP_FinalizeContext(devdata->auth_parameter.chap_local_ctx);
	CHAP_FinalizeContext(devdata->auth_parameter.chap_peer_ctx);
	SRP_FinalizeContext(devdata->auth_parameter.srp_ctx);

	my_kfree((void **) &devdata, "devdata");

	printk("max number of iovs in sendmsg %3d\n", max_send_niov);
	printk("max number of iovs in recvmsg %3d\n", max_recv_niov);

	/* check for any unfreed memory */
	my_kempty();
}

static int start_server_thread(char *ip_string, char *port_string, int tag)
{
	int i;
	struct portal_group *ptr;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	if (!devdata) {
		TRACE_ERROR("%s No device available\n", current->comm);
		goto out;
	}

	for (i = 0, ptr = iscsi_portal_groups; i < MAX_PORTAL; i++, ptr++) {
		if (!ptr->in_use)
			break;
	}

	if (i >= MAX_PORTAL) {
		TRACE_ERROR("%s No more portals permitted\n",
			    current->comm);
		goto out;
	}

	/* mark this portal in use with dynamic ip strings */
	ptr->in_use = 2;

	ptr->ip_string = my_kmalloc(strlen(ip_string) + 1, "ip_string");
	if (ptr->ip_string == NULL) {
		goto out0;
	}
	strcpy(ptr->ip_string, ip_string);

	ptr->port_string =
	    my_kmalloc(strlen(port_string) + 1, "port_string");
	if (ptr->port_string == NULL) {
		goto out1;
	}
	strcpy(ptr->port_string, port_string);

	ptr->tag = tag;
#ifdef RDMA_MODE
	ptr->dst_addr = ptr->src_addr = NULL;
#endif
	if (bring_up_portal(ptr)) {
		goto out2;
	}

	return 0;

      out2:
	my_kfree((void **) &ptr->port_string, "port_string");
      out1:
	my_kfree((void **) &ptr->ip_string, "ip_string");
      out0:
	ptr->in_use = 0;
      out:
	return -1;
}

static int stop_server_thread(char *ip_string, char *port_string)
{
	int i;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	for (i = 0; i < MAX_PORTAL; i++) {
		if (iscsi_portal_groups[i].in_use == 0)
			continue;

		if ((strcmp(iscsi_portal_groups[i].ip_string, ip_string) ==
		     0)
		    &&
		    (strcmp
		     (iscsi_portal_groups[i].port_string,
		      port_string) == 0)) {
			if (devdata->server_thr[i]) {
				/*  Mike Christie mikenc@us.ibm.com */
				TRACE(TRACE_DEBUG, "Sending kill to server thread %d\n", i);
				send_sig(ISCSI_SHUTDOWN_SIGNAL, devdata->server_thr[i], 1);
#if defined(__KERNEL__)
				down_interruptible(&devdata->server_sem);
#else
				pthread_join(devdata->server_thr[i]->threadid, NULL);
#endif /* __KERNEL__ */
			}
			free_sock_data(i);
			return 0;
		}
	}

	TRACE_ERROR("%s Can't find match with ip %s, port %s\n",
		    current->comm, ip_string, port_string);
	return -1;
}

static void free_sock_data(int i)
{
	struct portal_group *ptr = iscsi_portal_groups + i;

#ifndef RDMA_MODE
	if (devdata->server_socket[i]) {
		TRACE(TRACE_DEBUG, "iscsi_server releasing socket %d\n",
		      i);
		sock_release(devdata->server_socket[i]);
		devdata->server_socket[i] = NULL;
	}
#endif

	my_kfree((void **) &ptr->ip_address, "ip_address");
	if (ptr->in_use == 2) {
		my_kfree((void **) &ptr->ip_string, "ip_string");
		my_kfree((void **) &ptr->port_string, "port_string");
	}
	ptr->in_use = 0;
}

/*
 * Called by proc filesystem read operations
 */
static int iscsi_proc_info_read(char *buffer, char **start, off_t offset,
				int length, int inode)
{
	int i;
	int result;
	__u32 count;
	struct list_head *list_ptr;
	struct iscsi_session *session;
	struct iscsi_conn *conn;
	char ip_string[INET6_ADDRSTRLEN + 2], port_string[8];

	result = sprintf(buffer, "UNH-IOL iSCSI Target ");
#ifdef MEMORYIO
	result += sprintf(buffer + result, "in Memory Mode ");
#else
#ifdef DISKIO
	result += sprintf(buffer + result, "in Disk Mode ");
#else
#ifdef FILEIO
	result += sprintf(buffer + result, "in File Mode ");
#else
	result += sprintf(buffer + result, "in Generic Mode ");
#endif
#endif
#endif

	result +=
	    sprintf(buffer + result, "with BLOCKSIZE %d\n", BLOCKSIZE);

	result += sprintf(buffer + result, "Total of %d target devices\n",
			  target_count);

	count = 0;
	for (i = 0; i < MAX_PORTAL; i++) {
		if (iscsi_portal_groups[i].in_use)
			count++;
	}
	result += sprintf(buffer + result,
			  "Total of %u server threads\n", count);

	for (i = 0; i < MAX_PORTAL; i++) {
		if (!iscsi_portal_groups[i].in_use)
			continue;

		result += sprintf(buffer + result,
				  "    server_thread_%d at %s:%s,%d\n",
				  i, iscsi_portal_groups[i].ip_string,
				  iscsi_portal_groups[i].port_string,
				  iscsi_portal_groups[i].tag);
	}

	result += sprintf(buffer + result,
			  "max number of iovs in sendmsg %3d\n",
			  max_send_niov);
	result += sprintf(buffer + result,
			  "max number of iovs in recvmsg %3d\n",
			  max_recv_niov);

	/* clean bad sessions to get current information */
	clean_bad_stuff();

	/* non-destructive access to session lists */
	read_lock(&devdata->session_rwlock);

	count = 0;
	list_for_each(list_ptr, &devdata->session_list) {
		count++;
		TEST_LOOP_COUNTER(count, "session_list");
	}
	result +=
	    sprintf(buffer + result, "Total of %u sessions\n", count);
	if (count == 0) {
		result += print_config_info(*devdata->param_tbl,
					    buffer + result);
	}
	INIT_LOOP_COUNTER(count);
	list_for_each_entry(session, &devdata->session_list, sess_link) {
		BUMP_LOOP_COUNTER(count, "session_list");
		result += sprintf(buffer + result,
				  "    session for target %u is %p\n",
				  session->oper_param->TargetName,
				  session);
		result += sprintf(buffer + result,
				  "    session has %u connections\n",
				  session->nconn);
		INIT_LOOP_COUNTER(count2);
		list_for_each_entry(conn, &session->conn_list, conn_link) {
			BUMP_LOOP_COUNTER(count2, "conn_list");
			result += sprintf(buffer + result,
					  "        connection with CID "
					  "%u is %p on socket %p\n",
					  conn->cid, conn,
					  conn->conn_socket);
			if (cnv_inet_to_string(conn->ip_address,
					       ip_string,
					       port_string) > 0) {
				result +=
				    sprintf(buffer + result,
					    "%30s  %s:%s\n",
					    "Initiator Address", ip_string,
					    port_string);
			}
			if (cnv_inet_to_string(conn->local_ip_address,
					       ip_string,
					       port_string) > 0) {
				result +=
				    sprintf(buffer + result,
					    "%30s  %s:%s\n",
					    "Local Address", ip_string,
					    port_string);
			}
		}
		/* if target has lots of sessions, printing the session
		 * parameters for all of them here may overflow the proc 4K
		 * buffer.
		 */
		result += print_session_params(session->oper_param,
					       *session->session_params,
					       buffer + result);
		result += sprintf(buffer + result, "\n");
	}

	read_unlock(&devdata->session_rwlock);

	return result;
}

/*
 * Called by proc file system write operations
 */
static int iscsi_proc_info_write(char *buffer, char **start, off_t offset,
				 int length, int inode)
{
	int result;

	result = length;

	if (length == PAGE_SIZE)
		length--;

	/* make the buffer into a string */
	buffer[length] = '\0';
	TRACE(TRACE_DEBUG, "iscsi_proc_info string \"%s\"\n", buffer);

	if (length >= 16 && strncmp(buffer, "iscsi_target ip ", 16) == 0) {
		result = iscsi_proc_info_targ_ip(buffer, start, offset,
						 length, inode);
	} else if (length >= 20
		   && strncmp(buffer, "iscsi_target manage ", 20) == 0) {
		iscsi_proc_info_targ_manage(buffer, start, offset,
					    length, inode);
	} else {
		result = -EINVAL;
	}

	return result;
}

/*
 * Called for writing to the proc FS with the "iscsi_target manage" command
 */
static void iscsi_proc_info_targ_manage(char *buffer, char **start,
					off_t offset, int length,
					int inode)
{
	int param_neg_info;
	int int_val;
	char *ptr_to_keytext;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	/* this is for configuring the login
	 * parameters.  This buffer has expected
	 * interface, scan neg_info flags */
	param_neg_info = simple_strtoul(buffer + 20, &ptr_to_keytext, 0);

	/* Move a position to ignore space */
	ptr_to_keytext++;

	/* Now configure the parameters */
	if (param_neg_info == FORCE) {
		/* special to force login subphases */
		if (strcmp(ptr_to_keytext, "s") == 0) {
			devdata->force |= USE_SECURITY_PHASE;
		} else if (strcmp(ptr_to_keytext, "o")
			   == 0) {
			devdata->force |= USE_OPERATIONAL_PHASE;
		} else if (strcmp(ptr_to_keytext, "r")
			   == 0) {
			devdata->force |= USE_FULL_REPLIES;
		} else if (strcmp(ptr_to_keytext, "xok")
			   == 0) {
			devdata->force |= USE_REFLECT_XKEYS;
		} else if (strcmp(ptr_to_keytext, "i")
			   == 0) {
			devdata->force |= ALLOW_OFFERS_WHEN_IRRELEVANT;
		} else if (strcmp(ptr_to_keytext, "I")
			   == 0) {
			devdata->force |= ALLOW_REPLIES_WHEN_IRRELEVANT;
#ifdef RDMA_MODE
		} else if (strcmp(ptr_to_keytext, "iseroff") == 0) {
			devdata->force |= ISER_ISCSI_KEYS_OFF;
		} else if (strcmp(ptr_to_keytext, "rdmaLogin") == 0) {
			devdata->force |= USE_RDMA_LOGIN;
#endif
		} else if (strcmp(ptr_to_keytext, "tk1") == 0) {
			devdata->force |= USE_ONE_KEY_PER_TEXT;
		} else if (strncmp(ptr_to_keytext, "n=", 2) == 0) {
			devdata->nop_period =
			    simple_strtoul(ptr_to_keytext + 2, NULL, 0)
			    * HZ;
		} else if (strncmp(ptr_to_keytext, "v=", 2) == 0) {
			/* formerly set draft number to enforce */
			int_val =
			    simple_strtoul(ptr_to_keytext + 2, NULL, 0);
			if (int_val != DRAFT20) {
				TRACE(TRACE_ISCSI,
				      "Draft number %u ignored\n",
				      int_val / DRAFT_MULTIPLIER);
			}
		} else if (strncmp(ptr_to_keytext, "V=", 2) == 0) {
			TRACE(TRACE_ISCSI,
			      "Debug Verbose set to \"%s\"\n",
			      ptr_to_keytext + 2);
			int_val = simple_strtoul(ptr_to_keytext + 2,
						 NULL, 0);
			printk("trace_info = 0x%08x\n", int_val);
			TRACE_SET(int_val);
		} else if (strncmp(ptr_to_keytext, "p=", 2) == 0) {
			/* set phase-collapse choice */
			int_val =
			    simple_strtoul(ptr_to_keytext + 2, NULL, 0);
			if (int_val > 2 || int_val < 0) {
				TRACE_ERROR
				    ("%s Bad phase-collapse value %d\n",
				     current->comm, int_val);
			} else
				devdata->phase_collapse = int_val;
		} else if (strncmp(ptr_to_keytext, "Q=", 2) == 0) {
			/* set Maximum Queue Depth Allowed */
			int_val =
			    simple_strtoul(ptr_to_keytext + 2, NULL, 0);
			if (int_val > 255 || int_val <= 0) {
				TRACE_ERROR("%s Bad Max Queue Depth "
					    "Allowed %d\n", current->comm,
					    int_val);
			} else {
				devdata->max_queue_depth_allowed = int_val;
			}
		} else if (strncmp(ptr_to_keytext, "r2tp=", 5) == 0) {
			/* set the r2t period - SAI */
			devdata->r2t_period =
			    simple_strtoul(ptr_to_keytext + 5, NULL, 0)
			    * HZ;
		} else if (strcmp(ptr_to_keytext, "t") == 0) {
			/* chap and srp support - CHONG */
			TRACE(TRACE_ISCSI,
			      "target confirmation enabled\n");
			devdata->auth_parameter.auth_flags |=
			    USE_TARGET_CONFIRMATION;
		} else if (strcmp(ptr_to_keytext, "b") == 0) {
			TRACE(TRACE_ISCSI, "base64 number enabled\n");
			devdata->auth_parameter.auth_flags |= USE_BASE64;
			devdata->auth_parameter.chap_local_ctx->
			    number_format = BASE64_FORMAT;
			devdata->auth_parameter.chap_peer_ctx->
			    number_format = BASE64_FORMAT;
		} else if (strncmp(ptr_to_keytext, "px=", 3) == 0) {
			TRACE(TRACE_ISCSI, "CHAP peer secret set to %s\n",
			      ptr_to_keytext + 3);
			CHAP_SetSecret(ptr_to_keytext + 3,
				       devdata->auth_parameter.
				       chap_peer_ctx);
		} else if (strncmp(ptr_to_keytext, "pn=", 3) == 0) {
			TRACE(TRACE_ISCSI, "CHAP peer name set to %s\n",
			      ptr_to_keytext + 3);
			CHAP_SetName(ptr_to_keytext + 3,
				     devdata->auth_parameter.
				     chap_peer_ctx);
		} else if (strncmp(ptr_to_keytext, "lx=", 3) == 0) {
			TRACE(TRACE_ISCSI, "CHAP local secret set to %s\n",
			      ptr_to_keytext + 3);
			CHAP_SetSecret(ptr_to_keytext + 3,
				       devdata->auth_parameter.
				       chap_local_ctx);
		} else if (strncmp(ptr_to_keytext, "ln=", 3) == 0) {
			TRACE(TRACE_ISCSI, "CHAP local name set to %s\n",
			      ptr_to_keytext + 3);
			CHAP_SetName(ptr_to_keytext + 3,
				     devdata->auth_parameter.
				     chap_local_ctx);
		} else if (strncmp(ptr_to_keytext, "cl=", 3) == 0) {
			int_val =
			    simple_strtoul(ptr_to_keytext + 3, NULL, 0);
			if (int_val <= 0
			    || int_val > MAX_CHAP_BINARY_LENGTH) {
				TRACE_ERROR("%s invalid CHAP challenge "
					    "length %d\n", current->comm,
					    int_val);
			}
			TRACE(TRACE_ISCSI, "challenge length set to %d\n",
			      int_val);
			CHAP_SetChallengeLength(int_val,
						devdata->auth_parameter.
						chap_local_ctx);
		} else if (strncmp(ptr_to_keytext, "sx=", 3) == 0) {
			TRACE(TRACE_ISCSI, "SRP secret set to \"%s\"\n",
			      ptr_to_keytext + 3);
			SRP_SetSecret(ptr_to_keytext + 3,
				      devdata->auth_parameter.srp_ctx);
		} else if (strncmp(ptr_to_keytext, "sn=", 3) == 0) {
			TRACE(TRACE_ISCSI, "SRP name set to \"%s\"\n",
			      ptr_to_keytext + 3);
			SRP_SetName(ptr_to_keytext + 3,
				    devdata->auth_parameter.srp_ctx);
		} else {
			TRACE_ERROR("%s unknown force \"%s\"\n",
				    current->comm, ptr_to_keytext);
		}
	} else if (param_neg_info == RESTORE) {
		/* restore all settings to default values */

		/* reset parameter list to copy of defaults */
		param_tbl_uncpy(*devdata->param_tbl);
		param_tbl_init(*devdata->param_tbl);

		/* reset all other fields to 0 (or default) */
		devdata->force = 0;
		devdata->nop_period = 0;
		devdata->r2t_period = 0;
		devdata->max_queue_depth_allowed = QUEUE_DEPTH_ALLOWED;
		devdata->targ_snack_flg = 0;
		devdata->phase_collapse = 0;
		CHAP_FinalizeContext(devdata->auth_parameter.
				     chap_local_ctx);
		CHAP_FinalizeContext(devdata->auth_parameter.
				     chap_peer_ctx);
		SRP_FinalizeContext(devdata->auth_parameter.srp_ctx);
		devdata->auth_parameter.chap_local_ctx
		    = CHAP_InitializeContext();
		devdata->auth_parameter.chap_peer_ctx
		    = CHAP_InitializeContext();
		devdata->auth_parameter.srp_ctx = SRP_InitializeContext();

	} else if (param_neg_info == SNACK) {
		/* Error Recovery options to enable or disable SNACK - SAI */

		if (strncmp(ptr_to_keytext, "a=", 2) == 0) {
			if ((strcmp(ptr_to_keytext + 2, "y") == 0)
			    || strcmp(ptr_to_keytext + 2, "Y") == 0) {
				devdata->targ_snack_flg
				    |= DATACK1_SNACK_ENABLE;
			} else {
				devdata->targ_snack_flg
				    &= ~DATACK1_SNACK_ENABLE;
			}
		} else if (strncmp(ptr_to_keytext, "A=", 2) == 0) {
			if ((strcmp(ptr_to_keytext + 2, "y") == 0)
			    || strcmp(ptr_to_keytext + 2, "Y") == 0) {
				devdata->targ_snack_flg
				    |= DATACK2_SNACK_ENABLE;
			} else {
				devdata->targ_snack_flg
				    &= ~DATACK2_SNACK_ENABLE;
			}
		} else if (strncmp(ptr_to_keytext, "d=", 2) == 0) {
			if ((strcmp(ptr_to_keytext + 2, "y") == 0)
			    || strcmp(ptr_to_keytext + 2, "Y") == 0) {
				devdata->targ_snack_flg |=
				    DATA_SNACK_ENABLE;
			} else {
				devdata->targ_snack_flg &=
				    ~DATA_SNACK_ENABLE;
			}
		} else if (strncmp(ptr_to_keytext, "s=", 2) == 0) {
			if ((strcmp(ptr_to_keytext + 2, "y") == 0)
			    || strcmp(ptr_to_keytext + 2, "Y") == 0) {
				devdata->targ_snack_flg |=
				    STATUS_SNACK_ENABLE;
			} else {
				devdata->targ_snack_flg &=
				    ~STATUS_SNACK_ENABLE;
			}
		} else if (strncmp(ptr_to_keytext, "h=", 2) == 0) {

			if (simple_strtoul(ptr_to_keytext + 2, NULL, 0) ==
			    1) {
				devdata->targ_snack_flg |=
				    DATA_SNACK_RETRANSMIT;
			} else
			    if (simple_strtoul(ptr_to_keytext + 2, NULL, 0)
				== 2) {
				devdata->targ_snack_flg |=
				    DATA_SNACK_REJECT;
			}

			TRACE(TRACE_DEBUG,
			      "SNACK Flags : datasnack %d statsnack %d "
			      "DataSnackRexmit %d  DataSnackReject %d\n",
			      (devdata->
			       targ_snack_flg & DATA_SNACK_ENABLE),
			      ((devdata->
				targ_snack_flg & STATUS_SNACK_ENABLE) >>
			       1),
			      ((devdata->
				targ_snack_flg & DATA_SNACK_RETRANSMIT) >>
			       7),
			      ((devdata->
				targ_snack_flg & DATA_SNACK_REJECT) >> 6));
		} else {
			TRACE_ERROR("%s unknown snack \"%s\"\n",
				    current->comm, ptr_to_keytext);
		}
	} else {
		configure_parameter(param_neg_info, ptr_to_keytext,
				    *devdata->param_tbl);
	}
}

/*
 * Called for writing to the proc FS with the "iscsi_target ip" command
 */
static int iscsi_proc_info_targ_ip(char *buffer, char **start,
				   off_t offset, int length, int inode)
{
	int result = 0;
	int tag;
	int action;
	char *ptr;

	/* this buffer has expected preface, skip over
	 * ip address */
	char *ip_string, *port_string, *tag_string;

	ip_string = buffer + 16;
	ptr = strchr(ip_string, ' ');

	if (ptr == NULL || strncmp(ptr, " port ", 6) != 0) {
		result = -EINVAL;
		goto out;
	}
	*ptr = '\0';

	/* this buffer has expected preface, scan
	 * port number */
	port_string = ptr + 6;

	ptr = strchr(port_string, ' ');
	if (strncmp(ptr, " tag ", 5) != 0) {
		result = -EINVAL;
		goto out;
	}
	*ptr = '\0';

	/* this buffer has expected preface, scan
	 * target number */
	tag_string = ptr + 5;

	tag = simple_strtoul(tag_string, NULL, 0);

	ptr = strchr(tag_string, ' ');
	if (strncmp(ptr, " action ", 8) != 0) {
		result = -EINVAL;
		goto out;
	}

	/* this buffer has expected preface, scan
	 * action number */
	action = simple_strtoul(ptr + 8, NULL, 0);

	if (action == 0) {
		if (start_server_thread(ip_string, port_string, tag) < 0) {
			result = -EINVAL;
			goto out;
		}
	} else {
		if (stop_server_thread(ip_string, port_string) < 0) {
			result = -EINVAL;
			goto out;
		}
	}
      out:
	return result;
}
