/*	target/scsi_target.c
 *
 *	vi: set autoindent tabstop=8 shiftwidth=8 :
 *
 * 	This file has all the functions for scsi midlevel on target side
 *
 *	Copyright (C) 2001-2007 InterOperability Lab (IOL)
 *	University of New Hampshire (UNH)
 *	Durham, NH 03824
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *	USA.
 *
 *	The name of IOL and/or UNH may not be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
*/

#ifdef __KERNEL__

#define MEMORY_IO_DEBUG_PAGES
#define MEMORY_IO_FILL_VAL 0xeb

/* df revised begin */
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#ifndef KERNEL_VERSION		/* pre-2.1.90 didn't have it */
#define KERNEL_VERSION(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif
/* df revised end */

#include <linux/proc_fs.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
#include <scsi_device.h>	/* struct scsi_device */
#endif				/* LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18) */

#endif

#include "../common/scatterlist.h"

#include "scsi_target.h"
#include "../common/lun_packing.h"
#include "../common/iscsi_common.h"
#include "../common/unh_threads.h"

#include "../common/my_memory.h"

/* for use in alloc_pages() and __get_free_pages()
 * -- is the power of 2 such that 2^this_power pages are allocated at a time */
 /* eab -- PAGE_ALLOC_ORDER was originally 5.  This gives strange errors when
  * doing > 25 reads of 60000 blocks with
  * InitiatorRecvDataSegmentLength=16777215 */
#define PAGE_ALLOC_ORDER	8

#define PAGE_ALLOC_BLOCK	(PAGE_SIZE << PAGE_ALLOC_ORDER)

/* FUNCTION DECLARATIONS */

void scsi_target_process_thread(void *);

#if defined (FILEIO) || defined (GENERICIO)
static int build_filp_table(void);
static void close_filp_table(void);
#endif

static int abort_notify(struct SM *);
static void aen_notify(int, __u64);

static int hand_to_front_end(Target_Scsi_Cmnd *);

#ifdef CONFIG_PROC_FS
static int scsi_target_proc_info(char *buffer, char **start, off_t offset,
				 int length);
static int proc_scsi_target_gen_write(struct file *file, const char *buf,
				      unsigned long length, void *data);
static void remove_proc_target_dir_entries(Scsi_Target_Device *
					   the_device);
static int build_proc_target_dir_entries(Scsi_Target_Device * the_device);
#endif

/* holds pointer to newly created directory "/proc/scsi_target" */
struct proc_dir_entry *proc_scsi_target;

#if defined(DISKIO) && defined(TRUST_CDB)
/* doubly-linked circular list, one entry for every iscsi target
 * known to the scsi subsystem on this platform
 */
struct list_head target_map_list;
#else
/* matrix with one entry for every possible target and lun */
struct target_map_item target_map[MAX_TARGETS][MAX_LUNS];
#endif

/* mutex to control access to target_map array or list */
rwlock_t target_map_rwlock = RW_LOCK_UNLOCKED;

/*
 * Scsi_Target_Device: Maximum number of "active" targets at any time
 * Scsi_Target_Template: Maximum number of templates available
 *
 * The naming convention for variables is with the prefix "st" following
 * Eric's comments in the SCSI Initiator mid-level.
 *
 * The idea with the following variables is to have a generic linked
 * list of templates - there is one template for one "type" of device
 * whereas for each device there is a single entry in the device list
 */

Target_Emulator target_data;

/*
 * okay the following is to get something off the ground and working.
 * In the mode that the emulator is talking to a real disk drive, there
 * will be a need to map the id given to us by rx_cmnd to host, channel,
 * id as far as the SCSI mid-level is concerned. For the present moment
 * we map all of these to the same target. So all commands are directed
 * to different LUNs on a single device. I visualize this as being a
 * linked list of scsi disks attached to the system and whenever a
 * command is received, the rx_cmnd function does some id_lookup and map
 * the individual commands to one in the linked list. How this mapping
 * gets done is an upper level management function that I do not want to
 * deal with yet	ASHISH
 */
int target_count = 0;

/*
 * scsi_target_init_module: Initializes the SCSI target module
 * FUNCTION: carry out all the initialization stuff that is needed
 * INPUT: none
 * OUTPUT: 0: everything is okay
 * 	      else trouble
 */
int scsi_target_init_module(void)
{
	int r = -1;

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *generic;
#endif				/* CONFIG_PROC_FS */

	printk("scsi_target_init_module Starting\n");
	printk("scsi_target_init_module using %s\n", pr_label);

	/* initialize the global target_data struct */
	/*
	 * initialize the target semaphore as locked because you want
	 * the thread to block after entering the first loop
	 */
	init_MUTEX_LOCKED(&target_data.target_sem);
	init_MUTEX_LOCKED(&target_data.thread_sem);

	spin_lock_init(&target_data.msg_lock);
	target_data.next_device_id = 0;
	target_data.st_device_list = NULL;
	target_data.st_template_list = NULL;
	spin_lock_init(&target_data.cmd_queue_lock);
	INIT_LIST_HEAD(&target_data.cmd_queue);
	INIT_LIST_HEAD(&target_data.free_cmd_queue);
	atomic_set(&target_data.total_commands, 0);
	target_data.msgq_start = NULL;
	target_data.msgq_end = NULL;
	target_data.command_id = 0;
	target_data.thread_id = NULL;
	target_data.info = 0;

#ifdef MEMORYIO
	{			/* in memory mode, all luns in all targets are always in use */
#ifdef K26
		struct page *page;
#endif
		__u32 targ, lun;

		for (targ = 0; targ < MAX_TARGETS; targ++) {
			for (lun = 0; lun < MAX_LUNS; lun++) {
				target_map[targ][lun].in_use = 1;
			}
		}
		target_count = MAX_TARGETS * MAX_LUNS;

		printk("PAGE_SIZE %lu MAX_ORDER %u PAGE_ALLOC_ORDER %u "
		       "PAGE_ALLOC_BLOCK %lu\n", PAGE_SIZE, MAX_ORDER,
		       PAGE_ALLOC_ORDER, PAGE_ALLOC_BLOCK);
#ifdef K26

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
		sg_init_table(&memory_st_list, 1);
#endif

		page = alloc_pages(GFP_ATOMIC, PAGE_ALLOC_ORDER);
		if (!page) {
			printk("no space for memory_st_list.page\n");
			return -1;
		}

		/* the size parameters is the length of the buffer */
		sg_set_page(&memory_st_list, page, PAGE_ALLOC_BLOCK, 0);

		/*
		scatterlist_address(&memory_st_list) =
		    (dma_addr_t) (unsigned long) kmap(sg_page(&memory_st_list));
		    */
		/* this should always be 0 */
		memory_st_list.offset =
		    ((unsigned long) sg_virt(&memory_st_list))
		    & ~PAGE_MASK;
		scatterlist_length(&memory_st_list) = PAGE_ALLOC_BLOCK;
#else				/* K26 */
		scatterlist_address(&memory_st_list) =
		    (__u8 *) __get_free_pages(GFP_ATOMIC,
					      PAGE_ALLOC_ORDER);

		if (!scatterlist_address(&memory_st_list)) {
			printk("no space for memory_st_list.address\n");
			return -1;
		}
		memory_st_list.length = PAGE_ALLOC_BLOCK;
#endif				/* K26 */

#if defined(MEMORY_IO_DEBUG_PAGES)
		memset(sg_virt(&memory_st_list),
		       MEMORY_IO_FILL_VAL,
		       PAGE_ALLOC_ORDER);
#endif /* MEMORY_IO_DEBUG_PAGES */

	}
#endif				/* MEMORYIO */

#if defined(DISKIO) && defined(TRUST_CDB)
	INIT_LIST_HEAD(&target_map_list);
#endif				/* DISKIO && TRUST_CDB */

	/*
	 * create a list of devices that are available. This is a very
	 * simplistic mapping of these devices. Each new device will
	 * represent a new LUN. A new mapping function will have to
	 * solve this problem - LATER - Ashish
	 */
#if defined(FILEIO) || defined(GENERICIO)
	if (build_filp_table() < 0) {
		printk
		    ("scsi_target_init_module: build_filp_table returned error\n");
		goto out1;
	}
#endif				/* FILEIO || GENERICIO */

#ifdef CONFIG_PROC_FS
	proc_scsi_target = proc_mkdir("scsi_target", 0);
	if (!proc_scsi_target) {
		printk(KERN_ERR "cannot init /proc/scsi_target\n");
		r = -ENOMEM;
		goto out2;
	}
	printk("created directory /proc/scsi_target\n");
	generic =
	    create_proc_info_entry("scsi_target", 0, proc_scsi_target,
				   scsi_target_proc_info);
	if (!generic) {
		printk(KERN_ERR
		       "cannot init /proc/scsi_target/scsi_target\n");
		r = -ENOMEM;
		goto out3;
	}
	printk("created file /proc/scsi_target/scsi_target\n");
	generic->write_proc = proc_scsi_target_gen_write;
#endif				/* CONFIG_PROC_FS */

	kernel_thread((int (*)(void *)) scsi_target_process_thread,
		      NULL, 0);
	return 0;

/* error cleanup - Charles Coffing, 18-April-2005 */
#ifdef CONFIG_PROC_FS
      out3:
	remove_proc_entry("scsi_target", 0);
      out2:
#endif				/* CONFIG_PROC_FS */

#if defined(FILEIO) || defined(GENERICIO)
	close_filp_table();
      out1:
#endif				/* FILEIO || GENERICIO */
	return r;
}

/*
 * scsi_target_cleanup_module: Removal of the module from the kernel
 * FUNCTION: carry out the cleanup stuff
 * INPUT: None allowed
 * OUTPUT: None allowed
 */
void scsi_target_cleanup_module(void)
{
	Target_Scsi_Cmnd *cmd_curr;
	struct list_head *lptr, *next;
	__u32 count;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	printk("scsi_target_cleanup_module\n");

#ifdef GENERICIO
	/* close all the devices that you have opened */
	close_filp_table();

	if (target_data.signal_id != NULL) {
		printk("sending SIGKILL (%d) to thread id %#08lx\n",
		       SIGKILL, target_data.signal_id->threadid);
		send_sig(SIGKILL, target_data.signal_id, 1);
		down_interruptible(&target_data.signal_sem);
	}
#endif				/* GENERICIO */

#ifdef FILEIO
	close_filp_table();
#endif				/* FILEIO */

#if defined(DISKIO) && defined(TRUST_CDB)
	{
		struct target_map_item *this_item;

		write_lock(&target_map_rwlock);
		list_for_each_safe(lptr, next, &target_map_list) {
			list_del(lptr);
			this_item = list_entry(lptr, struct target_map_item, link);
			kfree(this_item);
		}
		write_unlock(&target_map_rwlock);
	}
#endif				/* DISKIO && TRUST_CDB */

#ifdef CONFIG_PROC_FS
	/* Don't show the /proc/scsi_target files */
	remove_proc_entry("scsi_target/scsi_target", 0);
	remove_proc_entry("scsi_target", 0);
#endif				/* CONFIG_PROC_FS */

	/* kill the thread */
	if (target_data.thread_id != NULL) {
#ifndef __KERNEL__
		printk("sending SIGKILL (%d) to thread id %#08lx\n",
		       SIGKILL, target_data.thread_id->threadid);
#endif				/* __KERNEL__ */
		send_sig(SIGKILL, target_data.thread_id, 1);

#ifdef DEBUG_SCSI_THREAD
		printk("%s, waiting on thread_sem\n",
		       current->comm);
#endif				/* DEBUG_SCSI_THREAD */

		down_interruptible(&target_data.thread_sem);

#ifdef DEBUG_SCSI_THREAD
		printk("%s, awake after thread_sem\n",
		       current->comm);
#endif				/* DEBUG_SCSI_THREAD */
	}

	/* deallocate all free Target_Scsi_Cmnd commands */
	count = 0;
	list_for_each_safe(lptr, next, &target_data.free_cmd_queue) {
		count++;
		list_del(lptr);
		cmd_curr = list_entry(lptr, Target_Scsi_Cmnd, link);
		kfree(cmd_curr);
	}

	if (count == atomic_read(&target_data.total_commands)) {
		printk("%s release %u Target_Scsi_Cmnd commands\n",
		       current->comm, count);
	} else {
		printk
		    ("%s release %u Target_Scsi_Cmnd commands but %u allocated\n",
		     current->comm, count,
		     atomic_read(&target_data.total_commands));
	}

#ifdef MEMORYIO

#ifdef K26
	if (sg_page(&memory_st_list)) {
		kunmap(sg_page(&memory_st_list));
		__free_pages(sg_page(&memory_st_list), PAGE_ALLOC_ORDER);
	}
#else				/* K26 */
	free_pages((long int) scatterlist_address(&memory_st_list),
		   PAGE_ALLOC_ORDER);
#endif				/* K26 */

#endif				/* MEMORYIO */

	printk("scsi_target module Exiting\n");
}

module_init(scsi_target_init_module);
module_exit(scsi_target_cleanup_module);

void target_info(void)
{
	Target_Scsi_Cmnd *cmd_curr;
	struct list_head *lptr, *next;
	int count;

	if (target_data.thread_id == NULL) {
		printk("scsi_target_process_thread id is NULL\n");
	} else {
		printk
		    ("scsi_target_process_thread TCB state %lu, pid %lu, info %u\n",
		     target_data.thread_id->state,
		     (unsigned long) target_data.thread_id->pid,
		     target_data.info);
		if (list_empty(&target_data.cmd_queue)) {
			printk("0 commands in queue\n");
		} else {
			count = 0;
			list_for_each_safe(lptr, next,
					   &target_data.cmd_queue) {
				count++;
				cmd_curr =
				    list_entry(lptr, Target_Scsi_Cmnd,
					       link);
				printk("%d command id %d status %d\n",
				       count, cmd_curr->id,
				       cmd_curr->state);
			}
		}
	}
}

/*
 * Returns 1 if any luns for this target are in use.
 * Returns 0 otherwise.
 */
int target_in_use(__u32 target_id)
{
	int result = 0;

	read_lock(&target_map_rwlock);

#if defined(DISKIO) && defined(TRUST_CDB)
	{
		struct target_map_item *this_item;

		list_for_each_entry(this_item, &target_map_list, link) {
			if (this_item->target_id == target_id
			    && this_item->in_use) {
				result = 1;
				break;
			}
		}
	}
#else				/* DISKIO && TRUST_CDB */
	if (target_id < MAX_TARGETS) {
		__u32 lun;

		for (lun = 0; lun < MAX_LUNS; lun++) {
			if (target_map[target_id][lun].in_use) {
				result = 1;
				break;
			}
		}
	}
#endif				/* DISKIO && TRUST_CDB */

	read_unlock(&target_map_rwlock);
	return result;
}

/*
 * register_target_template: to register a template with the midlevel
 * FUNCTION: to add the midlevel template to the linked list
 * INPUT: pointer to the template
 * OUTPUT: 0 for successful registration
 * 	   < 0 for trouble
 */
int register_target_template(Scsi_Target_Template * the_template)
{
	int check;
	Scsi_Target_Template *st_current;

	if (!the_template) {	/* huh */
		printk
		    ("register_target_template: refuse to register NULL template\n");
		return -1;
	}

	/* error if the template does not have a detect function */
	if (!the_template->detect) {
		printk
		    ("register_target_template: template does not have a detect function\n");
		return -1;
	}

	/* error if the template is already in the global template list */
	for (st_current = target_data.st_template_list; st_current != NULL;
	     st_current = st_current->next) {
		/* loop */
#ifdef DEBUG_REGISTER
		printk("Looping template %s\n", the_template->name);
#endif
		if (!strcmp(st_current->name, the_template->name)) {
			/* template is old */
			printk
			    ("register_target_template: Template %s already present\n",
			     the_template->name);
			return -1;
		}
	}

	/* add template to front of global template list */
	the_template->device_usage = 0;
	the_template->next = target_data.st_template_list;
	target_data.st_template_list = the_template;

#ifdef DEBUG_REGISTER
	printk("module count increased\n");
#endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
	MOD_INC_USE_COUNT;
#endif

	/* detect the device */
	check = the_template->detect(the_template);
	if (check < 0) {	/* error */
		printk("Hosts detected: %d .. removing template %s\n",
		       check, the_template->name);
		deregister_target_template(the_template);
		return check;
	}

	printk("Hosts detected: %d ... incrementing device usage\n",
	       check);
	return 0;
}

/*
 * deregister_target_template:
 * FUNCTION: removal of the target template from the list
 * INPUT: pointer to STT
 * OUTPUT: 0 for successful deregistration
 * 	   else < 0 for trouble
 */
int deregister_target_template(Scsi_Target_Template * the_template)
{
	Scsi_Target_Template *st_current, *st_prev;
	Scsi_Target_Device *st_dev_curr;

	if (!the_template) {	/* huh */
		printk("dereg...tmpt: refuse to remove NULL template\n");
		return -1;
	}

	/* remove all devices that are using this template */
	for (st_dev_curr = target_data.st_device_list;
	     st_dev_curr != NULL;) {
		if (st_dev_curr->template == the_template) {
#ifdef DEBUG_DEREGISTER
			printk("dereg...tmpt: match found for %s%u\n",
			       st_dev_curr->template->name,
			       st_dev_curr->dev_id);
#endif
			deregister_target_front_end(st_dev_curr);
			/* list has been perturbed, restart search from front again */
			st_dev_curr = target_data.st_device_list;
		} else {
			st_dev_curr = st_dev_curr->next;
		}
	}

	/* error if there are any devices still using this template */
	if (the_template->device_usage) {
		printk
		    ("dereg...tmpt: refuse to remove Template %s with non-zero "
		     "device usage %d\n", the_template->name,
		     the_template->device_usage);
		return -1;	/* should never get here */
	}

	/* remove template from device list */
	st_prev = NULL;
	for (st_current = target_data.st_template_list; st_current != NULL;
	     st_current = st_current->next) {
#ifdef DEBUG_DEREGISTER
		printk("Looping dereg template\n");
#endif
		if (st_current == the_template) {
#ifdef DEBUG_DEREGISTER
			/* match found */
			printk("dereg...tmpt: Template match found\n");
#endif
			/* this template can be removed */
			if (!st_prev)	/* First element */
				target_data.st_template_list =
				    target_data.st_template_list->next;
			else	/* middle of the road */
				st_prev->next = st_current->next;
			break;
		} else
			st_prev = st_current;
	}

#ifdef DEBUG_DEREGISTER
	printk("Decreasing module count\n");
#endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
	MOD_DEC_USE_COUNT;
#endif

	return 0;
}

/*
 * register_target_front_end:
 * FUNCTION: 	to register the individual device with the mid-level
 * 		to allocate the device an device id etc
 * 		start a thread that will be responsible for the target
 * INPUT:	pointer to a struct STD
 * OUTPUT:	Scsi_Target_Device - if everything is okay
 * 		else NULL if there is trouble
 */
Scsi_Target_Device *register_target_front_end(Scsi_Target_Template * tmpt)
{
	Scsi_Target_Device *the_device;

	if (!tmpt) {
		printk
		    ("register_target_front_end: Cannot register NULL device template !!!\n");
		return NULL;
	}

	the_device = kmalloc(sizeof(Scsi_Target_Device), GFP_KERNEL);
	if (!the_device) {
		printk
		    ("register_target_front_end: Could not allocate space for device\n");
		return NULL;
	}
#ifdef DEBUG_DEREGISTER
	printk("register_target_front_end: the_device %p tmpt %p\n",
	       the_device, tmpt);
#endif

	/* fill up the struct */
	the_device->template = tmpt;
	the_device->dev_id = target_data.next_device_id++;

	/* add to front of global device list */
	the_device->next = target_data.st_device_list;
	target_data.st_device_list = the_device;

#ifdef DEBUG_DEREGISTER
	printk("reg..end: device %s%u added\n", tmpt->name,
	       the_device->dev_id);
#endif

	/* Added by naren for proc support */
#ifdef CONFIG_PROC_FS
	build_proc_target_dir_entries(the_device);
#endif

	tmpt->device_usage++;
	return the_device;
}

/*
 * deregister_target_front_end:
 * FUNCTION:	to allow removal of the individual device from the
 * 		midlevel
 * 		free up the device id number for future use
 * 		close the thread responsible for the target after making
 * 		sure that all existing commands have been responded to
 * 		we need to do serious error checking since this can be
 * 		called by different front-ends
 * 		CANNOT BE CALLED FROM INTERRUPT CONTEXT
 * INPUT:	pointer to a struct STD to be removed
 * OUTPUT:	int - 0 if everything is okay
 * 		else < 0 if there is trouble
 */
int deregister_target_front_end(Scsi_Target_Device * the_device)
{
	Scsi_Target_Device *curr, *previous = NULL;
	Target_Scsi_Cmnd *cmnd;
	unsigned long flags;

	if (!the_device) {
		printk("dereg...end: cannot remove NULL device\n");
		return -1;
	}

	if (!the_device->template) {
		printk
		    ("dereg...end: cannot remove device with NULL template\n");
		return -1;
	}

	if (!the_device->template->device_usage) {
		printk
		    ("dereg...end: 0 device usage in template %s when deregistering device\n",
		     the_device->template->name);
		return -1;
	}

	/*
	 * go through the device list till we get to this device and
	 * then remove it from the list
	 */
	for (curr = target_data.st_device_list; curr != NULL;
	     curr = curr->next) {
		if (curr == the_device) {
#ifdef DEBUG_DEREGISTER
			printk("dereg..end: Have a match\n");
#endif
			break;
		} else
			previous = curr;
	}

	if (!curr) {		/* No match found */
		printk("dereg..end: No matching device found\n");
		return -1;
	}

	/* reduce device usage */
	curr->template->device_usage--;

	/* remove it from the list */
	if (previous)		/* not the first device */
		previous->next = curr->next;
	else
		target_data.st_device_list = curr->next;

	/* release the device */
	if (curr->template->release) {
		if (curr->template->release(curr)) {
			printk("dereg...end: release device failed\n");
			return -1;
		}
	}

	/* mark all commands corresponding to this device for dequeuing */
	spin_lock_irqsave(&target_data.cmd_queue_lock, flags);

	list_for_each_entry(cmnd, &target_data.cmd_queue, link) {
		if (cmnd->dev_id == curr->dev_id)
			cmnd->state = ST_DEQUEUE;
	}

	spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);

	/* wake up scsi_target_process_thread so it can dequeue stuff */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}

	/* freeing things */
/* error cleanup - Charles Coffing, 18-April-2005 */
#ifdef CONFIG_PROC_FS
	remove_proc_target_dir_entries(curr);
#endif
	kfree(curr);

	return 0;
}

/**
 * Performs cleanup on the target_process_thread.
 */
static void target_process_shutdown_thread(void *aux)
{
#ifndef __KERNEL__
	struct task_struct *current;
	current = pthread_getspecific(current_key);
#endif
	up(&target_data.thread_sem);
	printk("%s Exiting pid %d\n", current->comm, current->pid);
}

/*
 * scsi_target_process_thread: this is the mid-level target thread that
 * is responsible for processing commands.
 */
void scsi_target_process_thread(void *param)
{
	int i, found;
#ifdef GENERICIO
#ifdef __KERNEL__
	mm_segment_t old_fs;
#endif
	struct file *dev_file;
#endif				/* GENERICIO */
	struct scatterlist *st_list;
	Target_Scsi_Cmnd *cmd_curr;
	Target_Scsi_Message *msg;
	unsigned long flags;
	__u32 lun;
#ifdef DISKIO
	struct scsi_device *this_device = NULL;
	struct target_map_item *this_item;
#endif				/* DISKIO */
	struct list_head *lptr, *next;

#ifndef __KERNEL__
	struct task_struct *current;
#endif

	lock_kernel();

	unh_daemonize("target_thread");

	siginitsetinv(&current->blocked, SHUTDOWN_SIGS);

	/* mark that this target thread is alive */
	target_data.thread_id = current;

	unlock_kernel();

#ifdef DEBUG_SCSI_THREAD
	printk("%s Starting thread id %#08lx\n",
	       current->comm, (unsigned long) current->pid);
#endif				/* DEBUG_SCSI_THREAD */

#if !defined(__KERNEL__)
	pthread_cleanup_push(target_process_shutdown_thread, NULL);
#endif /* !__KERNEL__ */
	while (1) {

#ifdef DEBUG_SCSI_THREAD
#ifdef K26
		printk("%s going to sleep with irqs %sabled\n",
		       current->comm, irqs_disabled()? "dis" : "en");
#endif				/* K26 */
#endif				/* DEBUG_SCSI_THREAD */

#ifdef DEBUG_SCSI_THREAD
		printk("%s  waiting on target_sem, thread id %#08lx\n",
		       current->comm, (unsigned long) current->pid);
#endif				/* DEBUG_SCSI_THREAD */

		target_data.info = 1;
		if (down_interruptible(&target_data.target_sem))
			goto scsi_thread_out;
		target_data.info = 0;

#ifdef DEBUG_SCSI_THREAD
		printk("%s awake after target_sem, thread id %#08lx\n",
		       current->comm, (unsigned long) current->pid);
#endif

		/* is message received */
		while (target_data.msgq_start) {
			/* house keeping */
			spin_lock_irqsave(&target_data.msg_lock, flags);
			msg = target_data.msgq_start;
			target_data.msgq_start = msg->next;
			if (!target_data.msgq_start)
				target_data.msgq_end = NULL;
			spin_unlock_irqrestore(&target_data.msg_lock,
					       flags);

			/* execute function */
			switch (msg->fn) {
			case TMF_ABORT_TASK:
				{
					Target_Scsi_Cmnd *cmnd;
					cmnd = (Target_Scsi_Cmnd *) msg->value;
					found = 0;
					spin_lock_irqsave(&target_data.cmd_queue_lock, flags);
					list_for_each_entry(cmd_curr, &target_data.cmd_queue, link) {
						if ((cmd_curr->id == cmnd->id) && (cmd_curr->lun == cmnd->lun)) {
							found = 1;
							break;
						}
					}
					spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);

					if (found) {
						cmd_curr->abort_code = CMND_ABORTED;
						//if (cmd_curr->state != ST_PROCESSING)
						//      cmd_curr->state = ST_DEQUEUE;
					} else {
						printk("%s no command with id %d lun %d in list\n",
						     current->comm,
						     cmnd->id, cmnd->lun);
					}
					if (abort_notify(msg)) {
						printk("%s err aborting command with id %d lun %d\n",
						     current->comm,
						     cmnd->id, cmnd->lun);
						kfree(msg);
						goto scsi_thread_out;
					}
					break;
				}

			case TMF_LUN_RESET:
				{
					/* BAD BAD REALLY BAD */
					__u64 lun =
					    *((__u64 *) msg->value);

					spin_lock_irqsave(&target_data.
							  cmd_queue_lock,
							  flags);
					list_for_each_entry(cmd_curr,
							    &target_data.
							    cmd_queue,
							    link) {
						if (cmd_curr->lun == lun)
							scsi_release
							    (cmd_curr);
					}
					spin_unlock_irqrestore
					    (&target_data.cmd_queue_lock,
					     flags);

					aen_notify(msg->fn, lun);
					break;
				}

			case TMF_TARGET_WARM_RESET:
			case TMF_TARGET_COLD_RESET:
				{
					spin_lock_irqsave(&target_data.
							  cmd_queue_lock,
							  flags);
					list_for_each_entry(cmd_curr,
							    &target_data.
							    cmd_queue,
							    link) {
						scsi_release(cmd_curr);
					}
					spin_unlock_irqrestore
					    (&target_data.cmd_queue_lock,
					     flags);

					aen_notify(msg->fn, 0);
					break;
				}

			default:
				{
					printk
					    ("%s Bad message fn code %d\n",
					     current->comm, msg->fn);
					break;
				}
			}	/* switch */

			kfree(msg);
		}


#ifdef DEBUG_SCSI_THREAD
#ifdef K26
		printk
		    ("%s awake with irqs %sabled, search command queue\n",
		     current->comm, irqs_disabled()? "dis" : "en");
#endif				/* K26 */
#endif				/* DEBUG_SCSI_THREAD */

		/* There is a harmless race here.
		 * This loop is the ONLY place a command can be removed from the queue.
		 * So once we get lptr, it cannot be made invalid elsewhere.
		 * However, if a new element is added to the end of the queue
		 * (the ONLY place new elements are ever added) as lptr is being
		 * picked up or as next is being picked up or after next has been
		 * picked up, then we might not see this new element on the
		 * next iteration of this loop.
		 * That should not cause a problem, because the ONLY place a new
		 * command can be added to the queue is in rx_cmnd(), and after
		 * adding the new command to the queue, rx_cmnd() also does an "up"
		 * on target_sem, which will restart the outer loop of this thread
		 * and we will be back into this loop again in no time.
		 * Why do we say this race is harmless?
		 * The loop initialization generated by list_for_each_safe:
		 *              lptr = (&target_data.cmd_queue)->next
		 * and the code:
		 *              next = lptr->next;
		 * will always work properly, because list_add_tail() always sets
		 * the referenced "next" field last when adding a new element to
		 * the queue (i.e., AFTER all other pointer fields are set up
		 * correctly).   Therefore, there is a race with list_add_tail(),
		 * but we always get a pointer to a valid structure (either the head
		 * of the list, in which case we don't process the new element, or
		 * the new element which is completely filled in, in which case we
		 * do process the new element).
		 */
		list_for_each_safe(lptr, next, &target_data.cmd_queue) {
			cmd_curr =
			    list_entry(lptr, Target_Scsi_Cmnd, link);

#ifdef DEBUG_SCSI_THREAD
			printk("%s lptr %p next %p cmd_curr %p id %d "
			       "state %d\n",
			       current->comm, lptr, next, cmd_curr,
			       cmd_curr->id, cmd_curr->state);
#endif				/* DEBUG_SCSI_THREAD */

			/* is command received */
			if (cmd_curr->state == ST_NEW_CMND) {

				lun = cmd_curr->lun;

#if defined(DISKIO)
				read_lock(&target_map_rwlock);
#if defined(TRUST_CDB)
				list_for_each_entry(this_item, &target_map_list, link) {
					if (this_item->target_id == cmd_curr->target_id && this_item->in_use) {
						this_device = this_item->the_device;
						break;
					}
				}
#else				/* TRUST_CDB */
				if (cmd_curr->target_id < MAX_TARGETS) {
					if (lun < MAX_LUNS) {
						this_item = &target_map[cmd_curr->target_id][lun];
						if (this_item->in_use) {
							this_device = this_item->the_device;
						}
					}

					if (!this_device) {
						this_item = &target_map[cmd_curr->target_id][0];
						if (this_item->in_use) {
							printk("%s No lun %u for target id %u, "
							     "using lun 0 instead\n",
							     current->comm,
							     lun,
							     cmd_curr->target_id);
							this_device = this_item->the_device;
							lun = 0;
						}
					}
				}
#endif				/* TRUST_CDB */
				read_unlock(&target_map_rwlock);

				if (this_device) {
/* Ming Zhang, mingz@ele.uri.edu */
					if (!(cmd_curr->req)) {
#ifdef K26
						cmd_curr->req =
						    scsi_allocate_request
						    (this_device,
						     GFP_ATOMIC);
#else				/* K26 */
						cmd_curr->req =
						    scsi_allocate_request
						    (this_device);
#endif				/* K26 */
					}
				} else {
					printk
					    ("%s No device for lun %u target id %u\n",
					     current->comm, lun,
					     cmd_curr->target_id);
					goto scsi_thread_out;
				}
#else				/* DISKIO */
				if (!(cmd_curr->req)) {
					cmd_curr->req =
					    kmalloc(sizeof(Scsi_Request),
						    GFP_ATOMIC);
				}

				if (cmd_curr->req) {
					memset(cmd_curr->req, 0,
					       sizeof(Scsi_Request));

					if (lun >= MAX_LUNS)
						cmd_curr->req->sr_allowed =
						    1;
				}
#endif				/* DISKIO */

				if (!cmd_curr->req) {
					printk
					    ("%s No space for Scsi_Request\n",
					     current->comm);
				} else {

					memcpy(cmd_curr->req->sr_cmnd,
					       cmd_curr->cmd,
					       cmd_curr->len);

					if (handle_cmd(cmd_curr)) {
						printk
						    ("%s Error in handle_cmd for command %d\n",
						     current->comm,
						     cmd_curr->id);
						/* is bailing out a good idea? */
						goto scsi_thread_out;
					}
				}
			}

			/* is a command pending */
			if (cmd_curr->state == ST_PENDING) {
				/* call the rdy_to_xfer function */
				if (hand_to_front_end(cmd_curr)) {
					printk
					    ("%s Error in hand_to_front_end for command %d\n",
					     current->comm, cmd_curr->id);
					goto scsi_thread_out;
				}
			}

			/* is data received */
			if (cmd_curr->state == ST_TO_PROCESS) {
				/*
				 * we have received the data - does this
				 * go off to handle_cmd again ?
				 */
				if (handle_cmd(cmd_curr)) {
					printk
					    ("%s Error in handle_cmd for command %d\n",
					     current->comm, cmd_curr->id);
					/* is bailing out a good idea? */
					goto scsi_thread_out;
				}
			}
#ifdef GENERICIO

			/* is command processed */
			if (cmd_curr->state == ST_PROCESSED) {
				/* we have some reading to do */
				dev_file = cmd_curr->fd;
				if ((dev_file) && (dev_file->f_op)
				    && (dev_file->f_op->read)) {

#ifdef __KERNEL__
					old_fs = get_fs();
					set_fs(get_ds());
#endif

					i = dev_file->f_op->read(dev_file,
#ifdef K26
								 (char
								  __user *)
								 cmd_curr->
								 sg,
#else				/* K26 */
								 (__u8 *)
								 cmd_curr->
								 sg,
#endif				/* K26 */
								 sizeof
								 (sg_io_hdr_t),
								 &dev_file->
								 f_pos);
#ifdef __KERNEL__
					set_fs(old_fs);
#endif

					if ((i < 0) && (i != -EAGAIN)) {
						printk
						    ("%s Read error %d\n",
						     current->comm, i);
						goto scsi_thread_out;
					}
					cmd_curr->state = ST_DONE;
				}
			}
#endif				/* GENERICIO */

			/* is command done */
			if (cmd_curr->state == ST_DONE) {
				/*
				 * hand it off to the front end driver
				 * to transmit
				 */

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
#ifdef DISKIO
				/*
				 * for DISKIO under K26, should reassign sg_dma_address here
				 */
				/*
				for (i = 0; i < cmd_curr->req->sr_use_sg;
				     i++) {
					sg_virt(((struct scatterlist *) cmd_curr->req-> sr_buffer) + i) =
					    (dma_addr_t) page_address(((struct scatterlist *) cmd_curr->req->sr_buffer)[i].page);
					((struct scatterlist *) cmd_curr->
					 req->sr_buffer)[i].offset =
	   (unsigned long)
					    sg_virt(((struct
							     scatterlist *)
							    cmd_curr->req->
							    sr_buffer) +
							   i) & ~PAGE_MASK;
				}
				*/
#endif				/* DISKIO */
#endif				/* K26 */

				if (hand_to_front_end(cmd_curr)) {
					printk
					    ("%s Error in hand_to_front_end for command %d\n",
					     current->comm, cmd_curr->id);
					goto scsi_thread_out;
				}
			}

			/* can command be dequeued */
			if (cmd_curr->state == ST_DEQUEUE) {

#if defined(DEBUG_SCSI_THREAD)
				printk("%s: looking at dequeueing command\n", __func__);
#endif /* DEBUG_SCSI_THREAD */

				/*
				 * dequeue this command and then free it
				 */

#ifdef GENERICIO
				/* free up the SCSI generic stuff */
				if (cmd_curr->sg->dxferp)
					kfree(cmd_curr->sg->dxferp);
				kfree(cmd_curr->sg->sbp);
				kfree(cmd_curr->sg);
#endif				/* GENERICIO */

				if (cmd_curr->req) {
					/* free up pages */
					st_list = (struct scatterlist *)
					    cmd_curr->req->sr_buffer;
					if (st_list) {
						for (i = 0;
						     i <
						     cmd_curr->req->
						     sr_use_sg; i++) {

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
#ifdef MEMORYIO
							if (sg_page(&st_list[i]) ==
							    sg_page(&memory_st_list))
								continue;
#endif				/* MEMORYIO */
							if (sg_page(&st_list[i])) {
								kunmap(sg_page(&st_list[i]));
								__free_pages
								    (sg_page(&st_list[i]),
								     cmd_curr->
								     page_alloc_order);
							}
#else				/* K26 */
#ifdef MEMORYIO
							if (scatterlist_address(&st_list[i]) == scatterlist_address(&memory_st_list))
								continue;
#endif				/* MEMORYIO */
							if (scatterlist_address(&st_list[i])) {
/*
#ifdef DEBUG_SCSI_THREAD
*/
								if (cmd_curr->page_alloc_order) {
									printk
									    ("free page[%u] len %u order %u addr "
									     "%p\n",
									     i,
									     scatterlist_length
									     (&st_list
									      [i]),
									     cmd_curr->
									     page_alloc_order,
									     scatterlist_address
									     (&st_list
									      [i]));
								}
/* MEMORYIO
#endif
*/

								free_pages((long int) scatterlist_address(&st_list[i]), cmd_curr->page_alloc_order);
							}
#endif				/* K26 */
						}

#ifdef DEBUG_SCSI_THREAD
						printk
						    ("%s command %p id %d - freed %d st_list items\n",
						     current->comm,
						     cmd_curr,
						     cmd_curr->id, i);
#endif

						/* free up scatterlist if it was dynamically alloced */
						if (st_list !=
						    cmd_curr->st_list) {
							kfree(st_list);
						}
					}

					/* free up Scsi_Request */
#ifdef DISKIO
					scsi_release_request(cmd_curr->
							     req);
#else
					kfree(cmd_curr->req);
#endif
				}

				/* finally move this Target_Scsi_Cmnd to free command list */
				spin_lock_irqsave(&target_data.
						  cmd_queue_lock, flags);
				list_move_tail(lptr,
					       &target_data.
					       free_cmd_queue);
				spin_unlock_irqrestore(&target_data.
						       cmd_queue_lock,
						       flags);

#ifdef DEBUG_SCSI_THREAD
				printk("%s command %p - all free\n",
				       current->comm, cmd_curr);
#endif
			}
		}

	}
#if !defined(__KERNEL__)
	pthread_cleanup_pop(0);
#endif /* !__KERNEL__ */
      scsi_thread_out:
	target_process_shutdown_thread(NULL);
	return;
}

/*
 * rx_cmnd: this function is the basic function called by any front end
 * when it receives a SCSI command. The rx_cmnd function then fills up
 * a Target_Scsi_Cmnd struct, adds to the queue list, awakens the mid-
 * level thread and then returns the struct to the front end driver
 * INPUT:	device, target_id, lun, SCSI CDB as an unsigned char
 * 		length of the command (or size of scsi_cdb array if
 * 		unavailable
 * OUTPUT:	Target_Scsi_Cmnd struct or NULL if things go wrong
 */
Target_Scsi_Cmnd *rx_cmnd(Scsi_Target_Device * device, __u64 target_id,
			  __u64 lun, __u8 * scsi_cdb, int len, int datalen,
			  int in_flags, Target_Scsi_Cmnd ** result_command)
{
	Target_Scsi_Cmnd *command;
	unsigned long flags;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

	*result_command = command = NULL;

	if (!target_data.thread_id) {
		printk("rx_cmnd: No Mid-level running !!!!\n");
		return NULL;
	}

	if (!device) {
		printk("rx_cmnd: No device given !!!!\n");
		return NULL;
	}
#ifdef DEBUG_RX_CMND
#ifdef K26
	printk("rx_cmnd: enter with irqs %sabled\n",
	       irqs_disabled()? "dis" : "en");
#endif
#endif

	/* try to reuse existing command structure from the free command list */
	spin_lock_irqsave(&target_data.cmd_queue_lock, flags);
	if (!list_empty(&target_data.free_cmd_queue)) {
		command =
		    list_entry(target_data.free_cmd_queue.next,
			       Target_Scsi_Cmnd, link);
		list_del(&command->link);
	}
	spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);

	if (!command) {
		/* no free Target_Scsi_Cmnd command, try to allocate a new one */
		command = kmalloc(sizeof(Target_Scsi_Cmnd), GFP_ATOMIC);
		if (command == NULL) {
			printk
			    ("rx_cmnd: No space for Target_Scsi_Cmnd command\n");
			return NULL;
		}
		/* keep count of the total number of these commands ever allocated */
		atomic_inc(&target_data.total_commands);
	}
	*result_command = command;

#ifdef DEBUG_RX_CMND
	printk("rx_cmnd: filling up command struct %p\n", command);
#endif

	/* fill in Target_Scsi_Cmnd */
	memset(command, 0, sizeof(Target_Scsi_Cmnd));
	command->state = ST_NEW_CMND;
	command->abort_code = CMND_OPEN;
	command->device = device;
	command->dev_id = device->dev_id;

	/*ramesh@global.com
	   added data length and flgs to the command structure */
	command->datalen = datalen;
	command->flags = in_flags;

	/* cdeng change target_id later if lun doesn't match */
	command->target_id = target_id;
	command->lun = unpack_lun((__u8 *) & lun);
	if ((len <= MAX_COMMAND_SIZE) && (len > 0))
		command->len = len;
	else {
		printk("setting cmd len to %d instead of %d\n",
		       MAX_COMMAND_SIZE, len);
		command->len = MAX_COMMAND_SIZE;
	}
	memcpy(command->cmd, scsi_cdb, command->len);

#if defined (FILEIO) || defined (GENERICIO)
	/* fill in the file pointer */
	if (command->target_id < MAX_TARGETS && command->lun < MAX_LUNS) {
		struct target_map_item *this_item;

		read_lock(&target_map_rwlock);
		this_item = &target_map[command->target_id][command->lun];
		if (this_item->in_use) {
			command->fd = this_item->the_file;
#if defined (GENERICIO)
			command->blk_size = this_item->bytes_per_block;
#endif				/* GENERICIO */
		}
		read_unlock(&target_map_rwlock);
	}

	if (command->fd == NULL) {
		printk
		    ("%s No target for command with targetid %u, lun %u\n",
		     current->comm, command->target_id, command->lun);

		/*
		 * Arne Redlich <agr1@sourceforge.net>:
		 * kfree()'ing the command and returning NULL here will cause
		 * the connection / session to be killed, rendering the target
		 * unusable with initiators that don't understand/use the
		 * REPORT_LUNS response and instead rely on probing LUNs by
		 * INQUIRYs to a range of (possibly not available) LUNs
		 * - which is actually the case with the Cisco (3.4.3)
		 * and UNH (1.5.3, 1.5.4) initiators. Instead, doing nothing
		 * here and letting handle_cmd() generate a TYPE_NO_LUN INQUIRY
		 * response catches this quite gracefully.
		 */

	}
#endif				/* FILEIO || GENERICIO */

	spin_lock_irqsave(&target_data.cmd_queue_lock, flags);
#ifdef DEBUG_RX_CMND
	printk("rx_cmnd: locked cmd_queue_lock for %p\n", command);
#endif

	command->id = ++target_data.command_id;
	/* check this to make sure you dont have a command with this id ?????
	 * IGNORE FOR NOW
	 */
	if (!command->id) {
		/* FOR WRAP AROUNDS */
		command->id = ++target_data.command_id;
	}

	list_add_tail(&command->link, &target_data.cmd_queue);

#ifdef DEBUG_RX_CMND
	printk("rx_cmnd: unlock cmd_queue_lock for %p with id %d\n",
	       command, command->id);
#endif
	spin_unlock_irqrestore(&target_data.cmd_queue_lock, flags);

	/* wake up scsi_target_process_thread */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}
#ifdef DEBUG_RX_CMND
#ifdef K26
	printk("rx_cmnd: exit with irqs %sabled\n",
	       irqs_disabled()? "dis" : "en");
#endif
#endif

	return command;
}

/*
 * scsi_rx_data: This function is called by the lower-level driver to
 * tell the mid-level that data corresponding to a command has been
 * received. This function can be called from within an interrupt
 * handler (??). All this function does currently is to change the
 * state of a command and then wake the mid-level thread to deal with
 * this command.
 * INPUT: scsi command for which data has been received (MUST not be NULL)
 * OUTPUT: 0 if okay else < 0
 */
int scsi_rx_data(Target_Scsi_Cmnd * cmnd)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef DEBUG_SCSI_THREAD
	printk
	    ("%s in scsi_rx_data, command %p id %d change state from %d to %d\n",
	     current->comm, cmnd, cmnd->id, cmnd->state, ST_TO_PROCESS);
#endif
	cmnd->state = ST_TO_PROCESS;

	/* wake up the mid-level scsi_target_process_thread */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}

	return 0;
}

/*
 * scsi_target_done: This is the function called by the low-level driver
 * to signify that it has completed the execution of a given scsi cmnd
 * This function needs to remove the resources that have been allocated
 * to the given command, dequeue the command etc. This function will be
 * called from within the context of an interrupt handler. So, it may
 * be best to leave it up to the mid-level thread to actually deal with
 * these functions. Right now, I am setting it up so that the status of
 * the command is changed and the mid-level is awakened.
 * INPUT: scsi command to be dequeued
 * OUTPUT: 0 if everything is okay
 * 	   < 0 if there is trouble
 */
int scsi_target_done(Target_Scsi_Cmnd * cmnd)
{
#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef DEBUG_SCSI_THREAD
	printk
	    ("%s in scsi_target_done, command %p id %d change state from %d to %d\n",
	     current->comm, cmnd, cmnd->id, cmnd->state, ST_DEQUEUE);
#endif
	cmnd->state = ST_DEQUEUE;

	/* awaken scsi_target_process_thread to dequeue stuff */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}

	return 0;
}

/*
 * scsi_release: This function is called by a low-level driver when it
 * determines that it does not responses to certain commands. Situations
 * like this happen when for instance a LIP is received in a Fibre
 * Channel Loop or when a Logout is received in iSCSI before command
 * execution is completed. The low-level driver may no longer care about
 * receiving responses for those commands. This function can be called
 * from within interrupt context
 * INPUT: command to release
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
int scsi_release(Target_Scsi_Cmnd * cmnd)
{
	cmnd->abort_code = CMND_RELEASED;

	/*
	 * if a command is processing, it is not nice to
	 * dequeue a command, as we will get a response
	 * anyways. This may be an inherent race condn
	 * We catch what we can and move on. A second
	 * check performed when hand_to_front_end is
	 * called should catch the remaining commands
	 * - Ashish
	 */
	if (cmnd->state != ST_PROCESSING)
		cmnd->state = ST_DEQUEUE;

	/* wake up scsi_process_target_thread so it can dequeue stuff */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}

	return 0;
}

/*
 * rx_task_mgmt_fn: This function is called by a low-level driver to
 * indicate to the Mid-level that it has received a management function.
 * This function will decide the action to be taken in response to this
 * management function. This function will in turn create a message list
 * in the mid-level. CAN BE CALLED FROM INTERRUPT CONTEXT
 * INPUT: device, function, command - if relevant
 * OUTPUT: message or NULL if there is trouble
 * Definition of value:
 * 1. ABORT TASK
 * 	value = Target_Scsi_Cmnd
 * 2. ABORT TASK SET
 * 	This function cannot be directly performed since the Mid-Level
 * 	has no knowledge of what Initiators are logged in to the front-
 * 	end. This functionality can be achieved by issuing ABORTS to
 * 	various commands identified within the Task Set by the front
 * 	-end. value = N/A
 * 3. CLEAR ACA
 * 	I dont understand this one
 * 4. CLEAR TASK SET
 * 	Implement similar to ABORT Task Set - I know that all semantics
 * 	for this cannot be implemented (Any ideas ?)
 * 5. LUN RESET
 * 	set value = pointer to LUN
 * 6. TARGET RESET
 * 	set value = NULL
 */
struct SM *rx_task_mgmt_fn(struct STD *dev, int fn, void *value)
{
	unsigned long flags;
	Target_Scsi_Message *msg;

	if ((fn < TMF_ABORT_TASK) && (fn > TMF_TASK_REASSIGN)) {
		printk
		    ("rx_task_mgmt_fn: Invalid value %d for Task Mgmt function\n",
		     fn);
		return NULL;
	}
	if ((fn == TMF_ABORT_TASK_SET) || (fn == TMF_CLEAR_ACA)
	    || (fn == TMF_CLEAR_TASK_SET)) {
		printk
		    ("rx_task_mgmt_fn: Task mgmt function %d not implemented\n",
		     fn);
		return NULL;
	}
	if ((fn == TMF_ABORT_TASK) && (value == NULL)) {
		printk("rx_task_mgmt_fn: Cannot abort a NULL command\n");
		return NULL;
	}

	msg = kmalloc(sizeof(Target_Scsi_Message), GFP_ATOMIC);
	if (!msg) {
		printk("rx_task_mgmt_fn: No space for scsi message\n");
		return NULL;
	}

	msg->next = NULL;
	msg->prev = NULL;
	msg->device = dev;
	msg->value = value;
	msg->fn = fn;

	spin_lock_irqsave(&target_data.msg_lock, flags);

	if (!target_data.msgq_start) {
		target_data.msgq_start = target_data.msgq_end = msg;
	} else {
		target_data.msgq_end->next = msg;
		target_data.msgq_end = msg;
	}

	spin_unlock_irqrestore(&target_data.msg_lock, flags);

	/* wake up scsi_target_process_thread */
	if (atomic_read(&target_data.target_sem.count) <= 0) {
		up(&target_data.target_sem);
	}
	return msg;
}

#if defined (FILEIO) || defined (GENERICIO)
/*
 * build_filp_table: builds up a table of open file descriptors.
 * INPUT: nothing
 * OUTPUT: 0 if all ok, < 0 if there was trouble
 */
static int build_filp_table(void)
{
	char *tmp;
	int error;
	int targ, lun, max_files, max_luns;
#ifdef __KERNEL__
	mm_segment_t old_fs;
#endif
	struct file *dev_file;
	struct target_map_item *this_item;

	max_files = MAX_FILE_TARGETS;
	if (max_files > MAX_TARGETS)
		max_files = MAX_TARGETS;
	max_luns = MAX_FILE_LUNS;
	if (max_luns > MAX_LUNS)
		max_luns = MAX_LUNS;
	for (targ = 0; targ < max_files; targ++) {
		for (lun = 0; lun < max_luns; lun++) {
			this_item = &target_map[targ][lun];
#if defined (GENERICIO)
			if (targ == 0)
				sprintf(this_item->file_name, "/dev/sg%d",
					lun);
			else
				sprintf(this_item->file_name, "dev/sg%d%d",
					targ, lun);
#endif
#if defined (FILEIO)
			sprintf(this_item->file_name,
				"scsi_disk_file_%d_%d", targ, lun);
#endif
			printk("opening device %s\n",
			       this_item->file_name);

#ifdef __KERNEL__
			old_fs = get_fs();
			set_fs(get_ds());
#endif

			tmp = getname(this_item->file_name);

#ifdef __KERNEL__
			set_fs(old_fs);
			error = PTR_ERR(tmp);
			if (IS_ERR(tmp)) {
				printk
				    ("build_filp_table: getname returned error %d\n",
				     error);
				return -1;
			}
#endif

			/* open a file or a scsi_generic device */
#ifdef GENERICIO
			dev_file =
			    filp_open(tmp, O_RDWR | O_NONBLOCK, 0600);
#endif

#ifdef FILEIO
			dev_file =
			    filp_open(tmp, O_RDWR | O_NONBLOCK | O_CREAT,
				      0600);
#endif

			putname(tmp);
			if (IS_ERR(dev_file)) {
				error = PTR_ERR(dev_file);
				if (targ + lun == 0) {
					printk
					    ("build_filp_table: filp_open returned error %d\n",
					     error);
					return -1;
				} else
					goto out_of_loop;
			} else {
				/* mark this device "in-use" and save info about it */
				this_item->the_file = dev_file;
				this_item->max_blocks = FILESIZE;
				this_item->bytes_per_block = BLOCKSIZE;
				this_item->in_use = 1;

				printk("opened file %s (filp=%p) as "
				       "id: %d lun: %d\n",
				       this_item->file_name,
				       this_item->the_file,
				       targ, lun);

				target_count++;
			}
		}
	}

      out_of_loop:

#ifdef GENERICIO
	/* also we spawn a second thread to receive signals */
	init_MUTEX_LOCKED(&target_data.signal_sem);
	init_MUTEX_LOCKED(&target_data.sig_thr_sem);
	kernel_thread((int (*)(void *)) signal_process_thread, NULL, 0);
#endif
	return 0;
}

/*
 * close_filp_table: to close all the open file descriptors
 * INPUT: None
 * OUTPUT: None
 */
static void close_filp_table(void)
{
	__u32 targ, lun;
	struct target_map_item *this_item;

	write_lock(&target_map_rwlock);
	for (targ = 0; targ < MAX_TARGETS; targ++) {
		for (lun = 0; lun < MAX_LUNS; lun++) {
			this_item = &target_map[targ][lun];
			if (this_item->the_file) {
				filp_close(this_item->the_file, NULL);
				this_item->the_file = NULL;
				this_item->in_use = 0;
			}
		}
	}
	write_unlock(&target_map_rwlock);
}

#endif				/* FILEIO || GENERICIO */

/*
 * : allocates scatter-gather buffers for the received command
 * The assumption is that all front-ends use scatter-gather and so do
 * the back-ends ... this may change.
 * INPUT: command for which space has to be allocated, space needed
 * rw_buffers parameter is 0 for anything other than SCSI READ/WRITE
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int get_space(Target_Scsi_Cmnd * cmnd, __u32 space /* in bytes */ ,
	      int rw_buffers)
{
	int ret = -1;
	Scsi_Request *req = cmnd->req;
	/* We assume that scatter gather is used universally */
	struct scatterlist *st_list;
	__u32 buff_needed, i;
	__u32 count;
	__u32 page_order, page_block;
#if defined(__KERNEL__)
	struct page *page;
#endif /* __KERNEL__ */

/*
	printk("Entering %s space=%u, rw_buffers=%d\n", __func__,
	       space, rw_buffers);
*/

	/* first find the biggest allocation order that is necessary */
#ifdef MEMORYIO
	if (rw_buffers) {
		page_order = PAGE_ALLOC_ORDER;
		page_block = PAGE_ALLOC_BLOCK;
	} else
#endif				/* MEMORYIO */
	{
		if (space > (PAGE_ALLOC_BLOCK / 2)) {
			page_order = PAGE_ALLOC_ORDER;
			page_block = PAGE_ALLOC_BLOCK;
		} else {
			for (page_order = 0, page_block = PAGE_SIZE;
			     space > page_block;
			     page_order++, page_block <<= 1) {
			}
		}
	}
	cmnd->page_alloc_order = page_order;

	/* now get enough scatter gather entries (normally need only 1) */
	if (!(buff_needed = req->sr_use_sg)) {
		for (buff_needed = 1, i = space;
		     i > page_block; buff_needed++, i -= page_block) {
		}
		req->sr_use_sg = buff_needed;
	}
#ifdef DEBUG_HANDLE_CMD
	printk("get space %u order %u needed %u\n",
	       space, page_order, buff_needed);
#endif

	/* now allocate space for the st_list */
	if (!(st_list = req->sr_buffer)) {
		req->sr_bufflen = space;
		req->sr_sglist_len =
		    buff_needed * sizeof(struct scatterlist);
		if (buff_needed > MAX_ST_LIST_SLOTS) {
#ifdef DEBUG_HANDLE_CMD
			printk("allocate %u bytes for %u st_list slots\n",
			       req->sr_sglist_len, buff_needed);
#endif /* DEBUG_HANDLE_CMD */
			st_list = kmalloc(req->sr_sglist_len, GFP_ATOMIC);
			if (!st_list) {
				printk
				    ("get_space: No space for st_list\n");
				goto out;
			}
		} else {
			/* use the preallocated scatter-gather list array in the command */

#ifdef DEBUG_HANDLE_CMD
			printk
			    ("use prealloc s-g-list %p for %u bytes in %u st_list slots\n",
			     cmnd->st_list, req->sr_sglist_len,
			     buff_needed);
#endif

			st_list = cmnd->st_list;
		}
		req->sr_buffer = st_list;
		sg_init_table(st_list, buff_needed);
	}

	/* finally get space for the buffers themselves */
	for (i = 0, count = space; i < buff_needed;
	     i++, count -= page_block) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		if (sg_page(&st_list[i]))
			continue;
#ifdef MEMORYIO
		if (rw_buffers) {
			memcpy(st_list + i, &memory_st_list,
			       sizeof(struct scatterlist));
		} else
#endif				/* MEMORYIO */
		{
			page = alloc_pages(GFP_ATOMIC, page_order);
			if (!page) {
				printk(KERN_ERR "%s: failed to allocate "
				       "memory for sg entry\n",
				       __func__);
				goto out;
			}
			sg_set_page(&st_list[i],
			            page,
				    1 << page_order, 0);

			if (!sg_page(&st_list[i])) {
				/* this is probably an extra check now
				 * that there is a check on the
				 * alloc_pages() call. */
				printk(KERN_ERR "%s: failed to allocate "
				       "memory for sg entry\n",
				       __func__);
				goto out;
			}
			/*
			scatterlist_address(st_list + i) =
			    (dma_addr_t) (unsigned long) kmap(sg_page(&st_list[i]));
			    */
			/* this should always be 0 */
			/*
			st_list[i].offset = (unsigned long)
			    (sg_virt(st_list + i) & ~PAGE_MASK);
			    */
		}
		if (count > page_block)
			scatterlist_length(st_list + i) = page_block;
		else
			scatterlist_length(st_list + i) = count;
#else				/* K26 */
		if (scatterlist_address(&st_list[i]))
			continue;
#ifdef MEMORYIO
		if (rw_buffers) {
			memcpy(st_list + i, &memory_st_list,
			       sizeof(struct scatterlist));
		} else
#endif				/* MEMORYIO */
		{
			scatterlist_address(&st_list[i]) =
			    (__u8 *) __get_free_pages(GFP_ATOMIC,
						      page_order);

			if (!scatterlist_address(&st_list[i])) {
				printk("get space[%u] No pages order %u\n",
				       i, page_order);
				goto out;
			}
		}
		if (count > page_block) {
			scatterlist_length(&st_list[i]) = page_block;
		} else {
			scatterlist_length(&st_list[i]) = count;
		}
#endif				/* K26 */

#ifdef DEBUG_HANDLE_CMD
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		if (page_order) {
			printk("get space[%u] len %u order %u addr %p\n",
			       i, scatterlist_length(st_list + i), page_order,
			       (void *) (unsigned long) sg_virt(st_list + i));
		}
#else				/* K26 */
		if (page_order) {
			printk("get space[%u] len %u order %u addr %p\n",
			       i, scatterlist_length(&st_list[i]),
			       page_order,
			       scatterlist_address(&st_list[i]));
		}
#endif				/* K26 */
#endif				/* DEBUG_HANDLE_CMD */
	}

	ret = 0;

out:
/*
	printk("Leaving %s space=%u, rw_buffers=%d, ret=%d\n", __func__,
	       space, rw_buffers, ret);
*/


	return ret;
}

#if !defined(DISKIO) || !defined(TRUST_CDB)
/*
 * Returns size of space allocated for LUN list > 0 if all ok,
 * -1 on error after giving message
 */
int allocate_report_lun_space(Target_Scsi_Cmnd * cmnd)
{
	int i, luns, size;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef DEBUG_HANDLE_CMD
	printk("REPORT_LUNS received\n");
#endif

	/* perform checks on Report LUNS - LATER */
	if (cmnd->req->sr_cmnd[2] != 0) {
		printk("%s Select_Report in report_luns not zero\n",
		       current->comm);
	}

	/* set data direction */
	cmnd->req->sr_data_direction = SCSI_DATA_READ;

	/* get length */
	if (cmnd->target_id >= MAX_TARGETS) {
		printk("%s target id %u >= MAX_TARGETS %u\n",
		       current->comm, cmnd->target_id, MAX_TARGETS);
		return -1;
	}

	luns = 0;
	read_lock(&target_map_rwlock);
	for (i = 0; i < MAX_LUNS; i++) {
		if (target_map[cmnd->target_id][i].in_use)
			luns++;
	}
	read_unlock(&target_map_rwlock);

	if (luns == 0) {
		printk("%s No luns in use for target id %u\n",
		       current->comm, cmnd->target_id);
		return -1;
	}

	printk("%s REPORT_LUNS: target id %u reporting %d luns\n",
	       current->comm, cmnd->target_id, luns);

	/* allocate space */
	size = luns * 8;
	if (get_space(cmnd, size + 8, 0)) {
		printk("%s get_space returned error for %d\n",
		       current->comm, cmnd->id);
		return -1;
	}

	return size;
}

/*
 * get_allocation_length: This function looks at the received command
 * and calculates the size of the buffer to be allocated in order to
 * execute the command.
 * INPUT: pointer to command received
 * OUTPUT: buffer needing to be allocated or < 0 if there is an error
 */
static __u32 __attribute__ ((no_instrument_function))
    get_allocation_length(__u8 * cmd)
{
	__u32 err = 0;

	switch (cmd[0]) {
	case INQUIRY:
	case MODE_SENSE:
	case REQUEST_SENSE:
		{
			err = cmd[ALLOC_LEN_6];
#ifdef DEBUG_ALLOCN_LEN
			printk
			    ("get_allocation_length: INQUIRY/MODE SENSE/MODE SELECT length %d\n",
			     err);
#endif
			break;
		}

	case WRITE_10:
	case READ_10:
	case VERIFY:
		{
			err =
			    (cmd[ALLOC_LEN_10] << BYTE) +
			    cmd[ALLOC_LEN_10 + 1];
			err *= BLOCKSIZE;
#ifdef DEBUG_ALLOCN_LEN
			printk
			    ("get_allocation_length: READ_10/WRITE_10 length %d\n",
			     err);
#endif
			break;
		}

		/* cdeng, August 24 2002, Report luns */
	case REPORT_LUNS:
		{
			/* Bjorn Thordarson, 10 May 2004 */
			/*****
			err = (cmd[6] << 24) + (cmd[7] << 16) + (cmd[8] << 8) + cmd[9];
			*****/
			err = 0;

#ifdef DEBUG_ALLOCN_LEN
			printk
			    ("get_allocation_length: REPORT_LUNS length %d - FIXME\n",
			     err);
#endif
			break;
		}

#ifdef TAPE_DEVICE
	case READ_6:
	case WRITE_6:
		{
			err = (cmd[TAPE_ALLOC_LEN] << (BYTE * 2)) +
			    (cmd[TAPE_ALLOC_LEN + 1] << BYTE) +
			    cmd[TAPE_ALLOC_LEN + 2];
			if (cmd[1] & 1) {
				err *= BLOCKSIZE;	/*ramesh - need to check the block size, right now it is fixed */
			}
#ifdef DEBUG_ALLOCN_LEN
			printk
			    ("get_allocation_length: READ_6/WRITE_6 length %d\n",
			     err);
#endif
			break;
		}
#else				/* TAPE_DEVICE */

	case READ_6:
	case WRITE_6:
		{
			err = cmd[ALLOC_LEN_6] * BLOCKSIZE;
			if (err == 0)
				err = 256 * BLOCKSIZE;
#ifdef DEBUG_ALLOCN_LEN
			printk
			    ("get_allocation_length: READ_6/WRITE_6 length %d\n",
			     err);
#endif
			break;
		}
#endif				/* TAPE_DEVICE */

	default:
		{
			printk
			    ("get_allocation_length: Unknown command 0x%02x\n",
			     cmd[0]);
			break;
		}
	}

	return err;
}

/* perform checks on and sets up newly-received READ */
int __attribute__ ((no_instrument_function))
    setup_new_read(Target_Scsi_Cmnd * cmnd)
{
	int err = 1;
	int to_xfer;

#ifdef DEBUG_HANDLE_CMD
	printk("%s READ_%d received\n", pr_label,
	       cmnd->req->sr_cmnd[0] == READ_6 ? 6 : 10);
#endif

	/* set data direction for the READ */
	cmnd->req->sr_data_direction = SCSI_DATA_READ;

	/* get length */
	to_xfer = get_allocation_length(cmnd->req->sr_cmnd);

	if (to_xfer < 0) {
		printk
		    ("%s handle_cmd: get_allocation_length returned error %d "
		     "for %d\n", pr_label, to_xfer, cmnd->id);
		err = -1;
#ifdef GENERICIO
	} else
	    if (get_space(cmnd, to_xfer / BLOCKSIZE * cmnd->blk_size, 1)) {
#else
	} else if (get_space(cmnd, to_xfer, 1)) {
#endif
		err = -1;
	}

	return err;
}

/* perform checks on and sets up newly-received WRITE */
int __attribute__ ((no_instrument_function))
    setup_new_write(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	int to_xfer;

#ifdef DEBUG_HANDLE_CMD
	printk("%s WRITE_%d received\n", pr_label,
	       cmnd->req->sr_cmnd[0] == WRITE_6 ? 6 : 10);
#endif

	/* set data direction for the WRITE */
	cmnd->req->sr_data_direction = SCSI_DATA_WRITE;

	/* get length */
	to_xfer = get_allocation_length(cmnd->req->sr_cmnd);

	if (to_xfer < 0) {
		printk("%s handle_cmd: get_allocation_length returned "
		       "error %d for %d\n", pr_label, to_xfer, cmnd->id);
		err = -1;
#ifdef GENERICIO
	} else
	    if (get_space(cmnd, to_xfer / BLOCKSIZE * cmnd->blk_size, 1)) {
#else
	} else if (get_space(cmnd, to_xfer, 1)) {
#endif
		/* come back later */
		printk("%s: WARNING, get_space() cmnd=%p failed, coming "
			"back later\n", __func__, cmnd);
	} else {
#if defined(DEBUG_HANDLE_CMD)
		printk("%s: get_space() cmnd=%p succeeded\n", __func__, cmnd);
#endif /* DEBUG_HANDLE_CMD */
		cmnd->state = ST_PENDING;
 	}
	return err;
}

/* perform checks on and sets up newly-received command other than READ, WRITE*/
int __attribute__ ((no_instrument_function))
    setup_new_command(Target_Scsi_Cmnd * cmnd, int *to_xfer)
{
	int err = 1;

	/* set data direction for the command */
	cmnd->req->sr_data_direction = SCSI_DATA_READ;

	/* get length */
	*to_xfer = get_allocation_length(cmnd->req->sr_cmnd);

	if (*to_xfer < 0) {
		printk
		    ("%s handle_cmd: get_allocation_length returned error %d "
		     "for %d\n", pr_label, *to_xfer, cmnd->id);
		err = -1;
	} else if (get_space(cmnd, *to_xfer, 0)) {
		err = 0;	/* come back later */
	}
	return err;
}

#endif				/* !DISKIO || !TRUST_CDB */

#if defined (FILEIO) || defined (MEMORYIO)
/*
 * get_inquiry_response: This function fills in the buffer to respond to
 * a received SCSI INQUIRY. This function is relevant when the emulator
 * is responding to commands itself and not transmitting it onto a SCSI
 * HBA attached to the system
 * INPUT: Scsi_Request ptr to the inquiry, length of expected response
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
/* Bjorn Thordarson, 9-May-2004 */
int get_inquiry_response(Scsi_Request * req, int len, int type)
{
	__u8 *inq;
	__u8 *ptr, local_buffer[36];

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer =
	    (void *) (unsigned long) sg_virt((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer =
	    scatterlist_address((struct scatterlist *) req->sr_buffer);
#endif

	/* SPC-2 says in section 7.3.2 on page 82:
	 *  "The standard INQUIRY data shall contain at least 36 bytes."
	 */
	if (len >= 36) {
		/* caller's buffer is big enough to hold all 36 bytes */
		ptr = buffer;
		memset(buffer, 0x00, len);
	} else {
		/* fill a local 36-byte buffer */
		ptr = local_buffer;
		memset(local_buffer, 0x00, 36);
	}
	/* Bjorn Thordarson, 9-May-2004 */
	/*****
	ptr[0] = TYPE_DISK;
	*****/
	ptr[0] = type;
	ptr[2] = 4;		/* Device complies to this standard - SPC-2 */
	ptr[3] = 2;		/* data in format specified in this standard */
	ptr[4] = 31;		/* n - 4 = 35 - 4 = 31 for full 36 byte data */
	ptr[6] = 0x80;		/* BQue = 1 basic task management supported */

	/* SPC-2 says in section 7.3.2 on page 86:
	 * "ASCII data fields shall contain only graphic codes (i.e., code
	 * values 20h through 7Eh). Left-aligned fields shall place any
	 * unused bytes at the end of the field (highest offset) and the
	 * unused bytes shall be filled with space characters (20h)."
	 */

	/* 8 byte ASCII Vendor Identification of the target - left aligned */
	memcpy(&ptr[8], "UNH-IOL ", 8);

	/* 16 byte ASCII Product Identification of the target - left aligned */
#if defined (FILEIO)
	memcpy(&ptr[16], "file-mode target", 16);
#else
	memcpy(&ptr[16], "in-memory target", 16);
#endif

	/* 4 byte ASCII Product Revision Level of the target - left aligned */
	memcpy(&ptr[32], "1.2 ", 4);

	if (len < 36) {
		/* fill as much as possible of caller's too-small buffer */
		memcpy(buffer, ptr, len);
	}

	req->sr_result = DID_OK << 16;

	if (req->sr_allowed == 1) {

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		inq =
		    (void *) (unsigned long) sg_virt((struct scatterlist *) req->
					    sr_buffer);
#else
		inq =
		    scatterlist_address((struct scatterlist *) req->
					sr_buffer);
#endif
		inq[0] = 0x7f;
	}

	return 0;
}

/*
 * get_read_capacity_response: This function fills up the buffer to
 * respond to a received READ_CAPACITY. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the READ_CAPACITY
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
/* Bjorn Thordarson, 9-May-2004 */
int get_read_capacity_response(Target_Scsi_Cmnd * cmnd)
{
	__u32 blocksize = BLOCKSIZE, nblocks;
	__u8 *buffer;

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

/* RDR */
	nblocks = FILESIZE;
#ifdef FILEIO
	read_lock(&target_map_rwlock);
	if (cmnd->target_id >= MAX_TARGETS || cmnd->lun >= MAX_LUNS
	    || !target_map[cmnd->target_id][cmnd->lun].in_use) {
		printk("%s target id %u lun %u not in use\n",
		       current->comm, cmnd->target_id, cmnd->lun);
		read_unlock(&target_map_rwlock);
		return -1;
	}

	nblocks = target_map[cmnd->target_id][cmnd->lun].max_blocks;
	read_unlock(&target_map_rwlock);
#endif				/* FILEIO */

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	buffer =
	    (void *) (unsigned long) sg_virt((struct scatterlist *) cmnd->req->
				    sr_buffer);
#else				/* K26 */
	buffer =
	    scatterlist_address((struct scatterlist *) cmnd->req->
				sr_buffer);
#endif				/* K26 */

	memset(buffer, 0x00, READ_CAP_LEN);

	/* last block on the disk is (nblocks-1) */
	buffer[0] = ((nblocks - 1) >> (BYTE * 3)) & 0xFF;
	buffer[1] = ((nblocks - 1) >> (BYTE * 2)) & 0xFF;
	buffer[2] = ((nblocks - 1) >> (BYTE)) & 0xFF;
	buffer[3] = (nblocks - 1) & 0xFF;

	buffer[4] = (blocksize >> (BYTE * 3)) & 0xFF;
	buffer[5] = (blocksize >> (BYTE * 2)) & 0xFF;
	buffer[6] = (blocksize >> BYTE) & 0xFF;
	buffer[7] = blocksize & 0xFF;

	cmnd->req->sr_result = DID_OK << 16;
	return 0;
}

/* cdeng August 24 2002
 * get_mode_sense_response: This function fills up the buffer to
 * respond to a received MODE_SENSE. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the MODE_SENSE
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int get_mode_sense_response(Scsi_Request * req, __u32 len)
{
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer =
	    (void *) (unsigned long) sg_virt((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer =
	    scatterlist_address((struct scatterlist *) req->sr_buffer);
#endif

	memset(buffer, 0x00, len);

	buffer[0] = 0x0b;	// no. of bytes that follow == 11

	buffer[3] = 0x08;	// block descriptor length

	buffer[10] = 0x02;	// density code and block length

	req->sr_result = DID_OK << 16;
	return 0;
}
#endif				/* defined (FILEIO) || defined (MEMORYIO) */

#ifndef DISKIO
/* cdeng August 24 2002
 * get_report_luns_response: This function fills up the buffer to
 * respond to a received REPORT_LUNS. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the REPORT_LUNS
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
int get_report_luns_response(Target_Scsi_Cmnd * cmnd, __u32 len)
{
	int i;
	__u8 *limit, *next_slot;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer = (void *)
	    (unsigned long) sg_virt((struct scatterlist *) cmnd->req->sr_buffer);
#else
	__u8 *buffer =
	    scatterlist_address((struct scatterlist *) cmnd->req->
				sr_buffer);
#endif

	next_slot = buffer + 8;	/* first lun goes here */
	limit = next_slot + len;	/* address after end of entire list */
	memset(buffer, 0x00, len + 8);

	/* SAM-2, section 4.12.2 LUN 0 address
	 * "To address the LUN 0 of a SCSI device the peripheral device
	 * address method shall be used."
	 *
	 * What this means is that whenever the LUN is 0, the full 8 bytes of
	 * the LUN field in the iscsi PDU will also be 0. Because we have
	 * already zeroed out the entire buffer above,
	 * we don't need to do anything else here for lun 0.
	 */

	if (cmnd->target_id < MAX_TARGETS) {
		read_lock(&target_map_rwlock);
		for (i = 0; i < MAX_LUNS && next_slot < limit; i++) {
			if (target_map[cmnd->target_id][i].in_use) {
				pack_lun(i, 0, next_slot);
				next_slot += 8;;
			}
		}
		read_unlock(&target_map_rwlock);
	}

	/* lun list length */
	((__u32 *) buffer)[0] = cpu_to_be32(len);

#ifdef DEBUG_HANDLE_CMD
	/* dump out the complete REPORT LUNS response buffer */
	dump_buffer(buffer, len + 8);
#endif

	/* change status */
	cmnd->state = ST_DONE;
	cmnd->req->sr_result = DID_OK << 16;
	return 0;
}
#endif				/* !DISKIO */

/*
 * hand_to_front_end: the scsi command has been tackled by the mid-level
 * and is now ready to be handed off to the front-end either because it
 * is done or because data is needed to process this command.
 * This function is separated only because it makes
 * the main code a lot cleaner.
 * INPUT: struct scsi_cmnd that needs to be handed off
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int hand_to_front_end(Target_Scsi_Cmnd * the_command)
{
	Scsi_Target_Device *curr_device;
	int retval = 0;
#ifdef GENERICIO
	__u8 *temp;
#endif

#ifndef __KERNEL__
	struct task_struct *current;

	current = pthread_getspecific(current_key);
#endif

#ifdef DEBUG_SCSI_THREAD
#ifdef K26
	printk("%s enter hand_to_front_end with irqs %sabled\n",
	       current->comm, irqs_disabled()? "dis" : "en");
#endif
#endif

	/* get the device template corresponding to the device_id */
	for (curr_device = target_data.st_device_list; curr_device != NULL;
	     curr_device = curr_device->next) {
		if (curr_device->dev_id == the_command->dev_id)
			break;
	}

	if (curr_device == NULL) {	/* huh - should not get here */
		/*
		 * That there is not a device with this id may mean
		 * one of two things - there has been a mistake or
		 * that the device is no longer there. For now, the
		 * former is more probable
		 */
		printk("%s hand_to_front_end: No device with dev_id %u\n",
		       current->comm, the_command->dev_id);
		retval = -1;
		goto out;
	}

	/*
	 * In the time that the command was processed, it could have
	 * been aborted/released. We need to check this. If so, then
	 * the command state is changed to ST_DEQUEUE and returned
	 */
	if (the_command->abort_code != CMND_OPEN) {
		the_command->state = ST_DEQUEUE;
	} else if (the_command->state == ST_DONE) {
#ifdef DEBUG_SCSI_THREAD
		printk
		    ("%s hand_to_front_end call xmit_response for %p id %d change "
		     "state from %d to %d\n", current->comm, the_command,
		     the_command->id, the_command->state, ST_HANDED);
#endif
		if ((curr_device->template) && (curr_device->template->xmit_response)) {
			/*
			 * if it is generic then we need to get the
			 * blocksize information from the response to
			 * the READ_CAPACITY
			 */
#ifdef GENERICIO
			if (the_command->cmd[0] == READ_CAPACITY) {
#ifdef K26
				temp = page_address(((struct scatterlist *)
						     the_command->req->
						     sr_buffer)->page);
#else				/* K26 */
				temp =
				    ((struct scatterlist *) the_command->
				     req->sr_buffer)->address;
#endif				/* K26 */
				if (the_command->target_id < MAX_TARGETS
				    && the_command->lun < MAX_LUNS) {
					write_lock(&target_map_rwlock);
					target_map[the_command->
						   target_id][the_command->
							      lun]
					    .bytes_per_block =
					    (temp[4] << (BYTE * 3)) +
					    (temp[5] << (BYTE * 2)) +
					    (temp[6] << (BYTE)) + temp[7];
					write_unlock(&target_map_rwlock);
				}
			}
#endif				/* GENERICIO */

			the_command->state = ST_HANDED;
			if (curr_device->template->xmit_response(the_command)) {
				printk("%s hand_to_front_end: Error in xmit_response for %p "
				       "id %d\n", current->comm, the_command,
				       the_command->id);
				retval = -1;
			}
		} else {
			printk("%s hand_to_front_end: No xmit_response function\n",
			       current->comm);
			retval = -1;
		}
	} else if (the_command->state == ST_PENDING) {
#ifdef DEBUG_SCSI_THREAD
		printk
		    ("%s hand_to_front_end: call to rdy_to_xfer for %p id: %d\n",
		     current->comm, the_command, the_command->id);
#endif
		if ((curr_device->template)
		    && (curr_device->template->rdy_to_xfer)) {
			the_command->state = ST_XFERRED;
			if (curr_device->template->
			    rdy_to_xfer(the_command)) {
				printk
				    ("%s hand_to_front_end: Error in rdy_to_xfer for %p "
				     "id %d\n", current->comm, the_command,
				     the_command->id);
				retval = -1;
			}
		} else {
			printk
			    ("%s hand_to_front_end: No rdy_to_xfer function\n",
			     current->comm);
			retval = -1;
		}
	} else {
		printk
		    ("%s hand_to_front_end: command %p id: %d bad state %d\n",
		     current->comm, the_command, the_command->id,
		     the_command->state);
		retval = -1;
	}

      out:

#ifdef DEBUG_SCSI_THREAD
#ifdef K26
	printk("%s exit hand_to_front_end with irqs %sabled\n",
	       current->comm, irqs_disabled()? "dis" : "en");
#endif
#endif

	return retval;
}

/*
 * abort_notify: This function is used to notify to the front end driver
 * that a command has been successfully aborted
 * INPUT: Target_Scsi_Cmnd
 * OUTPUT: none
 */
static int abort_notify(Target_Scsi_Message * msg)
{
	Target_Scsi_Cmnd *cmnd;
	Scsi_Target_Device *curr_device;

	if (msg && msg->value)
		cmnd = (Target_Scsi_Cmnd *) msg->value;
	else {
		printk("abort_notify: NULL cmnd in the msg\n");
		return -1;
	}

	/* get the device template corresponding to the device_id */
	for (curr_device = target_data.st_device_list; curr_device != NULL;
	     curr_device = curr_device->next) {
		if (curr_device->dev_id == cmnd->dev_id)
			break;
	}

	if (curr_device == NULL) {
		printk("abort_notify: No device with dev_id %u\n",
		       cmnd->dev_id);
		return -1;
	}

	if ((curr_device->template)
	    && (curr_device->template->task_mgmt_fn_done))
		curr_device->template->task_mgmt_fn_done(msg);
	else {
		printk
		    ("abort_notify: Unable to notify front end about abort\n");
		return -1;
	}

	return 0;
}

/*
 * aen_notify: This function is used to notify to the low-level driver
 * that task management function has been received and that they need to
 * contact all initiators logged in to them and let them know. NOTE: the
 * aen_notify will also call the device that sent the task management
 * function - it needs to inform other initiators (than the one that
 * originated the Task Management function) about the received Mgmt fn
 * INPUT: management function and lun
 * OUTPUT: None
 */
static void aen_notify(int fn, __u64 lun)
{
	Scsi_Target_Device *dev;

	for (dev = target_data.st_device_list; dev != NULL;
	     dev = dev->next) {
		if (dev->template && dev->template->report_aen)
			dev->template->report_aen(fn, lun);
		else
			printk("aen_notify: Unable to notify device %u\n",
			       dev->dev_id);
	}
}


/* Define the functions that handle the proc interface */
#ifdef CONFIG_PROC_FS

/* dispatch_scsi_info is the central dispatcher
 * It is the interface between the proc-fs and the SCSI subsystem code
 */
static int
proc_scsi_target_read(char *buffer, char **start, off_t offset, int length,
		      int *eof, void *data)
{
	Scsi_Target_Device *tpnt = data;
	int n = 0;

	/*****
	printk("Entering proc_scsi_target_read\n");
	printk("buffer %p offset %d, length %d, eof %p, data %p\n",
			buffer, (int) offset, length, eof, tpnt);
	if (tpnt) {
		printk("template %p\n", tpnt->template);
		if (tpnt->template) {
			printk("proc_info %p\n", tpnt->template->proc_info);
		}
	}
	*****/

	if (tpnt->template->proc_info == NULL) {
/*		n = generic_proc_info(buffer, start, offset, length,
				      hpnt->hostt->info, hpnt); */
	} else {
		n = (tpnt->template->
		     proc_info(buffer, start, offset, length, 0, 0));
	}
	*eof = (n < length);
	return n;
}

#define PROC_BLOCK_SIZE (3*1024)	/* 4K page size, but our output routines
					 * use some slack for overruns
					 */

static int
proc_scsi_target_write(struct file *file, const char *buf,
		       unsigned long count, void *data)
{
	Scsi_Target_Device *tpnt = data;
	ssize_t ret = 0;
	char *page;
	char *start;

	if (count > PROC_BLOCK_SIZE)
		return -EOVERFLOW;

	if (!(page = (char *) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	if (copy_from_user(page, buf, count))
		return -EFAULT;

	if (tpnt->template->proc_info == NULL)
		ret = -ENOSYS;
	else
		ret =
		    tpnt->template->proc_info(page, &start, 0, count,
					      tpnt->dev_id, 1);
	free_page((ulong) page);
	return ret;
}

/*
 * Get the printable representation of a scsi device type.
 */
const char *printable_dev_type(char type)
{
	switch(type) {
	case TYPE_DISK:
		return "disk";
	case TYPE_TAPE:
		return "tape";
	case TYPE_PRINTER:
		return "printer";
	case TYPE_PROCESSOR:
		return "processor";
	case TYPE_WORM:
		return "worm";
	case TYPE_ROM:
		return "rom";
	case TYPE_SCANNER:
		return "scanner";
	case TYPE_MOD:
		return "mod";
	case TYPE_MEDIUM_CHANGER:
		return "medium_changer";
	case TYPE_COMM:
		return "comm";
	case TYPE_RAID:
		return "raid";
	case TYPE_ENCLOSURE:
		return "enclosure";
	case TYPE_RBC:
		return "rbc";
	case TYPE_NO_LUN:
		return "no_lun";
	default:
		return "<unknown>";
	}
}


#ifndef MEMORYIO
static int __attribute__ ((no_instrument_function))
    print_device_info(char *buffer, __u32 targ, __u32 lun,
		  struct target_map_item *this_item)
{
	int len = 0;

#if defined(DISKIO)
	struct scsi_device *the_device;

	if ((the_device = this_item->the_device)) {
		len =
		    sprintf(buffer,
			    "iscsi target %u: scsi device type %s, host %d, "
			    "channel %d, targetid %d, lun %d\n", targ,
			    printable_dev_type(the_device->type),
			    the_device->host->host_no, the_device->channel,
			    the_device->id, the_device->lun);
	} else {
		len =
		    sprintf(buffer, "iscsi target %u: no device!\n", targ);
	}
#endif				/* DISKIO */

#if defined(FILEIO) || defined(GENERICIO)
	if (this_item->the_file) {
		len =
		    sprintf(buffer,
			    "iscsi target %u lun %u: file %s, %u blocks, "
			    "%u bytes/block\n", targ, lun,
			    this_item->file_name, this_item->max_blocks,
			    this_item->bytes_per_block);
	} else {
		len =
		    sprintf(buffer, "iscsi target %u lun %u: no file!\n",
			    targ, lun);
	}
#endif				/* FILEIO || GENERICIO */

	return len;
}
#endif				/* !MEMORYIO */

static int
scsi_target_proc_info(char *buffer, char **start, off_t offset, int length)
{
	int len = 0;
#ifndef MEMORYIO
	struct target_map_item *this_item;
#endif

	len = sprintf(buffer, "unh_scsi_target using %s\n", pr_label);
	len +=
	    sprintf(buffer + len,
		    "%d target devices currently configured\n",
		    target_count);

#if defined(MEMORYIO)
	len +=
	    sprintf(buffer + len,
		    "%d configured iscsi target numbers: 0..%d\n",
		    MAX_TARGETS, MAX_TARGETS - 1);
	len +=
	    sprintf(buffer + len,
		    "each iscsi target has %d configured luns: " "0..%d\n",
		    MAX_LUNS, MAX_LUNS - 1);
#else				/* MEMORYIO */
	read_lock(&target_map_rwlock);

#if defined(DISKIO) && defined(TRUST_CDB)
	list_for_each_entry(this_item, &target_map_list, link) {
		if (this_item->in_use) {
			len +=
			    print_device_info(buffer + len,
					      this_item->target_id, 0,
					      this_item);
		}
	}
#else				/* DISKIO && TRUST_CDB */
	{
		__u32 targ, lun;

		for (targ = 0; targ < MAX_TARGETS; targ++) {
			for (lun = 0; lun < MAX_LUNS; lun++) {
				this_item = &target_map[targ][lun];
				if (this_item->in_use) {
					len +=
					    print_device_info(buffer + len,
							      targ, lun,
							      this_item);
				}
			}
		}
	}
	len +=
	    sprintf(buffer + len,
		    "%d configurable iscsi target numbers: " "0..%d\n",
		    MAX_TARGETS, MAX_TARGETS - 1);
	len +=
	    sprintf(buffer + len,
		    "each iscsi target has %d configurable "
		    "luns: 0..%d\n", MAX_LUNS, MAX_LUNS - 1);
#endif				/* DISKIO && TRUST_CDB */

	read_unlock(&target_map_rwlock);
#endif				/* MEMORYIO */
	return len;
}


/* RDR */
#if defined(FILEIO) || defined(GENERICIO)

/* Open a new device */
static int open_device(char *device, struct file **dev_file)
{
#ifdef __KERNEL__
	mm_segment_t old_fs;
#endif
	char *tmp;
	int error;

	printk("open device %s\n", device);
	old_fs = get_fs();
	set_fs(get_ds());
	tmp = getname(device);
	set_fs(old_fs);
	error = PTR_ERR(tmp);

	if (IS_ERR(tmp)) {
		printk("open_device: getname error %d\n", error);
		return -1;
	}

/* open a scsi_generic device */
#ifdef GENERICIO
	*dev_file =
	    filp_open(tmp, O_RDWR | O_NONBLOCK | O_LARGEFILE, 0600);
#endif

#ifdef FILEIO
	*dev_file =
	    filp_open(tmp, O_RDWR | O_NONBLOCK | O_CREAT | O_LARGEFILE,
		      0600);
#endif

	putname(tmp);
	if (IS_ERR(*dev_file)) {
		error = PTR_ERR(*dev_file);
		printk("open_device: filp_open error %d\n", error);
		return -1;
	}

	return 0;
}

/*
 * add this file to the matrix of open devices
 * Returns 0 if all ok, -1 on any error after giving message
 */
static int
add_file(__u32 target_id, __u32 lun, char *file_name, __u32 max_blocks,
	 __u32 block_size)
{
	int retval = -1;
	struct file *dev_file;
	struct target_map_item *this_item;

	if (target_id >= MAX_TARGETS) {
		printk("targetid=%u not in range [0..%d]\n", target_id,
		       MAX_TARGETS - 1);
	} else if (lun >= MAX_LUNS) {
		printk("lun=%u not in range [0..%d]\n", lun, MAX_LUNS - 1);
	} else if (strlen(file_name) >= MAX_FILE_NAME) {
		printk("filename %s longer than %d characters\n",
		       file_name, MAX_FILE_NAME - 1);
	} else if (block_size < 512 || block_size > 65536) {
		printk("blocksize %u not in range [512..65536]\n",
		       block_size);
	} else {
		write_lock(&target_map_rwlock);
		this_item = &target_map[target_id][lun];
		if (this_item->in_use) {
			printk
			    ("targetid %u lun %u already in use with filename %s\n",
			     target_id, lun, this_item->file_name);
		} else if (open_device(file_name, &dev_file)) {
			printk("add_file: Trouble in open_device(%s)\n",
			       file_name);
		} else {
			this_item->the_file = dev_file;
			this_item->max_blocks = max_blocks;
			this_item->bytes_per_block = block_size;
			strcpy(this_item->file_name, file_name);
			this_item->in_use = 1;

			printk("added file %s, targetid %u, lun %u\n",
			       file_name, target_id, lun);
			retval = 0;
		}
		write_unlock(&target_map_rwlock);
	}
	return retval;
}

/*
 * delete this file from the matrix of open devices
 * Returns 0 if all ok, -1 on any error after giving message
 */
static int del_file(__u32 target_id, __u32 lun)
{
	int retval = -1;
	struct target_map_item *this_item;

	if (target_id >= MAX_TARGETS) {
		printk("targetid %u not in range [0..%d]\n", target_id,
		       MAX_TARGETS - 1);
	} else if (lun >= MAX_LUNS) {
		printk("lun %u not in range [0..%d]\n", lun, MAX_LUNS - 1);
	} else {
		write_lock(&target_map_rwlock);
		this_item = &target_map[target_id][lun];
		if (!this_item->in_use) {
			printk("targetid %u lun %u not in use\n",
			       target_id, lun);
		} else {
			if (this_item->the_file) {
				filp_close(this_item->the_file, NULL);
				printk
				    ("closed file %s, targetid %u, lun %u\n",
				     this_item->file_name, target_id, lun);
			}
			this_item->the_file = NULL;
			this_item->in_use = 0;
			retval = 0;
		}
		write_unlock(&target_map_rwlock);
	}
	return retval;
}


static int
proc_scsi_target_gen_write(struct file *file, const char *buffer,
			   unsigned long length, void *data)
{
	char buf[512];
	char file_name[64];
	__u32 target_id, lun, max_blocks, block_size;

	if (length > 511)
		length = 511;
	if (copy_from_user(buf, buffer, length))
		return -EFAULT;
	buf[length] = 0;

	if (sscanf(buf, "addfile targetid=%u lun=%u filename=%63s "
		   "blocksinfile=%u blocksize=%u", &target_id, &lun,
		   file_name, &max_blocks, &block_size) == 5) {
		printk
		    ("addfile targetid=%u lun=%u filename=%s blocksinfile=%u "
		     "blocksize=%u\n", target_id, lun, file_name,
		     max_blocks, block_size);
		add_file(target_id, lun, file_name, max_blocks,
			 block_size);
	} else
	    if (sscanf(buf, "delfile targetid=%u lun=%u", &target_id, &lun)
		== 2) {
		printk("delfile targetid=%u lun=%u\n", target_id, lun);
		del_file(target_id, lun);
	} else {
		printk("scsi_target: Illegal input \"%s\"\n", buf);
	}

	return length;
}

#elif defined(DISKIO)

static int
proc_scsi_target_gen_write(struct file *file, const char *buffer,
			   unsigned long length, void *data)
{
	extern int add_scsi_device_by_addr(int, int, int, int );
	int err;
	char buf[512];
	__u32 host, chan, id, lun;

	if (length > 511)
		length = 511;
	if (copy_from_user(buf, buffer, length))
		return -EFAULT;
	buf[length] = 0;

	if (sscanf(buf, "adddevice host=%u chan=%u id=%u lun=%u",
	           &host, &chan, &id, &lun) == 4) {
		printk("adddevice host=%u chan=%u id=%u lun=%u",
		       host, chan, id, lun);
		err = add_scsi_device_by_addr(host, chan, id, lun);
		if (err)
			return err;
	} else
		printk("scsi_target: Illegal input \"%s\"\n", buf);

	return length;
}

#else				/* FILEIO || GENERICIO */

static int
proc_scsi_target_gen_write(struct file *file, const char *buf,
			   unsigned long length, void *data)
{
	printk("Function not implemented\n");
	return 0;
}

#endif				/* FILEIO || GENERICIO */

/* undoes the work of build_proc_target_dir_entries() */
static void remove_proc_target_dir_entries(Scsi_Target_Device * the_device)
{
	char name[12];

	sprintf(name, "%u", the_device->dev_id);
	remove_proc_entry(name, the_device->template->proc_dir);
	remove_proc_entry(the_device->template->name, proc_scsi_target);
}

/* Returns 0 if all ok, non-zero on error */
static int build_proc_target_dir_entries(Scsi_Target_Device * the_device)
{
	struct proc_dir_entry *p;
	char name[12];

	/* create the proc directory entry for the device
	 * this will be something like "/proc/scsi_target/iscsi_target"
	 */
	the_device->template->proc_dir =
	    proc_mkdir(the_device->template->name, proc_scsi_target);

	if (!the_device->template->proc_dir)
		return -ENOMEM;

	printk("created directory /proc/scsi_target/%s\n",
	       the_device->template->name);

	/* create the proc file entry for the device
	 * this will be something like "/proc/scsi_target/iscsi_target/0"
	 */
	sprintf(name, "%u", the_device->dev_id);
	p = create_proc_read_entry(name, S_IFREG | S_IRUGO | S_IWUSR,
				   the_device->template->proc_dir,
				   proc_scsi_target_read,
				   (void *) the_device);
	if (!p) {
		printk(KERN_ERR "Unable to register SCSI target %s in "
		       "/proc/scsi_target/%s\n", name,
		       the_device->template->name);
		remove_proc_entry(the_device->template->name,
				  proc_scsi_target);
		return -ENOMEM;
	}

	printk("created file /proc/scsi_target/%s/%s\n",
	       the_device->template->name, name);
	p->write_proc = proc_scsi_target_write;
	return 0;
}

#endif				/* CONFIG_PROC_FS */

#ifdef __KERNEL__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)

#include <scsi/scsi_device.h>	/* scsi_execute_async */

/*
 * Re-implementation of scsi_do_req() to use scsi_execute_async().
 *
 * NOTE: the `done' parameter differs from the pre-2.6.18 version in what it
 * takes for arguments.  Functions being passed here will need to have two
 * seperate implementations based on kernel version.
 */
void scsi_do_req(struct scsi_request *sreq, const unsigned char *cmnd,
		 void *buffer, unsigned bufflen,
		 void (*done) (void *, char *, int, int),
		 int timeout, int retries)
{
	scsi_execute_async(sreq->sr_device, cmnd, COMMAND_SIZE(cmnd[0]),
			   sreq->sr_data_direction, buffer, bufflen,
			   sreq->sr_use_sg, timeout, retries, sreq,
			   done, GFP_KERNEL);
}

/*
 * Re-implementation of scsi_allocate_request().
 */
struct scsi_request *scsi_allocate_request(struct scsi_device *sdev,
					   gfp_t gfp_mask)
{
	struct scsi_request *sreq;

	sreq = kzalloc(sizeof(*sreq), gfp_mask);
	if (likely(sreq))
		sreq->sr_device = sdev;

	return sreq;
}

/*
 * Re-implementation of scsi_release_request().
 */
void scsi_release_request(struct scsi_request *sreq)
{
	kfree(sreq);
}

#endif				/* >= 2.6.18 */
#endif

#ifdef MODULE
MODULE_AUTHOR("Bob Russell");
MODULE_DESCRIPTION("UNH");
MODULE_SUPPORTED_DEVICE("sd");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
EXPORT_SYMBOL(scsi_release);
EXPORT_SYMBOL(scsi_target_done);
EXPORT_SYMBOL(scsi_rx_data);
EXPORT_SYMBOL(register_target_template);
EXPORT_SYMBOL(deregister_target_template);
EXPORT_SYMBOL(register_target_front_end);
EXPORT_SYMBOL(deregister_target_front_end);
EXPORT_SYMBOL(rx_task_mgmt_fn);
EXPORT_SYMBOL(rx_cmnd);
EXPORT_SYMBOL(target_in_use);
EXPORT_SYMBOL(target_count);
EXPORT_SYMBOL(target_info);
#endif				/* K26 */
