/* unh_main_scsi_target.c */

#define _POSIX_C_SOURCE 200112L

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sched.h>
#include <pthread.h>

#include "../common/param_scan.h"
#include "../common/user_space.h"
#include "../target/scsi_midlevel.h"
#include "../common/iscsi_common.h"
#include "../common/unh_threads.h"

extern int module_init_scsi_target_init_module(void);
extern int module_init_iscsi_target_module_init(void);
extern void module_exit_scsi_target_cleanup_module(void);
extern void module_exit_iscsi_target_module_cleanup(void);

#define BUFSIZE	512


/* the global key for each thread's private "current" */
pthread_key_t current_key;


/* cleanup handler for the global key,
 * called on the exit of thread having non-null value of its private "current"
 */
extern void free_current(void *current);



/* tokenizes inbuf into at most "limit" tokens, storing pointers to them in
 * argv[].  counting starts from "start", but a
 * token starting with '#' ends the input prematurely.
 * returns final value of "start" or 0 if no tokens or more than "limit" tokens.
 */
int tokenize(char *inbuf, char *argv[], int start, int limit)
{
	char *next;

	if ((next = strtok(inbuf, WHITE_SPACE)) == NULL || *next == '#') {
		/* no tokens at all in the string */
		return 0;
	}

	while (start < limit) {
		argv[start++] = next;
		if ((next = strtok(NULL, WHITE_SPACE)) == NULL
		    || *next == '#')
			return start;
	}

	/* too many tokens, indicate none at all */
	return 0;
}

int process_input(char *inbuf);

/* returns 1 if should exit, 0 otherwise (just end of file) */
int read_command_file(FILE * fptr)
{
	char buffer[BUFSIZE];

	while (fgets(buffer, BUFSIZE, fptr) != NULL) {
		if (process_input(buffer))
			return 1;
	}
	return 0;
}

/* returns 1 if should exit, 0 otherwise (error or ok) */
int process_input(char *inbuf)
{
	int argc, i;
	char text_line[4096];
	char *argv[10];
	char *start;
	FILE *fptr;

	/*****
	printf("process_input: %s", inbuf);
	*****/

	argv[0] = "iscsi_manage";
	argv[1] = "target";

	/* first, tokenize the inbuf into argv */
	if ((argc = tokenize(inbuf, argv, 2, 10)) <= 2) {
		/* either no tokens or too many tokens */
		return 0;
	}

	printf("%d tokens\n", argc);
	for (i = 0; i < argc; i++)
		printf("%8d: %s\n", i, argv[i]);

	if (argc == 3) {
		if (strcmp(argv[2], "quit") == 0)
			return 1;
		if (strcmp(argv[2], "exit") == 0)
			return 1;
	}

	if (argc == 4 && strcmp(argv[2], "include") == 0) {
		if ((fptr = fopen(argv[3], "r")) == NULL) {
			perror(argv[3]);
			return 0;
		} else {
			i = read_command_file(fptr);
			fclose(fptr);
			return i;
		}
	}

	/* next, scan the tokens into proper line format */
	if (param_scan(argc, argv, 0, text_line))
		return 0;

	printf("text_line: %s\n", text_line);

	/* next, get the target to act on the command line */
	if (iscsi_proc_info(text_line, &start, 0, strlen(text_line), 0, 1))
		return 0;

	return 0;
}


int main(void)
{
	int i, err;
	struct task_struct *current;

	/* first, set up this main thread */
	printf("unh_main_scsi_target Starting thread id %#08lx\n",
	       pthread_self());
	fflush(stdout);

	/* second create global key for the private-to-each thread "current" */
	if ((err = pthread_key_create(&current_key, free_current)) != 0) {
		fprintf(stderr, "pthread_key_create current_key: %s\n",
			strerror(err));
		exit(EXIT_FAILURE);
	}

	/* third give this thread its private copy of "current" and init it */
	unh_daemonize("unh_main_scsi_target");

	/* fourth, initialize the SCSI target module */
	printf("unh_main_scsi_target: calling "
	       "module_init_scsi_target_init_module\n");
	if ((i = module_init_scsi_target_init_module())) {
		printf
		    ("unh_main_scsi_target: 1st initialization result = %d\n",
		     i);
	}
	fflush(stdout);
	sched_yield();

	/* fifth, initialize the iSCSI target module */
	printf("unh_main_scsi_target: calling "
	       "module_init_iscsi_target_module_init\n");
	if ((i = module_init_iscsi_target_module_init())) {
		printf
		    ("unh_main_scsi_target: 2nd initialization result = %d\n",
		     i);
	}
	fflush(stdout);
	sched_yield();

	/* sixth, go into loop (until read EOF) to allow target to function */
	read_command_file(stdin);

	/* seventh, clean up the iSCSI target module */
	printf("unh_main_scsi_target: calling "
	       "module_exit_iscsi_target_module_cleanup\n");
	fflush(stdout);
	module_exit_iscsi_target_module_cleanup();
	fflush(stdout);
	sched_yield();

	/* eighth, clean up the SCSI target module */
	printf("unh_main_scsi_target: calling "
	       "module_exit_scsi_target_cleanup_module\n");
	fflush(stdout);
	module_exit_scsi_target_cleanup_module();
	fflush(stdout);
	sched_yield();

	/* ninth, free this thread's private "current" structure */
	free_current(current);
	if ((err = pthread_setspecific(current_key, current)) != 0) {
		fprintf(stderr, "pthread_setspecific current_key: %s\n",
			strerror(err));
		exit(EXIT_FAILURE);
	}


	/* tenth, delete global key for the private-to-each thread "current" */
	if ((err = pthread_key_delete(current_key)) != 0) {
		fprintf(stderr, "pthread_key_delete current_key: %s\n",
			strerror(err));
	}

	/* eleventh, exit this thread (should be only one left running) */
	printf("unh_main_scsi_target Exiting thread id %#08lx\n",
	       pthread_self());
	fflush(stdout);
	pthread_exit((void *) (unsigned long) i);
}
